//
// Example Image processing program
// - take a cube of data and average along one axis.
//

/*

This example looks at a three dimensional chunk of data and averages
along one axis, producing a two dimensional chuck of data as
output. It is intended to be treated as an example image analysis
application.

Parellizing this operation is relatively straightforward. Simply break
the cube face perpendicular to the averaging axis into sections; the
volumes formed by sweeping each section along the averaging axis each
go to separate filters.

The simplest way to do this is to split the cube into 2D planes
parallel to the averaging axis; this is the approach I take.

It is probably not wise to split the cube into 1D slivers and pass
each sliver to a filter individually despite the fact that doing so
extracts more parallelism. Unless the depth of the cube along the
averaging axis is extreamly large, the overhead associated with having
that many filter objects would probably dominate the execution time.

*/

#include <fstream>
#include "DataCutter.h"
#include <assert.h>

const int edge = 4;
char OUTFILE[] = "avgcube.out";

/*
In this example, averaging is to be done along the z axis. 
*/

class data_cube {
	float data[edge*edge*edge];
public:
	data_cube()
	{
		//this pattern makes it possible to debug the result

		for (float *p=data; p < data+num3d(); p++)
			*p = 1;

		int e = num1d() - 1;

		(*this)(0,0,0) = 1;
		(*this)(0,0,e) = -1;

		(*this)(e,0,0) = 2;
		(*this)(e,0,e) = -2;

		(*this)(e,e,0) = 3;
		(*this)(e,e,e) = -3;

		(*this)(0,e,0) = 4;
		(*this)(0,e,e) = -4;
	}
	float operator()(int x, int y, int z) const
	{
		return data[z*num2d() + y*num1d() + x];
	}
	float& operator()(int x, int y, int z)
	{
		return data[z*num2d() + y*num1d() + x];
	}
	int num1d() const { return edge; }
	int num2d() const { return edge*edge; }
	int num3d() const { return edge*edge*edge; }
};

class data_plane {
	float data[edge*edge];
public:
	data_plane()
	{
		for (float *p=data; p < data+num2d(); p++)
			*p = rand();
	}
	float operator()(int x, int y) const
	{
		return data[y*num1d() + x];
	}
	float& operator()(int x, int y)
	{
		return data[y*num1d() + x];
	}
	int num1d() const { return edge; }
	int num2d() const { return edge*edge; }
};

ostream& operator<<(ostream &o, data_cube &c)
{
	for (int z=0; z < c.num1d(); z++) {
		for (int y=0; y < c.num1d(); y++) {
			for (int x=0; x < c.num1d(); x++)
				o << c(x,y,z) << '\t';
			o << endl;
		}
		o << "---\n";
	}
	return o;
}

ostream& operator<<(ostream &o, data_plane &p)
{
	for (int y=0; y < p.num1d(); y++) {
		for (int x=0; x < p.num1d(); x++)
			o << p(x,y) << '\t';
		o << endl;
	}
	return o;
}

data_cube cube;
data_plane plane;

//DATACUTTER HELPER CLASSES////////////////////////////////////////////////////

/*
These classes just serve to make the code look cleaner. They're also
simple enough that they'll be totally optimized away by the compiler.
(Except for the exceptions, of course.)
*/

class dc_error {
	DC_RTN_t err;
public:
	dc_error(DC_RTN_t e) : err(e) {}
	operator DC_RTN_t () { return err; }
};

class buffer {
	bool buf_owner;
	DC_Buffer_t *buf;
	void free()
	{
		if (buf != 0) {
			if (buf_owner)
				delete buf;
			else
				buf->consume();
		}
	}
	friend class outstream;
	DC_Buffer_t* under() { return buf; }
public:
	buffer(int count, int bytes) : buf_owner(true), buf(new DC_Buffer_t())
	{
		buf->New(count*bytes);
		buf->setSize(count*bytes);
	}
	buffer() : buf_owner(true), buf(0) {}
	~buffer() { free(); }
	void reset(int count, int bytes)
	{
		free();
		buf_owner = true;
		buf = new DC_Buffer_t();
		buf->New(count*bytes);
		buf->setSize(count*bytes);
	}
	void reset(DC_Buffer_t *b)
	{
		free();
		buf_owner = false;
		buf = b;
	}
	const void* p() const
	{
		return reinterpret_cast<const void*>(buf->getPtr());
	}
	void* p()
	{
		return reinterpret_cast<void*>(buf->getPtr());
	}
	unsigned int size() { return buf->getSize(); }
	bool valid() const { return buf != 0; }
};

class instream {
	DC_Filter_Base_t::arg_t &a;
	DC_PipeInStream_t *in;
public:
	instream(DC_Filter_Base_t::arg_t &arg, const char *str)
		: a(arg), in(a.insLookup(const_cast<char*>(str)))
	{
		if (in == 0) {
			cerr << a.sbFilterName << ": " << "instream lookup error\n";
			throw dc_error(DC_RTN_ABRT);
		}
	}
	bool operator>>(buffer &buf)
	{
		buf.reset(in->read());
		return buf.valid();
	}
};

class outstream {
	DC_Filter_Base_t::arg_t &a;
	DC_PipeOutStream_t *out;
public:
	outstream(DC_Filter_Base_t::arg_t &arg, const char *str)
		: a(arg), out(a.outsLookup(const_cast<char*>(str)))
	{
		if (out == 0) {
			cerr << a.sbFilterName << ": " << "outstream lookup error\n";
			throw dc_error(DC_RTN_ABRT);
		}
	}
	outstream& operator<<(buffer &buf)
	{
		if (!buf.valid())
			throw dc_error(DC_RTN_ABRT);

		int code = out->write(buf.under());
		if (code != DC_ERR_OK)
			cerr << a.sbFilterName << ": failed write to buffer\n";
		return *this;
	}
};

//FILTERS//////////////////////////////////////////////////////////////////////

class Producer : public DC_Filter_Base_t {
public:
	DC_RTN_t init(initarg_t &initarg)
	{
		return DC_RTN_OK;
	}
	DC_RTN_t process(arg_t &arg)
	{
		try {
			buffer buf(cube.num2d()+1, sizeof(float));
			outstream out(arg, "P-A");
			float *p = reinterpret_cast<float*>(buf.p());

			for (int y=0; y < cube.num1d(); y++) {
				//copy plane into buffer
				for (int z=0; z < cube.num1d(); z++)
					for (int x=0; x < cube.num1d(); x++)
						p[z*cube.num1d()+x] = cube(x,y,z);
				//last int in buffer is y value of that plane
				p[cube.num2d()] = y;
				//send buffer
				out << buf;
			}

			return DC_RTN_OK;
		} catch (dc_error e) {
			return e;
		}
	}  
	DC_RTN_t finalize(void)
	{
		return DC_RTN_OK;
	}
};

class Averager : public DC_Filter_Base_t {
public:
	DC_RTN_t init(initarg_t &initarg)
	{
		return DC_RTN_OK;
	}
	DC_RTN_t process(arg_t &arg)
	{
		try {
			buffer ibuf;
			buffer obuf(cube.num1d()+1, sizeof(float));
			instream in(arg, "P-A");
			outstream out(arg, "A-C");
			float *op = reinterpret_cast<float*>(obuf.p());

			while (in >> ibuf) {
				float *ip = reinterpret_cast<float*>(ibuf.p());
			
				//do averaging, place it in obuf
				for (float *f = op; f < op+cube.num1d(); f++)
					*f = 0.0;
				for (int z=0; z < cube.num1d(); z++)
					for (int x=0; x < cube.num1d(); x++)
						op[x] += ip[z*cube.num1d()+x];
				for (int x=0; x < cube.num1d(); x++)
					op[x] /= cube.num1d();
				op[cube.num1d()] = ip[cube.num2d()];
				//send obuf
				out << obuf;
			}
    
			return DC_RTN_OK;
		} catch (dc_error e) {
			return e;
		}
	}
	DC_RTN_t finalize(void)
	{
		return DC_RTN_OK; 
	}
};

class Consumer : public DC_Filter_Base_t {
	ofstream ofs;
public:
	DC_RTN_t init(initarg_t &initarg)
	{
		ofs.open(initarg.pwork->buf.getPtr());
		return DC_RTN_OK;
	}
	DC_RTN_t process(arg_t &arg)
	{
		try {
			buffer buf;
			instream in(arg, "A-C");
			
			while (in >> buf) {
				float *p = reinterpret_cast<float*>(buf.p());

				//mux slivers back together
				int y = static_cast<int>(p[buf.size()/sizeof(float) - 1]);
				for (unsigned int x=0; x < buf.size()/sizeof(float)-1; x++)
					plane(x,y) = p[x];
			}
    
			return DC_RTN_OK;
		} catch (dc_error e) {
			return e;
		}
	}  
	DC_RTN_t finalize(void)
	{
		ofs << cube << "========\n" << plane;
		ofs.close();
		return DC_RTN_OK;
	}
};

DC_Filter_Base_t *filter_factory(char *sbFilterName)
{
	if (strcmp(sbFilterName, "P") == 0) {
		return new Producer;
	} else if (strcmp(sbFilterName, "A") == 0) {
		return new Averager;
	} else if (strcmp(sbFilterName, "C") == 0) {
		return new Consumer;
	} else {
		cerr << "unknown filter: " << sbFilterName << endl;
	}
	return NULL;
}

//MAIN/////////////////////////////////////////////////////////////////////////

int main(int argc, char* argv[])
{
#ifdef _WIN32
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;
 
	wVersionRequested = MAKEWORD(2, 0);
 
	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 ) {
		cout << "Error, Winsock could not be initialized.";
		return 0;
	}  
#endif /* WIN32 */

	DC_FilterService_t DC;
	if (DC.init("example-avgcube", &filter_factory, &argc, &argv))
		exit(1);

	// Check whether this process is a remote copy for running filter instances
	if (DC.isRemoteProcess()) {
		// Run any filter instances as directed by console process
		DC.RemoteProcess();
		return 0;
	}

	// This is the console process

	DC_Work_t work; //static work, and no client-server front end
	DC_FilterLayout_t layout;
	DC_Placement_t placement;
	DC_FilterInstance_t* instance;

	// Describe set of filters and streams to execute
	layout.Add("P", NULL, "P-A");
	layout.Add("A", "P-A", "A-C");
	layout.Add("C", "A-C", NULL);

	// Determine where to place filters
	if (DC.PlacementPlanning(layout, work, placement) != 0) { 
		// Choose our own placement instead
		placement.Add("P", "<one_per_node>");
		placement.Add("A", "<one_per_node>", cube.num1d());
		placement.Add("C", "<one_per_node>");    
	}

	// Get a filter instance (existing or new) for executing work
	int status = DC.NewFilterInstance(layout, work, placement, instance);
	if (status != DC_ERR_OK)
		cerr << "failed GetFilterInstance: " << DC_strerror(status) << endl;
  
	// Append work to instance
	work.buf.Empty();
	work.buf.Append(OUTFILE, sizeof(OUTFILE));

	int wh = instance->AppendWork(work);
	if (wh < 0)
		cerr << "failed AppendWork: " << DC_strerror(wh) << endl;

	status = DC.WaitWork(wh);
	if (status == DC_ERR_InstanceFail)
		cerr << "failed WaitWork: " << DC_strerror(status) << endl;

	// Finish up
	status = DC.StopFilterInstance(instance);
	if (status != DC_ERR_OK)
		cerr << "failed StopFilterInstance: " << DC_strerror(status) << endl;

	cout << "finsihed writing to " << OUTFILE << endl;

	return 0;
}
