// ****************************************************************************
//  Name: example-ppfc.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 01/24/2002 : mdb - Created.
//
//  Desc: Sample filter application from manual.  A stand-alone application
//        that creates a single filter group instance that operates in detached
//        mode w.r.t. the console process.
// ****************************************************************************

#include "example-filters.h"

//
// filter object definitions
//

class P1 : public DC_Filter_Base_t {
  DC_Buffer_t buf;

public:  // DC_Filter_Base_t interface

  // when new "work" arrives as a result of the console process performing
  // an AppendWork() call, the init(), process(), and finalize() calls are
  // made for this "work".

  DC_RTN_t init(initarg_t &initarg) {
    buf.New(80);
    return DC_RTN_OK;
  };

  DC_RTN_t process(arg_t &arg) {
    cout << arg.sbFilterName << ": process() entry, work=\""
	 << arg.pwork->buf.getPtr() << "\"" << endl;

    // Get index of stream using name from layout.  Note: For this
    // simple filter layout, we could have just assumed the single
    // output stream is index 0.
    int i = arg.outsIndex("s1");
    if (i < 0) {
      cerr << arg.sbFilterName << ':' << DC_strerror(i) << endl;
      return DC_RTN_OK;
    }

    // Fill the pre-allocated buffer with "data".  This can be done in
    // many ways, but for this case, we use the built-in support for
    // writing strings to a buffer.
    buf.sprintf("** This is data from filter P1 for work %s",
		arg.pwork->buf.getPtr());
    buf.Append('\0');  // make a valid string

    // Write the buffer to the output stream.  In this case we use the
    // function "write", which makes a copy of the buffer as needed to
    // provide the expected semantics of the caller being able to write
    // to the buffer again after the write call returns.
    if (arg.outs[i].write(&buf) != DC_ERR_OK) {
      cout << arg.sbFilterName << ": failed write of buffer " << buf << endl;
    } else {
      cout << arg.sbFilterName << ": wrote buffer " << buf << endl;
    }

    // Since we are done processing, we exit the processing phase.  We
    // may be called again by the filtering service if more "work"
    // arrives.  We can indicate our desire to be eligible to process
    // future "work", by returning DC_RTN_EndOfWork.  If instead, we
    // wanted to force this filter to exit, we would return
    // DC_RTN_EndOfFilter.
    return DC_RTN_OK;
  };

  DC_RTN_t finalize(void) {
    return DC_RTN_OK;
  };
};


class P2 : public DC_Filter_Base_t {
  DC_Buffer_t buf;

public:  // DC_Filter_Base_t interface

  // when new "work" arrives as a result of the console process performing
  // an AppendWork() call, the init(), process(), and finalize() calls are
  // made for this "work".
  DC_RTN_t init(initarg_t &initarg) {
    buf.New(80);
    return DC_RTN_OK;
  };

  DC_RTN_t process(arg_t &arg) {
    cout << arg.sbFilterName << ": process() entry, work=\""
	 << arg.pwork->buf.getPtr() << "\"" << endl;

    // Get index of stream using name from layout.  Note: For this
    // simple filter layout, we could have just assumed the single
    // output stream is index 0.
    int i = arg.outsIndex("s2");
    if (i < 0) {
      cerr << arg.sbFilterName << ':' << DC_strerror(i) << endl;
      return DC_RTN_OK;
    }

    // Fill the pre-allocated buffer with "data".  This can be done in
    // many ways, but for this case, we use the built-in support for
    // writing strings to a buffer.
    buf.sprintf("** This is data from filter P2 for work %s",
		arg.pwork->buf.getPtr());
    buf.Append('\0');  // make a valid string

    // Write the buffer to the output stream.  In this case we use the
    // function "write", which makes a copy of the buffer as needed to
    // provide the expected semantics of the caller being able to write
    // to the buffer again after the write call returns.
    if (arg.outs[i].write(&buf) != DC_ERR_OK) {
      cout << arg.sbFilterName << ": failed write of buffer " << buf << endl;
    } else {
      cout << arg.sbFilterName << ": wrote buffer " << buf << endl;
    }

    // Since we are done processing, we exit the processing phase.  We
    // may be called again by the filtering service if more "work"
    // arrives.  We can indicate our desire to be eligible to process
    // future "work", by returning DC_RTN_EndOfWork.  If instead, we
    // wanted to force this filter to exit, we would return
    // DC_RTN_EndOfFilter.
    return DC_RTN_OK;
  };

  DC_RTN_t finalize(void) {
    return DC_RTN_OK;
  };
};


class F : public DC_Filter_Base_t {
public:  // DC_Filter_Base_t interface
  DC_RTN_t init(initarg_t &initarg) {
    return DC_RTN_OK;
  };

  DC_RTN_t process(arg_t &arg) {
    DC_Buffer_t *pbuf1, *pbuf2, bufOut;
    cout << arg.sbFilterName << ": process() entry, work=\""
	 << arg.pwork->buf.getPtr() << "\"" << endl;

    // Get indices of streams using name from layout.
    int i1 = arg.insIndex("s1");
    int i2 = arg.insIndex("s2");
    int i3 = arg.outsIndex("s3");
    if (i1 < 0) {
      cerr << arg.sbFilterName << ':' << DC_strerror(i1) << endl;
      return DC_RTN_OK;
    } else if (i2 < 0) {
      cerr << arg.sbFilterName << ':' << DC_strerror(i2) << endl;
      return DC_RTN_OK;
    } else if (i3 < 0) {
      cerr << arg.sbFilterName << ':' << DC_strerror(i3) << endl;
      return DC_RTN_OK;
    }

    // We are performing the following operation: Read one buffer from
    // each stream, combine them, and write the result to the output
    // stream When all buffers for the current unit-of-work (UOW) are
    // read from a stream, a NULL is returned.
    while (1) {
      if ((pbuf1 = arg.ins[i1].read()) == NULL) break;
      cout << arg.sbFilterName << ": read buffer from s1: " << *pbuf1 << endl;
      if ((pbuf2 = arg.ins[i2].read()) == NULL) {
	pbuf1->consume();
	break;
      }
      cout << arg.sbFilterName << ": read buffer from s2: " << *pbuf1 << endl;

      cout << arg.sbFilterName << ": processing buffers to produce output"
	   << endl;
      bufOut.New(pbuf1->getSize() + pbuf2->getSize() + 1);
      bufOut.Append(pbuf1);
      bufOut.Append('+');
      bufOut.Append(pbuf2);

      // we are done with the input buffers
      pbuf1->consume();
      pbuf2->consume();

      // write the output buffer
      if (arg.outs[i3].write(&bufOut) != DC_ERR_OK) {
	cout << arg.sbFilterName << ": failed write of buffer " << bufOut
	     << endl;
      } else {
	cout << arg.sbFilterName << ": wrote buffer " << bufOut << endl;
      }
    }
    return DC_RTN_OK;
  };

  DC_RTN_t finalize(void) {
    return DC_RTN_OK; 
  };
};


class C : public DC_Filter_Base_t {
public:  // DC_Filter_Base_t interface
  DC_RTN_t init(initarg_t &initarg) {
    return DC_RTN_OK;
  };

  DC_RTN_t process(arg_t &arg) {
    DC_Buffer_t *pbuf;
    cout << arg.sbFilterName << "::process() entry, work=\""
	 << arg.pwork->buf.getPtr() << "\"" << endl;

    // Get indices of streams using name from layout.
    int i = arg.insIndex("s3");
    if (i < 0) {
      cerr << arg.sbFilterName << ':' << DC_strerror(i) << endl;
      return DC_RTN_OK;
    }

    while (1) {
      if ((pbuf = arg.ins[i].read()) == NULL) break;

      cout << arg.sbFilterName << "::read " << *pbuf << endl;

      pbuf->consume();
    }
    //return arg.ins[i].isEndOfWork() ? DC_RTN_EndOfWork : DC_RTN_EndOfFilter;
    return DC_RTN_OK;
  };

  DC_RTN_t finalize(void) {
    return DC_RTN_OK;
  };
};


// this is a utility fcn used by the filtering service to create new
// application filter objects for a given name.  The name is specified
// in the layout and placement, and is mapped to a particular
// DC_Filter_Base_t derived object here.
DC_Filter_Base_t *filter_factory(char *sbFilterName) {
  cout << "app: creating a new object for filter " << sbFilterName << endl;
  if (strcmp(sbFilterName, "P1") == 0) {         return new P1;
  } else if (strcmp(sbFilterName, "P2") == 0) {  return new P2;
  } else if (strcmp(sbFilterName, "F") == 0) {   return new F;
  } else if (strcmp(sbFilterName, "C") == 0) {   return new C;
  } else {
    cerr << "app: unknown filter \"" << sbFilterName << "\" to create" << endl;
  }
  return NULL;
}


// this is still the entry point for both the console process (where
// the application is invoked), and for remote process used to start
// filters.
int main(int argc, char *argv[]) {

#ifdef _WIN32
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;
 
	wVersionRequested = MAKEWORD(2, 0);
 
	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 ) {
		cout << "Error, Winsock could not be initialized.";
		return 0;
	}  

#endif /* WIN32 */

  DC_FilterService_t DC;
  if (DC.init("example-ppfc", &filter_factory, &argc, &argv)) exit(1);

  // check if this process is a remote copy for running filter instances
  if (DC.isRemoteProcess()) {
    // run any filter instances as directed by console process
    DC.RemoteProcess();
    return 0;
  }
  // we are the console process

  DC_FilterLayout_t layout;
  DC_Work_t work;
  DC_Placement_t placement;
  DC_FilterInstance_t *pinstance;
  int ret;

  // describe set of filters and streams to use
  //
  // this can be done in one of two ways:
  // (1) use layout.ReadFile("app_layout.cfg")
  // (2) use layout.Add("filtername", "ins0name ins1name ...", "outs0name ...")
  //
#if 0
  layout.ReadFile("example-ppfc_layout.cfg");
#else
  layout.Add("P1", NULL, "s1");
  layout.Add("P2", NULL, "s2");
  layout.Add("F", "s1 s2", "s3");
  layout.Add("C", "s3", NULL);
#endif

  // determine where to place filters when they are run.  This is
  // simply a mapping of filter -> host
  if (DC.PlacementPlanning(layout, work, placement) != 0) {
    // automatic scheduler did not produce a placement, so we will
    // specify our own.  The hosts chosen will need to be running
    // an appd (started via "boot_util appd start host1 host2 ...").

    // we use a trick here for choosing the hosts: <one_per_node> is a
    // special value that means find the list of all appds that are
    // running, and assign filters to them in round-robin order.
    placement.Add("P1", "<one_per_node>");
    placement.Add("P2", "<one_per_node>");
    placement.Add("F", "<one_per_node>");
    placement.Add("C", "<one_per_node>");
  }


  // given the filter group layout and a desired placement, we can now
  // create a running instance that can be used to process work.
  if ((ret = DC.NewFilterInstance(layout, work, placement, pinstance))
      != DC_ERR_OK) {
    cerr << "app: failed NewFilterInstance() rtn=" << ret << endl
	 << DC_strerror(ret) << endl;
    exit(1);
  }

  // create some application-defined work to pass to the instance's filters
  int wh;
  work.buf.Empty();
  work.buf.sprintf("work1");
  work.buf.Append('\0');

  // append work to the instance
  if ((wh = pinstance->AppendWork(work)) < 0) {
    cerr << "app: failed AppendWork() work=" << work.buf
	 << " rtn=" << wh << endl
	 << DC_strerror(wh) << endl;
  }


  // wait for completion of work
  //
  // There are two ways to wait for work to be completed.
  // (1) wait for specific work to finish with "DC.WaitWork(wh)"
  // (2) wait for any work to finish with "wh = DC.WaitAnyWork()"
#if 1
  ret = DC.WaitWork(wh);
  cout << "app: wh=" << wh << " finished, ret=" << ret
       << endl;
#else
  // wait for any of the submitted work to finish
  int wh;
  while ((wh = DC.WaitAnyWork()) != DC_ERR_NoWorkFound) {
    if (wh < 0) {
      cerr << "app: failed WaitAnyWork() call" << DC_strerror(wh) << endl;
    } else {
      cout << "app: wh=" << wh << " finished, ret=" << DC.ExitStatus(wh)
	   << endl;
    }
  }
#endif

  return 0;
}


// ****************************************************************************
