//
// example-sort.cpp
//

#include <fstream>
#include <DataCutter.h>

const int setElements = 10;
const int numSets = 5;
const int filterInstances = 1;

// Used in C library's qsort routine
static int intcompare(const void *i, const void *j) {
  if (*(int*)i > *(int*)j)
    return 1;
  if (*(int*)i < *(int*)j)
    return -1;
  return 0;
}

//
// Filter object definitions
//
class Producer : public DC_Filter_Base_t {
public:
  int bufSize;
  DC_Buffer_t buf;
  
  DC_RTN_t init(initarg_t &initarg) {
    bufSize = setElements * sizeof(int);
    buf.New(bufSize);
    buf.setSize(bufSize);
    return DC_RTN_OK;
  }

  DC_RTN_t process(arg_t &arg) {
    int *p;

    p = (int*)buf.getPtr();

    // Get the output stream index
    int outId = arg.outsIndex("P-S");
    if (outId < 0) {
      cerr << arg.sbFilterName << ": " << DC_strerror(outId) << endl;
      return DC_RTN_ABRT;
    }

    // Create random numbers and write them to the stream
    srand((unsigned)p);
    for (int j = 0; j < numSets; ++j) {
      for (int i = 0; i < setElements; ++i) {
        p[i] = rand() % 100;
      }
      if (arg.outs[outId].write(&buf) != DC_ERR_OK) {
	cerr << arg.sbFilterName << ": failed write to buffer" << endl;
      }
    }

    // Indicate that this work cycle has ended
    return DC_RTN_OK;
  };
  
  DC_RTN_t finalize(void) {
    return DC_RTN_OK;
  };
};

// Sorter class
class Sorter : public DC_Filter_Base_t {
public:
  DC_RTN_t init(initarg_t &initarg) {
    return DC_RTN_OK;
  };

  DC_RTN_t process(arg_t &arg) {
    DC_Buffer_t *pbuf;

    // Get the input stream index
    int inId = arg.insIndex("P-S");
    if (inId < 0) {
      cerr << arg.sbFilterName << ": " << DC_strerror(inId) << endl;
      return DC_RTN_ABRT;
    }

    // Get the output stream index
    int outId = arg.outsIndex("S-C");
    if (outId < 0) {
      cerr << arg.sbFilterName << ": " << DC_strerror(outId) << endl;
      return DC_RTN_ABRT;
    }

    // Read, sort, and write out the number sets
    while ((pbuf = arg.ins[inId].read())) {
      qsort(pbuf->getPtr(), pbuf->getSize() / sizeof(int), sizeof(int), intcompare);
      if (arg.outs[outId].write(pbuf) != DC_ERR_OK) {
	cout << arg.sbFilterName << ": failed write on buffer"<< endl;
      }	
      pbuf->consume();
    }
    
    return DC_RTN_OK;
  };

  DC_RTN_t finalize(void) {
    return DC_RTN_OK; 
  };
};

// Consumer class
class Consumer : public DC_Filter_Base_t {
public:
  ofstream ofs;

  DC_RTN_t init(initarg_t &initarg) {
    char* outfile = initarg.pwork->buf.getPtr();
    ofs.open(outfile);
    return DC_RTN_OK;
  };
  
  DC_RTN_t process(arg_t &arg) {
    DC_Buffer_t *pbuf;

    // Get the stream index
    int inId = arg.insIndex("S-C");
    if (inId < 0) {
      cerr << arg.sbFilterName << ": " << DC_strerror(inId) << endl;
      return DC_RTN_OK;
    }

    // Read and output the sorted sets to a file
    while ((pbuf = arg.ins[inId].read())) {
      int *p = (int*)pbuf->getPtr();
      for (unsigned int i = 0; i < pbuf->getSize() / sizeof(int); ++i) {
	ofs << p[i] << ' ';
      }
      ofs << endl;
      pbuf->consume();
    }

    return DC_RTN_OK;
  };
  
  DC_RTN_t finalize(void) {
    ofs.close();
    return DC_RTN_OK;
  };
};

// Filter instantiation routine
DC_Filter_Base_t *filter_factory(char *sbFilterName) {
  if (strcmp(sbFilterName, "P") == 0) {
    return new Producer;
  } else if (strcmp(sbFilterName, "S") == 0) {
    return new Sorter;
  } else if (strcmp(sbFilterName, "C") == 0) {
    return new Consumer;
  } else {
    cerr << "unknown filter: " << sbFilterName << endl;
  }
  return NULL;
}

int main(int argc, char* argv[]) {
	
#ifdef _WIN32
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;
 
	wVersionRequested = MAKEWORD(2, 0);
 
	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 ) {
		cout << "Error, Winsock could not be initialized.";
		return 0;
	}  

#endif /* WIN32 */

  DC_FilterService_t DC;
  if (DC.init("example-sort", &filter_factory, &argc, &argv)) {
    exit(1);
  }

  // Check whether this process is a remote copy for running filter instances
  if (DC.isRemoteProcess()) {
    // Run any filter instances as directed by console process
    DC.RemoteProcess();
    return 0;
  }

  // This is the console process

  // Static work, and no client-server front end
  DC_Work_t work;

  // Describe set of filters and streams to execute
  DC_FilterLayout_t layout;
  layout.Add("P", NULL, "P-S");
  layout.Add("S", "P-S", "S-C");
  layout.Add("C", "S-C", NULL);

  // Determine where to place filters
  DC_Placement_t placement;
  if (DC.PlacementPlanning(layout, work, placement) != 0) { 
    // Choose our own placement instead
    placement.Add("P", "<one_per_node>");
    //placement.Add("S", "<one_per_node>", 2);
    placement.Add("S", "<one_per_node>");
    placement.Add("C", "<one_per_node>");    
  }

  // Get a filter instance (existing or new) for executing work
  int status;
  DC_FilterInstance_t* instance;

  if ((status = DC.NewFilterInstance(layout, work, placement, instance)) != DC_ERR_OK) {
    cerr << "failed GetFilterInstance: " << DC_strerror(status) << endl;
  }
  
  // Append work to instance
  int wh;

  for (int i = 1; i < argc; ++i) {
    work.buf.Empty();
	work.buf.Append(argv[i], (Types::uint)strlen(argv[i]) + 1);
    
    if ((wh = instance->AppendWork(work)) < 0) {
      cerr << "failed AppendWork: " << DC_strerror(wh) << endl;
    }
    status = DC.WaitWork(wh);
    if (status == DC_ERR_InstanceFail) {
      cerr << "failed WaitWork: " << DC_strerror(status) << endl;
      break;
    }
    cout << "finsihed writing to " << argv[i] << endl;
  }
  if ((status = DC.StopFilterInstance(instance)) != DC_ERR_OK) {
    cerr << "failed StopFilterInstance: " << DC_strerror(status) << endl;
  }
  return 0;
}

