// ****************************************************************************
//  Name: appd-client.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 04/09/2000 : mdb - Created.
//        11/08/2000 : mdb - Added waiting for an appd to appear.
//        04/13/2001 : mdb - Added ShutdownAll.
//
//  Desc: Client access routines for application execution daemon.
// ****************************************************************************

#include "lib/dc_standard.h"
#include "lib/dc_argv.h"
#include "lib/dc_configinfo.h"
#include "lib/dc_darray.h"
#include "lib/dc_socket.h"
#include "lib/dc_missing.h"
#include "lib/dc_memory.h"

#include "appd-client.h"
#include "dird-client.h"
#include "packet.h"

#if !defined(MAXHOSTNAMELEN)
#  define MAXHOSTNAMELEN 256
#endif


#if defined(DCDBG)
//#define DCDBG_APPD_PING
//#define DCDBG_APPD_PSTACK
//#define DCDBG_APPD_SHUTDOWN
#define DCDBG_APPD_FTP
#endif /* DCDBG */


/**
 * contact the dird and see if an appd is registered for a given host
 * rtn: port: <0=error, 0=no appd registered, >0=appd contact port
 */
int Appd_Find(char *sbAppdHost, bool fWait) {
  ConfigInfo cfg;
  char buf[MAXHOSTNAMELEN+32];

  snprintf(buf, sizeof(buf), "datacutter.appd.\"%s\"", sbAppdHost);
  if (Dird_Query(buf, cfg, NULL, 0, (fWait ? 1 : 0)) >= 0) {
    char *sb = cfg.Find(buf, "port");
    if (sb) return atoi(sb);
    return 0;  // no running appd found
  }
  return -1;  // dird not running
}

/**
 * contact the appd and see if alive
 * rtn: 0=success, !0=failure
 */
int Appd_Ping(char *sbAppdHost, int wAppdPort) {
  InetAddress  addrAppd;
  ClientSocket sockAppd;

  // connect to daemon
  if (sockAppd.Connect(sbAppdHost, wAppdPort, addrAppd) != 0) {
#if defined(DCDBG_APPD_PING)
    perror("Appd_Ping(): failed Connect()");
    cerr << addrAppd << endl;
#endif /* DCDBG_APPD_PING */
    return -1;
  }
#if defined(DCDBG_APPD_PING)
  cout << "Appd_Ping(): Connected to appd " << addrAppd << endl;
#endif /* DCDBG_APPD_PING */

  // build packet
  Packet pkt(TAG_APPD_PING, NULL);

  // send packet to daemon
  if (pkt.write(sockAppd) <= 0) {
    cout << "Appd_Ping(): Failed write of ping" << endl;
    sockAppd.close();
    return -2;
  }
#if defined(DCDBG_APPD_PING)
  cout << "Appd_Ping(): Sent ping to appd" << endl;
#endif /* DCDBG_APPD_PING */

  // wait for response
  pkt.setBuf(NULL);
  if (pkt.read(sockAppd) <= 0) {
    cout << "Appd_Ping(): Failed read of ack" << endl;
    sockAppd.close();
    return -3;
  }
  STestDelete(pkt.getPayload());  // throw away any payload
  if (pkt.getTag() != TAG_APPD_OK) {
    cout << "Appd_Ping(): Ack indicates failure ("
	 << pkt.getTag() << ")" << endl;
    sockAppd.close();
    return -4;
  }
#if defined(DCDBG_APPD_PING)
  cout << "Appd_Ping(): Recv ping ack -- appd is alive" << endl;
#endif /* DCDBG_APPD_PING */
  sockAppd.close();
  return 0;
}


/**
 * contact the appd and if alive, shut it down
 * rtn: 0=success, !0=failure
 */
int Appd_Shutdown(char *sbAppdHost, int wAppdPort) {
  InetAddress  addrAppd;
  ClientSocket sockAppd;

  // connect to daemon
  if (sockAppd.Connect(sbAppdHost, wAppdPort, addrAppd) != 0) {
    perror("Appd_Shutdown(): failed Connect()");
    cerr << addrAppd << endl;
    return -1;
  }
#if defined(DCDBG_APPD_SHUTDOWN)
  cout << "Appd_Shutdown(): Connected to appd " << addrAppd << endl;
#endif /* DCDBG_APPD_SHUTDOWN */

  // build packet
  Packet pkt(TAG_APPD_SHUTDOWN, NULL);

  // send packet to daemon
  if (pkt.write(sockAppd) <= 0) {
    cout << "Appd_Shutdown(): Failed write of pkt" << endl;
    sockAppd.close();
    return -2;
  }
#if defined(DCDBG_APPD_SHUTDOWN)
  cout << "Appd_Shutdown(): Sent shutdown to appd" << endl;
#endif /* DCDBG_APPD_SHUTDOWN */

  // wait for response
  pkt.setBuf(NULL);
  if (pkt.read(sockAppd) <= 0) {
    cout << "Appd_Shutdown(): Failed read of ack" << endl;
    sockAppd.close();
    return -3;
  }
  STestDelete(pkt.getPayload());  // throw away any payload
  if (pkt.getTag() != TAG_APPD_OK) {
    cout << "Appd_Shutdown(): Ack indicates failure ("
	 << pkt.getTag() << ")" << endl;
    sockAppd.close();
    return -4;
  }
  sockAppd.close();
  return 0;
}


/**
 * shutdown all appds we can find in dird
 * rtn: # of appds shutdown
 */
int Appd_ShutdownAll(void) {
  ConfigInfo cfg;
  int wRtn = 0;

  if (Dird_Query("datacutter.appd.*", cfg, NULL, 0, 0) >= 0) {
    // look through them all
    const char *sbSection = cfg.FirstSection();
    while (sbSection) {
      char *sbHost = cfg.Find(sbSection, "dns_name");
      int   wPort = atoi(cfg.Find(sbSection, "port"));
      cout << "stop appd on " << sbHost << ':' << wPort << endl;
      if (Appd_Shutdown(sbHost, wPort) == 0)
	++wRtn;
      
      sbSection = cfg.NextSection();
    }
  }
  return wRtn;
}

/**
 * request stack dump of remote processes, output to stdout
 */
int Appd_Pstack(char *sbAppdHost, int wAppdPort) {
  InetAddress  addrAppd;
  ClientSocket sockAppd;

  // connect to daemon
  if (sockAppd.Connect(sbAppdHost, wAppdPort, addrAppd) != 0) {
#if defined(DCDBG_APPD_PSTACK)
    perror("Appd_Pstack(): failed Connect()");
    cerr << addrAppd << endl;
#endif /* DCDBG_APPD_PSTACK */
    return -1;
  }
#if defined(DCDBG_APPD_PSTACK)
  cout << "Appd_Pstack(): Connected to appd " << addrAppd << endl;
#endif /* DCDBG_APPD_PSTACK */

  // build packet
  Packet pkt(TAG_APPD_PSTACK, NULL);

  // send packet to daemon
  if (pkt.write(sockAppd) <= 0) {
    cout << "Appd_Pstack(): Failed write of ping" << endl;
    sockAppd.close();
    return -2;
  }
#if defined(DCDBG_APPD_PSTACK)
  cout << "Appd_Pstack(): Sent ping to appd" << endl;
#endif /* DCDBG_APPD_PSTACK */

  // wait for response
  pkt.setBuf(NULL);
  if (pkt.read(sockAppd) <= 0) {
    cout << "Appd_Pstack(): Failed read of ack" << endl;
    sockAppd.close();
    return -3;
  }
  cout << "START \'" << sbAppdHost << "\' remote process dump ..."
       << endl
       << (pkt.getPayloadPtr() ? pkt.getPayloadPtr() : "")
       << "START \'" << sbAppdHost << "\' remote process dump."
       << endl;
  STestDelete(pkt.getPayload());  // throw away any payload
  if (pkt.getTag() != TAG_APPD_OK) {
    cout << "Appd_Pstack(): Ack indicates failure ("
	 << pkt.getTag() << ")" << endl;
    sockAppd.close();
    return -4;
  }
#if defined(DCDBG_APPD_PSTACK)
  cout << "Appd_Pstack(): Recv ping ack -- appd is alive" << endl;
#endif /* DCDBG_APPD_PSTACK */
  sockAppd.close();
  return 0;
}

/**
 * negotiate and send a file to a remote appd
 */
int Appd_Ftp(char *sbAppdHost, int wAppdPort,
	     File &f, unsigned int wChecksum, MemoryBuf &mb) {
  InetAddress  addrAppd;
  ClientSocket sockAppd;
  struct stat stat;

  if (!f.isopen() && f.open() == -1) {
    perror("Appd_Ftp(): failed open of file");
    return -1;
  }
  f.stat(&stat);

  // connect to daemon
  if (sockAppd.Connect(sbAppdHost, wAppdPort, addrAppd) != 0) {
    perror("Appd_Ftp(): failed Connect()");
    cerr << addrAppd << endl;
    return -1;
  }
#if defined(DCDBG_APPD_FTP)
  cout << "Appd_Ftp(): Connected to appd " << addrAppd << endl;
#endif /* DCDBG_APPD_FTP */

  // build packet
  mb.Empty();
  mb.sprintf("file=%s\n", f.get());
  mb.sprintf("mode=%u\n", stat.st_mode);
  mb.sprintf("size=%ld\n", stat.st_size);
  mb.sprintf("checksum=%u\n", wChecksum);
  Packet pkt(TAG_APPD_FTP, &mb);
  pkt.setExtra(0);  // this is the query

  // send packet to daemon
  if (pkt.write(sockAppd) <= 0) {
    cout << "Appd_Ftp(): Failed write of pkt" << endl;
    sockAppd.close();
    return -2;
  }
#if defined(DCDBG_APPD_FTP)
  cout << "Appd_Ftp(): Sent ftp(0) to appd" << endl;
#endif /* DCDBG_APPD_FTP */

  // wait for response
  pkt.setBuf(NULL);
  if (pkt.read(sockAppd) <= 0) {
    cout << "Appd_Ftp(): Failed read of ack" << endl;
    sockAppd.close();
    return -3;
  }
  if (pkt.getTag() != TAG_APPD_OK) {
    ConfigInfo cfgTmp;
    cfgTmp.ReadBuf(pkt.getPayloadPtr(),
		   pkt.getSizePayload());    
    cout << " - failed for appd on " << sbAppdHost << ':' << wAppdPort
	 << " - (" << cfgTmp["status"] << ") " << cfgTmp["msg"] << endl;
    sockAppd.close();
    return -4;
  }
  STestDelete(pkt.getPayload());  // throw away any payload

  if (pkt.getExtra() == 0) return 0;  // already has the file

  // send the file data blocks
  pkt.setBuf(&mb);
  pkt.setTag(TAG_APPD_FTP);
  pkt.setExtra(1);  // file data buffers
  if (!f.isopen() && f.open() < 0) {
    cout << "Appd_Ftp(): Failed open of file to send" << endl;
    sockAppd.close();
    return -5;
  }
  while (f.read(mb) > 0) {
    if (pkt.write(sockAppd) <= 0) {
      cout << "Appd_Ftp(): Failed write of ftp data pkt" << endl;
      sockAppd.close();
      return -2;
    }
  }
  pkt.setBuf(NULL);
  pkt.setExtra(2);  // done
  if (pkt.write(sockAppd) <= 0) {
    cout << "Appd_Ftp(): Failed write of ftp done pkt" << endl;
    sockAppd.close();
    return -2;
  }

  sockAppd.close();
  return 0;  
}


// ****************************************************************************
