// ****************************************************************************
//  Name: appd.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 08/05/1999 : mdb - Created.
//        08/18/1999 : mdb - Added Global.SL and packet handling.
//        09/12/1999 : mdb - Now use dird-client routine instead of manual.
//        09/13/1999 : mdb - Ctrl+C first kills any EXEC'ed processes, and the
//                           next one kills the appd process.
//        10/25/1999 : mdb - Added -host <hostname> option to 
//        11/03/1999 : mdb - Added -exit flag
//        11/13/1999 : mdb - Detects console socket going away and cleans up
//                           EXEC'ed children.
//        01/10/2000 : mdb - Added -int_host <hostname> option that causes a
//                           new int_name entry to be added to the dird entry.
//        04/08/2000 : mdb - Converted to use new ServerLoop_Base.
//        08/17/2000 : mdb - Fixed dangling ptr problem writing DONE pkt.
//        09/16/2000 : mdb - Added optional setting of DISPLAY envvar.
//        09/22/2000 : mdb - Removed unused -int_host option.
//        11/02/2000 : mdb - Set file handles to max for children.
//        03/04/2001 : mdb - Supports "setenv" option for EXEC.
//        05/07/2001 : mdb - Added an appd.cfg file, and translation of EXEC
//                           binary names based on [appd]binary_dir rules.
//        07/19/2001 : mdb - Added daemon capability (-daemon).
//        09/20/2001 : mdb - Fixed bug in binary_dir handling.
//        12/06/2001 : mdb - Check for output handler if (-outputhandler)
//        12/12/2001 : mdb - Added xauth cookie handling for daemon mode
//
//  Desc: DataCutter execution environment - remote node control daemon,
//        responsible for executing remote processes on a remote host.
// ****************************************************************************

#ifdef _WIN32
#include <windows.h>
#include <process.h>
#include "win32.h"
#else
#include <sys/wait.h>
#include <sys/resource.h>
#include <unistd.h>
#endif

#include "lib/dc_standard.h"
#include "lib/dc_configinfo.h"
#include "lib/dc_socket.h"
#include "lib/dc_exec.h"
#include "lib/dc_str.h"
#include "lib/dc_regex.h"
#include "lib/dc_file.h"
#ifndef _WIN32
#include "lib/dc_debug.h"
#endif

#include "appd.h"
#include "dird-client.h"
#include "server-loop-base.h"
#include "packet.h"
#include "util.h"

#if !defined(MAXHOSTNAMELEN)
#  define MAXHOSTNAMELEN 256
#endif

#define DETAILED_SIGCHLD_HANDLER

#if defined(DCDBG)
//#define DCDBG_DISPLAYSTATUS
#define DCDBG_FTP
#endif /* DCDBG */

typedef struct {
  pid_t pid;
  char *sbBinary;
  bool  fDone;
  int   wStatus;
  int   wKillAttempts;

  int   iClient;  // index of SL.rgClients[] that EXEC'ed this child
  BaseSocket *psock;  // open socket to process that EXEC'ed this child

#ifdef _WIN32
  HANDLE processHandle;
#endif

} Child_t;

#if defined(HAVE_INSTANTIATIONS)
// instantiate template classes we use
template class DArray<Child_t*>;
#endif /* HAVE_INSTANTIATIONS */


struct Global_s {
  char *sbVersion;
  Argv       a;
  ConfigInfo cfg;

  InetAddress addrLocal;
  ServerSocket sockLocal;

  bool fDaemon;  // running detached (-deamon) or in fg (maybe in xterm)
  bool fOutputHandler;
  bool fVerbose;

  char *sbDirdHost;
  int  wDirdPort;
  char sbDirdSection[MAXHOSTNAMELEN+17];
  char *sbBinaryDir;
#if 0
  bool fSetX11Display;
#endif
  bool fXauthCookieCached;

  bool fExitAfterExec;  // only handle a single exec
  bool fKillChildLostConnection;
  DArray<Child_t*> daChild;

  int fdSigPipeWrite;
} Global;

char sbBufTmp[1024];

char bufPacketOut[KB_32];
MemoryBuf mbPacketOut(bufPacketOut, KB_32, 0, false);

extern char **environ;

#ifdef _WIN32

BOOL g_win32Stop;
CRITICAL_SECTION g_cs;

int ErrorMessage(DWORD dwError)
{
	LPVOID lpMsgBuf;

    if(FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM, NULL,
            dwError, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),(LPTSTR) &lpMsgBuf,0,NULL ))
	{
		MessageBox(GetForegroundWindow(), (LPCTSTR) lpMsgBuf, TEXT("Error"), MB_OK);
	}

	return -1;
}

#endif

// return: 0=child killed, 1=no child to kill, 2=cannot kill child
int KillChild(Child_t *pchild) {
  // ASSUME: pchild->wKillAttempts is initialized correctly
  if (pchild->pid == -1) return 1;
  ++pchild->wKillAttempts;
  if (pchild->wKillAttempts >= 3) {
    // we give up here, since the process is not responding to kill
    return 2;
  }

#ifndef _WIN32
  if (kill(pchild->pid, SIGKILL) == -1) {
    cout << "KillChild: Error kill child pid " << pchild->pid << ' ';
    switch (errno) {
    case EINVAL: { cout << "(EINVAL)" << endl; break; }
    case EPERM:  { cout << "(EPERM)" << endl;  break; }
    case ESRCH:  { cout << "(ESRCH)" << endl;  break; }
    case EACCES: { cout << "(EACCES)" << endl;  break; }
    default:     { cout << '(' << errno << ')' << endl; }
    }
    return 2;
  } 
  else {
    cout << "KillChild: Killing child pid " << pchild->pid << endl;
  }

#endif

  return 0;
}


// display all currently running children
int DisplayStatus(void) {
  Child_t *pchild;
  char sbBuf[80];
  int nChild = 0;

  for (Types::uint j=0; j<Global.daChild.NumElem(); j++) {
    pchild = Global.daChild[j];
    if (pchild->pid == -1) continue;

    ++nChild;
    snprintf(sbBuf, sizeof(sbBuf), 
#if defined(LINUX)
	    "[%d] %d %c %d %d %d %d\n",
#else /* LINUX */
	    "[%d] %ld %c %d %d %d %d\n",
#endif /* LINUX */
	    j, pchild->pid, pchild->fDone ? 't' : 'f', pchild->wStatus,
	    pchild->wKillAttempts, pchild->iClient, pchild->psock->getFd());
    cout << sbBuf;
  }
  cout << nChild << " child processes running" << endl;
  return 0;
}


int CloseConsoleConnection(int iClient) {
  Child_t *pchild;
  bool fFirst = true;

  for (Types::uint j=0; j<Global.daChild.NumElem(); j++) {
    pchild = Global.daChild[j];
    if (pchild->pid != -1 && pchild->iClient == iClient) {
      if (fFirst) {
	fFirst = false;
	cout << "CloseConsoleConnection(" << iClient
	     << "): killing all EXECs for this client..." << endl;
      }
      if (Global.fKillChildLostConnection) KillChild(pchild);
    }
  }
  return 0;
}


int RegisterWithDird(void) {
  // build packet
  MemoryBuf mb(MAXHOSTNAMELEN*2 + 256);
  snprintf(Global.sbDirdSection, sizeof(Global.sbDirdSection),
	  "datacutter.appd.\"%s\"", Global.addrLocal.getHostName());
  mb.sprintf("[%s]\n", Global.sbDirdSection);
  mb.sprintf("pid=%ld\n", static_cast(long int,getpid()));
  mb.sprintf("dns_name=%s\n", Global.addrLocal.getHostName());
  mb.sprintf("ip_addr=%s\n", Global.addrLocal.getHostIpAddress());
  mb.sprintf("port=%d\n", Global.addrLocal.getPort());

  if (Dird_Add(mb, Global.sbDirdHost, Global.wDirdPort) != 0) {
    cout << "RegisterWithDird(): Failed write of registration info" << endl;
    return -2;
  }
  cout << "Registered with dird" << endl;
  return 0;
}
int UnRegisterWithDird(void) {
  // build packet
  MemoryBuf mb((int)strlen(Global.sbDirdSection)+10);
  mb.sprintf("%s", Global.sbDirdSection);

  if (Dird_Remove(mb, Global.sbDirdHost, Global.wDirdPort) != 0) {
    cout << "UnRegisterWithDird(): Failed write of removal command" << endl;
    return -2;
  }
  cout << "UnRegistered with dird" << endl;
  return 0;
}


// here we are given the pathname used by the console process to start
// a binary, and we translate this into the appropriate binary for the
// current machine as specified by the binary_dir rule for the current
// host.
//
// dir patterns can contain:
//     %F - filename of binary
//     %P - full pathname executed on console
//
// TODO: combine this with the InitBinaryDir() fcn, and look through
// all matching possibilities, including the console name until we
// find a file we can read and execute.
// 
char *BinaryDirTranslate_Alloc(char *sbBinary) {
  if (!Global.sbBinaryDir) return StrDup(sbBinary);  // %P
  // we replace %F and/or %P patterns in the binary dir
  Str str(Global.sbBinaryDir);
  str.Replace("%P", sbBinary);

  // find start of filename
  char *p = sbBinary + strlen(sbBinary) - 1;
  for ( ; p != sbBinary; --p) {
    if (*p == '/') {
      ++p;
      break;
    }
  }
  str.Replace("%F", p);
  //cout << "TRANSLATE: after= " << str << endl;
  return str.Dup();
}


class AppdServerLoop : public ServerLoop_Base {
protected:
  File fileFtp;  // ASSUME: only one ftp operation at a time

public:
  int ClientProcess(int i) {
    int wTag, wExtra;
    Packet packetIn(NULL);
    Packet packetOut(&mbPacketOut);
    ConfigInfo cfgTmp;
    Types::uint j;
    Child_t *pchild;

    if (packetIn.read(*(rgClients[i].psock)) <= 0) {
#if defined(DCDBG_DISPLAYSTATUS)
      cout << "Status before lost a console connection (fd="
	   << rgClients[i].psock->getFd() << "):\n"
	   << DisplayStatus() << endl;
#endif /* DCDBG_DISPLAYSTATUS */
      cout << "closed from other side fd=" << rgClients[i].psock->getFd() << endl;
      CloseConsoleConnection(i);
#if defined(DCDBG_DISPLAYSTATUS)
      cout << "Status after lost a console connection:\n"
	   << DisplayStatus() << endl;
#endif /* DCDBG_DISPLAYSTATUS */
      return SLE_CB_CLOSECONN;  // some error reading the client socket
    }
    cout << "read pkt from fd=" << rgClients[i].psock->getFd() << endl;

    //char *sbPayload = bufPacketIn;
    //char *sbPayload = packetIn.getPayload()->getPtr();

    wTag = packetIn.getTag();
    wExtra = packetIn.getExtra();
    switch (wTag) {
    case TAG_APPD_PING:
      // send the ack
      mbPacketOut.Empty();
      packetOut.setTag(TAG_APPD_OK);
      if (packetOut.write(*(rgClients[i].psock)) <= 0) {
	return SLE_CB_CLOSECONN;  // some error writing the client socket
      }
      break;

    case TAG_APPD_SHUTDOWN:
      cout << "SHUTDOWN ..." << endl;
      // first check for any child processes, and pass to children
      for (j=0; j<Global.daChild.NumElem(); j++) {
	pchild = Global.daChild[j];

	// BUGBUG: This will cause a bunch of SIGCHLD signals to be generated
	// that will never be seen by a call to SigProcess().
	if (KillChild(pchild) == 2) {
	  cout << "SigProcess: Child cannot be killed - ignoring it." << endl;
	  continue;
	}
      }
      UnRegisterWithDird();  // do before sending ack to avoid boot_util race
      // send the ack
      mbPacketOut.Empty();
      packetOut.setTag(TAG_APPD_OK);
      packetOut.write(*(rgClients[i].psock));
      return SLE_CB_SHUTDOWN;
      break;

    case TAG_APPD_EXEC:
      cout << "EXEC ..." << endl;
      cfgTmp.ReadBuf(packetIn.getPayloadPtr(),
		     packetIn.getSizePayload());
      cout << cfgTmp;

#if defined(DCDBG_DISPLAYSTATUS)
      cout << "Status before EXEC:\n" << DisplayStatus() << endl;
#endif /* DCDBG_DISPLAYSTATUS */

      ExecuteApp(cfgTmp, i);
      
#if defined(DCDBG_DISPLAYSTATUS)
      cout << "Status after EXEC:\n" << DisplayStatus() << endl;
#endif /* DCDBG_DISPLAYSTATUS */
      break;

    case TAG_APPD_PSTACK:
      cout << "PSTACK ..." << endl;
      // fill the ack/response buffer with process stack dumps
      mbPacketOut.Empty();
      for (Types::uint j=0; j<Global.daChild.NumElem(); j++) {
	Child_t *pchild = Global.daChild.Get(j);
	if (pchild->pid == -1 || pchild->fDone) continue;

#ifndef _WIN32
	char *p;

	p = DBG_StackDump(mbPacketOut.getPtrFree(), mbPacketOut.getFree(),
			  true, pchild->pid, pchild->sbBinary);
	if (p) {  // update the new size of the buffer contents
	  mbPacketOut.setSize(mbPacketOut.getSize() +
			      (p-mbPacketOut.getPtrFree()));
	}

#endif
	  }
      mbPacketOut.Append('\0');
      packetOut.setTag(TAG_APPD_OK);
      if (packetOut.write(*(rgClients[i].psock)) <= 0) {
	return SLE_CB_CLOSECONN;  // some error writing the client socket
      }
      break;

    case TAG_APPD_FTP:
      cout << "FTP ..." << endl;
      if (wExtra == 0) {  // initial request for ftp
	bool fRequestFile = false;
	int   wErr = 0;
	char *sbErr = NULL;

	cfgTmp.ReadBuf(packetIn.getPayloadPtr(),
		       packetIn.getSizePayload());
#if defined(DCDBG_FTP)
	cout << "debug: ftp: initial request ..." << endl << cfgTmp;
#endif /* DCDBG_FTP */

	// check for file and get checksum
	fileFtp.set(cfgTmp["file"]);
	int mode = atoi(cfgTmp["mode"]);
	if (fileFtp.open(O_RDONLY) == -1) {  // (not found, bad perm?)
	  fRequestFile = true;
	  if (fileFtp.creat(mode) == -1) {
	    wErr = 1;  // cannot create file
	    sbErr = strerror(errno);
	  }
	} else {  // found file
	  // compare file size
	  off_t wFtpSize = atol(cfgTmp["size"]);
	  struct stat stat;
	  fileFtp.stat(&stat);
	  if (wFtpSize != stat.st_size) {
	    fRequestFile = true;
	  }
#if defined(DCDBG_FTP)
	  cout << "debug: ftp: found file (size=" << stat.st_size
	       << "), ftp file size=" << wFtpSize << endl;
#endif /* DCDBG_FTP */

	  // compare file checksum
	  unsigned int wFtpSum = static_cast(unsigned int,
					     atoi(cfgTmp["checksum"]));
	  unsigned int wFileSum = fileFtp.checksum();
	  if (wFtpSum != wFileSum) {
	    // now check if we can write to the file
	    fileFtp.close();
	    if (fileFtp.open(O_WRONLY|O_TRUNC) == -1) {
	      fileFtp.unlink();
	      if (fileFtp.creat(mode) == -1) {
		// we cannot write to the existing file
		wErr = 1;
		sbErr = strerror(errno);
	      }
	    }
	    fRequestFile = true;
	  }
#if defined(DCDBG_FTP)
	  cout << "debug: ftp: found file (sum=" << wFileSum
	       << "), ftp file sum=" << wFtpSum << endl;
#endif /* DCDBG_FTP */
	}
	// send back ack
	mbPacketOut.Empty();
	if (wErr) {
	  packetOut.setTag(TAG_APPD_ERROR);
	  mbPacketOut.sprintf("status=%d\n", wErr);
	  if (sbErr) mbPacketOut.sprintf("msg=%s\n", sbErr);
	  fileFtp.close();
	} else {
	  packetOut.setTag(TAG_APPD_OK);
	  if (fRequestFile) {
	    packetOut.setExtra(1);
	  } else {
	    packetOut.setExtra(0);
	    fileFtp.close();
	  }
	}
#if defined(DCDBG_FTP)
	cout << "debug: ftp: sending ack ("
	     << "wErr=" << wErr << ", fRequestFile=" << fRequestFile
	     << ')' << endl;
#endif /* DCDBG_FTP */
	if (packetOut.write(*(rgClients[i].psock)) <= 0) {
	  fileFtp.unlink();
	  return SLE_CB_CLOSECONN;  // some error writing the client socket
	}
      } else if (wExtra == 1) {  // transfer file blocks
	fileFtp.write(*(packetIn.getPayload()));
#if defined(DCDBG_FTP)
	cout << "debug: ftp: wrote " << packetIn.getSizePayload()
	     << " bytes to file \"" << fileFtp.get() << "\"" << endl;
#endif /* DCDBG_FTP */
      } else { 	// done with ftp
	fileFtp.close();
#if defined(DCDBG_FTP)
	cout << "debug: done with ftp of file \"" << fileFtp.get() << "\""
	     << endl;
#endif /* DCDBG_FTP */
      }
      break;

    default:
      cerr << ErrorHeader(wERROR)
	   << "ClientProcess(): Ignoring packet with unknown tag #"
	   << wTag << "." << endl
	   << packetIn << endl;
    }
    return SLE_CB_OK;
  };

  // BUGBUG: How to realize which socket SIGPIPE belongs to?  Maybe we don't
  // have to know, since select() will later return with activity, and will be
  // closed.

  int SigProcess(int pipeval) {  // called with signals blocked
  
#ifndef _WIN32

	Child_t *pchild;
    int ret;
  
    if (pipeval == SIGHUP || pipeval == SIGINT) {
      // first check for any child processes, and pass to children
      bool fAny = false;
      for (Types::uint j=0; j<Global.daChild.NumElem(); j++) {
	pchild = Global.daChild[j];

	ret = KillChild(pchild);
	if (ret == 2) {
	  cout << "SigProcess: Child cannot be killed - ignoring it." << endl;
	  continue;
	}
	if (ret == 0) fAny = true;
      }
      if (!fAny) {
	// if none, we exit
	cout << "SigProcess(): SLE_CB_SHUTDOWN returned due to";
	if (pipeval == SIGHUP) { cout << " SIGHUP" << endl; };
	if (pipeval == SIGINT) { cout << " SIGINT" << endl; };
	UnRegisterWithDird();
	return SLE_CB_SHUTDOWN;
      }

    } else if (pipeval == SIGCHLD) {
      int wStatus;
      int nChild = 0;
      cout << "SigProcess(): Found SIGCHLD has occurred" << endl;
      // check all pids
      for (Types::uint j=0; j<Global.daChild.NumElem(); j++) {
	Child_t *pchild = Global.daChild.Get(j);
	if (pchild->pid == -1 || pchild->fDone) continue;

	if (waitpid(pchild->pid, &wStatus, WNOHANG) != 0) {

	  // BUGBUG: This only makes sense if WIFEXITED()
	  pchild->wStatus = WEXITSTATUS(wStatus);

	  cout << "Process done " << pchild->pid << "(" << j << ")" << endl;

	  // send notification down socket we received EXEC on
	  MemoryBuf mb(128);
	  mb.sprintf("pid=%ld\n", static_cast(long int,pchild->pid));
	  mb.sprintf("status=%d\n", pchild->wStatus);
	  if (WIFEXITED(wStatus)) {
	    mb.sprintf("done=EXITED\n");  // normal termination
	  } else if (WIFSIGNALED(wStatus)) {
	    mb.sprintf("done=TERMINATED signal %d\n", WTERMSIG(wStatus));
	  }
#if !defined(AIX) && defined(WCOREDUMP)
	  else if (WCOREDUMP(wStatus)) {
	    mb.sprintf("done=COREDUMP\n");
	  }
#endif /* AIX */
	  Packet pkt(TAG_APPD_DONE, &mb);

	  // 09/15/1999 : mdb - Must be done regardless of pkt.write failure.
	  // get rid of the child record now
	  pchild->fDone = true;  // mark as done
	  pchild->pid = -1;      // mark for reuse

	  // has this connection been closed already?
	  int map = (int)getClientMapValue(pchild->iClient);
	  if (map) {
	    // this socket is not known to be closed
	    pkt.write(*pchild->psock); // ignore a write error
	  }
	  cout << "Sent DONE packet" << endl;
	} else {
	  ++nChild;
	}

	if (Global.fExitAfterExec) {
	  UnRegisterWithDird();
	  return SLE_CB_SHUTDOWN;  // we want to exit
	}
      }
      cout.flush();
      cerr.flush();
      cout << "appd: " << nChild << " DataCutter applications running" << endl;
    }
  
#if defined(DCDBG_DISPLAYSTATUS)
    cout << "Status after SigProcess():\n" << DisplayStatus() << endl;
#endif /* DCDBG_DISPLAYSTATUS */
    cout << "SigProcess(): SLE_CB_OK returned." << endl;

#endif /* _WIN32 */

    return SLE_CB_OK;
  
  };

  int MapProcess(int i, enum eMapReason e) {
    switch (e) {
    case ADD_CONN:
      //cout << "MapProcess(" << i << ",ADD_CONN)"<< endl;
      setClientMapValue(i, (void*)1);
      break;
    case DISABLE_CONN:
      break;
    case ENABLE_CONN:
      break;
    case CLOSE_CONN:
      //cout << "MapProcess(" << i << ",CLOSE_CONN)"<< endl;
      setClientMapValue(i, (void*)0);
      break;
    }

    return SLE_CB_OK;
  };

  // extra helper member fcns

  int ExecuteApp(ConfigInfo &cfgTmp, int iClient) {
    Argv a;
    Str str;
    Child_t *pchild;
    char *sb;
    Types::uint j;
    DArray<char*> *pda;
    bool fCloseInput;
    pid_t pid;

#ifndef	_WIN32
	Exec e;
#else
	// used in starting a	console	in win32 enviornment
	PROCESS_INFORMATION piProcInfo; 
	STARTUPINFO siStartInfo; 

	ZeroMemory( &piProcInfo, sizeof(PROCESS_INFORMATION) );
	ZeroMemory( &siStartInfo,	sizeof(STARTUPINFO)	);

	siStartInfo.cb = sizeof(STARTUPINFO);	
#endif

    // build argv for this command - may have to be more general
#if 1
    // will exec directly
    char *sbBinary = BinaryDirTranslate_Alloc(cfgTmp["cmd"]);
    a.Append(sbBinary);
    pda = cfgTmp.FindAll("arg");
    for (j=0; j<pda->NumElem(); ++j) {
      a.Append(pda->Get(j));
    }
    delete pda;
#else
    // use /bin/sh to exec
    a.Append("/bin/sh");
    a.Append("-c");
    Str strExec;
    char *sb;
    strExec.Cat("exec ");
    strExec.Cat(cfgTmp["cmd"]);

    pda = cfgTmp.FindAll("arg");
    for (Types::uint i=0; i<pda->NumElem(); ++i) {
      if (i>0) strExec.Cat(' ');
      strExec.Cat();
    }
    delete pda;
#endif

    fCloseInput = cfgTmp["debug"] ? false : true;

    // gen output filename
    sb = cfgTmp["outerr"];
    if (sb && strlen(sb) > 0) {
      snprintf(sbBufTmp, sizeof(sbBufTmp), "%s.%ld", sb, (long)getpid());
    } else {
      sbBufTmp[0] = '\0';
    }

#ifndef _WIN32
    // spawn child
    if ((pid = fork()) < 0) {
      // error
      perror("fork");
      return -1;
	} else if (pid == 0) {
	// child
      int fd;
      char bufTmp[80];

      // cmd exec process
      cout << "Child pid " << getpid();
      if (strlen(sbBufTmp)>0) {
	cout << ": output to '" << sbBufTmp << "'" << endl;
      } else {
	cout << endl;
      }
      if (fCloseInput) close(0);
      if (strlen(sbBufTmp) > 0 &&
	  ((fd = open(sbBufTmp, O_WRONLY|O_CREAT|O_TRUNC, 0644)) < 0 ||
	   close(1) || dup2(fd, 1) < 0 ||
	   close(2) || dup2(fd, 2) < 0 ||
	   close(fd))) {
	perror("cmd exec process stdout/stderr redirection failed");
	//exit(1);
      }

	   // close any of the client sockets still open
      CloseAllConn();
	
#endif

      // dump useful info into output stream before app is exec'ed

      // indicate host we are running on
      cout << "host: " << Global.addrLocal.getHostName() << endl;

      // setcwd
      sb = cfgTmp["cwd"];
      if (sb && strlen(sb) > 0) {
	str.Set(sb);
	str.ReplaceEnvVars();
	cout << "cwd:  " << str.Get() << endl;

#ifndef _WIN32

	e.Chdir(str.Get());
    
#endif
	  
	  }

#ifndef _WIN32

      // X11 remote display
      char *sbDisplay;
      sbDisplay = sb = cfgTmp["display"];
      if (sb && strlen(sb) > 0) {	  
	snprintf(bufTmp, sizeof(bufTmp), "DISPLAY=%s", sb);
	putenv(bufTmp);
	cout << "X11 DISPLAY: " << sb << endl;
      }
      char *sbXauthMethod = cfgTmp["xauth_method"];
      if (sbXauthMethod) {
	sb = cfgTmp["xauth_cookie"];
	if (sb && strlen(sb) > 0 && !Global.fXauthCookieCached) {
	  Argv a;
	  Exec e;
	  int fdStdin, fdStdout, fdStderr;
	  fdStdout = fdStderr = Exec::FD_PARENT;
	  if (strcmp(sbXauthMethod, "list-add") == 0) {
	    fdStdin = Exec::FD_PIPE;
	    a.Append("/bin/sh");
	    a.Append("-c");
	    a.Append("exec xauth -q -");
	    //a.Append("-q");
	    //a.Append("-");
	    //cout << a << endl;
	    pid_t pid = e.ForkExec(a, fdStdin, fdStdout, fdStderr);
	    if (write(fdStdin, "add ", 4) < 0 ||
		write(fdStdin, sbDisplay, strlen(sbDisplay)) < 0 ||
		write(fdStdin, " . ", 3) < 0 ||
		write(fdStdin, sb, strlen(sb)) < 0 ||
		write(fdStdin, "\n", 1) < 0) {
	      perror("write add remote cookie");
	    }
	    char *p = strchr(sbDisplay, ':');
	    if (p && p>sbDisplay) {
	      if (write(fdStdin, "add ", 4) < 0 ||
		  write(fdStdin, sbDisplay, p-sbDisplay) < 0 ||
		  write(fdStdin, "/unix", 5) < 0 ||
		  write(fdStdin, p, strlen(p)) < 0 ||
		  write(fdStdin, " . ", 3) < 0 ||
		  write(fdStdin, sb, strlen(sb)) < 0 ||
		  write(fdStdin, "\n", 1) < 0) {
		perror("write add /unix cookie");
	      }
	    }
	    close(fdStdin);
	    e.Wait(pid);
	    cout << "xauth list-add: installed xauth cookie" << endl;
	  } else if (strcmp(sbXauthMethod, "nextract-nmerge") == 0) {
	    fdStdin = Exec::FD_PIPE;
	    // use stdin to avoid cookie visible in cmd line
	    // BUGBUG: this still appears to be broken
            a.Append("/bin/sh");
            a.Append("-c");
	    a.Append("exec xauth nmerge -");
	    //a.Append("nmerge");
	    //a.Append("-");
	    //cout << a << endl;
	    pid_t pid = e.ForkExec(a, fdStdin, fdStdout, fdStderr);
	    if (write(fdStdin, sb, strlen(sb)) < 0) perror("write cookie");
	    if (write(fdStdin, "\n", 1) < 0) perror("write newline");
	    close(fdStdin);
	    e.Wait(pid);
	    cout << "xauth nextract-nmerge: installed xauth cookie" << endl;
	  }
	}
      }

      // setenv
      // TAHSIN 09/21/2002: DArray<char*> ... was commented out
      DArray<char*> *pda = cfgTmp.FindAll("setenv");
      for (Types::uint j=0; j<pda->NumElem(); ++j) {
	char *sb = pda->Get(j);
	putenv(sb);
	cout << "setenv: " << sb << endl;
      }
      delete pda;

      // set number of open files to max
      struct rlimit rl;
      if (getrlimit(RLIMIT_NOFILE, &rl) != 0) {
	cout << "nofiles: cannot get number of open files info" << endl;
      } else {
	rl.rlim_cur = rl.rlim_max;  // set to max
	if (setrlimit(RLIMIT_NOFILE, &rl) != 0) {
	  cout << "nofiles: cannot set number of open files to max" << endl;
	} else {
	  cout << "nofiles: set to " << rl.rlim_cur << endl;
	}
      }

      // exec command
      cout << a << endl;
      e.Execvp(a);  // never returns
      exit(0);
    }
    // appd process continues

#else

  // Launch things WIN32 style
	  // throw create process stuff	in here

	/* the path to the console app must be in the PATH enviornment variable */
	Str command;
	a.toString(command,false);
	if(!CreateProcess(NULL, (LPSTR)command.Get(), NULL, NULL, NULL, NULL/*CREATE_NEW_CONSOLE*/, 
					  NULL,	NULL, &siStartInfo,	&piProcInfo))
	{
		return ErrorMessage(GetLastError());
	}
	pid =	piProcInfo.dwProcessId;

#endif

	// save child info
    for (j=0; j<Global.daChild.NumElem(); j++) {
      pchild = Global.daChild[j];
      if (pchild->pid == -1) break;  // found an empty spot
    }
    if (j == Global.daChild.NumElem()) {
      j = Global.daChild.Append((pchild = new Child_t));  // no spots
    }
    pchild->pid = pid;
    pchild->sbBinary = sbBinary;
    pchild->fDone = 0;
    pchild->wStatus = -1;  // init
    pchild->wKillAttempts = 0;  // init
    pchild->iClient = iClient;
    pchild->psock = rgClients[iClient].psock;

#ifdef _WIN32
	pchild->processHandle = piProcInfo.hProcess;
#endif

    cout << "Started process " << pid << " (" << j << ") ";
    if (strlen(sbBufTmp)>0) {
      cout << ": output to '" << sbBufTmp << "'" << endl;
    } else {
      cout << endl;
    }

    return 0;
  };
};

#ifndef _WIN32

static void sig_int(int, siginfo_t *, void *) {
  //cout << ErrorHeader(wWARNING) << "sig_int(): Handler called ..." << endl;
  int n = SIGINT;
  if (Global.fdSigPipeWrite != -1)
    write(Global.fdSigPipeWrite, &n, sizeof(n));  // signal to select()
}
static void sig_hup(int, siginfo_t *, void *) {
  //cout << ErrorHeader(wWARNING) << "sig_hup(): Handler called ..." << endl;
  int n = SIGHUP;
  if (Global.fdSigPipeWrite != -1)
    write(Global.fdSigPipeWrite, &n, sizeof(n));  // signal to select()
}
static void sig_pipe(int, siginfo_t *, void *) {
  //cout << ErrorHeader(wWARNING) << "sig_pipe(): Handler called ..." << endl;
  // just return, we don't need to inform select explicitly
}
#if defined(DETAILED_SIGCHLD_HANDLER)
static void sig_chld(int, siginfo_t *psi, void *) {
#if 0 /* cannot call cout from sig handler */
  cout << ErrorHeader(wWARNING) << "sig_chld(): Handler called ..." << endl;
  switch (psi->si_code) {
  case CLD_EXITED: cout << "CLD_EXITED";
    break;
  case CLD_KILLED: cout << "CLD_KILLED";
    break;
  case CLD_DUMPED: cout << "CLD_DUMPED";
    break;
  case CLD_TRAPPED: cout << "CLD_TRAPPED";
    break;
  case CLD_STOPPED: cout << "CLD_STOPPED";
    break;
  case CLD_CONTINUED: cout << "CLD_CONTINUED";
    break;
  default: cout << "CLD_(" << psi->si_code << ")";
    break;
  }
  cout << ",psi->si_pid=" << psi->si_pid
       << ",psi->si_status=" << psi->si_status
       << endl;
#endif /* 0 */
  // BUGBUG: We need to record this extended status info somewhere for when we
  // send out the DONE pkt, so we can indicate if it exited or dumped, or
  // whatever.

  int n = SIGCHLD;
  if (Global.fdSigPipeWrite != -1)
    write(Global.fdSigPipeWrite, &n, sizeof(n));  // signal to select()
}
#else /* DETAILED_SIGCHLD_HANDLER */
static void sig_chld(int, siginfo_t *, void *) {
  //cout << ErrorHeader(wWARNING) << "sig_chld(): Handler called ..." << endl;
  int n = SIGCHLD;
  if (Global.fdSigPipeWrite != -1)
    write(Global.fdSigPipeWrite, &n, sizeof(n));  // signal to select()
}
#endif /* DETAILED_SIGCHLD_HANDLER */
#endif /* _WIN32 */

void cleanup(void) {
  // child structs
  for (Types::uint j=0; j<Global.daChild.NumElem(); j++) {
    Child_t *pchild = Global.daChild[j];
    if (pchild) {
      if (pchild->sbBinary) delete [] pchild->sbBinary;
      delete pchild;
    }
  }

  if (Global.sbBinaryDir) delete [] Global.sbBinaryDir;
}

// connect to the required socket and send the given init ack id to
// let some higher level entity know we have initialized.
int InitAck(char *sbInitAck) {
  InetAddress  addr;
  ClientSocket sock;
  char *sbHost, *p;
  int   wPort, wInitId;
  int   wRtn = 0;
  
  if (!sbInitAck) return 0;  // cmdline option not given
  sbHost = StrDup(sbInitAck);
  if (!(p = strchr(sbHost, ':'))) return 1;  // port must be specified
  *p = '\0';
  wPort = atoi(++p);
  if (!(p = strchr(p, ','))) return 1;  // init ack id must be specified
  *p = '\0';
  wInitId = atoi(++p);
  if (sock.Connect(sbHost, wPort, addr) != 0) wRtn = 2;
  Packet packetOut(TAG_BOOTUTIL_INITACK, NULL);
  packetOut.setExtra(wInitId);
  packetOut.write(sock);
  sock.Close();
  delete [] sbHost;
  return wRtn;
}

char *InitBinaryDir_Alloc(ConfigInfo &cfg, const char *sbHostName) {
  DArray<char*> *pdaRules = cfg.FindAll("appd", "binary_dir");
  DC_Regex re;
  char *sbRtn = NULL;
  //cout << "DEBUG: hostname=" << sbHostName << endl;
  for (unsigned int i=0; i<pdaRules->NumElem(); ++i) {
    //cout << "DEBUG: daRules[" << i << "] = >" << pdaRules->Get(i) << '<' << endl;
    Str str(pdaRules->Get(i));
    DArray<char*> *pdaTokens = str.TokenArray();
    if (pdaTokens->NumElem() != 2) {
      cerr << "InitBinaryDir(): invalid rule \""
           << pdaRules->Get(i) << "\"" << endl;
    } else if (re.setPattern(pdaTokens->Get(0)) != 0) {
      cerr << "InitBinaryDir(): invalid pattern \""
           << pdaTokens->Get(0) << "\"" << endl;
    } else if (re.match(sbHostName)) {
      sbRtn = StrDup(pdaTokens->Get(1));  // break at end of iteration
    }
    Str::FlushTokenArray(pdaTokens);
    delete pdaTokens;
    if (sbRtn) break;
  }
  if (pdaRules) delete pdaRules;
  if (sbRtn) {
    cout << "Binary translation: " << sbRtn << endl;
  }
  return sbRtn;
}

#ifdef _WIN32
void *RemoveFinishedFilters(void *arg)
{
	AppdServerLoop *appServerLoop = reinterpret_cast<AppdServerLoop *>(arg);
	
	for(;;)
	{
		EnterCriticalSection(&g_cs);
		if(g_win32Stop) 
		{
			pthread_exit(0);
		}
		LeaveCriticalSection(&g_cs);

		for (Types::uint j=0; j<Global.daChild.NumElem(); j++)
		{
			Child_t *pchild = Global.daChild[j];
			if (pchild->pid != -1)
			{
				DWORD ExitCode;
				if (GetExitCodeProcess(pchild->processHandle, &ExitCode))
				{
					if (ExitCode != STILL_ACTIVE)
					{
					
						// do stuff to send the DONE packet
						// send notification down socket we received EXEC on
						MemoryBuf mb(128);
						mb.sprintf("pid=%ld\n", static_cast(long int,pchild->pid));
						mb.sprintf("status=%d\n", pchild->wStatus);
						if (ExitCode == 0) {
							mb.sprintf("done=EXITED\n");  // normal termination
						} 
						else 
						{
							mb.sprintf("done=TERMINATED signal %d\n", ExitCode);
						}
						Packet pkt(TAG_APPD_DONE, &mb);

						// 09/15/1999 : mdb - Must be done regardless of pkt.write failure.
						// get rid of the child record now
						pchild->fDone = true;  // mark as done
						pchild->pid = -1;      // mark for reuse

						// has this connection been closed already?
					
						int map = (int)appServerLoop->getClientMapValue(pchild->iClient);
						if (map) {
							// this socket is not known to be closed
							cout << "DONE packet sent" << endl;
							pkt.write(*pchild->psock); // ignore a write error
						
						}
					}
				}
			}
			
			Sleep(250);
		}
	}
	
	return 0;
}

#endif

int Init(void) {
  char *sb;
  int fdOut = -1;

#if 0
  Global.fSetX11Display = true;
#endif
  Global.fXauthCookieCached = false;
  Global.sbBinaryDir = NULL;

  Global.fDaemon = false;
  if (Global.a.Find("-daemon")) {

#ifndef _WIN32
	pid_t pid;
    if ((pid = fork()) < 0) {
      perror("fork");
      return -1;
    } else if (pid != 0) {  // parent
      exit(0);      
    }
#endif
    // child
    Global.fDaemon = true;
  }
  // CLH - added verbose flag
  Global.fVerbose = Global.a.NextFind("-verbose");
  
  InetAddress  addrOutput;
  ClientSocket sockOutput;  // ok because we never return and call dtor

  // format: host:port
  Global.fOutputHandler = false;
  if ((sb = Global.a.NextFind("-outputhandler")) != NULL) {
    char *p = strchr(sb, ':');
    if (p) { *p = '\0'; ++p; }

    // connect to output handler
    if (sockOutput.Connect(sb, atoi(p), addrOutput) != 0) {
      perror("appd: failed Connect() to output handler");
      cerr << addrOutput << endl;
      return -1;
    }
    Global.fOutputHandler = true;
    fdOut = sockOutput.getFd();
    if (Global.fVerbose)
      cout << "appd: Connected to output handler (fd=" << fdOut << ") "
	   << addrOutput << endl;
  }

  if (Global.fDaemon) {
    if (!Global.fOutputHandler) {
      // redirect all stdout/stderr to local tmp files
      char sbOut[128];
      char *sbTmpDir = getenv("TMPDIR");
      if (!sbTmpDir) sbTmpDir = "/tmp";
      snprintf(sbOut, sizeof(sbOut), "%s/appd.out.%d",
	       sbTmpDir, static_cast(int,getpid()));
      if (Global.fVerbose)
	cout << "appd: entering daemon mode, output to " << sbOut << endl;

      if ((fdOut = open(sbOut, O_WRONLY|O_CREAT|O_TRUNC, 0666)) < 0) {
	cerr << ErrorHeader(wERROR) << "appd: Cannot redirect to '"
	     << sbOut << "'.\n"
	     << ErrorIndent() << strerror(errno) << endl;
      }
    }
#ifndef _WIN32
    setsid();
#endif

    close(0); // stdin;
  }

  bool fOk = false;
  if (Global.addrLocal.Set(0) == 0) {  // use any free port
    // for some machines, the default gethostname() will return a hostname
    // unable to be seen outside this machine -- allow a cmdline override.
    if ((sb = Global.a.NextFind("-host")) != NULL) {
      Global.addrLocal.setHostName(sb, true);
    }
    if (Global.sockLocal.CreateBindListen(Global.addrLocal) == 0) {
      fOk = true;
    }
  }
  if (!fOk) {
    cerr << ErrorHeader(wERROR) << "Init(): Failed local port setup." << endl;
    return -1;
  }

  // redirect output to /tmp file or output handler socket
  if (fdOut >= 0) {
    cout.flush();
    cerr.flush();
    dup2(fdOut, 1);  // stdout
    dup2(fdOut, 2);  // stderr
    close(fdOut);
  }

  if (Global.fVerbose) {
    if (Global.fOutputHandler) {
    // format: name,options
      cout << "@@OUTPUTHANLDER@@ name "
	   << "appd@" << Global.addrLocal.getHostName() << endl;
    }
  }
  
  cout << Global.sbVersion << endl;
  if (Global.fVerbose) {
    cout << "pid=" << getpid()
	 << ", listening on " << Global.addrLocal << endl;
    if (Global.fOutputHandler) {
      // format: name,options
      cout << "@@OUTPUTHANLDER@@ name "
	   << "appd@" << Global.addrLocal.getHostName() << endl;
    }
  }

#ifndef _WIN32
  // install sig handlers
  Global.fdSigPipeWrite = -1;
  struct sigaction s;
  s.sa_sigaction = sig_int;
  sigemptyset(&s.sa_mask);
  s.sa_flags = SA_SIGINFO;
  if (sigaction(SIGINT, &s, NULL) != 0) {
    perror("Cannot install SIGINT handler");
    exit(1);
  }
  s.sa_sigaction = sig_hup;
  sigemptyset(&s.sa_mask);
  s.sa_flags = SA_SIGINFO;
  if (sigaction(SIGHUP, &s, NULL) != 0) {
    perror("Cannot install SIGHUP handler");
    exit(1);
  }
  s.sa_sigaction = sig_pipe;
  sigemptyset(&s.sa_mask);
  s.sa_flags = SA_SIGINFO;
  if (sigaction(SIGPIPE, &s, NULL) != 0) {
    perror("Cannot install SIGPIPE handler");
    exit(1);
  }
  s.sa_sigaction = sig_chld;
  sigemptyset(&s.sa_mask);
  s.sa_flags = SA_SIGINFO | SA_RESTART;
  if (sigaction(SIGCHLD, &s, NULL) != 0) {
    perror("Cannot install SIGCHLD handler");
    exit(1);
  }
#endif /* WIN32*/

  // look for config file
  if (Global.cfg.ReadFile(SYSAPPDCFG) == 0) {
    // binary_dir
    // note: hostname in addrLocal is overridden by -host argument
    Global.sbBinaryDir = InitBinaryDir_Alloc(Global.cfg,
					     Global.addrLocal.getHostName());
#if 0
    Global.fSetX11Display = FindParmBoolDefault(Global.cfg, "appd",
						"set_x11_display",
						Global.fSetX11Display);
#endif
  }

  if (Global.a.Find("-exit")) {
    Global.fExitAfterExec = true;
  } else {
    Global.fExitAfterExec = false;
  }
  if (Global.a.Find("-nokillchild")) {
    Global.fKillChildLostConnection = false;
  } else {
    Global.fKillChildLostConnection = true;
  }
  if ((sb = Global.a.NextFind("-dird-host")) == NULL) {
    Global.sbDirdHost = DIRD_DEFAULT_HOST;
  } else {
    Global.sbDirdHost = sb;
  }
  if ((sb = Global.a.NextFind("-dird-port")) == NULL) {
    Global.wDirdPort = DIRD_DEFAULT_PORT;
  } else {
    Global.wDirdPort = atoi(sb);
  }

  char *sbInitAck = Global.a.NextFind("-init-ack");
  InitAck(sbInitAck);

  if (RegisterWithDird() != 0) return -1;

  // setup server loop
  AppdServerLoop SL;
  SL.Init(Global.sockLocal);
#ifndef _WIN32
  sigset_t maskSigAny;
  sigemptyset(&maskSigAny);
  sigaddset(&maskSigAny, SIGHUP);
  sigaddset(&maskSigAny, SIGINT);
  Global.fdSigPipeWrite = SL.setSig(maskSigAny);
#else
  
  g_win32Stop = FALSE;
  pthread_t finishedFilterThread;
  int resultWIN32 = pthread_create(&finishedFilterThread, NULL, RemoveFinishedFilters, (void *)&SL);

#endif

  while (1) {
    switch (SL.OnePass()) {
#ifdef _WIN32
	  case ServerLoop_Base::SLE_ERROR_WIN:
#else
	  case ServerLoop_Base::SLE_ERROR:
#endif
	  cout << ErrorHeader(wERROR) << "failed OnePass() call." << endl;
      // what else to do?
      break;
    case ServerLoop_Base::SLE_SHUTDOWN:
      // we have already UnRegistered with Dird
      Global.sockLocal.Close();
      cleanup();

#ifdef _WIN32
	  EnterCriticalSection(&g_cs);
	  g_win32Stop = TRUE;
	  LeaveCriticalSection(&g_cs);
#endif

	  return 0;
      break;
    }
  }
  return 0;  // never get here
}

int main(int argc, char *argv[]) {

// Initializes Winsock stack on Win32 platforms
#ifdef _WIN32
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;
 
	wVersionRequested = MAKEWORD(2, 0);
 
	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 ) {
		cout << "Error, Winsock could not be initialized.";
		return 0;
	}
  
	InitializeCriticalSection(&g_cs);

#endif /* WIN32 */


  Global.sbVersion = "DataCutter appd v1.4 (02/14/2002)";

  (void) StandardInit(argc, argv);
  Global.a.Set(argc, argv);

#ifndef _WIN32
  DBG_init(argv[0], environ);
#endif

  int wRtn = Init();

  (void) StandardCleanup();
  if (wRtn != 0) {
    cout << "appd: exiting with status " << wRtn << endl;
  }
  cout.flush();
  cerr.flush();

#ifdef _WIN32
	DeleteCriticalSection(&g_cs);
#endif

  return wRtn;
}


// ****************************************************************************
