// ****************************************************************************
//  Name: boot_util.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 04/17/2000 : mdb - Created.
//        04/20/2000 : mdb - appd/dird/all working.  Still need to start appds
//                           asynchronously (maybe thread per appd), or a list
//                           of outstanding ports to accept on.  Need to split
//                           StartDaemon() to accomodate this.
//                           BUGBUG: shutdown w/Ctrl+C ok, shutdown all - will
//                           cause the dird next exe bind to fail?!?  No idea
//                           why, the code paths appear to do the same thing.
//                           Maybe some race with closing sockets?
//        07/13/2000 : mdb - Pass HostInfo version of dird host to appd.
//        08/13/2000 : mdb - Can start/stop multiple appd at once.
//        09/22/2000 : mdb - Use canonical name for appd/dird -host arg.
//        10/26/2000 : mdb - Added [-debug] and changed run dir name from tmp.
//        10/28/2000 : mdb - Added [-wait] to wait for daemons to finish, so
//                           scheduler on linux cluster will not exit too soon.
//        11/08/2000 : mdb - Fixed bug in "all stop" using [appstream.appd.*]
//                           Added "appd wait" to deal with parallel scheduler.
//                         - Added "dird <add|remove|query>" cmds to deal with
//                           scheduler to have more expts running at one time.
//                           Also modified wait to pass a specific host.
//        11/09/2000 : mdb - Fixed bug where appd wait should use canonical.
//        12/03/2000 : mdb - Changed path assumptions for binaries.
//        04/13/2001 : mdb - Added "appd stop -" to shutdown all running
//                           appd found in the dird.  This is useful for cases
//                           where a dumb scheduler (pbs) leaves processes
//                           running on the nodes after the allocation is done.
//        06/07/2001 : mdb - New status command from all appds.
//        07/20/2001 : mdb - Changed [-wait] to [-sleep].
//        07/27/2001 : mdb - Added "appd uptime" command.
//        12/06/2001 : mdb - "dird stop" now does "appd stop -" first.
//                         - If use -noxterm with appd start, check first for
//                           a dird registered outputhandler.
//        12/15/2001 : mdb - Modified daemon startup to remove serialization,
//                           and start all in parallel before init acks are
//                           recvd.  dird and appd now send an initid to
//                           determine which daemon has started.
//                           
//  Desc: Standlaone utility for doing various things such as booting the
//        runtime system, for shutting it down, and performing profiling of
//        cpu and network (TODO).
// ****************************************************************************

#include <signal.h>
#include <assert.h>

#include "lib/dc_standard.h"
#include "lib/dc_configinfo.h"
#include "lib/dc_exec.h"
#include "lib/dc_socket.h"
#include "lib/dc_darray.h"
#include "lib/dc_file.h"

#include "dird.h"
#include "dird-client.h"
#include "appd-client.h"
#include "hostinfo.h"
#include "packet.h"
#include "util.h"


// this enables tty mode for xterms, and other settings to aid in debugging
//#define BOOT_UTIL_DCDBG_MODE

//
// defaults
//
#if 1
// many rows of short xterms
const int wGeomCols       = 80;
const int wGeomRows       = 24;
const int wGeomBaseX      = 429;
const int wGeomBaseY      = 0;
const int wGeomIncrX      = 0;
const int wGeomIncrY      = 199;
const int wGeomGroupNum   = 4;
const int wGeomGroupIncrX = 100;
const int wGeomGroupIncrY = 0;
#else
// tall xterms - better for debugging
const int wGeomCols       = 80;
const int wGeomRows       = 75;
const int wGeomBaseX      = 100;
const int wGeomBaseY      = 0;
const int wGeomIncrX      = 313;
const int wGeomIncrY      = 0;
const int wGeomGroupNum   = 3;
const int wGeomGroupIncrX = 0;
const int wGeomGroupIncrY = 70;
#endif

const int wDaemonTimeout = 20;  // num seconds to wait for daemon ack

// BUGBUG: need a more general way to handle this -- probably use
// an envvar, or a datacutter-config.sh file in the path, or a
// ~/.datacutterrc file.
char *sbRemoteRunDir       = "work/dc/run"; // relative to home
char *sbRemoteDaemonBinDir = "../src"; // relative to rundir

typedef enum {
  eAppdCmdStatus=0, eAppdCmdUptime, eAppdCmdPstack, eAppdCmdFtp
} eAppdCmd_t;
typedef enum {
  DAEMONSTATE_START, DAEMONSTATE_INIT, DAEMONSTATE_EXIT
} eDaemonState_t;
typedef enum {
  DAEMON_DIRD, DAEMON_APPD
} eDaemonType_t;

typedef struct {
  pid_t  pid;
  char  *sbDesc;
  eDaemonType_t  eType;
  eDaemonState_t eState;
  int    wInitId;
} daemon_t;

typedef struct {
  char *sbHost;
  int   wPort;
  bool  fMark;
} appd_command_t;

struct Global_s {
  HostInfo hi;
  bool     fDebug;
  bool     fNoXterm;
  bool     fVerbose;
  bool     fOutputHandler;
  char    *sbOutputHandler;

  // these are for the [-sleep] option and for recv initacks
  bool     fSleep;
  int      ndaemonlist;
  daemon_t daemonlist[128];

  InetAddress  addrLocal;
  ServerSocket sockLocal;
  int          wInitId;
  int          nInitNeeded;

  // state for appd_commands
  DArray<appd_command_t*> daAppdCommand;
} Global;


class StartInfo {
public:
  eDaemonType_t eDaemonType;
  int   nXterm;
  char *sbHost;
  int   wPort;
  char *sbUser;
  char *sbStartDir;
  char *sbHostDirdConnect;  // only used for DAEMON_APPD

  StartInfo(void) : nXterm(-1), sbHost(0), wPort(0), sbUser(0),
		    sbStartDir(0), sbHostDirdConnect(0) {};
};


int dird_start(void);  // fwd ref
int dird_status(void);
int appd_start(char *);
int appd_wait(char *);
int appd_stop(char *);

// contact the dird and see if any output handler has been registered
// rtn: port: <0=error, 0=no handler registered, >0=handler registered
// updates Global.fOutputHandler,sbOutputHandler
int FindOutputHandler(void) {
  ConfigInfo cfg;
  char buf[128];

  Global.fOutputHandler = true;  // indicates we checked the dird

  snprintf(buf, sizeof(buf), "datacutter.outputhandler");
  if (Dird_Query(buf, cfg, NULL, 0, 0) >= 0) {
    char *sbHost = cfg.Find(buf, "dns_name");
    char *sbPort = cfg.Find(buf, "port");
    if (sbHost && sbPort) {
      size_t wLen = strlen(sbHost) + 20;
      Global.sbOutputHandler = new char[wLen];
      snprintf(Global.sbOutputHandler, wLen, "%s:%s", sbHost, sbPort);
      // TODO: pass this through HostInfo to get correct name mapping
      // TODO: actually attempt to connect to the host:port to make sure
      // it is alive (need to map hostinfo from this host for the test)
      return 1;  // found
    }
    return 0;  // no running handler found
  }
  return -1;  // dird not running
}

char *FixRelativePathname_Alloc(char *sbPath, char *sbRelativeTo) {
  if (sbPath[0] == '/') return StrDup(sbPath);  // not relative
  // get path to prepend
  char *p = strrchr(sbRelativeTo, '/');
  if (!p) return StrDup(sbPath);  // no path to prepend
  // prepend the path
  char *sbNew = new char[strlen(sbPath)+(p-sbRelativeTo+1)+1];
  sbNew[0] = '\0';
  strncat(sbNew, sbRelativeTo, p-sbRelativeTo+1);
  strcat(sbNew, sbPath);
  return sbNew;
}

// this uses DEPRECATED syntax of console cfg ...
void LoadAppConsoleCfg(char *sbFile_in, ConfigInfo &cfg) {
  char *sbFile;
  if (sbFile_in) {
    cfg.ReadFile(sbFile_in);
  } else {
    cfg.ReadFile("console.cfg");
  }
  sbFile = cfg.Find("console", "application_cfg");
  if (sbFile) {
    sbFile = FixRelativePathname_Alloc(sbFile, sbFile_in);
    cfg.ReadFile(sbFile, false);
    delete [] sbFile;
  }
  sbFile = cfg.Find("console", "placement_cfg");
  if (sbFile) {
    sbFile = FixRelativePathname_Alloc(sbFile, sbFile_in);
    cfg.ReadFile(sbFile, false);
    delete [] sbFile;
  }
}

static void sig_alrm(int, siginfo_t*, void*) { return; }

// check the listen socket for init acks
// rtn: number of init acks still expected
int CheckForInitAcks(bool fBlock) {
  int fd;
  if (Global.nInitNeeded == 0) return Global.nInitNeeded;
  InetAddress addrAccept;

  while (Global.nInitNeeded > 0) {
    if (fBlock) {
      if (wDaemonTimeout > 0) {
		  struct sigaction s;
	s.sa_sigaction = sig_alrm;
	sigemptyset(&s.sa_mask);
	s.sa_flags = SA_SIGINFO;
	if (sigaction(SIGALRM, &s, NULL) == 0) {
	  alarm(wDaemonTimeout);
	}
      }
      fd = Global.sockLocal.Accept(addrAccept);
      if (fd == -1) {
	cerr << ErrorHeader(wWARNING)
	     << "daemon init timeout for:" << endl;
	for (int i=0; i<Global.ndaemonlist; ++i) {
	  if (Global.daemonlist[i].eState != DAEMONSTATE_START) continue;
	  cerr << " - " << Global.daemonlist[i].sbDesc << endl;
	}
	return -1;  // alarm timeout?
      }
    } else {
      fd = Global.sockLocal.IAccept(addrAccept);
    }
    if (fd < 0) break;
    // read init ack (format: one line, ascii integer)
    int ret;
    Packet packetIn;
    BaseSocket sock;
    sock.getFd() = fd;
    if ((ret = packetIn.read(sock)) <= 0) {
      cerr << "debug: failed read of init ack packet" << endl;
    } else if (packetIn.getTag() != TAG_BOOTUTIL_INITACK) {
      cerr << "error: received non-expected init ack tag" << endl;
    } else {
      int wInitAck = packetIn.getExtra();
      bool fFound = false;
      for (int i=0; i<Global.ndaemonlist; ++i) {
	if (Global.daemonlist[i].wInitId != wInitAck) continue;
	fFound = true;
	Global.daemonlist[i].eState = DAEMONSTATE_INIT;
	--Global.nInitNeeded;
	if (Global.fVerbose)
	  cout << (Global.daemonlist[i].eType == DAEMON_DIRD ? "dird" : "appd")
	       << " initialized on " << Global.daemonlist[i].sbDesc
	       << " (" << Global.nInitNeeded << " more)" << endl;
	break;
      }
      if (!fFound) {
	cerr << "ERROR: received init ack we never sent!" << endl;
      }
    }
    close(fd);
  }
  return Global.nInitNeeded;
}

// BUGBUG: case when xterm not used does not work
// rtn: <0=fail, >0=success pid of xterm or rsh/ssh
int StartDaemon(StartInfo &si) {
  char buf[2048];
  Argv a, aRemote;
  Exec e;
  char *sbName, *sbBgColor;
  int eExecFlags;

  if (Global.ndaemonlist == 0) {
    Global.addrLocal.Set(0);
    Global.sockLocal.CreateBindListen(Global.addrLocal);
  }

  if (si.eDaemonType == DAEMON_DIRD) {
    sbName = "dird";
    sbBgColor = "yellow";
  } else if (si.eDaemonType == DAEMON_APPD) {
    sbName = "appd";
    sbBgColor = "gray80";
  } else {
    return -1;
  }

  if (Global.fVerbose) {
    cout << "starting " << sbName << " on " << si.sbHost << " ... ";
    cout.flush();
  }

  int wInitId = Global.wInitId++;

  // use an xterm?
  if (si.nXterm >= 0) {
    a.Append("xterm");
    a.Append("-title");
    if (si.sbUser) {
      snprintf(buf, sizeof(buf), "%s: (host=%s@%s)", sbName,
	      si.sbUser, si.sbHost);
    } else {
      snprintf(buf, sizeof(buf), "%s: (host=%s)", sbName, si.sbHost);
    }
    a.Append(StrDup(buf), true);
    a.Append("-sl");
    a.Append("1000");
    a.Append("-bg");
    a.Append(sbBgColor);
    a.Append("-fg");
    a.Append("black");
    a.Append("-font");
    a.Append("5x7");
    a.Append("-geom");
    if (si.eDaemonType == DAEMON_DIRD) {
      snprintf(buf, sizeof(buf), "+%d+%d", 0, 0);
    } else if (si.eDaemonType == DAEMON_APPD) {
      int nGroup   = si.nXterm / wGeomGroupNum;
      int nInGroup = si.nXterm % wGeomGroupNum;
      snprintf(buf, sizeof(buf), "%dx%d+%d+%d", wGeomCols, wGeomRows,
	      wGeomBaseX + nGroup*wGeomGroupIncrX + nInGroup*wGeomIncrX,
	      wGeomBaseY + nGroup*wGeomGroupIncrY + nInGroup*wGeomIncrY);
    }
    a.Append(StrDup(buf), true);
    a.Append("-e");
    eExecFlags = ( Exec::REXEC_FLAGS_ALLOCATETTY | \
		   Exec::REXEC_FLAGS_AUTOADDHOSTKEYS );
  } else {
    // Note: Since daemon mode implies the ssh tunnel to the remote node
    //       will stop once the daemon forks, jitd will not work since
    //       we have no X11 display to write to.  Users will need to realize
    //       this and setup a fixed working DISPLAY, or configure a single
    //       X11 tunnel from a shared node (like a cluster front end), and
    //       allow the nodes to use the shared ~/.Xauthority file.
    //       Note: the option "[console]remote_x11_display = xauth list-add"
    //       is used to address this need.
    eExecFlags = ( Exec::REXEC_FLAGS_AUTOADDHOSTKEYS |\
		   Exec::REXEC_FLAGS_NOREADSTDIN |\
		   Exec::REXEC_FLAGS_NOFORWARDX11 );
  }

  // we use /bin/sh here for expansion of envvars on the remote end
  snprintf(buf, sizeof(buf), "/bin/sh -c \"");
  if (si.sbStartDir) {
    snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), "cd %s ; ", si.sbStartDir);
  }
  // 09/22/2000 : mdb - use the canonical name for the -host arg
  char *sbHostCanonical = Global.hi.CanonicalName_Alloc(si.sbHost);

  // 04/16/02 : CLH - Removed dependecy on a particular location for DC executables.
  // Users can use their PATH env variable instead
  //sprintf(buf+strlen(buf), "%s/%s -host %s", sbRemoteDaemonBinDir, sbName, sbHostCanonical);
  snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), "%s -host %s", sbName, sbHostCanonical);
  delete [] sbHostCanonical;
  if (si.eDaemonType == DAEMON_APPD && si.sbHostDirdConnect) {
    snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), " -dird-host %s", si.sbHostDirdConnect);
  }
  if (si.wPort > 0) {
    if (si.eDaemonType == DAEMON_DIRD) {
      snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), " -port %d", si.wPort);
    } else {
      snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), " -dird-port %d", si.wPort);
    }
  }
  if (si.nXterm == -1) {
    snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), " -daemon");
  }
  if (si.eDaemonType == DAEMON_APPD && Global.sbOutputHandler) {
    snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), " -outputhandler %s", Global.sbOutputHandler);
  }
#if defined(BOOT_UTIL_DCDBG_MODE)
  if (si.eDaemonType == DAEMON_APPD) {
    snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), " -nokillchild");
  }
#endif /* BOOT_UTIL_DCDBG_MODE */

  // 07/12/2000 : mdb - Fixed bug where local hostname was not run thru hi
  char *sbHostTmp = Global.hi.ConnectNameFrom_Alloc(si.sbHost);
  snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), " -init-ack %s:%d,%d",
	  sbHostTmp, Global.addrLocal.getPort(), wInitId);
  delete [] sbHostTmp;

  //sprintf(buf+strlen(buf), "; sleep 10");
  snprintf(buf+strlen(buf), sizeof(buf)-strlen(buf), "\"");
  e.AppendRcmd(a, Exec::REXEC_SSH, si.sbHost, buf, si.sbUser, eExecFlags);

  int fdStdin = Exec::FD_PARENT;
  int fdStdout = Exec::FD_PARENT;
  int fdStderr = Exec::FD_PARENT;
  if (Global.fDebug) {
    cout << "About to execute:" << endl << a << endl;
  }
  pid_t pid = e.ForkExec(a, fdStdin, fdStdout, fdStderr);
  // we will check for this later ...

  // add to array of pids
  Global.daemonlist[Global.ndaemonlist].pid = pid;
  snprintf(buf, sizeof(buf), "%s on %s", sbName, si.sbHost);
  Global.daemonlist[Global.ndaemonlist].sbDesc = StrDup(buf);
  Global.daemonlist[Global.ndaemonlist].eType = si.eDaemonType;
  Global.daemonlist[Global.ndaemonlist].eState = DAEMONSTATE_START;
  Global.daemonlist[Global.ndaemonlist].wInitId = wInitId;
  ++Global.ndaemonlist;
  if (Global.fVerbose) cout << "pid=" << pid << endl;

  ++Global.nInitNeeded;
  // to avoid filling the accept queue, we check for any replies
  CheckForInitAcks(false);

  return pid;
}


// rtn: <0=fail, 0=already running, >0=success pid of xterm or rsh/ssh
int StartDird(StartInfo &si) {
  // check if it is running already
  if (Dird_Ping(si.sbHost, si.wPort) == 0) return 0;
  si.eDaemonType = DAEMON_DIRD;
  return StartDaemon(si);
}


// rtn: <0=fail, 0=already running, >0=success pid of xterm or rsh/ssh
int StartAppd(StartInfo &si) {
  // check if it is running already
  int wPort = Appd_Find(si.sbHost);
  if (wPort < 0) {  // failure: no dird running
    if (dird_start() < 0) return wPort;  // failure: cannot start dird
    wPort = Appd_Find(si.sbHost);
    if (wPort < 0) return wPort;  // failure: still cannot find dird 
  }  
  if (wPort > 0 && Appd_Ping(si.sbHost, wPort) == 0) return 0;
  si.eDaemonType = DAEMON_APPD;

  //char *sbHostDird = Global.hi.CanonicalName_Alloc(DIRD_DEFAULT_HOST);
  //si.sbHostDirdConnect = Global.hi.ConnectName(si.sbHost, sbHostDird);

  int wRtn = StartDaemon(si);

  //DeleteNull(sbHostDird);
  return wRtn;
}

// ASSUME: dird must already be running
int DoAllAppd(bool fStart, StartInfo &si, ConfigInfo &cfg) {
  const char *sbFilter = NULL, *sbSection;
  char *sbHost;
  int   rtn, n = 0;
  int wRtn = 0;
  char *sbHostDird = Global.hi.CanonicalName_Alloc(si.sbHost);

  sbSection = cfg.FirstSection();
  while (sbSection != NULL) {
    // anything with a section name "filter.*" is considered a filter
    if (strncmp(sbSection, "filter.", 7) == 0 && strlen(sbSection) > 7) {
      // BUGBUG: not checking to insure no whitespace is in the filter name
      sbFilter = sbSection+7;
      // find the placement for this filter
      sbHost = cfg.Find("placement", sbFilter);
      if (!sbHost) {
	cerr << ErrorHeader(wERROR)
	     << "missing placement spec for filter '" << sbFilter << "'"
	     << endl;
      } else {
	DeleteNull(si.sbHost);  // in case already alloc by caller
	si.sbHost = Global.hi.ConnectNameTo_Alloc(sbHost);
	si.sbHostDirdConnect = Global.hi.ConnectName(si.sbHost, sbHostDird);
	if (fStart) {
	  si.nXterm = (Global.fNoXterm ? -1 : n++);
	  FindOutputHandler();  // sets Global.{f,sb}OutputHandler
	  rtn = StartAppd(si);
	  if (rtn == 0) {
	    //cerr << ErrorHeader(wWARNING) << "appd already running on "
	    // << si.sbHost << endl;
	  } else if (rtn > 0) {
	    if (Global.fVerbose) cout << "appd started on " << si.sbHost << endl;
	  } else {
	    cerr << ErrorHeader(wERROR) << "could not start appd on "
		 << si.sbHost << endl;
	    wRtn = 1;
	  }
	  rtn = Appd_Find(si.sbHost, true);
	  if (rtn <= 0) {
	    cerr << ErrorHeader(wWARNING) 
		 << "no appd running on " << sbHost << endl;
	  }
	} else {
	  int wPort = Appd_Find(si.sbHost);  // rtn port
	  if (wPort <= 0) {
	    //cerr << ErrorHeader(wWARNING) 
	    // << "no appd running on " << si.sbHost << endl;
	  } else {
	    Appd_Shutdown(si.sbHost, wPort);
	  }
	}
      }
    }
    sbSection = cfg.NextSection();
  }
  if (!sbFilter) {
    wRtn = 1;
  }
  DeleteNull(sbHostDird);
  return wRtn;
}

int dird_start(void) {
  StartInfo si;

  if (getenv("DATACUTTER_DIRDHOST")) {
    si.sbHost = Global.hi.ConnectNameTo_Alloc(getenv("DATACUTTER_DIRDHOST"));
  } else {
    si.sbHost = Global.hi.ConnectNameTo_Alloc(DIRD_DEFAULT_HOST);
  }
  if (getenv("DATACUTTER_DIRDPORT")) {
    si.wPort = atoi(getenv("DATACUTTER_DIRDPORT"));
  } else {
    si.wPort = DIRD_DEFAULT_PORT;
  }  
  //si.sbStartDir = sbRemoteRunDir;
  si.nXterm = (Global.fNoXterm ? -1 : 0);
  int wRtn = StartDird(si);
  if (wRtn == 0) {
    cerr << ErrorHeader(wWARNING) << "dird already running on "
	 << si.sbHost << endl;
  } else if (wRtn > 0) {
    if (Global.fVerbose) cout << "dird started on " << si.sbHost << endl;
  } else {
    cerr << ErrorHeader(wERROR) << "could not start dird on "
	 << si.sbHost << endl;
  }
  delete [] si.sbHost;
  return wRtn;
}

int dird_stop(void) {
  char *sbHost;
  int wPort;
  if (getenv("DATACUTTER_DIRDHOST")) {
    sbHost = Global.hi.ConnectNameTo_Alloc(getenv("DATACUTTER_DIRDHOST"));
  } else {
    sbHost = Global.hi.ConnectNameTo_Alloc(DIRD_DEFAULT_HOST);
  }
  if (getenv("DATACUTTER_DIRDPORT")) {
    wPort = atoi(getenv("DATACUTTER_DIRDPORT"));
  } else {
    wPort = DIRD_DEFAULT_PORT;
  }  
  int wRtn = Dird_Shutdown(sbHost, wPort);
  if (wRtn == 0) {
    if (Global.fVerbose)
      cout << "shutdown dird" << endl;
  }
  delete [] sbHost;
  return wRtn;
}

int dird_status(void) {
  char *sbHost;
  int wPort;
  if (getenv("DATACUTTER_DIRDHOST")) {
    sbHost = Global.hi.ConnectNameTo_Alloc(getenv("DATACUTTER_DIRDHOST"));
  } else {
    sbHost = Global.hi.ConnectNameTo_Alloc(DIRD_DEFAULT_HOST);
  }
  if (getenv("DATACUTTER_DIRDPORT")) {
    wPort = atoi(getenv("DATACUTTER_DIRDPORT"));
  } else {
    wPort = DIRD_DEFAULT_PORT;
  }  
  int wRtn = Dird_Ping(sbHost, wPort);
  if (wRtn == 0) {
    cout << "dird on " << sbHost << ':' << wPort << " is alive." << endl;
  } else {
    cout << "dird is not running" << endl;
  }
  delete [] sbHost;
  return wRtn;
}

int dird_add(char *sbSection, int nEntry, char *sbEntry[]) {
  char *sbHost;
  int wPort;
  if (getenv("DATACUTTER_DIRDHOST")) {
    sbHost = Global.hi.ConnectNameTo_Alloc(getenv("DATACUTTER_DIRDHOST"));
  } else {
    sbHost = Global.hi.ConnectNameTo_Alloc(DIRD_DEFAULT_HOST);
  }
  if (getenv("DATACUTTER_DIRDPORT")) {
    wPort = atoi(getenv("DATACUTTER_DIRDPORT"));
  } else {
    wPort = DIRD_DEFAULT_PORT;
  }  
  int wRtn = 0;
  if (wPort <= 0) {
    cerr << ErrorHeader(wWARNING) 
	 << "no dird running on " << sbHost << endl;
  } else {
    MemoryBuf mb(1024);
    mb.sprintf("[%s]\n", sbSection);
    for (int i=0; i<nEntry; ++i) {
      mb.sprintf("%s\n", sbEntry[i]);
    }
    wRtn = Dird_Add(mb, sbHost, wPort);
    if (wRtn != 0) {
      cerr << ErrorHeader(wERROR)
	   << "failed sending add command to dird on " << sbHost << endl;
    }
  }
  delete [] sbHost;
  return wRtn;
}
int dird_query(char *sbSection, char *sbName) {
  char *sbHost;
  int wPort;
  if (getenv("DATACUTTER_DIRDHOST")) {
    sbHost = Global.hi.ConnectNameTo_Alloc(getenv("DATACUTTER_DIRDHOST"));
  } else {
    sbHost = Global.hi.ConnectNameTo_Alloc(DIRD_DEFAULT_HOST);
  }
  if (getenv("DATACUTTER_DIRDPORT")) {
    wPort = atoi(getenv("DATACUTTER_DIRDPORT"));
  } else {
    wPort = DIRD_DEFAULT_PORT;
  }  
  int wRtn = 0;
  if (wPort <= 0) {
    cerr << ErrorHeader(wWARNING) 
	 << "no dird running on " << sbHost << endl;
  } else {
    ConfigInfo cfg;
    // We assume that the required name will be in the first section
    // that shows up in the dird.  To do this better would require new
    // dird-client api routines and DIRD query packets.
    wRtn = Dird_Query(sbSection, cfg, sbHost, wPort, 1);
    if (wRtn > 0) {
      // now grab out the required name(s)
      bool fFirst = true;
      const char *sbSectionCfg, *sbNameCfg;
      sbSectionCfg = cfg.FirstSection();
      while (sbSectionCfg) {
	sbNameCfg = cfg.FirstName();
	while (sbNameCfg) {
	  if (strcmp(sbName, sbNameCfg) == 0) {
	    if (fFirst) {
	      fFirst = false;
	    } else {
	      cout << ' ';
	    }
	    cout << cfg.CurrentValue();
	  }
	  sbNameCfg = cfg.NextName();
	}
	sbSectionCfg = cfg.NextSection();
      }
      if (!fFirst) cout << endl;
    }
  }
  delete [] sbHost;
  return wRtn;
}
int dird_remove(char *sbSection) {
  char *sbHost;
  int wPort;
  if (getenv("DATACUTTER_DIRDHOST")) {
    sbHost = Global.hi.ConnectNameTo_Alloc(getenv("DATACUTTER_DIRDHOST"));
  } else {
    sbHost = Global.hi.ConnectNameTo_Alloc(DIRD_DEFAULT_HOST);
  }
  if (getenv("DATACUTTER_DIRDPORT")) {
    wPort = atoi(getenv("DATACUTTER_DIRDPORT"));
  } else {
    wPort = DIRD_DEFAULT_PORT;
  }  
  int wRtn = 0;
  if (wPort <= 0) {
    cerr << ErrorHeader(wWARNING) 
	 << "no dird running on " << sbHost << endl;
  } else {
    MemoryBuf mb((int)(strlen(sbSection)+1));
    mb.sprintf("%s", sbSection);
    wRtn = Dird_Remove(mb, sbHost, wPort);
    if (wRtn != 0) {
      cerr << ErrorHeader(wERROR)
	   << "failed sending remove command to dird on " << sbHost << endl;
    }
  }
  delete [] sbHost;
  return wRtn;
}

int appd_start(char *sbHost_in) {
  static int wCount=0;
  StartInfo si;
  if (strcmp(sbHost_in, "-") == 0) {
    cerr << ErrorHeader(wERROR)
	 << "cannot use \"-\" syntax for starting an appd" << endl;
    return 1;
  }
  if (dird_status() != 0) {
    if (dird_start() < 0) return 1;
    int ret;
    while ((ret = CheckForInitAcks(true)) > 0)
      ;
    if (ret < 0) exit(1);
  }
  si.sbHost = Global.hi.ConnectNameTo_Alloc(sbHost_in);
  //si.sbStartDir = sbRemoteRunDir;
  if (getenv("DATACUTTER_DIRDHOST")) {
    si.sbHostDirdConnect = Global.hi.ConnectNameTo_Alloc(getenv("DATACUTTER_DIRDHOST"));
  }
  if (getenv("DATACUTTER_DIRDPORT")) {
    si.wPort = atoi(getenv("DATACUTTER_DIRDPORT"));
  }
  si.nXterm = (Global.fNoXterm ? -1 : wCount++);
  FindOutputHandler();  // sets Global.{f,sb}OutputHandler
  int wRtn = StartAppd(si);
  if (wRtn == 0) {
    cerr << ErrorHeader(wWARNING) << "appd already running on "
	 << si.sbHost << endl;
  } else if (wRtn > 0) {
    if (Global.fVerbose) cout << "appd started on " << si.sbHost << endl;
  } else {
    cerr << ErrorHeader(wERROR) << "could not start appd on "
	 << si.sbHost << endl;
  }
  delete [] si.sbHost;
  return wRtn;
}
int appd_stop(char *sbHost_in) {
  int wRtn;
  if (strcmp(sbHost_in, "-") == 0) {
    wRtn = Appd_ShutdownAll();
    if (wRtn > 0) {
      cout << "shutdown " << wRtn << " appds" << endl;
    }
  } else {
    char *sbHost = Global.hi.ConnectNameTo_Alloc(sbHost_in);
    wRtn = Appd_Find(sbHost);  // rtn port
    if (wRtn <= 0) {
      cerr << ErrorHeader(wWARNING) 
	   << "no appd running on " << sbHost << endl;
    } else {
      wRtn = Appd_Shutdown(sbHost, wRtn);
    }
    delete [] sbHost;
  }
  return wRtn;
}
int appd_wait(char *sbHost_in) {
  char *sbHost = Global.hi.CanonicalName_Alloc(sbHost_in, true);
  cout << "looking for " << sbHost_in << " or " << sbHost << endl;
  int wRtn = Appd_Find(sbHost, true);  // rtn port
  if (wRtn <= 0) {
    cerr << ErrorHeader(wWARNING) 
	 << "no appd running on " << sbHost << endl;
  }
  delete [] sbHost;
  return wRtn;
}
int appd_waitn(int n) {
  ConfigInfo cfg;
  // wait until n appd's are available -- crude hack for scheduler use
  int ret = Dird_Query("datacutter.appd.*", cfg, NULL, 0, n);
  if (ret != n) {
    cerr << ErrorHeader(wERROR)
	 << "waited for " << n << " reponses, but only received " << ret
	 << endl;
    return 1;
  }
  return 0;
}
// reads in all the appd entries from the dird, and caches them for
// use with one or more appd_command() calls.
int appd_command_find_appds(void) {
  ConfigInfo cfg;
  if (Dird_Query("datacutter.appd.*", cfg) >= 0) {
    const char *sbSection;
    char *sbHost, *sbDnsName;
    int wPort;
    sbSection = cfg.FirstSection();
    while (sbSection != NULL) {
      sbDnsName = cfg.Find(sbSection, "dns_name");
      sbHost = Global.hi.ConnectNameTo_Alloc(sbDnsName);
      wPort = atoi(cfg.Find(sbSection, "port"));
      if (Appd_Ping(sbHost, wPort) < 0) {
	// appd is not really there, so cleanup dird entry
	cout << "note: appd on " << sbHost << ':' << wPort << " is not alive"
	     << " (removing from dird)" << endl;
	MemoryBuf mbTmp(sbSection, (int)strlen(sbSection), (int)strlen(sbSection));
	Dird_Remove(mbTmp);
	DeleteNull(sbHost);
      } else {
	// save state for later commands
	appd_command_t *pNew = new appd_command_t;
	pNew->sbHost = sbHost;
	pNew->wPort = wPort;
	pNew->fMark = true;
	Global.daAppdCommand.Append(pNew);
      }
      sbSection = cfg.NextSection();
    }
  }
  return 0;
}
int appd_command_subset_init(bool fMark) {
  // mark all as not used
  unsigned int i;
  for (i=0; i<Global.daAppdCommand.NumElem(); ++i) {
    appd_command_t *p = Global.daAppdCommand.Get(i);
    p->fMark = fMark;
  }
  return 0;
}
int appd_command_subset_check(char *sbHost_in) {
  if (strcmp(sbHost_in, "-") == 0) {
    appd_command_subset_init(true);
    return 0;
  }
  // check for a hostname
  char *sbHost = Global.hi.ConnectNameTo_Alloc(sbHost_in);
  unsigned int i;
  for (i=0; i<Global.daAppdCommand.NumElem(); ++i) {
    appd_command_t *p = Global.daAppdCommand.Get(i);
    if (strcmp(sbHost, p->sbHost) == 0) {
      p->fMark = true;
      break;
    }
  }
  if (i == Global.daAppdCommand.NumElem()) {  // not found?
    cerr << ErrorHeader(wWARNING)
	 << "no appd found running on host " << sbHost_in
	 << " (" << sbHost << ")" << endl;
  }
  delete [] sbHost;
  return 0;
}
// run a "command" against the marked entries in the Global.daAppdCommand state
int appd_command(eAppdCmd_t eCmd, char *sbArg) {
  int wRtn = 0, ret;
  unsigned int i;
  File f;
  unsigned int wChecksum;

  if (eCmd == eAppdCmdFtp) {  // open and compute sum here once for all appds
    f.set(sbArg);
    if (f.open() == -1) {
      cerr << "error: failed open of file \"" << sbArg << '\"' << endl;
      return 1;
    }
    cout << "file \"" << sbArg << "\" ..." << endl;
    wChecksum = f.checksum();
  }

  for (i=0; i<Global.daAppdCommand.NumElem(); ++i) {
    appd_command_t *p = Global.daAppdCommand.Get(i);
    if (!p->fMark) continue;  // not marked
    
    if (eCmd == eAppdCmdStatus) {
      cout << "appd on " << p->sbHost << ':' << p->wPort << " is alive."
	   << endl;
    } else if (eCmd == eAppdCmdUptime) {
      // TODO: send a status query packet to each appd, and have it
      // return remote process info, such as apps running, state of
      // processes, etc.
    } else if (eCmd == eAppdCmdPstack) {
      // TODO: send a status query packet to each appd, and have it
      // return remote process info, such as apps running, state of
      // processes, etc.
      ret = Appd_Pstack(p->sbHost, p->wPort);
    } else if (eCmd == eAppdCmdFtp) {
      //cout << "file \"" << sbArg << "\" to appd on "
      //   << p->sbHost << ':' << p->wPort << " ..." << endl;
      MemoryBuf mb(MB_1);
      ret = Appd_Ftp(p->sbHost, p->wPort, f, wChecksum, mb);
      if (ret) wRtn = ret;
    } else {
      assert(0);
    }
  }
  return ret;
}

int all_start(char *sbFile) {
  StartInfo si;
  ConfigInfo cfg;
  LoadAppConsoleCfg(sbFile, cfg);
  char env1[80], env2[80];

  char *sbTmp;
  sbTmp = FindParm(cfg, "console", "dird_host", false);
  if (sbTmp) {
    si.sbHost = Global.hi.ConnectNameTo_Alloc(sbTmp);
    sprintf(env1, "DATACUTTER_DIRDHOST=%s", sbTmp);
    putenv(env1);
  } else if (getenv("DATACUTTER_DIRDHOST")) {
    si.sbHost = Global.hi.ConnectNameTo_Alloc(getenv("DATACUTTER_DIRDHOST"));
  } else {
    si.sbHost = Global.hi.ConnectNameTo_Alloc(DIRD_DEFAULT_HOST);
  }
  sbTmp = FindParm(cfg, "console", "dird_port", false);
  if (sbTmp) {
    si.wPort = atoi(sbTmp);
    sprintf(env2, "DATACUTTER_DIRDPORT=%s", sbTmp);
    putenv(env2);
  } else if (getenv("DATACUTTER_DIRDPORT")) {
    si.wPort = atoi(getenv("DATACUTTER_DIRDPORT"));
  } else {
    si.wPort = DIRD_DEFAULT_PORT;
  }  
  si.nXterm = (Global.fNoXterm ? -1 : 0);

  int wRtn = StartDird(si);
  if (wRtn == 0) {
    cerr << ErrorHeader(wWARNING) << "dird already running on "
	 << si.sbHost << endl;
  } else if (wRtn > 0) {
    if (Global.fVerbose) cout << "dird started on " << si.sbHost << endl;
  } else {
    cerr << ErrorHeader(wERROR) << "could not start dird on "
	 << si.sbHost << endl;
    exit(1);
  }
  int ret;
  while ((ret = CheckForInitAcks(true)) > 0)
    ;
  wRtn = DoAllAppd(true, si, cfg);
  if (wRtn) {
    cerr << "placement section not found, no appds started" << endl;
    int tmp = Dird_Shutdown(si.sbHost, si.wPort);
    if (tmp == 0) {
      if (Global.fVerbose)
	cout << "shutdown dird" << endl;
    }
  }
  delete [] si.sbHost;
  return wRtn;
}

int all_stop(char *sbFile) {
  //StartInfo si;
  ConfigInfo cfg;
  LoadAppConsoleCfg(sbFile, cfg);
  char env1[80], env2[80];

  char *sbTmp;
  sbTmp = FindParm(cfg, "console", "dird_host", false);
  if (sbTmp) {
    sprintf(env1, "DATACUTTER_DIRDHOST=%s", sbTmp);
    putenv(env1);
  }
  sbTmp = FindParm(cfg, "console", "dird_port", false);
  if (sbTmp) {
    sprintf(env2, "DATACUTTER_DIRDPORT=%s", sbTmp);
    putenv(env2);
  }

  int wRtn = 0;
  //if (sbFile && strcmp(sbFile, "-") == 0) {  // all registered appds, dird
    // shutdown all appds and the dird
  if (Dird_Query("datacutter.appd.*", cfg) >= 0) {
    const char *sbSection;
    char *sbHost;
    int wPort;
    sbSection = cfg.FirstSection();
    while (sbSection != NULL) {
      sbHost = Global.hi.ConnectNameTo_Alloc(cfg.Find(sbSection, "dns_name"));
      wPort = atoi(cfg.Find(sbSection, "port"));
      if (Appd_Shutdown(sbHost, wPort) < 0) {
	// appd is not really there, so cleanup dird entry
	MemoryBuf mbTmp(sbSection, (int)strlen(sbSection), (int)strlen(sbSection));
	Dird_Remove(mbTmp);
      }
      DeleteNull(sbHost);
      sbSection = cfg.NextSection();
    }
    //}
    // TODO: should sleep a bit, then query dird for number still running
    sleep(1);
    if (getenv("DATACUTTER_DIRDHOST")) {
      sbHost = Global.hi.ConnectNameTo_Alloc(getenv("DATACUTTER_DIRDHOST"));
    } else {
      sbHost = Global.hi.ConnectNameTo_Alloc(DIRD_DEFAULT_HOST);
    }
    if (getenv("DATACUTTER_DIRDPORT")) {
      wPort = atoi(getenv("DATACUTTER_DIRDPORT"));
    } else {
      wPort = DIRD_DEFAULT_PORT;
    }  
    wRtn = Dird_Shutdown(sbHost, wPort);
    if (wRtn == 0) {
      if (Global.fVerbose)
	cout << "shutdown dird" << endl;
    }
    delete [] sbHost;
    //dird_stop();
    //} else {  // just appds for this application
    //LoadAppConsoleCfg(sbFile, cfg);
    // we leave the dird running
    //wRtn = DoAllAppd(false, si, cfg);
  }
  return wRtn;
}

int hostinfo_name(char *sbHost_in, bool fDns) {
  char *sbHost = Global.hi.CanonicalName_Alloc(sbHost_in, fDns);
  cout << sbHost_in << " (" << sbHost << ')' << endl;
  DeleteNull(sbHost);
  return 0;
}
int hostinfo_connect(char *sbHostFrom_in, char *sbHostTo_in, bool fDns) {
  char *sbHostFrom = Global.hi.CanonicalName_Alloc(sbHostFrom_in, fDns);
  char *sbHostTo   = Global.hi.CanonicalName_Alloc(sbHostTo_in, fDns);
  char *sbConnect  = Global.hi.ConnectName(sbHostFrom, sbHostTo);
  cout << "from: " << sbHostFrom_in << " (" << sbHostFrom << ')' << endl
       << "to:   " << sbHostTo_in << " (" << sbHostTo << ')' << endl
       << "connect name= " << sbConnect << endl;
  DeleteNull(sbHostFrom);
  DeleteNull(sbHostTo);
  return 0;
}

void usage(char *sb = NULL, int wExit = -1) {
  if (sb) {
    cerr << ErrorHeader(wERROR) << sb << endl;
  }
  cerr << "usage: "  << sbProgramName
       << " [-sleep] [-debug] [-noxterm] [-verbose] <command>" << endl
       << "   options:  -sleep = block until all started daemons exit" << endl
       << "             -debug = show commands being executed" << endl
       << "             -noxterm = start appds in daemon mode" << endl
       << "   commands: all <start|stop> [console_cfg_file|-]" << endl
       << "             dird <start|stop|status>" << endl
       << "             appd <start|stop> {<host>...|-}" << endl
       << "             appd <status|uptime>" << endl
       << "             appd pstack {<host>...|-}" << endl
       << "             appd ftp {<host>...|-} -- <file>..." << endl
       << endl
       << "             dird add <section> <name> <value>" << endl
       << "             dird query <section> <name>" << endl
       << "             dird remove <section>" << endl
       << "             appd wait <host>..." << endl
       << "             appd waitn n" << endl
    //<< "             appd exec <cmd args...>" << endl
       << endl
       << "             hostinfo [-dns] name <host>..." << endl
       << "             hostinfo [-dns] connect <host_from> <host_to>" << endl
       << endl
       << "             netprobe <host1> <host2>" << endl
       << endl;
  if (wExit != -1) {
    (void) StandardCleanup();
    exit(wExit);
  }
}


int main(int argc, char *argv[]) {
  int wRtn = 0, ret;
  int iArg;

  (void) StandardInit(argc, argv);

  // process flags
  Global.fDebug = false;
  Global.fNoXterm = false;
  Global.fSleep = false;
  for (iArg=1; iArg<argc; ++iArg) {
    if (argv[iArg][0] != '-') break;  // no more flags
    if (strcmp(argv[iArg], "-debug") == 0) {
      Global.fDebug = true;
    } else if (strcmp(argv[iArg], "-noxterm") == 0) {
      Global.fNoXterm = true;
    } else if (strcmp(argv[iArg], "-sleep") == 0) {
      Global.fSleep = true;
    } else if (strcmp(argv[iArg], "-verbose") == 0) {
      Global.fVerbose = true;
    } else {
      usage("unknown flag", 1);
    }
  }

  Global.fOutputHandler = false;
  Global.sbOutputHandler = NULL;
  Global.ndaemonlist = 0;
  Global.wInitId = 0;
  Global.nInitNeeded = 0;

  if (argc < iArg+2) usage(NULL, 1);

  if (strcmp(argv[iArg], "all") == 0) {
    if (argc != iArg+2 && argc != iArg+3)
      usage("wrong number of arguments for all", 1);
    Global.hi.Init();
    if (strcmp(argv[iArg+1], "start") == 0) {
      wRtn = all_start((argc > iArg+1 ? argv[iArg+2] : NULL));
    } else if (strcmp(argv[iArg+1], "stop") == 0) {
      wRtn = all_stop((argc > iArg+1 ? argv[iArg+2] : NULL));
    } else {
      usage("unknown all subcommand", 1);
    }

  } else if (strcmp(argv[iArg], "dird") == 0) {
    if (argc < iArg+2) usage("wrong number of arguments for dird", 1);
    Global.hi.Init();
    if (strcmp(argv[iArg+1], "start") == 0) {
      wRtn = dird_start();
    } else if (strcmp(argv[iArg+1], "stop") == 0) {
      appd_stop("-");
      wRtn = dird_stop();
    } else if (strcmp(argv[iArg+1], "status") == 0) {
      wRtn = dird_status();
    } else if (strcmp(argv[iArg+1], "add") == 0) {
      if (argc < iArg+3) usage("wrong number of arguments for dird add", 1);
      wRtn = dird_add(argv[iArg+2], argc-(iArg+3), argv+(iArg+3));
    } else if (strcmp(argv[iArg+1], "query") == 0) {
      if (argc < iArg+4) usage("wrong number of arguments for dird query", 1);
      wRtn = dird_query(argv[iArg+2], argv[iArg+3]);
    } else if (strcmp(argv[iArg+1], "remove") == 0) {
      if (argc < iArg+3) usage("wrong number of arguments for dird remove", 1);
      wRtn = dird_remove(argv[iArg+2]);
    } else {
      usage("unknown dird subcommand", 1);
    }

  } else if (strcmp(argv[iArg], "appd") == 0) {
    Global.hi.Init();
    if (strcmp(argv[iArg+1], "start") == 0) {
      if (argc < iArg+3) usage("wrong number of arguments for appd", 1);
      for (int j=iArg+2; j<argc; ++j) {
	ret = appd_start(argv[j]);
	if (ret) wRtn = ret;
      }
    } else if (strcmp(argv[iArg+1], "stop") == 0) {
      if (argc < iArg+3) usage("wrong number of arguments for appd", 1);
      for (int j=iArg+2; j<argc; ++j) {
	ret = appd_stop(argv[j]);
	if (ret) wRtn = ret;
      }
    } else if (strcmp(argv[iArg+1], "wait") == 0) {
      if (argc < iArg+3) usage("wrong number of arguments for appd", 1);
      for (int j=iArg+2; j<argc; ++j) {
	ret = appd_wait(argv[j]);
	if (ret) wRtn = ret;
      }
    } else if (strcmp(argv[iArg+1], "waitn") == 0) {
      if (argc < iArg+3) usage("wrong number of arguments for appd", 1);
      int n = atoi(argv[iArg+2]);
      if (n <= 0) {
	usage("invalid value for n", 1);
      } else {
	wRtn = appd_waitn(n);
      }
    } else if (strcmp(argv[iArg+1], "status") == 0) {
      if (argc != iArg+2) usage("wrong number of arguments for appd", 1);
      appd_command_find_appds();
      wRtn = appd_command(eAppdCmdStatus, NULL);
    } else if (strcmp(argv[iArg+1], "uptime") == 0) {
      if (argc != iArg+2) usage("wrong number of arguments for appd", 1);
      appd_command_find_appds();
      wRtn = appd_command(eAppdCmdUptime, NULL);
    } else if (strcmp(argv[iArg+1], "pstack") == 0) {
      if (argc < iArg+3) usage("wrong number of arguments for appd", 1);
      appd_command_find_appds();
      // update the set of "marked" appds to consider for the command
      if (argc == iArg+2) {  // no appds listed -> use all
	appd_command_subset_init(true);
      } else {
	appd_command_subset_init(false);
	for (int j=iArg+2; j<argc; ++j) {
	  appd_command_subset_check(argv[j]);
	}
      }
#if 1
      cout << "DEBUG(pstack): found " << Global.daAppdCommand.NumElem()
	   << " appds" << endl;
      int tmp = 0;
      for (unsigned int i=0; i<Global.daAppdCommand.NumElem(); ++i) {
	appd_command_t *p = Global.daAppdCommand.Get(i);
	if (p->fMark) ++tmp;
      }
      cout << "DEBUG(pstack): of those appds, " << tmp
	   << " marked for command" << endl;
#endif
      // now execute the command
      wRtn = appd_command(eAppdCmdPstack, NULL);
    } else if (strcmp(argv[iArg+1], "ftp") == 0) {
      if (argc < iArg+3) usage("wrong number of arguments for appd", 1);
      appd_command_find_appds();
      int iFileArgs = iArg+2;
      for (int j=iArg+2; j<argc; ++j) {
	if (strcmp(argv[j], "--") == 0) {
	  iFileArgs = j+1;
	  break;
	}
      }
      // update the set of "marked" appds to consider for the command
      if (iFileArgs == iArg+2) {  // no appds listed -> use all
	appd_command_subset_init(true);
      } else {
	appd_command_subset_init(false);
	for (int j=iArg+2; j<iFileArgs; ++j) {
	  appd_command_subset_check(argv[j]);
	}
      }
#if 1
      cout << "DEBUG: found " << Global.daAppdCommand.NumElem()
	   << " appds" << endl;
      int tmp = 0;
      for (unsigned int i=0; i<Global.daAppdCommand.NumElem(); ++i) {
	appd_command_t *p = Global.daAppdCommand.Get(i);
	if (p->fMark) ++tmp;
      }
      cout << "DEBUG: of those appds, " << tmp
	   << " marked for command" << endl;
#endif
      // now execute the command for all given files
      for (int j=iFileArgs; j<argc; ++j) {
	ret = appd_command(eAppdCmdFtp, argv[j]);
	if (ret) wRtn = ret;
      }
    } else {
      usage("unknown appd subcommand", 1);
    }

  } else if (strcmp(argv[iArg], "hostinfo") == 0) {
    bool fDns = false;
    if (strcmp(argv[iArg+1], "-dns") == 0) {
      fDns = true;
      ++iArg;
    }
    if (argc < iArg+2)
      usage("no subcommand for hostinfo", 1);

    Global.hi.Init();
    if (strcmp(argv[iArg+1], "name") == 0) {
      if (argc < iArg+2)
	usage("wrong number of arguments for hostinfo name", 1);
      for (int j=iArg+2; j<argc; ++j) {
	ret = hostinfo_name(argv[j], fDns);
	if (ret) wRtn = ret;
      }
    } else if (strcmp(argv[iArg+1], "connect") == 0) {
      if (argc < iArg+3)
	usage("wrong number of arguments for hostinfo connect", 1);
      ret = hostinfo_connect(argv[iArg+2], argv[iArg+3], fDns);
      if (ret) wRtn = ret;
    } else {
      usage("unknown hostinfo subcommand", 1);
    }

  } else if (strcmp(argv[iArg], "netprobe") == 0) {  // TODO
    if (argc != iArg+3) usage("wrong number of arguments for netprobe", 1);
    // We attempt to contact appd processes on the two hosts, and then
    // send a netprobe packet to the first.  This appd will contact the
    // appd on the second host, and determine the application level
    // latency and bw characteristics of the path in between.    

  } else {
    usage("unknown command", 1);
  }

  while (CheckForInitAcks(true) > 0)
    ;

  if (Global.fSleep) {
    Exec e;
    int  status;
    cout << "Waiting for " << Global.ndaemonlist << " daemons to finish ..."
	 << endl;
    for (int i=0; i<Global.ndaemonlist; ++i) {
      status = e.Wait(Global.daemonlist[i].pid);
      if (status == -1) {
	cout << "errno = " << strerror(errno) << endl;
      }
      cout << "[" << i << "].pid=" << Global.daemonlist[i].pid
	   << " finished "
	   << Global.daemonlist[i].sbDesc
	   << " (status=" << status << ")" << endl;
    }
  }

  // cleanup
  for (int i=0; i<Global.ndaemonlist; ++i) {
    delete [] Global.daemonlist[i].sbDesc;
  }
  for (unsigned int j=0; j<Global.daAppdCommand.NumElem(); ++j) {
    appd_command_t *p = Global.daAppdCommand.Get(j);
    delete [] p->sbHost;
    delete p;
  }

  (void) StandardCleanup();
  cout.flush();
  cerr.flush();
  return (wRtn == 0 ? 0 : 1);
}


// ****************************************************************************
