// ****************************************************************************
//  Name: consoleprocessstate.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 08/11/2000 : mdb - Created.
//        11/20/2000 : mdb - Reimplemented stopinstance and stopremoteprocess.
//        11/29/2000 : mdb - Converted to use DC_Mutex/DC_Cond.
//        03/13/2001 : mdb - Sweeping changes to support multiple copies.
//        08/07/2001 : mdb - Added [console]filterexit_stats_file support.
//        12/10/2001 : mdb - New checks and handling for X11 remote display.
//        01/23/2002 : mdb - Increased hard-coded size of buffer used for
//                           starting remote process (from KB_64 to KB_512).
//  Desc: State and supporting routines for the console process.
// ****************************************************************************

#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <assert.h>

#ifdef _WIN32
#include <process.h>
#include "win32.h"
#endif

#include <iostream>
#include <fstream>
using namespace std;

#include "lib/dc_str.h"
#include "lib/dc_memory.h"
#include "lib/dc_exec.h"

#include "dc_filterservice.h"
#include "consoleprocessstate.h"
#include "packet.h"
#include "packet-tags.h"
#include "dird-client.h"
#include "appd.h"
#include "server-loop-base.h"
#include "util.h"
#include "thread-reader.h"
#include "remoteprocessstate.h"
#include "dc_instance.h"

#if !defined(MAXHOSTNAMELEN)
#  define MAXHOSTNAMELEN 256
#endif

#if defined(DCDBG)
//#define DCDBG_FILTERSERVICE_CONSOLE_STARTUP
//#define DCDBG_FILTERSERVICE_CONSOLEOUT
//#define DCDBG_FILTERSERVICE_APPENDWORK
//#define DCDBG_FILTERSERVICE_FILTEREXIT
//#define DCDBG_FILTERSERVICE_SHUTDOWN
#endif /* DCDBG */

//#define DEBUGGER_USE_GDB
//#define DEBUGGER_USE_DBX


/**
 * ctor
 */
DC_ConsoleProcessState::DC_ConsoleProcessState(DC_FilterService_t *pDC_in)
  : pDC(pDC_in), sbXauthMethod(NULL), sbXauthCookie(NULL) {
}

/**
 * dtor
 */
DC_ConsoleProcessState::~DC_ConsoleProcessState(void) {
  Types::uint i;

  for (i=0; i<daPlacementSpec.NumElem(); ++i) {
    if (daPlacementSpec[i]) delete daPlacementSpec[i];
  }
  daPlacementSpec.Flush();

  TestDelete(sbXauthMethod);
  TestDelete(sbXauthCookie);
}

int DC_ConsoleProcessState::init(void) {
  char *sbTmp;
  char bufTmp[80];

  char *sb;
  char *sbDirdHost;
  int  wDirdPort;

  wNextInstanceNum = 0;
  cInstanceRunning = 0;

  cRemoteProcesses = 0;

  pRemoteFilter = NULL;

  // setup hostinfo for resolving unreachable hostname combinations
  hi.Init();

  // CLH - moved forward to allow redirection of dird host/port
  // read console config info
  // order of search: <appname>.cfg, console.cfg
  snprintf(bufTmp, sizeof(bufTmp), "%s.cfg", pDC->sbAppName);
  pDC->cfgConsole.SetOpt(ConfigInfo_Section::ADD_DUPNAME_REPLACE);
  if (pDC->cfgConsole.ReadFile((sbTmp="console.cfg"), false) &&
      pDC->cfgConsole.ReadFile((sbTmp=SYSCONSOLECFG), false)) {
    // this is not a fatal error
    cout << "*** Note: No default console config file found." << endl;
  }
  pDC->cfgConsole.ReadFile((sbTmp=bufTmp), false);

  sbTmp = FindParm(pDC->cfgConsole, "console", "dird_host", false);
  if (sbTmp) {
    sbDirdHost = sbTmp;
  } else if ((sb = pDC->_ArgvNextExtract(pDC->pargc, pDC->pargv, "-datacutter-dird-host"))) {
    sbDirdHost = sb;
  } else if (getenv("DATACUTTER_DIRDHOST")) {
    sbDirdHost = getenv("DATACUTTER_DIRDHOST");
  } else {
    sbDirdHost = DIRD_DEFAULT_HOST;
  }
  sbTmp = FindParm(pDC->cfgConsole, "console", "dird_port", false);
  if (sbTmp) {
    wDirdPort = atoi(sbTmp);
  } else if ((sb = pDC->_ArgvNextExtract(pDC->pargc, pDC->pargv, "-datacutter-dird-port"))) {
    wDirdPort = atoi(sb);
  } else if (getenv("DATACUTTER_DIRDPORT")) {
    wDirdPort = atoi(getenv("DATACUTTER_DIRDPORT"));
  } else {
    wDirdPort = DIRD_DEFAULT_PORT;
  }

  // query dird for all appd nodes
  // BUGBUG: This can be large.  Lazily ask for only required ones?
  if (Dird_Query("datacutter.appd.*", cfgDir, sbDirdHost, wDirdPort) < 0) {
    cerr << "init(): Failed query of dird to get appd contact info.\n";
    return -1;
  }
  //cout << "Dird_Query (begin)...\n" << cfgDir << "Dird_Query (end)." << endl;
  if (cfgDir.NumSection() == 0) {
    cerr << "DC_ConsoleProcessState::init(): no contact info for appds found."
	 << endl
	 << "  - make sure appd process are started on required remote hosts."
	 << endl;
    return -1;
  }

  // CLH - old location at which console.cfg was read

  sbTmp = FindParm(pDC->cfgConsole, "console", "application_cfg", false);
  if (sbTmp != NULL) {
    // this is done later with filter instances
    cerr << "DEPRECATED [console]application_cfg = " << sbTmp << endl;
  }
  sbTmp = FindParm(pDC->cfgConsole, "console", "placement_cfg", false);
  if (sbTmp != NULL) {
    // this is done later with filter instances
    cerr << "DEPRECATED [console]placement_cfg = " << sbTmp << endl;
  }
#if 0  // we will look for these later when we actually need them
  sbHostNameForced = FindParm(pDC->cfgConsole,
			      "console", "hostname_forced", false);
  sbConsoleOutputDir = FindParm(pDC->cfgConsole,
				"console", "console_output_dir", false);
#endif
  sbFilterExitFmt = FindParm(pDC->cfgConsole,
			     "console", "filterexit_stats_file", false);

  eStreamWritePolicy = DC_StreamWritePolicy_RR;  // default
  sbTmp = FindParm(pDC->cfgConsole, "console", "stream_write_policy", false);
  if (sbTmp) {
    if (StrCmp(sbTmp, "RR", true) == 0)
      eStreamWritePolicy = DC_StreamWritePolicy_RR;
    else if (StrCmp(sbTmp, "WRR", true) == 0)
      eStreamWritePolicy = DC_StreamWritePolicy_WRR;
    else if (StrCmp(sbTmp, "DD", true) == 0)
      eStreamWritePolicy = DC_StreamWritePolicy_DD;
    else {
      cout << ErrorHeader(wERROR)
	   << "unknown [console]stream_write_policy = \"" << sbTmp << '\"'
	   << endl;
      assert(0);
    }
  }

  // check for x11 remote display setting
  sbTmp = FindParm(pDC->cfgConsole, "console", "remote_x11_display", false);
  char *sbDisplay = getenv("DISPLAY");
  if (sbTmp && !sbDisplay) {
    cout << ErrorHeader(wERROR)
	 << "[console]remote_x11_display specified, but no $DISPLAY set"
	 << endl;
    sbTmp = NULL;
  }
  if (!sbTmp || StrNCmp(sbTmp, "none", 4, true) == 0) {
    eRemoteX11Display = REMX11_NONE;
  } else if (StrNCmp(sbTmp, "display", 7, true) == 0) {
    eRemoteX11Display = REMX11_DISPLAY;
  } else if (StrNCmp(sbTmp, "xauth", 5, true) == 0) {

#ifndef _WIN32
	  
	Exec e;
    Argv a;

    Str strArgs(sbTmp);
    sbXauthMethod = strArgs.TokenGet(1);

    eRemoteX11Display = REMX11_XAUTH;
    if (sbXauthMethod &&
	StrCmp(sbXauthMethod, "list-add", true) == 0) {
      a.Append("xauth");
      a.Append("list");
      a.Append(sbDisplay);
      Str strOutput;
      strOutput.SetNoDup(e.ExecCollectOutput_Alloc(a));
      sbXauthCookie = strOutput.TokenGet(2);
    } else if (sbXauthMethod &&
	       StrCmp(sbXauthMethod, "nextract-nmerge", true) == 0) {
      a.Append("xauth");
      a.Append("nextract");
      a.Append("-");
      a.Append(sbDisplay);
      sbXauthCookie = e.ExecCollectOutput_Alloc(a);
    } else {
      cout << ErrorHeader(wERROR)
	   << "unknown/missing [console]remote_x11_display, xauth <method> argument."
	   << endl;
      assert(0);      
    }

#endif


    //cout << "DEBUG: xauth cookie " << sbXauthCookie << endl;
  } else {
    cout << ErrorHeader(wERROR)
	 << "unknown [console]remote_x11_display = \"" << sbTmp << '\"'
	 << endl;
    assert(0);      
  }

  return 0;
}


/**
 * start remote libs added to daPlacementSpec by a filter instance
 */
//  11/09/2000 : mdb - Forgot to use hi. to get best connection name.
//
// thread: console process main thread only, not concurrent
int DC_ConsoleProcessState::StartRemoteProcesses(void) {
  int wRtn = DC_ERR_OK;
  Types::uint i;
  char *sbDebugSrcDir, *sbDebugExtraCmd;
  Packet pkt;
  MemoryBuf mb(KB_512);
  char bufTmp[(MAXHOSTNAMELEN*2) + 32];
  
  #ifndef _WIN32
    char *sbTmp;
  #endif

  sbDebugSrcDir = FindParm(pDC->cfgConsole, "console", "debug_src_dir", false);
  sbDebugExtraCmd = FindParm(pDC->cfgConsole, "console", "debug_extra_cmd", false);

  icStartRemoteProcesses.reset();
  int nStart = 0;

  mutexPlacementSpec.lock();

  for (i=0; i<daPlacementSpec.NumElem(); ++i) {
    DC_PlacementSpec *p = daPlacementSpec[i];
    if (!p || p->eState != DC_PlacementSpec::INIT) continue;

    if (p->fConsole) { // special case console process with a remote process

      if (pDC->pRemote) continue;  // did this for an earlier instance

      // We imitate the normal steps (in DC_FilterService_t::init())
      // that use an appd to create a remote process.
      pDC->pRemote = new DC_RemoteProcessState(pDC);

      // This usually happens upon recv HELLO_ACK, but since we send
      // no HELLO packet to ourselves, it must still be done.
      pDC->pRemote->ProcessConsoleCfgSettings();

      snprintf(bufTmp, sizeof(bufTmp), "%d,%s,%d,%s", i,
	      pDC->addrLocal.getHostName(), pDC->addrLocal.getPort(),
	      pDC->addrLocal.getHostName());
      if (pDC->pRemote->init(bufTmp) != 0) {
	cerr << "StartRemoteProcess(): Failed setup of RemoteProcess."
	     << endl;
	// BUGBUG: cleanup?
	wRtn = DC_ERR_Error;
	break;
      }
      // add the socket used to send HELLO to reader server loop -
      // send to reader command Q
      pDC->pReader->ReaderCmd_AddConn(&pDC->pRemote->sockConsole,
				      &pDC->pRemote->addrConsole);

      ++nStart;
      continue;
    }

    // use the appd to start this remote process on the remote nodes
    // as needed, and have them contact us.
    char *sbHostConnectTo = hi.ConnectNameTo_Alloc(p->sbHostName);
    if (p->sockAppd.Connect(sbHostConnectTo, p->wPortAppd, p->addrAppd) != 0) {
      snprintf(bufTmp, sizeof(bufTmp), 
	      "StartRemoteProcesses(): failed Connect to appd on %s:%d\n",
	      sbHostConnectTo, p->wPortAppd);
      perror(bufTmp);
      delete [] sbHostConnectTo;
      wRtn = DC_ERR_Error;  // cannot connect to appd on required host
      break;
    } 
    delete [] sbHostConnectTo;

    // add appd to reader server loop - send to reader command Q
    pDC->pReader->ReaderCmd_AddConn(&p->sockAppd, &p->addrAppd);

#if defined(DCDBG_FILTERSERVICE_CONSOLE_STARTUP)
    cout << "Connected to appd " << p->addrAppd << endl;
#endif /* DCDBG_FILTERSERVICE_CONSOLE_STARTUP */

    // build appd command packet to execute ourselves
    char *sbHostConnectFrom = hi.ConnectNameFrom_Alloc(p->sbHostName);

    // build command info for remote machine
    mb.Empty();

    // 04/17/02 CLH - modified to handle filter applications in 
    // locations unrelated to the location of appd

#ifndef _WIN32

	char* home = getenv("HOME");
    char* pwd = getenv("PWD");
    char* temp = strstr(pwd, home);
    if (temp) {
      while(*temp == *home) {
	++home;
	++temp;
      }
      mb.sprintf("cwd=%s\n", ++temp);
    } else {
      mb.sprintf("cwd=%s\n", ".");
    }

    if (eRemoteX11Display != REMX11_NONE) {
      sbTmp = getenv("DISPLAY");
      if (sbTmp[0] == ':') {  // no host specified in display name?
	mb.sprintf("display=%s%s\n", sbHostConnectFrom, sbTmp);
      } else {
	mb.sprintf("display=%s\n", sbTmp);
      }
      if (eRemoteX11Display == REMX11_XAUTH) {
	mb.sprintf("xauth_method=%s\n", sbXauthMethod);
	if (sbXauthCookie) mb.sprintf("xauth_cookie=%s\n", sbXauthCookie);
      }
    }
#endif

    // remote envvar settings
    DArray<char*> *pda = pDC->cfgConsole.FindAll("console", "remote_setenv");
    for (Types::uint j=0; j<pda->NumElem(); ++j) {
      mb.sprintf("setenv=%s\n", pda->Get(j));
    }
    delete pda;
    
    if (sbDebugSrcDir) {
      // first create the debugger command file
      snprintf(bufTmp, sizeof(bufTmp), "%s_debugger-command.%ld.%d",
	      pDC->sbAppName, static_cast(long,getpid()), i);
      ofstream of(bufTmp);
      if (! of) {
	perror("init(): failed create debug command file");
      }
#if defined(DEBUGGER_USE_GDB)
      of << "dir " << sbDebugSrcDir << endl;
      // this passes through sigpipe, since datacutter handles it
      of << "handle SIGPIPE nostop pass" << endl;
#elif defined(DEBUGGER_USE_DBX)
      of << "use " << sbDebugSrcDir << endl;
#endif
      if (sbDebugExtraCmd) of << sbDebugExtraCmd << endl;
      of << "run";
      // include app cmdline args that were specified when we were run
      for (int k=1; k<*(pDC->pargc); k++) {
	of << ' ' << (*(pDC->pargv))[k];
      }
      of << " -datacutter-remotelib-args ";
      of << i << ',' << sbHostConnectFrom << ',' << pDC->addrLocal.getPort()
	 << ',' << p->sbHostName;
      of << endl;
#if 0  /* this fails with: 'Invalid type combination in equality test.' */
      of << "if $_exitvalue == 0\n  quit\nend" << endl;
#endif
      of.close();

      // now create the command to run debugger instead
#if defined(DEBUGGER_USE_GDB)
      mb.sprintf("cmd=%s\n", "gdb");
      mb.sprintf("arg=--command=%s_debugger-command.%ld.%d\n",
		 pDC->sbAppName, static_cast(long,getpid()), i);
      mb.sprintf("arg=--quiet\n");
#elif defined(DEBUGGER_USE_DBX)
      mb.sprintf("cmd=%s\n", "dbx");
      mb.sprintf("arg=-c\n");
      mb.sprintf("arg=%s_debugger-command.%d\n", pDC->sbAppName, i);
#endif
      mb.sprintf("arg=%s\n", (*(pDC->pargv))[0]);
      mb.sprintf("debug\n");
    } else {
      mb.sprintf("cmd=%s\n", (*(pDC->pargv))[0]);
      // include app cmdline args that were specified when we were run
      for (int k=1; k<*(pDC->pargc); k++) {
	mb.sprintf("arg=%s\n", (*(pDC->pargv))[k]);
      }
      mb.sprintf("arg=-datacutter-remotelib-args\n");
      mb.sprintf("arg=%d,%s,%d,%s\n", i,
		 sbHostConnectFrom, pDC->addrLocal.getPort(), p->sbHostName);
      //mb.Dump();
    }
    delete [] sbHostConnectFrom;

    pkt.set(TAG_APPD_EXEC, &mb);

    if (pkt.write(p->sockAppd) <= 0) {
      snprintf(bufTmp, sizeof(bufTmp), 
	      "StartRemoteProcesses(): failed write to appd on %s:%d\n",
	      p->sbHostName, p->wPortAppd);
      perror(bufTmp);
      p->sockAppd.close();
      wRtn = DC_ERR_Error;  // failed write of exec command to appd
      break;
    }
#if defined(DCDBG_FILTERSERVICE_CONSOLE_STARTUP)
    cout << "Sent exec command to appd on " << p->addrAppd << endl;
#endif /* DCDBG_FILTERSERVICE_CONSOLE_STARTUP */
    ++nStart;
  }

  mutexPlacementSpec.unlock();

  // even if we had an error, we still collect the ones we expect
  icStartRemoteProcesses.initiator(nStart);  // reader wakes us

  return wRtn;
}

/**
 * called when recv HELLO packet
 * rtn: int - 0==success, !0==failed of HELLO_ACK failed
 */
// thread: reader
int DC_ConsoleProcessState::StartRemoteProcesses_Reader(ConfigInfo &cfg,
							BaseSocket *psock) {
  int wRtn = 0;
  bool fConsole;

  icStartRemoteProcesses.collectorPart1();

  Types::uint j = atoi(cfg.Find(NULL, "iplacement"));

  mutexPlacementSpec.lock();

  daPlacementSpec[j]->eState = DC_PlacementSpec::STARTED;
  daPlacementSpec[j]->psock = psock;
  daPlacementSpec[j]->wPortLib = atoi(cfg.Find(NULL, "port"));

  fConsole = daPlacementSpec[j]->fConsole;

  if (!fConsole) {
    ++cRemoteProcesses;
#if defined(DCDBG_FILTERSERVICE_CONSOLE_STARTUP)
    cout << "Remote Processes running ++ to " << cRemoteProcesses << endl;
#endif /* DCDBG_FILTERSERVICE_CONSOLE_STARTUP */
  }

  mutexPlacementSpec.unlock();

#if defined(DCDBG_FILTERSERVICE_CONSOLE_STARTUP)
  cout << "Recv HELLO pkt from remote lib " << j << " at port "
       << daPlacementSpec[j]->wPortLib << endl;
#endif /* DCDBG_FILTERSERVICE_CONSOLE_STARTUP */

  // this must be done before the .collectorPart2() routine lets the
  // console main thread go to avoid a race with the first
  // STARTINSTANCE packet.
  if (!fConsole) {
    // only send HELLO_ACK if we are not colocated with the remote process
    MemoryBuf mb;
    // add console cfg to packet
    mb.New(pDC->cfgConsole.SizeBytes());
    pDC->cfgConsole.WriteBuf(mb.getPtr(), mb.getFree());
    mb.setSize(mb.getMax());
    Packet packetOut(TAG_FILTERSERVICE_HELLO_ACK, &mb);

    if (packetOut.write(*psock) <= 0) {
      cerr << "StartRemoteProcesses_Reader(): failed write of HELLO_ACK"
	   << endl;
      wRtn = -1;
    }
#if defined(DCDBG_FILTERSERVICE_CONSOLE_STARTUP)
    cout << "[reader] Sent HELLO_ACK pkt (" << mb.getSize() << " bytes)"
	 << endl;
    //cout << pDC->cfgConsole;
#endif /* DCDBG_FILTERSERVICE_CONSOLE_STARTUP */
  }

  icStartRemoteProcesses.collectorPart2();

  return wRtn;
}

/**
 * look up the DC_ConsoleWork item and adjust counts (may not be found
 * if work was appended with fWait==false)
 */
// thread: reader
void DC_ConsoleProcessState::AppendWorkDone_Reader(int wInstanceNum,
						   int wWorkNum,
						   int wExitStatus) {
  Types::uint i, j;
  mutexConsoleWork.lock();

  for (i=0; i<daConsoleWork.NumElem(); ++i) {
    DC_ConsoleWork *pwork = daConsoleWork[i];
    if (pwork &&
	pwork->wWorkNum == wWorkNum &&
	pwork->wInstanceNum == wInstanceNum) {
      ++pwork->cFilters;

      // CLH - addred for new filter return types
      if (wExitStatus != DC_ERR_OK) {
	if (pwork->wExitStatus == DC_ERR_OK) {
	  DC_PlacementSpec *p;
	  Packet pkt(TAG_FILTERSERVICE_APPENDWORK_CANCEL, NULL);
	  pkt.setExtra(DC_EncodeInt(wInstanceNum, wWorkNum));
	  for (j=0; j<daPlacementSpec.NumElem(); ++j) {
	    p = daPlacementSpec[j];
	    if (!p || p->fConsole) 
	      continue;  
	    if (!p->psock) 
	      continue;
	    if (pkt.write(*(p->psock)) <= 0) {
	      cout << "AppendWorkDone_Reader(): Error sending APPENDWORK_CANCEL pkt to "
		   << p->addrAppd.getHostName() << endl;
	    }
	  }	  
	}
	pwork->wExitStatus = wExitStatus;
      }

      if (pwork->cFilters == pwork->nFilters) {
#if defined(DCDBG_FILTERSERVICE_APPENDWORK)
	cout << "[reader] AppendWorkDone_Reader(): i"
	     << wInstanceNum << ",w" << wWorkNum << " DONE." << endl;
#endif /* DCDBG_FILTERSERVICE_APPENDWORK */
	pwork->eState = DC_ConsoleWork::DONE;
	// signal in case the main thread is sleeping in a WaitWork() call
	cvSomeWorkDone.signal();
	
#if defined(STATS_FILTERSERVICE_WORK)
	mWorkRespTime.NewValue(pwork->tRespTime.Elapsed());
#if defined(DCDBG_FILTERSERVICE_APPENDWORK)
	cout << "[reader] Response time of i"
	     << wInstanceNum << ",w" << wWorkNum << " = "
	     << pwork->tRespTime.Elapsed() << endl;
#endif /* DCDBG_FILTERSERVICE_APPENDWORK */
#endif /* STATS_FILTERSERVICE_WORK */
      }
      break;
    };
  }

  mutexConsoleWork.unlock();
}

// ****************************************************************************
//  Name: StopAllFilterInstances
//  Date: 09/14/2000 : mdb - Created.
// ****************************************************************************
/**
 * Provides a "controlled" shutdown of all filter instances, by sending
 * a special work packet that is enqueued as regular work, and when
 * dequeued by the filter thread before an init-process-finalize
 * sequence of calls, will cause the execution to act like an
 * EndOfFilter was returned by the last process() call.
 *
 * Return: int - DC_ERR_xxx
 */
// thread: console process main
int DC_ConsoleProcessState::StopAllFilterInstances(void) {
  // there is a race between the result of sending the STOP work
  // packet and the instance ending on its own.  We need to lock a
  // flag and set it, and this lock is also used by the reader thread
  // when accumulating the FILTEREXIT packets.
  DC_FilterInstance_t *pinstance;

  mutexFilterInstance.lock();

#if defined(DCDBG_FILTERSERVICE_SHUTDOWN)
  cout << "StopAllFilterInstances(): sending STOPINSTANCE to filter instances ..."
       << endl;
#endif /* DCDBG_FILTERSERVICE_SHUTDOWN */
  // check all current instances
  Types::uint i;
  for (i=0; i<daFilterInstance.NumElem(); ++i) {
    pinstance = daFilterInstance[i];
    if (pinstance) pinstance->_StopInstance();
  }

#if defined(DCDBG_FILTERSERVICE_SHUTDOWN)
  cout << "StopAllFilterInstances(): waiting for " << cInstanceRunning
       << " instances to finish ..." << endl;
#endif /* DCDBG_FILTERSERVICE_SHUTDOWN */

  // wait to be signalled when all are done
  while (cInstanceRunning > 0) {
    //cvAllFilterInstancesDone.wait(mutexFilterInstance);
    cvFilterInstanceDone.wait(mutexFilterInstance);
  }

  mutexFilterInstance.unlock();

#if defined(DCDBG_FILTERSERVICE_SHUTDOWN)
  cout << "StopAllFilterInstances(): all instances finished." << endl;
#endif /* DCDBG_FILTERSERVICE_SHUTDOWN */
  return DC_ERR_OK;
}


/**
 * process the filterexit info
 */
// thread: reader
int DC_ConsoleProcessState::FilterExit_Reader(int wRet, ConfigInfo &cfg) {
  const char *sbSection = cfg.FirstSection();
  int   wInstanceNum = atoi(cfg.Find(sbSection, "instance"));
  char *sbFilter = cfg.Find(sbSection, "filter");

  if (sbFilterExitFmt) {
    cfgFilterExit.Merge(cfg, (ConfigInfo::ADD_DUPSEC_APPEND |
			      ConfigInfo_Section::ADD_DUPNAME_APPEND));
  }

  if (wRet != DC_ERR_OK) {
    cout << "[reader] FilterExit i" << wInstanceNum << ',' << sbFilter
	 << " ret=" << wRet << ' ' << DC_strerror(wRet) << endl;
  }
#if defined(DCDBG_FILTERSERVICE_FILTEREXIT)
  else {
    cout << "[reader] FilterExit i" << wInstanceNum << ',' << sbFilter
	 << " ret=" << wRet << ' ' << DC_strerror(wRet) << endl;
  }
#endif /* DCDBG_FILTERSERVICE_FILTEREXIT */
  return 0;
}

/**
 * every remote process will send a single INSTANCEEXIT packet when
 * all copies of all filters for this instance have exited in that
 * remote process.
 */
// thread: reader
int DC_ConsoleProcessState::InstanceExit_Reader(int wInstanceNum) {
  Types::uint i, foundidx=0;

  // update the instance state
  mutexFilterInstance.lock();

  DC_FilterInstance_t *pinstance;
  bool fFound = false;
  bool fDone = false;
  //bool fDoneAll = false;
  for (i=0; i<daFilterInstance.NumElem(); ++i) {
    pinstance = daFilterInstance[i];
    if (pinstance && pinstance->wInstanceNum == wInstanceNum) {
      fFound = true;
      foundidx = i;
      pinstance->_InstanceExit_Reader();
      if (pinstance->eState == DC_FilterInstance_t::DONE) {
	fDone = true;
	//if (--cInstanceRunning == 0) fDoneAll = true;
	--cInstanceRunning;
#if defined(DCDBG_FILTERSERVICE_FILTEREXIT)
	cout << "Instances running -- to " << cInstanceRunning << endl;
#endif /* DCDBG_FILTERSERVICE_FILTEREXIT */
	daFilterInstance[foundidx] = NULL;
	delete pinstance;
	cvFilterInstanceDone.signal();
      }
      break;
    }
  }
  mutexFilterInstance.unlock();

  if (!fFound) {
    cout << "InstanceExit_Reader(): throwing away stray INSTANCEEXIT packet."
	 << endl;
    return -1;  // assume this was a stray, and ignore it
  }

  if (fDone) {
    // TODO: cleanup of state for this instance?
    //daFilterInstance[foundidx] = NULL;
    //delete pinstance;

    if (sbFilterExitFmt) {  // output filter stats
      // TODO: one file per instance, or overwrite like it is now?
      char bufTmp[128];
      snprintf(bufTmp, 128, sbFilterExitFmt, pDC->sbAppName, wInstanceNum);
      cout << "Note: output filterexit stats to \"" << bufTmp << '\"' << endl;
#if defined(STATS_FILTERSERVICE_WORK)
      char bufTmp2[80];
      // argc, argv
      snprintf(bufTmp2, sizeof(bufTmp2), "%d", *(pDC->pargc));
      cfgFilterExit.AddEntry("stats", "argc", bufTmp2);
      for (int i=0; i<*(pDC->pargc); ++i) {
	snprintf(bufTmp2, sizeof(bufTmp2), "arg%d", i);
	cfgFilterExit.AddEntry("stats", bufTmp2, (*(pDC->pargv))[i]);
      }
      // mWorkRespTime
      snprintf(bufTmp2, sizeof(bufTmp2), "%ld", mWorkRespTime.getNum());
      cfgFilterExit.AddEntry("stats", "mWorkRespTime.n", bufTmp2);
      snprintf(bufTmp2, sizeof(bufTmp2), "%f", mWorkRespTime.getMin());
      cfgFilterExit.AddEntry("stats", "mWorkRespTime.min", bufTmp2);
      snprintf(bufTmp2, sizeof(bufTmp2), "%f", mWorkRespTime.getMax());
      cfgFilterExit.AddEntry("stats", "mWorkRespTime.max", bufTmp2);
      snprintf(bufTmp2, sizeof(bufTmp2), "%f", mWorkRespTime.getSum());
      cfgFilterExit.AddEntry("stats", "mWorkRespTime.sum", bufTmp2);
      snprintf(bufTmp2, sizeof(bufTmp2), "%f", mWorkRespTime.getMean());
      cfgFilterExit.AddEntry("stats", "mWorkRespTime.mean", bufTmp2);
      snprintf(bufTmp2, sizeof(bufTmp2), "%f", mWorkRespTime.getStdDev());
      cfgFilterExit.AddEntry("stats", "mWorkRespTime.stddev", bufTmp2);
      snprintf(bufTmp2, sizeof(bufTmp2), "%f", mWorkRespTime.getVar());
      cfgFilterExit.AddEntry("stats", "mWorkRespTime.var", bufTmp2);
#endif
      cfgFilterExit.WriteFile(bufTmp);
      // 01/04/2002 : mdb - This causes multiple instances that output to the
      //                    same file to lose previous stats.
      //cfgFilterExit.Flush();
    } else {
      cout << cfgFilterExit;
    }
  }
  //if (fDoneAll) {
  //cvAllFilterInstancesDone.signal();
  //}
  return 0;
}


// ****************************************************************************
//  Name: StopRemoteProcesses
//  Date: 09/16/2000 : mdb - Created.
// ****************************************************************************
/**
 * Attempts a "controlled" shutdown of all started remote processes.
 * When the console process exits, anything left over will be killed
 * by the appd anyway.
 *
 * Return: int - DC_ERR_xxx
 */
// thread: console process main
int DC_ConsoleProcessState::StopRemoteProcesses(void) {
  // send READEREXIT packets to all remote processes.  if all other threads
  // have exited nicely, this will cause the remote process to exit.

  mutexPlacementSpec.lock();

#if defined(DCDBG_FILTERSERVICE_SHUTDOWN)
  cout << "StopRemoteProcesses(): sending READEREXIT to remote processes ..."
       << endl;
#endif /* DCDBG_FILTERSERVICE_SHUTDOWN */
  Types::uint i;
  DC_PlacementSpec *p;
  Packet pkt(TAG_FILTERSERVICE_READEREXIT, NULL);
  for (i=0; i<daPlacementSpec.NumElem(); ++i) {
    p = daPlacementSpec[i];
    if (!p || p->fConsole) continue;  // skip the console (do that later)
    // 12/10/2000 : mdb - we never successfully connected to this appd?
    if (!p->psock) continue;
    if (pkt.write(*(p->psock)) <= 0) {
      cout << "StopRemoteProcesses(): Error sending READEREXIT pkt to "
	   << p->addrAppd.getHostName() << endl;
    }
  }

#if defined(DCDBG_FILTERSERVICE_SHUTDOWN)
  cout << "StopRemoteProcesses(): waiting for " << cRemoteProcesses
       << " remote processes to exit ..." << endl;
#endif /* DCDBG_FILTERSERVICE_SHUTDOWN */

  // wait to be signalled when all are done
  while (cRemoteProcesses > 0) {
    cvAllRemoteProcessesDone.wait(mutexPlacementSpec);
  }

  mutexPlacementSpec.unlock();

#if defined(DCDBG_FILTERSERVICE_SHUTDOWN)
  cout << "StopRemoteProcesses(): all remote processes exited." << endl;
#endif /* DCDBG_FILTERSERVICE_SHUTDOWN */
  return 0;
}

// thread: reader
int DC_ConsoleProcessState::StopRemoteProcesses_Reader(void) {
  mutexPlacementSpec.lock();
  --cRemoteProcesses;
#if defined(DCDBG_FILTERSERVICE_SHUTDOWN)
  cout << "Remote Processes running -- to " << cRemoteProcesses << endl;
#endif /* DCDBG_FILTERSERVICE_SHUTDOWN */
  if (cRemoteProcesses == 0) cvAllRemoteProcessesDone.signal();
  mutexPlacementSpec.unlock();

  // BUGBUG: We should also check to see if any instances were running
  // on this node, and clean up the mess.
  return 0;
}

void DC_ConsoleProcessState::DumpStats(void) {
#if defined(STATS_FILTERSERVICE_WORK)
  cout << "mWorkRespTime= " << mWorkRespTime << endl;
#endif /* STATS_FILTERSERVICE_WORK */
}

ostream &operator<<(ostream &os, DC_PlacementSpec &spec) {
  return os << "Placement{"
	    << spec.sbHostName << ':'
	    << spec.wPortAppd << ','
	    << "eState="
	    << (spec.eState == DC_PlacementSpec::INIT ? "INIT" :
		spec.eState == DC_PlacementSpec::STARTED ? "STARTED" :
		"STOPPED" ) << ','
	    << "wPortLib=" << spec.wPortLib << ','
	    << '}';
}


// ****************************************************************************
