// -*- c++ -*-
// ****************************************************************************
//  Name: consoleprocessstate.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 08/11/2000 : mdb - Created.
//        11/29/2000 : mdb - Converted to use DC_Mutex/DC_Cond.
//  Desc: DataCutter Filter Service user library.
// ****************************************************************************

#if !defined(H_DC_CONSOLEPROCESSSTATE)
#define H_DC_CONSOLEPROCESSSTATE

#include "lib/dc_configinfo.h"
#include "lib/dc_mutex.h"
#include "lib/dc_cond.h"
#include "lib/dc_initcollect.h"
#include "hostinfo.h"
#include "dc_policy.h"

#define STATS_FILTERSERVICE_WORK
#if defined(STATS_FILTERSERVICE_WORK)
#  include "lib/dc_timer.h"
#  include "lib/dc_metric.h"
#endif /* STATS_FILTERSERVICE_WORK */

class DC_FilterInstance_t;  // fwd ref
class DC_RemoteFilterCopy;  // fwd ref from remoteprocessstate.h

/**
 * keeps track of a single Append'ed piece of work.  The console
 * process state contains a collection of these -- one per
 * AppendWork() call.  Exported work handles are indexes into an array
 * of these object in the ConsoleProcessState object.  If flagged as
 * detached during the AppendWork() call, this entry is not added.
 * Otherwise, this entry entry remains until the caller asks for its
 * exit status, at which point it is reaped.
 */
class DC_ConsoleWork {
  friend class DC_FilterInstance_t;
public:
  DC_Work_t work;
  int wWorkNum;
  int wInstanceNum;  ///< instance this was appended to

  /**
   * INIT - after ctor called
   * RUNNING - the work has been submitted to all filters in the instance
   * DONE - instance has finished the work
   * WAIT - wait has been entered
   */
  enum { INIT, RUNNING, DONE, WAIT } eState;

  /// state to accumulate APPENDWORK_ACK
  int nPlacements, cPlacements;
  /// state to accumulate APPENDWORK_DONE
  int nFilters, cFilters;

  // TODO: not implemented
  // this is set when eState becomes DONE (should never be < 0)
  // - 0 == all filters finished work
  // - 1 == at least one filter failed
  int wExitStatus;

#if defined(STATS_FILTERSERVICE_WORK)
  Timer tRespTime;
#endif /* STATS_FILTERSERVICE_WORK */

public:
  DC_ConsoleWork(void) : eState(INIT), wExitStatus(0) {};
  ~DC_ConsoleWork(void) {};
};


/**
 *  One remote process exists on a given node for a single
 *  application, regardless how many instances there are.  The console
 *  process state object contains a collection of these, and all instances
 *  for that node reference one of these objects.
 */
class DC_PlacementSpec {
  friend ostream &operator<<(ostream &os, DC_PlacementSpec &spec);
public:
  bool  fConsole;  ///< T==this placement is console process for pipe thru case
  char *sbHostName;  ///< this is the canonical name

  /// appd contact info (only for fConsole == F)
  int   wPortAppd;
  /// connection to the appd that will be used to spawn the remote process
  InetAddress addrAppd;
  ClientSocket sockAppd;

  /// remote lib state (always set)
  int   wPortLib;
  BaseSocket *psock;  ///< ptr to Reader array of clients
  /**
   * INIT - after ctor called
   * STARTED - the remote lib is running
   * STOPPED - the remote lib has been stopped (recv APPD_DONE packet)
   */
  enum { INIT, STARTED, STOPPED } eState;

  /// remote process ctor
  DC_PlacementSpec(char *sbHostName_in, int wPort_in)
    : fConsole(false),
      sbHostName(StrDup(sbHostName_in)), wPortAppd(wPort_in),
      eState(INIT) {};
  /// console special case ctor
  DC_PlacementSpec(char *sbHostName_in)
    : fConsole(true),
      sbHostName(StrDup(sbHostName_in)), wPortAppd(0),
      eState(INIT) {};
  ~DC_PlacementSpec(void) {
    TestDeleteNull(sbHostName);
  };
};


/// Represents the runtime state for the Console Process
class DC_ConsoleProcessState {
public:
  DC_FilterService_t *pDC;
  HostInfo hi;
  ConfigInfo cfgDir;

  int wNextInstanceNum;  // BUGBUG: int rollover?
  int cInstanceRunning;  ///< count of running instances
  DArray<DC_FilterInstance_t*> daFilterInstance;
  DC_Mutex mutexFilterInstance;
  DC_Cond  cvFilterInstanceDone;  ///< used for stopping instances

  DArray<DC_ConsoleWork*> daConsoleWork;
  DC_Mutex mutexConsoleWork;
  DC_Cond  cvSomeWorkDone;

  /// used for starting an instance with a pseudo-filter in the console process
  DC_RemoteFilterCopy *pRemoteFilter;
  DC_Mutex mutexRemoteFilter;
  DC_Cond  cvRemoteFilter;

  int cRemoteProcesses;  ///< count of running remote processes
  DArray<DC_PlacementSpec*> daPlacementSpec;
  DC_Mutex mutexPlacementSpec;
  DC_Cond  cvAllRemoteProcessesDone;

  DC_InitiatorCollector_t icStartRemoteProcesses;

  /// local console resources (not implemented in current version)
  int fdConsoleOutput;
  bool fConsoleOutput;

  char *sbFilterExitFmt;
  ConfigInfo cfgFilterExit;  ///< collect filterexit pkt stats for output

  DC_StreamWritePolicy_t eStreamWritePolicy;  ///< default policy

  enum { REMX11_NONE, REMX11_DISPLAY, REMX11_XAUTH } eRemoteX11Display;
  char *sbXauthMethod;
  char *sbXauthCookie;

#if defined(STATS_FILTERSERVICE_WORK)
  Metric mWorkRespTime;
#endif /* STATS_FILTERSERVICE_WORK */

public:
  DC_ConsoleProcessState(DC_FilterService_t *pDC_in);
  ~DC_ConsoleProcessState(void);
  int init(void);
  int StartRemoteProcesses(void);
  int StartRemoteProcesses_Reader(ConfigInfo &cfg, BaseSocket *psock);
  void AppendWorkDone_Reader(int wInstanceNum, int wWorkNum, int wExitStatus);
  int FilterExit_Reader(int wRet, ConfigInfo &cfg);
  int InstanceExit_Reader(int wInstanceNum);
  int StopAllFilterInstances(void);
  int StopRemoteProcesses(void);
  int StopRemoteProcesses_Reader(void);

  void DumpStats(void);
};


#endif /* H_DC_CONSOLEPROCESSSTATE */
// ****************************************************************************
