// -*- c++ -*-
// ****************************************************************************
//  Name: dc_buffer.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 07/26/2000 : mdb - Created.
//  Desc: DataCutter buffer class.
// ****************************************************************************
/** @file dc_buffer.h */

#if !defined(H_DC_BUFFER)
#define H_DC_BUFFER

#include "lib/dc_memory.h"

/// The buffer class
/**
 * A DC_Buffer_t is the unit of transfer for streams.  This class is
 * simply a container for a chunk of contiguous memory, which can be
 * statically or dynamically allocated.  These buffers have a maximum
 * size, and keep track of how much of that maximum is occupied.
 * Data that will fit into the unused portion of the buffer can be
 * appended.  Extraction occurs from the head of the buffer,
 * maintaining a extraction pointer for multiple extraction
 * operations.  An application can also just get a pointer and length
 * of the data in the buffer, and use it in place.
 *
 * When finished with a buffer, the consume() method should be
 * called, which may free the space.  Heap allocated memory is
 * handled correctly based on the constructor used.  Generally
 * avoiding static DC_Buffer_t objects with the use of consume() is
 * advised, unless setConsume() with the value true is explicitly called.
*/
class DC_Buffer_t : public MemoryBuf {
  friend ostream &operator<<(ostream &os, DC_Buffer_t &buf);
  friend class DC_PipeInStream_t;
  friend class DC_PipeOutStream_t;
  friend class DC_ReaderServerLoop;
  friend class DC_StreamBufferQueue;
  friend class DC_RemoteOutStreamCopy;
private:
  enum { DATA=0, EOS, EOW } eType;
  bool fConsume;

public:
  /// Optionally used by application (ex: list of buffers)
  DC_Buffer_t *pNext;

public:
  DC_Buffer_t(void) : eType(DATA), fConsume(true) {};
  DC_Buffer_t(int wMax)
    : MemoryBuf(wMax), eType(DATA), fConsume(true) {};
  DC_Buffer_t(char *pBuf, int wMax, int wSize)
    : MemoryBuf(pBuf, wMax, wSize), eType(DATA), fConsume(false) {};
  DC_Buffer_t(const DC_Buffer_t &buf);

  //DC_Buffer_t(const MemoryBuf &rhs);
  //DC_Buffer_t &operator=(const MemoryBuf &rhs);  // handled by base class?

  /// Release a buffer
  /**
   * Called on a buffer returned from read that a filter will not
   * need any longer.  (Depending how memory was allocated for the 
   * buffer data area, this may or may not free the actual buffer space.)
   */
#if 1
  void consume(void) { if (fConsume) delete this; };
#else
  void consume(void) {
    size=-1;
    max=-1;
    pBuf=0;
    if (fConsume) delete this;
  };
#endif
  /// Get the consume state
  bool getConsume(void) { return fConsume; };
  ///  Set the consume state
  void setConsume(bool fConsume_in) { fConsume=fConsume_in; };
  // Note: use of fConsume mimics (exactly?) fFree in MemoryBuf base class?!?

#if 0  // These fcns are all available from the base class
  /// Allocate space in a buffer
  char *New(Types::uint wMax_in);
  /// Free the space allocated to a buffer
  void Delete(void);
  void Set(char *pBuf_in, int wMax_in, int wSize_in, bool fFree_in);
  /// Clear the contents of a buffer
  void Empty(void);
  /// Get the total size of a buffer
  Types::uint getMax(void);
  /// Get a pointer to the beginning of the buffer
  char *getPtr(void);
  /// Get the amount of used space
  Types::uint getSize(void);
  /// Get the amount of free space remaining
  Types::uint getFree(void);
  /// Get a pointer to the beginning of the free space
  char *getPtrFree(void);

  char *getPtrExtract(void);
  /// Get the amount of data left to extract
  Types::uint getSizeExtract(void);
  /// Reset the pointer 
  Types::uint resetExtract(Types::uint w=0);
  /// Set the size of space used
  Types::uint setSize(Types::uint wSize_in);
  /// Set the size of the space used
  int setSize(Types::uint size);
  /// Append data
  char *Append(char *p, Types::uint size);
  /// Append data
  char *Append(<type>);
  /// Use a format string to print to a buffer
  int sprintf(char *format, ...);

  int Extract(char *p, Types::uint size);

  int Extract(*<type>);
#endif /* 0 */
};

#endif /* H_DC_BUFFER */
