// ****************************************************************************
//  Name: dc_filterservice.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 08/05/1999 : mdb - Created.
//        08/20/1999 : mdb - Added info send to remote lib at start.
//                         - Remote lib connects to each other.
//        08/21/1999 : mdb - Starting threads for each filter.
//        08/23/1999 : mdb - Fixed bug where connecting to peers was broken.
//                         - Reader thread added, and read/write to streams.
//        08/31/1999 : mdb - write_nocopy() now does a consume() of the buffer
//                           when not colocated with the sink.  added 
//                           write_nocopy_noconsume() for when this is wrong.
//        09/10/1999 : mdb - Fixed stream fEOS semantics to be:
//                           1. set after read an EOS buf from the stream, or
//                           2. if the sink is not colocated, we set this on
//                              the source when writing the EOS buf.
//                         - Fixed race of starting filter threads both on
//                           a single node, and between remote nodes.  It uses
//                           a new barrier to the console and a thread barrier.
//        09/15/1999 : mdb - Added thread cancellation handling to reader.
//                         - Fixed write() to not consume after write_nocopy().
//        09/16/1999 : mdb - Changed application schema parsing.
//        10/18/1999 : mdb - Ported to xlC and AIX.
//        10/23/1999 : mdb - Added CONFIG_USE_DCE for pthreads xlC_r4 diffs.
//        12/13/1999 : mdb - init() filter fcn now uses initarg_t.
//        01/19/2000 : mdb - Fixed reader thread cancellation problem that
//                           caused deadlock on Linux.  Not using cancel now.
//        01/20/2000 : mdb - ConsoleOutput() outputs to an optional log file
//                           if [console]console_output_dir is set.
//        03/14/2000 : mdb - Automatic stats generation and send to console.
//        03/29/2000 : mdb - Added CloseOutputWaitInput() fcn for cleaner app
//                           exit w/o sig_pipe's.  App can instead close any
//                           way it wants before process() exits to avoid this
//                           default behavior.
//        04/07/2000 : mdb - Allow close of input streams, and shutdown queue
//                           which empties queue and the Reader drops pkts if
//                           not colocated.  Rewrote various parts to deal with
//                           this gracefully.
//        04/08/2000 : mdb - Split portions into thread-reader.cpp and
//                           thread-filter.cpp
//        08/11/2000 : mdb - Major structural and semantic rewrite: DC api,
//                           unit of work, console instances, changed main().
//        11/29/2000 : mdb - Converted to use DC_Mutex/DC_Cond.
//        03/13/2001 : mdb - Sweeping changes to support multiple copies.
//  Desc: DataCutter Filter Service user library.
// ****************************************************************************

#ifdef _WIN32
#include "win32.h"
#include <iostream>

using namespace std;
#endif

#include "lib/dc_str.h"
#include "lib/dc_memory.h"
#include "lib/dc_threadclass.h"
#include "lib/dc_barrier.h"

#include "dc_filterservice.h"
#include "consoleprocessstate.h"
#include "remoteprocessstate.h"
#include "packet.h"
#include "packet-tags.h"
#include "dird.h"
#include "appd.h"
#include "server-loop-base.h"
#include "dc_instance.h"

#include "thread-reader.h"
#include "thread-filter.h"

#include <stdarg.h>


/* UVC:  startinstance bug:
#define FILTER_SERVICE_MAX_Q_ITEMS   (20)
#define FILTER_SERVICE_MAX_CLIENTS  (256)
*/
#define FILTER_SERVICE_MAX_Q_ITEMS   KB_2
#define FILTER_SERVICE_MAX_CLIENTS   KB_4


#if defined(DCDBG)
//#define DCDBG_FILTERSERVICE_PROCESS_SPECS
//#define DCDBG_FILTERSERVICE_READER_THREAD
//#define DCDBG_FILTERSERVICE_NEWINSTANCE
//#define DCDBG_FILTERSERVICE_WORK
//#define DCDBG_FILTERSERVICE_FINDAPPCONFIG
//#define DCDBG_FILTERSERVICE_SHUTDOWN
#endif /* DCDBG */

extern char **environ;

PerfCallback_f g_fcnPerfCallback;

// Network Performance TODOs:
// 1. create dummy sockets between remote nodes
// 2. time latency between remote nodes, and compute bw*delay
// 3. use bw*delay to set send and recv window sizes (os specific)
// 4. then create the sockets with the new window size setting
// OR use an auto-tune tcp lib

#ifndef _WIN32
static void sig_pipe(int, siginfo_t *, void *) {
#if 0 /* cout is not async safe */
  cout << "\n\a\a ***************************************"
       << ErrorHeader(wWARNING) << "sig_pipe(): Handler called ..." << endl
       << "\a\a ***************************************" << endl;
#endif
}
#endif

// ****************************************************************************
//  Name: ArgvNextExtract
//  Date: 08/05/1999 : mdb - Created.
//        03/19/2000 : mdb - Fixed bug if multiple instances of flag.
// ****************************************************************************
/**
 *  Remove a flag and remove/return the next argument from argv[].
 *
 *  Inputs: pargc, pargv - command line args
 *  Return: char* - !NULL=found ok, NULL=not found
 */
char *DC_FilterService_t::_ArgvNextExtract(int *pargc, char ***pargv,
					   char *sbFlag) {
    char *sbRtn=NULL;
    int i, j;
    bool fFound = false;
    
    /* i - the next one to look at
     * j - the next spot in result
     */
    for (i=0,j=0; i<*pargc; ++i) 
    {
	if (!fFound && strcmp(sbFlag, (*pargv)[i]) == 0) {  /* found? */
	    fFound = true;
	    if (i+1 < *pargc) {  /* is there a following arg? */
		sbRtn = (*pargv)[i+1];
		i++;  /* skip the next arg */
	    }
	} 
	else {
	    (*pargv)[j] = (*pargv)[i];
	    j++;  /* leave in result argv */
	}
    }
    *pargc -= (i-j);  /* reduce argc by how many skipped */
     
    return sbRtn;
}


char *DC_FilterService_t::FindParm(ConfigInfo &cfg,
				   char *sbSection, char *sbName,
				   bool fExitNotFound) 
{
    char *sbRtn;
    if ((sbRtn = cfg.Find(sbSection, sbName)) == NULL) {
	if (fExitNotFound) 
	{
	    cerr << ErrorHeader(wERROR) << "DC_FilterService_t::FindParm(): "
		 << "no \"[" << sbSection << "] " << sbName << "\" specified."
		 << endl;
	    exit(1);
	}
    }
    return sbRtn;
}


// ****************************************************************************
//  Name: DC_FilterService_t (ctor/dtor)
//  Date: 08/05/1999 : mdb - Created.
// ****************************************************************************
DC_FilterService_t::DC_FilterService_t(void)
    : sbAppName(NULL), pargc(NULL), pargv(NULL),
      fcnFilterFactory(NULL), fcnPolicyFactory(NULL),
      pConsole(NULL), pRemote(NULL), pReader(NULL) 
{
}

DC_FilterService_t::~DC_FilterService_t(void) 
{
    // make sure all running instances are stopped cleanly and all work finished
    if (pConsole) {
	pConsole->DumpStats();
	
#if defined(DCDBG_FILTERSERVICE_SHUTDOWN)
	cout << "~DC_FilterService_t(): Shutting down filtering service ..." << endl;
#endif /* DCDBG_FILTERSERVICE_SHUTDOWN */
	
	// BUGBUG: deathstar is causing some shutdown problem ...
//#define DCDBG_DISABLE_SHUTDOWN
	
#if !defined(DCDBG_DISABLE_SHUTDOWN)
	pConsole->StopAllFilterInstances();
#endif /* !DCDBG_DISABLE_SHUTDOWN */
#if defined(DCDBG_FILTERSERVICE_SHUTDOWN)
	cout << "~DC_FilterService_t(): All filter instances stopped." << endl;
#endif /* DCDBG_FILTERSERVICE_SHUTDOWN */
	
#ifdef _WIN32
	Sleep(10);  // Hack to fix hanging on win32 machines
#endif
	// BUGBUG: this causes intermittent hangs, so it is disabled for now.
#if !defined(DCDBG_DISABLE_SHUTDOWN)
	pConsole->StopRemoteProcesses();
#endif /* !DCDBG_DISABLE_SHUTDOWN */
#if defined(DCDBG_FILTERSERVICE_SHUTDOWN)
	cout << "~DC_FilterService_t(): All remote processes stopped." << endl;
#endif /* DCDBG_FILTERSERVICE_SHUTDOWN */
    }
    
    // signal local reader to exit
    Packet pkt(TAG_FILTERSERVICE_READEREXIT, NULL);
    pkt.write(sockReader);
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
    cout << "Joining reader thread ..." << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
    pReader->thread_join(NULL);
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
    cout << "Done with reader thread." << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
    
    // dealloc
    if (pConsole) delete pConsole;
    if (pRemote) delete pRemote;
    if (pReader) delete pReader;
    if (sbAppName) delete [] sbAppName;
    (void) StandardCleanup();
    
    cout << "DataCutter Filter Service exit." << endl;
    cout.flush();
}


// ****************************************************************************
//  Name: init
//  Date: 08/05/1999 : mdb - Created.
//        11/07/2000 : mdb - Allow comma seperated -datacutter-cfg args.
// ****************************************************************************
/**
 *  Initializes system and determines if we are the console.
 *
 *  Inputs: sbAppName_in - name of the application
 *          fcnFilterFactory_in - user fcn ptr that creates filter objects
 *          pargc, pargv - command line args
 *  Return: int - 0=ok, !0=error
 */
int DC_FilterService_t::init(char *sbAppName_in,
			     FilterFactory_f fcnFilterFactory_in,
			     int *pargc_in, char ***pargv_in,
                             PerfCallback_f fcnPerfCallback_in) {
  char *sbVersion = "DataCutter Filter Service library v2.1 (07/02/2002)";
  char *sb;

  (void) StandardInit(*pargc_in, *pargv_in);
  cout << sbVersion << endl;

#ifndef _WIN32
  DBG_init((*pargv_in)[0], environ);
  DBG_InstallCoreSigHandlers();
#endif

  if (sbAppName_in) {
    sbAppName = StrDup(sbAppName_in);
  } else {
    // NULL passed in for appname, use the basename of the binary that was run
    if ((sb = strrchr(*pargv_in[0], '/'))) {
      sbAppName = StrDup(++sb);
    } else {
      sbAppName = StrDup(*pargv_in[0]);
    }
  }

  if ((fcnFilterFactory = fcnFilterFactory_in) == NULL) {
    cerr << "DC_FilterService_t::init(): fcnFilterFactory cannot be NULL"
	 << endl;
    exit(1);
  }
  g_fcnPerfCallback = fcnPerfCallback_in;
  
  if (pargc_in) pargc = pargc_in;
  if (pargv_in) pargv = pargv_in;
  if (! pargc || ! pargv) {
    cerr << "DC_FilterService_t::init(): cannot use NULL for &argc or &argv."
	 << endl;
    exit(1);
  }

  fCalledRun = false;

#ifndef _WIN32
  // catch SIGPIPE to avoid runtime abort
  struct sigaction s;
  s.sa_sigaction = sig_pipe;
  sigemptyset(&s.sa_mask);
  s.sa_flags = SA_SIGINFO;
  if (sigaction(SIGPIPE, &s, NULL) != 0) {
    cerr << "DC_FilterService_t::init(): Cannot install SIGPIPE handler: "
	 << strerror(errno) << endl;
  }
#endif

  // setup the server socket for eventual responses
  bool fOk = false;
  if (addrLocal.Set(0) == 0) {  // use any free port
    // use HostInfo instead
    //if (sbHostNameForced) addrLocal.setHostName(sbHostNameForced, true);
    if (sockLocal.CreateBindListen(addrLocal) == 0) {
      // CLH - commented out until I get a verbose flag setup
      //cout << "pid=" << getpid()
      //<< ", listening on " << addrLocal << endl;
      fOk = true;
    }
  }
  if (!fOk) {
    cerr << ErrorHeader(wERROR) << "DC_FilterService_t::init():"
	 << "Failed setup of local port.\n";
    return -1;
  }

  // reader server loop
  pReader = new DC_ReaderServerLoop(this);
  pReader->ReaderCmdQ.init("ReaderCmdQ", FILTER_SERVICE_MAX_Q_ITEMS);
  pReader->Init(sockLocal, FILTER_SERVICE_MAX_CLIENTS);

  // look for the special arg passed to remote libs by the console
  if ((sb = _ArgvNextExtract(pargc, pargv, "-datacutter-remotelib-args"))) {
    // we are a remote lib - extract the args
    pRemote = new DC_RemoteProcessState(this);

    // this sends the HELLO packet to the console process
    if (pRemote->init(sb) != 0) {
      cerr << "DC_FilterService_t::init(): Failed setup of RemoteProcess."
	   << endl;
      return DC_ERR_Error;
    }

    // expose to reader the socket used for HELLO/HELLO_ACK with console
    pReader->AddExternalConn(&pRemote->sockConsole, &pRemote->addrConsole);
  } else {
    // is the console
    pConsole = new DC_ConsoleProcessState(this);

    // look for command line args before we fill cfgConsole
    while ((sb = _ArgvNextExtract(pargc, pargv, "-datacutter-cfg"))) {
      cout << "Found -datacutter-cfg arg \"" << sb << "\"" << endl;
      // replace comma with a newline (avoid escaped commas)
      char *sbTmp = StrDup(sb);
      for (char *p=sbTmp; *p; ++p) {
	if (*p == ',' && (p==sbTmp || *(p-1) != '\\')) *p = '\n';
      }
      cfgConsole.ReadStr(sbTmp, false);
      delete [] sbTmp;
      cout << cfgConsole << "--" << endl;
    }

    if (pConsole->init() != 0) {
      cerr << "DC_FilterService_t::init(): Failed setup of ConsoleProcess."
	   << endl;
      return DC_ERR_Error;
    }
  }

  // start reader
  int ret = pReader->thread_create(NULL, NULL, true);
  if (ret != 0) {
    cout << "DC_FilterService_t::init(): "
	 << "failed reader thread creation (" << ret << ")" << endl;
    return DC_ERR_Error;  // major failure
  }

#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
  cout << "Created thread for reader ..." << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */

  // create socket connection to reader
  if (sockReader.Connect("localhost", addrLocal.getPort(),
			 addrReader, false) != 0) {  // reader thread
    perror("DC_FilterService(): init(): failed Connect() to local reader");
    cerr << addrLocal << endl;
    return DC_ERR_Error;
  }

  return DC_ERR_OK;
}

void DC_FilterService_t::setPolicyFactory(PolicyFactory_f fcnPolicyFactory_in) {
  fcnPolicyFactory = fcnPolicyFactory_in;
}

/**
 * called by remote process main thread
 */
int DC_FilterService_t::RemoteProcess(void) {
  if (!pRemote) return DC_ERR_NotRemoteProcess;
  // we want to block until the reader thread finishes
  // any commands to start instances will be handled by the reader
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
  cout << "DC_FilterService_t::RemoteProcess(): "
       << "main joining on reader ..." << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
  int wRtn;
  if ((wRtn = pReader->thread_join(NULL)) != 0) {
    cout << "DC_FilterService_t::RemoteProcess(): "
	 << "failed reader thread join (" << wRtn << ")" << endl;
    return DC_ERR_Error;
  }
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
  cout << "DC_FilterService_t::RemoteProcess(): "
       << "reader done, main continues ..." << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
  return DC_ERR_OK;
}

int DC_FilterService_t::PlacementPlanning(DC_FilterLayout_t &layout,
					  DC_Work_t &work,
					  DC_Placement_t &placement) {
  if (!pConsole) return DC_ERR_NotConsoleProcess;
  return -1;  // TODO
}

int DC_FilterService_t::ReuseFilterInstance(DC_FilterLayout_t &layout,
					    DC_Work_t &work,
					    DC_Placement_t &placement,
					    DC_FilterInstance_t *pinstance) {
  if (!pConsole) return DC_ERR_NotConsoleProcess;
  return -1;  // TODO
}


// ****************************************************************************
//  Name: NewFilterInstance
//  Date: 08/10/2000 : mdb - Created.
// ****************************************************************************
/**
 *  Creates a new filter instance context for use by the application.
 *
 *  Inputs: layout - describes set of filters and constraints to instantiate
 *          work - uow to add to the new instance
 *          placement - describes where filters should be started
 *  Output: pinstance - ptr to a new instance context to use
 *  Return: int - DC_ERR_OK=success, DC_ERR_xxx=failure
 */
int DC_FilterService_t::NewFilterInstance(DC_FilterLayout_t &layout,
					  DC_Work_t &work,
					  DC_Placement_t &placement,
					  DC_FilterInstance_t *&pinstance) {
  if (!pConsole) return DC_ERR_NotConsoleProcess;

  // ASSUME: the only thread that will ever call this function is the
  // console process main thread.  This function is not thread-safe.
  // The interaction and synchronization between the console process
  // main thread and the console process reader thread depends on
  // these assumptions for correctness.  Note that all other reader
  // functions should work ok concurrently with this new instance
  // creation.

#if defined(DCDBG_FILTERSERVICE_NEWINSTANCE)
  cout << "NewFilterInstance(): enter ..." << endl;
#endif /* DCDBG_FILTERSERVICE_NEWINSTANCE */

  // just replace passed in pointer value
  pinstance = new DC_FilterInstance_t;

  // look for vacated spot first, if none append
  Types::uint i;
  bool fFound = false;
  for (i=0; i<pConsole->daFilterInstance.NumElem(); ++i) {
    if (! pConsole->daFilterInstance[i]) { fFound = true; break; };
  }
  if (fFound) {
    pConsole->daFilterInstance.Set(i, pinstance);
  } else {
    i = pConsole->daFilterInstance.Append(pinstance);
  }

  // make sure there is a remote lib process on each needed host
  pinstance->_Reset(this, pConsole);
  if (pinstance->_ProcessLayout(layout) != 0) return DC_ERR_InvalidLayout;
  // the _ProcessPlacement() call also pre-fills pConsole->daPlacementSpec
  if (pinstance->_ProcessPlacement(placement)) return DC_ERR_InvalidPlacement;
  if (pConsole->StartRemoteProcesses()) return DC_ERR_AppdCommFailed;
  if (pinstance->_StartInstance()) return DC_ERR_RemoteProcessCommFailed;

  pConsole->mutexFilterInstance.lock();
  ++pConsole->cInstanceRunning;
#if defined(DCDBG_FILTERSERVICE_NEWINSTANCE)
  cout << "Instances running ++ to " << pConsole->cInstanceRunning << endl;
#endif /* DCDBG_FILTERSERVICE_NEWINSTANCE */
  pConsole->mutexFilterInstance.unlock();

#if defined(DCDBG_FILTERSERVICE_NEWINSTANCE)
  cout << "NewFilterInstance(): exit." << endl;
#endif /* DCDBG_FILTERSERVICE_NEWINSTANCE */
  return DC_ERR_OK;
}

// ****************************************************************************
//  Name: GetFilterInstance
//  Date: 09/10/2000 : mdb - Created.
// ****************************************************************************
/**
 *  Convenience routine to first try to reuse an existing filter
 *  instance, and if that fails, create a new one.
 *
 *  Inputs: layout - describes set of filters and constraints to instantiate
 *          work - uow to add to the new instance
 *          placement - describes where filters should be started
 *  Output: pinstance - ptr to context of the instance to use
 *  Return: int - DC_ERR_OK=success, DC_ERR_xxx=failure
 */
int DC_FilterService_t::GetFilterInstance(DC_FilterLayout_t &layout,
					  DC_Work_t &work,
					  DC_Placement_t &placement,
					  DC_FilterInstance_t *&pinstance) {
  if (!pConsole) return DC_ERR_NotConsoleProcess;
  if (!pinstance ||
      ReuseFilterInstance(layout, work, placement, pinstance) != 0) {
    // create a new one, lose old ptr pinstance
    int ret = NewFilterInstance(layout, work, placement, pinstance);
    if (ret != 0) return ret;  // failed
  }
  return DC_ERR_OK;
}

// ****************************************************************************
//  Name: StopFilterInstance
//  Date: 09/22/2000 : mdb - Created.
// ****************************************************************************
/**
 *  Attempts a controlled shutdown of the given filter instance.
 *
 *  Inputs: pinstance - ptr to context of the instance to use
 *          fWait     - true=wait for instance to stop, false=async stop
 *  Output: pinstance - NULL
 *  Return: int - DC_ERR_OK=success, DC_ERR_xxx=failure
 */
int DC_FilterService_t::StopFilterInstance(DC_FilterInstance_t *&pinstance,
					   bool fWait) {

  if (!pConsole) return DC_ERR_NotConsoleProcess;
  if (!pinstance) return DC_ERR_Error;  // invalid pinstance arg

  // the stop one instance case is called by the application.
  unsigned int i;
  pConsole->mutexFilterInstance.lock();
  for (i = 0; i < pConsole->daFilterInstance.NumElem(); ++i) {
    if (pinstance == pConsole->daFilterInstance[i])
      break;
  }
  int wRtn = pinstance->_StopInstance();
  //if (fWait) pinstance->_WaitUntilInstanceDone();
  if (fWait) {
    while (pConsole->daFilterInstance[i] != NULL)
      pConsole->cvFilterInstanceDone.wait(pConsole->mutexFilterInstance);
  }
  pConsole->mutexFilterInstance.unlock();
  if (wRtn != DC_ERR_OK) return wRtn;
  if (pinstance->_isDetached()) {
    // we only do this for the non-(<console> pseudo filter) case
    pinstance = NULL;  // cleanup will happen later
  }
  return wRtn;
}

// ****************************************************************************
//  Name: ProbeWork
//  Desc: Check the status of the given piece of work
//  Date: 04/24/2002 : mgray - Created.
//
//  Inputs: wh - valid work handle from an AppendWork() call
//  Return: int - <0 == DC_ERR_xxx, >=0 = exit status
//          status - DC_ConsoleWork::eState (INIT, RUNNING, DONE, WAIT)
// ****************************************************************************
// thread: console process main
int DC_FilterService_t::ProbeWork(DC_WorkHandle_t wh_in, int *status) {
  if (!pConsole) return DC_ERR_NotConsoleProcess;
  pConsole->mutexConsoleWork.lock();

  unsigned int wh = static_cast(unsigned int,wh_in);

  // validate
  if (wh > pConsole->daConsoleWork.NumElem() ||
      ! pConsole->daConsoleWork[wh] ||
      pConsole->daConsoleWork[wh]->eState == DC_ConsoleWork::INIT) {
    pConsole->mutexConsoleWork.unlock();
    return DC_ERR_InvalidWorkHandle;
  }

  DC_ConsoleWork *pwork = pConsole->daConsoleWork[wh];

  *status = pwork->eState;
  pConsole->mutexConsoleWork.unlock();
  return DC_ERR_OK;
}




// ****************************************************************************
//  Name: WaitWork
//  Date: 09/12/2000 : mdb - Created.
// ****************************************************************************
/**
 *  Block until the given piece of work is finish, reap the entry, and
 *  return the exit status.
 *
 *  Inputs: wh - valid work handle from an AppendWork() call
 *  Return: int - <0 == DC_ERR_xxx, >=0 = exit status
 */
// thread: console process main
int DC_FilterService_t::WaitWork(DC_WorkHandle_t wh_in) {
  if (!pConsole) return DC_ERR_NotConsoleProcess;
  pConsole->mutexConsoleWork.lock();

  unsigned int wh = static_cast(unsigned int,wh_in);

  // validate
  if (wh > pConsole->daConsoleWork.NumElem() ||
      ! pConsole->daConsoleWork[wh] ||
      pConsole->daConsoleWork[wh]->eState == DC_ConsoleWork::INIT) {
    pConsole->mutexConsoleWork.unlock();
    return DC_ERR_InvalidWorkHandle;
  }

  DC_ConsoleWork *pwork = pConsole->daConsoleWork[wh];

  // sleep if still running
  while (pwork->eState == DC_ConsoleWork::RUNNING) {
#if defined(DCDBG_FILTERSERVICE_WORK)
    cout << "Work " << pwork->wWorkNum
         << " for instance " << pwork->wInstanceNum
	 << " still RUNNING ... sleep" << endl;
#endif /* DCDBG_FILTERSERVICE_WORK */
    pConsole->cvSomeWorkDone.wait(pConsole->mutexConsoleWork);
  }
  // tag this work as "waited for", to avoid finding it in WaitAnyWork()
  pwork->eState = DC_ConsoleWork::WAIT;

  int wRtn = pwork->wExitStatus;
  // reap this entry
  delete pwork;
  pConsole->daConsoleWork[wh] = NULL;

  pConsole->mutexConsoleWork.unlock();
  return wRtn;
}

// ****************************************************************************
//  Name: WaitAnyWork
//  Date: 09/13/2000 : mdb - Created.
//        11/06/2000 : mdb - Fixed failed mutex unlock if DC_ERR_NoWorkFound.
// ****************************************************************************
/**
 *  Block until any work is finished, and return this work handle.
 *
 *  Return: int - <0 == DC_ERR_xxx, >=0 = work handle to use in ExitStatus()
 */
// thread: console process main
DC_WorkHandle_t DC_FilterService_t::WaitAnyWork(void) {
  if (!pConsole) return DC_ERR_NotConsoleProcess;
  pConsole->mutexConsoleWork.lock();

  // validate - make sure anything to wait for
  Types::uint i;
  DC_ConsoleWork *pwork;
  bool fFound = false;
  for (i=0; i<pConsole->daConsoleWork.NumElem(); ++i) {
    pwork = pConsole->daConsoleWork[i];
    if (! pwork) continue;

    if (pwork->eState == DC_ConsoleWork::DONE) {
      // found one without waiting
      pConsole->mutexConsoleWork.unlock();
      return i;
    } else if (pwork->eState == DC_ConsoleWork::RUNNING) {
      fFound = true;
    }
  }
  if (! fFound) {
    pConsole->mutexConsoleWork.unlock();
    return DC_ERR_NoWorkFound;
  }

  // there is work running -- sleep
  while (1) {
#if defined(DCDBG_FILTERSERVICE_WORK)
    cout << "WaitAnyWork: sleeping ..." << endl;
#endif /* DCDBG_FILTERSERVICE_WORK */
    pConsole->cvSomeWorkDone.wait(pConsole->mutexConsoleWork);
    // wake up - check big predicate
    fFound = false;
    for (i=0; i<pConsole->daConsoleWork.NumElem(); ++i) {
      pwork = pConsole->daConsoleWork[i];
      if (! pwork) continue;
      if (pwork->eState == DC_ConsoleWork::DONE) { fFound = true; break; }
    }
    if (fFound) break;
  }
  // tag this work as "waited for", to avoid finding it in WaitAnyWork()
  pwork->eState = DC_ConsoleWork::WAIT;

  pConsole->mutexConsoleWork.unlock();
  return i;
}

// ****************************************************************************
//  Name: ExitStatus
//  Date: 09/13/2000 : mdb - Created.
// ****************************************************************************
/**
 *  Reap the work entry, and return the exit status.
 *
 *  Inputs: wh - valid work handle from an AppendWork() or WaitAnyWork() call
 *  Return: int - <0 == DC_ERR_xxx, >=0 = exit status
 */
// thread: console process main
int DC_FilterService_t::ExitStatus(DC_WorkHandle_t wh_in) {
  if (!pConsole) return DC_ERR_NotConsoleProcess;
  int wRtn = DC_ERR_OK;
  pConsole->mutexConsoleWork.lock();
  unsigned int wh = static_cast(unsigned int,wh_in);
  if (wh > pConsole->daConsoleWork.NumElem() ||
      ! pConsole->daConsoleWork[wh] ||
      pConsole->daConsoleWork[wh]->eState == DC_ConsoleWork::INIT) {
    wRtn = DC_ERR_InvalidWorkHandle;
  } else if (pConsole->daConsoleWork[wh]->eState == DC_ConsoleWork::RUNNING) {
    wRtn = DC_ERR_WorkInProgress;
  } else {
    DC_ConsoleWork *pwork = pConsole->daConsoleWork[wh];
    wRtn = pwork->wExitStatus;
    // reap this entry
    delete pwork;
    pConsole->daConsoleWork[wh] = NULL;
  }
  pConsole->mutexConsoleWork.unlock();
  return wRtn;
}


void DC_FilterService_t::FilterLock(void) {
  if (pRemote) pRemote->_FilterLock(&pRemote->mutexFilterLock);
}
void DC_FilterService_t::FilterUnlock(void) {
  if (pRemote) pRemote->_FilterUnlock(&pRemote->mutexFilterLock);
}


#if 0  // TODO: need to come back to this once everything else is settled
// ****************************************************************************
//  Name: ConsolePrintf
//  Date: 08/05/1999 : mdb - Created.
// ****************************************************************************
/**
 *  Sends text to the console process.
 *
 *  Inputs: sbFormat, ... - printf args
 *  Return: int -
 */
int DC_FilterService_t::ConsolePrintf(char *sbFormat, ...) {
  int wRtn;
  _FilterLock(&mutexFilterInternal); // shared bufTmp
  va_list ap;
  va_start(ap, sbFormat);
  vsprintf(bufTmp, sbFormat, ap);
  va_end(ap);

  MemoryBuf buf;
  buf.Set(bufTmp, strlen(bufTmp), strlen(bufTmp), false);
  wRtn = ConsoleOutput(&buf, false);
  _FilterUnlock(&mutexFilterInternal);
  return wRtn;
}

int DC_FilterService_t::ConsoleOutput(MemoryBuf *pbuf, bool fLock) {
  int wRtn = 0;
  if (fLock) _FilterLock(&mutexFilterInternal);
  if (pConsole != NULL) {
    if (fdConsoleOutput == -1) {  // output to screen
      cout.write(pbuf->getPtr(), pbuf->getSize()); 
    } else {  // output to log file
      if (pbuf->write(fdConsoleOutput) != static_cast(int,pbuf->getSize())) {
	cout << "DC_FilterService_t::ConsoleOutput(): "
	     << "Failed write of output to console log file" << endl;
      } else {
	fConsoleOutput = true;
      }
    }
  } else {
    Packet pkt(TAG_FILTERSERVICE_CONSOLEOUT, pbuf);
    Types::uint iFilter = fCalledRun
      ? _FindFilterByTid(pthread_self()) : Types::uintMax;
    pkt.setExtra(iFilter); // so we can tell which filter is writing
    // BUGBUG: This extra iFilter info is never used
    if (pkt.write(sockConsole) <= 0) {
      cout << "DC_FilterService_t::ConsoleOutput(): "
	   << "Failed write of output to console process" << endl;
      wRtn = 1;
    }
  }
  if (fLock) _FilterUnlock(&mutexFilterInternal);
  return wRtn;
}
#endif /* 0 */

// ****************************************************************************
//  Name: FindAppConfig, FindAppConfigInt, FindAppConfigDbl
//  Date: 01/07/2000 : mdb - Created.
//        11/08/2000 : mdb - Added a filter level lock since Find operation
//                           is not thread-safe due to cycle detection.
// ****************************************************************************
/**
 *  Searches in the console config info for an entry in the seciton
 *  with the applications name (as specified in the init() call).
 *
 *  Inputs: sbName - entry to find
 *          wId    - sub section value (ie: [AppName.wId])
 *                   -1 = just use the AppName as the section name
 *          sbFilterName - NULL=don't use, else [AppName.FilterName]
 *                                           or [AppName.FilterName.wId]
 *  Return: char * - NULL=not found, !NULL=value (must be freed by caller!)
 */
char *DC_FilterService_t::FindAppConfig(const char *sbName,
					const char *sbFilterName,
					int wId) {
  char sbSection[80];
  if (sbFilterName) {
    if (wId == -1) {
      snprintf(sbSection, 80, "%s.%s", sbAppName, sbFilterName);
    } else {
      snprintf(sbSection, 80, "%s.%s.%d", sbAppName, sbFilterName, wId);
    }
  } else {
    if (wId == -1) {
      snprintf(sbSection, 80, "%s", sbAppName);
    } else {
      snprintf(sbSection, 80, "%s.%d", sbAppName, wId);
    }
  }
#if defined(DCDBG_FILTERSERVICE_FINDAPPCONFIG)
  //cout << cfgConsole << endl;
#endif /* DCDBG_FILTERSERVICE_FINDAPPCONFIG */
  if (pRemote) pRemote->_FilterLock(&pRemote->mutexFilterLock_Internal);
  char *sbFind = cfgConsole.FindDup(sbSection, sbName);
  if (pRemote) pRemote->_FilterUnlock(&pRemote->mutexFilterLock_Internal);
#if defined(DCDBG_FILTERSERVICE_FINDAPPCONFIG)
  cout << "FindAppConfig(): [" << sbSection << "]" << sbName << "= " << sbFind
       << endl;
#endif /* DCDBG_FILTERSERVICE_FINDAPPCONFIG */
  return sbFind;
}

long DC_FilterService_t::FindAppConfigInt(const char *sbName,
					  const char *sbFilterName,
					  int wId, bool fInterpretUnits) {
  char *sbAlloc = FindAppConfig(sbName, sbFilterName, wId);
  if (!sbAlloc) return 0;
  long wRtn = fInterpretUnits ? Atoi(sbAlloc) : atoi(sbAlloc);
  delete [] sbAlloc;
  return wRtn;
}

double DC_FilterService_t::FindAppConfigDbl(const char *sbName,
					    const char *sbFilterName,
					    int wId) {
  char *sbAlloc = FindAppConfig(sbName, sbFilterName, wId);
  if (!sbAlloc) return 0.0;
  double rRtn = atof(sbAlloc);
  delete [] sbAlloc;
  return rRtn;
}


// ****************************************************************************
