// -*- c++ -*-
// ****************************************************************************
//  Name: dc_filterservice.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 08/05/1999 : mdb - Created.
//        01/10/2002 : mdb - arg_t subclasses initarg_t to avoid duplication.
//        01/23/2002 : mdb - Misc cleanups suggested by Umit.
//  Desc: DataCutter Filter Service user library.
// ****************************************************************************
/** @file dc_filterservice.h */

#if !defined(H_DC_FILTERSERVICE)
#define H_DC_FILTERSERVICE

#include <pthread.h>

#include "lib/dc_standard.h"
#include "lib/dc_argv.h"
#include "lib/dc_configinfo.h"
#include "lib/dc_socket.h"
#include "lib/dc_memory.h"
#include "lib/dc_map.h"
#include "lib/dc_timer.h"
#include "lib/dc_metric.h"
#include "lib/dc_mutex.h"

#include "dc_error.h"
#include "dc_buffer.h"
#include "dc_filterlayout.h"
#include "dc_work.h"
#include "dc_placement.h"

// Forward Declarations
class DC_FilterService_t; 
class DC_PipeInStream_t;
class DC_PipeOutStream_t;
// From consoleprocessstate.h
class DC_ConsoleProcessState;
// From remoteprocessstate.h
class DC_RemoteInStreamCopy;
class DC_RemoteOutStreamCopy;
class DC_RemoteFilterCopy;
class DC_RemoteProcessState;
// From thread-reader.h
class DC_ReaderServerLoop;
// From dc_instance.h
class DC_FilterInstance_t;


typedef void (*PerfCallback_f)(int inst, const char *sbFilterName,
                               int rank, double *sa, long *sal);
extern PerfCallback_f g_fcnPerfCallback;


// Supported return values for user-defined interface methods
/*
 * These return values allow the filter to communicate with
 * the runtime system.  
 */
typedef enum {
  DC_RTN_OK,     // Put EOW on stream, return DC_ERR_OK on WaitWork
  DC_RTN_CONT,   // Put EOW on stream, return DC_ERR_WorkFail on WaitWork
  // CLH - There's no way to kill off a filter and allow it's cohorts to continue
  // without a signifigant code overhaul, so this is on hold for the moment.
  //DC_RTN_TERM,   // Put EOS on stream, return DC_ERR_RemoteFilterFail on WaitWork, kill filter
  DC_RTN_ABRT    // Put EOS on stream, return DC_ERR_InstanceFail on WaitWork, kill group instance
} DC_RTN_t;

/// Base filter class
/**
 * Applications should subclass this abstract base class for every filter,
 * then provide definitions for the virtual functions init, process, and
 * finalize.
 */
class DC_Filter_Base_t {
public:
  DC_FilterService_t *pDC; ///< The service which instantiates the filter in a remote process
  const char *sbFilterName;  ///< The name of the filter

  /// Argument passed to the init function containing a work description
  /**
   * A filter group is a logical description of interconnected
   * filters.  At runtime, an instance of the filter group is
   * created, which may optionally contain more than one transparent
   * copies of any filter.  These transparent copies execute
   * collectively in parallel to process work within the filter
   * group instance.  Within a filter group instance, all the
   * transparent copies of a particular filter on a single host is
   * called a copy set.  The rank and max functions allow the filter to
   * determine its rank in this two level heirarchy, which can be
   * useful for read style filters that process disk resident data
   * to partition the data among transparent copies of itself.
   */
  class initarg_t {
    friend class DC_RemoteFilterCopy;
  protected:
    DC_RemoteFilterCopy *pFilter;
  public:
    int         argc;
    char      **argv;
    const char *sbFilterName, *sbFullName, *sbCopyName;
    DC_Work_t  *pwork;

    ~initarg_t(void);

    /// Our rank within the single copyset on the current local host
    int rankWithinCopySet(void);
    /// The number of transparent copies on the current local host
    int maxWithinCopySet(void);
    /// Total copyset info for this filter group instance
    int rankCopySets(void);

    int maxCopySets(void);
    /// Total filter transparent copy info for this filter group instance
    int rankCopies(void);

    int maxCopies(void);
  };

  /// Argument passed to the process function containing input and output streams
  class arg_t : public initarg_t {
    friend class DC_RemoteProcessState;
    friend class DC_FilterInstance_t;
    friend class DC_RemoteFilterCopy;
  protected:
    int _init(DC_RemoteFilterCopy *pFilter_in, DC_FilterService_t *pDC);
    int _resetInsEOW(void);
    int DrainInput(int wWorkNum);

  public:
    arg_t(void);
    ~arg_t(void);

    int                 nins;
    DC_PipeInStream_t  *ins;
    int                 nouts;
    DC_PipeOutStream_t *outs;

    /// Returns an index to ins based on the stream name provided, or -1 if not found
    int insIndex(char *sbName);  
    /// Returns an index to outs based on the stream name provided, or -1 if not found
    int outsIndex(char *sbName); 
    /// Returns an input stream based on the stream name provided, or NULL if not found
    DC_PipeInStream_t  *insLookup(char *sbName);  
    /// Returns an output stream based on the stream name provided, or NULL if not found
    DC_PipeOutStream_t *outsLookup(char *sbName); 
    /// Blocks until a buffer is available for any input stream
    DC_Buffer_t *insReadAny(int *piFound = NULL, bool *rgStreamSet = NULL);
    /// Provides the number of streams in ins that already hit EOW
    int insNumEOW(void);
    /// Returns true if all the streams in ins are at EOW
    bool insAllEOW(void);
    /// Closes all outgoing streams and waits for all input streams to see EOS
    int CloseOutputWaitInput(int wWorkNum, bool fEndOfWork=false);
  };

public:
  DC_Filter_Base_t(void) : pDC(NULL), sbFilterName(NULL) {};
  // 11/29/2001 : mdb - (Umit found) sbFilterName ptr is shared, don't delete
  virtual ~DC_Filter_Base_t(void) {};

  /// Searches the console config file for the section [AppName.FilterName]
  // (do not call in ctor -- pDC is NULL until first init() call)
  char *FindAppFilterConfig(const char *sbName, int wId = -1);
  long FindAppFilterConfigInt(const char *sbName, 
			      int wId = -1,
			      bool fInterpretUnits = true);
  double FindAppFilterConfigDbl(const char *sbName, int wId = -1);

  /// Filter service provided method
  void FilterLock(void);
  /// Filter service provided method
  void FilterUnlock(void);

  /// User-defined interface method
  virtual DC_RTN_t init(initarg_t &arg) = 0;
  /// User-defined interface method
  virtual DC_RTN_t process(arg_t &arg) = 0;
  /// User-defined interface method
  virtual DC_RTN_t finalize(void) = 0;
};

/// Base stream policy
/**
 * Applications should subclass this abstract base class to customize
 * a stream's write policy when there are transparent copies.
 */
class DC_Policy_Base_t {
public:
  /**
   * Inputs: wMax     - total number of transparent copies stream writes to
   *         sbStream - name of output stream
   *         buf      - actual buffer written by the filter
   *         pUserArg - user defined argument (from write() call)
   * Return: int - 0..wMax-1 copyset to write to
   */
  virtual int write_copyset(int wMax, char *sbStream,
			    DC_Buffer_t *pbuf, void *pUserArg) {
    return 0;
  };
};

/// Input stream
/**
 * This is passed into filter objects for accessing the streams, and
 * are local to each filter thread (other than the pointers to
 * potentially shared (colocated) DC_Remote objects).
 */
class DC_PipeInStream_t {
  friend class DC_RemoteProcessState;
  friend class DC_RemoteFilterCopy;
  friend class DC_Filter_Base_t::arg_t;
private:
  DC_RemoteFilterCopy   *pFilter;
  DC_RemoteInStreamCopy *pInStream;
  bool fEOW;  // T==EOW buffer was read by user from stream
  bool fEOS;  // T==EOS buffer was read by user from stream
  DC_FilterService_t *pDC;

  DC_Buffer_t *read_u(void);

public:
  char *getName(void);
  bool isEndOfWork(void) { return fEOW; };
  bool isEndOfStream(void) { return fEOS; };

  DC_Buffer_t *read(void);
  int close(void);
};

/// Output stream
/**
 * This is passed into filter objects for accessing the streams, and
 * are local to each filter thread (other than the pointers to
 * potentially shared (colocated) DC_Remote objects).
 *
 * write() does a deep copy of the buffer object and memory region as
 *   needed to allow the caller to modify the buffer immediately after
 *   this call returns.  For example, a colocated sink will cause the
 *   buffer object and memory region to be duplicated and placed directly
 *   in the consumer's queue.
 *
 * pUserArg is used for when there is a user-defined stream write
 * policy in use.  This extra arg is passed to the
 * policy.write_copyset() callback to help decide which copyset the
 * buffer should be directed to.  A typical example would be to more
 * simply pass along metadata, such as spatial location information,
 * along with a data chunk to avoid needing to decode the whole data
 * buffer to decide where it should be sent.
 *
 * write_nocopy() uses the given buffer.  If colocated with the sink
 *   this will be deposited directly in that Queue.  Use of stack
 *   allocated DC_Buffer_t objects for this call is not recommended
 *   due to undesirable results if the stack frame is removed before
 *   the object is dequeued/used by a colocated consumer.
 *
 * pUserArg is used for when there is a user-defined stream write
 * policy in use.  This extra arg is passed to the
 * policy.write_copyset() callback to help decide which copyset the
 * buffer should be directed to.  A typical example would be to more
 * simply pass along metadata, such as spatial location information,
 * along with a data chunk to avoid needing to decode the whole data
 * buffer to decide where it should be sent.
 */
class DC_PipeOutStream_t {
  friend class DC_RemoteProcessState;
  friend class DC_RemoteFilterCopy;
  friend class DC_Filter_Base_t::arg_t;
private:
  DC_RemoteFilterCopy    *pFilter;
  DC_RemoteOutStreamCopy *pOutStream;
  bool fEOW;  ///< T==EOW buffer was written by user to stream
  bool fEOS;  ///< T==EOS buffer was written by user to stream
  DC_FilterService_t *pDC;

public:
  char *getName(void);
  bool isEndOfWork(void) { return fEOW; };
  bool isEndOfStream(void) { return fEOS; };

  int write(DC_Buffer_t *pbuf, void *pUserArg=NULL);
  inline int write(DC_Buffer_t &buf, void *pUserArg=NULL) {
    buf.setConsume(false);  // assume a static buf since not a ptr
    return write(&buf, pUserArg);
  }
  int write_nocopy(DC_Buffer_t *pbuf, void *pUserArg=NULL);
  /**
   * Be VERY careful using this version with local vars on the
   * stack, since it could be needed long after the write call
   * if we are colocated with the consumer.
   */
  inline int write_nocopy(DC_Buffer_t &buf, void *pUserArg=NULL) {
    buf.setConsume(false);  // assume a static buf since not a ptr
    return write_nocopy(&buf, pUserArg);
  }
  int close(bool fEndOfWork=false);
};




/// Main controlling class
/**
 * Applications should instantiate one of these.
 */
class DC_FilterService_t {
  friend class DC_ConsoleProcessState;
  friend class DC_FilterInstance_t;
  friend class DC_RemoteProcessState;
  friend class DC_RemoteFilterCopy;
  friend class DC_ReaderServerLoop;
  friend class DC_PipeOutStream_t;
  friend class DC_PipeInStream_t;
public:
  typedef DC_Filter_Base_t *(*FilterFactory_f)(char *);
  typedef DC_Policy_Base_t *(*PolicyFactory_f)(char *);

protected:
  char *sbAppName;
  int *pargc;
  char ***pargv;
  FilterFactory_f fcnFilterFactory;
  PolicyFactory_f fcnPolicyFactory;
    
    
  DC_ConsoleProcessState *pConsole;
  DC_RemoteProcessState  *pRemote;
  DC_ReaderServerLoop    *pReader;

  // common reader state
  InetAddress  addrReader;  ///< For localhost communication with reader
  ClientSocket sockReader;

  InetAddress  addrLocal;  ///< Listen socket
  ServerSocket sockLocal;

  /// Currently used for ConsoleOutput/ConsolePrintf and >1 filter fd writes and FindAppConfigXXX()
  DC_Mutex mutexFilterInternal;

  /// Runtime state
  bool fCalledRun;

private:
  char *_ArgvNextExtract(int *pargc, char ***pargv, char *sbFlag);
  char *FindParm(ConfigInfo &cfg, char *sbSection, char *sbName,
		 bool fExitNotFound=true);
  int _ProcessSpecs(ConfigInfo &cfg);
  int _ProcessPeerLibSpecs(ConfigInfo &cfg);
  int _ConsoleNode(void);
  Types::uint find_filter(char *sbName);
  Types::uint _FindFilterByTid(pthread_t tid);

  /**
   * console process: fills from local console.cfg file
   * remote process: fills from HELLO_ACK packet
   * main/filters: access through api calls for the [<appname>] sections
   */
  ConfigInfo cfgConsole;

public:
  DC_FilterService_t(void);
  ~DC_FilterService_t(void);

  int init(char *sbAppName_in, 
	   FilterFactory_f fcnFilterFactory_in,
	   int *pargc, 
	   char ***pargv,
           PerfCallback_f fcnPerfCallback_in = NULL);

  void setPolicyFactory(PolicyFactory_f fcnPolicyFactory_in);

  bool isRemoteProcess(void) { return (pConsole == NULL); };

  int RemoteProcess(void);

  char *getAppName(void) const { return sbAppName; };

  int PlacementPlanning(DC_FilterLayout_t &layout, DC_Work_t &work,
			DC_Placement_t &placement);

  int ReuseFilterInstance(DC_FilterLayout_t &layout, DC_Work_t &work,
			  DC_Placement_t &placement,
			  DC_FilterInstance_t *pinstance);
  int NewFilterInstance(DC_FilterLayout_t &layout, DC_Work_t &work,
			DC_Placement_t &placement,
			DC_FilterInstance_t *&pinstance);
  int GetFilterInstance(DC_FilterLayout_t &layout, DC_Work_t &work,
			DC_Placement_t &placement,
			DC_FilterInstance_t *&pinstance);
  int StopFilterInstance(DC_FilterInstance_t *&pinstance, bool fWait=true);

    
  // check status of a piece of work
  int ProbeWork(DC_WorkHandle_t wh, int *status);
    
  /// Wait for a unit of work to finish
  int WaitWork(DC_WorkHandle_t wh);

  //int WaitAllWork(void); // wait for all appended work to finish
  //int WaitAllWork(DC_FilterInstance_t &*pinstance);

  /// Wait for any appended work to finish
  DC_WorkHandle_t WaitAnyWork(void);

  //DC_WorkHandle_t DC.WaitAnyWork(DC_FilterInstance_t &*pinstance);

  int ExitStatus(DC_WorkHandle_t wh);

  void FilterLock(void);

  void FilterUnlock(void);

#if 0 // TODO: come back and re-implement this
  int ConsolePrintf(char *sbFormat, ...);
  int ConsoleOutput(MemoryBuf *pbuf, bool fLock=true);
  int ConsoleOutput(MemoryBuf &buf, bool fLock=true) {
    return ConsoleOutput(&buf, fLock);
  };
#endif

  /// Examines the section [AppName] of the console configuration file
  char *FindAppConfig(const char *sbName, const char *sbFilterName=NULL,
		      int wId=-1);

  /// Examines the section [AppName] of the console configuration file
  long FindAppConfigInt(const char *sbName, const char *sbFilterName=NULL,
			int wId=-1, bool fInterpretUnits=true);

  /// Examines the section [AppName] of the console configuration file
  double FindAppConfigDbl(const char *sbName, const char *sbFilterName=NULL,
			  int wId=-1);
};

#endif /* H_DC_FILTERSERVICE */
// ****************************************************************************
