// ****************************************************************************
//  Name: dc_instance.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 08/10/2000 : mdb - Created.
//        09/20/2000 : mdb - Bugfix using iFilterSource/Sink instead of
//                           iPlacementSource/Sink in _StartInstance().
//        11/20/2000 : mdb - Reworked stopinstance to handle counting itself.
//        03/10/2000 : mdb - Changed to support transparant set of copies
//                           on multiple hosts.
//        03/13/2001 : mdb - Sweeping changes to support multiple copies.
//        05/01/2001 : mdb - Long standing shutdown bug - we now count remote
//                           instances.
//        07/26/2001 : mdb - Added write policy subsets per copyset, new
//                           placement syntax/parsing.
//        09/20/2001 : mdb - Fixed bug in _StartInstance() counting the pseudo
//                           <console> filter, causing shutdown hangs.
//        09/26/2001 : mdb - Fixed bug: test for multiple <console> within a
//                           filter instance was incorrectly checking between
//                           *all* filter instances. (Umit ran into this)
//        09/27/2001 : mdb - Fixed nasty bug in StartInstance() where the
//                           <console> pRemoteFilter reader-console handoff
//                           was not reset causing the first instance to be
//                           used for all.
//        01/10/2002 : mdb - Compute and transfer 3 sets of rank/max values.
//  Desc: DataCutter Filter Service user library.
// ****************************************************************************

#include "lib/dc_threadclass.h"

#include "dc_filterservice.h"
#include "consoleprocessstate.h"
#include "dc_instance.h"
#include "dc_error.h"
#include "packet.h"
#include "packet-tags.h"
#include "util.h"

#ifdef _WIN32
#include "win32.h"
#endif


#define INSTANCE_DEFAULT_SINKQ_SIZE   (5)
#define INSTANCE_DEFAULT_WORKQ_SIZE  (10)


#if defined(DCDBG)
//#define DCDBG_FILTERSERVICE_PROCESS_SPECS
//#define DCDBG_FILTERSERVICE_APPENDWORK
//#define DCDBG_FILTERSERVICE_STARTINSTANCE
//#define DCDBG_FILTERSERVICE_STOPINSTANCE
#define DCDBG_FILTERSERVICE_ONEPERNODE_ASSIGNMENT
#endif /* DCDBG */


/**
 * look through copy set, and check if this placement is present
 */
DC_ConsoleFilterSpec::CopySet_t *
DC_ConsoleFilterSpec::find_placement(Types::uint iPlacement, Types::uint *piFound) {
  for (Types::uint m=0; m<daCopySet.NumElem(); ++m) {
    if (daCopySet[m].iPlacement == iPlacement) {
      if (piFound) *piFound = m;
      return &(daCopySet[m]);
    }
  }
  return NULL;
}

/**
 * ctor
 */
DC_FilterInstance_t::DC_FilterInstance_t(void)
  : pDC(NULL), pConsole(NULL), sbLayoutName(NULL), eState(INIT),
    wNextWorkNum(0),
    wRemoteInstances(0), cRemoteInstanceExit(0) {
    //wRemoteFilters(0), cRemoteFilterExit(0) {
}

/**
 * dtor
 */
DC_FilterInstance_t::~DC_FilterInstance_t(void) {
  _Delete();
}

void DC_FilterInstance_t::_Delete(void) {
  TestDeleteNull(sbLayoutName);
  if (eState == INIT) return;
  Types::uint i;
  for (i=0; i<daFilterSpec.NumElem(); ++i) {
    if (daFilterSpec[i]) delete daFilterSpec[i];
  }
  daFilterSpec.Flush();
  for (i=0; i<daStreamSpec.NumElem(); ++i) {
    if (daStreamSpec[i]) delete daStreamSpec[i];
  }
  daStreamSpec.Flush();
}

Types::uint DC_FilterInstance_t::find_filter(const char *sbName) {
  DC_ConsoleFilterSpec *pFilter;
  for (Types::uint i=0; i<daFilterSpec.NumElem(); ++i) {
    pFilter = daFilterSpec[i];
    if (pFilter && strcmp(sbName, pFilter->sbName) == 0) {
      return i;
    }
  }
  return Types::uintMax;  // not found
}

/**
 * look for a stream by name, and if not found, create a new one
 */
DC_ConsoleStreamSpec *DC_FilterInstance_t::find_stream(const char *sbName,
						       Types::uint &iStream) {
  DC_ConsoleStreamSpec *pStream;
  for (Types::uint i=0; i<daStreamSpec.NumElem(); ++i) {
    pStream = daStreamSpec[i];
    if (pStream && strcmp(sbName, pStream->sbName) == 0) {  // found
      iStream = i;
      return pStream;
    }
  }
  // not found -- create a new one
  pStream = new DC_ConsoleStreamSpec(sbName);
  pStream->i = daStreamSpec.Append(pStream);
  iStream = pStream->i;
  return pStream;
}


// TODO: in case instance reused by client, and eState is done, we need to
// empty da's etc.
int DC_FilterInstance_t::_Reset(DC_FilterService_t *pDC_in, 
				DC_ConsoleProcessState *pConsole_in) {
  mutex.lock();

  _Delete(); // in case this is being reused, reader should drop any pkts

  pDC = pDC_in;
  pConsole = pConsole_in;
  wInstanceNum = (pConsole->wNextInstanceNum)++;

  mutex.unlock();
  return 0;
}

/**
 * parse and validate the configinfo based layout, and fill out the
 * specs in our instance.
 *
 
 * case "thru":
 * Indicated by a single filter with the name <console>.  Can only
 * have ins (we may remove this restiction in the future if needed).
 * This filter is virtual, and no thread is created, as done for a
 * normal remote process filter.  The thread of the console process
 * main is used instead.  All the remote process structures are created
 * but the ins[] struct necessarily needs to be different, because
 * instances are dynamically created and stopped.  The console's
 * filterinstance obj is where the ins will be parked.
 */
int DC_FilterInstance_t::_ProcessLayout(DC_FilterLayout_t &layout) {
  const char *sb, *sbSection;
  Types::uint i, iFilter, iStream;
  Str str;
  DC_ConsoleFilterSpec *pFilter;
  DC_ConsoleStreamSpec *pStream;

  sbLayoutName = StrDup(layout.getName());
  if (!sbLayoutName) sbLayoutName = StrDup("<unnamed>");

  // process the filter entries
  sbSection = layout.FirstSection();
  while (sbSection != NULL) {
    // anything with a section name "filter.*" is considered a filter
    if (strncmp(sbSection, "filter.", 7) == 0 && strlen(sbSection) > 7) {
      // BUGBUG: not insuring absence of whitespace in the filter name
      sb = sbSection+7;

      // make sure filter name is unique
      for (i=0; i<daFilterSpec.NumElem(); ++i) {
	pFilter = daFilterSpec[i];
	if (pFilter && strcmp(sb, pFilter->sbName) == 0) {
	  cerr << ErrorHeader(wERROR)
	       << "_ProcessLayout(): filter names must be unique: \""
	       << sb << "\"" << endl;
	  return -1;
	}
      }

      pFilter = new DC_ConsoleFilterSpec(sb);
      iFilter = daFilterSpec.Append(pFilter);

#if defined(DCDBG_FILTERSERVICE_PROCESS_SPECS)
      cout << "_ProcessLayout(): Found " << *pFilter
	   << endl;
#endif /* DCDBG_FILTERSERVICE_PROCESS_SPECS */

#if 0  /* not sure if I will keep this */
      // process the optional copies=no entry
      sb = FindParm(layout, sbSection, "copies", false);
      if (!sb || StrCmp(sb, "yes") == 0 || StrCmp(sb, "true") == 0) {
	pFilter->fAllowCopies = true;
      } else if (StrCmp(sb, "no") == 0 || StrCmp(sb, "false") == 0) {
	pFilter->fAllowCopies = false;
      } else {
	cerr << "_ProcessLayout(): invalid copies= value \""
	     << sb << "\" for filter \""
	     << pFilter->sbName << "\"" << endl;
	return -1;
      }
#endif

      pFilter->wSinkQueueSize = FindParmIntDefault(pDC->cfgConsole,
						   "console",
						   "sink_queue_size",
						   INSTANCE_DEFAULT_SINKQ_SIZE);
      pFilter->wWorkQueueSize = FindParmIntDefault(pDC->cfgConsole,
						   "console",
						   "work_queue_size",
						   INSTANCE_DEFAULT_WORKQ_SIZE);

      // process the ins and outs entries
      str.Set(FindParm(layout, sbSection, "ins", false));
      while ((sb = str.TokenNoDup()) != NULL) {
	pStream = find_stream(sb, iStream);
	if (pStream->iFilterSink != Types::uintMax) {
	  cerr << ErrorHeader(wERROR)
	       << "_ProcessLayout(): cannot have two sinks for stream \""
	       << sb << "\"" << endl;
	  return -1;
	}
	pStream->iFilterSink = iFilter;
	daFilterSpec[iFilter]->daStreamSpecIn.Append(pStream);
      }
      str.Set(FindParm(layout, sbSection, "outs", false));
      while ((sb = str.TokenNoDup()) != NULL) {
	pStream = find_stream(sb, iStream);
	if (pStream->iFilterSource != Types::uintMax) {
	  cerr << "_ProcessLayout(): cannot have two sources for stream \""
	       << sb << "\"" << endl;
	  return -1;
	}
	pStream->iFilterSource = iFilter;
	daFilterSpec[iFilter]->daStreamSpecOut.Append(pStream);
      }
    }
    sbSection = layout.NextSection();
  }
  if (daFilterSpec.NumElem() == 0) {
    cerr << "_ProcessLayout(): no filters found." << endl;
    return -1;
  }

  // check streams for missing source or sink
  for (i=0; i<daStreamSpec.NumElem(); ++i) {
    pStream = daStreamSpec[i];
    if (pStream->iFilterSource == Types::uintMax) {
      cerr << "_ProcessLayout(): missing source for stream \""
	   << pStream->sbName << "\"" << endl;
      return -1;
    } else if (pStream->iFilterSink == Types::uintMax) {
      cerr << "_ProcessLayout(): missing sink for stream \""
	   << pStream->sbName << "\"" << endl;
      return -1;
    }

    // check for stream specific settings in [stream.*]
    char bufTmp[80];
    snprintf(bufTmp, sizeof(bufTmp), "stream.%s", pStream->sbName);
    // optional items
    char *sb = layout.Find(bufTmp, "write_policy");
    if (sb) {
      pStream->eStreamWritePolicy = DC_StreamWritePolicy_UD;
      pStream->sbWritePolicyUD = StrDup(sb);
    } else {
      pStream->eStreamWritePolicy = pConsole->eStreamWritePolicy;
    }
    //char *sb = layout.Find(bufTmp, "queue_size");
    //if (sb) pStream->wQueueSize = atoi(sb);

#if defined(DCDBG_FILTERSERVICE_PROCESS_SPECS)
    cout << "_ProcessLayout(): "
	 << "Found stream <" << pStream->sbName << ", from "
	 << daFilterSpec[pStream->iFilterSource]->sbName
	 << '(' << pStream->iFilterSource << ") to "
	 << daFilterSpec[pStream->iFilterSink]->sbName
	 << '(' << pStream->iFilterSink << ")>" << endl;
#endif /* DCDBG_FILTERSERVICE_PROCESS_SPECS */
  }
  return 0;
}

/**
 * parse and validate the placement information, check which remote
 * processes are not in daPlacementSpec and add them.
 */
// thread: console main
int DC_FilterInstance_t::_ProcessPlacement(DC_Placement_t &placement) {
  Types::uint i, j, k;
  char bufTmp[128], *sbHostName, *sb;
  Str str;
  DArray<char*> *pdaOnePerNode = NULL;
  unsigned int cOnePerNode;
  int iPlacement;
  bool fConsoleCheck = false;

  fDetached = true;

  // only look for placement entries that match filters we are going to run
  for (i=0; i<daFilterSpec.NumElem(); ++i) {
    DC_ConsoleFilterSpec *pFilter = daFilterSpec[i];
    if (strcmp(pFilter->sbName, "<console>") == 0) { // console case
      if (fConsoleCheck) {
	// ERROR: there should never be >1 placement spec w/fConsole
	// true, if there is only one <console> filter
	cerr << ErrorHeader(wERROR)
	     << "_ProcessPlacement(): "
	     << "cannot have more than one <console> filter" << endl;
	return -1;
      }
      fConsoleCheck = true;

      // here we do not require or even look for a placement entry for the
      // pseudo-filter in the console process.  we always place it where the
      // console was started.
      fDetached = false;

      // add a new entry if not a duplicate
      pConsole->mutexPlacementSpec.lock();

      for (j=0; j<pConsole->daPlacementSpec.NumElem(); ++j) {
	DC_PlacementSpec *p = pConsole->daPlacementSpec[j];
	if (p && p->fConsole) break;
      }
      if (j < pConsole->daPlacementSpec.NumElem()) {
	iPlacement = j;
      } else {
	// add a new entry to the consoleprocessstate daPlacementSpec
	sbHostName = pConsole->
	  hi.CanonicalName_Alloc(pDC->addrLocal.getHostName());

	iPlacement = pConsole->
	  daPlacementSpec.AppendByVal(new DC_PlacementSpec(sbHostName));

	delete [] sbHostName;
      }
      DC_ConsoleFilterSpec::CopySet_t cs;
      cs.iPlacement = iPlacement;
      cs.sbName = StrDup(pFilter->sbName);
      cs.fAllowCopies = false;
      cs.nCopiesLow = cs.nCopiesHigh = cs.nCopies = 1;
      cs.rankTotalCopiesStart = pFilter->maxTotalCopies;
      pFilter->daCopySet.Append(cs);
      pFilter->maxTotalCopies += cs.nCopies;

      // TODO: filter name on the console may have <name>:<cs-name> syntax
      if (strchr(pFilter->sbName, ':')) {
	cout << "ERROR: not implemented ':' copyset names for <console>"
	     << endl;
	cout.flush();
	exit(1);
      }

      pConsole->mutexPlacementSpec.unlock();

#if defined(DCDBG_FILTERSERVICE_PROCESS_SPECS)
      cout << "_ProcessPlacement(): "
	   << "Place pseudo-filter " << pFilter->sbName
	   << " on the console host, iPlacement=" << iPlacement << endl;
#endif /* DCDBG_FILTERSERVICE_PROCESS_SPECS */

    } else { // normal remote process case, using an appd

      // Current hostname handling for placmement and dird name: The
      // placement name must be the canonical hostname of where the
      // placement is, because this is how we lookup the appd contact
      // info.  The dns_name is redundant in the appd entry at the
      // dird.

      // BUGBUG: We currently assume that the existence of a
      // registered appd on a particular host means we can make use of
      // it.  Problems with this can occur if the userid of the appd
      // process does not match the userid of the console process.

      if (!placement.FirstSection("placement")) {
	cerr << ErrorHeader(wERROR)
	     << "_ProcessPlacement(): no [placement] section found." << endl;
	return DC_ERR_InvalidPlacement;
      }
      const char *sbTmp;
      bool  fPlacementFound = false;
      for (sbTmp=placement.FirstName(); sbTmp; sbTmp=placement.NextName()) {
	const char *p = strchr(sbTmp, ':');
	if ((!p && strcmp(sbTmp, pFilter->sbName) != 0) ||
	    (p && strncmp(sbTmp, pFilter->sbName, p-sbTmp) != 0)) continue;

	// have a placement entry for this filter
	fPlacementFound = true;
	
	char *sbPlacementName;
	int   nCopiesLow = -1, nCopiesHigh = -1;

	// check for optional specification of copies low and high
	str.Set(placement.CurrentValue());
	sbPlacementName = str.TokenNoDup();  // hostname
	if ((sb = str.TokenNoDup()) != NULL) {  // low
	  int n = atoi(sb);
	  if (n == 0) continue;
	  if (n < 0 || n > 100) {
	    cerr << ErrorHeader(wERROR)
		 << "_ProcessPlacement(): invalid number of copies value \""
		 << sb << "\"" << endl;
	    return DC_ERR_InvalidPlacement;
	  }
	  nCopiesLow = n;
	  if ((sb = str.TokenNoDup()) != NULL) {  // high
	    int n = atoi(sb);
	    if (n == 0) continue;
	    if (n < 0 || n > 100) {
	      cerr << ErrorHeader(wERROR)
		   << "_ProcessPlacement(): invalid number of copies value \""
		   << sb << "\"" << endl;
	      return DC_ERR_InvalidPlacement;
	    }
	    nCopiesHigh = n;
	  }
	}

	if (strcmp(sbPlacementName, "<one_per_node>") == 0) {
	  if (! pdaOnePerNode) {  // first one encountered
	    pdaOnePerNode = pConsole->cfgDir.FindAll(NULL, "dns_name");
	    cOnePerNode = 0;
	  }
	  // just assign the next node round-robin
	  sbPlacementName
	    = StrDup(pdaOnePerNode->Get(cOnePerNode++ %
					pdaOnePerNode->NumElem()));
#if defined(DCDBG_FILTERSERVICE_ONEPERNODE_ASSIGNMENT)
	  cout << "<one_per_node> assigning "
	       << pFilter->sbName << " to " << sbPlacementName << endl;
#endif /* DCDBG_FILTERSERVICE_ONEPERNODE_ASSIGNMENT */
	} else {
	  //cout << pConsole->hi.cfg;
	  sbPlacementName = pConsole->
	    hi.CanonicalName_Alloc(sbPlacementName, true);
	}
    
	// lookup dir entries, create placement objs, and update filter spec
	snprintf(bufTmp, sizeof(bufTmp), "datacutter.appd.\"%s\"", sbPlacementName);

	char *sbSection = bufTmp;
	char *sbHostName = FindParm(pConsole->cfgDir, sbSection, "dns_name",
				    false);
	if (! sbHostName) {
	  cerr << "_ProcessPlacement(): missing entry for appd on host \""
	       << sbPlacementName << "\"" << endl;
	  delete [] sbPlacementName;
	  return DC_ERR_InvalidPlacement;  // appd entry not found
	}
	sbHostName = pConsole->hi.CanonicalName_Alloc(sbHostName);
	int wPort = atoi(FindParm(pConsole->cfgDir, sbSection, "port"));

#if defined(DCDBG_FILTERSERVICE_PROCESS_SPECS)
	cout << "_ProcessPlacement(): "
	     << "Place filter " << pFilter->sbName
	     << " on " << sbHostName << ':' << wPort << endl;
#endif /* DCDBG_FILTERSERVICE_PROCESS_SPECS */

	// add a new entry if not a duplicate
	pConsole->mutexPlacementSpec.lock();

	for (k=0; k<pConsole->daPlacementSpec.NumElem(); ++k) {
	  DC_PlacementSpec *p = pConsole->daPlacementSpec[k];
	  if (p && !p->fConsole &&
	      strcmp(sbHostName, p->sbHostName) == 0) break;
	}
	if (k < pConsole->daPlacementSpec.NumElem()) {
	  // this is a previously added placement -- check if this filter
	  // is already placed there.  This means there were two placement
	  // entries for the same (filter,host) pair, which will cause
	  // the runtime to hang.
	  for (Types::uint l=0; l<pFilter->daCopySet.NumElem(); ++l) {
	    if (pFilter->daCopySet[l].iPlacement == k) {
	      cerr << "_ProcessPlacement(): filter \"" << pFilter->sbName
		   << "\" cannot have >1 entries for the host \""
		   << sbHostName << ':' << wPort << endl;
	      delete [] sbPlacementName;
	      pConsole->mutexPlacementSpec.unlock();
	      return DC_ERR_InvalidPlacement;  // dup (filter,host) pair found
	    }
	  }
	  iPlacement = k;
	} else {
	  // add a new entry to the consoleprocessstate daPlacementSpec
	  iPlacement = pConsole->daPlacementSpec
	    .AppendByVal(new DC_PlacementSpec(sbHostName, wPort));
	}
	DC_ConsoleFilterSpec::CopySet_t cs;
	cs.iPlacement = iPlacement;
	cs.sbName = StrDup(sbTmp);
	cs.fAllowCopies = true;  // TODO: re-enable [filter]copies=yes|no ?
	cs.nCopiesLow = nCopiesLow;
	cs.nCopiesHigh = nCopiesHigh;
	cs.nCopies = (nCopiesLow != -1 ? nCopiesLow : 1);
	cs.rankTotalCopiesStart = pFilter->maxTotalCopies;
	pFilter->daCopySet.Append(cs);
	pFilter->maxTotalCopies += cs.nCopies;

	pConsole->mutexPlacementSpec.unlock();

	delete [] sbPlacementName;
	delete [] sbHostName;
      }
      if (!fPlacementFound) {
	cerr << ErrorHeader(wERROR)
	     << "_ProcessPlacement(): no placement for filter \""
	     << pFilter->sbName << "\"" << endl;
	return DC_ERR_InvalidPlacement;
      }
    }
  }

  // now look for stream settings
  for (i=0; i<daStreamSpec.NumElem(); ++i) {
    DC_ConsoleStreamSpec *pStream = daStreamSpec[i];
    
    DArray<char*> *pdaPlacement = placement.FindAll("placement",
						    pStream->sbName);
    if (pdaPlacement->NumElem() == 0) continue;  // just use default settings

    for (Types::uint j=0; j<pdaPlacement->NumElem(); ++j) {
      char *sb, *sbSource, *sbSink;
      bool fNone = false;
      DC_StreamWritePolicy_t eStreamWritePolicy;    
      str.Set(pdaPlacement->Get(j));
      if ((sb = str.TokenNoDup()) == NULL) continue;
      if (strcmp(sb, "write") == 0) {
	// <stream> = write <source_re> <sink_re> <policy> <args>
	if ((sb = str.TokenNoDup()) == NULL) continue;
	sbSource = sb;
	if ((sb = str.TokenNoDup()) == NULL) continue;
	sbSink = sb;
	if ((sb = str.TokenNoDup()) == NULL) continue;
	// policy
	if (StrCmp(sb, "<none>", true) == 0) {
	  fNone = true;
	} else {
	  eStreamWritePolicy = DC_StreamWritePolicy_toenum(sb);
	  if (eStreamWritePolicy == DC_StreamWritePolicy_UD) {
	    // this does not make sense here, since specified in filterlayout
	    cerr << "_ProcessPlacement(): stream \"" << pStream->sbName
		 << "\" write rule invalid -- UD policy specified in layout"
		 << endl;
	    continue;
	  }
	}
	// add to array of rules
	DC_ConsoleStreamSpec::WriteRules_t *pwr
	  = new DC_ConsoleStreamSpec::WriteRules_t;
	if (pwr->reSource.setPattern(sbSource)) {
	  cerr << "_ProcessPlacement(): stream \"" << pStream->sbName
	       << "\" write rule has an invalid <source-re> \""
	       << sbSource << '\"' << endl;
	  return DC_ERR_InvalidPlacement;
	}
	if (pwr->reSink.setPattern(sbSink)) {
	  cerr << "_ProcessPlacement(): stream \"" << pStream->sbName
	       << "\" write rule has an invalid <sink-re> \""
	       << sbSink << '\"' << endl;
	  return DC_ERR_InvalidPlacement;
	}
	pwr->fNone = fNone;
	pwr->eStreamWritePolicy = eStreamWritePolicy;
	pStream->daWriteRules.Append(pwr);
      } else if (strcmp(sb, "sink_queue_size") == 0) {
	// TODO:
      }
    }
    delete pdaPlacement;
  }

  // cleanup
  if (pdaOnePerNode) delete pdaOnePerNode;
  return 0;
}


/**
 * generate and send the start instance command for each remote process
 */
// thread: console process main
int DC_FilterInstance_t::_StartInstance(void) {
  Packet packetOut;
  MemoryBuf mb(KB_32);
  Types::uint i, j, k, l, m, n;
  int   nFilters, nCopies;
  DC_PlacementSpec *pPlacement;
  DC_ConsoleFilterSpec *pFilter;
  DC_ConsoleStreamSpec *pStream;

  // compute number of instanceexit pkts we will eventually wait for
  mutex.lock();
  wRemoteInstances = 0;
  for (i=0; i<pConsole->daPlacementSpec.NumElem(); ++i) {
    pPlacement = pConsole->daPlacementSpec[i];

    // do we use this remote process for our instance?
    bool fUse = false;
    nFilters = nCopies = 0;
    for (j=0; j<daFilterSpec.NumElem(); ++j) {
      if (daFilterSpec[j]->find_placement(i) && !daFilterSpec[j]->fConsole) {
	fUse = true;
	break;
      }
    }
    if (fUse) ++wRemoteInstances;
  }
  mutex.unlock();
#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
  cout << "DEBUG: found " << wRemoteInstances << " remote processes for i"
       << wInstanceNum << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */

  DC_ConsoleFilterSpec::CopySet_t *pcs;

  for (i=0; i<pConsole->daPlacementSpec.NumElem(); ++i) {
    pPlacement = pConsole->daPlacementSpec[i];

#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
    cout << "DEBUG: remote process on "
	 << pPlacement->sbHostName << ':' << pPlacement->wPortLib << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */

    // pass 1 - count the number of filters for this remote process
    nFilters = nCopies = 0;
    for (j=0; j<daFilterSpec.NumElem(); ++j) {
      pcs = daFilterSpec[j]->find_placement(i);
      if (pcs) {
	++nFilters;
	nCopies += pcs->nCopies;
      }
    }
    if (nFilters == 0) continue;

    // pass 2 - build the command buffer
    mb.Empty();
    mb.sprintf("layout_name=%s\n", sbLayoutName);
    mb.sprintf("num_filters=%d\n", nFilters);
    mb.sprintf("num_copies=%d\n", nCopies);
    nFilters=0;
    for (j=0; j<daFilterSpec.NumElem(); ++j) {
      Types::uint ics;
      pFilter = daFilterSpec[j];
      pcs = pFilter->find_placement(i, &ics);
      if (!pcs) continue;

#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
      cout << "DEBUG: filter " << pFilter->sbName << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */

      mb.sprintf("[%d]\n", nFilters++);
      mb.sprintf("filter=%s\n", pFilter->sbName);
      if (pFilter->fConsole) mb.sprintf("console=true\n");
      mb.sprintf("copies=%s\n", (pcs->fAllowCopies ? "yes" : "no"));
      mb.sprintf("numcopies_initial=%d\n", pcs->nCopies);
      mb.sprintf("numcopies_low=%d\n", pcs->nCopiesLow);
      mb.sprintf("numcopies_high=%d\n", pcs->nCopiesHigh);
      mb.sprintf("sink_queue_size=%d\n", pFilter->wSinkQueueSize);
      mb.sprintf("work_queue_size=%d\n", pFilter->wWorkQueueSize);
      mb.sprintf("worknum=%d\n", 0);  // hard code this for now
      mb.sprintf("rankcopysets=%u\n", ics);
      mb.sprintf("maxcopysets=%u\n", pFilter->daCopySet.NumElem());
      mb.sprintf("rankcopies=%d\n", pcs->rankTotalCopiesStart);
      mb.sprintf("maxcopies=%d\n", pFilter->maxTotalCopies);
      mb.sprintf("nins=%d\n", pFilter->daStreamSpecIn.NumElem());
      for (k=0; k<pFilter->daStreamSpecIn.NumElem(); ++k) {
	pStream = pFilter->daStreamSpecIn[k];

#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
	cout << "DEBUG: input stream " << pStream->sbName << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */

	mb.sprintf("ins%d_name=%s\n", k, pStream->sbName);
	mb.sprintf("ins%d_filter=%s\n", k,
		   daFilterSpec[pStream->iFilterSource]->sbName);
	for (l=0,n=0;
	     l<daFilterSpec[pStream->iFilterSource]->daCopySet.NumElem();
	     ++l) {
	  DC_ConsoleFilterSpec::CopySet_t csSource
	    = daFilterSpec[pStream->iFilterSource]->daCopySet[l];

#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
	  cout << "DEBUG: ins copyset " << l << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */

	  // check for write rule match
	  DC_ConsoleStreamSpec::WriteRules_t *pWriteRules = NULL;
	  for (m=0; m<pStream->daWriteRules.NumElem(); ++m) {
	    if (pStream->daWriteRules[m]->reSource.match(csSource.sbName) &&
		pStream->daWriteRules[m]->reSink.match(pcs->sbName)) {
#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
	      cout << "DEBUG: ins matched copyset rule " << m
		   << " for source=" << csSource.sbName
		   << " sink=" << pcs->sbName
		   << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */
	      pWriteRules = pStream->daWriteRules[m];
	      break;  // only first match is considered
	    }
	  }
	  if (pWriteRules && pWriteRules->fNone) continue;
	  if (!pWriteRules) {
	    if (pStream->daWriteRules.NumElem() > 0) continue;
	  }

#if 0 /* this is not used; instead we pass the policy in STREAMCONNECT */
	  if (pStream->eStreamWritePolicy == DC_StreamWritePolicy_UD) {
	    // override with user defined policy
	    mb.sprintf("ins%d_write_policy%d=UD\n", k, n);
	    mb.sprintf("ins%d_write_policy%d_UD=%s\n", k, n,
		       pStream->sbWritePolicyUD);
	  } else {
	    mb.sprintf("ins%d_write_policy%d=%s\n", k, n,
		       DC_StreamWritePolicy_tostr((pWriteRules
						   ? pWriteRules->
						   eStreamWritePolicy
						   : pConsole->
						   eStreamWritePolicy)));
	  }
#endif
	  //mb.sprintf("ins%d_copysetid%d=%d\n", k, n, l);
	  if (csSource.iPlacement == i) {
	    mb.sprintf("ins%d_host%d=<colocated>\n", k, n);
	  } else {
	    mb.sprintf("ins%d_host%d=%s\n", k, n,
		       pConsole->daPlacementSpec[csSource.iPlacement]
		       ->sbHostName);
	    mb.sprintf("ins%d_port%d=%d\n", k, n,
		       pConsole->daPlacementSpec[csSource.iPlacement]
		       ->wPortLib);
	  }
	  mb.sprintf("ins%d_numcopies%d=%d\n", k, n, csSource.nCopies);

	  ++n;
	}
      }
      DC_StreamWritePolicy_t eStreamWritePolicy;
      mb.sprintf("nouts=%d\n", pFilter->daStreamSpecOut.NumElem());
      for (k=0; k<pFilter->daStreamSpecOut.NumElem(); ++k) {
	pStream = pFilter->daStreamSpecOut[k];

#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
	cout << "DEBUG: output stream " << pStream->sbName << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */

	mb.sprintf("outs%d_name=%s\n", k, pStream->sbName);
	mb.sprintf("outs%d_filter=%s\n", k,
		   daFilterSpec[pStream->iFilterSink]->sbName);
	bool fOutputPolicy = true;  // do once per filter
	for (l=0,n=0;
	     l<daFilterSpec[pStream->iFilterSink]->daCopySet.NumElem();
	     ++l) {
	  DC_ConsoleFilterSpec::CopySet_t csSink
	    = daFilterSpec[pStream->iFilterSink]->daCopySet[l];

#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
	  cout << "DEBUG: outs copyset " << l << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */

	  // check for write rule match
	  DC_ConsoleStreamSpec::WriteRules_t *pWriteRules = NULL;
	  for (m=0; m<pStream->daWriteRules.NumElem(); ++m) {
	    if (pStream->daWriteRules[m]->reSource.match(pcs->sbName) &&
		pStream->daWriteRules[m]->reSink.match(csSink.sbName)) {
#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
	      cout << "DEBUG: outs matched copyset rule " << m
		   << " for source=" << pcs->sbName
		   << " sink=" << csSink.sbName
		   << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */
	      pWriteRules = pStream->daWriteRules[m];
	      break;  // only first match is considered
	    }
	  }
	  if (pWriteRules && pWriteRules->fNone) continue;
	  if (!pWriteRules) {
	    if (pStream->daWriteRules.NumElem() > 0) continue;
	  }

	  // we can only support one stream write policy per outs
	  if (fOutputPolicy) {
	    fOutputPolicy = false;
	    if (pStream->eStreamWritePolicy == DC_StreamWritePolicy_UD) {
	      eStreamWritePolicy = DC_StreamWritePolicy_UD;
	      // override with user defined policy
	      mb.sprintf("outs%d_write_policy=UD\n", k);
	      mb.sprintf("outs%d_write_policy_UD=%s\n", k,
			 pStream->sbWritePolicyUD);
	    } else {
	      eStreamWritePolicy = (pWriteRules
				    ? pWriteRules->eStreamWritePolicy
				    : pConsole->eStreamWritePolicy);
	      mb.sprintf("outs%d_write_policy=%s\n", k,
			 DC_StreamWritePolicy_tostr(eStreamWritePolicy));
	    }
	  } else {
	    // make sure the policy rules don't imply mixing of write policies
	    if (pStream->eStreamWritePolicy != DC_StreamWritePolicy_UD &&
		eStreamWritePolicy != (pWriteRules
				       ? pWriteRules->eStreamWritePolicy
				       : pConsole->eStreamWritePolicy)) {
	      cerr << "_StartInstance(): ERROR: stream \"" << pStream->sbName
		   << "\" write rules cause policy "
		   << DC_StreamWritePolicy_tostr(eStreamWritePolicy)
		   << " and policy "
		   << DC_StreamWritePolicy_tostr((pWriteRules
						  ? pWriteRules
						  ->eStreamWritePolicy
						  : pConsole
						  ->eStreamWritePolicy))
		   << " to be used." << endl
		   << "  Add more rules using <none> or specifying more cases"
		   << endl;
	      return DC_ERR_InvalidPlacement;
	    }
	  }
	  //mb.sprintf("outs%d_copysetid%d=%d\n", k, n, l);
	  if (csSink.iPlacement == i) {
	    mb.sprintf("outs%d_host%d=<colocated>\n", k, n);
	  } else {
	    char *sbConnectTo = pConsole->hi
	      .ConnectName(pConsole->daPlacementSpec[i]
			   ->sbHostName,
			   pConsole->daPlacementSpec[csSink.iPlacement]
			   ->sbHostName);
	    //cout << "*** DEBUG: Connecting " << pConsole->daPlacementSpec[i]->sbHostName << " to " << pConsole->daPlacementSpec[csSink.iPlacement]->sbHostName << " using the name " << sbConnectTo << endl;

	    mb.sprintf("outs%d_host%d=%s\n", k, n, sbConnectTo);
	    mb.sprintf("outs%d_port%d=%d\n", k, n,
		       pConsole->daPlacementSpec[csSink.iPlacement]
		       ->wPortLib);
	  }
	  mb.sprintf("outs%d_numcopies%d=%d\n", k, n, csSink.nCopies);

	  ++n;
	}
      }
    }
#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
    cout << "STARTINSTANCE packet\n" << mb.getPtr() << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */

    // send to remote process for execution
    packetOut.set(TAG_FILTERSERVICE_STARTINSTANCE, &mb);
    packetOut.setExtra(wInstanceNum);
    if (packetOut.write(*(pPlacement->psock)) <= 0) {
      cerr << "_StartInstance(): failed write to remote process on "
	   << pPlacement->sbHostName << ':' << pPlacement->wPortLib << endl
	   << " fd=" << pPlacement->psock->getFd()
	   << " - " << strerror(errno) << endl;
      // TODO: tear down the partial instance, and possibly other
      // instances that share this placement host?
      mutex.lock();
      --wRemoteInstances;  // we will never hear from this one!
      mutex.unlock();
      return -1;
    }
#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
    cout << "Sent STARTINSTANCE i" << wInstanceNum << " command to host "
	 << pPlacement->sbHostName << ':'
	 << pPlacement->wPortLib << " ..." << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */
  }
  eState = WAIT;

  if (!fDetached) {
    // We wait here if a pseudo-filter is needed.  The reader will
    // signal us when the remote filter state has been created.  The
    // ptr to this state is passed using the consoleprocessstate
    // object.  This assumes that all calls trying to start a new
    // filter instance happen from a single thread.  This state could
    // be inside this instance object, but that complicates the reader
    // in trying to find it, and is not necessary under the above
    // assumption.
    pConsole->mutexRemoteFilter.lock();
    while (pConsole->pRemoteFilter == NULL) {
#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
      cout << "_StartInstance(): waiting for local reader to signal i"
	   << wInstanceNum << " processed ..." << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */
      pConsole->cvRemoteFilter.wait(pConsole->mutexRemoteFilter);
    }
    pConsole->mutexRemoteFilter.unlock();

    // now the reader has setup the pseudo-filter data structs, so
    // we can setup the console main thread interface object to
    // resemble the arg passed to the process() fcn of filter code.
    arg._init(pConsole->pRemoteFilter, pDC);

    pConsole->pRemoteFilter = NULL;  // reset
  }

  return 0;
}


// ****************************************************************************
//  Name: AppendWork
//  Date: 09/10/2000 : mdb - Created.
// ****************************************************************************
/**
 *  Append the given work buffer to this running instance.
 *
 *  Inputs: work - uow to add to the new instance
 *          fWait - T=indicates if user will eventually call DC.WaitWork()
 *                  F=automatic reap of work state when finished
 *  Return: int - 0+=work handle, <0=failure
 */
// thread: console process main
DC_WorkHandle_t DC_FilterInstance_t::AppendWork(DC_Work_t &work, bool fWait) {
  if (eState == INIT) return DC_ERR_InstanceNotRunning;

  int wRtn = DC_ERR_OK;  // for !fWait case, this is never changed
  Packet packetOut;
  Types::uint i, j;
  int   nFilters, nTotalFilters, nPlacements;
  DC_PlacementSpec *pPlacement;
  char bufTmp[80];

  int wWorkNum = wNextWorkNum++;

  // NOTE: we can always keep another struct in the instance object
  // with the iPlacement values for only this instance.  In practice
  // this may be ok to have them all in a single array, since num
  // Placments and num Filters are small, so we walk thru the
  // placements and check all the filters for this instance to see if
  // we are running there.


  // add this to the pConsole->daConsoleWork struct (this needs to be
  // done before we send any APPENDWORK pkts to avoid an unlikely but
  // possible race with the reader receiving the APPENDWORK_DONE pkt)
  if (fWait) {
    // count numbers of later responses to expect
    nTotalFilters = nPlacements = 0;
    for (i=0; i<pConsole->daPlacementSpec.NumElem(); ++i) {
      pPlacement = pConsole->daPlacementSpec[i];

      // accumulate number of filters for this placement
      nFilters=0;
      for (j=0; j<daFilterSpec.NumElem(); ++j) {
	DC_ConsoleFilterSpec::CopySet_t *pcs
	  = daFilterSpec[j]->find_placement(i);

	if (pcs && !daFilterSpec[j]->fConsole) {
	  nFilters += pcs->nCopies;
	  //++nFilters;
	}
      }
      if (nFilters == 0) continue;

      ++nPlacements;
      nTotalFilters += nFilters;
    }

    pConsole->mutexConsoleWork.lock();

    DC_ConsoleWork *pwork = new DC_ConsoleWork();
    pwork->work.buf.DeepCopy(work.buf);  // save until work is done
    pwork->wWorkNum = wWorkNum;
    pwork->wInstanceNum = wInstanceNum;
    pwork->eState = DC_ConsoleWork::RUNNING;

    pwork->nPlacements = nPlacements;
    pwork->cPlacements = 0;
    pwork->nFilters = nTotalFilters;
    pwork->cFilters = 0;

    // look for vacated spot first, if none append
    bool fFound = false;
    for (i=0; i<pConsole->daConsoleWork.NumElem(); ++i) {
      if (! pConsole->daConsoleWork[i]) { fFound = true; break; };
    }
    if (fFound) {
      pConsole->daConsoleWork.Set(i, pwork);
    } else {
      i = pConsole->daConsoleWork.Append(pwork);
    }
    wRtn = i;  // handle to return

#if defined(STATS_FILTERSERVICE_WORK)
    pwork->tRespTime.Start();
#endif /* STATS_FILTERSERVICE_WORK */

#if defined(DCDBG_FILTERSERVICE_APPENDWORK)
    cout << "AppendWork(): added ConsoleWork entry " << i
	 << " wait nPlacements=" << pwork->nPlacements
	 << ", nFilters=" << pwork->nFilters
	 << endl;
#endif /* DCDBG_FILTERSERVICE_APPENDWORK */

    pConsole->mutexConsoleWork.unlock();
  }

  // send one per remote process (even if >1 filters running in this
  // instance). the remote reader will sort it out and append to each
  // filter's work queue.
  for (i=0; i<pConsole->daPlacementSpec.NumElem(); ++i) {
    pPlacement = pConsole->daPlacementSpec[i];

    // count number of filters for this placement
    nFilters=0;
    for (j=0; j<daFilterSpec.NumElem(); ++j) {
      if (daFilterSpec[j]->find_placement(i) && !daFilterSpec[j]->fConsole)
	++nFilters;
    }
    if (nFilters == 0) continue;

    // send the new work
    // BUGBUG: backpressure?  size of work?
    packetOut.set(TAG_FILTERSERVICE_APPENDWORK, &(work.buf));
    // Encode wInstanceNum and wWorkNum into wExtra packet header
    // value.  This is needed to allow for muxing the application data
    // for the *not* detached case.  We need to differentiate the data
    // as to what instance/work num generated it.
    packetOut.setExtra(DC_EncodeInt(wInstanceNum, wWorkNum));
    if (packetOut.write(*(pPlacement->psock)) <= 0) {
      snprintf(bufTmp, sizeof(bufTmp),
	      "AppendWork(): failed write to remote process on %s:%d\n",
	      pPlacement->sbHostName, pPlacement->wPortLib);
      perror(bufTmp);
      // TODO: some of the nodes may have gotten the new work, while
      // others have not.  what to do??? -- tear down the instance?
      return DC_ERR_RemoteProcessCommFailed;
    }
  }

  // Wait for APPENDWORK_ACK pkts?  No because we want this to be as
  // fast a startup as possible.  Failures need to be handled in a
  // less intrusive way.

  // BUGBUG: if this fails, or later the APPENDWORK_ACKS come back
  // indicating failure, we need to cleanup the possible
  // DC_ConsoleWork entry

#if defined(DCDBG_FILTERSERVICE_APPENDWORK)
  cout << "AppendWork(): instance " << wInstanceNum << ' ' << work.buf << endl;
#endif /* DCDBG_FILTERSERVICE_APPENDWORK */
  eState = WORK;

  return wRtn;
}


// ****************************************************************************
//  Name: _StopInstance (assume locked externally)
//  Date: 09/14/2000 : mdb - Created.
//        05/01/2001 : mdb - Changed from FILTEREXIT pkts for use with copies.
// ****************************************************************************
/**
 *  Sends stop messages to all remote processes for this instance.
 *  The reader will later recv the INSTANCEEXIT packets and adjust the
 *  instance state accordingly.
 *
 *  Return: int - number of INSTANCEEXIT packets to expect for this instance
 */
// thread: console process main
int DC_FilterInstance_t::_StopInstance(void) {
  int nFilters;
  Types::uint i, j;

  mutex.lock();

  if (eState == INIT || eState == DONE) { eState = DONE; return DC_ERR_OK; }
  if (eState == STOP) {  // stop operation already in progress
    return DC_ERR_OK;
  }

  // WAIT or WORK state continues here

  Packet packetOut;
  DC_PlacementSpec *pPlacement;
  char bufTmp[80];

  // send one per remote process.  remote reader will handle >1 filter
  for (i=0; i<pConsole->daPlacementSpec.NumElem(); ++i) {
    pPlacement = pConsole->daPlacementSpec[i];

    // count number of filters for this placement
    nFilters=0;
    for (j=0; j<daFilterSpec.NumElem(); ++j) {
      if (daFilterSpec[j]->find_placement(i) && !daFilterSpec[j]->fConsole)
	++nFilters;
    }
    if (nFilters == 0) continue;

    // send the stop message
    packetOut.set(TAG_FILTERSERVICE_STOPINSTANCE, NULL);
    packetOut.setExtra(wInstanceNum);
    if (packetOut.write(*(pPlacement->psock)) <= 0) {
      snprintf(bufTmp, sizeof(bufTmp),
	      "_StopInstance(): failed write to remote process on %s:%d\n",
	      pPlacement->sbHostName, pPlacement->wPortLib);
      perror(bufTmp);
      // cleanup - mark as done and continue
      eState = DC_FilterInstance_t::DONE;
    }
  }

#if defined(DCDBG_FILTERSERVICE_STOPINSTANCE)
  cout << "_StopInstance(): instance " << wInstanceNum << endl;
#endif /* DCDBG_FILTERSERVICE_STOPINSTANCE */
  eState = STOP;

  mutex.unlock();
  return DC_ERR_OK;
}

// thread: console process reader
int DC_FilterInstance_t::_InstanceExit_Reader(void) {
  mutex.lock();

  eState = STOP;

  ++cRemoteInstanceExit;
#if defined(DCDBG_FILTERSERVICE_STOPINSTANCE)
  cout << "DC_FilterInstance_t::_InstanceExit_Reader(): "
       << cRemoteInstanceExit << "/" << wRemoteInstances << " exited"
       << endl;
#endif /* DCDBG_FILTERSERVICE_STOPINSTANCE */
  if (cRemoteInstanceExit == wRemoteInstances) {
    eState = DONE;  // this is the last one
#if defined(DCDBG_FILTERSERVICE_STOPINSTANCE)
    cout << "_InstanceExit_Reader(): i" << wInstanceNum << " done, signal ..."
	 << endl;
#endif /* DCDBG_FILTERSERVICE_STOPINSTANCE */
    //cvDone.signal();
  }

  mutex.unlock();
  return 0;
}

// CLH - no longer called
int DC_FilterInstance_t::_WaitUntilInstanceDone(void) {
  mutex.lock();

  while (eState != DONE) {
    //cvDone.wait(mutex);
  }
#if defined(DCDBG_FILTERSERVICE_STOPINSTANCE)
  cout << "_WaitUntilInstanceDone(): i" << wInstanceNum << " done."
       << endl;
#endif /* DCDBG_FILTERSERVICE_STOPINSTANCE */

  mutex.unlock();
  return 0;
}


ostream &operator<<(ostream &os, DC_ConsoleStreamSpec &spec) {
  return os << "Stream{"
	    << spec.sbName
	    << ",ifilter=" << spec.iFilterSource << ' ' << spec.iFilterSink
	    << '}';
}
ostream &operator<<(ostream &os, DC_ConsoleFilterSpec &spec) {
  os << "Filter{"
     << spec.sbName << ",copyset=";
  for (Types::uint m=0; m<spec.daCopySet.NumElem(); ++m) {
    os << '{' << spec.daCopySet[m].iPlacement
       << ' ' << spec.daCopySet[m].nCopies << "},";
  }
  return os << '}';
}


// ****************************************************************************
