// -*- c++ -*-
// ****************************************************************************
//  Name: dc_instance.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 08/10/2000 : mdb - Created.
//  Desc: DataCutter Filter Service user library.
// ****************************************************************************
/** @file dc_instance.h */

#if !defined(H_DC_INSTANCE)
#define H_DC_INSTANCE

#include "lib/dc_standard.h"
#include "lib/dc_argv.h"
#include "lib/dc_configinfo.h"
#include "lib/dc_socket.h"
#include "lib/dc_memory.h"
#include "lib/dc_map.h"
#include "lib/dc_timer.h"
#include "lib/dc_metric.h"
#include "lib/dc_mutex.h"
#include "lib/dc_cond.h"
#include "lib/dc_regex.h"

#include "packet-tags.h"
#include "dc_work.h"
#include "dc_filterlayout.h"
#include "dc_placement.h"
#include "dc_policy.h"


class DC_ConsoleStreamSpec {
  //friend ostream &operator<<(ostream &os, DC_ConsoleStreamSpec &spec);
public:
  char *sbName;
  Types::uint i;  ///< index of the containing daStreamSpec[]
  Types::uint iFilterSource, iFilterSink;
  DC_StreamWritePolicy_t eStreamWritePolicy;  ///< console default or UD override
  char *sbWritePolicyUD;

  /// rules for writer subsets
  class WriteRules_t {
  public:
    DC_Regex reSource;
    DC_Regex reSink;
    bool     fNone;
    DC_StreamWritePolicy_t eStreamWritePolicy;    
  };
  DArrayPtr<WriteRules_t*> daWriteRules;

  DC_ConsoleStreamSpec(const char *sbName_in)
    : sbName(StrDup(sbName_in)),
      i(Types::uintMax),
      iFilterSource(Types::uintMax), iFilterSink(Types::uintMax),
      eStreamWritePolicy(DC_StreamWritePolicy_RR),
      sbWritePolicyUD(NULL) {
  };
  ~DC_ConsoleStreamSpec(void) {
    TestDeleteNull(sbName);
    TestDeleteNull(sbWritePolicyUD);
    for (Types::uint j=0; j<daWriteRules.NumElem(); ++j)
      delete daWriteRules[j];
  };
};


class DC_ConsoleFilterSpec {
  friend ostream &operator<<(ostream &os, DC_ConsoleFilterSpec &spec);
public:
  // name syntax: <filter-name>[:<copy-set-name>]
  char *sbName;  ///< just the filter name
  class CopySet_t {
  public:
    Types::uint iPlacement;  ///< index into Placement array
    char *sbName;  ///< full name including optional copyset name
    bool fAllowCopies;
    int  nCopies;            ///< number of copies to run there initially
    int  nCopiesLow, nCopiesHigh;
    int  rankTotalCopiesStart;
  };
  DArray<CopySet_t> daCopySet;
  int maxTotalCopies;

  bool fConsole;  ///< T==this is a pseudo-filter at the console
  int  wSinkQueueSize, wWorkQueueSize;

  DArray<DC_ConsoleStreamSpec*> daStreamSpecIn;   ///< inputs to this filter
  DArray<DC_ConsoleStreamSpec*> daStreamSpecOut;  ///< outputs from this filter

  DC_ConsoleFilterSpec(const char *sbName_in)
    : sbName(StrDup(sbName_in)), maxTotalCopies(0) {
    fConsole = (strcmp(sbName_in, "<console>") == 0);
  };
  ~DC_ConsoleFilterSpec(void) {
    TestDeleteNull(sbName);
    for (Types::uint i=0; i<daCopySet.NumElem(); ++i)
      TestDeleteNull(daCopySet[i].sbName);
  };

  CopySet_t *find_placement(Types::uint iPlacement, Types::uint *piFound=NULL);
};


/**
 * Encapsulates the state of a set of filters being executed, and work
 * being directed at them.  This is of use within the Console Process
 * called by the application main() or front end.
 *
 * Multiple instances all exist within a single remote process.  That
 * remote process is assumed to exist before this class can do anything.
 */
class DC_FilterInstance_t {
  friend class DC_FilterService_t;
  friend class DC_ConsoleProcessState;
protected:
  DC_FilterService_t *pDC;
  DC_ConsoleProcessState *pConsole;
  int wInstanceNum;
  char *sbLayoutName;

  /** false if [filter.<console>] is detected in filterlayout, which
   * causes a pseudo-filter to be created for when results are
   * funnelled back through the main console thread.
   */
  bool fDetached;

  /**
  * INIT - after ctor called
  * WAIT - filters instantiated on remote libs, no work to do
  * WORK - processing work
  * STOP - currently being stopped
  * DONE - instance is finished
  */
  enum { INIT, WAIT, WORK, STOP, DONE } eState;

  int wNextWorkNum;  // BUGBUG: int rollover?

  int wRemoteInstances;    ///< total remote processes used for this instance
  int cRemoteInstanceExit; ///< count of remote processes that finished
  //DC_Cond cvDone;

  DArray<DC_ConsoleFilterSpec*> daFilterSpec;
  DArray<DC_ConsoleStreamSpec*> daStreamSpec;

  DC_Mutex mutex;

public:
  /// fDetached == F -> used by console process to recv results
  DC_Filter_Base_t::arg_t arg;

protected:
  Types::uint find_filter(const char *sbName);
  DC_ConsoleStreamSpec *find_stream(const char *sbName, Types::uint &iStream);

  bool _isDetached(void) { return fDetached; };

  void _Delete(void);
  int _Reset(DC_FilterService_t *pDC_in, DC_ConsoleProcessState *pConsole_in);
  int _ProcessLayout(DC_FilterLayout_t &layout);
  int _ProcessPlacement(DC_Placement_t &placement);
  int _StartInstance(void);
  int _StopInstance(void);
  int _WaitUntilInstanceDone(void);

  int _InstanceExit_Reader(void);

public:
  DC_FilterInstance_t(void);
  ~DC_FilterInstance_t(void);

  /// appends the given work to a running instance, returns handle
  DC_WorkHandle_t AppendWork(DC_Work_t &work, bool fWait=true);
};


#endif /* H_DC_INSTANCE */
// ****************************************************************************
