// -*- c++ -*-
// ****************************************************************************
//  Name: dc_sets.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 06/22/2001 : mdb - Created DC_UFSetForest.
//        06/24/2001 : mdb - Added DC_SetForest to allow enumeration of sets.
//        08/03/2001 : mdb - Added move and delete, and worked a bit for speed.
//                           Should combine these into a single class to avoid
//                           uneeded source code complexity.
//  Desc: Simple set classes.
// ****************************************************************************

#if !defined(H_SETS)
#define H_SETS

#include "lib/dc_standard.h"
#include "lib/dc_darray.h"


class DC_SetItem_Base {
  friend class DC_UFSetForest_t;
  friend class DC_SetForest_t;
private:
  DC_SetItem_Base *pSetItem_Parent;
  int wSetItem_Id;
public:
  DC_SetItem_Base(void) : pSetItem_Parent(NULL), wSetItem_Id(-1) {};
};

// ****************************************************************************
//  Class: DC_UFSetForest_t (optimized for Union-Find operations)
// ****************************************************************************

class DC_UFSetForest_t {
protected:
  DArrayPtr<DC_SetItem_Base*> daSetItems;  // all items added to any set
  int wSet;  // assigned from 1.., 0 is invalid

  unsigned int iSetItemsIter;
  int wSetIdIter;  // cached state - setid for current iteration

  // unlink member - remove a single element from its set linkage
  // - if this is the set rep, set *ppItemNewRepOut to the replacement
  int _unlink_member(DC_SetItem_Base &itemDel, bool fDelete,
		     DC_SetItem_Base **ppItemNewRepOut=NULL) {
    unsigned int i;
    DC_SetItem_Base *pItem, *pItemRep, *pItemNewRep;
    if (ppItemNewRepOut) *ppItemNewRepOut = NULL;
    if (&itemDel != itemDel.pSetItem_Parent) {
      // this is the simpler case -- we just check all items and any
      // that have the to-move-item as parent are fixed.
      pItemNewRep = itemDel.pSetItem_Parent;
    } else {  // set rep is the one to move
      // this case is more difficult, because we need to fix all
      // possible descendants to point to another representative which
      // may not exist (set of 1), and even if it does exist we don't
      // know what it will be until a full pass is made.
      pItemNewRep = NULL;
      for (i=0; i<daSetItems.NumElem(); ++i) {
	pItem = daSetItems[i];
	if (pItem == &itemDel) continue;  // skip us
	pItemRep = find_pathcompression(pItem);
	if (pItemRep == &itemDel) {  // in our set
	  pItemNewRep = pItem;
	  // swap id's to avoid messing up iterators
	  int wSetId_Tmp = pItem->wSetItem_Id;
	  pItem->wSetItem_Id = itemDel.wSetItem_Id;
	  itemDel.wSetItem_Id = wSetId_Tmp;
	  break;  // we only need one
	}
      }
      if (pItemNewRep && ppItemNewRepOut) *ppItemNewRepOut = pItemNewRep;
    }
    if (pItemNewRep) {
      // change anyone pointing to us to pItemNewRep
      for (i=0; i<daSetItems.NumElem(); ++i) {
	pItem = daSetItems[i];
	if (pItem->pSetItem_Parent == &itemDel) {
	  pItem->pSetItem_Parent = pItemNewRep;
	}
      }
    }

    if (fDelete) {
      itemDel.pSetItem_Parent = NULL;
      int wRet = itemDel.wSetItem_Id;
      itemDel.wSetItem_Id = -1;  // unassigned
      unsigned int iDel = daSetItems.Find(&itemDel);
      daSetItems.Delete(iDel);
      // update active iterator?
      if (iSetItemsIter > 0 && iSetItemsIter >= iDel) --iSetItemsIter;
      return wRet;
    }
    return 0;
  };

  inline int _find_set(DC_SetItem_Base *pItemRep) {
    return (pItemRep ? pItemRep->wSetItem_Id : -1);
  }

public:
  DC_UFSetForest_t(void) : wSet(0), iSetItemsIter(0), wSetIdIter(0) {};
  ~DC_UFSetForest_t(void) {};

  // make-set
  // rtn value should be saved for future find/union calls
  int make_set(DC_SetItem_Base &item, int wUniqueSetId=-1) {
    daSetItems.AppendByVal(&item);
    item.pSetItem_Parent = &item;
    if (wUniqueSetId != -1) {
      item.wSetItem_Id = wUniqueSetId;
    } else {
      item.wSetItem_Id = wSet;
      if (++wSet < 0) wSet = 0;  // incr with rollover
    }
    return item.wSetItem_Id;
  };

  // find - return representative id
  DC_SetItem_Base *find_pathcompression(DC_SetItem_Base *pItem) {
    if (pItem->pSetItem_Parent && pItem != pItem->pSetItem_Parent) {
      pItem->pSetItem_Parent = find_pathcompression(pItem->pSetItem_Parent);
    }
    return pItem->pSetItem_Parent;
  };
  inline int find_set(DC_SetItem_Base &item) {
    return _find_set(find_pathcompression(&item));
  };
  inline int find_set(DC_SetItem_Base *pItem) {
    return _find_set(find_pathcompression(pItem));
  };

  // union - combine sets such that all have the same representative
  int union_sets(DC_SetItem_Base &item1, DC_SetItem_Base &item2) {
    DC_SetItem_Base *pItemRep1 = find_pathcompression(&item1);
    DC_SetItem_Base *pItemRep2 = find_pathcompression(&item2);
    if (!pItemRep1 || !pItemRep2) return -1;
    pItemRep2->pSetItem_Parent = pItemRep1;
    return pItemRep1->wSetItem_Id;
  };

  // in the same set?
  bool equal_set(DC_SetItem_Base &item1, DC_SetItem_Base &item2) {
    return find_set(item1) == find_set(item2) ? true : false;
  };

  // delete member - remove a single element from its set
  inline int delete_member(DC_SetItem_Base &itemDel) {
    return _unlink_member(itemDel, true);
  };

  // move member - move a single element from one set to another
  int move_member(DC_SetItem_Base &itemMove, DC_SetItem_Base &itemDst) {
    DC_SetItem_Base *pItemRepMove = find_pathcompression(&itemMove);
    DC_SetItem_Base *pItemRepDst = find_pathcompression(&itemDst);
    if (!pItemRepMove || !pItemRepDst) return -1;
    if (pItemRepMove == pItemRepDst) return _find_set(pItemRepDst);  // valid?

    _unlink_member(itemMove, false);
    itemMove.pSetItem_Parent = pItemRepDst;
    return _find_set(pItemRepDst);
  };

  int members_set(int wSetId, DArrayPtr<DC_SetItem_Base*> &da) {
    unsigned int i;
    da.Flush();
    for (i=0; i<daSetItems.NumElem(); ++i) {
      DC_SetItem_Base *pItem = daSetItems[i];
      if (wSetId == find_set(*pItem)) {
	da.Append(pItem);
      }
    }
    return static_cast(int,i);
  };

  // manual iterator (NULL=done)
  // - use this for first call, and next calls
  DC_SetItem_Base *members_iter(int wSetId) {
    if (wSetIdIter == -1 || wSetIdIter != wSetId) {  // first or reset?
      wSetIdIter = wSetId;
      iSetItemsIter = 0;
    }
    while (iSetItemsIter < daSetItems.NumElem()) {
      DC_SetItem_Base *pItem = daSetItems[iSetItemsIter++];
      if (wSetId == find_set(*pItem)) return pItem;
    }
    iSetItemsIter = 0;
    return NULL;  // done
  };

  // manual iterator (NULL=done)
  // - use this for first call, and next calls
  // - this traverses *all* set items regardless of set membership, and
  //   is useful to avoid special cases where the wSetId changes for some
  //   elements when, say the representative is deleted, and the setid
  //   changes for later items in the set.
  DC_SetItem_Base *members_iter(void) {
    if (wSetIdIter == -1) {  // first or reset?
      iSetItemsIter = 0;
    }
    while (iSetItemsIter < daSetItems.NumElem()) {
      DC_SetItem_Base *pItem = daSetItems[iSetItemsIter++];
      return pItem;
    }
    iSetItemsIter = 0;
    return NULL;  // done
  };

  void members_iter_reset(void) {
    wSetIdIter = -1; iSetItemsIter = 0;
  };
};


// ****************************************************************************
//  Class: DC_SetForest_t (more general set class)
// ****************************************************************************


// also keeps a list of set representatives, so they can be enumerated.
class DC_SetForest_t : public DC_UFSetForest_t {
protected:
  DArrayPtr<DC_SetItem_Base*> daSetReps;
  unsigned int iSetRepsIter;

public:
  DC_SetForest_t(void) : iSetRepsIter(0) {};
  ~DC_SetForest_t(void) {};

  // make-set
  // rtn value should be saved for future find/union calls
  int make_set(DC_SetItem_Base &item, int wUniqueSetId=-1) {
    // keep track of all set reps
    daSetReps.AppendByVal(&item);
    return DC_UFSetForest_t::make_set(item, wUniqueSetId);
  };

  // union - combine sets such that all have the same representative
  int union_sets(DC_SetItem_Base &item1, DC_SetItem_Base &item2) {
    DC_SetItem_Base *pItemRep2 = find_pathcompression(&item2);
    // base class will make item1 the rep, so remove item2 from array of reps
    unsigned int iDel = daSetReps.Find(pItemRep2);
    daSetReps.Delete(iDel);
    // update active iterator?
    if (iSetRepsIter > 0 && iSetRepsIter >= iDel) --iSetRepsIter;
    return DC_UFSetForest_t::union_sets(item1, *pItemRep2);
  };

  // delete member - remove a single element from its set
  int delete_member(DC_SetItem_Base &itemDel) {
    DC_SetItem_Base *pItemNewRep;
    int wOldSetId = DC_UFSetForest_t::_unlink_member(itemDel, true,
						     &pItemNewRep);
    Types::uint iDel = daSetReps.Find(&itemDel);
    if (iDel < daSetReps.NumElem()) {  // we were a set rep
      if (pItemNewRep) {
	daSetReps[iDel] = pItemNewRep;  // replace
      } else {
	daSetReps.Delete(&itemDel);  // remove
	// update active iterator?
	if (iSetRepsIter > 0 && iSetRepsIter >= iDel) --iSetRepsIter;
      }
    }
    return wOldSetId;
  };

  // move member - move a single element from one set to another
  int move_member(DC_SetItem_Base &itemMove, DC_SetItem_Base &itemDst) {
    DC_SetItem_Base *pItemRepMove = find_pathcompression(&itemMove);
    DC_SetItem_Base *pItemRepDst = find_pathcompression(&itemDst);
    if (!pItemRepMove || !pItemRepDst) return -1;
    if (pItemRepMove == pItemRepDst) return _find_set(pItemRepDst);  // valid?

    DC_SetItem_Base *pItemNewRep;
    DC_UFSetForest_t::_unlink_member(itemMove, false,
				     &pItemNewRep);
    itemMove.pSetItem_Parent = pItemRepDst;
    Types::uint iDel = daSetReps.Find(&itemMove);
    if (iDel < daSetReps.NumElem()) {  // we were a set rep
      if (pItemNewRep) {
	daSetReps[iDel] = pItemNewRep;  // replace
      } else {
	daSetReps.Delete(&itemMove);  // remove
	// update active iterator?
	if (iSetRepsIter > 0 && iSetRepsIter >= iDel) --iSetRepsIter;
      }
    }
    return _find_set(pItemRepDst);
  }

  // to allow caller to enumerate all sets
  DArrayPtr<DC_SetItem_Base*> &get_sets(void) { return daSetReps; };
  int num_sets(void) { return daSetReps.NumElem(); };

  // allows enumeration of sets, combined with base::members_set()
  int get_sets(DArray<int> &da) {
    unsigned int i;
    da.Flush();
    for (i=0; i<daSetReps.NumElem(); ++i) {
      da.Append(daSetReps[i]->wSetItem_Id);
    }
    return static_cast(int,i);
  };

  // manual iterator (-1=done)
  // - use this for first call, and next calls
  int sets_iter(void) {
    if (iSetRepsIter < daSetReps.NumElem()) {
      return daSetReps[iSetRepsIter++]->wSetItem_Id;
    }
    iSetRepsIter = 0;
    return -1;  // done
  };
  void sets_iter_reset(void) { iSetRepsIter = 0; };
};


#endif /* H_SETS */
// ****************************************************************************
