// ****************************************************************************
//  Name: dird-client.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 09/12/1999 : mdb - Created.
//
//  Desc: Client access routines to directory daemon data.
// ****************************************************************************

#include "lib/dc_standard.h"
#include "lib/dc_argv.h"
#include "lib/dc_configinfo.h"
#include "lib/dc_darray.h"
#include "lib/dc_socket.h"
#include "lib/dc_missing.h"
#include "lib/dc_memory.h"

#include "dird-client.h"
#include "packet.h"

#if defined(DCDBG)
//#define DCDBG_DIRD_PING
//#define DCDBG_DIRD_ADD
//#define DCDBG_DIRD_QUERY
//#define DCDBG_DIRD_SHUTDOWN
#endif /* DCDBG */

/**
 * contact the dird and see if alive
 * rtn: 0=success, !0=failure
 */
int Dird_Ping(char *sbDirdHost, int wDirdPort) {
  InetAddress  addrDird;
  ClientSocket sockDird;

  if (!sbDirdHost) {
    sbDirdHost = getenv("DATACUTTER_DIRDHOST");
    if (!sbDirdHost) sbDirdHost = DIRD_DEFAULT_HOST;
  }
  if (wDirdPort==0) {
    char* sbTmp = getenv("DATACUTTER_DIRDPORT");
    if (!sbTmp) wDirdPort = DIRD_DEFAULT_PORT;
    else wDirdPort = atoi(sbTmp);
  }
  // connect to dird
  if (sockDird.Connect(sbDirdHost, wDirdPort, addrDird) != 0) {
#if defined(DCDBG_DIRD_PING)
    perror("Dird_Ping(): failed Connect()");
    cerr << addrDird << endl;
#endif /* DCDBG_DIRD_PING */
    return -1;
  }
#if defined(DCDBG_DIRD_PING)
  cout << "Dird_Ping(): Connected to dird " << addrDird << endl;
#endif /* DCDBG_DIRD_PING */

  // build packet
  Packet pkt(TAG_DIRD_PING, NULL);

  // send packet to dird
  if (pkt.write(sockDird) <= 0) {
    cout << "Dird_Ping(): Failed write of ping" << endl;
    sockDird.close();
    return -2;
  }
#if defined(DCDBG_DIRD_PING)
  cout << "Dird_Ping(): Sent ping to dird" << endl;
#endif /* DCDBG_DIRD_PING */

  // wait for response
  pkt.setBuf(NULL);
  if (pkt.read(sockDird) <= 0) {
    cout << "Dird_Ping(): Failed read of ack" << endl;
    sockDird.close();
    return -3;
  }
  STestDelete(pkt.getPayload());  // throw away any payload
  if (pkt.getTag() != TAG_DIRD_OK) {
    cout << "Dird_Ping(): Ack indicates failure ("
	 << pkt.getTag() << ")" << endl;
    sockDird.close();
    return -4;
  }
  sockDird.close();
  return 0;
}

/**
 * add the configinfo in the buf to the directory
 */
int Dird_Add(MemoryBuf &buf,
	     char *sbDirdHost, int wDirdPort) {
  InetAddress  addrDird;
  ClientSocket sockDird;

  if (!sbDirdHost) {
    sbDirdHost = getenv("DATACUTTER_DIRDHOST");
    if (!sbDirdHost) sbDirdHost = DIRD_DEFAULT_HOST;
  }
  if (wDirdPort==0) {
    char* sbTmp = getenv("DATACUTTER_DIRDPORT");
    if (!sbTmp) wDirdPort = DIRD_DEFAULT_PORT;
    else wDirdPort = atoi(sbTmp);
  }

  // connect to dird
  if (sockDird.Connect(sbDirdHost, wDirdPort, addrDird) != 0) {
    perror("Dird_Add(): failed Connect()");
    cerr << addrDird << endl;
    return -1;
  }
#if defined(DCDBG_DIRD_ADD)
  cout << "Dird_Add(): Connected to dird " << addrDird << endl;
#endif /* DCDBG_DIRD_ADD */

  // build packet
  Packet pkt(TAG_DIRD_ADD, &buf);

  // send packet to dird
  if (pkt.write(sockDird) <= 0) {
    cout << "Dird_Add(): Failed write of query" << endl;
    sockDird.close();
    return -2;
  }
#if defined(DCDBG_DIRD_ADD)
  cout << "Dird_Add(): Sent add to dird" << endl;
#endif /* DCDBG_DIRD_ADD */

  // wait for response
  pkt.setBuf(NULL);
  if (pkt.read(sockDird) <= 0) {
    cout << "Dird_Add(): Failed read of ack" << endl;
    sockDird.close();
    return -3;
  }
  STestDelete(pkt.getPayload());  // throw away any payload
  if (pkt.getTag() != TAG_DIRD_OK) {
    cout << "Dird_Add(): Ack indicates failure ("
	 << pkt.getTag() << ")" << endl;
    sockDird.close();
    return -4;
  }
  sockDird.close();
  return 0;
}

/**
 * given a query string, this matches ConfigInfo section names in the directory
 * 
 * return: <0=error, else=number of matches
 */
int Dird_Query(char *sbQuery, ConfigInfo &cfg, 
	       char *sbDirdHost, int wDirdPort, int nWait) {
  InetAddress  addrDird;
  ClientSocket sockDird;

  if (!sbDirdHost) {
    sbDirdHost = getenv("DATACUTTER_DIRDHOST");
    if (!sbDirdHost) sbDirdHost = DIRD_DEFAULT_HOST;
  }
  if (wDirdPort==0) {
    char* sbTmp = getenv("DATACUTTER_DIRDPORT");
    if (!sbTmp) wDirdPort = DIRD_DEFAULT_PORT;
    else wDirdPort = atoi(sbTmp);
  }
  if (!sbQuery) return -1;

  // connect to dird
  if (sockDird.Connect(sbDirdHost, wDirdPort, addrDird) != 0) {
    perror("Dird_Query(): failed Connect()");
    cerr << addrDird << endl;
    return -1;
  }
#if defined(DCDBG_DIRD_QUERY)
  cout << "Dird_Query(): Connected to dird " << addrDird << endl;
#endif /* DCDBG_DIRD_QUERY */

  // build packet
  MemoryBuf mb(strlen(sbQuery)+32);
  mb.sprintf("query=%s\n", sbQuery);
  if (nWait) mb.sprintf("wait=%d\n", nWait);
  Packet pkt(TAG_DIRD_QUERY, &mb);

  // send packet to dird
  if (pkt.write(sockDird) <= 0) {
    cout << "Dird_Query(): Failed write of query" << endl;
    sockDird.close();
    return -2;
  }
#if defined(DCDBG_DIRD_QUERY)
  cout << "Dird_Query(): Sent query to dird" << endl;
#endif /* DCDBG_DIRD_QUERY */

  // wait for response
  Packet packetIn;  // no payload buf allocated
  if (packetIn.read(sockDird) < 0) {
    cout << "Dird_Query(): Failed read of query response" << endl;
    sockDird.close();
    return -2;
  }

  cfg.Flush();
  cfg.ReadBuf(packetIn.getPayloadPtr(), packetIn.getSizePayload());
#if defined(DCDBG_DIRD_QUERY)
  cout << "Dird_Query(): query response received ..." << endl << cfg;
#endif /* DCDBG_DIRD_QUERY */

  // cleanup allocated payload
  delete packetIn.getPayload();
  
  sockDird.close();
  return cfg.NumSection();
}

/**
 * add the configinfo in the buf to the directory
 */
int Dird_Remove(MemoryBuf &buf,
		char *sbDirdHost, int wDirdPort) {
  InetAddress  addrDird;
  ClientSocket sockDird;

  if (!sbDirdHost) {
    sbDirdHost = getenv("DATACUTTER_DIRDHOST");
    if (!sbDirdHost) sbDirdHost = DIRD_DEFAULT_HOST;
  }
  if (wDirdPort==0) {
    char* sbTmp = getenv("DATACUTTER_DIRDPORT");
    if (!sbTmp) wDirdPort = DIRD_DEFAULT_PORT;
    else wDirdPort = atoi(sbTmp);
  }

  // connect to dird
  if (sockDird.Connect(sbDirdHost, wDirdPort, addrDird) != 0) {
    perror("Dird_Remove(): failed Connect()");
    cerr << addrDird << endl;
    return -1;
  }
#if defined(DCDBG_DIRD_REMOVE)
  cout << "Dird_Remove(): Connected to dird " << addrDird << endl;
#endif /* DCDBG_DIRD_REMOVE */

  // build packet
  Packet pkt(TAG_DIRD_REMOVE, &buf);

  // send packet to dird
  if (pkt.write(sockDird) <= 0) {
    cout << "Dird_Query(): Failed write of query" << endl;
    sockDird.close();
    return -2;
  }
#if defined(DCDBG_DIRD_REMOVE)
  cout << "Dird_Remove(): Sent remove to dird" << endl;
#endif /* DCDBG_DIRD_REMOVE */

  // wait for response
  pkt.setBuf(NULL);
  if (pkt.read(sockDird) <= 0) {
    cout << "Dird_Ping(): Failed read of ack" << endl;
    sockDird.close();
    return -3;
  }
  STestDelete(pkt.getPayload());  // throw away any payload
  if (pkt.getTag() != TAG_DIRD_OK) {
    cout << "Dird_Ping(): Ack indicates failure ("
	 << pkt.getTag() << ")" << endl;
    sockDird.close();
    return -4;
  }
  sockDird.close();
  return 0;
}

/**
 * contact the dird and if alive, shut it down
 * rtn: 0=success, !0=failure
 */
int Dird_Shutdown(char *sbDirdHost, int wDirdPort) {
  InetAddress  addrDird;
  ClientSocket sockDird;

  if (!sbDirdHost) {
    sbDirdHost = getenv("DATACUTTER_DIRDHOST");
    if (!sbDirdHost) sbDirdHost = DIRD_DEFAULT_HOST;
  }
  if (wDirdPort==0) {
    char* sbTmp = getenv("DATACUTTER_DIRDPORT");
    if (!sbTmp) wDirdPort = DIRD_DEFAULT_PORT;
    else wDirdPort = atoi(sbTmp);
  }

  // connect to daemon
  if (sockDird.Connect(sbDirdHost, wDirdPort, addrDird) != 0) {
    perror("Dird_Shutdown(): failed Connect()");
    cerr << addrDird << endl;
    return -1;
  }
#if defined(DCDBG_DIRD_SHUTDOWN)
  cout << "Dird_Shutdown(): Connected to appd " << addrDird << endl;
#endif /* DCDBG_DIRD_SHUTDOWN */

  // build packet
  Packet pkt(TAG_DIRD_SHUTDOWN, NULL);

  // send packet to daemon
  if (pkt.write(sockDird) <= 0) {
    cout << "Dird_Shutdown(): Failed write of pkt" << endl;
    sockDird.close();
    return -2;
  }
#if defined(DCDBG_DIRD_SHUTDOWN)
  cout << "Dird_Shutdown(): Sent shutdown to dird" << endl;
#endif /* DCDBG_DIRD_SHUTDOWN */

  // wait for response
  pkt.setBuf(NULL);
  if (pkt.read(sockDird) <= 0) {
    cout << "Dird_Shutdown(): Failed read of ack" << endl;
    sockDird.close();
    return -3;
  }
  STestDelete(pkt.getPayload());  // throw away any payload
  if (pkt.getTag() != TAG_DIRD_OK) {
    cout << "Dird_Shutdown(): Ack indicates failure ("
	 << pkt.getTag() << ")" << endl;
    sockDird.close();
    return -4;
  }
  sockDird.close();
  return 0;
}


// ****************************************************************************
