// ****************************************************************************
//  Name: dird.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 08/10/1999 : mdb - Created.
//        09/16/1999 : mdb - Removed checkpointing dir after every entry.
//        11/03/1999 : mdb - Increased static query result buffer to 64k.
//                           (TODO: need to compute size exactly)
//        01/11/2000 : mdb - Added Remove request.
//        04/08/2000 : mdb - Converted to use new ServerLoop_Base.
//        04/17/2000 : mdb - Added multiple queries in one packet.
//        07/13/2000 : mdb - Load dir-load.db on startup, and control load of
//                           dir.db with -restore-state option.
//        11/07/2000 : mdb - Added an -query-wait option that causes the query
//                           to be placed on a pending list and later checked
//                           when an ADD occurs.
//
//  Desc: Simple centralized store of host/daemon/app information.
// ****************************************************************************

#include "lib/dc_standard.h"
#include "lib/dc_argv.h"
#include "lib/dc_configinfo.h"
#include "lib/dc_socket.h"
#include "lib/dc_missing.h"
#include "lib/dc_memory.h"
#include "lib/dc_lists.h"

#include "dird.h"
#include "server-loop-base.h"
#include "packet.h"

#ifdef _WIN32
#include <signal.h>
#include <process.h>
#include "win32.h"
#endif

#if defined(DCDBG)
//#define DCDBG_DIRD_QUERY
//#define DCDBG_DIRD_PENDING
#endif /* DCDBG */


//#define DIRD_CHECKPOINT_ADD
//#define DIRD_CHECKPOINT_REMOVE

struct Global_s {
  bool fVerbose;  // CLH - added to reduce messages
  int fDebug;    // CLH - added to allow debugging at runtime
  char *sbVersion;
  Argv       a;

  char *sbHost;
  int wPort;

  ConfigInfo cfgDir;  // the current "directory"

  int fdSigPipeWrite;
} Global;


// currently we only handle exact matches "blah" or queries with a single
// '*' at the end "blah.*"
bool DirdQuery_IsMatch(const char *sbQuery, const char *sbSection) {
  size_t wLenQuery;
  if (strcmp(sbQuery, sbSection) == 0) return true;  // exact match
  wLenQuery = strlen(sbQuery);
  if (wLenQuery > 0) {
    if (sbQuery[wLenQuery-1] == '*') {
      if (strncmp(sbQuery, sbSection, wLenQuery-1) == 0) return true;  // match
    }
  }
  return false;
}


// look through the configinfo directory and copy matches into the buffer
// return: number of sections (and all entries) written
int DirdQuery(ConfigInfo &cfgQuery, MemoryBuf &mb, int nWait) {
  const char *sbNameQuery, *sbQuery;
  const char *sbSectionDir, *sbNameDir;
  int wRtn = 0;

  // the query cfg is expected to contain only the global section, and
  // entries of the form query=<text>, with an optional wait=<n> flag
  // which is the minimum number of sections to return as a result.
  if (! cfgQuery.FirstSection()) return 0;
  for (sbNameQuery = cfgQuery.FirstName() ;
       sbNameQuery ;
       sbNameQuery = cfgQuery.NextName()) {
    if (strcmp(sbNameQuery, "query") != 0) continue;
    sbQuery = cfgQuery.CurrentValue();

#if defined(DCDBG_DIRD_QUERY)
    cout << "DirdQuery(): query for \"" << sbQuery << "\" ..." << endl;
#endif /* DCDBG_DIRD_QUERY */
    sbSectionDir = Global.cfgDir.FirstSection();
    while (sbSectionDir != NULL) {
      if (DirdQuery_IsMatch(sbQuery, sbSectionDir)) {
#if defined(DCDBG_DIRD_QUERY)
	cout << "DirdQuery(): matches with \"" << sbSectionDir << "\"" << endl;
#endif /* DCDBG_DIRD_QUERY */
	// output this section
	mb.sprintf("[%s]\n", sbSectionDir);
	sbNameDir = Global.cfgDir.FirstName();
	while (sbNameDir != NULL) {
	  mb.sprintf("%s=%s\n", sbNameDir, Global.cfgDir.CurrentValue());
	  sbNameDir = Global.cfgDir.NextName();
	}
	++wRtn;
      }
      sbSectionDir = Global.cfgDir.NextSection();
    }
  }
#if defined(DCDBG_DIRD_QUERY)
  cout << "DirdQuery(): done, rtn=" << wRtn << " sections." << endl;
#endif /* DCDBG_DIRD_QUERY */
  return wRtn;
}


char bufPacketIn[KB_4];
MemoryBuf mbPacketIn(bufPacketIn, KB_4, 0, false);  // precompute for speed

char bufPacketOut[KB_64];  // BUGBUG: may not be enough for big queries
MemoryBuf mbPacketOut(bufPacketOut, KB_64, 0, false);


// for keeping track of pending queries
typedef struct pending_s {
  int        i;    // connection query came from
  int        n;    // number expected for this query
  ConfigInfo cfg;  // query
} pending_t;


class DirdServerLoop : public ServerLoop_Base {
public:
  DLList<pending_t*> listPending;

public:
  int ClientProcess(int i) {
    int wTag, n, wTmp;
    Packet packetIn(&mbPacketIn);
    Packet packetOut(&mbPacketOut);
    ConfigInfo cfgTmp;
    char *sbTmp;

    if (packetIn.read(*(rgClients[i].psock)) <= 0) {
      return SLE_CB_CLOSECONN;  // some error reading the client socket
    }

    //char *sbPayload = bufPacketIn;
    //char *sbPayload = packetIn.getPayload()->getPtr();

    wTag = packetIn.getTag();
    switch (wTag) {
    case TAG_DIRD_PING:
      if (Global.fVerbose) 
	cout << "PING," << *(rgClients[i].paddr) << endl;
      // send the ack
      mbPacketOut.Empty();
      packetOut.setTag(TAG_DIRD_OK);
      if (packetOut.write(*(rgClients[i].psock)) <= 0) {
	return SLE_CB_CLOSECONN;  // some error writing the client socket
      }
      break;

    case TAG_DIRD_SHUTDOWN:
      if (Global.fVerbose) 
	cout << "SHUTDOWN," << *(rgClients[i].paddr) << endl;
      // send the ack
      mbPacketOut.Empty();
      packetOut.setTag(TAG_DIRD_OK);
      packetOut.write(*(rgClients[i].psock));
      return SLE_CB_SHUTDOWN;
      break;

    case TAG_DIRD_ADD:
      if (Global.fVerbose) 
	cout << "ADD," << *(rgClients[i].paddr) << endl;
      //cout << "cfgDir was ..." << endl << Global.cfgDir;
      cfgTmp.ReadBuf(packetIn.getPayloadPtr(),
		     packetIn.getSizePayload());
      if (Global.fVerbose) 
	cout << cfgTmp;
      Global.cfgDir.Merge(cfgTmp);
      //cout << "cfgDir is now ..." << endl << Global.cfgDir;
      if (Global.fVerbose) 
	cout << "directory now has " << Global.cfgDir.NumSection() << " entries"
	     << endl;
      cfgTmp.Flush();
#if defined(DIRD_CHECKPOINT_ADD)
      cout << "Checkpointing directory data to " << DIRD_FILE << endl;
      Global.cfgDir.WriteFile(DIRD_FILE);
#endif /* DIRD_CHECKPOINT_ADD */

      // send the ack
      mbPacketOut.Empty();
      packetOut.setTag(TAG_DIRD_OK);
      if (packetOut.write(*(rgClients[i].psock)) <= 0) {
	return SLE_CB_CLOSECONN;  // some error writing the client socket
      }

      // now check if any of the pending entries have been satisfied
      {
	int n=listPending.StartIter();
	pending_t *p;
	for (int j=0; j<n; ++j) {
	  p = listPending.Next();
	  
	  // check if this one is satisfied by the ADD
	  mbPacketOut.Empty();
	  wTmp = DirdQuery(p->cfg, mbPacketOut, p->n);
	  if (wTmp >= p->n) {
	    if (Global.fVerbose) 
	      cout << "PENDING,finished pending query" << endl;
#if defined(DCDBG_DIRD_PENDING)
	    cout << "pending cfg: " << p->cfg;
#endif /* DCDBG_DIRD_PENDING */
	    // send the result
	    packetOut.setTag(TAG_DIRD_OK);
	    if (packetOut.write(*(rgClients[p->i].psock)) <= 0) {
	      //return SLE_CB_CLOSECONN;  // error writing to client socket
	    }
	    // remove from list
	    listPending.DelIter();
	    delete p;
	  } else {
#if defined(DCDBG_DIRD_PENDING)
	    cout << "PENDING,skipping unfinished pending query." << endl;
#endif /* DCDBG_DIRD_PENDING */
	    mbPacketOut.Empty();
	  }
	}
      }
      break;

    case TAG_DIRD_REMOVE:
      if (Global.fVerbose) 
	cout << "REMOVE," << *(rgClients[i].paddr) << endl;
      //cout << "cfgDir was ..." << endl << Global.cfgDir;
      // copy chars out of buf into a zero term string
      sbTmp = new char[packetIn.getSizePayload()+1];
      memcpy(sbTmp, packetIn.getPayloadPtr(), packetIn.getSizePayload());
      sbTmp[packetIn.getSizePayload()] = '\0';
      n = Global.cfgDir.RemoveSection(sbTmp);
      delete [] sbTmp;
      //cout << "cfgDir is now ..." << endl << Global.cfgDir;
      if (Global.fVerbose) 
	cout << "directory now has " << Global.cfgDir.NumSection() << " entries"
	     << endl;
#if defined(DIRD_CHECKPOINT_REMOVE)
      cout << "Checkpointing directory data to " << DIRD_FILE << endl;
      Global.cfgDir.WriteFile(DIRD_FILE);
#endif /* DIRD_CHECKPOINT_REMOVE */

      // send the ack
      mbPacketOut.Empty();
      mbPacketOut.sprintf("num_removed=%d\n", n);
      packetOut.setTag(TAG_DIRD_OK);
      if (packetOut.write(*(rgClients[i].psock)) <= 0) {
	return SLE_CB_CLOSECONN;  // some error writing the client socket
      }
      break;

    case TAG_DIRD_QUERY:
      if (Global.fVerbose)
	cout << "QUERY," << *(rgClients[i].paddr) << ",";
      cfgTmp.ReadBuf(packetIn.getPayloadPtr(), packetIn.getSizePayload());

      // NOTE: wait count applies to *all* the query= lines in the packet.
      n = Atoi(cfgTmp.Find("wait"));  // check for wait flag
#if defined(DCDBG_DIRD_QUERY)
      cout << "QUERY: wait=" << n << endl;
#endif /* DCDBG_DIRD_QUERY */

      mbPacketOut.Empty();
      wTmp = DirdQuery(cfgTmp, mbPacketOut, n);
      if (wTmp >= n) {
	cfgTmp.Flush();
	if (Global.fVerbose) 
	  cout << ',' << mbPacketOut.getSize() << " bytes" << endl;
	// send the result
	packetOut.setTag(TAG_DIRD_OK);
	if (packetOut.write(*(rgClients[i].psock)) <= 0) {
	  return SLE_CB_CLOSECONN;  // some error writing the client socket
	}
      } else {
	if (Global.fVerbose) 
	  cout << ", pending (found " << wTmp << '/' << n << ") ..." << endl;
	// add to the pending list
	pending_t *p = new pending_t;
	p->i = i;
	p->n = n;
	p->cfg.Merge(cfgTmp);
	listPending.InsTailByVal(p);

	cfgTmp.Flush();
	mbPacketOut.Empty();
      }
      break;

    default:
      cerr << ErrorHeader(wERROR)
	   << "ClientProcess(): Ignoring packet with unknown tag #"
	   << wTag << "." << endl
	   << packetIn << endl;
      // send the ack
      mbPacketOut.Empty();
      packetOut.setTag(TAG_DIRD_ERROR);
      if (packetOut.write(*(rgClients[i].psock)) <= 0) {
	return SLE_CB_CLOSECONN;  // some error writing the client socket
      }
    }
    return SLE_CB_OK;
  }

  int MapProcess(int i, enum eMapReason e) {
    switch (e) {
    case ADD_CONN:
      break;
    case DISABLE_CONN:
      break;
    case ENABLE_CONN:
      break;
    case CLOSE_CONN:
      // check for any orphaned entries in pending list and delete them
      // now check if any of the pending entries have been satisfied
      {
	int n=listPending.StartIter();
	pending_t *p;
	for (int j=0; j<n; ++j) {
	  p = listPending.Next();
	  if (p->i == i) {
#if defined(DCDBG_DIRD_PENDING)
	    cout << "MapProcess(): discarding pending entry (i="
		 << p->i << ",n=" << p->n << ")" << endl;
	    cout << "pending cfg: " << p->cfg;
#endif /* DCDBG_DIRD_PENDING */

	    listPending.DelIter();
	    delete p;
	  }
	}
      }
      break;
    }
    return SLE_CB_OK;
  }

  // BUGBUG: How to realize which socket SIGPIPE belongs to?  Maybe we don't
  // have to know, since select() will later return with activity, and will be
  // closed.

  int SigProcess(int pipeval) {  // called with signals blocked
#ifndef _WIN32
	if (pipeval == SIGHUP || pipeval == SIGINT) {
      cout << "SigProcess(): SLE_CB_SHUTDOWN returned due to";
      if (pipeval == SIGHUP) cout << " SIGHUP" << endl;
      if (pipeval == SIGINT) cout << " SIGINT" << endl;
      return SLE_CB_SHUTDOWN;
    }
#endif
    cout << "SigProcess(): SLE_CB_OK returned." << endl;
    return SLE_CB_OK;
  }
};


// connect to the required socket and send the given init ack id to
// let some higher level entity know we have initialized.
int InitAck(char *sbInitAck) {
  InetAddress  addr;
  ClientSocket sock;
  char *sbHost, *p;
  int   wPort, wInitId;
  int   wRtn = 0;
  
  if (!sbInitAck) return 0;  // cmdline option not given
  sbHost = StrDup(sbInitAck);
  if (!(p = strchr(sbHost, ':'))) return 1;  // port must be specified
  *p = '\0';
  wPort = atoi(++p);
  if (!(p = strchr(p, ','))) return 1;  // init ack id must be specified
  *p = '\0';
  wInitId = atoi(++p);
  if (sock.Connect(sbHost, wPort, addr) != 0) wRtn = 2;
  Packet packetOut(TAG_BOOTUTIL_INITACK, NULL);
  packetOut.setExtra(wInitId);
  packetOut.write(sock);
  sock.Close();
  delete [] sbHost;
  return wRtn;
}


void usage(char *sb = NULL) {
  if (sb) {
    cerr << ErrorHeader(wERROR) << sb << endl;
  }
  cerr << "usage: " << sbProgramName
       << " [-port <port>] [-init-ack <host>[:port]] [-verbose]" << endl;
}

#ifndef _WIN32
static void sig_int(int, siginfo_t *, void *) {
  //cout << "sig_int(): Handler called ..." << endl;
  int n = SIGINT;
  if (Global.fdSigPipeWrite != -1)
    write(Global.fdSigPipeWrite, &n, sizeof(n));  // signal to select()
}

static void sig_hup(int, siginfo_t *, void *) {
  //cout << "sig_hup(): Handler called ..." << endl;
  int n = SIGHUP;
  if (Global.fdSigPipeWrite != -1)
    write(Global.fdSigPipeWrite, &n, sizeof(n));  // signal to select()
}

static void sig_pipe(int, siginfo_t *, void *) {
  //cout << "sig_pipe(): Handler called ..." << endl;
  // just return, we don't need to inform select explicitly
}
#endif

int Init(void) {
  InetAddress  addrLocal;
  ServerSocket sockLocal;
  char *sb;

  // CLH - added verbose flag and debug level
  Global.fVerbose = Global.a.Find("-verbose");
  //if (Global.fVerbose)
  cout << Global.sbVersion << endl;

  bool fRestoreState = Global.a.NextFind("-restore-state");

  if ((sb = Global.a.NextFind("-port")) == NULL) {
    Global.wPort = DIRD_DEFAULT_PORT;
  } else {
    Global.wPort = atoi(sb);
  }
  char *sbInitAck = Global.a.NextFind("-init-ack");

  // CLH - added verbose flag and debug level
  Global.fVerbose = Global.a.NextFind("-verbose");
  if ((sb = Global.a.NextFind("-debug")) == NULL) {
    Global.fDebug = 0;
  } else {
    Global.fDebug = atoi(sb);
  }

#ifndef _WIN32

  // install sig handlers
  Global.fdSigPipeWrite = -1;
  struct sigaction s;
  s.sa_sigaction = sig_int;
  sigemptyset(&s.sa_mask);
  s.sa_flags = SA_SIGINFO;
  if (sigaction(SIGINT, &s, NULL) != 0) {
    perror("Cannot install SIGINT handler");
    exit(1);
  }
  s.sa_sigaction = sig_hup;
  sigemptyset(&s.sa_mask);
  s.sa_flags = SA_SIGINFO;
  if (sigaction(SIGHUP, &s, NULL) != 0) {
    perror("Cannot install SIGHUP handler");
    exit(1);
  }
  s.sa_sigaction = sig_pipe;
  sigemptyset(&s.sa_mask);
  s.sa_flags = SA_SIGINFO;
  if (sigaction(SIGPIPE, &s, NULL) != 0) {
    perror("Cannot install SIGPIPE handler");
    exit(1);
  }

#endif

  // read in current directory state
  Global.cfgDir.SetOpts(ConfigInfo::ADD_DUPSEC_REPLACE |
			ConfigInfo::OUT_EXPAND_EACH |
			ConfigInfo_Section::ADD_DUPNAME_APPEND);
  Global.cfgDir.ReadFile(DIRD_FILE_LOAD);
  if (Global.cfgDir.NumSection() > 0) {
    //cout << "DIRD_FILE_LOAD: " << Global.cfgDir << endl;
    if (Global.fVerbose)
      cout << "pre-load complete (" << Global.cfgDir.NumSection()
	   << ") from " << DIRD_FILE_LOAD << endl;
  }
  if (fRestoreState) {
    Global.cfgDir.ReadFile(DIRD_FILE, false);
  }
  if (Global.cfgDir.NumSection() > 0) {
    if (Global.fVerbose)
      cout << "restoring previous state ("
	   << Global.cfgDir.NumSection() << ") from " << DIRD_FILE << endl;
  }

  bool fOk = false;
  if (addrLocal.Set(Global.wPort) == 0) {
    Global.sbHost = addrLocal.getHostName();

    if (sockLocal.CreateBindListen(addrLocal) == 0) {
      Global.wPort = addrLocal.getPort();
      if (Global.fVerbose)
	cout << "pid=" << getpid() << ", listening on " << addrLocal << endl;
      fOk = true;
    }
  }
  if (!fOk) {
    cerr << ErrorHeader(wERROR) << "Init(): Failed setup of local port." << endl;
    return -1;
  }

  InitAck(sbInitAck);

  // setup server loop
  DirdServerLoop SL;
  SL.Init(sockLocal);

#ifndef _WIN32
  sigset_t maskSigAny;
  sigemptyset(&maskSigAny);
  sigaddset(&maskSigAny, SIGHUP);
  sigaddset(&maskSigAny, SIGINT);
  Global.fdSigPipeWrite = SL.setSig(maskSigAny);
#endif

  while (1) {
    switch (SL.OnePass()) {

#ifdef _WIN32
    case ServerLoop_Base::SLE_ERROR_WIN:
#else
    case ServerLoop_Base::SLE_ERROR:
#endif
      cerr << ErrorHeader(wERROR) << "failed OnePass() call." << endl;
      // what else to do?
      break;
    case ServerLoop_Base::SLE_SHUTDOWN:
      sockLocal.Close();
      Global.cfgDir.WriteFile(DIRD_FILE);
      return 0;
      break;
    }
  }
  // never get here
  return 0;
}

int main(int argc, char *argv[]) {

// For WIN32, we must initialize winsock
#ifdef WIN32

  WORD wVersionRequested;
  WSADATA wsaData;
  int err;
 
  wVersionRequested = MAKEWORD(2, 0);
 
  err = WSAStartup( wVersionRequested, &wsaData );
  if ( err != 0 ) {
	 cout << "Error, Winsock could not be initialized.";
     return 0;
  }

#endif


  Global.sbVersion = "DataCutter dird v1.1 (08/08/2002)";

  (void) StandardInit(argc, argv);
  Global.a.Set(argc, argv);
  Init();
  (void) StandardCleanup();
  cout.flush();
  cerr.flush();

  return 0;
}


// ****************************************************************************
