// ****************************************************************************
//  Name: hostinfo.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 04/12/2000 : mdb - Created.
//        07/13/2000 : mdb - Moved rules into [host.info] section so there is
//                           just a single Dird_Query() call.
//        09/28/2000 : mdb - Fixed 2 conditional memleaks in ConnectName.
//        03/13/2001 : mdb - Fixed cut/paste bug in ConnectName where invalid
//                           loop bound for pdaAliases was used.
//        05/07/2001 : mdb - Converted to use seperate DC_Regex.
//        07/19/2001 : mdb - Changed Init() to add an alias for canonical name.
//  Desc: 
// ****************************************************************************
#include "hostinfo.h"
#include "lib/dc_str.h"
#include "lib/dc_regex.h"
#include "../src/dird-client.h"

#ifdef _WIN32
#include "win32.h"
#include <winsock.h>
#else
#include <netdb.h>
#endif

#include <stdio.h>
#include <regex.h>

#if !defined(MAXHOSTNAMELEN)
#  define MAXHOSTNAMELEN 256
#endif


#if defined(DCDBG)
//#define DCDBG_RE_MATCH
//#define DCDBG_HOST_RESOLVE
#endif /* DCDBG */


/**
 * Check if string matches the extended regular expression
 * rtn: 1==match, otherwise 0 (no match, errors)
 */
int HostInfo::re_match(const char *sbString, char *sbPattern) {
  DC_Regex re;
#if defined(DCDBG_RE_MATCH)
  cerr << "re_match(): attempting match of '" << sbString << "' and '"
       << sbPattern << "'" << endl;
#endif /* DCDBG_RE_MATCH */
  if (re.setPattern(sbPattern) != 0 ||
      !re.match(sbString)) {
    return 0;
  }
  return 1;  // match
}


int HostInfo::Init(void) {
  // ASSUME: in the current directory
  if (InitFromFile("host.cfg") == 0) {
    InitFromDird();
  }
  //cout << "BEFORE:" << cfg;
  // turn canonical name in section name into an alias to be used for re match
  const char *sbSection, *sbName;
  char *sb;
  size_t len;
  bool fFound;
  sbSection = cfg.FirstSection();
  while (sbSection != NULL) {
    if (strncmp(sbSection, "host.info.\"", 11) == 0) {
      len = strlen(sbSection)-11-1;
      fFound = false;
      sbName = cfg.FirstName();
      while (sbName != NULL) {
	if (strcmp(sbName, "alias") == 0 &&
	    strncmp(sbSection+11, cfg.CurrentValue(), len) == 0) {
	  fFound = true;
	  break;
	}
	sbName = cfg.NextName();
      }
      if (!fFound) {  // add a new entry here
	sb = new char[len+1];
	memcpy(sb, sbSection+11, len);
	sb[len] = '\0';
	cfg.AddEntry(const_cast(char*,sbSection), "alias", sb);
	delete [] sb;
      }
    }
    sbSection = cfg.NextSection();
  }
  //cout << "AFTER:" << cfg;
  return cfg.NumSection();
}

int HostInfo::InitFromDird(char *sbHost, int wPort) {
  ConfigInfo cfgTmp;
  char buf[64];
  snprintf(buf, sizeof(buf), "host.info*");
  if (Dird_Ping(sbHost, wPort) == 0 &&
      Dird_Query(buf, cfgTmp, sbHost, wPort) > 0) {
    cfg.Merge(cfgTmp);
  }
  return cfg.NumSection();
}

int HostInfo::InitFromFile(char *sbFile) {
  ConfigInfo cfgTmp;
  cfgTmp.ReadFile(sbFile);
  cfg.Merge(cfgTmp);
  return cfg.NumSection();
}

/**
 * given a hostname, find the canonical name.
 * This can be inefficient, since we are searching through all the entries.
 * return value is allocated
 */
char *HostInfo::CanonicalName_Alloc(char *sbHost, bool fGetHostByName) {
  const char *sbSection, *sbName;
  char *sb;

  if (fGetHostByName) {
    struct hostent *phost = gethostbyname(sbHost);
    if (phost) sbHost = phost->h_name;
  }

  sbSection = cfg.FirstSection();
  while (sbSection != NULL) {
    if (strncmp(sbSection, "host.info.\"", 11) == 0) {
      sbName = cfg.FirstName();
      while (sbName != NULL) {
	if (strcmp(sbName, "alias") == 0 &&
	    strcmp(sbHost, cfg.CurrentValue()) == 0) {
	  size_t len = strlen(sbSection)-11-1;
	  sb = new char[len+1];
	  memcpy(sb, sbSection+11, len);
	  sb[len] = '\0';
	  return sb;
	}
	sbName = cfg.NextName();
      }
    }
    sbSection = cfg.NextSection();
  }
  return StrDup(sbHost);
}

/**
 * inputs are expected to be canonical hostnames
 * rtn: string is not allocated, and points to the cfg alias entry
 */
char *HostInfo::ConnectName(char *sbHostFrom, char *sbHostTo) {
#if defined(DCDBG_HOST_RESOLVE)
  cout << "ConnectName(): entry from=" << sbHostFrom
       << " to=" << sbHostTo << endl;
#endif /* DCDBG_HOST_RESOLVE */
  DArray<char*> *pdaRules = cfg.FindAll("host.info", "rule");
  for (unsigned int i=0; i<pdaRules->NumElem(); ++i) {
#if defined(DCDBG_HOST_RESOLVE)
    cout << "ConnectName(): daRules[" << i << "] = >" << pdaRules->Get(i)
	 << "<" << endl;
#endif /* DCDBG_HOST_RESOLVE */
    Str str(pdaRules->Get(i));
    DArray<char*> *pdaTokens = str.TokenArray();
    if (pdaTokens->NumElem() < 3) {
      cerr << "ConnectName(): invalid rule \""
	   << pdaRules->Get(i) << "\"" << endl;
      continue;
    }
    // now check if the current host pair match this rule
    if (re_match(sbHostFrom, pdaTokens->Get(0)) &&
	re_match(sbHostTo, pdaTokens->Get(1))) {
#if defined(DCDBG_HOST_RESOLVE)
      cout << "ConnectName(): *** Found rule match" << endl;
#endif /* DCDBG_HOST_RESOLVE */
      // find alias that matches this rule
      char buf[MAXHOSTNAMELEN+16];
      snprintf(buf, sizeof(buf), "host.info.\"%s\"", sbHostTo);
      DArray<char*> *pdaAliases = cfg.FindAll(buf, "alias");
      for (unsigned int j=0; j<pdaAliases->NumElem(); ++j) {
	char *sbRtn = pdaAliases->Get(j);
	if (re_match(sbRtn, pdaTokens->Get(2))) {
	  // found the correct alias to use
#if defined(DCDBG_HOST_RESOLVE)
	  cout << "ConnectName(): matched rule and found an alias to use"
	       << endl
	       << "  - rule:  " << pdaRules->Get(i) << endl
	       << "  - from:  " << sbHostFrom << endl
	       << "  - to:    " << sbHostTo << endl
	       << "  - alias: " << sbRtn << endl;
#endif /* DCDBG_HOST_RESOLVE */
	  delete pdaAliases;
	  Str::FlushTokenArray(pdaTokens);
	  delete pdaTokens;
	  delete pdaRules;
	  return sbRtn;
	}
      }
      // we matched the rule, but cannot find a suitable alias
      cerr << "ConnectName(): matched rule but cannot find matching alias"
	   << endl
	   << "  - rule: " << pdaRules->Get(i) << endl
	   << "  - from: " << sbHostFrom << endl
	   << "  - to:   " << sbHostTo << endl;
      if (pdaAliases) delete pdaAliases;
      break;
    }
    Str::FlushTokenArray(pdaTokens);
    delete pdaTokens;
  }
  if (pdaRules) delete pdaRules;

  return sbHostTo;
}

/**
 * input is not expected to be a canonical hostname
 * assume "from host" is the host we are executing on
 * rtn: string is not allocated, and points to the cfg alias entry
 */
char *HostInfo::ConnectNameTo_Alloc(char *sbHostTo) {
  char buf[MAXHOSTNAMELEN];
  gethostname(buf, sizeof(buf));
  char *sbFrom = CanonicalName_Alloc(buf, false);
  char *sbTo = CanonicalName_Alloc(sbHostTo, true);
  char *sbRtn = StrDup(ConnectName(sbFrom, sbTo));
  delete [] sbFrom;
  delete [] sbTo;
  return sbRtn;
}

/**
 * assume "to host" is the host we are executing on
 */
char *HostInfo::ConnectNameFrom_Alloc(char *sbHostFrom) {
  char buf[MAXHOSTNAMELEN];
  gethostname(buf, sizeof(buf));
  char *sbFrom = CanonicalName_Alloc(sbHostFrom, true);
  char *sbTo = CanonicalName_Alloc(buf, false);
  char *sbRtn = StrDup(ConnectName(sbFrom, sbTo));
  delete [] sbFrom;
  delete [] sbTo;
  return sbRtn;
}


// ****************************************************************************
