package dc.jfilters;

//
//  A DC_Buffer_t is the unit of transfer for streams.  This class is
//  simply a container for a chunk of contiguous memory, which can be
//  statically or dynamically allocated.  These buffers have a maximum
//  size, and keep track of how much of that maximum is occupied.
//  Data that will fit into the unused portion of the buffer can be
//  appended.  Extraction occurs from the head of the buffer,
//  maintaining a extraction pointer for multiple extraction
//  operations.  An application can also just get a pointer and length
//  of the data in the buffer, and use it in place.
//
//  When finished with a buffer, the consume() method should be
//  called, which may free the space.  Heap allocated memory is
//  handled correctly based on the constructor used.  Generally
//  avoiding static DC_Buffer_t objects with the use of consume() is
//  advised, unless setConsume(true) is explicitly called.
//

public class DC_Buffer_t {
    //  friend ostream &operator<<(ostream &os, DC_Buffer_t &buf);
    //  friend class DC_PipeInStream_t;
    //  friend class DC_PipeOutStream_t;
    //  friend class DC_ReaderServerLoop;
    //  friend class DC_StreamBufferQueue;
    //  friend class DC_RemoteOutStreamCopy;

    public DC_Buffer_t getPNext() {
        _getPNext();

        return pNext;
    }

    public void setPNext(DC_Buffer_t newPNext) {
        if (newPNext != null) {
            pNext = newPNext;
            _setPNext(pNext);
        }
        else {
            pNext = null;
        }
    }


  /* Default constructor */
    public DC_Buffer_t() {
        this(true);
    }

   /* Private constructor to create just the skeleton class in Java */
   private DC_Buffer_t(boolean constructC) {
        if (constructC == true) {
	     System.out.println("Created a C++ class in BUffer");
            _initialize();
        }
   }

    public DC_Buffer_t(int wMax) {
        _initialize(wMax);
    }

  /* Clone current buffer */
    public Object clone()
    throws CloneNotSupportedException {
        DC_Buffer_t buff = (DC_Buffer_t) super.clone();
        _clone(buff);
        return buff;
    }

    protected void finalize() {
        _delete();
    }


    //
    // filter objects can call the following fcns
    //

    // called on a buffer returned from read that a filter will not
    // need any longer.
    // (depending how memory was allocated for the buffer data area,
    // this may or may not free the actual buffer space)

    public void consume() {
        _consume();
    }

    public boolean getConsume() {
        return _getConsume();
    }

    public void setConsume(boolean newFConsume) {
        _setConsume(newFConsume);
    }

    // Note: use of fConsume mimics (exactly?) fFree in MemoryBuf base class?!?


    // *** These fcns come from the base class ***
    //char *New(Types::uint wMax_in);
    public byte[] New(long wMax_in) throws IllegalArgumentException {
        if (wMax_in < 0) {
            throw new IllegalArgumentException("wMax_in must be positive in DC_Buffer_t.New");
        }
        return _new(wMax_in);
    }

    //void Delete(void);
    public void Delete() {
        _delete_funct();
    }

    //void Set(char *pBuf_in, int wMax_in, int wSize_in, bool fFree_in);
    public void Set(byte[] pBuf, int wMax_in, int wSize, boolean fFree) {
        _set(pBuf, wMax_in, wSize, fFree);
    }

    //void Empty(void);
    public void Empty() {
        _empty();
    }

    //Types::uint getMax(void);      // total size of buf
    public long getMax() {
        return _getMax();
    }

    //char *getPtr(void);      // to beginning
    public byte[] getBuffer() {
        return _getBuffer();
    }

    //Types::uint getSize(void);     // amount used
    public long getSize() {
        return _getSize();
    }

    //Types::uint getFree(void);     // free space at end
    public long getFree() {
        return _getFree();
    }

    //char *getPtrFree(void);  // to free space
    public byte[] getFreeBuffer() {
        return _getFreeBuffer();
    }

    //char *getPtrExtract(void);
    public byte[] getExtractBuffer() {
        return _getExtractBuffer();
    }

    //Types::uint getSizeExtract(void); // amt left to extract
    public long getSizeExtract() {
        return _getSizeExtract();
    }

    //Types::uint resetExtract(Types::uint w=0);
    public long resetExtract() {
        return resetExtract(0);
    }

    public long resetExtract(long w) throws IllegalArgumentException {
        if (w < 0) {
            throw new IllegalArgumentException("w must be positive in DC_Buffer_t.resetExtract(long)");
        }
        return _resetExtract(w);
    }

    //Types::uint setSize(Types::uint wSize_in);
    public long setSize(long wSize_in) throws IllegalArgumentException {
        if (wSize_in < 0) {
            throw new IllegalArgumentException("wSize_in must be positive in DC_Buffer_t.setSize");
        }
        return _setSize(wSize_in);
    }

    //char *Append(char *p, Types::uint size);
    public byte[] Append(byte[] p, long size) throws IllegalArgumentException {
        if (size < 0) {
            throw new IllegalArgumentException("size must be positive in DC_Buffer_t.Append");
        }
        else if (size > p.length) {
            throw new IllegalArgumentException("size must be smaller than " + p.length + " in DC_Buffer_t.Append");
        }
        return _append(p, size);
    }

    //how to implement in java???
    //int sprintf(char *format, ...);


    //int Extract(char *p, Types::uint size);
    public int Extract(byte[] p, long size) {
        if (size < 0) {
            throw new IllegalArgumentException("size must be positive in DC_Buffer_t.Extract");
        }
        else if (size > p.length) {
            throw new IllegalArgumentException("size must be smaller than p.length in DC_Buffer_t.Extract");
        }
        return _extract(p, size);
    }

    //NATIVE FOUNDATION BELOW
    private native byte[] _new(long wMax_in);
    private native void _delete_funct();
    private native void _set(byte[] pBuf, int wMax_in, int wSize, boolean fFree);
    private native void  _empty();
    private native long _getMax();
    private native byte[] _getBuffer();
    private native long _getSize();
    private native long _getFree();
    private native byte[] _getFreeBuffer();
    private native byte[] _getExtractBuffer();
    private native long _getSizeExtract();
    private native long _resetExtract(long w);
    private native byte[] _append(byte[] p, long size);
    private native int _extract(byte[] p, long size);
    private native int _setSize(long size);

    private DC_Buffer_t pNext = null;
    private native void _getPNext();
    private native void _setPNext(DC_Buffer_t newPNext);

    private native void _initialize();
    private native void _initialize(int wMax);
    private native void _clone(DC_Buffer_t buff);
    private native void _delete();

    private native void _consume();
    private native boolean  _getConsume();
    private native void _setConsume(boolean newFConsume);
}
