#include "dc_jnimap.h"
#include <reg/NativeReg.h>

DC_JNIMap *DC_JNIMap::instance = 0;
DC_Mutex  *DC_JNIMap::lock = new DC_Mutex();

/*
 * Get a reference to a JNIMap.  
 * The JNIMap class uses the singleton pattern to ensure only one map exists
 */
DC_JNIMap* DC_JNIMap::GetJNIMap() {
   if (instance == 0) {
      instance = new DC_JNIMap();
   }   

   return instance;
}


/*
 * Private constructor for the JNIMap, initializes the internal maps
 * The JNIMap class uses the singleton pattern to ensure only one map exists
 */
DC_JNIMap::DC_JNIMap() {
   javaToC = new JINT2JLONG();
   javaToC->clear();
}

/*
 * Destructor wich cleans up the internal maps
 */
DC_JNIMap::~DC_JNIMap() {
   delete javaToC;
   delete lock;
}

/*
 * Get a C++ object using a Java class
 */
jlong DC_JNIMap::find(JNIEnv *env, jobject javaObject) {
   jint javaCode = NativeReg::ConstructNReg()->GetHashCode(env, javaObject);

   jlong cObject = 0;
   
   JINT2JLONG::iterator location;

   if (javaCode != 0) {
      location = javaToC->find(javaCode);
      if (location != javaToC->end()) {
         cObject = (*location).second;
      }
   }

   return cObject;
}

/*
 * Add a link between a give Java object and a C++ object
 * Assumes that there is no pre-existing link for the Java Object
 */
void DC_JNIMap::addPairNoOverwrite(JNIEnv *env, jobject javaObject, jlong cObject) {
   jint javaCode = NativeReg::ConstructNReg()->GetHashCode(env, javaObject);

   if (javaCode != 0) {
      javaToC->insert(JINT2JLONG::value_type(javaCode, cObject)); 
   }
}

/*
 * Add a link between a given Java object and a C++ object
 * When there is a pre-existing C++ link, it is first removed and the C++ object is destroyed
 */
void DC_JNIMap::addPair(JNIEnv *env, jobject javaObject, jlong cObject) {
  jint javaCode = NativeReg::ConstructNReg()->GetHashCode(env, javaObject);
  
  if (javaCode != 0) {
     jlong existingCObject = find(env, javaObject);
     
     if (existingCObject != 0) {
        delete reinterpret_cast<void *>(existingCObject);
	remove(env, javaObject);
     }

     javaToC->insert(JINT2JLONG::value_type(javaCode, cObject));
  }
}

/*
 * Remove a link between a given Java object and a C++ object
 */
bool DC_JNIMap::remove(JNIEnv *env, jobject javaObject) {
   jint javaCode = NativeReg::ConstructNReg()->GetHashCode(env, javaObject);
   bool removed = false;
   
   JINT2JLONG::iterator location;
   
   if (javaCode != 0) {
      location = javaToC->find(javaCode);
      if (location != javaToC->end()) {
         javaToC->erase(location);
         
         removed = true;
      }
   }

   return removed;   
}
