// ****************************************************************************
//  Name: dc_argv.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 05/19/1997 : mdb - Created.
//  Desc: Utilities for manipulating argc argv pairs.
// ****************************************************************************

#include "dc_argv.h"

#define ARGV_DefaultExtentSize 10


// ****************************************************************************
//  Class: Argv
// ****************************************************************************

Argv::Argv(void) {
  nElem = maxElem = 0; 
  rgElem = NULL;
  rgfDelete = NULL;
}

Argv::~Argv(void) { 
  Flush();
  TestDelete(rgElem);
  TestDelete(rgfDelete);
}

void Argv::Extend(void) {
  Types::uint nNewSize = maxElem + ARGV_DefaultExtentSize;
  char **rgTmp = rgElem;
  bool *rgfTmp = rgfDelete;
  rgElem = new char*[(maxElem = nNewSize)];
  rgfDelete = new bool[maxElem];
  if (rgTmp) {
    // copy over old elements
    for (Types::uint i=0; i<nElem; i++) {
      rgElem[i] = rgTmp[i];
      rgfDelete[i] = rgfTmp[i];
    }
    rgElem[nElem] = NULL;
    delete [] rgTmp;
    delete [] rgfTmp;
  }
}


// ****************************************************************************
//  Name: Set
//  Date: 07/23/1997 : mdb - Created.
// ****************************************************************************
/**
 * Use a given argc and argv.
 * @param argc
 * @param argv - arg list to use
 */
void Argv::Set(int argc, char *argv[], bool fDelete) {
  TestDelete(rgElem);
  TestDelete(rgfDelete);
  nElem = argc;
  maxElem = nElem + 1;
  rgElem = new char*[maxElem];
  rgfDelete = new bool[maxElem];
  for (Types::uint i=0; i<nElem; i++) {
    rgElem[i] = argv[i];
    rgfDelete[i] = fDelete;
  }
  rgElem[nElem] = NULL;
}

void Argv::Set(DArray<char*> da, bool fDelete) {
  TestDelete(rgElem);
  TestDelete(rgfDelete);
  nElem = da.NumElem();
  maxElem = nElem + 1;
  rgElem = new char*[maxElem];
  rgfDelete = new bool[maxElem];
  for (Types::uint i=0; i<nElem; i++) {
    rgElem[i] = da.Get(i);
    rgfDelete[i] = fDelete;
  }
  rgElem[nElem] = NULL;
}


// ****************************************************************************
//  Name: Append
//  Date: 07/23/1997 : mdb - Created.
// ****************************************************************************
/**
 * Add an arg to the end of the list.
 * @param sbNewArg - arg to add
 */
int Argv::Append(char *sbNewArg, bool fDelete) {
  if (nElem+1 >= maxElem) Extend();

  rgElem[nElem] = sbNewArg;
  rgfDelete[nElem] = fDelete;
  nElem++;
  rgElem[nElem] = NULL;
  return nElem;
}

int Argv::Append(DArray<char*> da, bool fDelete) {
  Types::uint n = da.NumElem();
  for (Types::uint i=0; i<n; i++) {
    Append(da.Get(i), fDelete);
  }
  return nElem;
}


// ****************************************************************************
//  Name: Flush
//  Date: 07/23/1997 : mdb - Created.
// ****************************************************************************
/**
 * Empties all in list
 */
void Argv::Flush(void) {
  if (rgElem != NULL) {
    for (Types::uint i=0; i<nElem; i++) {
      if (rgfDelete[i]) TestDelete(rgElem[i]);
    }
    nElem = 0;
    rgElem[nElem] = NULL;
  }
}


// ****************************************************************************
//  Name: Find
//  Date: 06/12/1997 : mdb - Created.
//        07/23/1997 : mdb - Moved into Argv class.
// ****************************************************************************
/** 
 * Detects the occurrence of sbFlag.
 * @param sbFlag - flag to look for
 * @return bool - true=found, false=not found
 */
bool Argv::Find(char *sbFlag) {
  for (Types::uint i=0; i<nElem; i++) {
    if (strcmp(sbFlag, rgElem[i]) == 0) return true;
  }
  return false;
}


// ****************************************************************************
//  Name: NextFind
//  Date: 05/19/1997 : mdb - Created.
//        07/23/1997 : mdb - Moved into Argv class.
// ****************************************************************************
/**
 * Find the first occurrence of sbFlag and the following argument.
 * @param sbFlag - flag to look for
 * @return char* - NULL=not found, else=argument
 */
char *Argv::NextFind(char *sbFlag) {
  for (Types::uint i=0; i<nElem; i++) {
    if (strcmp(sbFlag, rgElem[i]) == 0) {  /* found? */
      return ((i+1 < nElem) ? rgElem[i+1] : (char*)NULL);
    }
  }
  return NULL;
}


// ****************************************************************************
//  Name: NextExtract
//  Date: 05/19/1997 : mdb - Created.
//        07/23/1997 : mdb - Moved into Argv class.
// ****************************************************************************
/** 
 * Remove the first occurrence of sbFlag and the following argument.
 * @param sbFlag - flag to look for
 * @return char* - NULL=not found, else=argument
 */
char *Argv::NextExtract(char *sbFlag) {
  char *sbRtn=NULL;
  Types::uint i, j, fFound = 0;
  
  /* i - the next one to look at
   * j - the next spot in result
   */
  for (i=0,j=0; i<nElem; i++) {
    if (!fFound && strcmp(sbFlag, rgElem[i]) == 0) {  /* found? */
      fFound = 1;
      if (i+1 < nElem) {  /* is there a following arg? */
        sbRtn = rgElem[i+1];
        i++;  /* skip the next arg */
      }
    } else {
      rgElem[j] = rgElem[i];
      j++;  /* leave in result argv */
    }
  }
  nElem -= (i-j);  /* reduce by how many skipped */
  rgElem[nElem] = NULL;
  
  return sbRtn;
}


// ****************************************************************************
//  Name: toString
//  Date: 04/18/2000 : mdb - Created.
// ****************************************************************************
/**
 * Flattens the argv into a string with spaces between args and quotes
 * around args with spaces in them.
 * @param fAddQuotes - add quotes to any arg that has spaces
 * @param strOut - str to fill
 * @return int - 0=success
 */
int Argv::toString(Str &strOut, bool fAddQuotes) {
  strOut.Set(NULL);
  for (int i=0; i<GetNum(); ++i) {
    if (i>0) strOut.Cat(' ');
    if (fAddQuotes && strchr(rgElem[i], ' ') && rgElem[i][0] != '\"') {
      strOut.Cat('\"');
      strOut.Cat(rgElem[i]);
      strOut.Cat('\"');
    } else {
      strOut.Cat(rgElem[i]);
    }
  }
  return 0;
}


// ****************************************************************************
//  Name: operator<<
//  Date: 07/23/1997 : mdb - Created.
// ****************************************************************************
/** 
 * Output an argv object.
 */
ostream &operator<<(ostream &os, Argv &argv) {
  for (Types::uint i=0; i<argv.nElem; i++) {
    os << "[" << i << "] " << argv.rgElem[i] << endl;
  }
  return os;
}

// ****************************************************************************
