// ****************************************************************************
//  Name: dc_barrier.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 09/11/1999 : mdb - Created.
//        10/23/1999 : mdb - Changes for DCE use.
//        08/13/2000 : mdb - Added correct resetting for repeated use.
//  Desc: Resuable thread barrier implementation.
//  Note: There is no pthread_cleanup code for when a thread is cancelled
//        holding a barrier lock.
// ****************************************************************************

#include "dc_barrier.h"
#include <iostream>


//#define DCDBG_BARRIER

#if defined(DCDBG_BARRIER)
#  include <unistd.h>
#endif /* DCDBG_BARRIER */


//
//  class DC_Barrier
//

void DC_Barrier::_init(void) {
#if defined(DCDBG_BARRIER)
  cerr << "DC_Barrier (" << hex << (unsigned int)this << dec << ") "
       << "initialized to " << n << endl;
#endif /* DCDBG_BARRIER */
}

DC_Barrier::DC_Barrier(void) : x(0), n(0) {
  _init();
}
DC_Barrier::DC_Barrier(int num) : x(0), n(num) {
  _init();
}

/**
 * only can be called before it is used
 */
void DC_Barrier::set(int num) {
#if defined(DCDBG_BARRIER)
  cerr << "DC_Barrier (" << hex << (unsigned int)this << dec << ") "
       << "set to " << n << endl;
#endif /* DCDBG_BARRIER */
  n = num;
}

/**
 * note: fSlowWakeup is a hack for testing purposes to control ordering
 */
void DC_Barrier::sync(bool fSlowWakeup) {
#if defined(DCDBG_BARRIER)
  cerr << "DC_Barrier(" << hex << (unsigned int)this << dec << ") "
       << "sync() thread t" << hex << (int)pthread_self() << dec
       << " entry (before lock) ... " << endl;
  cerr.flush();
#endif /* DCDBG_BARRIER */
  mutex.lock();
  bool fFirstHalf = (x < n);  // save for our wakeup
  ++x;
#if defined(DCDBG_BARRIER)
  cerr << "DC_Barrier(" << hex << (unsigned int)this << dec << ") "
       << "sync() thread t" << hex << (int)pthread_self() << dec
       << " incremented to " << x << "/" << n << endl;
  cerr.flush();
#endif /* DCDBG_BARRIER */
  if (x == 2*n || x == n) {
#if defined(DCDBG_BARRIER)
    cerr << "DC_Barrier(" << hex << (unsigned int)this << dec << ") "
	 << "sync() thread t" << hex << (int)pthread_self() << dec
	 << " last to sync - wake up all others ..." << endl;
    cerr.flush();
#endif /* DCDBG_BARRIER */
    // BUGBUG: the only problem is we may be waking those that have
    // rushed ahead to the next barrier, in which case their while
    // test will fail, and they will go back to sleep
    
    // this is the last to hit the barrier -- wake up the others
    cv.broadcast();
    if (x == 2*n) x = 0;  // reset to first half
  } else {
    // still more to arrive at sync() call
    // - check if we are in the next half of the barrier space
    //   which avoid the [Almasi] p.588 problem where one thread
    //   can leave the barrier and reach the next barrier before
    //   the rest of the threads have woken up, causing deadlock.
#if defined(DCDBG_BARRIER)
    cout << "DC_Barrier(" << hex << (unsigned int)this << dec << ") "
	 << "sync() thread t" << hex << (int)pthread_self() << dec
	 << " while: fFirstHalf=" << (fFirstHalf ? "t" : "f")
	 << ",x=" << x << ",n=" << n << endl;
#endif /* DCDBG_BARRIER */
    while ((x < n) == fFirstHalf) {
#if defined(DCDBG_BARRIER)
      cerr << "DC_Barrier(" << hex << (unsigned int)this << dec << ") "
	   << "sync() thread t" << hex << (int)pthread_self() << dec
	   << " going to sleep (" << x << "/" << n << ")" << endl;
      cerr.flush();
#endif /* DCDBG_BARRIER */
      cv.wait(mutex);
#if defined(DCDBG_BARRIER)
      if (fSlowWakeup) sleep(5);
      cout << "DC_Barrier(" << hex << (unsigned int)this << dec << ") "
	   << "sync() thread t" << hex << (int)pthread_self() << dec
	   << " wakeup: fFirstHalf=" << (fFirstHalf ? "t" : "f")
	   << ",x=" << x << ",n=" << n << endl;
#endif /* DCDBG_BARRIER */
    }
  }
#if defined(DCDBG_BARRIER)
  cerr << "DC_Barrier(" << hex << (unsigned int)this << dec << ") "
       << "sync() thread t" << hex << (int)pthread_self() << dec
       << " unlock ... " << endl;
  cerr.flush();
#endif /* DCDBG_BARRIER */
  mutex.unlock();
#if defined(DCDBG_BARRIER)
  cerr << "DC_Barrier(" << hex << (unsigned int)this << dec << ") "
       << "sync() thread t" << hex << (int)pthread_self() << dec
       << " exit." << endl;
  cerr.flush();
#endif /* DCDBG_BARRIER */
}


// ****************************************************************************
