// ****************************************************************************
//  Name: dc_configinfo.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 05/20/1997 : mdb - Created.
//        06/11/1997 : mdb - Added ability to iterate through for Renato.
//        07/31/1997 : mdb - Added auto-expansion of macros.
//        08/13/1999 : mdb - Added reading from strings, buf, dump to file.
//                         - Changed/renamed from ConfigFile to ConfigInfo.
//        01/07/2000 : mdb - Added "//" comments and "#include" of files.
//        01/09/2000 : mdb - Changed so #include will work as expected.
//        01/11/2000 : mdb - Added RemoveSection() method.
//        04/03/2000 : mdb - Rewrote to add several features/fixes.
//                           (lazy expansion, >1 section name, SizeBytes)
//  Desc: Reading of application configuration information.
// ****************************************************************************

#include "dc_standard.h"

#ifdef _WIN32
#include "win32.h"
#endif

#include <fstream>
#include <sstream>
using namespace std;

#include "dc_configinfo.h"
#include "dc_str.h"

#include <assert.h>


//#define DCDBG_CONFIG
//#define DCDBG_CONFIG_ADD
//#define DCDBG_CONFIG_MERGE


// *************************
//  Class: ConfigInfo_Entry
// *************************

/**
 * copy and steal rhs
 */
ConfigInfo_Entry &ConfigInfo_Entry::operator=(ConfigInfo_Entry &rhs) {
  if (this == &rhs) return *this;  // self-assignment
  sbName = rhs.sbName; rhs.sbName = NULL;
  sbValue = rhs.sbValue; rhs.sbValue = NULL;
  sbExpanded = rhs.sbExpanded; rhs.sbExpanded = NULL;
  wCycle = rhs.wCycle;
  return *this;
}


// ***************************
//  Class: ConfigInfo_Section
// ***************************

/**
 * copy and steal rhs: this is needed so dtor can cleanup its state in fcns
 * like AppendSection() without deleting the state in the DArray and w/o
 * doing deep copies of temps when the temp will be deleted anyway.
 */
ConfigInfo_Section &ConfigInfo_Section::operator=(ConfigInfo_Section &rhs) {
  if (this == &rhs) return *this;  // self-assignment
  sbName = rhs.sbName; rhs.sbName = NULL;
  prgEntry = rhs.prgEntry; rhs.prgEntry = NULL;
  return *this;
}

unsigned int ConfigInfo_Section::FindEntry(const char *sbEntryName) {
  if (prgEntry) {
    unsigned int nEntry = prgEntry->NumElem();
    for (unsigned int j=0; j<nEntry; ++j) {
      if (StrCmp(sbEntryName, prgEntry->Get(j).sbName, true) == 0) {
	return j;  // found name
      }
    }
  }
  return (unsigned int)-1;  // not found
}

unsigned int ConfigInfo_Section::AppendEntry(const char *sbEntryName,
					     const char *sbEntryValue) {
  if (!prgEntry) prgEntry = new DArray<ConfigInfo_Entry>;
  ConfigInfo_Entry e(sbEntryName, sbEntryValue);
  return prgEntry->Append(e);
}

unsigned int ConfigInfo_Section::ReplaceEntry(unsigned int iName,
					       const char *sbNewValue) {
  if (!prgEntry) return (unsigned int)-1;  // error
  if (iName > prgEntry->NumElem()) return (unsigned int)-1;  // error
  prgEntry->Get(iName).SetValue(sbNewValue);
  return iName;
}

unsigned int ConfigInfo_Section::AddEntry(const char *sbEntryName,
					   const char *sbEntryValue,
					   int opts_in) {
  if (opts_in) {
    unsigned int i = FindEntry(sbEntryName);
    if (i != (unsigned int)-1) {
      if (opts_in & ADD_DUPNAME_REPLACE) {
#if defined(DCDBG_CONFIG_ADD)
	cout << "AddEntry(): DUPNAME_REPLACE >" << sbEntryName << "< >"
	     << sbEntryValue << "<" << endl;
#endif /* DCDBG_CONFIG_ADD */
	return ReplaceEntry(i, sbEntryValue);
      }
    }
  }
  // default
#if defined(DCDBG_CONFIG_ADD)
  cout << "AddEntry(): appending >" << sbEntryName << "< >"
       << sbEntryValue << "<" << endl;
#endif /* DCDBG_CONFIG_ADD */
  return AppendEntry(sbEntryName, sbEntryValue);
}

void ConfigInfo_Section::Flush(void) {
  if (!prgEntry) return;
  unsigned int nEntry = prgEntry->NumElem();
  for (unsigned int j=0; j<nEntry; ++j) {
    TestDeleteNull(prgEntry->Get(j).sbName);
    TestDeleteNull(prgEntry->Get(j).sbValue);
    TestDeleteNull(prgEntry->Get(j).sbExpanded);
  }
  prgEntry->Flush();  // empty/reset the DArray
  SDeleteNull(prgEntry);
}


// *******************
//  Class: ConfigInfo
// *******************

/**
 * this size depends on the formatted output of operator<<()
 */
unsigned int ConfigInfo::SizeBytes(void) {
  size_t wSize = 0;
  unsigned int nSections = rgSection.NumElem();
  for (unsigned int i=0; i<nSections; ++i) {
    unsigned int len = (unsigned int)strlen(rgSection.Get(i).sbName);
    if (len > 0) wSize += len + 4;

    DArray<ConfigInfo_Entry> *prgEntry = rgSection.Get(i).prgEntry;
    if (prgEntry) {
      unsigned int nEntry = prgEntry->NumElem();
      for (unsigned int j=0; j<nEntry; ++j) {
	ConfigInfo_Entry *pEntry = &(prgEntry->Get(j));

	wSize += strlen(pEntry->sbName) + 2;
	if (opts & ConfigInfo::OUT_EXPAND_EACH) {
	  char *sbTmp = ExpandEntryMacros(pEntry, rgSection.Get(i).sbName,
					  false);
	  if (sbTmp) {
	    wSize += strlen(sbTmp);
	    delete [] sbTmp;
	  }
	} else {
	  wSize += strlen(pEntry->sbValue);
	}
      }
    }
  }
  return (unsigned int)(++wSize);  // for null byte
}

// ****************************************************************************
//  Name: Flush
//  Date: 05/21/1997 : mdb - Created.
// ****************************************************************************
/**
 * Free up alloced space for strings in darrays.
 */ 
void ConfigInfo::Flush(void) {
  unsigned int nSections = rgSection.NumElem();
  for (unsigned int i=0; i<nSections; ++i) {
    DeleteNull(rgSection.Get(i).sbName);
    rgSection.Get(i).Flush();
  }
  rgSection.Flush();  // empty/reset the DArray

  iSection = iEntry = 0;  // reset iterator
  wCycle = 0;

  daCurrentSection.Flush();
}

unsigned int ConfigInfo::FindSection(const char *sbSectionName) {
  unsigned int nEntry = rgSection.NumElem();
  for (unsigned int j=0; j<nEntry; ++j) {
    if (StrCmp(sbSectionName, rgSection.Get(j).sbName, true) == 0) {
      return j;  // found name
    }
  }
  return (unsigned int)-1;  // not found
}

unsigned int ConfigInfo::AppendSection(const char *sbSectionName) {
  ConfigInfo_Section section(sbSectionName);
  return rgSection.Append(section);
}

unsigned int ConfigInfo::ReplaceSection(unsigned int iSection_in) {
  if (iSection_in > rgSection.NumElem())
      return (unsigned int)-1;  // error
  // we assume the same section name will be used
  rgSection.Get(iSection_in).Flush();
  return iSection_in;
}

unsigned int ConfigInfo::AddSection(const char *sbSectionName) {
  if (opts) {
    unsigned int i = FindSection(sbSectionName);
    if (i != (unsigned int)-1) {
      if (opts & ADD_DUPSEC_USEOLD || strlen(sbSectionName) == 0) {
#if defined(DCDBG_CONFIG_ADD)
	cerr << "AddSection(): DUPSEC_USEOLD >" << sbSectionName << "<"
	     << endl;
#endif /* DCDBG_CONFIG_ADD */
	return i;
      } else if (opts & ADD_DUPSEC_REPLACE) {
#if defined(DCDBG_CONFIG_ADD)
	cerr << "AddSection(): DUPSEC_REPLACE >" << sbSectionName << "<"
	     << endl;
#endif /* DCDBG_CONFIG_ADD */
	return ReplaceSection(i);
      }
    }
  }
  // default
#if defined(DCDBG_CONFIG_ADD)
  cerr << "AddSection(): appending >" << sbSectionName << "<" << endl;
#endif /* DCDBG_CONFIG_ADD */
  return AppendSection(sbSectionName);
}

// ****************************************************************************
//  Name: RemoveSection
//  Date: 01/11/2000 : mdb - Created.
// ****************************************************************************
/**
 * Searches for and removes the given section and all entries.
 * @param sbSection - section name to remove
 * @return int - number of sections removed
 */
int ConfigInfo::RemoveSection(const char *sbSection) {
  int nRemoved = 0;
  unsigned int iSection2;
  while ((iSection2 = FindSection(sbSection)) != (unsigned int)-1) {
    // remove this section
    ++nRemoved;
    rgSection.Get(iSection2).Flush();  // remove all the name-value entries
    TestDeleteNull(rgSection.Get(iSection2).sbName);
    rgSection.Remove(iSection2);  // remove from the section array
  }
  if (rgSection.NumElem() == 0) Flush();  // did we remove them all?
  return nRemoved;
}

void ConfigInfo::AddEntry(unsigned int iSection_in,
                          const char *sbName, const char *sbValue) {
  if (iSection_in > rgSection.NumElem())
      return;  // error
  rgSection.Get(iSection_in).AddEntry(sbName, sbValue, opts);
}

void ConfigInfo::AddEntry(const char *sbSection, const char *sbName, const char *sbValue) {
  unsigned int iSection2 = AddSection(sbSection);
  AddEntry(iSection2, sbName, sbValue);
}


// ****************************************************************************
//  Name: ReadFile
//  Date: 05/20/1997 : mdb - Created.
//        05/22/1997 : mdb - Added default config file search.
//        06/16/1997 : mdb - Fixed bug where sbProgramName is not initialized.
//        01/09/2000 : mdb - Split into an Internal version to be able to pass
//                           in state recursively.
//        02/09/2000 : mdb - Fixed always return 0 bug in last change.
// ****************************************************************************
/**
 * Use the given file as a config file.
 * @param sbFile - filename, NULL=look for default files
 * @return int - 0=ok, 1=failed open or close
 */
int ConfigInfo::ReadFile(const char *sbFile, bool fFlushBeforeRead) {
  int wRtn;
  if (fFlushBeforeRead) Flush();
  wRtn = ReadFile_Internal(sbFile);
  daCurrentSection.Flush();
  return wRtn;
}

/**
 * also called recursively for #include files
 */
int ConfigInfo::ReadFile_Internal(const char *sbFile) {
  char sbBuf[256];
  ifstream fin;
 
  if (sbFile) {
    fin.open(sbFile, ios::in);
    if (fin.fail()) return 1;
  } else {  // NULL = look for default names
    snprintf(sbBuf, sizeof(sbBuf), "%s.cfg",
	    (sbProgramName ? sbProgramName : "ProgramName"));
    fin.open(sbBuf, ios::in);
    if (fin.fail()) {
      snprintf(sbBuf, sizeof(sbBuf), "default.cfg");
      fin.open(sbBuf, ios::in);
      if (fin.fail()) {
	snprintf(sbBuf, sizeof(sbBuf), "%s/%s.cfg", getenv("HOME"),
		(sbProgramName ? sbProgramName : "ProgramName"));
	fin.open(sbBuf, ios::in);
	if (fin.fail()) {
	  snprintf(sbBuf, sizeof(sbBuf), "%s/default.cfg", getenv("HOME"));
	  fin.open(sbBuf, ios::in);
	  if (fin.fail()) return 1;
	}
      }
    }
    sbFile = sbBuf;
  }

  fin >> *this;
  fin.close();
  return 0;
}


// ****************************************************************************
//  Name: ReadStr
//  Date: 08/13/1999 : mdb - Created.
// ****************************************************************************
/**
 * Use the given string as the contents of a config file.
 * @param sbMem - buffer in memory containing configinfo data
 * @return int - 0=ok, 1=failed
 */
int ConfigInfo::ReadStr(const char *sbMem, bool fFlushBeforeRead) {
  if (!sbMem) return 1;
  istringstream sin(sbMem);
  if (fFlushBeforeRead) Flush();
  sin >> *this;
  daCurrentSection.Flush();
  return 0;
}


// ****************************************************************************
//  Name: ReadBuf
//  Date: 08/13/1999 : mdb - Created.
//        12/13/1999 : mdb - Fixed bug handling empty or NULL buffer.
// ****************************************************************************
/**
 * Use the given buffer as the contents of a config file.
 * @param pBuf - ptr of memory buffer
 * @param wLen - len of memory buffer
 * @return int - 0=ok, 1=failed
 */
int ConfigInfo::ReadBuf(const char *pBuf, int wLen, bool fFlushBeforeRead) {
  if (!(pBuf && wLen>0)) return 1;
  istringstream sin(string(pBuf, wLen));
  if (fFlushBeforeRead) Flush();
  sin >> *this;
  daCurrentSection.Flush();
  return 0;
}


// ****************************************************************************
//  Name: Merge
//  Date: 08/13/1999 : mdb - Created.
// ****************************************************************************
/** 
 * Add the entries of the given config to the current one.  What happens
 * for duplicate section names and entries with sections depends on the
 * value of the opts parameter.  By default 
 * Entries from duplicate section names are added to the existing
 * section.  Duplicate entries for a given section are just added.
 * @param cfgMerge - config entries to merge
 */
void ConfigInfo::Merge(ConfigInfo &cfgMergeFrom, int opts_in) {
  int i, j, nEntry;
  unsigned int iSectionMergeTo;

  int opts_save;
  if (opts_in != -1) {
    opts_save = GetOpts();
    SetOpts(opts_in);
  }

#if defined(DCDBG_CONFIG_MERGE)
  cout << "Merge(): entry opts=" << opts << ' ';
  if (GetOpts() & ADD_DUPSEC_USEOLD)  cout << "DUPSEC_USEOLD ";
  if (GetOpts() & ADD_DUPSEC_REPLACE) cout << "DUPSEC_REPLACE ";
  if (GetOpts() & ADD_DUPSEC_APPEND)  cout << "DUPSEC_APPEND ";
  if (GetOpts() & ConfigInfo_Section::ADD_DUPNAME_APPEND)
    cout << "DUPNAME_APPEND ";
  if (GetOpts() & ConfigInfo_Section::ADD_DUPNAME_REPLACE)
    cout << "DUPNAME_REPLACE ";
  cout << " ...\n";
#endif /* DCDBG_CONFIG_MERGE */

  // go through passed in config entries
  int nSections = cfgMergeFrom.rgSection.NumElem();
  for (i=0; i<nSections; ++i) {
    char *sbSection = cfgMergeFrom.rgSection.Get(i).sbName;

#if defined(DCDBG_CONFIG_MERGE)
    cout << "Merge(): looking at source section \"" << sbSection << "\"\n";
#endif /* DCDBG_CONFIG_MERGE */

    iSectionMergeTo = AddSection(sbSection);

    // now copy the entries
    DArray<ConfigInfo_Entry> *prgEntryMergeFrom
      = cfgMergeFrom.rgSection.Get(i).prgEntry;
    if (!prgEntryMergeFrom) continue;  // empty

#if defined(DCDBG_CONFIG_MERGE)
    cout << "Merge(): Copying the entries from section " << i
	 << " of src to section " << iSectionMergeTo << " of dest\n";
#endif /* DCDBG_CONFIG_MERGE */

    nEntry = prgEntryMergeFrom->NumElem();
    for (j=0; j<nEntry; ++j) {
      // add this entry
      rgSection.Get(iSectionMergeTo)
	.AddEntry(prgEntryMergeFrom->Get(j).sbName,
		  prgEntryMergeFrom->Get(j).sbValue, GetOpts());
    }
  }

  if (opts_in != -1) {
    SetOpts(opts_save);
  }

#if defined(DCDBG_CONFIG_MERGE)
  cout << "Merge(): exit...\n";
#endif /* DCDBG_CONFIG_MERGE */
}


// ****************************************************************************
//  Name: WriteFile
//  Date: 08/13/1999 : mdb - Created.
//        08/18/1999 : mdb - Changed to call general Write().
// ****************************************************************************
/** 
 * Save the current entries to the given file.
 * @param sbFile - filename, NULL="filesave.cfg"
 * @return int - 0=ok, 1=failed save
 */
int ConfigInfo::WriteFile(const char *sbFile) {
  ofstream fout;

  if (sbFile) {
    fout.open(sbFile, ios::out);
    if (fout.fail()) return 1;
  } else {  // NULL = look for default names
    fout.open("filesave.cfg", ios::out);
    if (fout.fail()) return 1;
  }

  fout << *this;
  fout.close();

  return fout.fail() ? 1 : 0;
}


// ****************************************************************************
//  Name: WriteBuf
//  Date: 08/18/1999 : mdb - Created.
// ****************************************************************************
/** 
 * Write the config info to the given buffer.
 * @param pBuf - ptr of memory buffer
 * @param wLen - len of memory buffer
 * @return int - 0=ok, 1=failed
 */
int ConfigInfo::WriteBuf(char *pBuf, int wLen) {
  ostringstream sout(string(pBuf, wLen));
  sout << *this;
  sout << '\0';
  return 0;
}


// ****************************************************************************
//  Name: FindEntry (internal)
//  Date: 05/20/1997 : mdb - Created.
//        07/31/1997 : mdb - Made an internal method for reuse.
//        08/07/1997 : mdb - Added auto heirarchy checking.
// ****************************************************************************
/**
 * Attempts to find an entry.
 * @param sbSection - which section to check, ""=global section, NULL=wildcard (flat search only) 
 * @param sbName - name to look for
 * @param fScope - true=auto check up in scope, false=check here only
 * @return ConfigInfo_Entry_t * - NULL=not found, else=entry found
 */
ConfigInfo_Entry *ConfigInfo::FindEntry(const char *sbSection,
					const char *sbName,
					bool fScope) {
  DArray<ConfigInfo_Entry> *prgEntry;
  Str str;

  str.Set(sbSection);
  while (1) {
    // first find the section
    int nSections = rgSection.NumElem();
    for (int i=0; i<nSections; ++i) {
      if (!sbSection ||
	  StrCmp(str.Get(), rgSection.Get(i).sbName, true) == 0) {
	// found the section, now look for name
	prgEntry = rgSection.Get(i).prgEntry;
	if (prgEntry) {
	  int nEntry = prgEntry->NumElem();
	  for (int j=0; j<nEntry; ++j) {
	    if (StrCmp(sbName, prgEntry->Get(j).sbName, true) == 0) {
	      // found a name match in this entry
	      return &(prgEntry->Get(j));
	    }
	  }
	}
	if (sbSection) break; // only continue into other sections if wildcard
      }
    }

    if (fScope) {
      int wLen = str.Len();
      str.RChop(".");
      if (str.Len() == wLen) break;
    } else {
      break;
    }
  }
  return NULL;
}


// ****************************************************************************
//  Name: Find
//  Date: 05/20/1997 : mdb - Created.
//  Note: This is *not* threadsafe due to cycle detection.
// ****************************************************************************
/**
 * Attempts to find an entry.
 * @param sbSection - which section to check, ""=global section, NULL=wildcard
 * @param sbName - name to look for
 * @return char * - NULL=not found, else=value found
 */ 
char *ConfigInfo::Find(const char *sbSection, const char *sbName) {
  ConfigInfo_Entry *pEntry = FindEntry(sbSection, sbName, true);
  if (pEntry) {
    return ExpandEntryMacros(pEntry, sbSection, true);
  }
  return (char*)NULL;
}


// ****************************************************************************
//  Name: FindDup
//  Date: 11/08/2000 : mdb - Created.
// ****************************************************************************
/** 
 * Attempts to find an entry, while avoiding internal state that would
 * cause thread-safety issues.
 * @param sbSection - which section to check, ""=global section, NULL=wildcard
 * @param sbName - name to look for
 * @return char * - NULL=not found, else=value found (must be freed by caller)
 */
char *ConfigInfo::FindDup(const char *sbSection, const char *sbName) {
  ConfigInfo_Entry *pEntry = FindEntry(sbSection, sbName, true);
  if (pEntry) {
    return ExpandEntryMacros(pEntry, sbSection, false);
  }
  return (char*)NULL;
}


// ****************************************************************************
//  Name: FindAll
//  Date: 05/22/1997 : mdb - Created.
//        08/13/1999 : mdb - Changed to return a pointer to the allocated da
//                           instead of a reference.  There was no easy way
//                           to free the reference, and it was a leak.
//  BUGBUG: There is still the thread safety issue of this returning ptrs to
//          the internal sbExpanded state, and another thread changing them.
// ****************************************************************************
/**
 * Attempts to find all entries.
 * @param sbSection - which section to check, ""=global section, NULL=wildcard
 * @param sbName - name to look for
 * @return DArray<char*>* - result set (must be freed by caller)
 */
DArray<char*> *ConfigInfo::FindAll(const char *sbSection, const char *sbName) {
  DArray<char*> *pdaRtn = new DArray<char*>;
  DArray<ConfigInfo_Entry> *prgEntry;

  // first find the section
  int nSections = rgSection.NumElem();
  for (int i=0; i<nSections; i++) {
    if (!sbSection ||
	StrCmp(sbSection, rgSection.Get(i).sbName, true) == 0) {
      // found the section, now look for name
      prgEntry = rgSection.Get(i).prgEntry;
      if (!prgEntry) continue;
      int nEntry = prgEntry->NumElem();
      for (int j=0; j<nEntry; j++) {
	if (StrCmp(sbName, prgEntry->Get(j).sbName, true) == 0) {
	  // found a name match, return the value
	  pdaRtn->AppendByVal(ExpandEntryMacros(&(prgEntry->Get(j)), sbSection,
						true));
	}
      }
      if (sbSection) break; // only continue into other sections if wildcard
    }
  }
  return pdaRtn;
}


// ****************************************************************************
//  Name: ExpandEntryMacros
//  Date: 07/31/1997 : mdb - Created.
//        08/05/1997 : mdb - Fixed recursive problem to only expand if the
//                           result won't directly cause a cycle.
//        01/08/2000 : mdb - Change to allow for expansion of nested macros.
//        11/08/2000 : mdb - Added option to not store result internally, and
//                           to steal instead of duping from local Str.
// ****************************************************************************
/**
 * Expands an entry's macros by recursive expanding.
 * @param pEntry - which entry to expand, or NULL
 * @param sbSection - scope of parent (if not found, will try "" scope)
 * @param fSaveResult - T==store alloc result internally in config entry
 * @return char * - NULL=not found, else=value found (an allocated string if fSaveResult==F)
 */
char *ConfigInfo::ExpandEntryMacros(ConfigInfo_Entry *pEntry,
				    const char *sbSection,
				    bool fSaveResult) {
  ++wCycle;
  if (wCycle == (unsigned int)-1) {  // reset all cycle counters
    unsigned int nSections = rgSection.NumElem();
    for (unsigned int i=0; i<nSections; ++i) {
      DArray<ConfigInfo_Entry> *prgEntry = rgSection.Get(i).prgEntry;
      if (!prgEntry) continue;
      unsigned int nEntry = prgEntry->NumElem();
      for (unsigned int j=0; j<nEntry; ++j) {
	prgEntry->Get(j).wCycle = 0;
      }
    }
    wCycle = 0;
  }
  return ExpandEntryMacros_Aux(pEntry, sbSection, fSaveResult);
}

/**
 * returns an allocated string if fSaveResult==F
 */
char *ConfigInfo::ExpandEntryMacros_Aux(ConfigInfo_Entry *pEntry,
					const char *sbSection,
					bool fSaveResult) {
  Str str;
  char *p, *q;
  char *sbFullName, *sbName;
  char *sbVal;
  bool fAnyChanges = false;

  char * CloseCurlyBrace = "}";
  char * OpenCurlyBrace = "${";

  if (pEntry == NULL) return NULL;  // not found
  if (pEntry->wCycle == wCycle) return NULL;  // cycle detected

  pEntry->wCycle = wCycle;  // for transitive cycles

  // look through value for macro vars to expand
  str.Set(pEntry->sbValue);
  p = str.Get();
  while (1) {
    if ((p = strstr(p, "${")) && (q = StrEndDelim(p, CloseCurlyBrace, OpenCurlyBrace))) {
      sbFullName = StrNDup(p, (int)(q-p+1));
      sbName = StrNDup(p+2, (int)(q-p-2));
      // first try to expand using section scope, otherwise try "" scope
      if (((sbVal = ExpandEntryMacros_Aux(FindEntry(sbSection, sbName, true),
					  sbSection, fSaveResult))) &&
	  !strstr(sbVal, sbFullName)) {
	str.Replace(sbFullName, sbVal);
	fAnyChanges = true;
	p = str.Get();
      } else {
	p += 2;  // skip ahead if not a replaced var
      }
      if (!fSaveResult) TestDelete(sbVal);  // 04/04/2001 : mdb - Memory leak.
      DeleteNull(sbFullName);
      DeleteNull(sbName);
    } else {
      break;  // no more
    }
  }

  pEntry->wCycle = 0;  // undo for next time

  if (fSaveResult) {
    if (fAnyChanges) {
      // store here to avoid caller having to clean up
      TestDelete(pEntry->sbExpanded);
      pEntry->sbExpanded = str.Steal();
    }
    return pEntry->sbExpanded ? pEntry->sbExpanded : pEntry->sbValue;
  }
  return str.Steal();
}


// ****************************************************************************
//  Name: operator>>
//  Date: 05/21/1997 : mdb - Created.
//        12/19/1997 : mdb - Added esc newline for continuation to next line.
//        08/13/1999 : mdb - Changed to take an istream as arg.
//        03/17/2000 : mdb - Changed is.get to is.getline for new C++ std.
//        04/03/2000 : mdb - Renamed from Read(), and added obj state.
//                         - Support >1 section names by splitting on input.
// ****************************************************************************
/**
 * Parses the config info stream to find all entries.
 */
istream &operator>>(istream &is, ConfigInfo &cfg) {
  const int cchBuf = KB_64;
  char sbBuf[cchBuf];
  char *sb;
  char *pch;
  size_t wLen, cch;
  unsigned int k;

  const char * const EmptyString = "";

  while (1) {
    sb = sbBuf;
    cch = cchBuf;
    while (1) {
      if (! is.getline(sb, (int)cch, '\n')) return is;

      wLen = strlen(sb);
      cch -= wLen;
      if (wLen > 0 && sb[wLen-1] == '\\') {  // escaped end of line
	sb[wLen-1] = ' ';  // replace the backslash with a space
	sb += wLen;
      } else {
	break;
      }
    }

    StrTrimWhite(sbBuf);
    wLen = strlen(sbBuf);
    if (wLen > 0) {
      if (strncmp(sbBuf, "#include ", 9) == 0 && wLen > 9) {
	// TODO: add expanding of ${xxx} within include filename?
	// extract filename and include it
	cfg.ReadFile_Internal(sbBuf+9);
	continue;
      }
      if (sbBuf[0] == '#') continue;  // comment
      if (strncmp(sbBuf, "//", 2) == 0) continue;  // c++ comment
      if (sbBuf[0] == '[') {  // section header
	pch = sbBuf + strlen(sbBuf) - 1;
	if (*pch != ']') {
	  cerr << ErrorHeader(wERROR) 
	       << "bad section header '" << sbBuf << "'.\n";
	  continue;
	}
	*pch = '\0';
	StrTrimWhite(sbBuf+1);
	// add a new section
	Str strTmp(sbBuf+1);
	DArray<char*> *pdaTmp = strTmp.TokenArrayNoDup(", ");
	cfg.daCurrentSection.Flush();
	for (k=0; k<pdaTmp->NumElem(); ++k) {
#if defined(DCDBG_CONFIG_ADD)
	  cout << "DEBUG: token(" << k+1 << "/" << pdaTmp->NumElem() << ")="
	       << pdaTmp->Get(k) << endl;
#endif /* DCDBG_CONFIG_ADD */
	  cfg.daCurrentSection.AppendByVal(cfg.AddSection(pdaTmp->Get(k)));
	}
	SDeleteNull(pdaTmp);

      } else {  // entry line
	pch = strchr(sbBuf, '=');
	if (cfg.daCurrentSection.NumElem() == 0) {
	  // found global option w/o explicit section
	  cfg.daCurrentSection.AppendByVal(cfg.AddSection(EmptyString));
	}
	for (k=0; k<cfg.daCurrentSection.NumElem(); ++k) {
	  if (pch) {
	    *pch = '\0';  // split the string
	    cfg.AddEntry(cfg.daCurrentSection[k], StrTrimWhite(sbBuf),
			 StrTrimWhite(pch+1));
	  } else {
	    cfg.AddEntry(cfg.daCurrentSection[k], StrTrimWhite(sbBuf), EmptyString);
	  }
	}
      }
    }
  }
}


// ****************************************************************************
//  Name: FirstSection
//  Date: 06/11/1997 : mdb - Created.
//        07/26/2001 : mdb - Added optional starting section.
// ****************************************************************************
/**
 * Point iteration to first section.
 */ 
const char *ConfigInfo::FirstSection(const char *sbSection) {
  if (!sbSection) {
    iSection = -1;
    return NextSection();
  }
  if ((iSection = FindSection(sbSection)) == -1) {
    return NULL;  // not found
  }
  return rgSection.Get(iSection).sbName;
}

// ****************************************************************************
//  Name: NextSection
//  Date: 06/11/1997 : mdb - Created.
// ****************************************************************************
/**
 * Point iteration to the next section.
 */
const char *ConfigInfo::NextSection(void) {
  if (iSection+1 < static_cast(int,rgSection.NumElem())) {
    ++iSection;
    return rgSection.Get(iSection).sbName;
  }
  return NULL;
}

// ****************************************************************************
//  Name: FirstName
//  Date: 06/11/1997 : mdb - Created.
// ****************************************************************************
/** 
 * Point iteration to first name within the current section.
 */
const char *ConfigInfo::FirstName(void) {
  iEntry = -1;
  return NextName();
}

// ****************************************************************************
//  Name: NextName
//  Date: 06/11/1997 : mdb - Created.
// ****************************************************************************
/**
 * Point iteration to the next name within the current section.
 */
const char *ConfigInfo::NextName(void) {
  if (iSection < static_cast(int,rgSection.NumElem()) &&
      rgSection.Get(iSection).prgEntry) {
    DArray<ConfigInfo_Entry> *prgEntry;
    prgEntry = rgSection.Get(iSection).prgEntry;
    if (iEntry+1 < static_cast(int,prgEntry->NumElem())) {
      ++iEntry;
      return prgEntry->Get(iEntry).sbName;
    }
  }
  return NULL;
}

// ****************************************************************************
//  Name: CurrentValue
//  Date: 06/11/1997 : mdb - Created.
// ****************************************************************************
/**
 * Get value from the current section and name.
 */
const char *ConfigInfo::CurrentValue(void) {
  if (iSection < static_cast(int,rgSection.NumElem()) &&
      rgSection.Get(iSection).prgEntry) {
    DArray<ConfigInfo_Entry> *prgEntry;
    prgEntry = rgSection.Get(iSection).prgEntry;
    if (iEntry < static_cast(int,prgEntry->NumElem())) {
      return ExpandEntryMacros(&(prgEntry->Get(iEntry)),
			       rgSection.Get(iSection).sbName, true);
    }
  }
  return NULL;
}


// ****************************************************************************
//  Name: operator>>
//  Date: 08/18/1999 : mdb - Created.
//        10/22/2001 : mdb - Fixed bug where global section header is needed
//                           if "" section entries are added later.
// ****************************************************************************
/**
 * Output the current entries to the given stream.
 */
ostream &operator<<(ostream &os, ConfigInfo &cfg) {
  unsigned int nSections = cfg.rgSection.NumElem();
  for (unsigned int i=0; i<nSections; ++i) {
    // only output section names that are not the global section
    if (strlen(cfg.rgSection.Get(i).sbName) > 0 || i > 0) {
      os << '[' << cfg.rgSection.Get(i).sbName << "]\n";
    }

    DArray<ConfigInfo_Entry> *prgEntry = cfg.rgSection.Get(i).prgEntry;
    if (!prgEntry) continue;
    unsigned int nEntry = prgEntry->NumElem();
    for (unsigned int j=0; j<nEntry; ++j) {
      os << prgEntry->Get(j).sbName << '=';
      if (cfg.opts & ConfigInfo::OUT_EXPAND_EACH) {
	os << cfg.ExpandEntryMacros(&(prgEntry->Get(j)),
				    cfg.rgSection.Get(i).sbName, true);
      } else {
	os << prgEntry->Get(j).sbValue;
      }
      os << endl;
    }
    os << endl;
  }
  return os;
}


// ****************************************************************************
