// -*- c++ -*-
// ****************************************************************************
//  Name: dc_configinfo.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 05/20/1997 : mdb - Created.
//        06/11/1997 : mdb - Added ability to iterate through for Renato.
//        07/31/1997 : mdb - Added auto-expansion of macros.
//        08/13/1999 : mdb - Added reading from strings, buf, dump to file.
//                         - Changed/renamed from ConfigFile to ConfigInfo.
//        01/07/2000 : mdb - Added "//" comments and "#include" of files.
//        01/09/2000 : mdb - Added state to make #include work as expected.
//        01/11/2000 : mdb - Added RemoveSection() method.
//        04/03/2000 : mdb - Rewrote to add several features/fixes.
//                           (lazy expansion, >1 section name, SizeBytes)
//        11/08/2000 : mdb - Added a new FindDup to deal with the thread
//                           safety problem of internal expansion strings.
//  Desc: Reading of application configuration information.
// ****************************************************************************
/** @file dc_configinfo.h */

#if !defined(H_CONFIGINFO)
#define H_CONFIGINFO

#include <fstream>

#define DCDBG_TRACE
#include "dc_standard.h"
#include "dc_darray.h"

class ConfigInfo;  // fwd ref

// *************************
//  Class: ConfigInfo_Entry
// *************************

/// An entry of application configuration information.
class ConfigInfo_Entry {
  friend class ConfigInfo;
  friend class ConfigInfo_Section;
  friend std::ostream &operator<<(std::ostream &os, ConfigInfo &cfg);
  friend std::istream &operator>>(std::istream &is, ConfigInfo &cfg);

protected:
  char   *sbName;
  char   *sbValue;
  char   *sbExpanded;
  unsigned int wCycle;  // used to detect cycles in dynamic expansion

public:
  ConfigInfo_Entry(void)
    : sbName(NULL), sbValue(NULL), sbExpanded(NULL), wCycle(0) {
    DBG_TRACE("ConfigInfo_Entry::Constructor()");
  };

  ConfigInfo_Entry(const char *sbName_in, const char *sbValue_in)
    : sbExpanded(NULL), wCycle(0) {
    DBG_TRACE("ConfigInfo_Entry::Constructor(with args)");
    sbName = StrDup(sbName_in);
    sbValue = StrDup(sbValue_in);
  };

  ~ConfigInfo_Entry(void) {
    DBG_TRACE("ConfigInfo_Entry::Destructor()");
    TestDelete(sbName); TestDelete(sbValue); TestDelete(sbExpanded);
  };

  ConfigInfo_Entry &operator=(ConfigInfo_Entry &rhs);

  void SetValue(const char *sbValue_in) {
    TestDelete(sbValue);
    sbValue = StrDup(sbValue_in);
  };

  void Flush(void) {
    TestDelete(sbName);
    TestDelete(sbValue);
    TestDelete(sbExpanded);
  };
};


// ***************************
//  Class: ConfigInfo_Section
// ***************************

/// A section of application configuration information.
class ConfigInfo_Section {
  friend class ConfigInfo;
  friend std::ostream &operator<<(std::ostream &os, ConfigInfo &cfg);
  friend std::istream &operator>>(std::istream &is, ConfigInfo &cfg);

protected:
  char *sbName;
  DArray<ConfigInfo_Entry> *prgEntry;

public:
  ConfigInfo_Section(void) : sbName(NULL), prgEntry(NULL) {
    DBG_TRACE("ConfigInfo_Section::Constructor()");
  };

  ConfigInfo_Section(const char *sbName_in) : sbName(NULL), prgEntry(NULL) {
    DBG_TRACE("ConfigInfo_Section::Constructor(with args)");
    sbName = StrDup(sbName_in);
  };

  ~ConfigInfo_Section(void) {
    DBG_TRACE("ConfigInfo_Section::Destructor()");
    TestDelete(sbName); TestDelete(prgEntry);
  };

  ConfigInfo_Section &operator=(ConfigInfo_Section &rhs);

  unsigned int FindEntry(const char *sbEntryName);
  unsigned int AppendEntry(const char *sbEntryName, const char *sbEntryValue);
  unsigned int ReplaceEntry(unsigned int iName, const char *sbNewValue);

  enum add_opts { // should be bitwise disjoint from main opts
    ADD_DUPNAME_APPEND  = 1,
    ADD_DUPNAME_REPLACE = 2
  };

  unsigned int AddEntry(const char *sbEntryName, const char *sbEntryValue, int opts=-1);

  int NumEntry(void) { return prgEntry ? prgEntry->NumElem() : 0; };
  void Flush(void);
};


// *******************
//  Class: ConfigInfo
// *******************

/// Application configuration information.
class ConfigInfo {
  friend std::ostream &operator<<(std::ostream &os, ConfigInfo &cfg);
  friend std::istream &operator>>(std::istream &is, ConfigInfo &cfg);

protected:
  DArray<ConfigInfo_Section> rgSection;
  int iSection, iEntry;  // iterator
  unsigned int wCycle;
  int opts;

  // add state
  DArray<int> daCurrentSection;

  unsigned int FindSection(const char *sbSectionName);
  unsigned int AppendSection(const char *sbSectionName);
  unsigned int ReplaceSection(unsigned int iSection);

  ConfigInfo_Entry *FindEntry(const char *sbSection, const char *sbName,
			      bool fScope);
  char *ExpandEntryMacros_Aux(ConfigInfo_Entry *pEntry, const char *sbSection,
			      bool fSaveResult);
  char *ExpandEntryMacros(ConfigInfo_Entry *pEntry, const char *sbSection,
			  bool fSaveResult);

  int ReadFile_Internal(const char *sbFile);

public:
  enum add_opts { // should be bitwise disjoint from Section opts
    ADD_DUPSEC_USEOLD   = 4,
    ADD_DUPSEC_REPLACE  = 8,
    ADD_DUPSEC_APPEND   = 16,
    OUT_EXPAND_EACH     = 32
  };

  ConfigInfo(void) : iSection(0), iEntry(0), wCycle(0) {
    DBG_TRACE("ConfigInfo::Constructor()");
    //opts = ADD_DUPSEC_USEOLD | ConfigInfo_Section::ADD_DUPNAME_APPEND;
    opts = ADD_DUPSEC_USEOLD | OUT_EXPAND_EACH | ConfigInfo_Section::ADD_DUPNAME_APPEND;
  };

  ~ConfigInfo(void) {
    DBG_TRACE("ConfigInfo::Destructor()");
    Flush();
  };

  unsigned int SizeBytes(void);
  int NumSection(void) { return rgSection.NumElem(); };
  void Flush(void);

  void SetOpts(int opts_in) { opts = opts_in; };
  void SetOpt(int opt) { opts |= opt; };
  void ClrOpt(int opt) { opts &= ~opt; };
  int  GetOpts(void) { return opts; };

  unsigned int AddSection(const char *sbSectionName);
  int RemoveSection(const char *sbSection);

  void AddEntry(unsigned int iSection, const char *sbName, const char *sbValue);
  void AddEntry(const char *sbSection, const char *sbName, const char *sbValue);

  int ReadFile(const char *sbFile, bool fFlushBeforeRead=true);  // file
  int ReadStr(const char *sbMem, bool fFlushBeforeRead=true);    // memory string
  int ReadBuf(const char *pBuf, int wLen, bool fFlushBeforeRead=true); // memory

  int WriteFile(const char *sbFile);  // file
  int WriteBuf(char *pBuf, int wLen); // memory

  char *FindDup(const char *sbSection, const char *sbName);
  char *Find(const char *sbSection, const char *sbName);
  char *Find(const char *sbName) { return Find("", sbName); };
  char *operator[](char *sbName) { return Find("", sbName); };
  DArray<char*> *FindAll(const char *sbSection, const char *sbName);
  DArray<char*> *FindAll(const char *sbName) { return FindAll("", sbName); };

  void Merge(ConfigInfo &cfgMerge, int opts=-1);

  // iterator
  const char *FirstSection(const char *sbSection=NULL);
  const char *NextSection(void);
  const char *FirstName(void);
  const char *NextName(void);
  const char *CurrentValue(void);
};


#endif /* H_CONFIGINFO */
// ****************************************************************************
