// -*- c++ -*-
// ****************************************************************************
//  Name: dc_darray.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 05/21/1997 : mdb - Created based on chaos++ example.
//        06/16/1997 : mdb - operator<< is not compiled correctly,
//                           probably since it is a template class.
//        01/11/2000 : mdb - Added Remove() method.
//        04/03/2000 : mdb - Can set extent size.
//  Desc: Implementation of a dynamic array class.
// ****************************************************************************
/** @file dc_darray.h */

#if !defined(H_DARRAY)
#define H_DARRAY

#include "dc_standard.h"
#include "dc_debug.h"

// turn off debugging for this file
#if defined(DCDBG_TRACE)
  #undef DBG_TRACE
  #define DBG_TRACE(STR)
#endif

#define DARRAY_DefaultExtentSize 10

//
// Notes: T must support assignment and have a constructor with no args
//

/// Implementation of a dynamic array class.
template <class T>
class DArray {
protected:
  T*    rgElem;
  Types::uint maxElem;
  Types::uint nElem;
  Types::uint wExtentSize;

  void ExtendArrayTo(Types::uint nNewSize) {
    DBG_TRACE("DArray::ExtendArrayTo()");
    T* rgTmp = rgElem;
    rgElem = new T[(maxElem = nNewSize)];
    if (rgTmp) {
      // copy over old elements
      for (Types::uint i=0; i<nElem; i++) {
	rgElem[i] = rgTmp[i];
      }
      delete [] rgTmp;
    }
  };

  void ExtendArray(Types::uint nExtent) {
    DBG_TRACE("DArray::ExtendArray()");
    ExtendArrayTo(maxElem + nExtent);
  };

public:
  DArray() {
    DBG_TRACE("DArray::Constructor()");
    rgElem = NULL;
    maxElem = nElem = 0;
    wExtentSize = DARRAY_DefaultExtentSize;
  };

  DArray(Types::uint maxElem_in,
	 Types::uint wExtentSize_in = DARRAY_DefaultExtentSize) {
    DBG_TRACE("DArray::Constructor(maxElem_in)");
    if ((maxElem = maxElem_in) == 0) {
      rgElem = NULL;
    } else {
      rgElem = new T[(this->nElem = maxElem_in)];
    }
    nElem = 0;
    wExtentSize = wExtentSize_in;
  };

  ~DArray() {
    DBG_TRACE("DArray::Destructor()");
    if (maxElem > 0) {
      DeleteNull(rgElem);
    }
  };

public:
  Types::uint NumElem(void) {
    return nElem;
  };

  void SetExtentSize(Types::uint wExtentSize_in) {
    wExtentSize = wExtentSize_in;
  };

#if 0
  // element assignment/lookup
  T& operator[](Types::uint i) {
    DBG_TRACE("T& Darray::operator[]");
    if (i >= maxElem) ExtendArrayTo(i+wExtentSize);
    if (i >= nElem) nElem = i+1;
    return rgElem[i];
  };

#else
  // element assignment (called for both lookup and assignment)
  T& operator[](Types::uint i) {
    DBG_TRACE("T& Darray::operator[] (lvalue)");
    if (i >= maxElem) ExtendArrayTo(i+wExtentSize);
    if (i >= nElem) nElem = i+1;
    return rgElem[i];
  };

  // element lookup (I don't know if/when this is ever called)
  const T& operator[](Types::uint i) const {
    DBG_TRACE("const T& Darray::operator[] (rvalue)");
    //if (i >= maxElem) ExtendArrayTo(i+wExtentSize);
    //if (i >= nElem) nElem = i+1;
    return rgElem[i];
  };
#endif /* 0 */

  // element lookup
  T& Get(Types::uint i) {
    DBG_TRACE("T& Darray::Get()");
    if (i >= nElem) {
      cerr << "dc_darray.h: Out of bounds array access: "
	   << i << " (0.." << nElem-1 << ")\n";
      return rgElem[0];  // dummy value
    }
    return rgElem[i];
  }

  T& GetFast(Types::uint i) { return rgElem[i]; };

  T& GetLast(void) { return rgElem[nElem-1]; };

  // element assignment
  T& Set(Types::uint i, T& t) {
    DBG_TRACE("T& Darray::Set()");
    if (i >= maxElem) ExtendArrayTo(i+wExtentSize);
    if (i >= nElem) nElem = i+1;
    rgElem[i] = t;
    return rgElem[i];
  };

  T& SetByVal(Types::uint i, T t) {
    DBG_TRACE("T& Darray::SetByVal()");
    if (i >= maxElem) ExtendArrayTo(i+wExtentSize);
    if (i >= nElem) nElem = i+1;
    rgElem[i] = t;
    return rgElem[i];
  };
  

  // element append
  // 04/05/2001 : mdb - Removed const on these, because I added an operator=()
  //                    to dc_configinfo.cpp that modifies the rhs, thus const
  //                    is too strict here.  Perhaps better would have been to
  //                    fix dc_configinfo.cpp to not need the odd operator=().
  //Types::uint Append(const T& elem) {
  Types::uint Append(T& elem) {
    DBG_TRACE("DArray::Append()");
    if (nElem == maxElem) ExtendArray(wExtentSize);
    rgElem[nElem++] = elem;
    return nElem-1;  // index of where added
  };

  //Types::uint AppendByVal(const T elem) {
  Types::uint AppendByVal(T elem) {
    DBG_TRACE("DArray::AppendByVal()");
    if (nElem == maxElem) ExtendArray(wExtentSize);
    rgElem[nElem++] = elem;
    return nElem-1;  // index of where added
  };

#if 0  /* this fails for many cases where operator==() is not defined */
  // this assumes that operator== is defined, and that no cleanup is
  // needed when replacing the elem.  (e.g., pointers)
  Types::uint ReplaceOrAppend(const T& elem, const T elemReplace) {
    bool fFound = false;
    for (Types::uint k=0; k<nElem; ++k) {
      if (rgElem[k] == elemReplace) {
	rgElem[k] = elem;
	return k;
      }
    }
    return Append(elem);
  };
#endif /* 0 */

  /// remove the last element
  Types::uint DeAppend(void) {
    DBG_TRACE("DArray::DeAppend()");
    nElem--;
    return nElem-1;  // new last index
  }

  /// whole array assignment
  DArray<T>& operator=(const DArray<T>& darray) {
    DBG_TRACE("DArray::operator=()");
    if (this != &darray) {  // avoid self copies
      if (maxElem > 0) delete [] rgElem;
      maxElem = nElem = darray.nElem;
      if (maxElem > 0) {
	rgElem = new T[maxElem];
      }
      for (Types::uint i=0; i<nElem; i++) {
	rgElem[i] = darray.rgElem[i];
      }
    }
    return *this;
  };

  /// empty entire structure
  DArray<T>& Flush(void) {
    DBG_TRACE("DArray::Flush()");
    TestDeleteNull(rgElem);
    maxElem = nElem = 0;
    return *this;
  }

  /// reverse whole array
  Types::uint Reverse(void) {
    T elemTmp;
    for (Types::uint i=0; i<nElem/2; ++i) {
      elemTmp = rgElem[i];
      rgElem[i] = rgElem[nElem - i - 1];
      rgElem[nElem - i - 1] = elemTmp;
    }
    return nElem;
  }

  /// remove an element
  Types::uint Remove(Types::uint i) {
    DBG_TRACE("DArray::Remove()");
    if (i >= nElem) return nElem;  // don't do anything
    if (nElem == 1) {
      Flush();
    } else {
      // copy all the elements
      for (Types::uint j=i+1; j<nElem; ++j) {
	rgElem[j-1] = rgElem[j];
      }
      --nElem;
    }
    return nElem;
  }

#if 0
  friend ostream& operator<<(ostream& os, DArray<T>& darray) {
    for (Types::uint i=0; i<darray.nElem; i++) {
      os << "[" << i << "] " << darray.rgElem[i] << endl;
    }
    return os;
  };
#endif /* 0 */
};


/// Implementation of a dynamic array class pointer.
template <class T>
class DArrayPtr : public DArray<T> {
public:
  ~DArrayPtr(void) {
    // This assumes all heap allocated objects, which is not always true
    //     if (maxElem > 0) {
    //       for (Types::uint i=0; i<maxElem; ++i) {
    // 	STestDeleteNull(rgElem[i]);
    //       }
    //     }
  };

public:
  // return: 0..nElem-1 if found, nElem if not found
  Types::uint Find(const T& elem) {
    for (Types::uint k=0; k<nElem; ++k) {
      if (rgElem[k] == elem) return k;
    }
    return nElem;
  };

  // this assumes that operator== is defined, and that no cleanup is
  // needed when replacing the elem.  (e.g., pointers)
  Types::uint ReplaceOrAppend(const T& elem, const T elemReplace) {
    Types::uint k;
    if ((k = Find(elemReplace)) < nElem) {
      rgElem[k] = elem;
      return k;
    }
    return AppendByVal(elem);
  };

  Types::uint AppendIfNotExist(const T&elem) {
    Types::uint k;
    if ((k = Find(elem)) == nElem) {
      return AppendByVal(elem);
    }
    return k;
  };

  // delete and shift the following elements
  Types::uint Delete(Types::uint iDel) {
    if (iDel < nElem) {
      for (Types::uint k=iDel+1; k<nElem; ++k) {
	rgElem[k-1] = rgElem[k];
      }
      --nElem;
    }
    return nElem;
  };

  Types::uint Delete(const T& elem) {
    return Delete(Find(elem));
  };

  // insert and shift the following elements
  Types::uint Insert(Types::uint iIns, const T& elem) {
    Types::uint iAppend;
    if (nElem == 0) return AppendByVal(elem);  // ignore iIns here as needed
    iAppend = AppendByVal(NULL);
    for (Types::uint k=iAppend; k>iIns; --k) {
      rgElem[k] = rgElem[k-1];
    }
    rgElem[iIns] = elem;
    return nElem;
  };
};


#endif /* H_DARRAY */
// ****************************************************************************
