// ****************************************************************************
//  Name: dc_debug.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 05/08/1997 : mdb - Created.
//        03/29/2001 : mdb - Added Linux sig handlers for a meaningful core.
//        01/03/2002 : mdb - Added DBG_StackDump().
//        01/10/2002 : mdb - Fixed bug in DBG_InstallSigHandlers() not deal
//                           with NULL input correctly (renato found).
//  Desc: Implementation of debug support routines.
// ****************************************************************************

#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <assert.h>

#include "dc_standard.h"
#include "dc_debug.h"
#include "dc_exec.h"

int DebugTrace::wDepth = 0;
DebugTrace *DebugTrace::pRoot = NULL;

DebugTrace::DebugTrace(char *sbFcnName, int fPrint) {
  sbName = sbFcnName;
  pPrev = pRoot;
  pRoot = this;
  this->fPrint = fPrint;
  if (fPrint) {
    fprintf(DCDBG_OUTPUT, "%*s>%s entry...\n",
	    DCDBG_INDENT_PER_DEPTH * wDepth, "", sbName);
    wDepth++;
  }
};

DebugTrace::~DebugTrace(void) {
  pRoot = pPrev;
  if (fPrint) {
    --wDepth;
    fprintf(DCDBG_OUTPUT, "%*s>%s exit.\n",
	    DCDBG_INDENT_PER_DEPTH * wDepth, "", sbName);
  }
};

/**
 * A very simple attempt to show a call stack.  Only routines that
 * have the DBG_TRACE macro call will be shown.
 */
void DebugTrace_Dump(char *sbMsg) {
  DebugTrace *p = DebugTrace::pRoot;

  fprintf(DCDBG_OUTPUT, "%*sDebugTrace Dump: %s ...\n",
	  DCDBG_DEPTH, "", sbMsg);
  if (!p) {
    fprintf(DCDBG_OUTPUT, "%*s-(empty)\n", DCDBG_DEPTH, "");
    return;
  }

  while (p) {
    fprintf(DCDBG_OUTPUT, "%*s-%s\n", DCDBG_DEPTH, "", p->sbName);
    p = p->pPrev;
  }
}


#include "dc_argv.h"

#include <sys/resource.h>
#include <sys/time.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

#if 0
// The default behavior in linux is to not dump core for a process
// that shares memory pages, and because linux threads shared mem
// pages, the offending thread is killed w/o dumping core.  Instead
// the last thread killed will dump the core file.  This hack installs
// sig handlers that should cause a core file for the offending
// thread.  (based on a posting by David Schwartz
// <davids@webmaster.com> to comp.programming.threads on 2001-01-10
// talking about how to get a meaningful core file on linux kernels)
//
// this still suffers from a single core file hit in parallel in a
// shared fs
//
//static void sig_handler_dump_core(int sig) {
static void sig_handler_dump_core(int, siginfo_t *, void *) {
  struct rlimit rl;
  if (fork() == 0) abort();
  rl.rlim_cur = 0;
  rl.rlim_max = 0;
  setrlimit(RLIMIT_CORE, &rl);
  getrlimit(RLIMIT_CORE, &rl);
  rl.rlim_cur = 0;
  setrlimit(RLIMIT_CORE, &rl);
  chdir("/");
  chdir("/proc");
  abort();
}
#endif /* 0 */

static char **DBG_envp = NULL;
static char  *DBG_sbProgramBinary = NULL;

void DBG_init(char *sbProgramBinary, char **envp) {
  DBG_sbProgramBinary = sbProgramBinary;
  DBG_envp = envp;
}


#define DBG_JITD_USE_XTERM
#define DBG_JITD_USE_BINARY_IN_PATH

struct DBG_handler_t {
  bool fStarted;
  Argv a;
  char **envp;
  int pfd1[2];
} DBG_handler;
static void sig_handler_start_debugger(int sig, siginfo_t *, void *) {
  pid_t pidChild;
  if (DBG_handler.fStarted) return;
  DBG_handler.fStarted = true;

  if ((pidChild = fork()) == 0) {  // child debugger process
#if defined(DBG_JITD_USE_BINARY_IN_PATH)
    execvp(DBG_handler.a.Get(0), DBG_handler.a.argv());
#else /* DBG_JITD_USE_BINARY_IN_PATH */
    execve(DBG_handler.a.Get(0), DBG_handler.a.argv(), DBG_handler.envp);
#endif /* DBG_JITD_USE_BINARY_IN_PATH */
    _exit(1);
    // If this fails, we are stuck -- no way to let parent know, which
    // will hang reexecuting the bad instruction.  Need a way to stop
    // it from happening.
  } else { // parent buggy main process
    //close(0);  // so we can interact with gdb instead
    // here we just go back and perform the stmt that caused the problem.
    // Once the debugger is attached we will stop as needed.
    if (sig == SIGABRT) while(1) ;  // return from abort will not stop
  }
}

// SIGILL, SIGBUS, SIGFPE, SIGSEGV, SIGSYS, SIGABRT
int DBG_InstallCoreSigHandlers(char *sbProgramBinary_in, char **envp_in) {
  int wRtn = 0;
  char tmp[80], tmp2[128];

  char *sbProgramBinary = (sbProgramBinary_in ?
			   sbProgramBinary_in : DBG_sbProgramBinary);

  //cout << "## Installing Core Signal Handlers for JIT debugging ..." << endl;
  DBG_handler.fStarted = false;
#if defined(DBG_JITD_USE_XTERM)
#if defined(DBG_JITD_USE_BINARY_IN_PATH)
  DBG_handler.a.Append("xterm");
#else /* DBG_JITD_USE_BINARY_IN_PATH */
  DBG_handler.a.Append("/usr/X11R6/bin/xterm");
#endif /* DBG_JITD_USE_BINARY_IN_PATH */
  DBG_handler.a.Append("-T");
  gethostname(tmp, 80);
  snprintf(tmp2, 128, "JIT Debugging for %s on host %s",
	   sbProgramBinary, tmp);
  DBG_handler.a.Append(StrDup(tmp2));
  DBG_handler.a.Append("-e");
#endif /* DBG_JITD_USE_XTERM */

  // CLH - added support for dbx (06/07/02)
#if defined(DEBUGGER_USE_GDB)
#if defined(DBG_JITD_USE_BINARY_IN_PATH)
  DBG_handler.a.Append("/usr/bin/gdb");
#else
  DBG_handler.a.Append("gdb");
#endif /* DBG_JITD_USE_BINARY_IN_PATH */
  DBG_handler.a.Append(sbProgramBinary);
  snprintf(tmp, 80, "%d", getpid());
  DBG_handler.a.Append(StrDup(tmp));
#elif defined(DEBUGGER_USE_DBX)
#if defined(DBG_JITD_USE_BINARY_IN_PATH)
  DBG_handler.a.Append("/usr/bin/dbx");
#else
  DBG_handler.a.Append("dbx");
#endif /* DBG_JITD_USE_BINARY_IN_PATH */
  snprintf(tmp, 80, "-a%d", getpid());
  DBG_handler.a.Append(StrDup(tmp));
  DBG_handler.a.Append(sbProgramBinary);
#endif
  DBG_handler.envp = envp_in ? envp_in : DBG_envp;

  struct sigaction s;
  s.sa_sigaction = sig_handler_start_debugger;
  sigemptyset(&s.sa_mask);
  s.sa_flags = SA_SIGINFO;
  if (sigaction(SIGILL, &s, NULL) != 0) {
    cerr << "DBG_InstallCoreSigHandlers(): Cannot install SIGILL handler: " << endl;
    // CLH - not MT safe - << strerror(errno) << endl;
    wRtn = 1;
  }
  s.sa_sigaction = sig_handler_start_debugger;
  sigemptyset(&s.sa_mask);
  s.sa_flags = SA_SIGINFO;
  if (sigaction(SIGBUS, &s, NULL) != 0) {
    cerr << "DBG_InstallCoreSigHandlers(): Cannot install SIGBUS handler: " << endl;
    // CLH - not MT safe - << strerror(errno) << endl;
    wRtn = 1;
  }
  s.sa_sigaction = sig_handler_start_debugger;
  sigemptyset(&s.sa_mask);
  s.sa_flags = SA_SIGINFO;
  if (sigaction(SIGFPE, &s, NULL) != 0) {
    cerr << "DBG_InstallCoreSigHandlers(): Cannot install SIGFPE handler: " << endl;
    // CLH - not MT safe - << strerror(errno) << endl;
    wRtn = 1;
  }
  s.sa_sigaction = sig_handler_start_debugger;
  sigemptyset(&s.sa_mask);
  s.sa_flags = SA_SIGINFO;
  if (sigaction(SIGSEGV, &s, NULL) != 0) {
    cerr << "DBG_InstallCoreSigHandlers(): Cannot install SIGSEGV handler: " << endl;
    // CLH - not MT safe - << strerror(errno) << endl;
    wRtn = 1;
  }
#if !defined(LINUX)  /* (not defined on redhat 6.1 ll.alliance.unm.edu) */
  s.sa_sigaction = sig_handler_start_debugger;
  sigemptyset(&s.sa_mask);
  s.sa_flags = SA_SIGINFO;
  if (sigaction(SIGSYS, &s, NULL) != 0) {
    cerr << "DBG_InstallCoreSigHandlers(): Cannot install SIGSYS handler: " << endl;
    // CLH - not MT safe - << strerror(errno) << endl;
    wRtn = 1;
  }
#endif /* !LINUX */
  s.sa_sigaction = sig_handler_start_debugger;
  sigemptyset(&s.sa_mask);
  s.sa_flags = SA_SIGINFO;
  if (sigaction(SIGABRT, &s, NULL) != 0) {
    cerr << "DBG_InstallCoreSigHandlers(): Cannot install SIGABRT handler: " << endl;
    // CLH - not MT safe - << strerror(errno) << endl;
    wRtn = 1;
  }
  return wRtn;
}

// yes, this is an ugly hack to get a stack dump, but I know of no
// other lightweight ELF decoders to approximate the same thing.
//
// pBufOut, wLenOut = where to write the output, if NULL,* then use stdout.
char *DBG_StackDump(char *pBufOut, int wLenOut, bool fCompactOutput,
		    int pid, char *sbProgramBinary_in, char **envp_in) {
  Argv a;
  Exec e;
  char tmp[40], tmpfile[80], buf[256];

  char *sbProgramBinary = (sbProgramBinary_in ?
			   sbProgramBinary_in : DBG_sbProgramBinary);
  char **envp = (envp_in ? envp_in : DBG_envp);
  int  fdStdin, fdStdout, fdStderr;
  int  wRtn;
  pid_t pidChild;

  if (!sbProgramBinary) {
    cerr << ErrorHeader(wERROR) << "DBG_StackDump(): must set binary name"
	 << endl;
    return NULL;
  }
  if (!envp) {
    cerr << ErrorHeader(wERROR) << "DBG_StackDump(): must set env ptr"
	 << endl;
    return NULL;
  }

  //cout << "DBG_StackDump() ..." << endl;
  a.Append("gdb");
  a.Append(sbProgramBinary);
  snprintf(tmp, sizeof(tmp), "%d", (pid == -1 ? getpid() : pid));
  a.AppendDup(tmp);
  //cout << a << endl;

  // run the command piped to collect stdout (ignore stderr as in csh)
  fdStdin = Exec::FD_PIPE;
  fdStdout = fdStderr = Exec::FD_PIPE;
  pidChild = e.ForkExec(a, fdStdin, fdStdout, fdStderr);
  assert(fdStdout >= 0);

  //snprintf(buf, sizeof(buf), "bt\n");
  snprintf(buf, sizeof(buf), "thread apply all bt\n");

  write(fdStdin, buf, strlen(buf));
  snprintf(buf, sizeof(buf), "quit\n");
  write(fdStdin, buf, strlen(buf));

  // read output from child
  bool fOutput = false;
  bool fFirstOnLine = true;
  while (1) {
    if ((wRtn = read(fdStdout, buf, sizeof(buf))) <= 0) break;
    if (fCompactOutput) {  // attempt to strip out gdb formatted stack frames
      for (int i=0; i<wRtn; ++i) {
	if (fFirstOnLine) {
	  fOutput = ((!fOutput && strchr("#", buf[i])) ||
		     (fOutput && strchr("# ", buf[i])));
	  fFirstOnLine = false;
	}
	if (fOutput) {
	  if (pBufOut) {
	    if (wLenOut-- > 0) *pBufOut++ = buf[i];
	  } else {
	    cout.put(buf[i]);
	  }
	}
	if (buf[i] == '\n') fFirstOnLine = true;
      }
    } else {  // output all gdb output
      if (pBufOut) {
	int wLenCopy = (wRtn <= wLenOut ? wRtn : wLenOut);
	memcpy(pBufOut, buf, wLenCopy);
	pBufOut += wLenCopy;
	wLenOut -= wLenCopy;
      } else {
	cout.write(buf, wRtn);
      }
    }
  }
  close(fdStdin);
  close(fdStdout);
  cout << endl;
  e.Wait(pidChild);
  //cout << "DBG_StackDump(): Child debugger finished." << endl;

  if (tmpfile) unlink(tmpfile);
  return (pBufOut ? pBufOut : (char*)1);
}


// ****************************************************************************
