// -*- c++ -*-
// ****************************************************************************
//  Name: dc_debug.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 05/08/1997 : mdb - Created.
//        05/21/1997 : mdb - Added DCDBG_OFF to shut off all output.
//  Desc: Implementation of debug support routines.
// ****************************************************************************
/** @file dc_debug.h */

#if !defined(H_DCDBG)
#define H_DCDBG

#include <stdio.h>

#if defined(DCDBG_OFF)

#else /* DCDBG_OFF */
#  define DCDBG_MAIN
#  define DCDBG_TRACE
#  define DCDBG_MSG
#endif /* DCDBG_OFF */


// Number of spaces to indent per level
#if !defined(DCDBG_INDENT_PER_DEPTH)
#define DCDBG_INDENT_PER_DEPTH 2
#endif

// Output stream to send debug messages to
#if !defined(DCDBG_OUTPUT)
#define DCDBG_OUTPUT stderr
#endif

// Control how DBG_MSG# output behaves
// indented with respect to DBG_TRACE depth
#define DCDBG_MSG_INDENT
// include file/line info
#define DCDBG_MSG_FILENO
#define DCDBG_MSG_LINENO


#if defined(DCDBG_MAIN)
#  define DBG(X) X
#else /* DCDBG_MAIN */
#  define DBG()
#endif /* DCDBG_MAIN */

#define DCDBG_DEPTH (DCDBG_INDENT_PER_DEPTH*DebugTrace::wDepth)

// Generate indentation and file/line info for prefix of output
#if defined(DCDBG_MSG_FILENO) && defined(DCDBG_MSG_LINENO)
#  define DCDBG_MSG_PREFIX_FMT  "%*s[%s:%d]"
#  define DCDBG_MSG_PREFIX_ARGS DCDBG_DEPTH, "", __FILE__, __LINE__
#elif defined(DCDBG_MSG_FILENO)
#  define DCDBG_MSG_PREFIX_FMT  "%*s[%s]"
#  define DCDBG_MSG_PREFIX_ARGS DCDBG_DEPTH, "", __FILE__
#elif defined(DCDBG_MSG_LINENO)
#  define DCDBG_MSG_PREFIX_FMT  "%*s[%d]"
#  define DCDBG_MSG_PREFIX_ARGS DCDBG_DEPTH, "", __LINE__
#else
#  define DCDBG_MSG_PREFIX_FMT  "%*s"
#  define DCDBG_MSG_PREFIX_ARGS DCDBG_DEPTH, ""
#endif


//
//  Inspired by code I saw once, but cannot remember where.
//

/**
 * Trace fcn calls - produces an indented call tree for routines that
 * add a local instance of this class to its entry points.  The auto
 * call of the destructor gives simple exit point output.
 */
class DebugTrace {
 public:
  static int wDepth;
  static DebugTrace *pRoot;
 protected:
  char       *sbName;
  DebugTrace *pPrev;
  int         fPrint;

 public:
  DebugTrace(char *sbFcnName, int fPrint = 1);
  ~DebugTrace(void);

  friend void DebugTrace_Dump(char *sbMsg);
};


//
//  DBG_TRACE = add trace object to call stack and output debug messages
//              for entry and exit.  Also, the depth is incremented forcing
//              DBG_MSG# output to also be indented.
//  DBG_TRACE_NOPRINT = similar to DBG_TRACE where a trace object is added to
//              the call stack, but the entry and exit points produce no
//              output, and the depth is not changed.
//
#if defined(DCDBG_TRACE)
#  define DBG_TRACE(STR)           DebugTrace _tmp_trace_var(STR)
#  define DBG_TRACE_NOPRINT(STR)   DebugTrace _tmp_trace_var(STR, 0)
#  define DBG_TRACE_DUMP(STR)      DebugTrace_Dump(STR)
#else
#  define DBG_TRACE(STR)
#  define DBG_TRACE_NOPRINT(STR)
#  define DBG_TRACE_DUMP(STR)
#endif


//
//  Inspired by uysal@cs and acha@cs code.
//
//  Debug output is indented to appropriate fcn call level to aid in
//  visually following debug messages.
//
#if defined(DCDBG_MSG)
#  define DBG_MSG(STR) \
          fprintf(DCDBG_OUTPUT, DCDBG_MSG_PREFIX_FMT STR, \
		  DCDBG_MSG_PREFIX_ARGS)
#  define DBG_MSG1(STR,A1) \
          fprintf(DCDBG_OUTPUT, DCDBG_MSG_PREFIX_FMT STR, \
		  DCDBG_MSG_PREFIX_ARGS, (A1))
#  define DBG_MSG2(STR,A1,A2) \
          fprintf(DCDBG_OUTPUT, DCDBG_MSG_PREFIX_FMT STR, \
		  DCDBG_MSG_PREFIX_ARGS, (A1), (A2))
#  define DBG_MSG3(STR,A1,A2,A3) \
          fprintf(DCDBG_OUTPUT, DCDBG_MSG_PREFIX_FMT STR, \
		  DCDBG_MSG_PREFIX_ARGS, (A1), (A2), (A3))
#  define DBG_MSG4(STR,A1,A2,A3,A4) \
          fprintf(DCDBG_OUTPUT, DCDBG_MSG_PREFIX_FMT STR, \
		  DCDBG_MSG_PREFIX_ARGS, (A1), (A2), (A3), (A4))
#  define DBG_MSG5(STR,A1,A2,A3,A4,A5) \
          fprintf(DCDBG_OUTPUT, DCDBG_MSG_PREFIX_FMT STR, \
		  DCDBG_MSG_PREFIX_ARGS, (A1), (A2), (A3), (A4), (A5))
#else
#  define DBG_MSG(STR)
#  define DBG_MSG1(STR,A1)
#  define DBG_MSG2(STR,A1,A2)
#  define DBG_MSG3(STR,A1,A2,A3)
#  define DBG_MSG4(STR,A1,A2,A3,A4)
#  define DBG_MSG5(STR,A1,A2,A3,A4,A5)
#endif

extern void DBG_init(char *sbProgramBinary, char **envp);
extern int DBG_InstallCoreSigHandlers(char *sbProgramBinary_in=NULL, char **envp_in=NULL);
extern char *DBG_StackDump(char *pBufOut, int wLenOut, bool fCompactOutput,
			   int pid=-1, char *sbProgramBinary_in=NULL, char **envp_in=NULL);

#endif /* H_DCDBG */
// ****************************************************************************
