// ****************************************************************************
//  Name: dc_aexec.cc
//  Athr: Michael D. Beynon (mdb)
//  Date: 05/19/1997 : mdb - Created.
//        10/28/2000 : mdb - Fixed bug in Wait() where EINTR was not handled.
//        12/10/2001 : mdb - Added ExecCollectOutput routines.
//        01/04/2002 : mdb - Attempt to avoid zombie in CollectOutput.
//  Desc: Routines for executing external commands (rsh, rexec, exec)
// ****************************************************************************

#include "dc_exec.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>


// ****************************************************************************
//  Class: Exec
// ****************************************************************************

int Exec::Chdir(char *sbCwd) {
  // change directory?
  if (sbCwd) {
    if (chdir(sbCwd) == -1) {
      cerr << ErrorHeader(wERROR) << "Exec::Chdir(): Failed chdir to '"
	   << sbCwd << "'.\n";
	// CLH - not MT safe - << ErrorIndent() << strerror(errno) << endl;
      return 1;
    }
  }
  return 0;
}

void Exec::Execvp(Argv &a) {
  if (a.GetNum() > 0) {
    // run command
    if (execvp(a.Get(0), a.argv()) == -1) {
      cerr << ErrorHeader(wERROR) << "Exec::Execvp(): Failed exec of '"
	   << a.Get(0) << "'.\n";
      // CLH - not MT safe - << ErrorIndent() << strerror(errno) << endl;
      exit(1);
    }
  }
}

// ****************************************************************************
//  Name: Exec
//  Date: 07/23/1997 : mdb - Created.
//        03/09/1998 : mdb - Supports both fd's for the same file.
//        12/11/2001 : mdb - Added a stdin fd.
// ****************************************************************************
/**
 * Execute the command represented by the Arg list.
 * @param fdStdin fd to use for command:
 *  \n>=0 = use this fd, 
 *  \nFD_PIPE(-1) = create pipe to cmd, 
 *  \nFD_PARENT(-2) = inheirit from parent, 
 *  \nFD_CLOSE(-3)  = close input
 *  \nif create pipe, these hold fd for pipe read
 * @param fdStdout fd to use for command:
 *  \n>=0 = use this fd, 
 *  \nFD_PIPE(-1) = create pipe to cmd, 
 *  \nFD_PARENT(-2) = inheirit from parent, 
 *  \nFD_CLOSE(-3)  = close input
 *  \nif create pipe, these hold fd for pipe read
 * @param fdStderr fd to use for command:
 *  \n>=0 = use this fd, 
 *  \nFD_PIPE(-1) = create pipe to cmd, 
 *  \nFD_PARENT(-2) = inheirit from parent, 
 *  \nFD_CLOSE(-3)  = close input
 *  \nif create pipe, these hold fd for pipe read
 * @return pid_t - pid of child command, <0 == error
 */
pid_t Exec::ForkExec(Argv &a, int &fdStdin, int &fdStdout, int &fdStderr,
		     char *sbCwd) {
  pid_t pid;
  int   fdChildStdin[2], fdChildStdout[2], fdChildStderr[2];  // [0]=r,[1]=w

  if ((fdStdin  == FD_PIPE && pipe(fdChildStdin)) ||
      (fdStdout == FD_PIPE && pipe(fdChildStdout)) ||
      (fdStderr == FD_PIPE && pipe(fdChildStderr))) {
    return -1;
  }

  if ((pid = fork()) < 0) {
    // error
    perror("Exec::ForkExec(): cannot fork");
    return -1;
  } else if (pid == 0) {
    // child
    if ((fdStdin == FD_PIPE && (dup2(fdChildStdin[0], STDIN_FILENO) < 0 ||
				close(fdChildStdin[0]) ||
				close(fdChildStdin[1]))) ||
	(fdStdin == FD_CLOSE && close(STDIN_FILENO)) ||
	(fdStdin >= 0 && (dup2(fdStdin, STDIN_FILENO) < 0 ||
			  close(fdStdin))) ||

	(fdStdout == FD_PIPE && (dup2(fdChildStdout[1], STDOUT_FILENO) < 0 ||
				 close(fdChildStdout[0]) ||
				 close(fdChildStdout[1]))) ||
	(fdStdout == FD_CLOSE && close(STDOUT_FILENO)) ||
	(fdStdout >= 0 && (dup2(fdStdout, STDOUT_FILENO) < 0 ||
			   close(fdStdout))) ||

	(fdStderr == FD_PIPE && (dup2(fdChildStderr[1], STDERR_FILENO) < 0 ||
				 close(fdChildStderr[0]) ||
				 close(fdChildStderr[1]))) ||
	(fdStderr == FD_CLOSE && close(STDERR_FILENO)) ||
	(fdStderr >= 0 && (dup2(fdStderr, STDERR_FILENO) < 0 ||
			   (fdStderr != fdStdout && close(fdStderr))))) {
      perror(a.Get(0));
      exit(1);
    }

    // change directory
    Chdir(sbCwd);

    // run command
    Execvp(a);
  }

  // parent
  if (fdStdin == FD_PIPE) {
    fdStdin = fdChildStdin[1]; // write to command stdin
    close(fdChildStdin[0]);
  }
  if (fdStdout == FD_PIPE) {
    fdStdout = fdChildStdout[0]; // read from command stdout
    close(fdChildStdout[1]);
  }
  if (fdStderr == FD_PIPE) {
    fdStderr = fdChildStderr[0]; // read from command stderr
    close(fdChildStderr[1]);
  }
  return pid;
}

pid_t Exec::ForkExec(Argv &a, char *sbCwd) {
  int fdStdout, fdStderr, fdStdin;
  fdStdin = FD_CLOSE;
  fdStdout = fdStderr = FD_PARENT;
  return ForkExec(a, fdStdin, fdStdout, fdStderr, sbCwd);
}

pid_t Exec::ForkExec(Argv &a, char *sbStdout, char *sbStderr, char *sbCwd) {
  int fdStdin, fdStdout, fdStderr;
  fdStdin = FD_CLOSE;
  if ((fdStdout = open(sbStdout, O_WRONLY|O_CREAT|O_TRUNC, 0666)) < 0) {
    cerr << ErrorHeader(wERROR) << "Exec::ForkExec(): Cannot redirect to '"
	 << sbStdout << "'.\n";
    // CLH - not MT safe - << ErrorIndent() << strerror(errno) << endl;
    return -1;
  }
  if (strcmp(sbStdout, sbStderr) == 0) {
    fdStderr = fdStdout;
  } else {
    if ((fdStderr = open(sbStderr, O_WRONLY|O_CREAT|O_TRUNC, 0666)) < 0) {
      cerr << ErrorHeader(wERROR) << "Exec::ForkExec(): Cannot redirect to '"
	   << sbStderr << "'.\n";
      // CLH - not MT safe - << ErrorIndent() << strerror(errno) << endl;
      if (fdStdout >= 0) close(fdStdout);
      return -1;
    }
  }
  return ForkExec(a, fdStdin, fdStdout, fdStderr, sbCwd);
}

int Exec::Wait(pid_t pid) {
  int stat;
  // wait for child to exit
  do {
    if (waitpid(pid, &stat, 0) < 0 && errno != EINTR) return -1;
  } while (! WIFEXITED(stat));
  return WEXITSTATUS(stat);
}

void Exec::AppendRcmd(Argv &aNew,
		      int eType, char *sbHost, char *sbRemoteCmd,
		      char *sbUser, int eFlags) {
  if (eType == Exec::REXEC_REMSH) {
    aNew.Append("remsh");
  } else if (eType == Exec::REXEC_RSH) {
    aNew.Append("rsh");
  } else if (eType == Exec::REXEC_SSH) {
    aNew.Append("ssh");
  } else {
    assert(0);
  }

  if (eFlags & REXEC_FLAGS_ALLOCATETTY) aNew.Append("-t");  // for ssh and gdb
  if (eFlags & REXEC_FLAGS_NOREADSTDIN) aNew.Append("-n");
  if (eFlags & REXEC_FLAGS_NOFORWARDX11) aNew.Append("-x");
  if (eFlags & REXEC_FLAGS_AUTOADDHOSTKEYS) {
    // 01/23/2002 : mdb - Umit found this option is not valid for commercial
    //                    SSH installed at OSU.  Removed for compatibility.
#if 0
    aNew.Append("-o");
    aNew.Append("StrictHostKeyChecking=no");
#endif
  }

  if (sbUser) {
    aNew.Append("-l");
    aNew.Append(sbUser);
  }

  aNew.Append(sbHost);
  aNew.Append(StrDup(sbRemoteCmd), true);
}


/**
 * execute the given command and collect the output into a string
 */
char *Exec::ExecCollectOutput_Alloc(Argv &aExec) {
  int fdStdin, fdStdout, fdStderr;
  Str str;
  pid_t pid;

  str.Set("");
  if (aExec.GetNum() == 0) return str.Dup();

  // run the command piped to collect stdout (ignore stderr as in csh)
  fdStdin = FD_CLOSE;
  fdStdout = fdStderr = FD_PIPE;
  pid = ForkExec(aExec, fdStdin, fdStdout, fdStderr);
  assert(fdStdout >= 0);

  const int maxbuf = 256;
  char  buf[maxbuf];
  int   wRtn;

  // read output from child
  while (1) {
    if ((wRtn = read(fdStdout, buf, maxbuf)) > 0) {
      str.NCat(buf, wRtn);
    } else {
      break;
    }
  }
  close(fdStdout);

  // 01/04/2001 : mdb - make sure it has stopped executing to avoid zombies
  kill(pid, SIGKILL);
  Wait(pid);

  str.RChopCh('\n');  // if last char is a newline, remove it

  // replace (as csh does) odd whitespace with a space
  str.Replace('\n', ' ');
  str.Replace('\r', ' ');
  str.Replace('\t', ' ');

  return str.Dup();
}

char *Exec::ExecCollectOutput_Alloc(char *sbExec) {
  Argv a;

  if (strlen(sbExec) == 0) {
    return ExecCollectOutput_Alloc(a);  // nothing to do
  }
  a.Append("/bin/sh");
  a.Append("-c");
  a.Append(sbExec);
  return ExecCollectOutput_Alloc(a);
}


// ****************************************************************************
