// -*- c++ -*-
// ****************************************************************************
//  Name: dc_hash.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 06/30/1995 : mdb - Created.
//        10/07/1997 : mdb - Turned into templates.
//        11/20/1997 : mdb - Changed signatures to make easier to use.
//                         - wUsedCount now reflects deletions too.
//                         - Fixed bug where all cells deleted -> endless loop
//                         - Changed Delete to not take a ptr to a value
//                           => always use ptr as class V for complex objects
//        11/21/1997 : mdb - Added wHopCount to avoid filling table with Del
//                           markers, by deleting cells with noone hopping over
//        12/03/1997 : mdb - Fixed problem with default hash fcns being < 0
//        12/07/1997 : mdb - Removed seperate HashTableCell template class
//                         - g++272 cannot handle member fcns that pass/return
//                           ptrs to the nested class.  As a hack, I changed
//                           ProbeInsert to use void* and did casting.  It
//                           would be cleaner if HashTable<K,V>::Cell* worked.
//        02/04/1998 : mdb - Added a Delete by ptr to cell.
//  Desc: Implementation of a hash table with double hashing.
// ****************************************************************************
/** @file dc_hash.h */

#ifndef H_HASH
#define H_HASH


#include "dc_standard.h"
#include "dc_random.h"

/* #define DCDBG_HASH */
/* #define DCDBG_HASH_PROBE */

#define DCDBG_HASH_FCNS  /* for validation */

// ****************************************************************************
//  Class Definitions
// ****************************************************************************

/// Implementation of a hash table with double hashing.
template<class K, class V>
class HashTable {
private:
  class Cell {
  public:
    K key;         // key to hash
    V value;       // user defined value
    bool bDel;     // T=deleted cell, F=filled cell
    int wHopCount; // number of other nodes that skipped us when added

  public:
    Cell(K &key_in, V &value_in) : key(key_in), value(value_in) {
      bDel = false; wHopCount = 0;
    };

    V &Delete(void) { bDel = true; return value; };  // not copied

    K &GetKey(void) { return key; };

    V &GetValue(void) { return value; };
    void SetValue(V &value) { this->value = value; };
  };

  int wM;                                // # cells in current hash table
  int wUsedCount;                        // # cells not (empty or deleted)
  Cell **pCells;                         // table of pointers to cells
  double dRehashLevel;                   // (0..1) % filled to force rehash
  int wIter;                             // idx of next item to return

  int (*h)(K &key, int wM);              // primary hash fcn
  int (*h2)(K &key, int wM);             // secondary hash fcn
  bool (*fcnEqual)(K &key1, K &key2); // equal test fcn
  ostream& (*fcnPrintKey)(ostream &os, K key);  // output key fcn
  ostream& (*fcnPrintValue)(ostream &os, V value);  // output value fcn

  int RehashSize(void);
  int ProbeInsert(K &key, V &value, int wM,
		  void **P,
		  void *POLD,
		  int wHopCountIncrement = 1);

  int h_default(K &key, int wM);   // built in primary hash fcn
  int h2_default(K &key, int wM);  // built in secondary hash fcn

  int Internal_Search(K &key, int &wProbeCount, int wHopCountIncrement = 0);

public:
  HashTable(void);
  HashTable(bool (*fcnEqual)(K&, K&),
	    ostream& (*fcnPrintKey)(ostream&, K) = NULL,
	    ostream& (*fcnPrintValue)(ostream&, V) = NULL,
	    int wM = 0,
	    int (*h)(K&, int) = NULL,
	    int (*h2)(K&, int) = NULL);
  ~HashTable(void);

  // same as Constructor, but allows later setting
  void init(bool (*fcnEqual)(K&, K&),
	    ostream& (*fcnPrintKey)(ostream&, K) = NULL,
	    ostream& (*fcnPrintValue)(ostream&, V) = NULL,
	    int wM = 0,
	    int (*h)(K&, int) = NULL,
	    int (*h2)(K&, int) = NULL);

  void setRehashLevel(double d) { dRehashLevel = d; };

  int Search(K &key, V &value);
  int SearchCell(K &key);
  V &GetValue(int iCell) { return pCells[iCell]->GetValue(); };
  int Insert(K &key, V &value);
  int Delete(K &key, V &value);

  int Delete(void *p);
  int Delete(int iCell);

  int GetRandom(Random &rand, V &value);

  void Flush(int (*fcnFlushValue)(V &value));

  int StartIter(void) { wIter = 0; return wUsedCount; };
  V &Next(void) {
    // return the cell pointed to by the current value of wIter.
    // setup wIter to point at the next cell for the future calls.
    for (; wIter < wM; wIter++) {
      if (pCells[wIter] && ! pCells[wIter]->bDel) break;
    }
    if (wIter >= wM) {  // the user called this too many times
      cerr << ErrorHeader(wFATAL) << "HashTable::Next(): iterated past end.\n";
      exit(1);
    }
    return pCells[wIter++]->GetValue();
  };

  void PrintKey(ostream &os, K &key);
  void PrintValue(ostream &os, V &value);
  void PrintKeyValue(ostream &os, K &key, V &value);
  void Print(ostream &os);

  int Check(ostream &os);
};


// ****************************************************************************
//  Name: HashTable::HashTable
//  Date: 07/09/1995 : mdb - Created.
//        11/20/1997 : mdb - Simplified call signature.
// ****************************************************************************
/**
 * Constructor.
 * @param fcnEqual - Callback fcn to determine if two keys are equal 
 * @param wM - initial size of hash table
 * @param h, h2 - primary hash fcn
 * @param h2 - secondary hash fcn
 * @param fcnPrintKey - Callback fcn to print the key field
 * @param fcnPrintValue - Callback fcn to print the value field
 */
template<class K, class V>
HashTable<K,V>::HashTable(void) {
};

template<class K, class V>
HashTable<K,V>::HashTable(bool (*fcnEqual)(K&, K&),
			  ostream& (*fcnPrintKey)(ostream&, K),
			  ostream& (*fcnPrintValue)(ostream&, V),
			  int wM,
			  int (*h)(K&, int),
			  int (*h2)(K&, int)) {
  init(fcnEqual, fcnPrintKey, fcnPrintValue, wM, h, h2);
};

template<class K, class V>
void HashTable<K,V>::init(bool (*fcnEqual)(K&, K&),
			  ostream& (*fcnPrintKey)(ostream&, K),
			  ostream& (*fcnPrintValue)(ostream&, V),
			  int wM,
			  int (*h)(K&, int),
			  int (*h2)(K&, int)) {
  if (wM < 1) wM = 7;  // default
  // adjust size to be prime to insure probe will cover all of table
  for (; !IsPrime(wM); wM++)
    ; // null
  this->wM = wM;

  wUsedCount = 0;
  dRehashLevel = 0.75;  // default

  this->fcnEqual = fcnEqual;
  this->h = h;
  this->h2 = h2;
  this->fcnPrintKey = fcnPrintKey;
  this->fcnPrintValue = fcnPrintValue;
  
  pCells = new Cell*[wM];  // array of pointers
  for (int i=0; i<wM; i++) {
    pCells[i] = NULL;  // all empty
  }
};


// ****************************************************************************
//  Name: HashTable::~HashTable
//  Date: 07/09/1995 : mdb - Created.
// ****************************************************************************
/**
 * Destructor.
 */
template<class K, class V>
HashTable<K,V>::~HashTable(void) {
  // delete cells
  if (pCells) {
    for (int i=0; i<wM; i++) {
      if (pCells[i]) delete pCells[i];
    }
    // delete pointers to cells
    delete [] pCells;
  }
};


// ****************************************************************************
//  Name: HashTable::h_default,h2_default
//  Date: 11/19/1997 : mdb - Created.
//        12/03/1997 : mdb - Use Types::uchar to avoid negative values.
// ****************************************************************************
/**
 * Built-in hash fcn if not specified.
 */
template<class K, class V>
int HashTable<K,V>::h_default(K &key, int wM) {
  Types::uchar *pBuf = (Types::uchar*) &key;
  int wLen = sizeof(key);
  int wRtn = 0;  // local return
  int i, wPi[20] = { 3,1,4,1,5,9,2,6,5,3,5,8,9,7,9,3,2,3,8,5 };
  for (i=0; i<wLen; i++) {
    wRtn += wPi[i]*(int)pBuf[i];
  }
  return wRtn % wM;
};
template<class K, class V>
int HashTable<K,V>::h2_default(K &key, int wM) {
  Types::uchar *pBuf = (Types::uchar*) &key;
  int wLen = sizeof(key);
  int wRtn = 0;  // local return
  int i, wE[20]  = { 2,7,1,8,2,8,1,8,2,8,4,5,9,0,4,5,2,3,5,4 };
  for (i=0; i<wLen; i++) {
    wRtn += wE[i]*(int)pBuf[i];
  }
  return (wRtn % (wM - 1)) + 1;
};


// ****************************************************************************
//  Name: HashTable::RehashSize
//  Date: 07/09/1995 : mdb - Created.
// ****************************************************************************
/**
 * Finds a "good" choice for the expanded hash table size.
 * @return int - new size
 */
template<class K, class V>
int HashTable<K,V>::RehashSize(void) {
  int wRtn;
  // find next prime number > twice the current size
  for (wRtn=2*wM+1; !IsPrime(wRtn); wRtn++)
    ; // null
  return wRtn;
};


// ****************************************************************************
//  Name: HashTable::Internal_Search
//  Date: 07/09/1995 : mdb - Created.
// ****************************************************************************
/**
 * Probe the hash table looking for the given key.
 * @param key  - key to look for
 * @param wProbeCount - number of probes to find item
 * @return int - index of pCells[]=found, -1=not found
 */
template<class K, class V>
int HashTable<K,V>::Internal_Search(K &key, int &wProbeCount,
				    int wHopCountIncrement) {
  int wH = h ? h(key, wM) : h_default(key, wM);
  int wH2 = h2 ? h2(key, wM) : h2_default(key, wM);
  int wProbe;

#if defined(DCDBG_HASH_FCNS)
  if (wH < 0 || wH > wM - 1) {
    cerr << ErrorHeader(wERROR) << "HashTable::Internal_Search(): "
	 << "Invalid primary hash fcn value = " << wH << endl;
  }
  if (wH2 < 0 || wH2 > wM - 1) {
    cerr << ErrorHeader(wERROR) << "HashTable::Internal_Search(): "
	 << "Invalid secondary hash fcn value = " << wH2 << endl;
  }
#endif /* DCDBG_HASH_FCNS */

  // search for first of: empty cell or our key (not deleted)
  for (wProbeCount=0; ; wProbeCount++) {
    wProbe = (wH + wProbeCount*wH2) % wM;
    //cout << "HashTable::Internal_Search(): wProbeCount=" << wProbeCount
    //     << ", wProbe=" << wProbe << ", wM=" << wM << endl;
    if (! pCells[wProbe]) {
      // empty cell
      break;
    } else if (wProbeCount > wM) {
      // cycled .. cannot be here
      break;
    } else if (pCells[wProbe]->bDel) {
      // deleted cell
      ; // skip this one
    } else if ((*fcnEqual)(pCells[wProbe]->key, key)) {
      // not empty, not deleted, keys match => found
      return wProbe;
    }

#if defined(DCDBG_HASH)
    cout << "HashTable<>::Internal_Search(): Probing past "
	 << wProbe << endl;
#endif /* DCDBG_HASH */
    if (wHopCountIncrement) {
      // here we are probing past another filled cell
      pCells[wProbe]->wHopCount += wHopCountIncrement;
    }
  }  // for

  // failed insert
  if (wHopCountIncrement == -1) {  // top level call
#if defined(DCDBG_HASH)
    cout << "HashTable<>::Internal_Search(): "
	 << "Search failed, fixing optimistic assumption\n";
#endif /* DCDBG_HASH */
    // need to go back and fix decremented hop counts
    return Internal_Search(key, wProbeCount, 1);
  }

  return -1;
};

template<class K, class V>
int HashTable<K,V>::SearchCell(K &key) {
  int wProbeCount;
  return Internal_Search(key, wProbeCount);
};

template<class K, class V>
int HashTable<K,V>::Search(K &key, V &value) {
  int wProbeCount;
  int i = Internal_Search(key, wProbeCount);

  if (i != -1) {  // found?
    value = pCells[i]->GetValue();  // copy it out
#if defined(DCDBG_HASH)
    cout << "HashTable<>::Search(): Found ";
    PrintKey(cout, key);
    cout << " in " << wProbeCount << " probes\n";
#endif /* DCDBG_HASH */
    return 0;
  }
  return 1;
};


// ****************************************************************************
//  Name: HashTable::ProbeInsert
//  Date: 07/09/1995 : mdb - Created.
// ****************************************************************************
/**
 * Probe the hash table and insert a new key.  Note this is generic
 * since it works with any table pointer.
 * @param key - key to add
 * @param value - user defined value for this key
 * @param p - cell table pointer to add into
 * @param pOld - previous cell table pointer:
 *  \nNULL=adding a new entry to the hash table
 *  \n!NULL=rehashing, so only move pointers between tables
 * @param wHopCountIncrement - used internally by routine, never call with this
 * @return Cell* - -1=duplicate, else=index of inserted cell
 */
template<class K, class V>
int HashTable<K,V>::ProbeInsert(K &key,
				V &value, int wM,
				void **P,
				void *POLD,
				int wHopCountIncrement) {
  HashTable<K,V>::Cell **p = (HashTable<K,V>::Cell**)P;
  HashTable<K,V>::Cell *pOld = (HashTable<K,V>::Cell*)POLD;
  int wH = h ? h(key, wM) : h_default(key, wM);
  int wH2 = h2 ? h2(key, wM) : h2_default(key, wM);

#if defined(DCDBG_HASH_FCNS)
  if (wH < 0 || wH > wM - 1) {
    cerr << ErrorHeader(wERROR) << "HashTable::ProbeInsert(): "
	 << "Invalid primary hash fcn value = " << wH << endl;
  }
  if (wH2 < 0 || wH2 > wM - 1) {
    cerr << ErrorHeader(wERROR) << "HashTable::ProbeInsert(): "
	 << "Invalid secondary hash fcn value = " << wH2 << endl;
  }
#endif /* DCDBG_HASH_FCNS */

  // search for first of: empty cell, deleted cell, or our key
  for (int i=0; ; i++) {
    int wProbe = (wH + i*wH2) % wM;
#if defined(DCDBG_HASH_PROBE)
    cout << "HashTable<>::ProbeInsert(): wProbe to " << wProbe << "\n";
#endif /* DCDBG_HASH */
    if (! p[wProbe]) {
      // empty cell
      if (pOld) {
	// a rehash insert -- only move pointer
	p[wProbe] = pOld;
      } else {
	// adding a new entry
#if defined(DCDBG_HASH_PROBE)
	cout << "HashTable<>::ProbeInsert(): Adding new entry at "
	     << wProbe << "\n";
#endif /* DCDBG_HASH */
	wUsedCount++;
	p[wProbe] = new Cell(key, value);
      }
      return wProbe;

    } else if (p[wProbe]->bDel) {
      // use the cell marked as deleted
#if defined(DCDBG_HASH_PROBE)
      cout << "HashTable<>::ProbeInsert(): Resusing deleted cell at "
	   << wProbe << "\n";
#endif /* DCDBG_HASH */
      p[wProbe]->key = key;
      p[wProbe]->value = value;
      p[wProbe]->bDel = false;
      wUsedCount++;
      return wProbe;

    } else if ((*fcnEqual)(p[wProbe]->key, key)) {
      // not empty, not deleted, keys match
#if defined(DCDBG_HASH_PROBE)
      cout << "HashTable<>::ProbeInsert(): Duplicate found at "
	   << wProbe << "\n";
#endif /* DCDBG_HASH */
      break;  // duplicate
    }

#if defined(DCDBG_HASH)
    cout << "HashTable<>::ProbeInsert(): Probing past " << wProbe << endl;
#endif /* DCDBG_HASH */
    // here we are probing past another filled cell
    p[wProbe]->wHopCount += wHopCountIncrement;

  }  // for

  // failed insert
  if (wHopCountIncrement == 1) {  // top level call
#if defined(DCDBG_HASH)
    cout << "HashTable<>::ProbeInsert(): "
	 << "Insert failed, fixing optimistic assumption\n";
#endif /* DCDBG_HASH */
    // need to go back and fix incremented hop counts
    return ProbeInsert(key, value, wM, (void**)p, (void*)pOld, -1);
  }
  
  return -1;
};


// ****************************************************************************
//  Name: HashTable::Insert
//  Date: 07/09/1995 : mdb - Created.
//        07/13/1995 : mdb - Fixed problem where wUsedCount should *not*
//                           be set to 0, since ProbeInsert() no longer
//                           blindly increments the counter.
// ****************************************************************************
/**
 * Insert a new key into the hash table and rehash afterwards if
 * needed.
 * @param key - key to add
 * @param value - user defined value for this key
 * @return Cell* - 1=duplicate, 0=ptr to inserted cell
 */
template<class K, class V>
int HashTable<K,V>::Insert(K &key, V &value) {
  int i, wProbe;

  // insert new key
  wProbe = ProbeInsert(key, value, wM, (void**)pCells, (void*)NULL);
  Cell *pRtn = pCells[wProbe];

  // check if we need to expand table and rehash
  if (pRtn && dRehashLevel > 0 &&
      (double)wUsedCount / (double)wM  >= dRehashLevel) {
    int wNewM = RehashSize();

#if defined(DCDBG_HASH)
    cout << "HashTable<>::Insert(): Resizing from "
	 << wM << " to " << wNewM << " cells.\n";
#endif /* DCDBG_HASH */

    // create new bigger table
    Cell **pNewCells = new Cell*[wNewM];
    for (i=0; i<wNewM; i++) {
      pNewCells[i] = NULL;  // all empty
    }

    // walk through old table and rehash into new table
    for (i=0; i<wM; i++) {
      // only do this for filled cells.  Del cells are purged.
      if (pCells[i] && ! pCells[i]->bDel) {
	ProbeInsert(pCells[i]->key, pCells[i]->value, wNewM, (void**)pNewCells,
		    (void*)pCells[i]);
      }
    }

    // install new table
    delete[] pCells;
    pCells = pNewCells;
    wM = wNewM;
  }

  return pRtn ? 0 : 1;
};


// ****************************************************************************
//  Name: HashTable::Delete
//  Date: 07/09/1995 : mdb - Created.
//        11/20/1997 : mdb - Now return value by ref
// ****************************************************************************
/**
 * Delete a key from the hash table.
 * @param key - key to delete
 * @param value - user defined value for this key (to allow outside deletion)
 * @return int - 0=deleted, !0=not found
 */
template<class K, class V>
int HashTable<K,V>::Delete(K &key, V &value) {
  int wProbeCount;
  int i = Internal_Search(key, wProbeCount, -1);

  if (i != -1) {  // found?
    if (pCells[i]->wHopCount == 0) {
      value = pCells[i]->GetValue();
      delete pCells[i];
      pCells[i] = NULL;
    } else {  // need to mark as Deleted
      value = pCells[i]->Delete();  // copied here
    }
    wUsedCount--;
    return 0;
  }
  return 1;
};


// ****************************************************************************
//  Name: HashTable::Delete
//  Date: 02/04/1998 : mdb - Created.
// ****************************************************************************
/**
 * Delete an entry from the hash table.
 * @param p - ptr to a cache cell entry
 * @return int - 0=deleted, !0=not found
 */
template<class K, class V>
int HashTable<K,V>::Delete(void *p) {
  int wRtn;
  V value;
  // want to make use of the Delete fcn attempt to reduce bDel entries
  // so we find the key, and simply call the normal delete
  Cell *pCell = (Cell*)p;
  if ((wRtn = Delete(pCell->key, value)) != 0) {
    cerr << ErrorHeader(wERROR)
	 << "HashTable::Delete(void*): Key pointed to not found in table.\n";
  }
  return wRtn;
};


// ****************************************************************************
//  Name: HashTable::Delete
//  Date: 02/04/1998 : mdb - Created.
//  Note: The cell index value can change as the hash table is automatically
//        expanded on insertions.
// ****************************************************************************
/** 
 * Delete an entry from the hash table.
 * @param iCell - index of cache cell entry
 * @return int - 0=deleted, !0=not found
 */
template<class K, class V>
int HashTable<K,V>::Delete(int iCell) {
  int wRtn;
  V value;
  // want to make use of the Delete fcn attempt to reduce bDel entries
  // so we find the key, and simply call the normal delete
  if (iCell < 0 || iCell >= wM) return -1;  // bad index value
  if ((wRtn = Delete(pCells[iCell]->key, value)) != 0) {
    cerr << ErrorHeader(wERROR)
	 << "HashTable::Delete(int): Key pointed to not found in table.\n";
  }
  return wRtn;
};


// ****************************************************************************
//  Name: HashTable::GetRandom
//  Date: 01/29/1998 : mdb - Created.
// ****************************************************************************
/**
 * Find a randomly chosen entry in the hash table.
 * @param rand - a Random class to use
 * @param V - value copied out of cell
 * @return int - index of pCells[]=found, -1=not found
 */
template<class K, class V>
int HashTable<K,V>::GetRandom(Random &rand, V &value) {
  int i, iStart;

  if (wUsedCount == 0) return -1;

  iStart = rand.GetRandRange(0, wM-1);
  // look forward from the random spot, wrapping around, until we find one
  for (i = iStart; i != iStart-1; i++) {
    if (i == wM) i=0;
    // check if a non-empty cell
    if (pCells[i] && !pCells[i]->bDel) {
      value = pCells[i]->GetValue();  // copy it out
      return i;
    }
  }
  return -1;  // none found: should never happen, because we check used count
};


// ****************************************************************************
//  Name: HashTable::Flush
//  Date: 02/05/1998 : mdb - Created.
// ****************************************************************************
/**
 * Empties the hash table, but leaves the size unchanged
 * @param fcnFlushValue - fcn to cleanup hash table value from a cell
 */
template<class K, class V>
void HashTable<K,V>::Flush(int (*fcnFlushValue)(V &value)) {
  for (int i=0; i<wM; i++) {
    if (pCells[i]) {  // either filled or deleted
      if (!pCells[i]->bDel) {
	if (fcnFlushValue) (*fcnFlushValue)(pCells[i]->GetValue());
	wUsedCount--;
      }
      // remove hash cell storage
      delete pCells[i];
      pCells[i] = NULL;
    }
  }
};


// ****************************************************************************
//  Name: HashTable::Print
//  Date: 07/09/1995 : mdb - Created.
//        07/11/1995 : mdb - Added formatting to keep indexes aligned.
//        02/11/1998 : mdb - Added seperate print fcns for key/value.
// ****************************************************************************
/**
 * Prints a representation of the hash table to the given stream.
 * @param os - stream to output to
 */
template<class K, class V>
void HashTable<K,V>::PrintKey(ostream &os, K &key) {
  if (fcnPrintKey) {
    (*fcnPrintKey)(os, key);
  } else {
    os << "<key>";
  }
};

template<class K, class V>
void HashTable<K,V>::PrintValue(ostream &os, V &value) {
  if (fcnPrintValue) {
    (*fcnPrintValue)(os, value);
  } else {
    os << "<value>";
  }
};

template<class K, class V>
void HashTable<K,V>::PrintKeyValue(ostream &os, K &key, V &value) {
  PrintKey(os, key);
  os << " -> ";
  PrintValue(os, value);
};

template<class K, class V>
void HashTable<K,V>::Print(ostream &os) {
  int wWidth = NumberDecimalDigits(wM-1);
  int wOld = os.width();
  int wH;
  cout << "HashTable: wM=" << wM << ", wUsedCount=" << wUsedCount
       << "\nindex: [wFirstHash,wHopCount] key -> value\n";
  for (int i=0; i<wM; i++) {
    os.width(wWidth);
    os << i;
    os.width(wOld);
    os << ": ";

    if (! pCells[i]) {
      os << "Empty\n";
    } else if (pCells[i]->bDel) {
      os << "[" << pCells[i]->wHopCount << "] ";
      os << "Del\n";
    } else {
      wH = h ? h(pCells[i]->key, wM) : h_default(pCells[i]->key, wM);
      os << "[" << wH << "," << pCells[i]->wHopCount << "] ";
      PrintKeyValue(os, pCells[i]->key, pCells[i]->value);
      os << endl;
    }
  }
};

template<class K, class V>
int HashTable<K,V>::Check(ostream &os) {
  int wH, wSum;

  os << "HashTable::Check(): Checking the validity of the hash table ...\n";
  os << "this= 0x" << hex << (int)this << dec
     << ", wM= " << wM
     << ", wUsedCount= " << wUsedCount
     << ", pCells= 0x" << hex << (int)pCells << dec
     << ", dRehashLevel= " << dRehashLevel
     << ", wIter= " << wIter << endl;

  if (pCells) {
    wSum = 0;
    for (int i=0; i<wM; i++) {
      os << i;
      os << ": ";

      if (! pCells[i]) {
	os << "Empty\n";
      } else if (pCells[i]->bDel) {
	os << "[" << pCells[i]->wHopCount << "] ";
	os << "Del\n";
      } else {
	wSum++;
	wH = h ? h(pCells[i]->key, wM) : h_default(pCells[i]->key, wM);
	os << "[" << wH << "," << pCells[i]->wHopCount << "] ";
	PrintKeyValue(os, pCells[i]->key, pCells[i]->value);
	os << endl;
      }
    }
  }
  if (wSum != wUsedCount) {
    os << "Count Error: Found " << wSum << " used cells, but there should be "
       << wUsedCount << ".\n";
  }
  return 0;
};


#endif // H_HASH
// ****************************************************************************

