// ****************************************************************************
//  Name: dc_initcollect.cpp
//  Athr: Michael D. Beynon (mdb) - beynon@cs.umd.edu
//  Date: 09/05/2000 : mdb - Created.
//  Desc: Specialized synchronization structure between two threads.
// ****************************************************************************

#include "dc_initcollect.h"
#include <iostream>
#include <errno.h>
#include <string.h>


//#define DCDBG_INITCOLLECT

//
//  class DC_InitiatorCollector_t
//

// ctor
DC_InitiatorCollector_t::DC_InitiatorCollector_t(bool fSync_in) {
  sem1.setExtLock(&mutex);
  sem2.setExtLock(&mutex);
  reset(fSync_in);
}

/**
 * should not be called when in use
 */
void DC_InitiatorCollector_t::reset(bool fSync_in) {
  mutex.lock();
  fSync = fSync_in;
  fDone = false;
  nResults = cResults = 0;
  sem1.init(0);
  sem2.init(0);
  if (!fSync) sem1.post();  // pre-release for async case
  mutex.unlock();
}

/**
 * complete initiator processing - know num results, wait for collector
 */
void DC_InitiatorCollector_t::initiator(int nResults_in) {
  if (nResults_in == 0) return;  // nothing to do
  mutex.lock();
  nResults = nResults_in;
  fDone = true;
#if defined(DCDBG_INITCOLLECT)
  cout << "initiator(" << hex << (unsigned int)this << dec << "): "
       << "nResults=" << nResults << " cResults=" << cResults
       << endl;
#endif /* DCDBG_INITCOLLECT */
  if (fSync) sem1.post();  // let collector go if any arrived
#if defined(DCDBG_INITCOLLECT)
  cout << "initiatorPart2(" << hex << (unsigned int)this << dec << "): "
       << "before sem2.WAIT ..."
       << endl;
#endif /* DCDBG_INITCOLLECT */
  if (!fSync && cResults >= nResults) sem2.post(); // collector already done
  sem2.wait();  // wait for collector to process all replies
  mutex.unlock();
}

/**
 * first half of initiator processing - know how many results to recv
 */
void DC_InitiatorCollector_t::initiatorPart1(int nResults_in) {
  if (nResults_in == 0) return;  // nothing to do
  mutex.lock();
  nResults = nResults_in;
  fDone = true;
#if defined(DCDBG_INITCOLLECT)
  cout << "initiatorPart1(" << hex << (unsigned int)this << dec << "): "
       << "nResults=" << nResults << " cResults=" << cResults
       << endl;
#endif /* DCDBG_INITCOLLECT */
  if (fSync) sem1.post();  // let collector go if any arrived
  mutex.unlock();
}

/**
 * second half of initiator processing - block until all recv by collector
 */
void DC_InitiatorCollector_t::initiatorPart2(void) {
  if (nResults == 0) return;  // nothing to do
  mutex.lock();
#if defined(DCDBG_INITCOLLECT)
  cout << "initiatorPart2(" << hex << (unsigned int)this << dec << "): "
       << "before sem2.WAIT ..."
       << endl;
#endif /* DCDBG_INITCOLLECT */
  if (!fSync && cResults >= nResults) sem2.post(); // collector already done
  sem2.wait();  // wait for collector to process all replies
  mutex.unlock();
}

/**
 * complete collector processing - recv a result, if last release initiator
 */
void DC_InitiatorCollector_t::collector(int nRecv) {
  mutex.lock();
  if (cResults == 0) sem1.wait();  // wait for initiatorDone()
  cResults += nRecv;
#if defined(DCDBG_INITCOLLECT)
  cout << "collector(" << hex << (unsigned int)this << dec << "): "
       << "nResults=" << nResults << " cResults=" << cResults
       << " " << (fDone && cResults >= nResults ? "sem2.POST" : "")
       << endl;
#endif /* DCDBG_INITCOLLECT */
  if (fDone && cResults >= nResults) sem2.post();
  mutex.unlock();
}

/**
 * first half of collector processing - recv a result (may block if sync)
 */
void DC_InitiatorCollector_t::collectorPart1(int nRecv) {
  mutex.lock();
  if (cResults == 0) sem1.wait();  // wait for initiatorDone()
  cResults += nRecv;
#if defined(DCDBG_INITCOLLECT)
  cout << "collectorPart1(" << hex << (unsigned int)this << dec << "): "
       << "nResults=" << nResults << " cResults=" << cResults
       << endl;
#endif /* DCDBG_INITCOLLECT */
  mutex.unlock();
}

/**
 * second half of collector processing - if last release initiator
 */
void DC_InitiatorCollector_t::collectorPart2(void) {
  mutex.lock();
#if defined(DCDBG_INITCOLLECT)
  cout << "collectorPart2(" << hex << (unsigned int)this << dec << "): "
       << "nResults=" << nResults << " cResults=" << cResults
       << " " << (fDone && cResults >= nResults ? "sem2.POST" : "")
       << endl;
#endif /* DCDBG_INITCOLLECT */
  if (fDone && cResults >= nResults) sem2.post();
  mutex.unlock();
}


// ****************************************************************************
