// -*- c++ -*-
// ****************************************************************************
//  Name: dc_lists.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 06/17/1997 : mdb - Created.
//        12/07/1997 : mdb - Removed seperate Node class from SLList.
//                         - Added DLList.
//        01/26/1998 : mdb - Hacked inserts to return ptr to list node.
//                         - Added delete by ptr to list node.
//                         - Every DLList node has a ptr to the parent class.
//  Desc: Simple list classes.
// ****************************************************************************
/** @file dc_lists.h */

#if !defined(H_LISTS)
#define H_LISTS

#include "dc_standard.h"


// ****************************************************************************
//  Class: SLList
// ****************************************************************************

/// Simple list class.
template <class T>
class SLList {

  /// nested node class to avoid another template defn
  class Item {
  public:
    T     data;
    Item *pNext;

    Item(T &dat, Item *pNxt) : data(dat), pNext(pNxt) {};
    ~Item() { pNext = NULL; };
  };

protected:
  Item *pHead, *pTail;
  int wNum;

  Item *pIter;  // points to next list element in iteration

public:
  SLList(void) { pHead = pTail = NULL; wNum = 0; pIter = NULL; };
  ~SLList(void) { Flush(); };
  void InsHead(T &data);
  void InsTail(T &data);
  int getNum(void) { return wNum; };
  T *getHead(void) { return (pHead ? &(pHead->data) : (T*)NULL); };
  T *getTail(void) { return (pTail ? &(pTail->data) : (T*)NULL); };
  int DelHead(T &data);
  int DelHead(void);

  void Flush(void);

  int StartIter(void) { pIter = pHead; return wNum; };
  T *pFirst(void);
  T *pNext(void);
  T &First(void);
  T &Next(void);
  bool Eol(void) { return (pIter ? false : true); };
};

template <class T>
inline void SLList<T>::InsHead(T &data) {
  Item *p = pHead;
  pHead = new Item(data, p);
  if (!pTail) pTail = pHead;
  wNum++;
}

template <class T>
inline void SLList<T>::InsTail(T &data) {
  Item *p = pTail;
  pTail = new Item(data, NULL);
  if (p) p->pNext = pTail;
  if (!pHead) pHead = pTail;
  wNum++;
}

template <class T>
inline int SLList<T>::DelHead(T &data) {
  Item *p = pHead;
  if (pHead) {
    data = pHead->data;
    pHead = pHead->pNext;
    if (!pHead) pTail = NULL;
    delete p;
    wNum--;
    return 0;
  }
  return 1;
}

template <class T>
inline int SLList<T>::DelHead(void) {
  Item *p = pHead;
  if (pHead) {
    pHead = pHead->pNext;
    if (!pHead) pTail = NULL;
    delete p;
    wNum--;
    return 0;
  }
  return 1;
}

template <class T>
inline void SLList<T>::Flush(void) {
  Item *p = pHead;
  while (p) {
    pHead = pHead->pNext;
    delete p;
    p = pHead;
  }
  pTail = pHead;  // NULL
  wNum = 0;
  pIter = NULL;
}

template <class T>
inline T *SLList<T>::pFirst(void) {
  pIter = pHead;
  return pIter ? pNext() : (T*)NULL;
}

template <class T>
inline T *SLList<T>::pNext(void) {
  if (pIter) {
    T *pdataRtn = &(pIter->data);
    pIter = pIter->pNext;
    return pdataRtn;
  }
  return (T*)NULL;
}

template <class T>
inline T &SLList<T>::First(void) {
  pIter = pHead;
  if (pIter) {
    return Next();
  }
  cerr << ErrorHeader(wERROR) << "SLList::First() NULL iterator.";
  exit(1);
  return Next();  // make compiler happy
}

template <class T>
inline T &SLList<T>::Next(void) {
  T *pdataRtn;
  if (pIter) {
    pdataRtn = &(pIter->data);
    pIter = pIter->pNext;
    return *pdataRtn;
  }
  cerr << ErrorHeader(wERROR) << "SLList::Next() NULL iterator.";
  exit(1);
  return *pdataRtn;  // make compiler happy
}


// ****************************************************************************
//  Class: DLList
// ****************************************************************************

/// Simple list class.
template <class T>
class DLList {

  /// nested node class to avoid another template defn
  class Item {
  public:
    T     data;
    Item *pNext, *pPrev;
    DLList<T> *pList;

    Item(T &dat, Item *pNxt, Item *pPrv, DLList<T> *pLst)
      : data(dat), pNext(pNxt), pPrev(pPrv), pList(pLst) {};
    ~Item() { pNext = pPrev = NULL; };
  };

protected:
  Item *pHead, *pTail;
  int wNum;

  Item *pIter;  // points to current list element in iteration (NULL=at start)

public:
  DLList(void) { pHead = pTail = NULL; wNum = 0; pIter = NULL; };
  ~DLList(void) { Flush(); };
  void *InsHead(T &data);
  void *InsHead(const T &data) { T d = data; return InsHead(d); };
  void *InsHeadByVal(T data) { T d = data; return InsHead(d); };
  void *InsTail(T &data);
  void *InsTail(const T &data) { T d = data; return InsTail(d); };
  void *InsTailByVal(T data) { T d = data; return InsTail(d); };
  int getNum(void) { return wNum; };
  T *getHead(void) { return (pHead ? &(pHead->data) : (T*)NULL); };
  T *getTail(void) { return (pTail ? &(pTail->data) : (T*)NULL); };
  int DelHead(T &data);
  int DelTail(T &data);
  int DelItem(void *p);
  int Unlink(void *p);

  void Flush(void);

  int StartIter(void) { pIter = NULL; return wNum; };
  T *pFirst(void);
  T *pNext(void);
  T &First(void);
  T &Next(void); 
  int DelIter(void);
  // InsBeforeIter()  to be added
  // InsAfterIter()  to be added
  bool Eol(void) { return (pIter == pTail ? true : false); };
};

template <class T>
inline void *DLList<T>::InsHead(T &data) {
  Item *p = pHead;
  pHead = new Item(data, p, NULL, this);
  if (p) p->pPrev = pHead;
  if (!pTail) pTail = pHead;
  wNum++;
  return (void*)pHead;
}

template <class T>
inline void *DLList<T>::InsTail(T &data) {
  Item *p = pTail;
  pTail = new Item(data, NULL, p, this);
  if (p) p->pNext = pTail;
  if (!pHead) pHead = pTail;
  wNum++;
  return (void*)pTail;
}

template <class T>
inline int DLList<T>::DelHead(T &data) {
  Item *p = pHead;
  if (pHead) {
    data = pHead->data;
    pHead = pHead->pNext;
    if (pHead) {
      pHead->pPrev = NULL;
    } else {
      pTail = NULL;
    }
    delete p;
    wNum--;
    return 0;
  }
  return 1;
}

template <class T>
inline int DLList<T>::DelTail(T &data) {
  Item *p = pTail;
  if (pTail) {
    data = pTail->data;
    pTail = pTail->pPrev;
    if (pTail) {
      pTail->pNext = NULL;
    } else {
      pHead = NULL;
    }
    delete p;
    wNum--;
    return 0;
  }
  return 1;
}

template <class T>
inline int DLList<T>::Unlink(void *p_in) {
  Item *p = (Item*)p_in;
  // BUGBUG: This will mess up a current iterator, so beware
  if (p->pNext) {
    p->pNext->pPrev = p->pPrev;
  } else {  // deleting tail
    pTail = p->pPrev;
  }
  if (p->pPrev) {
    p->pPrev->pNext = p->pNext;
  } else {  // deleting head
    pHead = p->pNext;
  }
  wNum--;
  return 0;
}

template <class T>
inline int DLList<T>::DelItem(void *p_in) {
  Unlink(p_in);
  delete (Item*)p_in;
  return 0;
}

template <class T>
inline void DLList<T>::Flush(void) {
  Item *p = pHead;
  while (p) {
    pHead = pHead->pNext;
    delete p;
    p = pHead;
  }
  pTail = pHead;  // NULL
  wNum = 0;
  pIter = NULL;
}

template <class T>
inline T *DLList<T>::pFirst(void) {
  pIter = NULL;
  return pNext();
}

template <class T>
inline T *DLList<T>::pNext(void) {
  Item *pTmp;
  if (pIter) {
    pTmp = pIter->pNext;
  } else {
    pTmp = pHead;
  }
  if (!pTmp) return (T*)NULL;  // EOL
  pIter = pTmp;
  return &(pIter->data);
}

template <class T>
inline T &DLList<T>::First(void) {
  pIter = NULL;
  return Next();
}

template <class T>
inline T &DLList<T>::Next(void) {
  Item *pTmp;
  if (pIter) {
    pTmp = pIter->pNext;
  } else {
    pTmp = pHead;
  }
  if (!pTmp) {  // EOL
    cerr << ErrorHeader(wERROR) << "DLList::Next() NULL iterator.";
    exit(1);
  }
  pIter = pTmp;
  return pIter->data;
}

template <class T>
inline int DLList<T>::DelIter(void) {
  Item *p = pIter;
  if (pIter) {
    pIter = pIter->pPrev;  // move to last one seen
    if (p->pNext) {
      p->pNext->pPrev = p->pPrev;
    } else {  // deleting tail
      pTail = p->pPrev;
    }
    if (p->pPrev) {
      p->pPrev->pNext = p->pNext;
    } else {  // deleting head
      pHead = p->pNext;
    }
    delete p;
    wNum--;
    return 0;
  }
  return 1;
}

#endif /* H_LISTS */
// ****************************************************************************
