// -*- c++ -*-
// ****************************************************************************
//  Name: dc_map.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 02/11/1998 : mdb - Created.
//  Desc: Implementation of a map class that uses the hashtable class.
// ****************************************************************************
/** @file dc_map.h */

#ifndef H_MAP
#define H_MAP

#include "dc_standard.h"
#include "dc_hash.h"

/* #define DCDBG_MAP */
/* #define DCDBG_MAP_SHOWTABLE */


// Note: Since this class relies on the template hash table, both must
//       be instantiated by the caller.


// ****************************************************************************
//  Class Definitions
// ****************************************************************************

/// Implementation of a map class that uses the hashtable class.
template<class K, class V>
class Map {
private:
  HashTable<K,V> hash;
  V valueNotFound;

public:
  Map(void) { };
  Map(V valueNotFound,
      bool (*fcnEqual)(K&, K&),
      ostream& (*fcnPrintKey)(ostream&, K) = NULL,
      ostream& (*fcnPrintValue)(ostream&, V) = NULL) {
    init(valueNotFound, fcnEqual, fcnPrintKey, fcnPrintValue);
  };

  // same as Constructor, but allows later setting
  void init(V valueNotFound_in,
	    bool (*fcnEqual)(K&, K&),
	    ostream& (*fcnPrintKey)(ostream&, K) = NULL,
	    ostream& (*fcnPrintValue)(ostream&, V) = NULL) {
    valueNotFound = valueNotFound_in;
    hash.init(fcnEqual, fcnPrintKey, fcnPrintValue);
  };

  int set(K &key, V &value);
  V  &get(K &key);
  int del(K &key);

  void Print(ostream &os) { hash.Print(os); };
  void Flush(void) { hash.Flush(NULL); };
};


// ****************************************************************************
//  Name: Map::set
//  Date: 02/11/1998 : mdb - Created.
// ****************************************************************************
/**
 * Adds a new mapping for the given key, value.
 * @param key - mapping to add
 * @param value - mapping to add
 * @return int - 0=success, 1=failed (key already mapped)
 */
template<class K, class V>
int Map<K,V>::set(K &key, V &value) {
  // add to hash table
  if (hash.Insert(key, value)) {
#if defined(DCDBG_MAP)
    cerr << ErrorHeader(wERROR)
	 << "Map::set(): Failed setting of new mapping: ";
    hash.PrintKeyValue(cerr, key, value);
    cerr << endl;
#endif /* DCDBG_MAP */
    return 1;
  }
#if defined(DCDBG_MAP)
  cout << "Map::set(): Added ";
  hash.PrintKeyValue(cout, key, value);
  cout << endl;
#if defined(DCDBG_MAP_SHOWTABLE)
  hash.Print(cout);
#endif /* DCDBG_MAP_SHOWTABLE */
#endif /* DCDBG_MAP */
  return 0;
}


// ****************************************************************************
//  Name: Map::get
//  Date: 02/11/1998 : mdb - Created.
// ****************************************************************************
/**
 * Lookup a mapping for the given key.
 * @param key - mapping to locate
 * @return V - valueNotFound=failed, otherwise=success
 */
template<class K, class V>
V &Map<K,V>::get(K &key) {
  int iCell;
  if ((iCell = hash.SearchCell(key)) == -1) {  // not found?
#if defined(DCDBG_MAP)
    cerr << ErrorHeader(wERROR) << "Map::get(): Search for ";
    hash.PrintKey(cerr, key);
    cerr << " mapping failed.\n";
#endif /* DCDBG_MAP */
    return valueNotFound;
  }
#if defined(DCDBG_MAP)
  cout << "Map::get(): Found ";
  hash.PrintKeyValue(cout, key, hash.GetValue(iCell));
  cout << endl;
#if defined(DCDBG_MAP_SHOWTABLE)
  hash.Print(cout);
#endif /* DCDBG_MAP_SHOWTABLE */
#endif /* DCDBG_MAP */
  return hash.GetValue(iCell);
}


// ****************************************************************************
//  Name: Map::del
//  Date: 02/11/1998 : mdb - Created.
// ****************************************************************************
/** 
 * Delete the mapping for the given key.
 * @param key - mapping to delete
 * @param int - 0=success, 1=failed (key not found)
 */
template<class K, class V>
int Map<K,V>::del(K &key) {
  V valueTmp;

  if (hash.Delete(key, valueTmp)) {
#if defined(DCDBG_MAP)
    cerr << ErrorHeader(wERROR) << "Map::del(): Search for ";
    hash.PrintKey(cerr, key);
    cerr << " mapping failed.\n";
#endif /* DCDBG_MAP */
    return 1;
  }
#if defined(DCDBG_MAP)
  cout << "Map::del(): Deleted ";
  hash.PrintKeyValue(cout, key, value);
  cout << endl;
#if defined(DCDBG_MAP_SHOWTABLE)
  hash.Print(cout);
#endif /* DCDBG_MAP_SHOWTABLE */
#endif /* DCDBG_MAP */
  // BUGBUG: Here we ignore possible cleanup of the key and value
  return 0;
}


#endif /* H_MAP */
// ****************************************************************************

