// ****************************************************************************
//  Name: memory.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 06/04/1997 : mdb - Created.
//        06/25/1997 : mdb - Moved all non-inlines from dc_memory.h
//        01/04/2000 : mdb - Added ExtractAppend() to MemoryBuf.
//        04/17/2001 : mdb - Elim redundant work in MemoryBuf::New,Delete,Set.
//  Desc: Simple memory classes.
// ****************************************************************************

#include "dc_memory.h"

#include <ctype.h>
#include <stdarg.h>

#ifdef _WIN32
#include "win32.h"
#endif

// ****************************************************************************
//  Internal Utility Routines
// ****************************************************************************


// ****************************************************************************
//  Name: _Memory_Dump
//  Date: 06/04/1997 : mdb - Created.
//        08/12/1999 : mdb - Moved to static routine.
// ****************************************************************************
/**
 * Output the block as hex integers.
 * @param pBuf - pointer of buffer to use
 * @param wNumBytes - length of buffer to use
 */
static void _Memory_Dump(void *pBuf, int wNumBytes) {
  Types::uint2 *rgw = (Types::uint2 *)pBuf;
  int i, j;
  int wElemSize = sizeof(Types::uint2);
  int nElem = wNumBytes / wElemSize;
  int wDigits = NumberDecimalDigits((long)nElem);
  int nElemPerLine = 14;
  int nLines = nElem/nElemPerLine;
  if (nLines*nElemPerLine < nElem) nLines++;

  for (i=0; i<nLines; i++) {
    // hex shorts
    printf("%*x:", wDigits, i*wElemSize);
    for (j=0; j<nElemPerLine; j++) {
      if ((i*nElemPerLine + j) >= nElem) break;
      printf(" %04x", rgw[i*nElemPerLine + j]);  // BUGBUG: endian order
    }
    printf("\n");
    // ascii chars
    printf("%*c:", wDigits, ' ');
    for (j=0; j<nElemPerLine; j++) {
      if ((i*nElemPerLine + j) >= nElem) break;
      Types::uint2 w = rgw[i*nElemPerLine + j];
      Types::uchar c1 = ((Types::uchar*)&w)[0];
      Types::uchar c2 = ((Types::uchar*)&w)[1];
      printf("  ");
      if (isalnum(c1) || ispunct(c1)) {
	printf("%c", c1);
      } else {
	printf(" ");
      }
      printf(" ");
      if (isalnum(c2) || ispunct(c2)) {
	printf("%c", c2);
      } else {
	printf(" ");
      }
    }
    printf("\n");
  }
}

// ****************************************************************************
//  Name: _Memory_Checksum, _Memory_ChecksumCh
//  Date: 06/04/1997 : mdb - Created.
//        08/12/1999 : mdb - Moved to static routines.
// ****************************************************************************
/**
 * Calculate the CRC-16 checksum of the block.
 * @param pBuf - pointer of buffer to use
 * @param wNumBytes - length of buffer to use
 * @return Types::uint - checksum value
 */
static Types::uint _Memory_ChecksumCh(Types::uint wSum, Types::uchar ch) {
  int i, fApplyGen;
  // take checksum so far, and process one char
  for (i=8; i; i--) {
    // need to apply generator after shift?
    fApplyGen = (wSum & 0x8000);
    // shift sum left, move msbit of char into lsbit of sum
    wSum = (wSum << 1) | ((ch & 0x80) >> 7);
    ch <<= 1;
    // apply generator
    //  - only need x^15 + x^2 + 1 now since x^16 will result in 1 xor 1 = 0
    //  - thus  1000 0000 0000 0101 => 0x8005
    if (fApplyGen) wSum ^= 0x8005;
  }
  return (wSum & 0xffff);
}

static Types::uint _Memory_Checksum(Types::uint wSum,
				    void *pBuf, int wNumBytes, bool fLast) {
  Types::uchar *rgch = (Types::uchar*)pBuf;
  int i;

  // process buffer
  for (i=0; i<wNumBytes; i++) {
    wSum = _Memory_ChecksumCh(wSum, *rgch);
    rgch++;
  }

  // handle appended r-1 zeros (16=2chars)
  if (fLast) {
    wSum = _Memory_ChecksumCh(wSum, '\0');
    wSum = _Memory_ChecksumCh(wSum, '\0');
  }

  return wSum;
}


// ****************************************************************************
//  Class: MemBuf
// ****************************************************************************


// ****************************************************************************
//  Name: Dump
//  Date: 06/04/1997 : mdb - Created.
//        08/12/1999 : mdb - Calls static routine.
// ****************************************************************************
/**
 * Output the block as hex integers.
 * @param wMaxBytes - a limit to the size of the output (-1 = all buffer)
 */
void MemBuf::Dump(int wMaxBytes) {
  int wNumBytes = (wMaxBytes == -1 ? wMax : StandardMin((int)wMax,wMaxBytes));

  printf("MemBuf::Dump(): Address=%x wMax=%d DumpLength=%d DumpChecksum=%d\n",
	 (Types::uint)pBuf, wMax, wNumBytes, Checksum(wMaxBytes));

  ::_Memory_Dump(pBuf, wNumBytes);
}


// ****************************************************************************
//  Name: Checksum
//  Date: 06/04/1997 : mdb - Created.
//        08/12/1999 : mdb - Calls static routines.
// ****************************************************************************
/**
 * Calculate the CRC-16 checksum of the block.
 * @param wMaxBytes - a limit to the size of the output (-1 = all buffer)
 * @return Types::uint - checksum value
 */
Types::uint MemBuf::Checksum(int wMaxBytes) {
  int wNumBytes = (wMaxBytes == -1 ? wMax : StandardMin((int)wMax,wMaxBytes));

  return ::_Memory_Checksum(0, pBuf, wNumBytes, true);
}


// ****************************************************************************
//  Class: MemBlock
// ****************************************************************************

char *MemBlock::New(Types::uint wSize) {
  TestDelete(pBuf);
  wMax = wSize;
  pBuf = new char[wMax];
  return pBuf;
}

void MemBlock::Delete(void) {
  TestDeleteNull(pBuf);
  wMax = 0;
  pTmp = NULL;
}

char *MemBlock::Append(char *pSrc, Types::uint wSize) {
  if (!pTmp) pTmp = pBuf;
  if ((pTmp - pBuf) + wSize > wMax) {
    cerr << ErrorHeader(wWARNING)
	 << "MemBlock::Append(): Cannot write " << wSize
	 << " bytes at offset " << (pTmp-pBuf) << " of a "
	 << wMax << " byte block.\n";
    return NULL;  // does not fit
  }
  memcpy(pTmp, pSrc, wSize);
  pTmp += wSize;
  return pTmp;
}

MemBlock &MemBlock::operator=(const MemBlock &mb) {
  Delete();
  pBuf = mb.pBuf;
  wMax = mb.wMax;
  pTmp = mb.pTmp;
  return *this;
}


// ****************************************************************************
//  Class: MemList
// ****************************************************************************

// ****************************************************************************
//  Name: Append
//  Date: 06/18/1997 : mdb - Created.
// ****************************************************************************
/**
 * Add a new MemBuf entry to end of list.
 * @param pSrc, wSize - ptr to create MemBuf
 * @param wSize - length to create MemBuf
 * @return char * - same as pSrc
 */
char *MemList::Append(char *pSrc, Types::uint wSize) {
  MemBuf mbTmp(pSrc, wSize);
  list.InsTail(mbTmp);
  wMax += wSize;
  return pSrc;
}

void MemList::Dump(int wMaxBytes) {
  MemBuf *pmb;
  cout << "MemList::Dump(): getNum=" << list.getNum() << endl;
  for (int i=list.StartIter(); i; i--) {
    pmb = list.pNext();
    pmb->Dump(wMaxBytes);
  }
}


// ****************************************************************************
//  Class: MemoryBuf - redone implementation I needed for appstreamd/packet.cpp
// ****************************************************************************


// 11/29/2001 : mdb - Manually inlined and minimized this common call for perf
char *MemoryBuf::New(Types::uint wMax_in) {
  if (pBuf && wMax > 0 && fFree) {
    delete [] pBuf; // assume array
  }
  if (wMax_in > 0) {
    pBuf=new char[wMax = wMax_in];
    wSize=wSizeExtract=0;
    fFree=true;
  } else {
    pBuf=NULL;
    wMax=wSize=wSizeExtract=0;
  }  
  return pBuf;
}

void MemoryBuf::Delete(void) {
  if (pBuf && wMax > 0 && fFree) {
    delete [] pBuf; // assume array
    pBuf=NULL;
  }
  wMax=wSize=wSizeExtract=0;
}

void MemoryBuf::Set(const char *pBuf_in, int wMax_in, int wSize_in,
		    bool fFree_in) {
  if (pBuf && wMax > 0 && fFree) {
    delete [] pBuf; // assume array
  }
  if (wMax_in > 0) {
    pBuf=const_cast(char*,pBuf_in); wMax=wMax_in; wSize=wSize_in;
    fFree=fFree_in;
  } else {
    pBuf=NULL;
    wMax=wSize=wSizeExtract=0;
  }
  wSizeExtract=0;
}

/**
 * does a shallow copy
 */
int MemoryBuf::ShallowCopy(const MemoryBuf &rhs) {
  if (&rhs != this) {
    Delete();
    pBuf = rhs.pBuf;
    wMax = rhs.wMax;
    wSize = rhs.wSize;
    fFree = false;
    wSizeExtract = rhs.wSizeExtract;   // unclear what is the right semantics
  }
  return 0;
}

/**
 * does a shallow copy
 */
const MemoryBuf &MemoryBuf::operator=(const MemoryBuf &rhs) {
  ShallowCopy(rhs);
  return *this;
}

/**
 * does a deep copy
 */
int MemoryBuf::DeepCopy(const MemoryBuf &rhs) {
  if (&rhs != this) {
    //New(rhs.wMax);   // BUGBUG: just big enough, or include free space?
    New(rhs.wSize);
    if (rhs.wSize > 0) {
      memcpy(pBuf, rhs.pBuf, rhs.wSize);  // leave free space uninit?
      wSize = rhs.wSize;
      wSizeExtract = rhs.wSizeExtract;   // unclear what is the right semantics
    }
  }
  return 0;
}

/**
 * we shallow copy the buffer, and remove it from the rhs
 */
int MemoryBuf::StealCopy(MemoryBuf &rhs) {
  if (&rhs != this) {
    Delete();
    // copy all fields
    pBuf = rhs.pBuf;
    wMax = rhs.wMax;
    wSize = rhs.wSize;
    fFree = rhs.fFree;
    wSizeExtract = rhs.wSizeExtract;
    // remove from rhs
    rhs.pBuf = NULL;
    rhs.wMax = 0;
    rhs.wSize = 0;
    rhs.fFree = false;
    rhs.wSizeExtract = 0;
    }
  return 0;
}


// ****************************************************************************
//  Name: Append
//  Date: 08/12/1999 : mdb - Created.
// ****************************************************************************
/**
 * Appends the given buffer to the end of the current buffer.
 * @param pAdd - pointer of buffer to use
 * @param wSizeAdd - length of buffer to use
 * @return char * - ptr after added region
 */
char *MemoryBuf::Append(char *pAdd, Types::uint wSizeAdd) {
   if (wSize + wSizeAdd > wMax) {
      if (pBuf && wMax > 0 && fFree) {
          // lets realloc it
         Types::uint newMax = wMax + (wSizeAdd > KB_1) ? wSizeAdd : KB_1;
         char *newpBuf = new char[newMax];
         memcpy(newpBuf, pBuf, wSize);
         delete [] pBuf;
         pBuf = newpBuf;
         wMax = newMax;
      }
      else {
         cerr << ErrorHeader(wWARNING)
              << "MemoryBuf::Append(): Cannot write " << wSizeAdd
              << " bytes at offset " << wSize << " of a "
              << wMax << " byte block.\n";
         return NULL;     // does not fit
      }
   }
   memcpy(pBuf+wSize, pAdd, wSizeAdd);
   wSize += wSizeAdd;
   return (pBuf+wSize);
}


// ****************************************************************************
//  Name: sprintf
//  Date: 08/12/1999 : mdb - Created.
//        10/18/2000 : mdb - Now use snprintf to avoid buffer overwrites.
//        09/27/2001 : mdb - Adds an implicit \0 null byte if there is room.
//  Note: Embedding a \0 in the format string will mess up the return value of
//        vsprintf, and also may not write the null byte.
// ****************************************************************************
/**
 * Does a std sprintf to the current buffer w/o checking for overflow.
 * @param sbFormat - standard printf arg 
 * @param args - standard printf arg
 * @return int - number of characters printed
 */
int MemoryBuf::sprintf(char *sbFormat, ...) {
  va_list ap;
  va_start(ap, sbFormat);
  int nBytes = vsnprintf(pBuf+wSize, wMax-wSize, sbFormat, ap);
  if (nBytes > 0) {
    setSize(getSize()+nBytes);
  }
  va_end(ap);
  return nBytes;
}

#if 1  /* ***DEBUG*** */
#include <signal.h>
#endif

// ****************************************************************************
//  Name: Extract
//  Date: 08/27/1999 : mdb - Created.
// ****************************************************************************
/**
 * Copies from the read ptr into the given ptr, and advances the
 * read ptr.
 * @param pBuf_in - ptr to buffer to write to (NULL = don't write)
 * @param wSize_in - length of buffer to write to
 * @return int - number of bytes extracted
 */
int MemoryBuf::Extract(char *pBuf_in, Types::uint wSize_in) {
  if (wSizeExtract + wSize_in > wSize) {
    cerr << ErrorHeader(wWARNING)
	 << "MemoryBuf::Extract(): Cannot read " << wSize_in
	 << " bytes at offset " << wSizeExtract << " of a "
	 << wSize << " byte block.\n";
#if 1
    raise(SIGSEGV);
#endif
    return 0;
  }
  if (pBuf_in) memcpy(pBuf_in, pBuf+wSizeExtract, wSize_in);
  wSizeExtract += wSize_in;
  return wSize_in;
}


// ****************************************************************************
//  Name: ExtractAppend
//  Date: 01/05/2000 : mdb - Created.
// ****************************************************************************
/**
 * This performs an extract from one buffer and an append into another
 * in a single operation (copy).
 */
int MemoryBuf::ExtractAppend(MemoryBuf &bufAppendTo, Types::uint wSize_in) {
  if (wSizeExtract + wSize_in > wSize) {
    cerr << ErrorHeader(wWARNING)
	 << "MemoryBuf::ExtractAppend(): Cannot read " << wSize_in
	 << " bytes at offset " << wSizeExtract << " of a "
	 << wSize << " byte block.\n";
    return 0;
  }
  bufAppendTo.Append(pBuf+wSizeExtract, wSize_in);
  wSizeExtract += wSize_in;
  return wSize_in;
}


// ****************************************************************************
//  Name: readMax
//  Date: 01/05/2000 : mdb - Created.
// ****************************************************************************
/** 
 * Attempts to read up to the max size of the buffer, and adjusts the
 * size according to what was actually read.  This is intended for
 * reading from a file in fixed size chunks until the last one.
 * Note this overwrites the previous contents.
 */
int MemoryBuf::readMax(int fd) {
  int nBytes;
  if (!getPtr()) return 0;
  nBytes = ::read(fd, getPtr(), getMax());
  if (nBytes >= 0) {
    setSize(nBytes);
  } else {
    setSize(0);
  }
  return nBytes;
}


// ****************************************************************************
//  Name: Dump
//  Date: 08/12/1999 : mdb - Created.
// ****************************************************************************
/**
 * Output the block as hex integers.
 */
void MemoryBuf::Dump(int wMaxBytes) {
  int wNumBytes = ((wMaxBytes == -1 ||
		    wMaxBytes > static_cast(int,getSize())) ?
		   getSize() : wMaxBytes);

  printf("MemoryBuf::Dump(): addr=%x max=%d size=%d checksum=%x\n",
	 (Types::uint)pBuf, wMax, getSize(), Checksum());
  ::_Memory_Dump(pBuf, wNumBytes);
  if (wNumBytes < static_cast(int,getSize()))
    cout << "(truncated MemoryBuf dump)" << endl;
}

ostream &operator<<(ostream &os, MemoryBuf &buf) {
  os << "MemoryBuf{addr=" << std::hex << (long)(buf.pBuf) << std::dec
     << ",size=" << buf.getSize()
     << ",max=" << buf.getMax()
     << ",fFree=" << buf.fFree
     << "}";
  return os;
}


// ****************************************************************************
//  Name: Checksum
//  Date: 08/12/1999 : mdb - Created.
// ****************************************************************************
/**
 * Calculate the CRC-16 checksum of the memory block.
 * @return Types::uint - checksum value
 */
Types::uint MemoryBuf::Checksum(int wMaxBytes) {
  int wNumBytes = ((wMaxBytes == -1 ||
		    wMaxBytes > static_cast(int,getSize())) ?
		   getSize() : wMaxBytes);
  return ::_Memory_Checksum(0, pBuf, wNumBytes, true);
}

/**
 * these are intended to be used to compute a checksum across multiple
 * blocks such as those seen for reading in a file in blocks.
 */
Types::uint MemoryBuf::ChecksumBlockInit(void) {
  return 0;
}

Types::uint MemoryBuf::ChecksumBlock(unsigned int wSum, int wMaxBytes) {
  int wNumBytes = ((wMaxBytes == -1 ||
		    wMaxBytes > static_cast(int,getSize())) ?
		   getSize() : wMaxBytes);
  return ::_Memory_Checksum(wSum, pBuf, wNumBytes, false);
}

Types::uint MemoryBuf::ChecksumBlockDone(unsigned int wSum) {
  return ::_Memory_Checksum(wSum, NULL, 0, true);
}


// ****************************************************************************
