// -*- c++ -*-
// ****************************************************************************
//  Name: dc_memory.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 06/03/1997 : mdb - Created.
//        06/25/1997 : mdb - Moved all non-inlines to memory.cc
//        03/31/1998 : mdb - Changed semantics of MemBlock to keep the append
//                           ptr and membuf ptr seperate.
//        09/15/1999 : mdb - Added StealCopy and ShallowCopy fcns to MemoryBuf.
//        02/18/2001 : mdb - Added more MemoryBuf::Extract() methods, and fixed
//                           odd behavior of MemoryBuf::setSize(0).
//        03/31/2001 : mdb - Moved many fcns into .cpp to avoid over inlining.
//  Desc: Simple memory classes.
// ****************************************************************************
/** @file dc_memory.h */

#if !defined(H_MEMORY)
#define H_MEMORY

#ifdef _WIN32
#include <io.h>
#else
#include <unistd.h>
#endif

#include "dc_standard.h"
#include "dc_lists.h"
#include "dc_str.h"


/// Simple memory class.
/**
 *  no memory allocated - just holds ptr and length
 */
class MemBuf {
protected:
  char *pBuf;
  Types::uint wMax;

public:
  MemBuf(void) { pBuf = NULL; wMax = 0; };
  MemBuf(char *pBuf, int wMax) { Set(pBuf, wMax); };
  virtual ~MemBuf() {};  // this is because other fcns are virtual
  
  void Set(char *pBuf_in, int wMax_in) {
    pBuf = pBuf_in; wMax = wMax_in;
  };
  
  void Dump(int wMaxBytes = -1);
  Types::uint Checksum(int wMaxBytes = -1);

  char *getPtr(void) { return pBuf; };
  virtual Types::uint getSize(void) { return wMax; };

  const MemBuf &operator=(const MemBuf &rhs) {
    if (&rhs != this) {
      pBuf = rhs.pBuf;
      wMax = rhs.wMax;
    }
    return *this;
  };

  int read(int fd) {
    return (getPtr() ? ::read(fd, getPtr(), getSize()) : 0);
  };
  int write(int fd) { 
    return (getPtr() ? ::write(fd, getPtr(), getSize()) : 0);
  };
};

/// Simple memory class.
/**
 * allocs memory, and copies to end of used portion of buffer
 */
class MemBlock : public MemBuf {
protected:
  char *pTmp;  // append ptr, NULL if no appends done

public:
  MemBlock(void) { pTmp = NULL; };
  MemBlock(Types::uint wSize) { pTmp = NULL; New(wSize); };
  ~MemBlock(void) { TestDeleteNull(pBuf); };

  char *New(Types::uint wSize);
  void Delete(void);

  // rtn full buffer size, unless we have been appending
  Types::uint getSize(void) { return pTmp ? (Types::uint)(pTmp-getPtr()) : wMax; };

  char *Append(char *pSrc, Types::uint wSize);
  char *Append(short wSrc) {
    return Append((char*)&wSrc, sizeof(short)); 
  };
  char *Append(int wSrc) {
    return Append((char*)&wSrc, sizeof(int));
  };
  char *Append(long wSrc) {
    return Append((char*)&wSrc, sizeof(long));
  };
  char *Append(Str str) {
    return Append(str.Get(), str.Len());
  };

  MemBlock &operator=(const MemBlock &mb);
};

/// Simple memory class.
/**
 * avoids copies, and keeps a list of memory buffers
 */
//  ***MDB*** - may want reimplement w/darray to avoid overuse of new/delete.
class MemList {
protected:
  SLList<MemBuf> list;
  Types::uint wMax;

public:
  MemList() { wMax = 0; };

  char *Append(short &wSrc) {
    return Append((char*)&wSrc, sizeof(short));
  };
  char *Append(int &wSrc) {
    return Append((char*)&wSrc, sizeof(int));
  };
  char *Append(long &wSrc) {
    return Append((char*)&wSrc, sizeof(long));
  };
  char *Append(char *pSrc, Types::uint wSize);

  // just calls the list methods
  int getNum(void)    { return list.getNum(); };
  int StartIter(void) { return list.StartIter(); };
  MemBuf *pNext(void) { return list.pNext(); };
  void Flush(void)    { list.Flush(); };

  void Dump(int wMaxBytes = -1);
  Types::uint getSize(void) { return wMax; };
};


// ****************************************************************************
//  MemoryBuf is a rewritten implementation I needed for DataCutter/packet.cpp
// ****************************************************************************

/// Simple memory class.
/**
 *  Provides a max size buffer, which can be appended to, and keeps track of
 *  the used size.  At deletion time, the buffer is freed only if set to be.
 */
class MemoryBuf {
  friend ostream &operator<<(ostream &os, MemoryBuf &buf);
protected:
  char *pBuf;   // ptr to start of buffer
  Types::uint wMax;   // max size of buffer
  Types::uint wSize;  // size of the "used" prefix in buffer
  bool  fFree;  // controls if the buffer is ever freed
  Types::uint wSizeExtract;


public:
  MemoryBuf(void)
    : pBuf(NULL), wMax(0), wSize(0), fFree(false), wSizeExtract(0) {};
  MemoryBuf(int wMax) { pBuf=NULL; New(wMax); };
  MemoryBuf(const char *pBuf_in, int wMax, int wSize, bool fFree=false) {
    pBuf=NULL;
    Set(pBuf_in, wMax, wSize, fFree);
  };
  ~MemoryBuf() { Delete(); };
  
  char *New(Types::uint wMax_in);
  void Delete(void);
  void Set(const char *pBuf_in, int wMax_in, int wSize_in, bool fFree_in);

  void Empty(void) { wSize=0; wSizeExtract=0; };
  
  char *getPtr(void) { return pBuf; };            // to beginning
  Types::uint getSize(void) { return wSize; };    // amount used at beg
  Types::uint getMax(void) { return wMax; };      // total size of buf
  Types::uint getFree(void) { return wMax-wSize; }; // free space at end
  char *getPtrFree(void) { return pBuf+wSize; };  // to free space

  char *getPtrExtract(void) { return pBuf+wSizeExtract; };
  Types::uint getSizeExtract(void) { return wSize-wSizeExtract; }; // amt left
  Types::uint resetExtract(Types::uint w=0) { return wSizeExtract=w; };

  // 02/18/2001 : mdb - Allow setting size to 0 (as in Empty() call)
  Types::uint setSize(Types::uint wSize_in) {
    if (wSize_in >= 0 && wSize_in <= wMax) wSize=wSize_in;
    return wSize;
  };

  bool getFreeFlag(void) { return fFree; };
  void setFreeFlag(bool fFree_in) { fFree=fFree_in; };

  // does a shallow copy (same buffer as rhs)
  int ShallowCopy(const MemoryBuf &rhs);
  const MemoryBuf &operator=(const MemoryBuf &rhs);

  // does a deep copy (dup buffer from rhs)
  int DeepCopy(const MemoryBuf &rhs);

  // we shallow copy the buffer, and remove it from the rhs
  int StealCopy(MemoryBuf &rhs);

  // 02/18/2001 : mdb - Added more complete set of Append/Extract prototypes
  char *Append(char *pAdd, Types::uint wSizeAdd);
  char *Append(char ch) { return Append(&ch, sizeof(char)); };
  char *Append(short w) { return Append((char*)&w, sizeof(short));  };
  char *Append(int w)   { return Append((char*)&w, sizeof(int)); };
  char *Append(long w)  { return Append((char*)&w, sizeof(long)); };
  char *Append(float r) { return Append((char*)&r, sizeof(float)); };
  char *Append(double r) { return Append((char*)&r, sizeof(double)); };
  char *Append(Str &str) { return Append(str.Get(), str.Len()); };
  char *Append(MemoryBuf &mb) {
    return Append(mb.getPtr(), mb.getSize());
  };
  char *Append(MemoryBuf *pmb) {
    return Append(pmb->getPtr(), pmb->getSize());
  };

  int Extract(char *pBuf_in, Types::uint wSize_in);
  int Extract(char *pch) { return Extract((char*)pch, sizeof(char)); };
  int Extract(short *pw) { return Extract((char*)pw, sizeof(short)); };
  int Extract(int *pw)   { return Extract((char*)pw, sizeof(int)); };
  int Extract(long *pw)  { return Extract((char*)pw, sizeof(long)); };
  int Extract(float *pr) { return Extract((char*)pr, sizeof(float)); };
  int Extract(double *pr) { return Extract((char*)pr, sizeof(double)); };

  // extract out of the current buf object, and append to the given buf object
  int ExtractAppend(MemoryBuf &bufAppendTo, Types::uint wSize_in);

  int read(int fd) {
    return (getPtr() ? ::read(fd, getPtr(), getSize()) : 0);
  };
  int readMax(int fd);
  int write(int fd) { 
    return (getPtr() ? ::write(fd, getPtr(), getSize()) : 0);
  };
  int sprintf(char *sbFormat, ...);

  void Dump(int wMaxBytes = -1);
  Types::uint Checksum(int wMaxBytes = -1);

  Types::uint ChecksumBlockInit(void);
  Types::uint ChecksumBlock(Types::uint wSum, int wMaxBytes = -1);
  Types::uint ChecksumBlockDone(Types::uint wSum);
};


#endif /* H_MEMORY */
// ****************************************************************************
