// ****************************************************************************
//  Name: dc_mutex.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 11/20/2000 : mdb - Created.
//        02/27/2001 : mdb - Added debugging code to track down usage bugs.
//        03/20/2001 : mdb - Fixed ret==-1 -> errno strerror reporting.
//        04/13/2001 : mdb - Changed debugging mutex calls to better handle
//                           cond implicit lock and unlock.  Realized debug
//                           version needs an internal mux to protect debug
//                           state, but this is hard to accomplish without
//                           a complex multiple lock backoff scheme.  Todo:
//                           only need one mux for debugging version + bool
//                           state for locked/unlocked (dumb).
//        04/23/2001 : mdb - Changed to use dc_pthread-wrappers.h
//  Desc: Simple synchronization wrappers.
// ****************************************************************************

#include <iostream>

#include <errno.h>
#include <string.h>
#include <signal.h>

#include "dc_mutex.h"
#include "dc_pthread-wrappers.h"
#include "dc_standard.h"


#if defined(LINUX)
#define DCDBG_USE_LINUX_DEBUG_MUTEX
#if defined(DCDBG_USE_LINUX_DEBUG_MUTEX)
//#define DCDBG_USE_LINUX_DEBUG_MUTEX_setkind
//#define DCDBG_USE_LINUX_DEBUG_MUTEX_settype
#endif
#endif /* LINUX */

#if defined(DCDBG)
//#define DCDBG_MUTEX_LOCKUNLOCK
#endif


//
//  class DC_Mutex
//

// ctor/dtor
DC_Mutex::DC_Mutex(char *sbName_in, pthread_mutexattr_t *pattr) {
  int ret;
#if defined(DCDBG_USE_MUTEX_CHECKS)
  char sbTmp[40];
  snprintf(sbTmp, 40, "%s(%lx)", (sbName_in ? sbName_in : "DC_Mutex"),
	   (unsigned long)this);
  sbName = StrDup(sbTmp);
  ret = Pthread_mutex_init(&mutexDebug, NULL);
  ret = Pthread_cond_init(&cvDebug, NULL);
  fLocked = false;
#endif /* DCDBG_USE_MUTEX_CHECKS */

#if defined(DCDBG_USE_LINUX_DEBUG_MUTEX)
  pthread_mutexattr_t attr;
  if (pattr == NULL) {
    ret = pthread_mutexattr_init(&attr);
    if (ret != 0) {
      cerr << "DC_Mutex(): failed pthread_mutexattr_init - " << endl;
      // CLH - not MT safe - << (ret != -1 ? strerror(ret) : strerror(errno)) << endl;
    }
    pattr = &attr;
  }
#if defined(DCDBG_USE_LINUX_DEBUG_MUTEX_setkind)
  ret = pthread_mutexattr_setkind_np(pattr, PTHREAD_MUTEX_ERRORCHECK_NP);
  if (ret != 0) {
    cerr << "DC_Mutex(): failed pthread_mutexattr_setkind_np - " << endl;
    // CLH - not MT safe - << (ret != -1 ? strerror(ret) : strerror(errno)) << endl;
  }
#elif defined(DCDBG_USE_LINUX_DEBUG_MUTEX_settype)
  ret = pthread_mutexattr_settype(pattr, PTHREAD_MUTEX_ERRORCHECK_NP);
  if (ret != 0) {
    cerr << "DC_Mutex(): failed pthread_mutexattr_settype - " << endl;
    // CLH - not MT safe - << (ret != -1 ? strerror(ret) : strerror(errno)) << endl;
  }
#endif
#endif /* DCDBG_USE_LINUX_DEBUG_MUTEX */

  ret = Pthread_mutex_init(&mutex, pattr);
  if (ret != 0) {
    cerr << "DC_Mutex(): failed pthread_mutex_init - " << endl;
    // CLH - not MT safe - << (ret != -1 ? strerror(ret) : strerror(errno)) << endl;
  }
}

DC_Mutex::~DC_Mutex(void) {
#if defined(DCDBG_USE_MUTEX_CHECKS)
  pthread_mutex_lock(&mutexDebug);
  if (fLocked) {
    cerr << ">>> " << sbName
	 << ".~DC_Mutex(): mutex is locked by [t" << (long)tidOwner << ']'
	 << endl;
  }
  TestDelete(sbName);
  pthread_mutex_unlock(&mutexDebug);
  pthread_cond_destroy(&cvDebug);
#endif /* DCDBG_USE_MUTEX_CHECKS */
  int ret = pthread_mutex_destroy(&mutex);
  if (ret != 0) {
    cerr << "~DC_Mutex(): failed pthread_mutex_destroy - " << endl;
    // CLH - not MT safe - << (ret != -1 ? strerror(ret) : strerror(errno)) << endl;
  }
}

void DC_Mutex::lock(void) {
#if defined(DCDBG_USE_MUTEX_CHECKS)
  pthread_mutex_lock(&mutexDebug);
  if (fLocked && pthread_equal(tidOwner, pthread_self())) {
    cerr << ">>> [t" << (long)pthread_self() << ']' << sbName
	 << ".lock(): current thread already lock owner." << endl;
    pthread_mutex_unlock(&mutexDebug);
    raise(SIGSEGV);
    return;
  }
  while (fLocked) {  // sleep waiting until we can get the lock
    pthread_cond_wait(&cvDebug, &mutexDebug);
  }
  fLocked = true;
  tidOwner = pthread_self();
  int ret = pthread_mutex_lock(&mutex);
  if (ret != 0) {
    cerr << "DC_Mutex::lock() failed pthread_mutex_lock - " << endl;
    // CLH - not MT safe - << (ret != -1 ? strerror(ret) : strerror(errno)) << endl;
    raise(SIGSEGV);
  }
#if defined(DCDBG_MUTEX_LOCKUNLOCK)
  cerr << ">>> [t" << (long)pthread_self() << ']' << sbName
       << ".lock(): ok." << endl;
#endif /* DCDBG_MUTEX_LOCKUNLOCK */
  pthread_mutex_unlock(&mutexDebug);
#else /* DCDBG_USE_MUTEX_CHECKS */
  int ret = pthread_mutex_lock(&mutex);
  if (ret != 0) {
    cerr << "DC_Mutex::lock() failed pthread_mutex_lock - " << endl;
    // CLH - not MT safe - << (ret != -1 ? strerror(ret) : strerror(errno)) << endl;
    raise(SIGSEGV);
  }
#endif /* DCDBG_USE_MUTEX_CHECKS */
}

void DC_Mutex::unlock(void) {
#if defined(DCDBG_USE_MUTEX_CHECKS)
  pthread_mutex_lock(&mutexDebug);
  if (!fLocked) {
    cerr << ">>> [t" << (long)pthread_self() << ']' << sbName
	 << ".unlock(): mutex is not locked." << endl;
    pthread_mutex_unlock(&mutexDebug);
    raise(SIGSEGV);
    return;
  } else if (!pthread_equal(tidOwner, pthread_self())) {
    cerr << ">>> [t" << (long)pthread_self() << ']' << sbName
	 << ".unlock(): current thread not lock owner." << endl;
    pthread_mutex_unlock(&mutexDebug);
    raise(SIGSEGV);
    return;
  }
  fLocked = false;
  int ret = pthread_mutex_unlock(&mutex);
  if (ret != 0) {
    cerr << "DC_Mutex::unlock() failed pthread_mutex_unlock - " << endl;
    // CLH - not MT safe - << (ret != -1 ? strerror(ret) : strerror(errno)) << endl;
    raise(SIGSEGV);
    return;
  }  
#if defined(DCDBG_MUTEX_LOCKUNLOCK)
  cerr << ">>> [t" << (long)pthread_self() << ']' << sbName
       << ".unlock(): ok." << endl;
#endif /* DCDBG_MUTEX_LOCKUNLOCK */
  pthread_mutex_unlock(&mutexDebug);
  pthread_cond_signal(&cvDebug);
#else /* DCDBG_USE_MUTEX_CHECKS */
  int ret = pthread_mutex_unlock(&mutex);
  if (ret != 0) {
    cerr << "DC_Mutex::unlock() failed pthread_mutex_unlock - " << endl;
    // CLH - not MT safe - << (ret != -1 ? strerror(ret) : strerror(errno)) << endl;
    raise(SIGSEGV);
    return;
  }  
#endif /* DCDBG_USE_MUTEX_CHECKS */
}

// return T if it was locked
bool DC_Mutex::trylock(void) {
#if defined(DCDBG_USE_MUTEX_CHECKS)
  pthread_mutex_lock(&mutexDebug);
  if (fLocked) {
    if (pthread_equal(tidOwner, pthread_self())) {
      cerr << ">>> [t" << (long)pthread_self() << ']' << sbName
	   << ".trylock(): current thread already lock owner." << endl;
      pthread_mutex_unlock(&mutexDebug);
      raise(SIGSEGV);
      return true;
    } else {
      pthread_mutex_unlock(&mutexDebug);
      return false;
    }
  }
  fLocked = true;
  tidOwner = pthread_self();
  int ret = pthread_mutex_trylock(&mutex);
  if (ret == EBUSY) return false;
  if (ret != 0) {
    cerr << "DC_Mutex::trylock() failed pthread_mutex_trylock - " << endl;
    // CLH - not MT safe - << (ret != -1 ? strerror(ret) : strerror(errno)) << endl;
    return false;
  }
#if defined(DCDBG_MUTEX_LOCKUNLOCK)
  cerr << ">>> [t" << (long)pthread_self() << ']' << sbName
       << ".trylock(): ok." << endl;
#endif /* DCDBG_MUTEX_LOCKUNLOCK */
  pthread_mutex_unlock(&mutexDebug);
#else /* DCDBG_USE_MUTEX_CHECKS */
  int ret = pthread_mutex_trylock(&mutex);
  if (ret == EBUSY) return false;
  if (ret != 0) {
    cerr << "DC_Mutex::trylock() failed pthread_mutex_trylock - " << endl;
    // CLH - not MT safe - << (ret != -1 ? strerror(ret) : strerror(errno)) << endl;
    return false;
  }
#endif /* DCDBG_USE_MUTEX_CHECKS */
  return true;
}

#if defined(DCDBG_USE_MUTEX_CHECKS)
// assume: locked
// called for cases where mutex ownership changes w/o calling lock/unlock
void DC_Mutex::implicitUnlock(void) {
  pthread_mutex_lock(&mutexDebug);
  fLocked = false;
#if defined(DCDBG_MUTEX_LOCKUNLOCK)
  cerr << ">>> [t" << (long)pthread_self() << ']' << sbName
       << ".implicitUnlock(): signaling waiters." << endl;
#endif /* DCDBG_MUTEX_LOCKUNLOCK */
  pthread_mutex_unlock(&mutexDebug);
  pthread_cond_signal(&cvDebug);
  // external mutex still locked, which cond_wait will unlock
}

// assume: unlocked
// called for cases where mutex ownership changes w/o calling lock/unlock
void DC_Mutex::implicitLock(void) {
  // external mutex was re-locked by cond_wait
  pthread_mutex_unlock(&mutex);  // avoid deadlock
  lock();
}
#endif /* DCDBG_USE_MUTEX_CHECKS */  


// ****************************************************************************
