// ****************************************************************************
//  Name: dc_queue.cpp
//  Athr: Michael D. Beynon (mdb) - beynon@cs.umd.edu
//  Date: 01/28/1999 : mdb - Created.
//        09/10/1999 : mdb - Added external mutex/cv.
//        10/23/1999 : mdb - Changes for DCE use.
//        01/06/2000 : mdb - Added linux port of pthread_get_expiration_np().
//        04/07/2000 : mdb - Restructured to allow for external lock/unlock
//                           of the internal mutex.
//                         - Changed many fcn signatures to return status to
//                           catch new shutdown() call.  This allows for
//                           caller signaling to producer/consumer.
//        09/15/2000 : mdb - Rewitten with less buggy version of multiple
//                           queue external lock/cv.
//        09/28/2000 : mdb - Fixed small leak of pmutex in certain cases.
//        11/16/2000 : mdb - Fixed bug where shutdown() used signal instead of
//                           broadcast to wake up cv waiters.
//        04/18/2001 : mdb - Removed mod operations and replaced with ifs.
//        04/23/2001 : mdb - Added template to have Q of structs (not ptrs).
//  Desc: pthread-safe concurrent queue implementation.
// ****************************************************************************

#include "dc_queue.h"
#include "dc_standard.h"
#include "dc_pthread-wrappers.h"

#include <errno.h>

#ifdef _WIN32
#include "win32.h"
#endif
	
#if defined(DCDBG_QUEUE)
#ifdef _WIN32
#include <iostream>
#else
#include <iostream.h>
#endif
#endif /* DCDBG_QUEUE */


// *******************************************************************
//  Note: This template definition .cpp file is included for
//  instantiation in dc_template-instantiation.cpp, or is included in
//  the header file for compilers that do not support it.
// *******************************************************************


//
//  DC_QueueT<>
//

template <class T>
DC_QueueT<T>::DC_QueueT(void)
  : sbName(NULL),  buf(NULL), fInternalMutex(true), pmutex(NULL) { max = 0;
}

// ctor
/*template <class T>
DC_QueueT<T>::DC_QueueT(void)
  : sbName(NULL), max(0), buf(NULL), fInternalMutex(true), pmutex(NULL) {
};*/

// dtor
template <class T>
DC_QueueT<T>::~DC_QueueT(void) {
  _delete();
};

template <class T>
void DC_QueueT<T>::_delete(void) {
  if (sbName) delete [] sbName;
  if (buf) {
    delete [] buf; buf = NULL;
  }
  if (fInternalMutex && pmutex) {
    delete pmutex;
  }
  // we do not destroy external mutex
}

/**
 * specify queue size to init storage.  optionally replace internal
 * mutex with an external one, and add a global cvNotEmpty, which are
 * neither initialized nor destroyed by this queue.
 * (can be called repeatedly -- undefined if called when a thread is blocked)
 */
template <class T>
void DC_QueueT<T>::init(char *sbName_in, int maxItems,
			DC_Mutex *pmutex_in,
			DC_Cond *pcvExternalNotEmpty_in,
			DC_Cond *pcvExternalEmpty_in) {
  _delete();  // cleanup previous state (if this is a repeated call)

  sbName = StrDup(sbName_in ? sbName_in : "DC_QueueT");

  // init new state
  if (maxItems <= 0) maxItems = 10;  // just fix a bad arg
  max = maxItems;
  buf = new T[max];

  iHead = 0;
  iTail = max - 1;
  num = 0;

  // init sync
  if (pmutex_in) {
    fInternalMutex = false;
    pmutex = pmutex_in;
  } else {
    fInternalMutex = true;
    char *sbTmp = new char[strlen(sbName)+10];
    snprintf(sbTmp, strlen(sbName)+10, "(%s.pmutex)", sbName);
    pmutex = new DC_Mutex(sbTmp);
    delete [] sbTmp;
  }
  pcvExternalNotEmpty = pcvExternalNotEmpty_in;
  pcvExternalEmpty = pcvExternalEmpty_in;
}

/**
 * check if queue is shutdown
 */
template <class T>
bool DC_QueueT<T>::isShutdown(bool fLock) {
  bool fRtn;
  if (fLock) pmutex->lock();
  fRtn = buf ? false : true;
  if (fLock) pmutex->unlock();
  return fRtn;
}

/**
 * shut down the queue.  Calls to operate on the queue will now fail.
 */
// Care needs to be used when calling this, because user allocated queue
// entries are leaked if not handled by the caller ...
//    Q.lockInternal();
//    while (Q.probe(false) > 0) {
//      if (Q.dequeue(&p, false) > 0) *** app specific cleanup of *p ***;
//    }
//    Q.shutdown(false);
//    Q.unlockInternal();
template <class T>
int DC_QueueT<T>::shutdown(bool fLock) {
  if (fLock) pmutex->lock();
  if (!buf) {
    if (fLock) pmutex->unlock();
    return QUEUE_SHUTDOWN;
  }
#if defined(DCDBG_TEST)
  cout << "DC_QueueT::shutdown(fLock=" << fLock << ")" << endl
       << "  max=" << max << ", num=" << num
       << ", iHead=" << iHead << ", iTail=" << iTail
       << endl;
#endif
  // any entries currently in queue need to be handled by caller
  delete [] buf; buf = NULL;
  max = num = 0;
  // in case any sleeping threads
  cvNotFull.broadcast();
  _BroadcastNotEmpty();

  if (fLock) pmutex->unlock();
  return QUEUE_OK;
}


// optional internal mutex interface
template <class T>
DC_Mutex *DC_QueueT<T>::getInternal(void) {
  return pmutex;
}

template <class T>
int DC_QueueT<T>::lockInternal(void) {
  pmutex->lock();
  return QUEUE_OK;
}

template <class T>
int DC_QueueT<T>::unlockInternal(void) {
  pmutex->unlock();
  return QUEUE_OK;
}

/**
 * signal waiters there is data to read
 */
template <class T>
void DC_QueueT<T>::_SignalNotEmpty(void) {
  cvNotEmpty.signal();
  if (pcvExternalNotEmpty) pcvExternalNotEmpty->signal();
  return;
}

template <class T>
void DC_QueueT<T>::_BroadcastNotEmpty(void) {
  cvNotEmpty.broadcast();
  if (pcvExternalNotEmpty) pcvExternalNotEmpty->broadcast();
  return;
}

/**
 * add item to the end of queue, block if the queue is full
 */
template <class T>
int DC_QueueT<T>::enqueue(T *item, bool fLock) {
  if (fLock) pmutex->lock();

  if (!buf) {
    if (fLock) pmutex->unlock();
    return QUEUE_SHUTDOWN;
  }
  while (num == max) {
#if defined(DCDBG_QUEUE)
    cout << "DC_QueueT " << hex << (int)this << dec << ": "
	 << "full with " << num << " entries ... cv.wait()\n";
#endif /* DCDBG_QUEUE */
    cvNotFull.wait(pmutex);
    if (!buf) {
      if (fLock) pmutex->unlock();
      return QUEUE_SHUTDOWN;
    }
  }
#if defined(DCDBG_QUEUE)
  cout << "DC_QueueT " << hex << (int)this << dec << ": "
       << "was empty ... cv.signal()\n";
#endif /* DCDBG_QUEUE */
  // add item to tail of queue
  if (++iTail == max) iTail = 0;
  buf[iTail] = *item;  // assumes T::operator=() is defined
  ++num;

  _SignalNotEmpty();
  //_BroadcastNotEmpty();

  if (fLock) pmutex->unlock();
  return QUEUE_OK;
}

/**
 * remove an item from queue.  If timeout<0 an empty queue will
 * block, otherwise wait up to timeout usec for an item to appear
 * before giving up.
 */
template <class T>
int DC_QueueT<T>::dequeue(T *item, int timeout, bool fLock) {
  bool fTimeout = false;
  struct timespec delay, abstime;

  if (timeout >= 0) {
    // timeout is in usec, and we want tv_sec and tv_nsec
    delay.tv_sec = timeout / 1000000;
    delay.tv_nsec = (timeout % 1000000) * 1000000000;
    pthread_get_expiration_np(&delay, &abstime);
  }

  if (fLock) pmutex->lock();

  if (!buf) {
    if (fLock) pmutex->unlock();
    return QUEUE_SHUTDOWN;
  }
  while (num == 0 && !fTimeout) {
    if (timeout < 0) {
      // BLOCK
#if defined(DCDBG_QUEUE)
      cout << "DC_QueueT " << hex << (int)this << dec << ": "
	   << "empty ... cv.wait()\n";
#endif /* DCDBG_QUEUE */
      cvNotEmpty.wait(pmutex);
    } else {
      // BLOCK with timeout
      if (cvNotEmpty.wait(pmutex, &abstime) != 0) {
	fTimeout = true;
      }
    }
    if (!buf) {
      if (fLock) pmutex->unlock();
      return QUEUE_SHUTDOWN;
    }
  }
  if (!fTimeout) {
#if defined(DCDBG_QUEUE)
    cout << "DC_QueueT " << hex << (int)this << dec << ": "
	 << "was full with " << num << " entries ... cv.signal()\n";
#endif /* DCDBG_QUEUE */
    // remove item from head of queue
    *item = buf[iHead];  // assumes T::operator=() is defined
    if (++iHead == max) iHead = 0;
    --num;

    cvNotFull.signal();
    if (num == 0 && pcvExternalEmpty) pcvExternalEmpty->signal();
    //cvNotFull.broadcast();
    //if (num == 0 && pcvExternalEmpty) pcvExternalEmpty->broadcast();
  }
  if (fLock) pmutex->unlock();
  return QUEUE_OK;
}

/**
 * return (without removing it) the first item in the queue
 */
template <class T>
int DC_QueueT<T>::head(T *item, bool fLock) {
  if (fLock) pmutex->lock();
  if (!buf) {
    if (fLock) pmutex->unlock();
    return QUEUE_SHUTDOWN;
  }
  if (num > 0) {
    *item = buf[iHead];  // assumes T::operator=() is defined
    //  } else {
    //    *item = NULL;
  }
  if (fLock) pmutex->unlock();
  return QUEUE_OK;
}


// >0 if the queue contains at least one item
template <class T>
int DC_QueueT<T>::probe(bool fLock) {
  int wRtn;
  if (fLock) pmutex->lock();
  if (!buf) {
    if (fLock) pmutex->unlock();
    return QUEUE_SHUTDOWN;
  }
  wRtn = num;
  if (fLock) pmutex->unlock();
  return wRtn;
}

/**
 * >0 if the queue contains space for at least one item
 * @return 0=full, <0=error, >0=num spaces free in queue
 */
template <class T>
int DC_QueueT<T>::probeSpace(bool fLock) {
  int wRtn;
  if (fLock) pmutex->lock();
  if (!buf) {
    if (fLock) pmutex->unlock();
    return QUEUE_SHUTDOWN;
  }
  wRtn = max-num;
  if (fLock) pmutex->unlock();
  return wRtn;
}


/**
 * @param fFull - true=full (enqueue would block), otherwise false
 * @return 0=ok, <0=error
 */
template <class T>
int DC_QueueT<T>::probeIsFull(bool &fFull, bool fLock) {
  if (fLock) pmutex->lock();
  if (!buf) {
    if (fLock) pmutex->unlock();
    return QUEUE_SHUTDOWN;
  }
  fFull = (num == max);
  if (fLock) pmutex->unlock();
  return QUEUE_OK;
}

/**
 * @param fEmpty - true=empty (dequeue would block), otherwise false
 * @return 0=ok, <0=error
 */
template <class T>
int DC_QueueT<T>::probeIsEmpty(bool &fEmpty, bool fLock) {
  if (fLock) pmutex->lock();
  if (!buf) {
    if (fLock) pmutex->unlock();
    return QUEUE_SHUTDOWN;
  }
  fEmpty = (num == 0);
  if (fLock) pmutex->unlock();
  return QUEUE_OK;
}


// ****************************************************************************
