// -*- c++ -*-
// ****************************************************************************
//  Name: dc_queue.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 01/28/1999 : mdb - Created.
//        08/21/1999 : mdb - Added constructor w/o args, and an init() fcn.
//        09/10/1999 : mdb - Added external mutex/cv.
//        04/07/2000 : mdb - Restructured to allow for external lock/unlock
//                           of the internal mutex.
//                         - Changed many fcn signatures to return status to
//                           catch new shutdown() call.  This allows for
//                           caller signaling to producer/consumer.
//        09/15/2000 : mdb - Rewitten with less buggy version of multiple
//                           queue external lock/cv.
//        11/16/2000 : mdb - Added external empty cv.
//        04/23/2001 : mdb - Added template to have Q of structs (not ptrs).
//  Desc: pthread-safe concurrent queue implementation.
// ****************************************************************************
/** @file dc_queue.h */

#if !defined(H_DC_QUEUE)
#define H_DC_QUEUE

#include "../common-config.h"

#include <stdlib.h>

#include "dc_mutex.h"
#include "dc_cond.h"


//#define DCDBG_QUEUE
#ifdef _WIN32
#include <wtypes.h>
#include <errno.h>
#include "unix_calls.h"
#endif

/// Pthread-safe concurrent queue implementation.
template <class T>
class DC_QueueT {
protected:
  char *sbName;  // symbolic name
  int   max;    // max size of queue
  int   num;    // current number of elements in queue
  int   iHead;  // next element to dequeue
  int   iTail;  // last element enqueued
  T    *buf;

  // producer-consumer sync (transparant via member fcns)
  bool      fInternalMutex;
  DC_Mutex *pmutex;
  DC_Cond   cvNotFull;
  DC_Cond   cvNotEmpty;
  DC_Cond  *pcvExternalNotEmpty;  // used for a collection of queues
  DC_Cond  *pcvExternalEmpty;  // used to signal a newly empty queue

  void _SignalNotEmpty(void);
  void _BroadcastNotEmpty(void);

public:
  enum err_val { QUEUE_OK=0, QUEUE_SHUTDOWN=-1 };

private:
  //  void _init(void);
  void _delete(void);

public:
  DC_QueueT(void);
  ~DC_QueueT(void);

  // specify queue size, init storage.  optionally replace internal
  // mutex with an external one, and add a global cvNotEmpty
  // (can be called repeatedly -- undefined if called when a thread is blocked)
  void init(char *sbName_in, int maxItems,
	    DC_Mutex *pmutex_in=NULL,
	    DC_Cond *pcvExternalNotEmpty_in=NULL,
	    DC_Cond *pcvExternalEmpty_in=NULL);

  // check if queue is shutdown
  bool isShutdown(bool fLock=true);
  // shut down the queue.  Calls to operate on the queue will now fail.
  int shutdown(bool fLock=true);

  // External access to the internal mutex, for use with fLock=false
  // calls to the member fcns.  In this case, the caller *must* have called
  // lockInternal() before any of the member fcns.
  // This is useful to avoid caller races without another level of locks
  // rtn: 0=ok
  DC_Mutex *getInternal(void);
  int lockInternal(void);
  int unlockInternal(void);

  // add item to the end of queue, block if the queue is full
  // rtn: 0=ok, <0=error
  int enqueue(T item, bool fLock=true) {
    return enqueue(&item, fLock);
  };
  // for efficiency to avoid byval copies, pass in a ptr, and enqueue *item
  int enqueue(T *item, bool fLock=true);

  // remove an item from queue.  If timeout<0 an empty queue will
  // block, otherwise wait up to timeout usec for an item to appear
  // before giving up.
  // rtn: 0=ok, <0=error
  int dequeue(T *item, int timeout=-1, bool fLock=true);

  // return (without removing it) the first item in the queue
  // rtn: 0=ok, <0=error
  int head(T *item, bool fLock=true);

  // >0 if the queue contains at least one item
  // rtn: 0=ok, <0=error, >0=num items in queue
  int probe(bool fLock=true);

  // >0 if the queue contains space for at least one item
  // rtn: 0=full, <0=error, >0=num spaces free in queue
  int probeSpace(bool fLock=true);

  // out: true=full (enqueue would block), otherwise false
  // rtn: 0=ok, <0=error
  int probeIsFull(bool &fFull, bool fLock=true);
  // out: true=empty (dequeue would block), otherwise false
  // rtn: 0=ok, <0=error
  int probeIsEmpty(bool &fEmpty, bool fLock=true);
};

#if !defined(HAVE_INSTANTIATIONS)
// for compilers that do not support explicit template instantiation,
// we need the template definition in addition to the declaration.
#include "dc_queue.cpp"
#endif /* !HAVE_INSTANTIATIONS */



// for backward compatibility with prev non-template version
typedef DC_QueueT<void*> DC_Queue;


#endif /* H_DC_QUEUE */
// ****************************************************************************
