// ****************************************************************************
//  Name: dc_random.cpp
//  Athr: Michael D. Beynon (mdb)/
//  Date: 01/28/1998 : mdb - Created from example in the [Jain] book.
//  Desc: Pseudo-random multiplicitive linear congruential generator (LCG).
// ****************************************************************************

//
//  Include Files
//
#include "dc_random.h"
#include <time.h>
#ifdef _WIN32
#include <winsock.h>
#else
#include <unistd.h>
#endif


// set of seeds that are 100k apart
static const int maxSeed = 50;
static int rgSeed[maxSeed] = {
  1,
  46831694,
  1841581359,
  1193163244,
  727633698,
  933588178,
  804159733,
  1671059989,
  1061288424,
  1961692154,
  1227283347,
  1171034773,
  276090261,
  1066728069,
  209208115,
  554590007,
  721958466,
  1371272478,
  675466456,
  1095462486,
  1808217256,
  2095021727,
  1769349045,
  904914315,
  373135028,
  717419739,
  881155353,
  1489529863,
  1521138112,
  298370230,
  1140279430,
  1335826707,
  706178559,
  110356601,
  884434366,
  962338209,
  1341315363,
  709314158,
  591449447,
  431918286,
  851767375,
  606179079,
  1500869201,
  1434868289,
  263032577,
  753643799,
  202794285,
  715851524,
  509392896,
  2116653394
};

// ctor
Random::Random(void) {
  char bufHost[128];
  wSeed = static_cast(int,time(NULL));
  if (gethostname(bufHost, 128) != -1) {
    int iMax = (int)strlen(bufHost);
    for (int i=0; i<iMax; ++i) {
      if (i%2 == 0) wSeed += static_cast(int,bufHost[i]);
      else wSeed -= static_cast(int,bufHost[i]);
    }
  }
  cBitsLeft = 0; fConserveBits = true; _Init();
};

// ctor
Random::Random(int wSeed_in) : wSeed(wSeed_in) { 
  cBitsLeft = 0; fConserveBits = true; _Init();
};

int Random::GetSeed(int i) {
  return rgSeed[i % maxSeed];
}


// ****************************************************************************
//  Name: Random::GetRand
//  Date: 01/28/1998 : mdb - Created based on [Jain, p443].
//  Note: Requires ints with at least 2^31 - 1 maxint value.
//        The computation of ax mod m (due to Schrange(1979)) is solved as:
//          ax mod m = g(x) + m h(x)
//            where  g(x) = a (x mod q) - r(x div q)
//                   h(x) = (x div q) - (ax div m)
//                   q = m div a
//                   r = m mod a
// ****************************************************************************
/**
 * Generate the next value in the pseudo-random sequence.
 * @return int - next value
 */
int Random::GetRand(void) {
  int wNew = a * (wSeed % q) - r * (wSeed / q);
  if (wNew > 0) {
    wSeed = wNew;
  } else {
    wSeed = wNew + m;
  }
  return wSeed;
}


// ****************************************************************************
//  Name: Random::ProbTest
//  Date: 06/21/1995 : mdb - Created.
// ****************************************************************************
/**
 * An easy way to do probabilistic coin flipping.
 * @param Qn - numerator of probability of success
 * @param Qd - denominator of probability of success
 * @return bool - true=Success, false=Failure
 */
bool Random::ProbTest(int Qn, int Qd) {
  int b = NumberBits(Qd);
  int wMask = (1 << b) - 1;
  int p;
  while (1) {
    if (cBitsLeft < b) {
      // generate a new random number
      wCurrentRandomValue = GetRand();
      cBitsLeft = 31;
    }
    // remove just the bits we need
    p = wCurrentRandomValue & wMask;
    wCurrentRandomValue >>= b;
    cBitsLeft -= b;
    if (p < Qn) {
      // hit
      return true;
    } else if (p < Qd) {
      // miss
      return false;
    }
    // not applicable -> throw away b bits and try again
  }  // while
}


// ****************************************************************************
//  Name: Random::GetRandRange
//  Date: 01/28/1998 : mdb - Created.
// ****************************************************************************
/**
 * An easy way to get a value in a range.
 * @param wLow - low value of range
 * @param wHigh - high value of range
 * @return int - uniformly distributed value in range [wLow..wHigh]  
 */
int Random::GetRandRange(int wLow, int wHigh) {
  int wTmp, wRange;
  int b, wMask, p;
  if (wLow > wHigh) {
    wTmp = wLow;
    wLow = wHigh;
    wHigh = wTmp;
  }
  wRange = wHigh - wLow;
  if ((b = NumberBits(wRange)) > 31) {
    cerr << ErrorHeader(wERROR) << "Random::GetRandRange(): "
	 << "Range too large, will require " << b
	 << " bits, and we only have 31\n";
  }
  wMask = (1 << b) - 1;
  if (fConserveBits) {
    if (cBitsLeft < b) {
      // generate a new random number
      wCurrentRandomValue = GetRand();
      cBitsLeft = 31;
    }
    // remove just the bits we need
    p = wCurrentRandomValue & wMask;
    wCurrentRandomValue >>= b;
    cBitsLeft -= b;
    // wrap around if too big (same # bits, but larger range)
    while (p > wRange) {
      p -= wRange;
    }
  } else {
    p = (int) (GetRand01() * (wRange+1));
  }
  return wLow + p;
}


// ****************************************************************************
