// ****************************************************************************
//  Name: dc_regex.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 05/07/2001 : mdb - Created.
//        06/16/2001 : mdb - Added REG_NOSUB workaround.
//        07/26/2001 : mdb - Added "*" -> ".*" special case handling.
//  Desc: Simple regex class.
// ****************************************************************************

#include <iostream>


#include "dc_regex.h"
#include "dc_standard.h"


#if defined(DCDBG)
#define DCDBG_RE_COMPILE
//#define DCDBG_RE_MATCH
#endif /* DCDBG */


// Note: the REG_NOSUB compilation flag does not appear to do what I
// expected based on the linux man page.  It seems to only report or
// not report the substring matches for regexec calls, thus we need
// to explicitly check for a non-substring match.


// ctor
DC_Regex::DC_Regex(void) : sbPattern(NULL) {
}

// dtor
DC_Regex::~DC_Regex(void) {
  if (sbPattern) {
    delete [] sbPattern;
    regfree(&re);
  }
}

/**
 * Compile the extended regex pattern
 * @return 0 == ok, else error
 */
int DC_Regex::setPattern(const char *sbPattern_in) {
  if (sbPattern) {
    delete [] sbPattern;
    sbPattern = NULL;
    regfree(&re);
  }
  if (strcmp(sbPattern_in, "*") == 0) sbPattern_in = ".*";  // special case
  int ret = regcomp(&re, sbPattern_in, REG_EXTENDED|REG_NEWLINE);
  if (ret != 0) {
#if defined(DCDBG_RE_COMPILE)
    size_t size = regerror(ret, &re, (char *)NULL, (size_t)0);
    char *buf = new char[size];
    regerror(ret, &re, buf, size);
    cerr << "DC_Regex::setPattern(): failed compilation '"
	 << sbPattern_in << "' : " << buf << endl;
    delete buf;
#endif /* DCDBG_RE_COMPILE */
    return 1;  // error
  }
  sbPattern = StrDup(sbPattern_in);
  return 0;
}

/**
 * Check if string matches the pattern
 * @return t==match, f==no match
 */
bool DC_Regex::match(const char *sb, bool fNoSub) {
  if (!sbPattern) {
    cerr << "DC_Regex::match(): ERROR: calling match() without setting pattern"
	 << endl;
    return false;
  }
#if defined(DCDBG_RE_MATCH)
  cerr << "DC_Regex::match(): attempting match of '" << sb << "' and '"
       << sbPattern << "'" << endl;
#endif /* DCDBG_RE_MATCH */
  if (!sb) return false;

  regmatch_t regmatch[1];
  int ret = regexec(&re, sb, 1, regmatch, 0);
  if (ret != 0) {
#if defined(DCDBG_RE_MATCH)
    size_t size = regerror(ret, &re, (char *)NULL, (size_t)0);
    char *buf = new char[size];
    regerror(ret, &re, buf, size);
    cerr << "DC_Regex::match(): failed matching of '" << sb << "' : "
         << buf << endl;
    delete buf;
#endif /* DCDBG_RE_MATCH */
    return false; // error
  }

  // check the offsets to make sure we have a full string match
  if (fNoSub &&
      !(regmatch[0].rm_so == static_cast(regoff_t,0) &&
	regmatch[0].rm_eo == static_cast(regoff_t,strlen(sb)))) {
    return false;
  }

#if defined(DCDBG_RE_MATCH)
  cerr << "DC_Regex::match(): SUCCESSFUL match of '" << sb << "'" << endl;
#endif /* DCDBG_RE_MATCH */
  return true;  // match
}


// ****************************************************************************
