// ****************************************************************************
//  Name: dc_socket.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 05/23/1997 : mdb - Created.
//        06/03/1997 : mdb - Changed BaseSocket::connect() to reflect problems
//                           seen by Titan.  If a connect call fails, it will
//                           only work if the socket is closed, and recreated.
//        06/19/1997 : mdb - Added read/write using MemList.
//        06/24/1997 : mdb - Member fcns to get/set socket options, inspired by
//                           how Socket++ handled it.
//        09/24/1997 : mdb - Added gethostbyname to other InetAddress::Set()
//                         - Need another InetAddress::Set() for IP address.
//        03/19/1998 : mdb - Made reads/writes with internal ints, endian aware
//        10/26/1999 : mdb - Added an extra addr field to InetAddress to allow
//                           recording of ip address for listen sockets using
//                           the INADDR_ANY wildcard.
//        01/10/2000 : mdb - Added optional fRetry arg for connect.
//        12/11/2000 : mdb - Finally made gethostbyname() calls thread safe.
//                           If calls are made elsewhere, the mutex will need
//                           to be extern'ed.  We did not use gethostbyname_r()
//                           because of its portability problems.
//        12/15/2001 : mdb - Changed Accept(),IAccept() to return -1 for none.
//  Desc: Interprocess communication.
// ****************************************************************************

#include "dc_socket.h"
#include "dc_machine.h"
#include "dc_missing.h"
#if defined(HAVE_PTHREAD)
#  include "dc_guard.h"
#endif

#if defined(AIX)
#  include <strings.h>  // for bzero() which is #def from FD_ZERO
#endif /* AIX */

#ifdef _WIN32
#include <winsock.h>
#define ECONNREFUSED WSAECONNREFUSED
#include "win32.h"
#endif

static const int maxBufTmp = 128;
static char sbBufTmp[maxBufTmp];

#if defined(HAVE_PTHREAD)
static DC_Mutex mutexGetHostBy;
#endif


//#define DCDBG_SOCKET

static void Debug_PrintIpAddress(ostream &os, char *p) {
#if 0
  os << (Types::uint)(p[0]);
  for (Types::uint j=1; j<sizeof(in.s_addr); j++) {
    os << '.' << (Types::uint)(p[j]);
  }
#else
  os << inet_ntoa(*(struct in_addr*)p);
#endif
}

#if defined(DCDBG_SOCKET)
static void Debug_PrintHostent(struct hostent *host) {
  cout << "Host: Official name:  " << host->h_name << endl;
  if (*host->h_aliases) {
    cout << "      Aliases: ";
    while (*host->h_aliases) {
      cout << *host->h_aliases << "  ";
      ++host->h_aliases;
    }
    cout << endl;
  }
  cout << "      Address: type=" << host->h_addrtype;
  switch (host->h_addrtype) {
  case AF_INET: cout << " (AF_INET)"; break;
  case AF_UNIX: cout << " (AF_UNIX)"; break;
  default: cout << " (unknown)";
  }
  cout << ", length=" << host->h_length;
  cout << ", list=";
  Types::uchar **rgsbTmp = (Types::uchar**)host->h_addr_list;
  while (*rgsbTmp) {
    Debug_PrintIpAddress(cout, (char*) *rgsbTmp);
    cout << "  ";
    ++rgsbTmp;
  }
  cout << endl;
}
#else  /* DCDBG_SOCKET */
#define Debug_PrintHostent(host)
#endif /* DCDBG_SOCKET */


// ****************************************************************************
//  Class: InetAddress
// ****************************************************************************


// ****************************************************************************
//  Name: Set
//  Date: 05/23/1997 : mdb - Created.
//        10/26/1999 : mdb - Fixed bug which INADDR_ANY was not used, causing
//                           bind to be too specific on multihomed hosts.
// ****************************************************************************
/**
 * Init an address.
 * @param wPort - port number (0 = use any free port)
 * @param wAddr - adapter to use (INADDR_ANY)  - not used any more
 */
int InetAddress::Set(Types::uint2 wPort, Types::uint4 wAddr) {  // server style
  struct hostent *host;
#if defined(HAVE_PTHREAD)
  DC_Guard guard(mutexGetHostBy);
#endif

  // find local machine name
  if (gethostname(sbBufTmp, maxBufTmp) == 0) {
    TestDelete(sbHostName);
    sbHostName = StrDup(sbBufTmp);
  }
  // dns lookup local host
  if ((host = gethostbyname(sbHostName)) == (struct hostent *)NULL ||
      ! *host->h_addr_list) {
    return -1;
  }
  Debug_PrintHostent(host);

  memset(&in, 0, sizeof(in));
  in.sin_family = AF_INET;
  in.sin_addr.s_addr = htonl(wAddr);
  in.sin_port = htons(wPort);

  memcpy(&addr, host->h_addr_list[0], host->h_length);
  return 0;
}

int InetAddress::Set(char *sbHost, Types::uint2 wPort) {  // client style
  struct hostent *host;
#if defined(HAVE_PTHREAD)
  DC_Guard guard(mutexGetHostBy);
#endif
  // dns lookup given host
  if ((host = gethostbyname(sbHost)) == (struct hostent *)NULL ||
      ! *host->h_addr_list) {
    cerr << ErrorHeader(wWARNING)
	 << "InetAddress::Set(): hostbyname information for "
	 << sbHost << " not found.\n";
    return -1;
  } else {
    Debug_PrintHostent(host);
  }
  TestDelete(sbHostName);
  sbHostName = StrDup(sbHost);

  memset(&in, 0, sizeof(in));
  in.sin_family = AF_INET;
  memcpy(&in.sin_addr, host->h_addr_list[0], host->h_length);
  in.sin_port = htons(wPort);
  return 0;
}

int InetAddress::Set(char *pAddr, int wLen, Types::uint2 wPort) {  // client style
  struct hostent *host;
#if defined(HAVE_PTHREAD)
  DC_Guard guard(mutexGetHostBy);
#endif
  // dns lookup given addr
  if ((host = gethostbyaddr(pAddr, wLen, AF_INET)) == (struct hostent *)NULL ||
      ! *host->h_addr_list) {
    cerr << ErrorHeader(wWARNING)
	 << "InetAddress::Set(): hostbyaddr information for ";
    Debug_PrintIpAddress(cerr, pAddr);
    cerr << " not found.\n";
    //return -1;
    // try to continue without the host entry, since we have the in_addr_t
    TestDeleteNull(sbHostName);
    sbHostName = StrDup("<unknown>");
    
    memset(&in, 0, sizeof(in));
    in.sin_family = AF_INET;
    memcpy(&in.sin_addr, pAddr, wLen);
    in.sin_port = htons(wPort);
  } else {
    Debug_PrintHostent(host);

    TestDeleteNull(sbHostName);
    sbHostName = StrDup(host->h_name);

    memset(&in, 0, sizeof(in));
    in.sin_family = AF_INET;
    memcpy(&in.sin_addr, host->h_addr, host->h_length);
    in.sin_port = htons(wPort);
  }
  return 0;
}

char *InetAddress::setHostName(char *sbHost, bool fProcessName) {
  TestDelete(sbHostName);
  sbHostName = StrDup(sbHost);
  if (fProcessName) {  // perform processing such as ip address lookup
    struct hostent *host;
#if defined(HAVE_PTHREAD)
    DC_Guard guard(mutexGetHostBy);
#endif

    // dns lookup local host
    if ((host = gethostbyname(sbHostName)) == (struct hostent *)NULL ||
	! *host->h_addr_list) {
      return NULL;
    }
    Debug_PrintHostent(host);

    // only replace the ip address if it was previously *not* the wildcard
    if (in.sin_addr.s_addr != htonl(INADDR_ANY)) {
      // replace ip address
      memcpy(&in.sin_addr, host->h_addr_list[0], host->h_length);
    }
    // always replace the extra addr
    memcpy(&addr, host->h_addr_list[0], host->h_length);
  }
  return sbHostName;
}

struct in_addr InetAddress::getHostIpAddress_In(void) {
  // here we check to see if this is a listen socket, and return the extra
  // stored addr instead.  This allows the listen to be for any interface,
  // but we can still get the ip addr for a setHostName() call.
  if (in.sin_addr.s_addr == htonl(INADDR_ANY)) {
    return addr;
  } else {
    return *(struct in_addr*)&(in.sin_addr.s_addr);
  }
}
ostream &operator<<(ostream &os, InetAddress &addr) {
  os << (addr.sbHostName ? addr.sbHostName : inet_ntoa(addr.in.sin_addr))
     << ':' << ntohs(addr.in.sin_port);
  return os;
}


// ****************************************************************************
//  Class: BaseSocket
// ****************************************************************************


int BaseSocket::getopt(option_e op, char *pBuf, int wLen, level_e lev) {
#if defined(AIX)
  unsigned long len;
#else
  socklen_t len;
#endif
  len = wLen;
  if (::getsockopt (fd, lev, op, pBuf, &len) == -1) {
    perror("BaseSocket::getopt(): failure");
  }
  return len;
}

int BaseSocket::setopt(option_e op, char *pBuf, int wLen, level_e lev) {
  if (::setsockopt (fd, lev, op, pBuf, wLen) == -1) {
    perror("BaseSocket::setopt(): failure");
    return -1;
  }
  return 0;
}

int BaseSocket::debug(int wValue) {
  int w = 0;
  getopt(so_debug, (char*)&w, sizeof(w));
  if (wValue != -1) setopt(so_debug, (char*)&wValue, sizeof(wValue));
  return w;
}

int BaseSocket::reuseaddr(int wValue) {
  int w = 0;
  getopt(so_reuseaddr, (char*)&w, sizeof(w));
  if (wValue != -1) setopt(so_reuseaddr, (char*)&wValue, sizeof(wValue));
  return w;
}

int BaseSocket::keepalive(int wValue) {
  int w = 0;
  getopt(so_keepalive, (char*)&w, sizeof(w));
  if (wValue != -1) setopt(so_keepalive, (char*)&wValue, sizeof(wValue));
  return w;
}

int BaseSocket::dontroute(int wValue) {
  int w = 0;
  getopt(so_dontroute, (char*)&w, sizeof(w));
  if (wValue != -1) setopt(so_dontroute, (char*)&wValue, sizeof(wValue));
  return w;
}

int BaseSocket::linger(int wValue, int wTime) {
  struct linger s, t;
  s.l_onoff = 0;
  s.l_linger = 0;
  getopt(so_linger, (char*)&s, sizeof(s));
  if (wValue != -1) {
    t.l_onoff = (wValue == 0 ? 0 : 1);
    t.l_onoff = (wTime == -1 ? s.l_linger : wTime);
    setopt(so_linger, (char*)&t, sizeof(t));
  }
  return (s.l_onoff ? s.l_linger : -1);  // rtn how much linger time was set
}

int BaseSocket::broadcast(int wValue) {
  int w = 0;
  getopt(so_broadcast, (char*)&w, sizeof(w));
  if (wValue != -1) setopt(so_broadcast, (char*)&wValue, sizeof(wValue));
  return w;
}

int BaseSocket::oobinline(int wValue) {
  int w = 0;
  getopt(so_oobinline, (char*)&w, sizeof(w));
  if (wValue != -1) setopt(so_oobinline, (char*)&wValue, sizeof(wValue));
  return w;
}

int BaseSocket::sndbuf(int wValue) {
  int w = 0;
  getopt(so_sndbuf, (char*)&w, sizeof(w));
  if (wValue >= 0) setopt(so_sndbuf, (char*)&wValue, sizeof(wValue));
  return w;
}

int BaseSocket::rcvbuf(int wValue) {
  int w = 0;
  getopt(so_rcvbuf, (char*)&w, sizeof(w));
  if (wValue >= 0) setopt(so_rcvbuf, (char*)&wValue, sizeof(wValue));
  return w;
}

int BaseSocket::gettype(void) {
  int w = 0;
  getopt(so_type, (char*)&w, sizeof(w));
  return w;
}

int BaseSocket::geterror(void) {
  int w = 0;
  getopt(so_error, (char*)&w, sizeof(w));
  return w;
}

int BaseSocket::nodelay(int wValue) {
  int w = 0; 
  getopt(tcp_nodelay, (char*)&w, sizeof(w));
  if (wValue >= 0) setopt(so_rcvbuf, (char*)&wValue, sizeof(wValue));
  return w;
}

int BaseSocket::create(int wFamily) {
  int wRtn;
  if ((wRtn = (int)::socket(wFamily, wType, 0)) < 0) {
    perror("BaseSocket::create(): failure");
  }
  return (fd = wRtn);
}

int BaseSocket::bind(BaseAddress &addr) {
  int wRtn;
  int wRetry = 0;
#if defined(AIX) || defined(LINUX)
  size_t wLength;
#else
  socklen_t wLength;
#endif

  wLength = addr.getLength();
  while (1) {
    if ((wRtn = ::bind(fd, addr.getSockAddr(), wLength)) < 0) {
      if (++wRetry < 10) {
	sleep(wRetry*5);
	if (wRetry == 1) {
	  cerr << ErrorHeader(wWARNING)
	       << "BaseSocket::bind() retrying ";
	  addr.Print(cerr);
	  cerr << " ...\n";
	}
	continue;
      } else {
	perror("BaseSocket::bind(): failure");
	break;
      }
    }
    // success
    if (::getsockname(fd, addr.getSockAddr(), &wLength) < 0) {
      perror("BaseSocket::bind(): getsockname() failure");
    }
    break;
  }
  return wRtn;
}

int BaseSocket::listen(int wBackLog) {
  int wRtn;
  if ((wRtn = ::listen(fd, wBackLog)) < 0) {
    perror("BaseSocket::listen(): failure");
  }
  return wRtn;
}

int BaseSocket::accept(BaseAddress &addrAccept) {
  int wRtn;
#if defined(AIX) || defined(LINUX)
  size_t wLength;
#else
  socklen_t wLength;
#endif
  while (1) {
    wLength = addrAccept.getLength();
    if ((wRtn = (int)::accept(fd, addrAccept.getSockAddr(), &wLength)) < 0) {
      if (errno == EINTR) continue;
      perror("BaseSocket::accept(): failure");
    } else {
      break;
    }
  }
  return wRtn;
}

int BaseSocket::connect(BaseAddress &addr, bool fRetry) {
  int wRtn;
  int wRetry = 0;

  while (1) {
    if ((wRtn = ::connect(fd, addr.getSockAddr(), addr.getLength())) < 0) {
      // cannot recall connect() on EINTR failure, as per [Stevens,p124]
      //if (errno == EINTR) continue;
      if (errno == ECONNREFUSED) break;
      if (fRetry && ++wRetry < 5) {
	close();
	create(addr.getAddressFamily());
	sleep(wRetry*5);
	if (wRetry == 1) {
	  perror("BaseSocket::connect(): failure");
	  cerr << ErrorHeader(wWARNING)
	       << "BaseSocket::connect() retrying ";
	  addr.Print(cerr);
	  cerr << " ...\n";
	}
	continue;
      } else {
	if (fRetry) perror("BaseSocket::connect(): failure");
      }
    }
    break;
  }
  return wRtn;
}

int BaseSocket::close(void) {
  int wRtn = -1;
  if (fd != -1) {  // not already closed

#ifdef _WIN32
	if ((wRtn = ::closesocket(fd)) != 0) {
#else
	if ((wRtn = ::close(fd)) < 0) {
#endif

      cout << "BaseSocket::close(): failure (fd=" << fd << ") : " << endl;
      // CLH - not mt safe - << strerror(errno) << endl;
      //perror("BaseSocket::close(): failure");
    }
    fd = -1;
  } else {
    cerr << ErrorHeader(wWARNING)
	 << "BaseSocket::close(): trying to close socket that is not open.\n";
  }
  return wRtn;
}

int BaseSocket::shutdown(int howto) {
  int wRtn = -1;
  if (fd != -1) {  // not already closed
    // arg2 = 2 -> SHUT_RDWR
    if ((wRtn = ::shutdown(fd, howto)) < 0) {
      perror("BaseSocket::shutdown(): failure");
    }
    fd = -1;
  } else {
    cerr << ErrorHeader(wWARNING)
	 << "BaseSocket::shutdown(): trying to shutdown socket that is not open.\n";
  }
  return wRtn;
}

// ****************************************************************************
//  Name: read
//  Date: 05/28/1997 : mdb - Created.
// ****************************************************************************
/**
 * Reads a block into a buffer.  If eof encountered, the requested
 * bytes may not all have been read.  If a stream socket, this may
 * involve several system calls to read, while this routine blocks
 * until the entire block is read.
 * @param pBuf - buffer to read into
 * @param nBytes - buffer to read into
 * @return int - number of characters read into buffer, <0=error
 */
#if 1
int BaseSocket::read(int fd, char *pBuf, int nBytes) {
  int n, nMore = nBytes;
  while (nMore > 0) {

#ifdef _WIN32
	while ((n = ::recv(fd, pBuf, nMore,NULL)) < 0) {
#else
	while ((n = ::read(fd, pBuf, nMore)) < 0) {
#endif

      if (errno == EINTR) continue;
      return n;  // error
    }
    if (n == 0) break;  // eof

    nMore -= n;
    pBuf += n;
  }
  return nBytes - nMore;
}

#else /* 1 */
int BaseSocket::read(int fd, char *pBuf, int nBytes) {
  int n, nMore = nBytes;
  cout << "BaseSocket::read(fd=" << fd << "): entry read nBytes="
       << nMore << " bytes\n";
  while (nMore > 0) {
    cout << "BaseSocket::read(fd=" << fd << "): still need to read nMore="
	 << nMore << " bytes\n";
    while ((n = ::read(fd, pBuf, nMore)) < 0) {
      if (errno == EINTR) {
	cout << "BaseSocket::read(fd=" << fd << "): failed read, rtn=" << n
	     << ", errno=EINTR\n";
	continue;
      }
      cout << "BaseSocket::read(fd=" << fd << "): failed read, rtn=" << n
	   << ", errno=" << errno << endl;
      return n;  // error
    }
    cout << "BaseSocket::read(fd=" << fd << "): rtn=" << n << endl;
    if (n == 0) break;  // eof

    nMore -= n;
    pBuf += n;
  }
  return nBytes - nMore;
}
#endif /* 1 */

int BaseSocket::read(MemList &ml) {
  int n, nRtn = 0;
  MemBuf *pmb;
  for (int i=ml.StartIter(); i; i--) {
    pmb = ml.pNext();
    if ((n = read(pmb->getPtr(), pmb->getSize())) < 0) return n;  // error
    nRtn += n;
  }
  return nRtn;
}

int BaseSocket::read(char *&sb) {
  int   n, nRtn = 0;
  Types::uint wSize;
  char *pBuf;
  sb = NULL;
  // first get size
  if ((n = read(&wSize, sizeof(wSize))) < 0) return n;  // error
  if (fEndianConvertRead) ConvertEndian(1, sizeof(Types::uint), &wSize);
  nRtn += n;
  // now read the string, delimited by null bytes
  if (wSize > 0) {
    pBuf = new char[wSize];
    if ((n = read(pBuf, wSize)) < 0) return n;  // error
    nRtn += n;
    sb = pBuf;
  }
  return nRtn;
}

int BaseSocket::read(Argv &a) {
  int   n, nRtn = 0;
  Types::uint i, wSize;
  char *pBuf;
  // first get size
  if ((n = read(&wSize, sizeof(wSize))) < 0) return n;  // error
  if (fEndianConvertRead) ConvertEndian(1, sizeof(Types::uint), &wSize);
  nRtn += n;
  // now read the strings, delimited by null bytes
  if (wSize > 0) {
    pBuf = new char[wSize];
    if ((n = read(pBuf, wSize)) < 0) return n;  // error
    nRtn += n;
    // build the argv struct by breaking at the null bytes
    a.Flush();
    a.Append(pBuf, true);
    for (i=1; i<wSize; i++) {
      if (pBuf[i-1] == '\0') {
	a.Append(pBuf+i, false);
      }
    }
  }
  return nRtn;
}


// ****************************************************************************
//  Name: write
//  Date: 05/28/1997 : mdb - Created.
// ****************************************************************************
/**
 * Writes a block from a buffer.  If eof encountered, the requested
 * bytes may not all have been written.  If a stream socket, this may
 * involve several system calls to write, while this routine blocks
 * until the entire block is written.
 * @param pBuf - buffer to read into
 * @param nMaxBytes - buffer to read into
 * @return int - number of characters written from buffer, <0=error
 */
#if 1
int BaseSocket::write(int fd, char *pBuf, int nBytes) {
  int n, nMore = nBytes;
  while (nMore > 0) {

#ifdef _WIN32
	while ((n = ::send(fd, pBuf, nMore, NULL)) != SOCKET_ERROR) {
#else
	while ((n = ::write(fd, pBuf, nMore)) < 0) {
#endif

      if (errno == EINTR) continue;
      return n;  // error
    }

    nMore -= n;
    pBuf += n;
  }
  return nBytes - nMore;
}

#else /* 1 */
int BaseSocket::write(int fd, char *pBuf, int nBytes) {
  int n, nMore = nBytes;
  while (nMore > 0) {
    cout << "BaseSocket::write(fd=" << fd << "): still need to write nMore="
	 << nMore << " bytes\n";
    while ((n = ::write(fd, pBuf, nMore)) < 0) {
      if (errno == EINTR) {
	cout << "BaseSocket::write(fd=" << fd << "): failed write, rtn=" << n
	     << ", errno=EINTR\n";
	continue;
      }
      cout << "BaseSocket::write(fd=" << fd << "): failed write, rtn=" << n
	   << ", errno=" << errno << endl;
      return n;  // error
    }
    cout << "BaseSocket::write(fd=" << fd << "): rtn=" << n << endl;

    nMore -= n;
    pBuf += n;
  }
  return nBytes - nMore;
}

#endif /* 1 */
int BaseSocket::write(MemList &ml) {
  int n, nRtn = 0;
  MemBuf *pmb;
  for (int i=ml.StartIter(); i; i--) {
    pmb = ml.pNext();
    if ((n = write(pmb->getPtr(), pmb->getSize())) < 0) return n;  // error
    nRtn += n;
  }
  return nRtn;
}

int BaseSocket::write(char *sb) {
  int   n, nRtn = 0;
  Types::uint wSize, wTmp;
  wSize = wTmp = (Types::uint)strlen(sb) + 1;
  // first write the size
  if (fEndianConvertWrite) ConvertEndian(1, sizeof(Types::uint), &wSize);
  if ((n = write(&wSize, sizeof(wSize))) < 0) return n;  // error
  if (fEndianConvertWrite) wSize = wTmp;
  nRtn += n;
  // now write the string, delimited by null bytes
  if (wSize > 0) {
    if ((n = write(sb, wSize)) < 0) return n;  // error
    nRtn += n;
  }
  return nRtn;
}

int BaseSocket::write(Argv &a) {
  int   i, n, nRtn = 0;
  Types::uint wSize, wTmp;
  // first determine size and send it
  wSize = 0;
  for (i=0; i<a.GetNum(); i++) {
	  wSize += (Types::uint)strlen(a.Get(i)) + 1;
  }
  if (fEndianConvertWrite) {
    wTmp = wSize;
    ConvertEndian(1, sizeof(Types::uint), &wSize);
  }
  if ((n = write(&wSize, sizeof(wSize))) < 0) return n;  // error
  if (fEndianConvertWrite) wSize = wTmp;
  nRtn += n;
  // now write the strings, delimited by null bytes
  if (wSize > 0) {
    for (i=0; i<a.GetNum(); i++) {
      //if ((n = write(a.Get(i), (int)strlen(a.Get(i)) + 1)) < 0) return n;  // error
      nRtn += n;
    }
  }
  return nRtn;
}


// ****************************************************************************
//  Name: readline
//  Date: 05/28/1997 : mdb - Created.
// ****************************************************************************
/**
 * Read a full line from the input stream, up to (but not including)
 * a newline or eof.  The buffer is null terminated.
 * This really needs to be read in larger pieces, but then other
 * read routines would have to know to check the common read buffer.
 * @param pBuf - buffer to read into
 * @param nMaxBytes - buffer to read into
 * @return int - number of characters read into buffer, <0=error
 */
int BaseSocket::readline(int fd, char *pBuf, int nMaxBytes, char chEol) {
  int  i, n;
  char ch;
  for (i=0; i<nMaxBytes-1; i++) {

#ifdef _WIN32
	if ((n = ::recv(fd, &ch, 1,NULL)) == 1) {
#else
	if ((n = ::read(fd, &ch, 1)) == 1) {
#endif

      if (ch == chEol) break;
      *pBuf++ = ch;
    } else if (n == 0) {  // eof
      break;
    } else {
      perror("");
      return -1;  // error
    }
  }
  *pBuf = '\0';
  return i;
}


// ****************************************************************************
//  Class: ServerSocket
// ****************************************************************************

int ServerSocket::CreateBindListen(BaseAddress &addr, int wBackLog) {
  if (create(addr.getAddressFamily()) >= 0) {
    if (bind(addr) == 0 && listen(wBackLog) == 0) {
      return 0;
    }
    close();
  }
  return -1;
}

int ServerSocket::IAccept(BaseAddress &addrAccept) {
  int wRtn;
  fd_set fdset;
  struct timeval tv = {0, 0};

  FD_ZERO(&fdset);
  FD_SET(fd, &fdset);
  if ((wRtn = ::select(fd+1, &fdset, NULL, NULL, &tv)) == -1) {
    if (errno != EINTR) perror("ServerSocket::IAccept(): failure");
  } else if (wRtn == 1) {
    return accept(addrAccept);  // new connection
  }
  return -1;  // no connections pending
}

int ServerSocket::Accept(BaseAddress &addrAccept) {
  fd_set fdset;

  FD_ZERO(&fdset);
  FD_SET(fd, &fdset);
  if (::select(fd+1, &fdset, NULL, NULL, NULL) == -1) {
    if (errno != EINTR) perror("ServerSocket::Accept(): failure");
  } else {
    return accept(addrAccept);  // new connection
  }
  return -1;  // no connections pending
}


// ****************************************************************************
//  Class: ClientSocket
// ****************************************************************************


int ClientSocket::Connect(BaseAddress &addr, bool fRetry) {
  if (create(addr.getAddressFamily()) >= 0 &&
      connect(addr, fRetry) == 0) {
    return 0;
  }
  return -1;
}

int ClientSocket::Connect(char *sbHost, Types::uint2 wPort, InetAddress &addr,
			  bool fRetry) {
  if (addr.Set(sbHost, wPort) == 0) {
    return Connect(addr, fRetry);
  }
  return -1;
}


// ****************************************************************************
