// -*- c++ -*-
// ****************************************************************************
//  Name: dc_socket.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 05/23/1997 : mdb - Created.
//        01/10/2000 : mdb - Added optional fRetry arg for connect.
//  Desc: Interprocess communication.
// ****************************************************************************
/** @file dc_socket.h */

#if !defined(H_SOCKET)
#define H_SOCKET

#include <stddef.h>
#include <sys/types.h>
#if defined(AIX)
#include <sys/select.h>
#endif /* AIX */
#ifndef _WIN32
  #include <sys/socket.h>
  #include <sys/un.h>
  #include <netdb.h>
  #include <netinet/in.h>
  #include <netinet/tcp.h>
  #include <arpa/inet.h>
  #include <sys/time.h>
#else
  #include <winsock.h>
  #include <time.h>
#endif

#include "dc_standard.h"
#include "dc_memory.h"
#include "dc_argv.h"


static const int DEFAULT_BACKLOG = SOMAXCONN;

/// Base class for host addressing
class BaseAddress {
public:
  virtual ~BaseAddress() { };

  virtual struct sockaddr *getSockAddr(void) = 0;
  virtual int getLength(void) = 0;
  virtual int getAddressFamily(void) = 0;
  virtual void Print(ostream &os) = 0;
};

/// Internet address class
class InetAddress : public BaseAddress {
  struct sockaddr_in in;
  char *sbHostName;
  struct in_addr addr;  // used for listen sockets
public:
  InetAddress(void) { sbHostName = NULL; };
  ~InetAddress(void) { TestDelete(sbHostName); };

  int Set(Types::uint2 wPort, Types::uint4 wAddr = INADDR_ANY);
  int Set(char *sbHost, Types::uint2 wPort);
  int Set(char *pAddr, int wLen, Types::uint2 wPort);

  struct sockaddr *getSockAddr(void) { return (struct sockaddr *)&in; };
  struct sockaddr_in *getSockAddr_In(void) { return &in; };
  int getLength(void) { return sizeof(struct sockaddr_in); };
  int getAddressFamily(void) { return AF_INET; };
  void Print(ostream &os) { os << *this; };

  char *getHostName(void) { return sbHostName; };
  char *setHostName(char *sbHost, bool fProcessName=false);
  struct in_addr getHostIpAddress_In(void);
  // BUGBUG: this is not threadsafe
  char *getHostIpAddress(void) {
    return inet_ntoa(getHostIpAddress_In());
  };

  Types::uint2 getPort(void) { return ntohs(in.sin_port); };
  struct in_addr *getInAddr(void) { return &(in.sin_addr); };

  friend ostream &operator<<(ostream &os, InetAddress &addr);
};

#ifndef _WIN32
/// Unix address class
class UnixAddress : public BaseAddress {
  struct sockaddr_un un;
public:
  int Set(char *sbPathName);

  struct sockaddr *getSockAddr(void) { return (struct sockaddr *)&un; };
  int getLength(void) { return sizeof(struct sockaddr_un); };
  int getAddressFamily(void) { return AF_UNIX; };
  void Print(ostream &os) { os << "(UnixAddress not implemented)"; };

  friend ostream &operator<<(ostream &os, UnixAddress &addr);
};
#endif

/// Base socket class
class BaseSocket {
public:
  enum type_e {
    sock_stream    = SOCK_STREAM,
    sock_dgram     = SOCK_DGRAM,
    sock_raw       = SOCK_RAW,
    sock_rdm       = SOCK_RDM,
    sock_seqpacket = SOCK_SEQPACKET
  };
  enum level_e {
    sol_socket     = SOL_SOCKET,
    ipproto_ip     = IPPROTO_IP,
    ipproto_tcp    = IPPROTO_TCP
  };
  enum option_e {
    // socket level options (SOL_SOCKET)
    so_debug       = SO_DEBUG,       // recording of debug info
#if !defined(LINUX)
    so_acceptconn  = SO_ACCEPTCONN,  // socket has had listen()
#endif /* !LINUX */
    so_reuseaddr   = SO_REUSEADDR,   // allow local address reuse
    so_keepalive   = SO_KEEPALIVE,   // keep connections alive
    so_dontroute   = SO_DONTROUTE,   // routing bypass for outmsg
    so_broadcast   = SO_BROADCAST,   // permission to broadcast
#if !defined(LINUX)
    so_useloopback = SO_USELOOPBACK, // bypass hardware when possible
#endif /* !LINUX */
    so_linger      = SO_LINGER,      // linger on close if data present
    so_oobinline   = SO_OOBINLINE,   // leave oob data in band
    so_sndbuf      = SO_SNDBUF,      // size of buffer for send
    so_rcvbuf      = SO_RCVBUF,      // size of buffer for receive
#if !defined(LINUX)
    so_sndlowat    = SO_SNDLOWAT,    // send low water mark
    so_rcvlowat    = SO_RCVLOWAT,    // receive low water mark
    so_sndtimeo    = SO_SNDTIMEO,    // send timeout
    so_rcvtimeo    = SO_RCVTIMEO,    // receive timeout
#endif /* !LINUX */
    so_error       = SO_ERROR,       // get and clear error status
    so_type        = SO_TYPE,        // get type of socket

    // tcp level options (IPPROTO_TCP)
    tcp_nodelay    = TCP_NODELAY     // disable Nagle's algorithm
  };

protected:
  int fd;
  int wType;

  int fEndianConvertRead;
  int fEndianConvertWrite;

public:
  BaseSocket(void) : fd(-1), wType(SOCK_STREAM),
		     fEndianConvertRead(0), fEndianConvertWrite(0) {
  };
#if 0
  ~BaseSocket(void) { if (fd != -1) close(); };
#endif /* 0 */

  void setType(int wSockType) { wType = wSockType; };

  void setEndian(int fConvertRead, int fConvertWrite) {
    fEndianConvertRead = fConvertRead;
    fEndianConvertWrite = fConvertWrite;
  };
  int getEndianConvertRead(void) { return fEndianConvertRead; };
  int getEndianConvertWrite(void) { return fEndianConvertWrite; };

  int getopt(option_e op, char *pBuf, int wLen, level_e lev = sol_socket);
  int setopt(option_e op, char *pBuf, int wLen, level_e lev = sol_socket);

  // use arg for setopt, and omit arg for getopt
  // socket level options
  int debug(int wValue=-1);
  int reuseaddr(int wValue=-1);
  int keepalive(int wValue=-1);
  int dontroute(int wValue=-1);
  int linger(int wValue=-1, int wTime=-1);
  int broadcast(int wValue=-1);
  int oobinline(int wValue=-1);
  int sndbuf(int wValue=-1);
  int rcvbuf(int wValue=-1);
  int gettype(void);
  int geterror(void);
  // tcp level options
  int nodelay(int wValue=-1);
    
  int create(int wFamily);
  int bind(BaseAddress &addr);
  int listen(int wBackLog = DEFAULT_BACKLOG);
  int accept(BaseAddress &addrAccept);
  int connect(BaseAddress &addr, bool fRetry=true);
  int close(void);
  int shutdown(int howto = 2); // 2 == SHUT_RDWR

  int read(int fd, char *pBuf, int nBytes);
  int read(char *pBuf, int nBytes) { return read(fd, pBuf, nBytes); };
  int read(void *pBuf, int nBytes) {
    return read(fd, (char*)pBuf, nBytes);
  };
  int read(MemBlock &mb) { return read(fd, mb.getPtr(), mb.getSize()); };
  int read(MemList &ml);
  int read(char *&sb);
  int read(Argv &a);

  int write(int fd, char *pBuf, int nBytes);
  int write(char *pBuf, int nBytes) { return write(fd, pBuf, nBytes); };
  int write(void *pBuf, int nBytes) {
    return write(fd, (char*)pBuf, nBytes);
  };
  int write(MemBlock &mb) { return write(fd, mb.getPtr(), mb.getSize()); };
  int write(MemList &ml);
  int write(char *sb);
  int write(Argv &a);

  int readline(int fd, char *pBuf, int nMaxBytes, char chEol = '\n');
  int readline(char *pBuf, int nMaxBytes, char chEol = '\n') {
    return readline(fd, pBuf, nMaxBytes, chEol);
  };

  int &getFd(void) { return fd; };
};

/// Client socket
class ClientSocket : public BaseSocket {
public:
  int Connect(BaseAddress &addr, bool fRetry=true);
  int Connect(char *sbHost, Types::uint2 wPort, InetAddress &addr,
	      bool fRetry=true);

  int Close(void) { return ((fd != -1) ? close() : 0); };
};

/// Server socket
class ServerSocket : public BaseSocket {
public:
  int CreateBindListen(BaseAddress &addr, int wBackLog = DEFAULT_BACKLOG);
  int IAccept(BaseAddress &addrAccept);  // non-blocking
  int Accept(BaseAddress &addrAccept);  // blocking

  int Close(void) { return ((fd != -1) ? close() : 0); };
};


#endif /* H_SOCKET */
// ****************************************************************************
