// ****************************************************************************
//  Name: dc_standard.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 06/17/1995 : mdb - Created.
//  Desc: Miscellaneous supporting routines.
// ****************************************************************************

#include "dc_standard.h"
#include <new>

#ifdef _WIN32
#include "win32.h"
#endif

//
//  Declarations
//
const int wWARNING = 0;
const int wERROR = 1;
const int wFATAL = 2;

char *sbProgramName = NULL;
char *sbProgramDir = NULL;

char *sbScopeName = NULL;  // externally settable, by threads for example


//#define DCDBG_STANDARD_NEW_HANDLER


static char sbBuffer[256] = "";


// ****************************************************************************
//  Name: StandardNewHandler
//  Date: 06/19/1995 : mdb - Created.
//        01/20/2000 : mdb - On the linux cluster, malloc can fail with EINTR.
//                           Note that wCount is for *all* calls to new, and
//                           not per call, since we have no way to distinguish.
//                         - Added no output version, and call abort() to dump.
// ****************************************************************************
/**
 * Catch failure of a new operation.
 */
static void StandardNewHandler(void) {
#if defined(DCDBG_STANDARD_NEW_HANDLER)
  static int wCount = 0;
  if (errno == EINTR && ++wCount < 25) {
    cerr << ErrorHeader(wWARNING) << "EINTR malloc, retry ..." << endl;
    return;
  }
#else /* DCDBG_STANDARD_NEW_HANDLER */
  if (errno == EINTR) return;  // try again
#endif /* DCDBG_STANDARD_NEW_HANDLER */
  cerr << ErrorHeader(wFATAL) << "out of free store." << endl;
  (void)StandardCleanup();
  abort();
}


// ****************************************************************************
//  Name: StandardInit
//  Date: 06/19/1995 : mdb - Created.
//        06/27/1995 : mdb - Sets program name for error messages.
//        05/27/1997 : mdb - Program name and program directory found.
// ****************************************************************************
/**
 * Currently only sets the new error handler.
 * Post-conditions:
 *      1. all new operations will be caught by StandardNewHandler().
 */
int StandardInit(int argc, char *argv[]) {
  int wRtn = 0;  // local return
  char *sb;
  int i;

  i = argc; // just to make -Wall shut up

  set_new_handler(&StandardNewHandler);

  // find executable name
  if ((sb = strrchr(argv[0], '/'))) {
    i = (int)(sb - argv[0]);
    sbProgramName = StrDup(++sb);
    sbProgramDir = StrDup(argv[0]);  // some wasted space
    sbProgramDir[i] = '\0';
  } else {
    sbProgramName = StrDup(argv[0]);
    sbProgramDir = StrDup(".");
  }
  return wRtn;
}


// ****************************************************************************
//  Name: StandardCleanup
//  Date: 06/28/1995 : mdb - Created.
// ****************************************************************************
/**
 * Comes here at end of program.
 */
int StandardCleanup(void) {
  int wRtn = 0;  // local return

  TestDeleteNull(sbProgramName);
  TestDeleteNull(sbProgramDir);

  return wRtn;
}


// ****************************************************************************
//  Name: StrCmp
//  Date: 04/24/1994 : mdb - Created.
//        06/17/1995 : mdb - Added fCaseSensitive flag.
//        06/03/1997 : mdb - Moved test out of inner loop.
// ****************************************************************************
/**
 * Similar to strcmp, but case sensitivity can be selected.
 * Pre-conditions:
 *   1. sbStr1 and sbStr2 should be a valid strings
 * @param sbStr1 - string to compare
 * @param sbStr2 - string to compare
 * @param fCaseSensitive - T=case sensitive compare, F=case insensitive compare
 * @return int - 0 = same, otherwise = diferent
 */
int StrCmp(const char *sbStr1, const char *sbStr2, bool fCaseSensitive) {

  //  Same length?
  if (strlen(sbStr1) != strlen(sbStr2)) {
    return 1;
  }

  if (fCaseSensitive) {
    while (*sbStr1 && (*sbStr1 == *sbStr2)) {
      ++sbStr1;
      ++sbStr2;
    }
  } else {
    while (*sbStr1 && (tolower(*sbStr1) == tolower(*sbStr2))) {
      ++sbStr1;
      ++sbStr2;
    }
  }
  return (*sbStr1 - *sbStr2);
}

int StrNCmp(const char *sbStr1, const char *sbStr2, int n,
	    bool fCaseSensitive) {
  if (fCaseSensitive) {
    while (*sbStr1 && (n > 0) && (*sbStr1 == *sbStr2)) {
      ++sbStr1;
      ++sbStr2;
      --n;
    }
  } else {
    while (*sbStr1 && (n > 0) && (tolower(*sbStr1) == tolower(*sbStr2))) {
      ++sbStr1;
      ++sbStr2;
      --n;
    }
  }
  return ((n == 0) ? 0 : (*sbStr1 - *sbStr2));
}


// ****************************************************************************
//  Name: StrDup
//  Date: 04/10/1994 : mdb - Created.
// ****************************************************************************
/**
 * Duplicate a string by creating new space and copying it.
 * @param sb - string to duplicate
 * @return char* - points to new string
 */
char *StrDup(const char *sb) {
  if (!sb) return NULL;
  char *sbTmp = new char[strlen(sb) + 1];
  strcpy(sbTmp, sb);
  return sbTmp;
}


// ****************************************************************************
//  Name: StrNDup
//  Date: 07/30/1997 : mdb - Created.
// ****************************************************************************
/**
 * Duplicate a string by creating new space and copying up to N of it.
 * @param sb - string to duplicate
 * @param n  - max chars to dup
 * @return char* - points to new string
 */
char *StrNDup(const char *sb, int n) {
  char *sbTmp;
  int  len;

  if (!sb) return NULL;
  len = (int)strlen(sb);

  if (len > n) len = n;

  sbTmp = new char[len + 1];
  strncpy(sbTmp, sb, len);
  sbTmp[len] = '\0';
  return sbTmp;
}


// ****************************************************************************
//  Name: StrNCopyNull
//  Date: 05/04/1994 : mdb - Created.
// ****************************************************************************
/**
 * Does basically a strncpy but insures a NULL char at the end.
 * @param sbDest - string to copy to (already allocated space)
 * @param sbSrc  - string to copy from
 * @param cchMax - max characters to be held in sbDest including NULL byte
 */
void StrNCopyNull(char *sbDest, const char *sbSrc, int cchMax) {
  strncpy(sbDest, sbSrc, cchMax-1);
  sbDest[cchMax-1] = '\0';
}


// ****************************************************************************
//  Name: StrToLower
//  Date: 06/18/1995 : mdb - Created.
// ****************************************************************************
/** 
 * Destructively converts a string to all lower case.
 * @param sb - string to convert
 * @return char* - redundantly points to input string
 */
char *StrToLower(char *sb) {
  for (char *pch = sb; *pch != '\0'; pch++) {
    *pch = tolower(*pch);
  }
  return sb;
}


// ****************************************************************************
//  Name: StrToUpper
//  Date: 06/18/1995 : mdb - Created.
// ****************************************************************************
/**
 * Destructively converts a string to all upper case.
 * @param sb - string to convert
 * @param char* - redundantly points to input string
 */
char *StrToUpper(char *sb) {
  for (char *pch = sb; *pch != '\0'; pch++) {
    *pch = toupper(*pch);
  }
  return sb;
}


// ****************************************************************************
//  Name: StrLeftTrimWhite
//  Date: 05/21/1997 : mdb - Created.
// ****************************************************************************
/**
 * Destructively removes whitespace from left side of string.
 * @param sb - string to convert
 * @return char* - redundantly points to input string
 */
char *StrLeftTrimWhite(char *sb) {
  char *sbSrc, *sbDst = sb;
  
  // look for first non-whitespace
  for (sbSrc=sb; isspace(*sbSrc) && *sbSrc != '\0'; sbSrc++) {
    // null
  }
  // copy rest of string
  while (*sbSrc != '\0') {
    *sbDst++ = *sbSrc++;
  }
  *sbDst = '\0';
  return sb;
}


// ****************************************************************************
//  Name: StrRightTrimWhite
//  Date: 05/21/1997 : mdb - Created.
// ****************************************************************************
/**
 * Destructively removes whitespace from right side of string.
 * @param sb - string to convert
 * @return char* - redundantly points to input string
 */
char *StrRightTrimWhite(char *sb) {
  char *pch;
  for (pch = sb+strlen(sb)-1; pch >= sb && isspace(*pch); pch--) {
    // null
  }
  pch[1] = '\0';
  return sb;
}


// ****************************************************************************
//  Name: StrTrimWhite
//  Date: 05/21/1997 : mdb - Created.
// ****************************************************************************
/** 
 * Destructively removes whitespace from both sides of string.
 * @param sb - string to convert
 * @return char* - redundantly points to input string
 */
char *StrTrimWhite(char *sb) {
  return StrLeftTrimWhite(StrRightTrimWhite(sb));
}


// ****************************************************************************
//  Name: MemNDup
//  Date: 09/26/1997 : mdb - Created.
// ****************************************************************************
/**
 * Duplicate a memory buffer by creating new space, and copying.
 * @param p - ptr to memory area to duplicate
 * @param n - chars to dup
 * @return char* - points to duplicate
 */
char *MemNDup(char *p, int n) {
  char *pTmp;

  pTmp = new char[n];
  strncpy(pTmp, p, n);
  return pTmp;
}


// ****************************************************************************
//  Name: strrstr
//  Date: 01/08/2000 : mdb - Created.
// ****************************************************************************
/**
 * Obvious counterpart to strstr, but does not seem to exist.
 * @param sbStr - string to search
 * @param sbSub - substring to look for
 * @return char* - pointer to beginning of substring in sb
 */
char *strrstr(const char *sbStr, const char *sbSub) {
  char ch, *p, *pSub = (char *)sbSub;
  int wLen;

  ch = *pSub++;
  if (ch == '\0') return (char *)sbStr; // arbitrary return (undefined)
  wLen = (int)strlen(pSub);

  for (p=(char *)sbStr+strlen(sbStr)-1; p >= sbStr; --p) {
    if (*p == ch && strncmp(p+1, pSub, wLen) == 0) return p;  // found
  }
  return NULL;
}


// ****************************************************************************
//  Name: StrEndDelim
//  Date: 01/08/2000 : mdb - Created.
// ****************************************************************************
/**
 * Find the ending delimiter in a string for a given ending delimiter.
 * @param sbStr - string to search (ptr must be at the start delim)
 * @param sbEnd - ending delimiter to look for
 * @param sbStart - starting delimiter (!NULL = handle nesting of delims)
 * @return char* - ptr to corresponding ending delimiter, NULL=not found
 */
char *StrEndDelim(char *sbStr, char *sbEnd, char *sbStart) {
  char *pEnd, *pStart;
  int wLenEnd, wLenStart, wCount = 0;

  if (!sbStart) return strstr(sbStr, sbEnd);  // not nested

  wLenEnd   = (int)strlen(sbEnd);
  wLenStart = (int)strlen(sbStart);

  sbStr += wLenStart;  // skip beginning start delim
  pEnd = strstr(sbStr, sbEnd);
  pStart = strstr(sbStr, sbStart);
  while (pEnd) {
    if (!pStart || pEnd < pStart) {
      if (wCount == 0) return pEnd;  // found
      --wCount;
      pEnd = strstr(pEnd+wLenEnd, sbEnd);
    } else {
      ++wCount;
      pStart = strstr(pStart+wLenStart, sbStart);
    }
  }
  return NULL;
}


// ****************************************************************************
//  Name: ErrorHeader
//  Date: 06/27/1995 : mdb - Created. (based on [Stroustrup p134])
// ****************************************************************************
/**
 * A flexible error printing routine.
 * @param wSeverity - 0=warning, 1=error, 2=fatal error
 * @return char* - resulting string
 */
char *ErrorHeader(int wSeverity) {

  snprintf(sbBuffer, sizeof(sbBuffer), "\n>>> ");
  if (sbProgramName != NULL) {
    snprintf(sbBuffer+strlen(sbBuffer), sizeof(sbBuffer)-strlen(sbBuffer), 
	     "%s: ", sbProgramName);
  }
  if (sbScopeName != NULL) {
    snprintf(sbBuffer+strlen(sbBuffer), sizeof(sbBuffer)-strlen(sbBuffer),
	     "%s: ", sbScopeName);
  }
  if (wSeverity == 2) {
    snprintf(sbBuffer+strlen(sbBuffer), sizeof(sbBuffer)-strlen(sbBuffer),
	     "%s: ", "Fatal Error");
  } else if (wSeverity == 1) {
    snprintf(sbBuffer+strlen(sbBuffer), sizeof(sbBuffer)-strlen(sbBuffer),
	     "%s: ", "Error");
  } else {
    snprintf(sbBuffer+strlen(sbBuffer), sizeof(sbBuffer)-strlen(sbBuffer),
	     "%s: ", "Warning");
  }

  return sbBuffer;
}
char *ErrorIndent(void) {
  char *p;
  int n = 6;
  if (sbProgramName != NULL) n += (int)strlen(sbProgramName);
  for (p=sbBuffer; n; n--) {
    *(p++) = ' ';
  }
  *p = '\0';

  return sbBuffer;
}


// ****************************************************************************
//  Name: NextChar
//  Date: 07/05/1995 : mdb - Created.
// ****************************************************************************
/**
 * Find out what the next char to read will be.
 * @param is - stream to read
 * @return char - next char
 */
char NextChar(istream &is) {
  char chRtn;

  if (is.get(chRtn)) {
    is.putback(chRtn);
  }
  return chRtn;
}


// ****************************************************************************
//  Name: EatWhite
//  Date: 07/05/1995 : mdb - Created. (based on [Stroustrup p333])
// ****************************************************************************
/**
 * Gobble spaces from input stream up to next non-space.
 * @param is - stream to read
 * @return istream& - same as input
 */
istream &EatWhite(istream &is) {
  char ch;

  while (is.get(ch)) {
    // too far?
    if (isspace(ch) == 0 || ch == '\n') {
      is.putback(ch);
      break;
    }
  }
  return is;
}


// ****************************************************************************
//  Name: EatLine
//  Date: 07/05/1995 : mdb - Created.
// ****************************************************************************
/**
 * Gobble everything up to the next end of line.
 * @param is - stream to read
 * @return istream& - same as input
 */
istream &EatLine(istream &is) {
  char ch;

  while (is.get(ch)) {
    // too far?
    if (ch == '\n') {
      break;
    }
  }
  return is;
}


// ****************************************************************************
//  Name: lg
//  Date: 06/17/1995 : mdb - Created.
// ****************************************************************************
/** 
 * Find log base two value.
 * @param x - number to find log of
 * @return double - log
 */
double lg(double x) {
  return log(x) / log(2.0);
}


// ****************************************************************************
//  Name: NumberBits
//  Date: 06/17/1995 : mdb - Created.
// ****************************************************************************
/**
 * Find number of bits needed to represent the given number.
 * @param x - number to test
 * @return int - number of bits
 */
int NumberBits(long x) {
  return (int) ceil(lg(x+1));
}


// ****************************************************************************
//  Name: NumberDecimalDigits
//  Date: 07/11/1995 : mdb - Created.
// ****************************************************************************
/**
 * Find number of decimal digits needed to represent the given number.
 * @param x - number to test
 * @param int - number of digits
 */
int NumberDecimalDigits(long x) {
  if (x < 9) {
    return 1;
  }
  return (int) ceil(log10(static_cast(double,x+1)));
}


// ****************************************************************************
//  Name: IsPrime
//  Date: 07/03/1995 : mdb - Created.
// ****************************************************************************
/**
 * Determine if an integer is prime.
 * @param wN - number to test
 * @return bool - T=prime, F=not prime
 */
bool IsPrime(long wN) {
  bool bRtn = true;

  if (wN < 0) {
    wN = -wN;  // force positive
  }

  if (wN <= 2) {
    bRtn = true;  // arbitrarily define these
  } else if ((wN % 2) == 0) {
    bRtn = false;
  } else {
    long wMax = (long) floor((sqrt(static_cast(double,wN))));
    for (long w = 3; w <= wMax; w+=2) {
      if ((wN % w) == 0) {
	bRtn = false;
	break;
      }
    } // for
  }
  return bRtn;
}


// ****************************************************************************
//  Name: Atoi
//  Date: 06/05/1997 : mdb - Created.
//        06/25/1997 : mdb - Changed to support non 1024 bases, and removed
//                           code to copy the string first, since atoi() seems
//                           smart enough to ignore the suffix.
// ****************************************************************************
/**
 * Convert from an ascii string to a number.
 * ###kb, ###mb, ###gb - (1024, 1024^2, 1024^3) bytes
 * ###k,  ###m,  ###g  - (10^3, 10^6, 10^9) number
 * @param sb - string to convert
 * @return long - value converted
 */
long Atoi(const char *sb, long wDefault) {
  long wRtn;
  if (!sb) return wDefault;
  int len = (int)strlen(sb);
  if (len > 2 && ((sb[len-2] == 'G' || sb[len-2] == 'g') &&
		  (sb[len-1] == 'B' || sb[len-1] == 'b'))) {
    wRtn = atoi(sb) * GB_1;
  } else if (len > 1 && (sb[len-1] == 'G' || sb[len-1] == 'g')) {
    wRtn = atoi(sb) * G_1;
  } else if (len > 2 && ((sb[len-2] == 'M' || sb[len-2] == 'm') &&
		  (sb[len-1] == 'B' || sb[len-1] == 'b'))) {
    wRtn = atoi(sb) * MB_1;
  } else if (len > 1 && (sb[len-1] == 'M' || sb[len-1] == 'm')) {
    wRtn = atoi(sb) * M_1;
  } else if (len > 2 && ((sb[len-2] == 'K' || sb[len-2] == 'k') &&
			 (sb[len-1] == 'B' || sb[len-1] == 'b'))) {
    wRtn = atoi(sb) * KB_1;
  } else if (len > 1 && (sb[len-1] == 'K' || sb[len-1] == 'k')) {
    wRtn = atoi(sb) * K_1;
  } else {
    wRtn = atoi(sb);
  }
  return wRtn;
}


// ****************************************************************************
//  Name: Atof
//  Date: 06/16/2001 : mdb - Created.
// ****************************************************************************
/**
 * Convert from an ascii string to a floating point number.
 * @param sb - string to convert
 * @return double - value converted
 */
double Atof(const char *sb, double rDefault) {
  if (!sb) return rDefault;
  return atof(sb);
}


// ****************************************************************************
//  Name: CalcBytes
//  Date: 06/05/1997 : mdb - Created.
//        12/29/1999 : mdb - Added optional scratch buffer to make MT-safe.
// ****************************************************************************
/**
 * Given a number of bytes, generate a volume string.
 * @param wBytes - used to calculate bytes
 * @return char* - string representation of bytes (This is *not* new'ed space, and may be overwritten.)
 */
char *CalcBytes_r(long wBytes, char *sbTmp) {
  double d; 

  if ((d = double(wBytes) / double(MB_1)) >= double(1)) {
    sprintf(sbTmp, "%.2f MB", d);
  } else if ((d = double(wBytes) / double(KB_1)) >= double(1)) {
    sprintf(sbTmp, "%.2f KB", d);
  } else {
    d = double(wBytes);
    sprintf(sbTmp, "%.0f bytes", d);
  }

  return sbTmp;
}
char *CalcBytes(long wBytes) {
  return CalcBytes_r(wBytes, sbBuffer);
}


// ****************************************************************************
//  Name: CalcRate
//  Date: 06/05/1997 : mdb - Created.
//        12/29/1999 : mdb - Added optional scratch buffer to make MT-safe.
// ****************************************************************************
/**
 * Given a number of bytes and time, generate a rate string.
 * @param wBytes - used to calculate rate
 * @param dTime - used to calculate rate
 * @return char* - string representation of rate (This is *not* new'ed space, and may be overwritten.)
 */
char *CalcRate_r(long wBytes, double dTime, char *sbTmp) {
  double d; 

  if (dTime > 0) {
    if ((d = (double(wBytes) / double(MB_1)) / dTime) >= double(1)) {
      sprintf(sbTmp, "%.2f MB/sec", d);
    } else if ((d = (double(wBytes) / double(KB_1)) / dTime) >= double(1)) {
      sprintf(sbTmp, "%.2f KB/sec", d);
    } else {
      d = double(wBytes) / dTime;
      sprintf(sbTmp, "%.2f bytes/sec", d);
    }
  } else {
    sprintf(sbTmp, "NaN bytes/sec");
  }

  return sbTmp;
}
char *CalcRate(long wBytes, double dTime) {
  return CalcRate_r(wBytes, dTime, sbBuffer);
}


// ****************************************************************************
//  Name: CalcPercent
//  Desc: Given a number and a total, generate a percent (rounded to an int).
//  Date: 02/06/1997 : mdb - Created.
//
//  Inputs: wNum, wTotal - number and total to calc percent with
//  Return: int - percent
// ****************************************************************************
/**
 * Given a number and a total, generate a percent (rounded to an int).
 * @param wNum - number and total to calc percent with
 * @param wTotal - number and total to calc percent with
 * @return int - percent
 */
int CalcPercent(int wNum, int wTotal) {
  return wTotal == 0 ? 0 : (int)((double)wNum/(double)wTotal*100.0);
}


// ****************************************************************************
