// ****************************************************************************
//  Name: dc_str.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 07/28/1997 : mdb - Created.
//        10/25/2000 : mdb - ctor now handles null str.
//        06/16/2001 : mdb - Changed semantics of token, to not reset itself
//                           automatically after NULL is returned.  User must
//                           manually call TokenReset.
//  Desc: Supporting string routines.
// ****************************************************************************

#include "dc_standard.h"
#include "dc_str.h"


Str::Str(void) {
  sb = NULL;
  wLen = wMax = 0;
  sbTmp = NULL;
}

Str::Str(const char *sbNew) {
  if (sbNew) {
    sb = StrDup(sbNew);
    wLen = (int)strlen(sbNew);
    wMax = wLen + 1;
  } else {
    sb = NULL;
    wLen = wMax = 0;
  }
  sbTmp = NULL;
}

Str::~Str(void) {
  TestDelete(sb);
}


// ****************************************************************************
//  Name: Set
//  Date: 07/30/1997 : mdb - Created.
//        06/16/2001 : mdb - Fixed bug forgetting to set sbTmp for token.
// ****************************************************************************
/**
 * (re)init to the given string
 * @param sbNew - string to init
 */
char *Str::Set(const char *sbNew) {
  TestDelete(sb);

  if (sbNew) {
    sb = StrDup(sbNew);
    wLen = (int)strlen(sbNew);
    wMax = wLen + 1;
  } else {
    sb = NULL;
    wLen = wMax = 0;
  }
  sbTmp = NULL;
  return sb;
}
char *Str::SetNoDup(char *sbNew) {
  TestDelete(sb);
  sb = sbNew;  // don't dup this, just take ownership of it, and free on dtor
  wLen = (int)strlen(sbNew);
  wMax = wLen + 1;
  sbTmp = NULL;
  return sb;
}


// ****************************************************************************
//  Name: Steal
//  Date: 11/08/2000 : mdb - Created.
// ****************************************************************************
/**
 * Take the contents of the allocated memory, and reset str to NULL.
 */
char *Str::Steal(void) {
  char *sbRtn = sb;
  if (sb) {
    sb = NULL;
    wLen = wMax = 0;
    sbTmp = NULL;
  }
  return sbRtn;
}


// ****************************************************************************
//  Name: Cat,NCat
//  Date: 07/30/1997 : mdb - Created.
//        04/27/1998 : mdb - Added NCat().
// ****************************************************************************
/**
 * strcat a string to the existing string.  (alloc as needed)
 * @param sbAdd - string to add
 */
char *Str::Cat(char chAdd) {
  char sbAdd[2];
  sbAdd[0] = chAdd;
  sbAdd[1] = '\0';
  return NCat(sbAdd, 1);
}

char *Str::Cat(char *sbAdd) {
  int wLenAdd = (int)strlen(sbAdd);
  return NCat(sbAdd, wLenAdd);
}

char *Str::NCat(char *sbAdd, int wLenAdd) {
  char *sbOld;

  // expand storage?
  if (wLen + wLenAdd + 1 > wMax) {
    wMax = wLen + wLenAdd + 1;
    sbOld = sb;
    sb = new char [wMax];
    if (sbOld) {
      strcpy(sb, sbOld);
      delete [] sbOld;
    } else {
      sb[0] = '\0';
    }
  }

  // now append
  strncat(sb, sbAdd, wLenAdd);
  wLen += wLenAdd;

  return sb;
}


// ****************************************************************************
//  Name: Replace
//  Date: 04/27/1998 : mdb - Created.
// ****************************************************************************
/** 
 * For all occurences of a char, replace it with another char.
 * Note that a zero byte is not allowed for either.
 * (no alloc needed)
 * @param chOld - char to look for
 * @param chNew - char to replace with
 */
char *Str::Replace(char chOld, char chNew) {
  char *p;

  if (!sb) return sb;
  if (chOld == '\0') return sb;  // not allowed
  if (chNew == '\0') return sb;  // not allowed

  for (p=sb; *p; p++) {
    if (*p == chOld) *p = chNew;
  }
  return sb;
}


// ****************************************************************************
//  Name: Replace
//  Date: 07/30/1997 : mdb - Created.
// ****************************************************************************
/**
 * For all occurences of a substring, replace it with another string.
 * (alloc as needed)
 * @param sbOld - string to look for
 * @param sbNew - string to replace with
 */
char *Str::Replace(char *sbOld, char *sbNew) {
  int wLenOld = (int)strlen(sbOld);
  int wLenNew = (int)strlen(sbNew);
  char *p, *s, *t, *sbResult;
  int   i, wLenResult, wMaxResult;

  if (!sb) return sb;

  // count how many replacements will be made
  for (i=0, p=sb; (p=strstr(p, sbOld)); i++) {
    p += wLenOld;
    if (p > sb+wLen) break;  // past eos
  }
  if (i==0) return sb;  // nothing to do

  wLenResult = wLen + i * (wLenNew - wLenOld);
  wMaxResult = wLenResult + 1;
  sbResult = new char[wMaxResult];

  p=strstr(sb, sbOld);
  s=sb;
  t=sbResult;
  while (1) {
    if (p) {
      if (p == s) {  // do a replacement here
	strncpy(t, sbNew, wLenNew);
	s += wLenOld;
	t += wLenNew;
	if (s > sb+wLen) break;  // past eos
	p = strstr(s, sbOld);
      } else {  // copy chars from orig str up to next replacement
	i = (int)(p-s);
	strncpy(t, s, i);
	s += i;
	t += i;
      }
    } else {
      // no more replacements to make - copy chars to end of orig str
      i = wLen - (int)(s-sb);
      strncpy(t, s, i);
      t += i;
      *t = '\0';
      break;
    }
  }
  
  TestDelete(sb);
  sb = sbResult;
  wLen = wLenResult;
  wMax = wMaxResult;
  return sb;
}


// ****************************************************************************
//  Name: ReplaceEnvVars
//  Date: 07/30/1997 : mdb - Created.
//        01/08/2000 : mdb - Changed to allow expansion of nested env vars.
// ****************************************************************************
/**
 * Look for ${NAME} patterns, and check to see if NAME is an envvar.
 * Replace with the envvar value if found, else do nothing.
 */
char *Str::ReplaceEnvVars(void) {
  char *p, *q;
  char *sbFullName, *sbName;
  char *sbVal;

  if (!sb) return sb;

  p = sb;
  while (1) {
    if ((p = strrstr(p, "${")) && (q = strchr(p, '}'))) {
      sbFullName = StrNDup(p, (int)(q-p+1));
      sbName = StrNDup(p+2, (int)(q-p-2));
      if ((sbVal = getenv(sbName))) {
	Replace(sbFullName, sbVal);
	p = sb;
      } else {
	p += 2;  // skip ahead if not an env var
      }
      delete [] sbFullName;
      delete [] sbName;
    } else {
      break;  // no more
    }
  }
  return sb;
}


// ****************************************************************************
//  Name: RChop
//  Date: 08/07/1997 : mdb - Created.
// ****************************************************************************
/**
 * Chop off the last delimeted token.
 * @param sbDelim - set of possible chars to use as the chop point
 */
char *Str::RChop(char *sbDelim) {
  char *p;

  if (!sb) return sb;

  // skip past last token
  for (p=sb + wLen - 1; p>=sb && !strchr(sbDelim, *p); p--)
    ;
  // eat extra delimiters
  for ( ; p>=sb && strchr(sbDelim, *p); p--)
    ;
  // chop string
  p[1] = '\0';
  wLen = (int)(p - sb + 1);

  return sb;
}

// ****************************************************************************
//  Name: RChopCh
//  Date: 04/27/1998 : mdb - Created.
// ****************************************************************************
/**
 * Chop off the last character if it matches the given char.
 * @param chChop - char to chop off the right end
 */
char *Str::RChopCh(char chChop) {
  char *p;

  if (!sb) return sb;

  p = sb + wLen - 1;
  if (p>=sb && *p == chChop) {
    *p = '\0';
    wLen--;
  }

  return sb;
}



// ****************************************************************************
//  Name: _Token_Internal
//  Date: 02/25/1998 : mdb - Created.
// ****************************************************************************
/**
 * Scans to get the next token.
 * @param sbDelim - indicate start/end of tokens
 * @param wMaxSize - max allowable token length (-1=no limit)
 * @param fNoDup - true=ok to chop up the str and not allocate space, false=do a dup copy of the token
 * @param fCount - true=don't do anything except advance sbTmp ptr, false=do real processing
 * @return char* - ptr to allocated space holding token, NULL=none left
 */
char *Str::_Token_Internal(char *sbDelim, int wMaxSize,
			   bool fNoDup, bool fCount) {
  char *sbStart = NULL;
  int   wSize = 0;

  if (!sbTmp) sbTmp = sb;  // first call?
  if (!sbTmp) return NULL;
  if (wMaxSize == -1) wMaxSize = wLen;
  while (wSize < wMaxSize && *sbTmp != '\0') {
    if (strchr(sbDelim, *sbTmp) != NULL) {
      if (!sbStart) {
        sbTmp++;
        continue;
      }
      break;
    }
    if (!sbStart) sbStart = sbTmp;
    sbTmp++;
    wSize++;
  }

  if (wSize == 0) {
    // we do *not* set sbTmp = NULL here -- user calls TokenReset() if needed.
    return NULL;
  }

  if (fCount) return sbStart;
  if (fNoDup) {
    if (*sbTmp != '\0') {  // middle of the string?
      *sbTmp = '\0';
      sbTmp++; // so the next StrTok will be past the added null
    }
    return sbStart;
  } else {
    return StrNDup(sbStart, wSize);
  }
}


// ****************************************************************************
//  Name: TokenNum
//  Date: 10/16/2000 : mdb - Created.
//        06/16/2001 : mdb - Fixed bug to lessen impact on current Token state.
// ****************************************************************************
/**
 * Scan for all possible tokens (from the current token position),
 * and return count.
 * @param sbDelim - indicate start/end of tokens
 * @param wMaxSize - max allowable token length (-1=no limit)
 * @return int - count of tokens found
 */
int Str::TokenNum(char *sbDelim, int wMaxSize) {
  int nRtn=0;
  char *sbTmpSave = sbTmp;
  while (_Token_Internal(sbDelim, wMaxSize, false, true) != NULL) {
    ++nRtn;
  }
  sbTmp = sbTmpSave;  // restore Token() state to where it was
  return nRtn;
}


// ****************************************************************************
//  Name: TokenArray
//  Date: 04/03/2000 : mdb - Created.
// ****************************************************************************
/**
 * Splits a string into tokens, and returns an array of StrDup() copies
 * @param sbDelim  - indicate start/end of tokens
 * @param wMaxSize - max allowable token length (-1=no limit)
 * @return DArray<char*>* - ptr to alloc DArray of alloc tokens
 */
DArray<char*> *Str::TokenArray(char *sbDelim, int wMaxSize) {
  DArray<char*> *pdaRtn = new DArray<char*>;
  char *sb;

  while ((sb = Token(sbDelim, wMaxSize)) != NULL) {
    pdaRtn->Append(sb);
  }
  return pdaRtn;
}

void Str::FlushTokenArray(DArray<char*> *pda) {
  for (unsigned int i=0; i<pda->NumElem(); ++i) {
    TestDeleteNull(pda->Get(i));
  }
}

DArray<char*> *Str::TokenArrayNoDup(char *sbDelim, int wMaxSize) {
  DArray<char*> *pdaRtn = new DArray<char*>;
  char *sb;

  while ((sb = TokenNoDup(sbDelim, wMaxSize)) != NULL) {
    pdaRtn->Append(sb);
  }
  return pdaRtn;
}


// ****************************************************************************
//  Name: TokenGet
//  Date: 12/12/2001 : mdb - Created.
// ****************************************************************************
// returns the token i ahead of the current position
/**
 * Returns the i'th token ahead of the current position.  There is not
 * a NoDup version, because that could cause this function to be non-
 * repeatable over many calls.
 * @param i - index of the token to return (0..k-1)
 * @param sbDelim - indicate start/end of tokens
 * @param wMaxSize - max allowable token length (-1=no limit)
 * @return char* - ptr to alloc token
 */
char *Str::TokenGet(int i, char *sbDelim, int wMaxSize) {
  char *sb;
  char *sbTmpSave = sbTmp;
  for (int j=i; j>0; --j) {  // skip all but last token
    if (_Token_Internal(sbDelim, wMaxSize, false, true) == NULL) break;
  }
  sb = _Token_Internal(sbDelim, wMaxSize, false, false);
  sbTmp = sbTmpSave;  // restore Token() state to where it was
  return sb;
}

ostream &operator<<(ostream &os, Str &str) {
  return os << str.sb;
}


// ****************************************************************************
