// ****************************************************************************
//  Name: dc_threadclass.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 02/25/2000 : mdb - Created loosely based on newsgroup discussion.
//                         - First impl where thread created in ctor and calls
//                           a virt fcn was incorrect because virt fcs cannot
//                           be called in ctor.
//        09/29/2000 : mdb - Cleaned up a bit, and removed unused items.
//        11/16/2000 : mdb - Added ThreadPool and ThreadWorkerClass.
//        04/23/2001 : mdb - Moved wrappers to dc_pthread-wrappers.h
//        
//  Desc: Implementation of a c++ thread class.
//        Both classes require a thread to be a seperate object, and each
//        object has only one run() call.
// ****************************************************************************
#include "dc_threadclass.h"
#include "dc_pthread-wrappers.h"

#include <iostream>
#include <errno.h>
#include <time.h>

#include "dc_debug.h"

#ifdef _WIN32
#include <windows.h>
#include "win32.h"
#endif

using namespace std;


// ****************************************************************************
//  ThreadClass (similar to Java style)
// ****************************************************************************

typedef struct {
    ThreadClass *pObj;
    void        *pArg;
    
    // used for barrier
    bool     fSignal;
    DC_Mutex *pmutex;
    DC_Cond  *pcv;
} ThreadClass_StartArg_t;

/**
 * actual C linkage routine called by pthreads
 */
extern "C" {
static void *ThreadClass_Start(void *p_in)
{
    ThreadClass_StartArg_t *p = (ThreadClass_StartArg_t*)p_in;
    if (p->fSignal) {
#if defined(DCDBG_THREADCLASS)
        std::cout << "ThreadClass_Start(): in C linkage wrapper - barrier" << std::endl;
#endif /* DCDBG_THREADCLASS */
        p->pmutex->lock();
        p->fSignal = false;
        p->pcv->signal();  // must be done while holding mutex
        p->pmutex->unlock();
        // mutex and cv are now invalid because caller will destroy them
    }
#if defined(DCDBG_THREADCLASS)
    std::cout << "ThreadClass_Start(): in C linkage wrapper - method call" << std::endl;
#endif /* DCDBG_THREADCLASS */
    void *pRtn = (p->pObj)->thread_run(p->pArg);
    delete p;
    return pRtn;
}
}

/**
 * called by external thread to create a new thread that will call run()
 */
int ThreadClass::thread_create(pthread_attr_t *pattr, void *pArg,
			       bool fBarrier)
{
    int ret;
    pthread_attr_t attr;
    bool fAttrInit = false;
    
    if (pattr == NULL) {
        if (fDetach) { // by default pthread_create creates "joinable" if attr is null
            // so if we want detached; make sure we create an attr and set "detached"
            if ((ret=pthread_attr_init(&attr)) != 0) return ret;  // error
            pattr = &attr;
            fAttrInit = true;        
            if ((ret=pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED))!= 0) {
                pthread_attr_destroy(&attr);
                return ret;
            }            
        }
    }
    
    // build arg to pass to the new thread
    ThreadClass_StartArg_t *p = new ThreadClass_StartArg_t;
    p->pObj = this;
    p->pArg = pArg;
    p->fSignal = fBarrier;
    if (fBarrier) {
        p->pmutex = new DC_Mutex;
        p->pcv = new DC_Cond;
    }
    
#if defined(DCDBG_THREADCLASS)
    std::cout << "ThreadClass(): creating thread ..." << std::endl;
#endif /* DCDBG_THREADCLASS */
    while (1) {
        ret = Pthread_create(&tid, pattr, ThreadClass_Start, (void*)p);
        if (ret != EAGAIN)
            break;

        cout << "*** Error: ThreadClass(): thread_create == EAGAIN, sleeping ..."
             << endl;
        
#ifdef _WIN32
        DWORD dwSecs = 1000;
        Sleep(dwSecs);
#else
        struct timespec tv;
        tv.tv_sec = 1;
        tv.tv_nsec = 0;
        nanosleep(&tv, NULL);
#endif
        
    }
    if (ret == 0) {
        fStarted = true;
        if (!fAttrInit && fDetach) // we used user's attr and user wanted to run detached; so call Pthread_detach
            Pthread_detach(tid);
    }
    if (fAttrInit) {
        pthread_attr_destroy(&attr);
    }
    
    if (fBarrier) {
        // block for new thread to signal us
        p->pmutex->lock();
        while (p->fSignal) {
            p->pcv->wait(p->pmutex);
        }
        p->pmutex->unlock();
        
        // now we can destroy the sync
        delete p->pcv;
        delete p->pmutex;
    }
    return ret;
}

/**
 * called by external thread to wait until run() finished
 */
int ThreadClass::thread_join(void *status) {
  return pthread_join(tid, &status);
}

/**
 * called by new thread to detach itself (race with join() in external thr)
 */
void ThreadClass::thread_detach(void)
{
    fDetach = true;
    if (pthread_equal(tid, pthread_self()) // new thread calling
        || fStarted) {  // external thread calling after start()
        Pthread_detach(tid);
    } // else external thread calling before start() we had already set fDetach to true
}


// ****************************************************************************
//  ThreadWorkerClass
// ****************************************************************************

int ThreadWorkerClass::thread_worker_init(DC_Queue *pQ_in)
{
    pQ = pQ_in;
    return 0;
}

void *ThreadWorkerClass::thread_run(void *pArg)
{
    if (!pQ)
        return NULL;
    
  // here we continuously wait on work from the shared work Q
    void *pWork;
    while (pQ->dequeue((void**)&pWork) == DC_Queue::QUEUE_OK) {
        if (thread_worker_run(pWork))
            break;
    }
    return NULL;
}

ThreadPool::ThreadPool(int wQueueMax,
		       int nWorker_in, ThreadWorkerClass *rgWorker_in)
  : rgWorker(rgWorker_in), nWorker(nWorker_in)
{
    
    // initialize work Q
    Q.init("poolQ", wQueueMax, NULL, NULL, &cvEmptyQ);
    
    // start worker threads
    for (int i=0; i<nWorker; ++i) {
        rgWorker[i].thread_worker_init(&Q);
        rgWorker[i].thread_create();  // we assume they are not already running
    }
    
    return;
}

int ThreadPool::AddWork(void *pWork)
{
    return Q.enqueue(pWork);
}

int ThreadPool::WaitWork(void)
{
    // wait for work to be finished
    Q.lockInternal();
    while (Q.probe(false) > 0) {
        cvEmptyQ.wait(Q.getInternal());
    }
    Q.unlockInternal();
    return 0;
}

int ThreadPool::Shutdown(void)
{
    // forcibly shutdown queue
    Q.shutdown();
    
    // wait for all workers to exit
    for (int i=0; i<nWorker; ++i) {
        rgWorker[i].thread_join(NULL);
    }
    return 0;
}

// ****************************************************************************
