// -*- c++ -*-
// ****************************************************************************
//  Name: dc_threadclass.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 02/25/2000 : mdb - Created loosely based on newsgroup discussion.
//        11/16/2000 : mdb - Added ThreadPool and ThreadWorkerClass.
//        04/23/2001 : mdb - Moved wrappers to dc_pthread-wrappers.h
//  Desc: Implementation of a c++ thread class.
// ****************************************************************************
/** @file dc_threadclass.h */

#if !defined(H_THREADCLASS)
#define H_THREADCLASS

#include "../common-config.h"

#include "dc_cond.h"
#include "dc_queue.h"


// ****************************************************************************
//  Name: ThreadClass
// ****************************************************************************
/// Interface for spawning threads (similar to Java)
/**
 * This design allow only one thread to be created per derived object
 * due to the class level state.
 */
class ThreadClass {
private:
  bool fStarted, fDetach;
public:
  pthread_t tid;

  ThreadClass() : fStarted(false), fDetach(false) {};
  virtual ~ThreadClass() {};
   
  // called by external thread to create a new thread that will call run()
  int thread_create(pthread_attr_t *pattr=NULL, void *pArg=NULL,
		    bool fBarrier=false);
  // called by external thread to wait until run() finished
  int thread_join(void *status=NULL);

  // called by new thread to detach itself
  // note: this is a race if join() is called by external thread
  void thread_detach(void);

  // defined by derived class -- this is what the new thread executes
  virtual void *thread_run(void *pArg) = 0;
};

// ****************************************************************************
//  Name: ThreadPool, ThreadWorkerClass
// ****************************************************************************
/// Interface for spawning a set of related worker threads
class ThreadWorkerClass : public ThreadClass {
private:
  DC_Queue *pQ;
public:
  ThreadWorkerClass() : pQ(NULL) {};
  virtual ~ThreadWorkerClass() {};

  // called by external thread to link this worker to a work Q
  int thread_worker_init(DC_Queue *pQ_in);

  // this is taken over by this class to control waiting on the work Q
  void *thread_run(void *pArg);

  // defined by derived class -- new thread executes this when work is found
  // return: 0==continue, !0==exit this worker thread
  virtual int thread_worker_run(void *pWork) = 0;
};

/// Interface for spawning a set of related worker threads
class ThreadPool {
private:
  DC_Queue Q;
  DC_Cond cvEmptyQ;
  ThreadWorkerClass *rgWorker;
  int nWorker;
public:
  ThreadPool(int wQueueMax, int nWorker_in, ThreadWorkerClass *rgWorker_in);

  int AddWork(void *pWork);
  int WaitWork(void);

  int Shutdown(void);
};


#endif /* H_THREADCLASS */
// ****************************************************************************
