// ****************************************************************************
//  Name: dc_timer.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 06/04/1997 : mdb - Created.
//  Desc: Simple timers.
// ****************************************************************************

#include "dc_timer.h"

#ifdef _WIN32
#include <time.h>
#include "win32.h"
#include "unix_calls.h"
using namespace std;
#else
#include <unistd.h>
#endif

#include <errno.h>

#include "dc_missing.h"


// this is always being used, despite a faster wclk.o on the SP
#if !defined(AIX) || 1
/*****************************************************************************
 *  Name: w_seconds
 *  Date: 11/29/1995 : mdb - Created.
 *****************************************************************************/
/**
 * Returns the current wall clock time in seconds.
 * @return double - number of seconds
 */
double w_seconds(void) {
  struct timeval tv;
#if 1 /* claim is second arg is never used on most systems */
  int ret = gettimeofday(&tv, NULL);
#else
  struct timezone tz;
  int ret = gettimeofday(&tv, &tz);
#endif
  if (ret != 0) {
    cerr << ErrorHeader(wERROR) << "w_seconds(): gettimeofday() failed: " << endl;
    // CLH - not MT safe - << strerror(errno) << endl;
    return 0;
  }
  return (double)tv.tv_sec + (double)tv.tv_usec / (double)1000000;
}
#endif /* !AIX */


ostream &operator<<(ostream &os, struct timeval &tv) {
  os << tv.tv_sec;
  if (tv.tv_usec) {  // only output usec if non-zero
    os << '.';
    os.width(6);
    os.fill('0');
    os << tv.tv_usec;
  }
  return os;
}



// ****************************************************************************
//  Class: Timer
// ****************************************************************************


// ****************************************************************************
//  Name: Reset
//  Date: 06/04/1997 : mdb - Created.
// ****************************************************************************
/**
 * Clears any elapsed time, and start time.
 */ 
void Timer::Reset(void) {
  tvElapsed.tv_sec = tvElapsed.tv_usec = 0;  // throw away any recorded time
  tvStart.tv_sec = tvStart.tv_usec = 0;  // stop the timer
  eState = Timer::INIT;
}


// ****************************************************************************
//  Name: Start
//  Date: 06/04/1997 : mdb - Created.
// ****************************************************************************
/**
 * Starts the timer.
 */
void Timer::Start(void) {
  // if already started, overwrite the start time with this one
#if 1 /* claim is second arg is never used on most systems */
  int ret = gettimeofday(&tvStart, NULL);
#else
  struct timezone tz;
  int ret = gettimeofday(&tvStart, &tz);
#endif
  if (ret != 0) {
    cerr << ErrorHeader(wERROR) << "Timer::Start(): gettimeofday() failed: " << endl;
    // CLH - not MT safe - << strerror(errno) << endl;
  }
  eState = Timer::STARTED;
}


// ****************************************************************************
//  Name: Stop
//  Date: 06/04/1997 : mdb - Created.
// ****************************************************************************
/** 
 * Stops the timer, and calculates the elapsed time.
 */
void Timer::Stop(void) {
  struct timeval  tvStop;
  if (eState != Timer::STARTED) return;
#if 1 /* claim is second arg is never used on most systems */
  int ret = gettimeofday(&tvStop, NULL);
#else
  struct timezone tz;
  int ret = gettimeofday(&tvStop, &tz);
#endif
  if (ret != 0) {
    cerr << ErrorHeader(wERROR) << "Timer::Stop(): gettimeofday() failed: " << endl;
    // CLH - not MT safe - << strerror(errno) << endl;
    return;
  }
  tvElapsed.tv_sec += tvStop.tv_sec - tvStart.tv_sec;
  tvElapsed.tv_usec += tvStop.tv_usec - tvStart.tv_usec;

  eState = Timer::STOPPED;
}


// ****************************************************************************
//  Name: _Correction
//  Date: 06/05/1997 : mdb - Created.
//        04/07/2000 : mdb - Forgot to check for usec overflow.
// ****************************************************************************
/**
 * Fixes the elapsed time to have positive usecs.
 */
void Timer::_Correction(struct timeval &tv) {
  // lazy correction for stop_usec < start_usec
  // ex: stop time 101:0001 - start time 100:9999 = 1:-9998
  while (tv.tv_usec < 0) {
    tv.tv_usec += 1000000;
    --tv.tv_sec;
  }
  while (tv.tv_usec > 1000000) {
    tv.tv_usec -= 1000000;
    ++tv.tv_sec;
  }
}


// ****************************************************************************
//  Name: Elapsed
//  Date: 06/05/1997 : mdb - Created.
// ****************************************************************************
/**
 * If running, stops the timer, and calc elapsed time.
 */
double Timer::Elapsed(void) {
  if (eState == Timer::STARTED) Stop();
  _Correction(tvElapsed);
  return (double)tvElapsed.tv_sec + (double)tvElapsed.tv_usec/(double)1000000;
}


// ****************************************************************************
//  Name: Snapshot
//  Date: 06/05/1997 : mdb - Created.
// ****************************************************************************
/**
 * Calc elapsed time w/o stopping the timer.
 */
double Timer::Snapshot(void) {
  struct timeval  tvSnap;
  if (eState == Timer::STOPPED) return Elapsed();  // timer stopped
  if (eState != Timer::STARTED) return 0.0;  // timer not started at all  
#if 1 /* claim is second arg is never used on most systems */
  int ret = gettimeofday(&tvSnap, NULL);
#else
  struct timezone tz;
  int ret = gettimeofday(&tvSnap, &tz);
#endif
  if (ret != 0) {
    cerr << ErrorHeader(wERROR) << "Timer::Snapshot(): gettimeofday() failed: " << endl;
    // CLH - not MT safe - << strerror(errno) << endl;
    return (double)0;
  }
  tvSnap.tv_sec = tvElapsed.tv_sec + tvSnap.tv_sec - tvStart.tv_sec;
  tvSnap.tv_usec = tvElapsed.tv_usec + tvSnap.tv_usec - tvStart.tv_usec;

  _Correction(tvSnap);
  return (double)tvSnap.tv_sec + (double)tvSnap.tv_usec/(double)1000000;
}


// ****************************************************************************
//  Name: operator<<
//  Date: 06/04/1997 : mdb - Created.
// ****************************************************************************
/**
 * Outputs current elapsed time.  Does not stop the timer if running.
 */
ostream &operator<<(ostream &os, Timer &t) {
  t._Correction(t.tvElapsed);

  os << t.tvElapsed;
  // os << " seconds";
  if (t.eState == Timer::STARTED) os << " (still running)";
  return os;
}


// ****************************************************************************
//  Class: CpuDelay
// ****************************************************************************

CpuDelay::CpuDelay(int wDelayPercent, double rMaxDelay_in) {
  rDelayFactor = 1.0 + ( static_cast(double,wDelayPercent) /
			 static_cast(double,100) );
  rMaxDelay = rMaxDelay_in;
  fNoDelay = (wDelayPercent == 0);
}

void CpuDelay::Begin(void) {
  t.Reset();
  t.Start();
}

double CpuDelay::End(bool fDebug) {
  if (! fNoDelay) {
    double real_elapsed = t.Snapshot();
    double delay_elapsed = real_elapsed * rDelayFactor;
    double delay_wait = delay_elapsed - real_elapsed;
    if (rMaxDelay > 0.0 && delay_wait > rMaxDelay) {
      if (fDebug) {
	cout << "CpuDelay::End(): max delay " << rMaxDelay << " < "
	     << delay_wait << endl;
      }
      delay_elapsed = real_elapsed + rMaxDelay;
    }
    if (fDebug) {
      cout << "CpuDelay::End(): busy wait start for "
	   << delay_elapsed - t.Snapshot() << " more seconds" << endl;
    }
    while (t.Snapshot() < delay_elapsed) {
      // busy wait to emulate computation taking longer
    }
    if (fDebug) {
      cout << "CpuDelay::End(): real=" << real_elapsed
	   << ", delayed=" << delay_elapsed
	   << " (" << rDelayFactor << "x)" << endl;
    }
  }
  return t.Snapshot();
}

void CpuDelay::BusyWait(double rSeconds) {
  if (rSeconds == 0.0) return;
  Timer tTmp;
  tTmp.Reset();
  tTmp.Start();
  cout << "CpuDelay::BusyWait(): for " << rSeconds << " seconds" << endl;
  while (tTmp.Snapshot() < rSeconds) {
    // busy wait
  }
}

void CpuDelay::Sleep(double rSeconds) {
  if (rSeconds == 0.0) return;
  cout << "CpuDelay::Sleep(): for " << rSeconds << " seconds" << endl;
  ::sleep((unsigned int)rSeconds);
}



// ****************************************************************************
