// ****************************************************************************
//  Name: packet.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 08/11/1999 : mdb - Created.
//        09/16/1999 : mdb - Now use Types class.
//  Desc: Simple packet framework - can be subclassed.
// ****************************************************************************

#include "packet.h"
#include "lib/dc_machine.h"

#if defined(DCDBG)
//#define DCDBG_PACKET
#endif /* DCDBG */

// these are used in the packet header to detect "valid" packets.
static Types::uint1 magic1 = 0xaa;
static Types::uint1 magic2 = 0x55;

// ****************************************************************************
//  Name: 
//  Desc: 
//  Date: 08/11/1999 : mdb - Created.
//
//  Inputs: 
//  Return: int - 0=added ok, !0=duplicate found
// ****************************************************************************


void Packet::_EncodeHeader(Types::uint4 rgw[]) {
  rgw[0] = (magic1 << 24) | (wTag << 8) | magic2;
  rgw[1] = wExtra;
  rgw[2] = sizePayload;
}

void Packet::_DecodeHeader(Types::uint4 rgw[]) {
  wTag = (rgw[0] & 0x00ffff00) >> 8;
  wExtra = rgw[1];
  sizePayload = rgw[2];
}

bool Packet::_IsValid(Types::uint4 rgw[]) {
  Types::uint1 mymagic1, mymagic2;
  
  mymagic1 = (rgw[0] >> 24) & 0xff;
  mymagic2 = rgw[0] & 0xff;
  return (mymagic1 == magic1 && mymagic2 == magic2) ? true : false;
}

bool Packet::IsValid(void) {
  bool fValid = true;
  if (_IsValid(rgwHeader)) {
    // TODO: here we set some same endian flag
    _DecodeHeader(rgwHeader);
  } else {
    Types::uint4 rgwHeaderCopy[PACKET_HEADER_INTS];
    ConvertEndian(PACKET_HEADER_INTS, sizeof(Types::uint4), rgwHeader, rgwHeaderCopy);
    if (_IsValid(rgwHeaderCopy)) {
      // TODO: here we set some opposite endian flag
      _DecodeHeader(rgwHeaderCopy);
    } else {
      fValid = false;
    }
  }
  return fValid;
}

/**
 * when reading a packet, if pmb is NULL, we will allocate a new buf
 * of the correct size, and return it.  Otherwise, we try to use the
 * buf we have.
 */
int Packet::read(BaseSocket &sock, bool fEmptyMemoryBuf) {
  int wRtn, wTmp;

  // read the fixed size header
  if ((wRtn = sock.read(&rgwHeader, sizeof(rgwHeader))) <= 0) {
    return wRtn;
  }
  if (! IsValid()) return -2;  // this also decodes the header

#if defined(DCDBG_PACKET)
  cout << "Packet::read(fd=" << sock.getFd() << "): read " << wRtn
       << " bytes of header (tag="
       << wTag << " extra=" << wExtra << " size=" << sizePayload << ")"
       << endl;
#endif /* DCDBG_PACKET */

  if (sizePayload > 0) {
    if (pmb) {
      if (fEmptyMemoryBuf) pmb->Empty();

      // make sure the free space left in the buffer is large enough
      // to hold the payload we are going to read.
      // BUGBUG: We are not validating the payloadSize as huge or negative
      if (sizePayload > static_cast(int,pmb->getFree())) {
	cerr << ErrorHeader(wERROR) << "Packet::read(): "
	     << "Attempting to read a packet payload of length " << sizePayload
	     << " into buffer ptr=" << std::hex << pmb->getPtrFree() << std::dec
	     << ",len=" << pmb->getFree() << ".\n";
	//
	// BUGBUG: now what to do? - we really need to read the rest of the
	// packet data from the socket, or we will be out of sync with packet
	// boundaries.  Maybe we need to have a way to expand the buffer by
	// allocating a new one ?!?  Looks like a MemBuf is not enough here.
	//
	// PROPOSE: read what we can, and read the rest into temp space and
	//          discard?
      }
    } else {
      // allocate a new buffer of the needed size
#if defined(DCDBG_PACKET)
      cout << "Packet::read(): Allocating a new MemoryBuf of size "
	   << sizePayload << endl;
#endif /* DCDBG_PACKET */
      pmb = new MemoryBuf(sizePayload);
    }

    // read the payload
    if ((wTmp = sock.read(pmb->getPtrFree(), sizePayload)) <= 0) {
      return wRtn;
    }
    pmb->setSize(sizePayload + pmb->getSize());  // set new size of buf
    wRtn += wTmp;
  } else {
    // zero length payload
    if (pmb) pmb->setSize(0);
    // BUGBUG: what about !pmb case here? new a zero len memorybuf?
  }
  return wRtn;
}

/**
 * writes the used portion at the start of a buffer
 */
int Packet::write(BaseSocket &sock) {
  int wRtn, wTmp;

  sizePayload = pmb ? pmb->getSize() : 0;
  _EncodeHeader(rgwHeader);
  
  // BUGBUG: maybe use a listio call to write both the header and payload

  // write the fixed size header
  if ((wRtn = sock.write(&rgwHeader, sizeof(rgwHeader))) <= 0) {
#if defined(DCDBG_PACKET)
    cout << "\aPacket::write(fd=" << sock.getFd() << "): FAILED ret=" << wRtn
	 << " header (tag="
	 << wTag << " extra=" << wExtra << " size=" << sizePayload << ")"
	 << endl;
#endif /* DCDBG_PACKET */
    return wRtn;
  }
#if defined(DCDBG_PACKET)
  cout << "Packet::write(fd=" << sock.getFd() << "): wrote " << wRtn
       << " bytes of header (tag="
       << wTag << " extra=" << wExtra << " size=" << sizePayload << ")"
       << endl;
#endif /* DCDBG_PACKET */

  // write the payload
  if (sizePayload > 0) {
    if ((wTmp = sock.write(pmb->getPtr(), sizePayload)) <= 0) {
#if defined(DCDBG_PACKET)
      cout << "\aPacket::write(fd=" << sock.getFd() << "): FAILED ret=" << wTmp
	   << " bytes of payload\n";
#endif /* DCDBG_PACKET */
      return wRtn;
    }
#if defined(DCDBG_PACKET)
    cout << "Packet::write(fd=" << sock.getFd() << "): wrote " << wTmp
	 << " bytes of payload\n";
#endif /* DCDBG_PACKET */
    wRtn += wTmp;
  }
  return wRtn;
}


ostream &operator<<(ostream &os, Packet &pkt) {
  os << "<[";
  os << "tag=" << pkt.wTag
     << ",extra=" << pkt.wExtra
     << ",len=" << pkt.sizePayload;
#if 0
  unsigned char *p = (unsigned char*)pkt.rgwHeader;
  cout.setf(ios::showbase);
  for (unsigned int i=0; i<sizeof(pkt.rgwHeader); i++) {
    if (i) cout << ' ';
    cout << hex << (unsigned int)(p[i]);
  }
  cout.unsetf(ios::showbase);
#endif
  os << "]payload=";
  if (pkt.pmb) {
    os << '\n';
    os.flush();
    //pkt.pmb->Dump();  // *pmb should be set to exactly the payload size
    pkt.pmb->Dump(128);  // truncate overly large output
  } else {
    os << "NULL";
  }
  os << ">";
  return os;
}


// ****************************************************************************
