// -*- c++ -*-
// ****************************************************************************
//  Name: packet.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 08/11/1999 : mdb - Created.
//        09/16/1999 : mdb - Now use Types class.
//  Desc: Simple packet framework - can be subclassed.
// ****************************************************************************

#if !defined(H_DC_PACKET)
#define H_DC_PACKET

#include "lib/dc_standard.h"
#include "lib/dc_memory.h"
#include "lib/dc_socket.h"


/**
 *  This uses an 12 byte header that detects endian problems by encoding
 *  an internal magic number, a two-byte user-defined tag, and the length
 *  of the payload.
 *    header byte 0 : always 10101010  (10101010=magic number)
 *    header byte 1 : msb of short tag (for big endian layout)
 *    header byte 2 : lsb of short tag (for big endian layout)
 *    header byte 3 : always 01010101  (01010101=magic number)
 *  This forces byte 0 to be different than byte 4, so automatic endian
 *  detection and handling is possible.
 *    header byte 4-7  : extra integer (user assignable)
 *    header byte 8-11 : length integer
 */
//  BUGBUG: We are not actually handling the endian problem now, other than
//          swapping bytes to make sure a header is valid.
class Packet {
  enum { PACKET_HEADER_INTS = 3 };

protected:
  MemoryBuf *pmb;

  Types::uint4 rgwHeader[PACKET_HEADER_INTS];
  int   wTag;
  int   wExtra;
  int   sizePayload;

  void _EncodeHeader(Types::uint4 rgw[]);
  void _DecodeHeader(Types::uint4 rgw[]);
  bool _IsValid(Types::uint4 rgw[]);

public:
  Packet(void) : pmb(NULL), wTag(0), wExtra(0), sizePayload(0) {};
  Packet(MemoryBuf *pmb_in) : wTag(0), wExtra(0) { setBuf(pmb_in); };
  Packet(int wTag_in, MemoryBuf *pmb_in) : wExtra(0) { set(wTag_in, pmb_in); };
  ~Packet(void) { };

  /// NOTE: caller must have freed the prev pmb if needed
  void set(int wTag_in, MemoryBuf *pmb_in) { wTag=wTag_in; setBuf(pmb_in); };
  void setBuf(MemoryBuf *pmb_in) { pmb = pmb_in; };

  inline int getTag(void) { return wTag; };
  inline int getExtra(void) { return wExtra; };
  inline int getSizePayload(void) { return sizePayload; };
  inline MemoryBuf *getPayload(void) { return pmb; };
  inline char *getPayloadPtr(void) {
    return static_cast(char*,(pmb ? pmb->getPtr() : NULL));
  };

  inline int setTag(Types::uint2 wTag_in) { return (wTag=wTag_in); };
  inline int setExtra(Types::uint4 wExtra_in) { return (wExtra=wExtra_in); };
  inline int setSizePayload(Types::uint4 sizePayload_in) {
    return (sizePayload = sizePayload_in);
  };
  
  /// reads into the free portion at the end of a buffer
  int read(BaseSocket &sock, bool fEmptyMemoryBuf=true);
  int write(BaseSocket &sock);

  bool IsValid(void);

  friend ostream &operator<<(ostream &os, Packet &pkt);
};


#endif /* H_DC_PACKET */
// ****************************************************************************
