// ****************************************************************************
//  Name: planner.cpp
//  Athr: Michael D. Beynon (mdb) - beynon@cs.umd.edu
//  Date: 06/14/2001 : mdb - Created.
//        06/22/2001 : mdb - Reads layout ok, reads dag app layout.
//        08/26/2001 : mdb - Almost output simple single sweep schedule.
//        09/12/2001 : mdb - Added initial copies support, and fixed more
//                           assertion failures due to double epsilon values.
//        09/18/2001 : mdb - Missing nCopies in ScaleCpuTimes calculation.
//        09/21/2001 : mdb - First version of copies working, using tuning run.
//        09/22/2001 : mdb - Remove unneeded cluster level mappings.
//                         - Added PsOutput for AppNode graph.
//        10/03/2001 : mdb - Remove tune readtime from scaled extent.
//        10/05/2001 : mdb - Fixed various stupid bugs causing non-sensical
//                           schedules.
//        10/06/2001 : mdb - Moved CheckReadStalls() into ScaleCpuTimes().
//        10/15/2001 : mdb - Fixed bug in PIPE rLastData abs and timeExtent rel
//        02/24/2002 : mdb - Changes for capacity-based network model.
//  Desc: Stand-alone planning stage for filter applications.
// ****************************************************************************
//  TODO:

//  - I created USE_COMPUTENETWORKXFER so far for my changes
//  - pass into ComputeNetworkXfer() a valid timeStart, so we can populate
//    netmodel
//  - change ComputeNetworkXfer() to use Check() or Assign().
//  - Need to add a routine, similar to ComputeEST(), which will do the Assign
//    from AssignMapNode() (maybe call ComputeEST() with an Assign flag?)
// ****************************************************************************
#include <assert.h>
#include <iostream>
#include <fstream>
#include <iomanip>


#include "lib/dc_standard.h"
#include "lib/dc_debug.h"
#include "lib/dc_configinfo.h"
#include "lib/dc_regex.h"
#include "lib/dc_str.h"
#include "lib/dc_timer.h"

#include "dc_heap.h"

//#define DEBUG_MISC
//#define DEBUG_COMPUTE_EFT
//#define DEBUG_COMPUTE_EST
//#define DEBUG_COMPUTE_LB
//#define DEBUG_CAPACITYMODEL_EFT
//#define DEBUG_CAPACITYMODEL_CHECKASSIGN
//#define DEBUG_CPU_ASSIGNMENT

// Delicate hack: this assumes the planner always uses nodes in the
// same order in a homogeneous cluster, thus if we examine an idle
// node, we will never do any better, so we prune the search there.
// Of course this fails miserably with pinned nodes not being the
// first ones in the list, since we then avoid considering them.
// Also, this assumes no original load on any of the hosts, hence all
// idle nodes are the same.  This is probably best used during
// controlled debugging, where avoiding reams of debugging messages is
// a good thing.
//#define ENABLE_IDLE_NODE_PRUNE

#define USE_COMPUTENETWORKXFER

#define DOUBLE_EPSILON 0.001

struct {
  bool  fDoScheduling;
  int   wFeatures;
  char *sbFeatureString;
} Global;

// always get: producer to consumer sweep, smp sched, pipeline model
// note: values should be bitwise disjoint, for or'ing together
typedef enum {
  PLANNER_USE_ALL=1,   // instead of pruning clusters, try all clusters
  PLANNER_USE_PIPE=2,  // when computing EST, use pred EWT instead of EFT
  PLANNER_NETMODEL=4,  // network sched model
  PLANNER_COPIES_1=8,  // create copies based on tuning run assumption
  PLANNER_COPIES_2=16, // create copies iteratively with actual mapping
  PLANNER_REFINE=32    // final refinement phase to correct last hop decision
} ePlannerFeatures;

typedef enum { COLOR_WHITE, COLOR_GRAY, COLOR_BLACK } eColor_t;

//
// Overall Todo:
//
// * implement available capacity for cpu/mem/network.
// => cpu done, todo mem/network
//
// * how to adjust network capacities?  Maybe a 2nd pass after
// compute network cost during ComputeEFT()?
//
// * deal with multiple copies placed on the same node(host).
// * When mapping, add the "prev mapped host" as a special case as
// the initial min value.  That way, we won't skip this possibility
// with doing the cluster-oriented assignment.
// => this appears to work
//
// * BUGBUG: store rCycleComp time in pred pAppNode, so when computing
// EST we can deal with the amount of required data for a cycle in the
// current filter is different than the output cycle from the preds.
// * ComputeEST() has a BUGBUG describing the mismatch between the
// sender's rEWT and the amount of data needed for a local cycle at
// the receiver.
//
// * BUGBUG: much of the cycle annotations could be made more specific
// on a per-stream basis.  For now, we just normalized the cycle behavior
// to be across all streams.  For one, the instant_firstwrite time is
// across all streams, which is incorrect if one stream appears to be
// pipelined, and others of the same filter are accumulator.
//

//
//  tuning run
//
//  - assume one filter per host is used for the tuning run.
//  - assume congestion-free well-behaved interconnect
//  - assume all homogeneous nodes (probably ok to relax this and scale)
//  - collect filterexit stats using [console]filterexit_stats_file= ...
//  - strong assumption that this run is representative of future runs.
//    If not we need a classification of possible workloads, and match
//    new queries to a representative run from the classification.
//  - assume uniform activity during uow -- not bursty comp or comm
//

//
//  cpu time
//
//  - modelled as nCpu individual resources, each with a %remaining field
//

//
//  overall structure
//
//  - single graph, mutated for adding/removing copies as we partition.
//  - for copies, we still have global state for a filter to say how much
//    work in total is to be done
//  - sets?  how do we store and manipulate the sets used to represent
//    partitions?  Each appnode can only be a member of one set.  Probably
//    simplest to just add a da on each cluster.
//
//  creating copies
//  pros:
//  - reduce end-to-end time of the filter by reducing service time.
//    t' = t/num_copies
//  - this should balance the pipeline, and result in less queuing before
//    computationally intensive filters, and better overlap in the pipeline
//  cons:
//  - increasing number of copies will increase the interarrival time of
//    data buffers read by the copies.  If this gets too large the sink
//    will stall, impacting the pipelining benefits (network i/o of next
//    buffer s overlapped with the processing of the current buffer)
//
//  model of smp
//  - single virtual cpu with combined mips power
//  - use simple earliest-finish-time (eft) computation during scheduling
//

//
// NEVER-ACTUALLY-TODO: domain-internal and domain-external should be
// removed and replaced by full hostname forms in the network
// connection specs.  I may never get to this.  For now, I will edit
// the placement output by hand as needed to get it to run in our
// particular cluster environment.
//

//
//  app layout decl
//

// fwd decl
class DC_AppNode;
class DC_AppLink;
class DC_TopologyCluster;
class DC_TopologyNode;
class DC_TopologyNodeInfo;

// wrapper of pLink to include rate and volume information
class DC_AppNodeInsOuts {
public:
  DC_AppLink *pLink;  // ptr to actual stream link object
  long        wVolumeKb;
  long        wNumber;
  double      rRate;  // input or output rate

  long        wVolumeKb_Cycle;  // flow per inter-uow "cycle"

  DC_AppNodeInsOuts(void)
    : pLink(NULL), wVolumeKb(0), wNumber(0), rRate(0.0) {};
};


// shared obj among all copies of a particular app node
class DC_AppNodeShared {
public:
  int nCopies;  // total number that exist (on any host) of this appnode
  DArrayPtr<DC_AppNode*> daCopies;

  int    nCopiesMin, nCopiesMax;

  DC_TopologyNodeInfo *pNodeInfoTuning;  // filled in by topology::ProcessApp()
  double rProcessTime, rReadTime, rWriteTime, rCompTime, rInstantWriteFirst;

  int    wNumber_Cycle;  // number of inter-uow "cycle"

public:
  // ctor
  DC_AppNodeShared(void) : pNodeInfoTuning(NULL) {};
};

class DC_AppNode {
public:
  char  *sbName;  // base name of node
  int    wCopyNum;
  DC_AppNodeShared    *pShared;

  DArrayPtr<DC_AppNodeInsOuts*>  daIns;
  DArrayPtr<DC_AppNodeInsOuts*>  daOuts;

  // actual host assignment
  bool             fPinned;  // marked as pinned from layout.cfg
  DC_TopologyNode *pNodeMap;     // final mapping to a node within cluster
  int              nCopiesMap;   // final mapping #copies on this node
  double           rEST;  // earliest start time
  double           rEWT;  // earliest write time (pipeline effect)
  double           rEFT;  // earliest finish time
  double           rCycleTimeProcess;  // actual cycle time
  double           rCycleTimeComp;  // part of cycle time for computation
  double           rCycleTimeStall;  // TODO: part of cycle waiting for input
  double           rCycleRate; // cycles per sec for this copy only

  // dfs info
  DC_AppNode *pDfsPred;
  int wDfsDiscover, wDfsFinish;
  eColor_t eColor;

  // dag node characteristics:
  // t_level  = max node and link costs from any entry, up to but not
  //            including, this node.
  // b_level  = max node and link costs from this node to any exit.
  // sb_level = max node costs from this node to any exit.
  // alap     = as-late-as-possible start time that would not increase
  //            the overall schedule length.
  double rWeight;
  double t_level, b_level, sb_level, alap;

public:
  DC_AppNode(char *sbName_in)
    : sbName(StrDup(sbName_in)), wCopyNum(0), pShared(NULL),
      fPinned(false), pNodeMap(NULL), rWeight(0.0) {};
  ~DC_AppNode(void) { TestDelete(sbName); };

  void PrintFullName(ostream &os) const {
    os << sbName;
    if (pShared && pShared->nCopies > 1) os << ':' << wCopyNum;
  };
};


// This represents a single writer subset for output.  A writer can
// only appear in one of these, and a reader can appear in any number
// of these.  When copies are created and subsets are needed, this
// may be divided into two that represent the same logical stream.
class DC_AppLink {
public:
  char  *sbName;
  double rCost;
  DArrayPtr<DC_AppNode*> daSrc;
  DArrayPtr<DC_AppNode*> daDst;
  bool   fScheduled;

public:
  DC_AppLink(char *sbName_in)
    : sbName(StrDup(sbName_in)), rCost(0.0), fScheduled(false) {};
  ~DC_AppLink(void) { TestDelete(sbName); };
};


class DC_App_t : public ConfigInfo {
  friend ostream &operator<<(ostream &os, DC_App_t &app);
public:
  DArrayPtr<DC_AppNode*> daNodes;  // all nodes ([0] is entry)
  DArrayPtr<DC_AppLink*> daLinks;   // all links between nodes

  // dfs state
  DArrayPtr<DC_AppNode*> *pdaDfsFinish;  // optionally filled by dfs()

  double cp_length;  // critical-path longest path

  double rEFT;     // filled in manually from outside
  double rOptEFT;  // filled in manually from outside

protected:
  DC_AppNode *find_node(char *sbName);
  DC_AppLink *find_link(char *sbName);

  int dfs_visit(DC_AppNode *pNode, int wTime);

public:
  DC_App_t(void) : pdaDfsFinish(NULL), cp_length(0) {};
  ~DC_App_t(void) {};

  int Process(ConfigInfo &cfgTuneRun);

  int dfs(void);
  int reverse_topological_sort(DArrayPtr<DC_AppNode*> &daSorted);
  int topological_sort(DArrayPtr<DC_AppNode*> &daSorted);

  void reset_color(eColor_t eColor);

  void ComputeLevels(void);
  DC_AppNode *CreateCopy(DC_AppNode *pOrig);
  void ComputeCycleValues(DC_AppNode *pAppNode);

  double GetMakeSpan(void);

  DC_AppNodeInsOuts *FindPredInsOuts(DC_AppNode *pAppNode, DC_AppLink *pLink);
  DC_AppNodeInsOuts *FindSuccInsOuts(DC_AppNode *pAppNode, DC_AppLink *pLink);

  int PsOutput(char *sbFile);
};


//
//  topology decl
//

class DC_TopologyCluster;  // fwd decl

template <class T>
class DC_CapacityModel {
public:
  // maximum capacity values
  int    nResources;
  double rCapacityMax;  // 0..max capacity
  double rContentionPenalty;  // time interval pct increase due to contention

  // current usage state (one "list" per resource)
  //
  // BUGBUG: to be perfectly accurate, we should record this, and when
  // we go above 100% utilization, the previously assigned appnode
  // time should increase, and all decisions made during scheduling
  // that were based on it.  Therefore the EFT for each assigned
  // AppNode should be sotred here, and accessed "live" during the
  // external scheduling process.
  //
  // -> for now we just assume the only appnode slowed down by >100%
  // util is the new one, and compute its EFT accordingly.  Had we
  // implemented our own thread scheduler, this could be possible.
  //
  class MemberDetail {
  public:
    // this was DC_AppNode *pAppNode, but was changed to allow DC_AppLink
    T          *pMemberObj;
    //DC_AppNode *pAppNode;
    double      rCapacityUse;
  };
  class CapacityTimelineEntry {
  public:
    double timeStart;
    double timeExtent;
    double rCapacity;
    DArrayPtr<MemberDetail*> daMembers;
  };
  class CapacityTimeline {
  public:
    DArrayPtr<CapacityTimelineEntry*> da;  // variable sized list of intervals
  };
  CapacityTimeline *rgTimeline;  // one for each resource

public:
  // ctor/dtor
  DC_CapacityModel(void)
    : nResources(0), rCapacityMax(1.0), rgTimeline(NULL) {};
  ~DC_CapacityModel(void) { delete [] rgTimeline; };

  void init(int nResources_in, double rCapacityMax_in,
	    double rContentionPenalty_in) {
    nResources   = nResources_in;
    rCapacityMax = rCapacityMax_in;
    rContentionPenalty = rContentionPenalty_in;
    rgTimeline   = new CapacityTimeline[nResources];
  };

protected:
  // split the given entry into two pieces, and insert the new entry
  // after the given one.
  void _SplitEntry(int iResource, int iEntry, CapacityTimelineEntry *pentry,
		   const double timeStart) {
    CapacityTimelineEntry *pentryNew = new CapacityTimelineEntry;
    pentryNew->timeStart = timeStart;
    pentryNew->timeExtent = (pentry->timeExtent -
			     (pentryNew->timeStart - pentry->timeStart));
    assert(pentryNew->timeExtent > DOUBLE_EPSILON);
    pentryNew->rCapacity = pentry->rCapacity;
    // copy all members to new
    for (unsigned int i=0; i<pentry->daMembers.NumElem(); ++i) {
      MemberDetail *pMember = pentry->daMembers[i];
      MemberDetail *pMemberNew = new MemberDetail;
      pMemberNew->pMemberObj = pMember->pMemberObj;
      pMemberNew->rCapacityUse = pMember->rCapacityUse;
      pentryNew->daMembers.Append(pMemberNew);
    }
    rgTimeline[iResource].da.Insert(iEntry+1, pentryNew);
    pentry->timeExtent -= pentryNew->timeExtent;
    assert(pentry->timeExtent > DOUBLE_EPSILON);
    return;
  };
  // the times are assumed to be scaled to these cpus
  double _ComputeEFT(double timeStart, double timeExtent, double rCapacityUse,
		     int iResource, bool fDoAssign, T *pMemberObj) {
    assert(rCapacityUse >= 0.0 && rCapacityUse <= rCapacityMax);
    assert(timeExtent > 0.0);

    double timeBusy = timeExtent * (rCapacityUse / rCapacityMax);  // 100% time
    assert(timeBusy > 0.0 && timeBusy <= timeExtent);

#if defined(DEBUG_CAPACITYMODEL_EFT)
    cout << "_ComputeEFT("
	 << "timeStart=" << timeStart << ','
	 << "timeExtent=" << timeExtent << ','
	 << "rCapacityUse=" << rCapacityUse << ','
	 << "iResource=" << iResource << ','
	 << (fDoAssign ? "true" : "false")
	 << (pMemberObj ? pMemberObj->sbName : "NULL")
	 << ')'
	 << " timeBusy=" << timeBusy << endl;
#endif /* DEBUG_CAPACITYMODEL_EFT */

    bool fStarted = false;
    for (unsigned int i=0; i<rgTimeline[iResource].da.NumElem(); ++i) {
      CapacityTimelineEntry *pentry = rgTimeline[iResource].da.Get(i);

#if defined(DEBUG_CAPACITYMODEL_EFT)
      cout << "DEBUG: checking pentry->timeStart=" << pentry->timeStart
	   << endl;
#endif /* DEBUG_CAPACITYMODEL_EFT */

      if (!fStarted) {
	if (timeStart >= pentry->timeStart) {  // timeStart after this entry?
	  if (timeStart+DOUBLE_EPSILON <
	      pentry->timeStart+pentry->timeExtent) {
	    // timeStart is more then EPSILON before the end of the entry 
	    // found: use this entry
	  } else {
	    continue;  // skip: timeStart is past extent of this entry
	  }
	} else {
	  // error: since previous entries are contiguous, we should
	  // have hit an overlapping region by now, and fStarted
	  // should be true.
	  assert(0);
	}
      }

      // state: first encountered overlap, or continuing after first overlap
      //
      // timeStart is somewhere in inside this entry
      fStarted = true;
      if (pentry->rCapacity <= 0.0) {  // no resource capacity left in entry?
	timeStart += pentry->timeExtent;
	continue;
      }

      // state: we are using capacity from the current entry

      // contention
      //
      // ASSUME: we assume that the more threads per entry, the higher
      // the context switch cost will be, thus we increase timeBusy
      // slightly to reflect this, using a pct of the entry
      // timeExtent.  We must do this before possible splitting, so
      // that the effect is the same for both cases of fDoAssign.
#if 1
#if 0
      // 10/16/2001 : mdb - We will eventually be grouping the same filters
      //                    per node, so no reason to penalize it.
      // 10/18/2001 : mdb - It turns out this should not be done, and rather
      //                    we should encourage spread out use of nodes.
      //
      // only count entry members that are not the same filter as us
      int wTmp = 0;
      unsigned int iMember;
      for (iMember=0; iMember<pentry->daMembers.NumElem(); ++iMember) {
	// BUGBUG: after change from pAppNode -> pMemberObj, this may not work
	if (pentry->daMembers[iMember]->pMemberObj->pShared != pMemberObj->pShared)
	  ++wTmp;
      }
      double rTmp = ((pentry->timeExtent - fabs(timeStart - pentry->timeStart))
		     * rContentionPenalty * (wTmp+1));
#else
      double rTmp = ((pentry->timeExtent - fabs(timeStart - pentry->timeStart))
		     * rContentionPenalty * (pentry->daMembers.NumElem()+1));
#endif
      // longer extent, and keep utilization the same
      timeBusy = timeBusy*(timeExtent+rTmp)/timeExtent;
      timeExtent += rTmp;
#else
      timeExtent += pentry->timeExtent * rContentionPenalty;
#endif

      // split off head? -- move to non-head for use now
      if (fabs(timeStart - pentry->timeStart) > DOUBLE_EPSILON &&
	  ((pentry->timeExtent - (timeStart - pentry->timeStart))
	   > DOUBLE_EPSILON)) {
#if defined(DEBUG_CAPACITYMODEL_EFT)
	cout << "DEBUG: split head" << endl;
#endif /* DEBUG_CAPACITYMODEL_EFT */
	if (fDoAssign) {
	  _SplitEntry(iResource, i, pentry, timeStart);
	  ++i; pentry = rgTimeline[iResource].da.Get(i);  // advance loop
	  assert(fabs(timeStart - pentry->timeStart) < DOUBLE_EPSILON);
	}
      }

      // ASSUME: we can use capacity with a lower util (which will
      // take longer), but not more capacity than requested if more is
      // available.

      // actual capacity amount we can use from this entry
      double rUse = (pentry->rCapacity < rCapacityUse
		     ? pentry->rCapacity : rCapacityUse);
      // the percentage of the total capacity we can use from this interval
      double rPctCapacity = rUse / rCapacityMax;
      // the amount of some-capacity time in the interval we will use
      double timeUseCapacity = (pentry->timeExtent -
				(timeStart - pentry->timeStart));
      // dedicated resource time we can get from the interval
      double timeUse = rPctCapacity * timeUseCapacity;
#if defined(DEBUG_CAPACITYMODEL_EFT)
      cout << "DEBUG(before): i=" << i
	   << ", rUse=" << rUse
	   << ", rPctCapacity=" << rPctCapacity
	   << ", timeUseCapacity=" << timeUseCapacity
	   << ", timeUse=" << timeUse
	   << ", timeBusy=" << timeBusy
	   << ", timeStart=" << timeStart
	   << endl;
#endif /* DEBUG_CAPACITYMODEL_EFT */
      // split tail? -- stay at non-tail to use now
      if ((timeUse - timeBusy) > DOUBLE_EPSILON) {  // more than we need?
	timeUse = timeBusy;
	timeUseCapacity = timeUse / rPctCapacity;
	assert((pentry->timeExtent - timeUseCapacity) > DOUBLE_EPSILON);
#if defined(DEBUG_CAPACITYMODEL_EFT)
	cout << "DEBUG: split tail" << endl;
#endif /* DEBUG_CAPACITYMODEL_EFT */
	if (fDoAssign) {
	  _SplitEntry(iResource, i, pentry, timeStart + timeUseCapacity);
	  // no need to advance loop or update entry reference,
	  // since the current entry is the one to modify
	}
	// NOTE: we will exit in this iteration, since timeUse == timeBusy
      }
      // we use part of this entry
      if (fDoAssign) {
	pentry->rCapacity -= rUse;
	//_AddToMembers(pentry->iMembers, pMemberObj);
	MemberDetail *pMember = new MemberDetail;
	pMember->pMemberObj = pMemberObj;
	pMember->rCapacityUse = rUse;
	pentry->daMembers.Append(pMember);
      }
      timeBusy -= timeUse;
      timeStart += timeUseCapacity;
#if defined(DEBUG_CAPACITYMODEL_EFT)
      cout << "DEBUG(after): i=" << i
	   << ", rUse=" << rUse
	   << ", timeUseCapacity=" << timeUseCapacity
	   << ", timeUse=" << timeUse
	   << ", timeBusy=" << timeBusy
	   << ", timeStart=" << timeStart
	   << endl;
      Dump("end of loop", cout);
#endif /* DEBUG_CAPACITYMODEL_EFT */
      // was <= 0.0
      if (timeBusy < DOUBLE_EPSILON)
	return timeStart;  // done before end of list
    }
#if defined(DEBUG_CAPACITYMODEL_EFT)
    cout << "DEBUG: end of previous entries list"
	 << ", timeBusy=" << timeBusy
	 << ", timeStart=" << timeStart
	 << endl;
#endif /* DEBUG_CAPACITYMODEL_EFT */

    // state: we have used any pre-existing entries that had available
    // capacity, and now we need to add new entries to handle all (or
    // the remaining part) of this assignment.

    // we need a "filler" entry between the last entry (if any) and
    // the new one.  This will insure that we have a contiguous set of
    // entries from time 0 to the last entry.
    if (fDoAssign) {
      CapacityTimelineEntry *pentryNew;
      if (rgTimeline[iResource].da.NumElem() == 0) {
	if (timeStart > 0.0) {
	  // empty list filler
	  pentryNew = new CapacityTimelineEntry;
	  pentryNew->timeStart = 0.0;
	  pentryNew->timeExtent = timeStart;
	  pentryNew->rCapacity = rCapacityMax;
	  rgTimeline[iResource].da.Append(pentryNew);
#if defined(DEBUG_CAPACITYMODEL_EFT)
	  cout << "DEBUG: adding filler entry to empty list" << endl;
#endif /* DEBUG_CAPACITYMODEL_EFT */
	}
      } else {
	CapacityTimelineEntry *pentryLast = rgTimeline[iResource].da.GetLast();
	//assert(timeStart >= (entryLast.timeStart + entryLast.timeExtent));
	// past end of list filler
	if ((timeStart - (pentryLast->timeStart + pentryLast->timeExtent))
	    > DOUBLE_EPSILON) {
	  pentryNew = new CapacityTimelineEntry;
	  pentryNew->timeStart = (pentryLast->timeStart +
				  pentryLast->timeExtent);
	  pentryNew->timeExtent = timeStart - pentryNew->timeStart;
	  pentryNew->rCapacity = rCapacityMax;
	  rgTimeline[iResource].da.Append(pentryNew);
#if defined(DEBUG_CAPACITYMODEL_EFT)
	  cout << "DEBUG: adding filler from end of list" << endl;
#endif /* DEBUG_CAPACITYMODEL_EFT */
	}
      }

#if defined(DEBUG_CAPACITYMODEL_EFT)
      cout << "DEBUG: adding new entry at end of list" << endl;
#endif /* DEBUG_CAPACITYMODEL_EFT */
	  pentryNew = new CapacityTimelineEntry;
      pentryNew->timeStart = timeStart;
      pentryNew->timeExtent = timeBusy / (rCapacityUse/rCapacityMax);
      pentryNew->rCapacity = rCapacityMax - rCapacityUse;

      MemberDetail *pMember = new MemberDetail;
      pMember->pMemberObj = pMemberObj;
      pMember->rCapacityUse = rCapacityUse;
      pentryNew->daMembers.Append(pMember);

      rgTimeline[iResource].da.Append(pentryNew);
    }

    return timeStart + timeBusy / (rCapacityUse/rCapacityMax);
  };
  // in: pMemberObj - may be NULL if !fDoAssign
  double _CheckAssign(double timeStart, double timeExtent,
		      double timeBusy, bool fDoAssign, T *pMemberObj) {
#if defined(DEBUG_CAPACITYMODEL_CHECKASSIGN)
    cout << "_CheckAssign("
	 << "timeStart=" << timeStart << ','
	 << "timeExtent=" << timeExtent << ','
	 << "timeBusy=" << timeBusy << ','
	 << (fDoAssign ? "true" : "false")
	 << ')' << endl;
#endif /* DEBUG_CAPACITYMODEL_CHECKASSIGN */

    double rEFT, rMinEFT;
    int iMinResource = -1;
    for (int iResource=0; iResource<nResources; ++iResource) {
      rEFT = _ComputeEFT(timeStart, timeExtent, timeBusy, iResource, false,
			 pMemberObj);
#if defined(DEBUG_CAPACITYMODEL_CHECKASSIGN)
      cout << "_CheckAssign(): CHECK return cpu#" << iResource
	   << " -- rEFT=" << rEFT << endl;
#endif /* DEBUG_CAPACITYMODEL_CHECKASSIGN */
      if (iMinResource == -1 || rMinEFT > rEFT) {
	iMinResource = iResource;
	rMinEFT = rEFT;
      }
    }
    if (fDoAssign) {
      rEFT = _ComputeEFT(timeStart, timeExtent, timeBusy, iMinResource, true,
			 pMemberObj);
#if defined(DEBUG_CAPACITYMODEL_CHECKASSIGN)
      cout << "_CheckAssign(): ASSIGN return cpu#" << iMinResource
	   << " -- rEFT=" << rEFT << ", rMinEFT=" << rMinEFT << endl;
#endif /* DEBUG_CAPACITYMODEL_CHECKASSIGN */
#if 1
      assert(fabs(rEFT - rMinEFT) < DOUBLE_EPSILON);
#else
      assert(rEFT == rMinEFT);
#endif
    }
    return rMinEFT;
  };

public:
  // if the resource usage is assigned, return the EFT.
  inline double Check(double timeStart, double timeExtent,
		      double rCapacityUse, T *pMemberObj) {
#if defined(DEBUG_CAPACITYMODEL_CHECKASSIGN)
    cout << "*** Check("
	 << "timeStart=" << timeStart << ','
	 << "timeExtent=" << timeExtent << ','
	 << "rCapacityUse=" << rCapacityUse
	 << ')' << endl;
#endif /* DEBUG_CAPACITYMODEL_CHECKASSIGN */
    return _CheckAssign(timeStart, timeExtent, rCapacityUse,
			false, pMemberObj);
  };

  // assign the given resource usage, record the state, and return the EFT.
  inline double Assign(double timeStart, double timeExtent,
		       double rCapacityUse, T *pMemberObj) {
#if defined(DEBUG_CAPACITYMODEL_CHECKASSIGN)
    cout << "*** Assign("
	 << "timeStart=" << timeStart << ','
	 << "timeExtent=" << timeExtent << ','
	 << "rCapacityUse=" << rCapacityUse
	 << ')' << endl;
#endif /* DEBUG_CAPACITYMODEL_CHECKASSIGN */
    return _CheckAssign(timeStart, timeExtent, rCapacityUse,
			true, pMemberObj);
  };

  // remove the given appnode.  This can cause fragmented entries left behind
  inline bool UnAssign(T *pMemberObj) {
    int iResource;
    unsigned int i, j;
    bool fFound = false;
    for (iResource=0; iResource<nResources; ++iResource) {
      // look through all regions until it is found
      for (i=0; i<rgTimeline[iResource].da.NumElem(); ++i) {
	CapacityTimelineEntry *pentry = rgTimeline[iResource].da[i];
	for (j=0; j<pentry->daMembers.NumElem(); ++j) {
	  MemberDetail *pMember = pentry->daMembers[j];
	  if (pMemberObj == pMember->pMemberObj) {
	    fFound = true;
	    pentry->rCapacity += pMember->rCapacityUse;
	    pentry->daMembers.Delete(j); --j;  // remove, and continue in loop
	    delete pMember;
	  }
	}
      }
      if (fFound) break;
    }
    if (fFound) {
      // cleanup pass: remove full capacity, no member entries from the end
      while (rgTimeline[iResource].da.NumElem() > 0) {
	CapacityTimelineEntry *pentry = rgTimeline[iResource].da.GetLast();
	if (fabs(pentry->rCapacity - rCapacityMax) < DOUBLE_EPSILON &&
	    pentry->daMembers.NumElem() == 0) {
	  rgTimeline[iResource].da.DeAppend();
	  delete pentry;
	}
      }
      // TODO: cleanup pass to coalesce neighbor entries with identical members
    }
    return fFound;
  };

  bool isAllEmpty(void) {
    for (int iResource=0; iResource<nResources; ++iResource) {
      if (rgTimeline[iResource].da.NumElem() > 0) return false;
    }
    return true;
  };

  double getMaxTime(void) {
    double rMaxTime = 0.0, rTmp;
    for (int iResource=0; iResource<nResources; ++iResource) {
      if (rgTimeline[iResource].da.NumElem() == 0) continue;
      CapacityTimelineEntry *pentry = rgTimeline[iResource].da.GetLast();
      rTmp = pentry->timeStart + pentry->timeExtent;
      if (rMaxTime < rTmp) rMaxTime = rTmp;
    }
    return rMaxTime;
  };

  void Dump(char *sb, ostream &os=cout) {
    if (sb) {
      os << sb << " ...  [timeStart timeExtent rCapacity {appnode usage,...})"
	 << endl;
    }
    for (int iResource=0; iResource<nResources; ++iResource) {
      os << "cpu #" << iResource << ": ";
      for (unsigned int i=0; i<rgTimeline[iResource].da.NumElem(); ++i) {
	CapacityTimelineEntry *pentry = rgTimeline[iResource].da[i];
	os << " [" << pentry->timeStart
	   << ' ' << pentry->timeExtent
	   << ' ' << pentry->rCapacity;

	os << " {";
	for (unsigned int j=0; j<pentry->daMembers.NumElem(); ++j) {
	  MemberDetail *pMember = pentry->daMembers[j];
	  if (j > 0) os << ',';
	  if (pMember->pMemberObj) {
	    pMember->pMemberObj->PrintFullName(os);
	    os << ' ' << pMember->rCapacityUse;
	  } else {
	    os << "NULL";
	  }
	}
	os << '}';

	os << ')';
      }
      os << endl;
    }
  };

  void PsOutputUtil(ostream &os, char *sbName);
};




// wrapper class that computes utilization from 0..1, and uses CapacityModel
class DC_SmpModel : public DC_CapacityModel<DC_AppNode> {
public:
  // maximum capacity values
  int    nCpu;
  double rMips;   // per processor
  long   wMemMb;  // total

public:
  // ctor/dtor
  DC_SmpModel(void) : nCpu(0), rMips(0.0), wMemMb(0) {};
  //~DC_SmpModel(void) {};

  void init(int nCpu_in, double rMips_in, long wMemMb_in) {
    nCpu = nCpu_in;
    rMips = rMips_in;
    wMemMb = wMemMb_in;
    DC_CapacityModel<DC_AppNode>::init(nCpu, 1.0, 0.001);
  };

  // ASSUME: times passed in are already scaled to the processor type
  inline double Check(double timeStart, double timeExtent, double timeBusy,
		      DC_AppNode *pAppNode) {
    return DC_CapacityModel<DC_AppNode>
      ::Check(timeStart, timeExtent, timeBusy/timeExtent, pAppNode);
  };

  // ASSUME: times passed in are already scaled to the processor type
  inline double Assign(double timeStart, double timeExtent, double timeBusy,
		       DC_AppNode *pAppNode) {
    return DC_CapacityModel<DC_AppNode>
      ::Assign(timeStart, timeExtent, timeBusy/timeExtent, pAppNode);
  };
};


// wrapper class that computes network utilization from 0..1, and uses
// CapacityModel
class DC_NetModel : public DC_CapacityModel<DC_AppLink> {
public:
  // maximum capacity values
  double rBandwidth;  // peak link bw (guaranteed not to exceed)
  double rLatency;    // not really used in model, but stored here

public:
  // ctor/dtor
  DC_NetModel(void) : rBandwidth(0.0), rLatency(0.0) {};
  //~DC_NetModel(void) {};

  void init(double rBandwidth_in, double rLatency_in) {
    rBandwidth= rBandwidth_in;
    rLatency = rLatency_in;
    // here we use a higher contention penalty, since collisions will
    // result in retransmissions and non-linear queueing delays
    DC_CapacityModel<DC_AppLink>::init(1, 1.0, 0.0001);
  };

  inline double Check(double timeStart, double timeExtent, long wBytes,
		      DC_AppLink *pAppLink) {
    return DC_CapacityModel<DC_AppLink>
      ::Check(timeStart, timeExtent, (double)wBytes/timeExtent, pAppLink);
  };

  inline double Assign(double timeStart, double timeExtent, long wBytes,
		       DC_AppLink *pAppLink) {
    return DC_CapacityModel<DC_AppLink>
      ::Assign(timeStart, timeExtent, (double)wBytes/timeExtent, pAppLink);
  };
};


class DC_TopologyNodeInfo {
public:
  int    nCpu;
  double rMips;  // per processor
  long   wMemMb;
public:
  //DC_TopologyNodeInfo(void) {};
  //~DC_TopologyNodeInfo(void) {};

  // this is probably not needed, since no dyn alloc memory in obj
  DC_TopologyNodeInfo &operator=(const DC_TopologyNodeInfo &rhs) {
    if (this != &rhs) {
      nCpu   = rhs.nCpu;
      rMips  = rhs.rMips;
      wMemMb = rhs.wMemMb;
    }
    return *this;
  };
};


class DC_TopologyNode {
public:
  char *sbName;  // base name of node
  DC_SmpModel smp;
  DC_TopologyCluster *pCluster;  // cluster we are in
  eColor_t eColor;
public:
  DC_TopologyNode(void) : sbName(NULL), pCluster(NULL) {};
  ~DC_TopologyNode(void) { TestDelete(sbName); };
};


class DC_TopologyClusterInfo : public DC_TopologyNodeInfo {
public:
  void aggregate(const DC_TopologyClusterInfo &a) {
    nCpu += a.nCpu;
    rMips += a.rMips;
    wMemMb += a.wMemMb;
  };
  void aggregate(const DC_TopologyNodeInfo &a, const int nNodes) {
    nCpu += a.nCpu * nNodes;
    rMips += (a.rMips * a.nCpu * nNodes);
    wMemMb += a.wMemMb * nNodes;
  };
};


class DC_TopologyNetworkCost {
  friend ostream &operator<<(ostream &os, DC_TopologyNetworkCost &c);
public:
  double rBandwidth; // units: Mbps
  double rLatency;   // units: ms
public:
  DC_TopologyNetworkCost(void)
    : rBandwidth(0.0), rLatency(0.0) {};
  DC_TopologyNetworkCost(double rBandwidth_in, double rLatency_in)
    : rBandwidth(rBandwidth_in), rLatency(rLatency_in) {};
  void aggregate(const DC_TopologyNetworkCost &c) {
    // bw is that of the smallest pipe, and latency is cumulative
    rBandwidth = (rBandwidth < c.rBandwidth ? rBandwidth : c.rBandwidth);
    rLatency = rLatency + c.rLatency;
  };
};

// bus within a cluster
class DC_TopologyNodeNetwork {
public:
  char  *sbName;  // map node name to specific interconnect name
#if defined(USE_COMPUTENETWORKXFER)
  DC_NetModel net;
#else
  DC_TopologyNetworkCost cost;        // capacity
#endif
public:
  DC_TopologyNodeNetwork(void) : sbName(NULL) {};
  ~DC_TopologyNodeNetwork(void) { TestDelete(sbName); };
};

// point-to-point between clusters
class DC_TopologyClusterNetwork {
public:
  char  *sbName;  // map node name to specific interconnect name
#if defined(USE_COMPUTENETWORKXFER)
  DC_NetModel net;
#else
  DC_TopologyNetworkCost cost;        // capacity
#endif
  DC_TopologyCluster    *pDstCluster;
public:
  DC_TopologyClusterNetwork(void) : sbName(NULL) {};
  ~DC_TopologyClusterNetwork(void) { TestDelete(sbName); };
};


class DC_TopologyCluster {
  friend ostream &operator<<(ostream &os, DC_TopologyCluster &c);
public:
  char *sbName;
  bool  fNodeCluster;  // t == daNodes > 0, f == daSubClusters > 0
  DC_TopologyCluster *pClusterParent;
  int   wClusterLevel;
  
  // both: aggregate of all contained clusters, recursively
  char  *sbType;
  DC_TopologyClusterInfo ClusterInfo;  // aggregate info
  DC_TopologyClusterInfo ClusterInfoRemain;

  // container: only has sub-clusters
  DArrayPtr<DC_TopologyCluster*> daSubClusters;
  DArrayPtr<DC_TopologyClusterNetwork*> daClusterNetwork;  // cluster-level

  // nodecluster: only has nodes
  DArrayPtr<DC_TopologyNode*> daNodes;  // set of nodes in the cluster
  DArrayPtr<DC_TopologyNodeNetwork*> daNodeNetwork;  // buses within cluster
  DC_TopologyNodeInfo NodeInfo;  // for homogenous set of nodes in this cluster


  // appnodes that are assigned to this cluster
  DArrayPtr<DC_AppNode*> daAppNodeAssign;

  char *sbDomainInternal;
  char *sbDomainExternal;
public:
  DC_TopologyCluster(void)
    : sbName(NULL), sbType(NULL),
      sbDomainInternal(NULL), sbDomainExternal(NULL) {};
  ~DC_TopologyCluster(void) {
    TestDelete(sbName);
    TestDelete(sbType);
    TestDelete(sbDomainInternal); TestDelete(sbDomainExternal);
  };
};


class DC_Topology_t : public ConfigInfo {
  friend ostream &operator<<(ostream &os, DC_Topology_t &topology);
public:
  DArrayPtr<DC_TopologyCluster*> daClusters;  // all clusters
  DArrayPtr<DC_TopologyCluster*> daClusterRoot;  // cluster root entries

protected:
  int _ProcessNodeCluster(char *sbSection, char *sbNodes,
			  DC_TopologyCluster *pCluster);
  int _ProcessClusterNetwork(DC_TopologyCluster *pCluster, char *sbSection);
  int _ProcessClusterList(char *sbClusterList,
			  DC_TopologyCluster *pClusterParent,
			  int wClusterLevel);

  void _ComputeAggregateInfo_Aux(DC_TopologyCluster *pCluster);
  int _ComputeAggregateInfo(void);

public:
  DC_Topology_t(void)  {};
  ~DC_Topology_t(void) {};

  int Process(void);
  int ProcessApp(DC_App_t &app, ConfigInfo &cfgTuneRun);

  // given a hostname (foo.cs.umd.edu, or host basename (foo), find
  // the cluster it is located in.
  DC_TopologyCluster *FindHost(const char *sbName, DC_TopologyNode **ppNode=NULL);

#if defined(USE_COMPUTENETWORKXFER)
  double ComputeNetworkXfer(DC_TopologyCluster *pClusterSrc,
			    DC_TopologyCluster *pClusterDst,
			    long wVolumeKb,
			    double timeStart, double timeExtent,
			    bool fDoAssign);
#else
  DC_TopologyNetworkCost ComputeNetworkCost(DC_TopologyCluster *pClusterSrc,
					    DC_TopologyCluster *pClusterDst);
#endif
  int AssignMapNode(DC_AppNode *pAppNode, DC_TopologyNode *pNode, int nCopies,
		    double rEST, double rLastData);
  void DC_Topology_t::ScaleCpuTimes(const DC_AppNode *pAppNode,
				    const double rMipsDest,
				    double *ptimeExtent, double *ptimeBusy,
				    double rLastData);
  double ComputeEST(DC_AppNode *pAppNode, DC_TopologyCluster *pCluster,
		    DC_TopologyNode *pNode, int *pwColocation,
		    double *prLastData);
  double ComputeEFT(DC_AppNode *pAppNode, DC_TopologyCluster *pCluster,
		    DC_TopologyNode *pNode, double rEST, double rLastData);
  int ComputeNeededCopies(DC_App_t &app, DC_AppNode *pAppNode);

  int PsOutput(char *sbFile);
  int PsOutput_Aux(ofstream &fout, DC_TopologyCluster *pCluster, int wLevel);

  double ComputeLowerBound(DC_App_t &app);
  double ComputeLowerBound_Aux(DC_AppNode *pAppNode);
};


//
//  app layout defn
//

ostream &operator<<(ostream &os, DC_AppNode &n) {
  n.PrintFullName(os);
  os << "(weight="   << n.rWeight
     << ",t_level="  << n.t_level
     << ",b_level="  << n.b_level
     << ",sb_level=" << n.sb_level
     << ",alap="     << n.alap
     << ')';
  return os;
}
ostream &operator<<(ostream &os, DC_App_t &t) {
  unsigned int i, j, k;
  os << "DC_App_t{" << endl;
  os << "  daNodes.NumElem() = " << t.daNodes.NumElem() << endl;
  for (i=0; i<t.daNodes.NumElem(); ++i) {
    if (t.daNodes[i]->t_level + t.daNodes[i]->b_level == t.cp_length) {
      os << " *";
    } else {
      os << "  ";
    }
    os << "daNodes[" << i << "] = " << *t.daNodes[i] << endl;
    os << "    ins = ";
    for (j=0; j<t.daNodes[i]->daIns.NumElem(); ++j) {
      os << t.daNodes[i]->daIns[j]->pLink->sbName
	 << "(cost=" << t.daNodes[i]->daIns[j]->pLink->rCost
	 << ",num=" << t.daNodes[i]->daIns[j]->wNumber
	 << ",vol=" << t.daNodes[i]->daIns[j]->wVolumeKb
	 << ",vol_cycle=" << t.daNodes[i]->daIns[j]->wVolumeKb_Cycle	
	 << ",src=";
      for (k=0; k<t.daNodes[i]->daIns[j]->pLink->daSrc.NumElem(); ++k) {
	if (k>0) os << ',';
	os << t.daNodes[i]->daIns[j]->pLink->daSrc[k]->sbName;
      }
      os << ") ";
    }
    os << endl;
    os << "    outs = ";
    for (j=0; j<t.daNodes[i]->daOuts.NumElem(); ++j) {
      os << t.daNodes[i]->daOuts[j]->pLink->sbName
	 << "(cost=" << t.daNodes[i]->daOuts[j]->pLink->rCost
	 << ",num=" << t.daNodes[i]->daOuts[j]->wNumber
	 << ",vol=" << t.daNodes[i]->daOuts[j]->wVolumeKb
	 << ",vol_cycle=" << t.daNodes[i]->daOuts[j]->wVolumeKb_Cycle	
	 << ",dst=";
      for (k=0; k<t.daNodes[i]->daOuts[j]->pLink->daDst.NumElem(); ++k) {
	if (k>0) os << ',';
	os << t.daNodes[i]->daOuts[j]->pLink->daDst[k]->sbName;
      }
      os << ") ";
    }
    os << endl;
    if (t.daNodes[i]->pShared) {
      os << "    shared = "
	 << "(nCopies=" << t.daNodes[i]->pShared->nCopies
	 << ",tProcess=" << t.daNodes[i]->pShared->rProcessTime
	 << ",tRead=" << t.daNodes[i]->pShared->rReadTime
	 << ",tWrite=" << t.daNodes[i]->pShared->rWriteTime
	 << ",tComp=" << t.daNodes[i]->pShared->rCompTime
	 << ",num_cycle=" << t.daNodes[i]->pShared->wNumber_Cycle
	 << ')' << endl;	
    }
  }
  os << "  cp_length = " << t.cp_length << endl;
  os << "}" << endl;
  return os;
}

// compute "normalized" flow for an appnode, based on the max number
// of output stream operations.
void DC_App_t::ComputeCycleValues(DC_AppNode *pAppNode) {
  unsigned int i;
  unsigned int nins = pAppNode->daIns.NumElem();
  unsigned int nouts = pAppNode->daOuts.NumElem();
  DC_AppNodeInsOuts *pInsOuts;

  // NOTE: this does not deal with the implications of copies.  Any
  // user of the cycle values will need to deduce the impact copies
  // will have.

  // ASSUME: copies will balance rates enough to get pipelining, so we
  // only need to worry about the first data

  // normalize sub-uow cycle processing to max output stream production
#if 1 /* ***MDB*** */
  long wNumber;
  if (Global.wFeatures & PLANNER_USE_PIPE) {
    wNumber = 1;
    for (i=0; i<nouts; ++i) {
      if (wNumber < pAppNode->daOuts[i]->wNumber)
	wNumber = pAppNode->daOuts[i]->wNumber;  // max
    }
    if (nouts == 0) {
      for (i=0; i<nins; ++i) {
	if (wNumber < pAppNode->daIns[i]->wNumber)
	  wNumber = pAppNode->daIns[i]->wNumber;  // max
      }
    }
  } else {
    // if no pipelining, the cycle is the entire volume
    wNumber = 1;
  }
#else
  long wNumber = 1;
  for (i=0; i<nouts; ++i) {
    if (wNumber < pAppNode->daOuts[i]->wNumber)
      wNumber = pAppNode->daOuts[i]->wNumber;  // max
  }
#endif
  pAppNode->pShared->wNumber_Cycle = wNumber;
#if defined(DEBUG_MISC)
  cout << "ComputeCycleValues(): " << pAppNode->sbName
       << " wNumber_Cycle=" << pAppNode->pShared->wNumber_Cycle << endl;
#endif /* DEBUG_MISC */

  // computed per-cycle values
  for (i=0; i<nins; ++i) {
    pInsOuts = pAppNode->daIns[i];
    pInsOuts->wVolumeKb_Cycle = pInsOuts->wVolumeKb / wNumber;
    if (pInsOuts->wVolumeKb_Cycle == 0) pInsOuts->wVolumeKb_Cycle = 1;
  }
  for (i=0; i<nouts; ++i) {
    pInsOuts = pAppNode->daOuts[i];
    pInsOuts->wVolumeKb_Cycle = pInsOuts->wVolumeKb / wNumber;
    if (pInsOuts->wVolumeKb_Cycle == 0) pInsOuts->wVolumeKb_Cycle = 1;
  }
  return;
}


double DC_App_t::GetMakeSpan(void) {
  double rMax;
  unsigned int i;
  assert(daNodes.NumElem() > 0);
  for (i=0; i<daNodes.NumElem(); ++i) {
    DC_AppNode *pAppNode = daNodes[i];
    if (i == 0 || rMax < pAppNode->rEFT) rMax = pAppNode->rEFT;
  }
  return rMax;
}


DC_AppNode *DC_App_t::find_node(char *sbName) {
  for (unsigned int i=0; i<daNodes.NumElem(); ++i) {
    DC_AppNode *pNode = daNodes[i];
    if (strcmp(sbName, pNode->sbName) == 0) return pNode;
  }
  return NULL;
}
DC_AppLink *DC_App_t::find_link(char *sbName) {
  for (unsigned int i=0; i<daLinks.NumElem(); ++i) {
    DC_AppLink *pLink = daLinks[i];
    if (strcmp(sbName, pLink->sbName) == 0) return pLink;
  }
  return NULL;
}

DC_AppNodeInsOuts *DC_App_t::FindPredInsOuts(DC_AppNode *pAppNode,
					     DC_AppLink *pLink) {
  for (unsigned int i=0; i<pAppNode->daOuts.NumElem(); ++i) {
    DC_AppNodeInsOuts *pInsOuts = pAppNode->daOuts[i];
    if (pLink == pInsOuts->pLink) return pInsOuts;
  }
  return NULL;
}
DC_AppNodeInsOuts *DC_App_t::FindSuccInsOuts(DC_AppNode *pAppNode,
					     DC_AppLink *pLink) {
  for (unsigned int i=0; i<pAppNode->daIns.NumElem(); ++i) {
    DC_AppNodeInsOuts *pInsOuts = pAppNode->daIns[i];
    if (pLink == pInsOuts->pLink) return pInsOuts;
  }
  return NULL;
}


// dfs
int DC_App_t::dfs_visit(DC_AppNode *pNode, int wTime) {
  pNode->eColor = COLOR_GRAY;
  pNode->wDfsDiscover = ++wTime;
  for (unsigned int i=0; i<pNode->daOuts.NumElem(); ++i) {
    for (unsigned int k=0; k<pNode->daOuts[i]->pLink->daDst.NumElem(); ++k) {
      DC_AppNode *pSucc = pNode->daOuts[i]->pLink->daDst[k];
      if (pSucc->eColor == COLOR_WHITE) {
	pSucc->pDfsPred = pNode;
	dfs_visit(pSucc, wTime);
      }
    }
  }
  pNode->eColor = COLOR_BLACK;
  pNode->wDfsFinish = ++wTime;
  if (pdaDfsFinish) pdaDfsFinish->Append(pNode);  // save reverse finish order
  return 0;
}
int DC_App_t::dfs(void) {
  unsigned int i;
  for (i=0; i<daNodes.NumElem(); ++i) {  // init - mark all nodes white
    DC_AppNode *pNode = daNodes[i];
    pNode->eColor = COLOR_WHITE;
    pNode->pDfsPred = NULL;
  }
  int wTime = 0;
  for (i=0; i<daNodes.NumElem(); ++i) {
    DC_AppNode *pNode = daNodes[i];
    if (pNode->eColor == COLOR_WHITE) dfs_visit(pNode, wTime);
  }
  return 0;
}

int DC_App_t::reverse_topological_sort(DArrayPtr<DC_AppNode*> &daSorted) {
  daSorted.Flush();
  pdaDfsFinish = &daSorted;  // set internal flag, so dfs() saves finish order
  dfs();
  pdaDfsFinish = NULL;
  return 0;
}
int DC_App_t::topological_sort(DArrayPtr<DC_AppNode*> &daSorted) {
  int wRtn = reverse_topological_sort(daSorted);
  daSorted.Reverse();
  return wRtn;
}

void DC_App_t::reset_color(eColor_t eColor) {
  for (unsigned int i=0; i<daNodes.NumElem(); ++i) {
    daNodes[i]->eColor = eColor;
  }
}

void DC_App_t::ComputeLevels(void) {
  unsigned int i, j, k;
  double rMax, rMax2, rMin, rTmp;
  DArrayPtr<DC_AppNode*> daTopoList;
  DC_AppNode *pNode, *pNodePred, *pNodeSucc;
  DC_AppLink *pLink;

  // t-level
  topological_sort(daTopoList);
  for (i=0; i<daTopoList.NumElem(); ++i) {
    pNode = daTopoList[i];
    rMax = 0.0;
    for (j=0; j<pNode->daIns.NumElem(); ++j) {
      pLink = pNode->daIns[j]->pLink;
      for (k=0; k<pLink->daSrc.NumElem(); ++k) {
	pNodePred = pLink->daSrc[k];
	rTmp = pNodePred->t_level + pNodePred->rWeight + pLink->rCost;
	if (rTmp > rMax) rMax = rTmp;
      }
    }
    pNode->t_level = rMax;
  }

  // b-level, sb-level
  cp_length = 0.0;
  reverse_topological_sort(daTopoList);
  for (i=0; i<daTopoList.NumElem(); ++i) {
    pNode = daTopoList[i];
    rMax = rMax2 = 0.0;
    for (j=0; j<pNode->daOuts.NumElem(); ++j) {
      pLink = pNode->daOuts[j]->pLink;
      for (k=0; k<pLink->daDst.NumElem(); ++k) {
	pNodeSucc = pLink->daDst[k];
	rTmp = pNodeSucc->b_level + pLink->rCost;
	if (rTmp > rMax) rMax = rTmp;
	rTmp = pNodeSucc->sb_level;
	if (rTmp > rMax2) rMax2 = rTmp;
      }
    }
    pNode->b_level = pNode->rWeight + rMax;
    if (pNode->b_level > cp_length) cp_length = pNode->b_level;
    pNode->sb_level = pNode->rWeight + rMax2;
  }

  // alap
  reverse_topological_sort(daTopoList);
  for (i=0; i<daTopoList.NumElem(); ++i) {
    pNode = daTopoList[i];
    rMin = cp_length;
    for (j=0; j<pNode->daOuts.NumElem(); ++j) {
      pLink = pNode->daOuts[j]->pLink;
      for (k=0; k<pLink->daDst.NumElem(); ++k) {
	pNodeSucc = pLink->daDst[k];
	rTmp = pNodeSucc->alap - pLink->rCost;
	if (rTmp < rMin) rMin = rTmp;
      }
    }
    pNode->alap = rMin - pNode->rWeight;
  }
}


// create a new appnode object based on the given one, and link it
// into the graph as needed.
DC_AppNode *DC_App_t::CreateCopy(DC_AppNode *pNodeOrig) {
  unsigned int i;
  DC_AppNode *pNode;
  DC_AppNodeInsOuts *pInsOuts;

  pNode = new DC_AppNode(pNodeOrig->sbName);
#if 1
  // find location of this one, and insert the new one in the list
  // after the last copy to preserve order in final output (easier
  // than fixing it later)
  for (i=0; i<daNodes.NumElem(); ++i) {
    if (daNodes[i] == pNodeOrig) break;
  }
#if 1 /* ***MDB*** */
  // now find the last copy of this one
  for (; i<daNodes.NumElem(); ++i) {
    if (daNodes[i]->pShared != pNodeOrig->pShared) break;
  }
  assert(i<=daNodes.NumElem());
  daNodes.Insert(i, pNode);
#else
  assert(i<daNodes.NumElem());
  daNodes.Insert(i+1, pNode);
#endif
#else
  daNodes.Append(pNode);
#endif

  // add another copy
  pNode->pShared = pNodeOrig->pShared;
  pNode->wCopyNum = (pNode->pShared->nCopies)++;
  pNode->pShared->daCopies.Append(pNode);

  // update values because there is a new copy to divide work between
  pNode->rWeight = pNodeOrig->rWeight;

  // NOTE: for now we share the pInsOuts wrappers between copies,
  // since otherwise we would have to assume the division of work will
  // be uniform.  This is deferred until later.  Uniform volume can be
  // computed later by pInsOuts->wVolumeKb / pNode->pShared->nCopies.

  // insert new copy into existing graph
  for (i=0; i<pNodeOrig->daIns.NumElem(); ++i) {
    pInsOuts = pNodeOrig->daIns[i];

    pNode->daIns.Append(pInsOuts);
    pInsOuts->pLink->daDst.Append(pNode);
  }
  for (i=0; i<pNodeOrig->daOuts.NumElem(); ++i) {
    pInsOuts = pNodeOrig->daOuts[i];

    pNode->daOuts.Append(pInsOuts);
    pInsOuts->pLink->daSrc.Append(pNode);
  }
  return pNode;
}


// validate configinfo entries, and build in-core data structures
// - *this contains the filter layout, and cfgTuneRun contains runtime stats
int DC_App_t::Process(ConfigInfo &cfgTuneRun) {
  const int maxbuf = 80;
  char buf[maxbuf];
  char buf2[maxbuf];
  char *sb;
  unsigned int i;
  Str str;
  DC_AppNode *pNode;
  DC_AppLink *pLink;
  DC_AppNodeInsOuts *pInsOuts;
  double rProcessRemove;

  // create node and specified link objects
  sb = const_cast(char*,FirstSection());
  while (sb) {
    if (strncmp(sb, "filter.", 7) == 0) {
      pNode = new DC_AppNode(sb+7);
      pNode->pShared = new DC_AppNodeShared;
      pNode->pShared->nCopies = 1;
      pNode->pShared->daCopies.Append(pNode);

#if 1
      // readfirst is the blocking time when the pipe fills, after
      // which execution essentially begins
      rProcessRemove = Atof(cfgTuneRun.Find(sb, "process.readfirst_time"));
      pNode->pShared->rProcessTime =
	Atof(cfgTuneRun.Find(sb, "process_time")) - rProcessRemove;
#else
      pNode->pShared->rProcessTime =
	Atof(cfgTuneRun.Find(sb, "process_time"));
#endif
      pNode->pShared->rReadTime =
	Atof(cfgTuneRun.Find(sb, "process.read_time")) +
	Atof(cfgTuneRun.Find(sb, "process.readany_time"));
      pNode->pShared->rWriteTime =
	Atof(cfgTuneRun.Find(sb, "process.write_time"));
      pNode->pShared->rCompTime =
	Atof(cfgTuneRun.Find(sb, "process.computation_time"));
      assert(pNode->pShared->rCompTime > 0.0);
#if 1
      // since we are not counting readfirst as part of process, we
      // need to remove it from here too
      pNode->pShared->rInstantWriteFirst =
	Atof(cfgTuneRun.Find(sb, "process.instant_writefirst"));
      if (pNode->pShared->rInstantWriteFirst > 0.0) {
	pNode->pShared->rInstantWriteFirst -= rProcessRemove;
      }
#else
      pNode->pShared->rInstantWriteFirst =
	Atof(cfgTuneRun.Find(sb, "process.instant_writefirst"));
#endif
      assert(pNode->pShared->rInstantWriteFirst >= 0.0);
      assert(pNode->pShared->rInstantWriteFirst <=
	     pNode->pShared->rProcessTime);

      pNode->rWeight = pNode->pShared->rCompTime;

      daNodes.Append(pNode);
    } else if (strncmp(sb, "stream.", 7) == 0) {
      pLink = new DC_AppLink(sb+7);
      daLinks.Append(pLink);
    }
    sb = const_cast(char*,NextSection());
  }

  // create links, fill set of entry nodes
  for (i=0; i<daNodes.NumElem(); ++i) {
    pNode = daNodes[i];

    snprintf(buf, maxbuf, "filter.%s", pNode->sbName);
    str.Set(Find(buf, "ins"));
    while ((sb = str.TokenNoDup())) {
      if (! (pLink = find_link(sb))) {  // no [stream.*] section found earlier?
	pLink = new DC_AppLink(sb);
	daLinks.Append(pLink);
      }
      pLink->daDst.Append(pNode);
      pInsOuts = new DC_AppNodeInsOuts;
      pInsOuts->pLink = pLink;
      snprintf(buf2, maxbuf, "filter.%s.ins.%s",
	       pNode->sbName, pLink->sbName);
      pInsOuts->wVolumeKb = Atoi(cfgTuneRun.Find(buf2, "volume"))/1024;
      pInsOuts->wNumber = Atoi(cfgTuneRun.Find(buf2, "number"));
      // ASSUME:
      // total UOW volume divided by the total processing time, will give us
      // an approximation of the service rate of the filter (Kb/sec)
      pInsOuts->rRate = ( static_cast(double,pInsOuts->wVolumeKb) /
			  Atof(cfgTuneRun.Find(buf, "process_time")) );
      pNode->daIns.Append(pInsOuts);
#if 1
      // this should be in terms of time, since node weight is also time
      pLink->rCost = Atof(cfgTuneRun.Find(buf2, "write_time"));
#else
      pLink->rCost = static_cast(double,pInsOuts->wVolumeKb);
#endif
    }
    str.Set(Find(buf, "outs"));
    while ((sb = str.TokenNoDup())) {
      if (! (pLink = find_link(sb))) {  // no [stream.*] section?
	pLink = new DC_AppLink(sb);
	daLinks.Append(pLink);
      }
      pLink->daSrc.Append(pNode);
      pInsOuts = new DC_AppNodeInsOuts;
      pInsOuts->pLink = pLink;
      snprintf(buf2, maxbuf, "filter.%s.outs.%s",
	       pNode->sbName, pLink->sbName);
      pInsOuts->wVolumeKb = Atoi(cfgTuneRun.Find(buf2, "volume"))/1024;
      pInsOuts->wNumber = Atoi(cfgTuneRun.Find(buf2, "number"));
      pInsOuts->rRate = ( static_cast(double,pInsOuts->wVolumeKb) /
			  Atof(cfgTuneRun.Find(buf, "process_time")) );
      pNode->daOuts.Append(pInsOuts);
      // pLink->rCost already set on input side
    }
  }

  // validate links
  for (i=0; i<daLinks.NumElem(); ++i) {
    pLink = daLinks[i];
    if (pLink->daSrc.NumElem() == 0) {
      cout << "error: link " << pLink->sbName << " has no src nodes" << endl;
      assert(0);
    }
    if (pLink->daDst.NumElem() == 0) {
      cout << "error: link " << pLink->sbName << " has no dst nodes" << endl;
      assert(0);
    }
  }
  // validate nodes
  for (i=0; i<daNodes.NumElem(); ++i) {
    pNode = daNodes[i];
    if ((pNode->daIns.NumElem() == 0) && (pNode->daOuts.NumElem() == 0)) {
      cout << "error: node " << pNode->sbName << " has no streams" << endl;
      assert(0);
    }
  }

  // validate entry node (exactly one, and is daNodes[0])
  int nEntry = 0;
  for (i=0; i<daNodes.NumElem(); ++i) {
    if (daNodes[i]->daIns.NumElem() == 0) {
      ++nEntry;
      if (i != 0) {  // swap this with the [0] item
	pNode = daNodes[0];
	daNodes[0] = daNodes[i];
	daNodes[i] = pNode;
      }
    }
  }
  if (nEntry == 0) {
    cerr << "DC_App_t::Process(): nEntry (=" << nEntry << ") should be > 0"
	 << endl;
    abort();
  } else if (nEntry > 1) {  // add a new replacement entry node
    // NOTE: it is unclear if forcing a single entry node makes sense
    // for us.  For related work static sched techniques, it *is*
    // needed.
    cout << "DC_App_t::Process(): nEntry (=" << nEntry
	 << "), creating a single dummy entry" << endl;
    DC_AppNode *pNode = new DC_AppNode("dummy_node");
    daNodes.Insert(0, pNode);
    pNode->rWeight = 0.0;
    for (i=1; i<daNodes.NumElem(); ++i) {
      if (daNodes[i]->daIns.NumElem() == 0) {
	pLink = new DC_AppLink("dummy_link");
	pLink->rCost = 0.0;
	daLinks.Append(pLink);
	pLink->daSrc.Append(pNode);
	pInsOuts = new DC_AppNodeInsOuts;
	pInsOuts->pLink = pLink;
	pNode->daOuts.Append(pInsOuts);
	pLink->daDst.Append(daNodes[i]);
	pInsOuts = new DC_AppNodeInsOuts;
	pInsOuts->pLink = pLink;
	daNodes[i]->daIns.Append(pInsOuts);
      }
    }
  }
  return 0;
}


//
//  topology defn
//

ostream &operator<<(ostream &os, DC_TopologyNodeInfo &i) {
  return os << "ncpu=" << i.nCpu << ' '
	    << "mips=" << i.rMips << ' '
	    << "mem_mb=" << i.wMemMb;
}

ostream &operator<<(ostream &os, DC_SmpModel &smp) {
  return os << "ncpu=" << smp.nCpu << ' '
	    << "mips=" << smp.rMips << ' '
	    << "mem_mb=" << smp.wMemMb;
  // TODO: some summary of the scheduled state on the smp cpus
}

#if defined(USE_COMPUTENETWORKXFER)
ostream &operator<<(ostream &os, DC_NetModel &net) {
  return os << "bw=" << net.rBandwidth << ' '
	    << "lat=" << net.rLatency;
  // TODO: some summary of the scheduled state on the smp cpus
}
#else
ostream &operator<<(ostream &os, DC_TopologyNetworkCost &c) {
  return os << "bw=" << c.rBandwidth << ' '
	    << "lat=" << c.rLatency;
}
#endif

ostream &operator<<(ostream &os, DC_TopologyCluster &c) {
  unsigned int i;
  if (c.fNodeCluster) {
    os << "*** Node Cluster \"" << c.sbName << '\"';
  } else {
    os << "*** Collection Cluster \"" << c.sbName << '\"';
  }
  os << " level= " << c.wClusterLevel << endl;
  if (c.pClusterParent) {
    os << "parent_cluster= \"" << c.pClusterParent->sbName << '\"' << endl;
  }
  os << "aggr_cluster_info: " << c.ClusterInfo << endl;
  os << "aggr_cluster_info_remain: " << c.ClusterInfo << endl;
  if (c.fNodeCluster) {
    os << "nodes= ";
    for (i=0; i<c.daNodes.NumElem(); ++i) {
      os << c.daNodes[i]->sbName << ' '
	 << "smp{" << c.daNodes[i]->smp << '}' << endl;
    }
    os << endl;
    os << "node_info: " << c.NodeInfo << endl;
    for (i=0; i<c.daNodeNetwork.NumElem(); ++i) {
      os << "node_network[" << i << "]= "
	 << c.daNodeNetwork[i]->sbName << ' '
#if defined(USE_COMPUTENETWORKXFER)
	 << c.daNodeNetwork[i]->net
#else
	 << c.daNodeNetwork[i]->cost
#endif
	 << endl;
    }
  } else {
    os << "sub_clusters= ";
    for (i=0; i<c.daSubClusters.NumElem(); ++i) {
      os << c.daSubClusters[i]->sbName << ' ';
    }
    os << endl;
  }
  for (i=0; i<c.daClusterNetwork.NumElem(); ++i) {
    os << "cluster_network[" << i << "] "
       << c.daClusterNetwork[i]->sbName << ' '
#if defined(USE_COMPUTENETWORKXFER)
       << c.daClusterNetwork[i]->net << ' '
#else
       << c.daClusterNetwork[i]->cost << ' '
#endif
       << " (to cluster \"" << c.daClusterNetwork[i]->pDstCluster->sbName
       << "\")" << endl;
  }
  os << "domain: internal= " << c.sbDomainInternal
     << ", external= " << c.sbDomainExternal << endl;

  return os;
}

ostream &operator<<(ostream &os, DC_Topology_t &t) {
  unsigned int i;
  os << "DC_Topology_t{" << endl;
  os << "  daClusterRoot.NumElem() = " << t.daClusterRoot.NumElem() << endl;
  for (i=0; i<t.daClusterRoot.NumElem(); ++i) {
    os << "  daClusterRoot[" << i << "] = " << t.daClusterRoot[i]->sbName
       << endl;
  }
  os << "  daClusters.NumElem() = " << t.daClusters.NumElem() << endl;
  os << "}" << endl;
  for (i=0; i<t.daClusters.NumElem(); ++i) {
    os << *t.daClusters[i];
  }
  return os;
}

int DC_Topology_t::_ProcessNodeCluster(char *sbSection, char *sbNodes,
				       DC_TopologyCluster *pCluster) {
  unsigned int i;
  char *sb;
  Str str;

  pCluster->sbType = StrDup(Find(sbSection, "cpu_type"));
  pCluster->NodeInfo.nCpu = Atoi(Find(sbSection, "cpu_num"));
  pCluster->NodeInfo.rMips = Atof(Find(sbSection, "cpu_mips"));
  pCluster->NodeInfo.wMemMb = Atoi(Find(sbSection, "mem_mb"));

  str.Set(sbNodes);
  while ((sb = str.TokenNoDup()) != NULL) {
    DC_TopologyNode *pNode = new DC_TopologyNode;
    pCluster->daNodes.Append(pNode);

    pNode->sbName = StrDup(sb);
    pNode->smp.init(pCluster->NodeInfo.nCpu, pCluster->NodeInfo.rMips,
		    pCluster->NodeInfo.wMemMb);
    pNode->pCluster = pCluster;
  }

  DArray<char*> *pdaTmp = FindAll(sbSection, "network");
  for (i=0; i<pdaTmp->NumElem(); ++i) {
    DC_TopologyNodeNetwork *pNodeNetwork = new DC_TopologyNodeNetwork;
    pCluster->daNodeNetwork.Append(pNodeNetwork);
    
    str.Set(pdaTmp->Get(i));
    str.TokenNoDup();  // ignore src
    str.TokenNoDup();  // ignore dest
    pNodeNetwork->sbName = str.Token();
#if defined(USE_COMPUTENETWORKXFER)
    double rBandwidth = Atof(str.TokenNoDup());
    double rLatency = Atof(str.TokenNoDup());
    pNodeNetwork->net.init(rBandwidth, rLatency);
#else
    pNodeNetwork->cost.rBandwidth = Atof(str.TokenNoDup());
    pNodeNetwork->cost.rLatency = Atof(str.TokenNoDup());
#endif
  }
  delete pdaTmp;
  return 0;
}

int DC_Topology_t::_ProcessClusterNetwork(DC_TopologyCluster *pCluster,
					  char *sbSection) {
  unsigned int i, j, k;
  Str str;
  DC_Regex reSrc, reDst;
  char  *sbNodeNetName;
  double rBandwidth;
  double rLatency;

  // add all possible comm links between the sub-clusters according to
  // the re patterns in the network cfg entries
  DArray<char*> *pdaTmp = FindAll(sbSection, "network");
  for (k=0; k<pdaTmp->NumElem(); ++k) {
    str.Set(pdaTmp->Get(k));
    if (str.TokenNum() < 4) {
      cerr << "ERROR: invalid network entry \"" << str << '\"' << endl;
      assert(0);
    }

    reSrc.setPattern(str.TokenNoDup());
    reDst.setPattern(str.TokenNoDup());
    sbNodeNetName = str.TokenNoDup();
    rBandwidth = Atof(str.TokenNoDup());
    rLatency = Atof(str.TokenNoDup());

    for (i=0; i<pCluster->daSubClusters.NumElem(); ++i) {
      for (j=0; j<pCluster->daSubClusters.NumElem(); ++j) {
	if (i==j) continue;
	// i is src, j is dst -- look thru all network entries for connections
	if (reSrc.match(pCluster->daSubClusters[i]->sbName) &&
	    reDst.match(pCluster->daSubClusters[j]->sbName)) {
	  DC_TopologyClusterNetwork *pNetwork = new DC_TopologyClusterNetwork;
	  pCluster->daSubClusters[i]->daClusterNetwork.Append(pNetwork);

	  pNetwork->pDstCluster = pCluster->daSubClusters[j];
	  pNetwork->sbName = StrDup(sbNodeNetName);
#if defined(USE_COMPUTENETWORKXFER)
	  pNetwork->net.init(rBandwidth, rLatency);
#else
	  pNetwork->cost.rBandwidth = rBandwidth;
	  pNetwork->cost.rLatency = rLatency;
#endif
	}
      }
    }
  }
  delete pdaTmp;  
  return 0;
}

int DC_Topology_t::_ProcessClusterList(char *sbClusterList,
				       DC_TopologyCluster *pClusterParent,
				       int wClusterLevel) {
  DC_TopologyCluster *pCluster;
  char *sb;
  char  buf[80];
  Str   str;

  str.Set(sbClusterList);
  while ((sb = str.TokenNoDup()) != NULL) {
    pCluster = new DC_TopologyCluster;
    daClusters.Append(pCluster);
    if (pClusterParent) {
      pClusterParent->daSubClusters.Append(pCluster);
    } else {
      daClusterRoot.Append(pCluster);
    }
    pCluster->pClusterParent = pClusterParent;
    pCluster->sbName = StrDup(sb);
    pCluster->wClusterLevel = wClusterLevel;

    snprintf(buf, 80, "topology.cluster.%s", sb);
    if ((sb = Find(buf, "clusters"))) {
      // Collection Cluster
      pCluster->fNodeCluster = false;
      _ProcessClusterList(sb, pCluster, wClusterLevel+1);
      _ProcessClusterNetwork(pCluster, buf);
    } else if ((sb = Find(buf, "nodes"))) {
      // Node Cluster
      pCluster->fNodeCluster = true;
      _ProcessNodeCluster(buf, sb, pCluster);
    } else {
      cerr << "ERROR: cluster \"" << buf
	   << "\" missing <clusters|nodes> entry" << endl;
      exit(1);
    }
    pCluster->sbDomainInternal = StrDup(Find(buf, "domain_internal"));
    pCluster->sbDomainExternal = StrDup(Find(buf, "domain_external"));
  }
  return 0;
}

void DC_Topology_t::
_ComputeAggregateInfo_Aux(DC_TopologyCluster *pCluster) {
  if (pCluster->fNodeCluster) {
    pCluster->ClusterInfo.aggregate(pCluster->NodeInfo,
				    pCluster->daNodes.NumElem());
    pCluster->ClusterInfoRemain = pCluster->ClusterInfo;
    return;
  }
  // container cluster
  for (unsigned int i=0; i<pCluster->daSubClusters.NumElem(); ++i) {
    DC_TopologyCluster *pClusterChild = pCluster->daSubClusters[i];
    _ComputeAggregateInfo_Aux(pClusterChild);
    pCluster->ClusterInfo.aggregate(pClusterChild->ClusterInfo);
    pCluster->ClusterInfoRemain = pCluster->ClusterInfo;
  }
}
int DC_Topology_t::_ComputeAggregateInfo(void) {
  for (unsigned int i=0; i<daClusterRoot.NumElem(); ++i) {
    _ComputeAggregateInfo_Aux(daClusterRoot[i]);
  }
  return 0;
}

// validate configinfo entries, and build in-core data structures
int DC_Topology_t::Process(void) {
  _ProcessClusterList(Find("topology", "cluster_root"), NULL, 0);
  _ComputeAggregateInfo();
  return 0;
}

// check for app nodes pre-pinned to hosts (we expect at least 1, possibly 2+)
int DC_Topology_t::ProcessApp(DC_App_t &app, ConfigInfo &cfgTuneRun) {
  unsigned int i, j;
  Str   str;
  char *sb;
  const int maxbuf = 80;
  char  buf[maxbuf];
  DC_AppNode         *pAppNode, *pAppNodeTmp;
  DC_TopologyCluster *pCluster;
  DC_TopologyNode    *pNode;

  // init as "to be processed"
  app.reset_color(COLOR_BLACK);

  // process
  for (i=0; i<app.daNodes.NumElem(); ++i) {
    pAppNode = app.daNodes[i];
    if (pAppNode->eColor != COLOR_BLACK) continue;
    pAppNode->eColor = COLOR_WHITE;

    app.ComputeCycleValues(pAppNode);  // this has nothing to do all else here

    snprintf(buf, maxbuf, "filter.%s", pAppNode->sbName);

    // check which host was used for the tuning run
    pCluster = FindHost(cfgTuneRun.Find(buf, "placement"));
    assert(pCluster);

    pAppNode->pShared->pNodeInfoTuning = &pCluster->NodeInfo;

    if (Global.fDoScheduling) {
      // pinned list: break space delimited list into tokens
      // note: more than 1 pinned location implies copies
      str.Set(app.Find(buf, "pin"));
    }
    // first create all copies, so we know the total which is used
    // during mapping
    unsigned int nCopiesPinned = (unsigned int)str.TokenNum();
#if defined(DEBUG_MISC)
    cout << "Found " << nCopiesPinned << " pinned "
	 << pAppNode->sbName << endl;
#endif /* DEBUG_MISC */
    for (j=1; j<nCopiesPinned; ++j) {  // create nCopies-1 new copies
      pAppNodeTmp = app.CreateCopy(pAppNode);
      pAppNodeTmp->eColor = COLOR_WHITE;  // ignore during this pass
    }
    // now map the copies
    for (j=0; j<nCopiesPinned; ++j) {
      pAppNodeTmp = pAppNode->pShared->daCopies[j];
      pAppNodeTmp->fPinned = true;

      // find the cluster/host where this appnode is pinned
      sb = str.TokenNoDup();
      cout << "Pinned: " << pAppNodeTmp->sbName << " on " << sb << endl;

      pCluster = FindHost(sb, &pNode);
      assert(pCluster);
      assert(pCluster->fNodeCluster);

      // TODO: how to compute arg AssignMapNode(..., rEST=0.0) below
      // for non-producers?

      if (0 /* DISABLED */ && pAppNodeTmp->daIns.NumElem() == 0) {  // producer
	// num copies = proportional to nCpus to get async i/o benefit
	// BUGBUG: this does not really create copies of the filters,
	// such that the scheduling knows about it. (TODO:)
	AssignMapNode(pAppNodeTmp, pNode, 2 * pCluster->NodeInfo.nCpu,
		      0.0, 0.0);
      } else {
	AssignMapNode(pAppNodeTmp, pNode, 1, 0.0, 0.0);
      }
    }

    // copies limiting spec
    str.Set(app.Find(buf, "copies"));
    pAppNode->pShared->nCopiesMin = Atoi(str.TokenNoDup(), -1);
    pAppNode->pShared->nCopiesMax = Atoi(str.TokenNoDup(), -1);
    if (pAppNode->pShared->nCopiesMin != -1 ||
	pAppNode->pShared->nCopiesMax != -1) {
      assert(nCopiesPinned == 0);  // cannot have []pin and []copies entries
    }
  }
  return 0;
}

DC_TopologyCluster *DC_Topology_t::FindHost(const char *sbName,
					    DC_TopologyNode **ppNode) {
  int wLen;
  const char *sb;
  if (!sbName) return NULL;
  if ((sb = strchr(sbName, '.')) == 0) {  // basename
    wLen = strlen(sbName);
  } else {
    wLen = sb-sbName;
  }
  // for now, we manually walk through all clusters until we find it.
  // If this gets to be expensive, we can use a hash table.
  DC_TopologyCluster *pCluster;
  unsigned int i, j;
  for (i=0; i<daClusters.NumElem(); ++i) {
    pCluster = daClusters[i];
    if (! pCluster->fNodeCluster) continue;
    for (j=0; j<pCluster->daNodes.NumElem(); ++j) {
      if (strncmp(pCluster->daNodes[j]->sbName, sbName, wLen) == 0) {
	if (ppNode) *ppNode = pCluster->daNodes[j];
	return pCluster;
      }
    }
  }
  return NULL;  // not found
}


#if 1
#if defined(USE_COMPUTENETWORKXFER)
// rewritten version that computes reall what we are looking for --
// the EFT of a xfer starting at a particular time.  fDoAssign controls
// if these are check or assign calls.
//
// timeExtent == 0.0 means "as fast as possible", and the NetModel
// will need to handle this
double DC_Topology_t::
ComputeNetworkXfer(DC_TopologyCluster *pClusterSrc,
		   DC_TopologyCluster *pClusterDst,
		   long wVolumeKb,
		   double timeStart, double timeExtent,
		   bool fDoAssign) {
  unsigned int i, j;
  double timeDone = timeStart;  // absolute time

  // TODO: timeExtent will be needed for NetModel use
  // TODO: use Check() or Assign() instead of the static peak bw/lat values

  if (pClusterSrc == pClusterDst) {  // all in one cluster, use bus(es)
    assert(pClusterSrc->fNodeCluster);
    // ASSUME: choose highest bw bus connection
    if (pClusterSrc->daNodeNetwork.NumElem() > 0) {
      unsigned int iMin=0;
      for (i=1; i<pClusterSrc->daNodeNetwork.NumElem(); ++i) {
	if (pClusterSrc->daNodeNetwork[i]->net.rBandwidth >
	    pClusterSrc->daNodeNetwork[iMin]->net.rBandwidth)
	  iMin = i;
      }
#if 0
      if (fDoAssign) {
	call pClusterSrc->daNodeNetwork[iMin]->net.Assign();
      } else {
	call pClusterSrc->daNodeNetwork[iMin]->net.Check();
      }
#endif
      timeDone += (pClusterSrc->daNodeNetwork[iMin]->net.rLatency +
		   ((double(wVolumeKb)
		     / pClusterSrc->daNodeNetwork[iMin]->net.rBandwidth)
		    * double(8.0*1024.0/1000000.0)));
    } else {  // no nodenetwork defined, assume one host
      // use default (inf bw, 0 lat)
    }
  } else {  // between clusters, use inter-cluster links
    // look through cluster networks
    bool fFound = false;
    for (j=0; j<pClusterSrc->daClusterNetwork.NumElem(); ++j) {
      if (pClusterSrc->daClusterNetwork[j]->pDstCluster == pClusterDst) {
	fFound = true;
	break;
      }
    }
    if (fFound) {
#if 0
      if (fDoAssign) {
	call pClusterSrc->daClusterNetwork[j]->net.Assign();
      } else {
	call pClusterSrc->daClusterNetwork[j]->net.Check();
      }
#endif
      timeDone += (pClusterSrc->daClusterNetwork[j]->net.rLatency +
		   ((double(wVolumeKb)
		     / pClusterSrc->daClusterNetwork[j]->net.rBandwidth)
		    * double(8.0*1024.0/1000000.0)));
    } else {
      cout << "debug: no cluster network connection found between "
	   << pClusterSrc->sbName << " - " << pClusterDst->sbName << endl;
    }
    // this assert could fail if the heirarchy of clusters is more
    // than 1 level deep, and we need to add parts of the following
    // version of this routine
    assert(fFound);
  }
  return timeDone;
}
#else /* USE_COMPUTENETWORKXFER */
DC_TopologyNetworkCost DC_Topology_t::
ComputeNetworkCost(DC_TopologyCluster *pClusterSrc,
		   DC_TopologyCluster *pClusterDst) {
  unsigned int i, j;
  DC_TopologyNetworkCost cost(9999999.0, 0.0);

  if (pClusterSrc == pClusterDst) {  // all in one cluster, use bus(es)
    assert(pClusterSrc->fNodeCluster);
    // ASSUME: choose highest bw bus connection
    if (pClusterSrc->daNodeNetwork.NumElem() > 0) {
      unsigned int iMin=0;
      for (i=1; i<pClusterSrc->daNodeNetwork.NumElem(); ++i) {
	if (pClusterSrc->daNodeNetwork[i]->cost.rBandwidth >
	    pClusterSrc->daNodeNetwork[iMin]->cost.rBandwidth)
	  iMin = i;
      }
      cost = pClusterSrc->daNodeNetwork[iMin]->cost;
    } else {  // no nodenetwork defined, assume one host
      // use default (inf bw, 0 lat)
    }
  } else {  // between clusters, use inter-cluster links
    // look through cluster networks
    bool fFound = false;
    for (j=0; j<pClusterSrc->daClusterNetwork.NumElem(); ++j) {
      if (pClusterSrc->daClusterNetwork[j]->pDstCluster == pClusterDst) {
	cost.aggregate(pClusterSrc->daClusterNetwork[j]->cost);
	fFound = true;
	break;
      }
    }
    if (!fFound) {
      cout << "debug: no cluster network connection found between "
	   << pClusterSrc->sbName << " - " << pClusterDst->sbName << endl;
    }
    // this assert could fail if the heirarchy of clusters is more
    // than 1 level deep, and we need to add parts of the following
    // version of this routine
    assert(fFound);
  }
  return cost;
}
#endif /* USE_COMPUTENETWORKXFER */
#else /* 1 */
// this version is overkill, since the topology processing routine
// handles all peer cluster connections
DC_TopologyNetworkCost DC_Topology_t::
ComputeNetworkCost(DC_TopologyCluster *pClusterSrc,
		   DC_TopologyCluster *pClusterDst) {
  unsigned int i, j;
  DC_TopologyCluster *pCluster;
  DArray<DC_TopologyCluster*> daSrcUp;
  DArray<DC_TopologyCluster*> daDstUp;

  // we need the shortest uptree-downtree path between these two, so
  // we first compute the complete path to the root, shorten each to
  // get just the minimum part in common at the root.
  for (pCluster=pClusterSrc; ; pCluster=pCluster->pClusterParent) {
    daSrcUp.Append(pCluster);
    if (!pCluster->pClusterParent) break;
  }
  for (pCluster=pClusterDst; ; pCluster=pCluster->pClusterParent) {
    daDstUp.Append(pCluster);
    if (!pCluster->pClusterParent) break;
  }
  // prune common root entries
  while (1) {
    // invariant is daSrcUp[<last>] == daDstUp[<last>] because of how
    // we contructed them
    int nSrcUp = daSrcUp.NumElem();
    int nDstUp = daDstUp.NumElem();
    if (nSrcUp > 1 && nDstUp > 1 &&
	daSrcUp[nSrcUp-2] == daDstUp[nDstUp-2]) {
      daSrcUp.DeAppend();
      daDstUp.DeAppend();
      continue;
    }
    break;
  }
  daSrcUp.DeAppend();  // only need one copy of the common element
  daDstUp.DeAppend();  // don't need either
  // combine into a single path
  for (i=daDstUp.NumElem()-1; ; --i) {
    daSrcUp.Append(daDstUp[i]);
    if (i==0) break;
  }

  // daSrcUp contains complete path
  for (i=0; i<daSrcUp.NumElem(); ++i)
    cout << "combined daSrcUp[" << i << "] = " << daSrcUp[i]->sbName << endl;

  // compute cost
  DC_TopologyNetworkCost cost(9999999.0, 0.0);
  if (daSrcUp.NumElem() == 1) {  // all in one cluster, use bus
    assert(pClusterSrc == pClusterDst);
    if (pClusterSrc->fNodeCluster) {
      // look through buses for this cluster, and choose fastest???
      unsigned int iMin=0;
      for (i=1; i<pClusterSrc->daNodeNetwork.NumElem(); ++i) {
	if (pClusterSrc->daNodeNetwork[i]->cost.rBandwidth >
	    pClusterSrc->daNodeNetwork[iMin]->cost.rBandwidth)
	  iMin = i;
      }
      cost = pClusterSrc->daNodeNetwork[iMin]->cost;
    }
  } else {  // between clusters, use inter-cluster links
    DC_TopologyCluster *pCluster1 = daSrcUp[0];
    DC_TopologyCluster *pCluster2;
    for (i=1; i<daSrcUp.NumElem(); ++i) {
      pCluster2 = daSrcUp[i];

      cout << "debug: checking " << pCluster1->sbName << " - " << pCluster2->sbName << endl;

      // we ignore nodecluster uplinks to container clusters
      if (! pCluster1->fNodeCluster) {
	// look through cluster networks
	bool fFound = false;
	for (j=0; j<pCluster1->daClusterNetwork.NumElem(); ++j) {
	  if (pCluster1->daClusterNetwork[j]->pDstCluster == pCluster2) {
	    cost.aggregate(pCluster1->daClusterNetwork[j]->cost);
	    fFound = true;
	    break;
	  }
	}
	if (!fFound) cout << "debug: no cluster network connection found between " << pCluster1->sbName << " - " << pCluster2->sbName << endl;
      } else {
	cout << "debug: skipping nodecluster " << pCluster1->sbName << endl;
      }

      pCluster1 = pCluster2;
    }
  }
  return cost;
}
#endif /* 1 */

// assigns the given appnode to the topology node
int DC_Topology_t::AssignMapNode(DC_AppNode *pAppNode,
				 DC_TopologyNode *pNode, int nCopies,
				 double rEST, double rLastData) {
#if defined(DEBUG_MISC)
  cout << "DEBUG: AssignMapNode(): "
       << pAppNode->sbName << " -> " << pNode->sbName
       << endl;
#endif /* DEBUG_MISC */

  pAppNode->pNodeMap = pNode;
  pAppNode->nCopiesMap = nCopies;  // this is not really used other
				   // than for 2 copies of producers.

  pAppNode->rEST = rEST;

  double timeExtent, timeBusy;
  ScaleCpuTimes(pAppNode, pNode->smp.rMips, &timeExtent, &timeBusy, rLastData-rEST);
#if defined(DEBUG_MISC)
  cout << "DEBUG: ScaleCpuTimes():"
       << " rEST=" << rEST
       << " timeExtent=" << timeExtent
       << " timeBusy=" << timeBusy
       << " nCopies=" << pAppNode->pShared->nCopies
       << endl;
#endif /* DEBUG_MISC */

#if defined(DEBUG_CPU_ASSIGNMENT)
  pNode->smp.Dump("before assign", cout);
#endif
  pAppNode->rEFT = pNode->smp.Assign(rEST, timeExtent, timeBusy, pAppNode);
#if defined(DEBUG_CPU_ASSIGNMENT)
  pNode->smp.Dump("after assign", cout);
#endif

  // used for our pipelined assumption: after the first cycle, there
  // is data to be sent.  Note that timeExtent is possibly scaled when
  // scheduled, resulting in rEFT.
#if 1
  // Here, we use the tuning run recorded rInstantWriteFirst.  This captures
  // accumulator vs. pipelined behavior better than assuming the write cycle
  // is over the whole interval.  The proportional time from the tuning run
  // is applied to the actual time once scheduled.
  pAppNode->rEWT = rEST + ((pAppNode->rEFT - pAppNode->rEST) *
			   (pAppNode->pShared->rInstantWriteFirst /
			    pAppNode->pShared->rProcessTime));
#else
  // We make the simplifying assumption that the cycles are
  // distributed evenly over the rEST..eEFT time period.
  assert(pAppNode->pShared->wNumber_Cycle >= 1);
  pAppNode->rEWT = rEST + ((pAppNode->rEFT - pAppNode->rEST)
			   / pAppNode->pShared->wNumber_Cycle);
#endif
  if (pAppNode->rEWT > pAppNode->rEFT) pAppNode->rEWT = pAppNode->rEFT;

#if defined(DEBUG_MISC)
  cout << "DEBUG: smp.Assign() + adj:"
       << " rEST=" << pAppNode->rEST
       << " rEFT=" << pAppNode->rEFT
       << " rEWT=" << pAppNode->rEWT
       << endl;
#endif /* DEBUG_MISC */

  // compute cycle output rate expected from this copy
  //
  // this is local to this appnode, and is not necessarrily the same
  // as the rates of the sources/sinks.  To relate those, we compute
  // cycle rate factors for each ins/outs when needed.

#if defined(DEBUG_MISC)
  cout << "DEBUG2: wNumber_Cycle=" << pAppNode->pShared->wNumber_Cycle
       << " nCopies=" << pAppNode->pShared->nCopies
       << endl;
#endif /* DEBUG_MISC */

  // ASSUME: we see parallelism with multiple copies.  For producers,
  // we assume the data is partitioned, which is similar to
  // non-producer copies where the fixed number of input buffers are
  // shared among all copies.  Overall, this should work for both.  We
  // also need to insure that the number of copies is taken into
  // account when predicting the scaled timeExtent (above).
  int nCycles = pAppNode->pShared->wNumber_Cycle / pAppNode->pShared->nCopies;
  if (nCycles < 1) nCycles = 1;
  pAppNode->rCycleRate = (double(nCycles) /
			  (pAppNode->rEFT - pAppNode->rEST));

  // Ideally we need timeBusy and timeExtent of a dedicated cpu, but
  // when actually scheduled, this may be longer.  We scale the times
  // based on a linear assumption, and divide by the number of cycles
  // to get the per-cycle times.  These are saved for later use in
  // computing the related rates.
  double rRatio = (pAppNode->rEFT - pAppNode->rEST)/(timeExtent);
  pAppNode->rCycleTimeProcess = ((timeExtent * rRatio) /
				 pAppNode->pShared->wNumber_Cycle);
  pAppNode->rCycleTimeComp = ((timeBusy * rRatio) /
			      pAppNode->pShared->wNumber_Cycle);

#if defined(DEBUG_MISC)
  cout << "DEBUG1: " << pAppNode->sbName
       << " wNumber_Cycle=" << pAppNode->pShared->wNumber_Cycle << ','
       << " rEST=" << pAppNode->rEST << ','
       << " rEWT=" << pAppNode->rEWT << ','
       << " rEFT=" << pAppNode->rEFT << ','
       << " rCycleRate=" << pAppNode->rCycleRate << ','
       << " rCycleTimeProcess=" << pAppNode->rCycleTimeProcess << ','
       << " rCycleTimeComp=" << pAppNode->rCycleTimeComp << endl;
#endif /* DEBUG_MISC */
  assert(pAppNode->rEST <= pAppNode->rEWT);
  assert(pAppNode->rEWT <= pAppNode->rEFT);

  // mark all ins as "scheduled"
  for (unsigned int i=0; i<pAppNode->daIns.NumElem(); ++i) {
    pAppNode->daIns[i]->pLink->fScheduled = true;
  }
  return 0;
}


// output: timeExtent, timeBusy - for use with DC_SmpModel calls.
void DC_Topology_t::ScaleCpuTimes(const DC_AppNode *pAppNode,
				  const double rMipsDest,
				  double *ptimeExtent, double *ptimeBusy,
				  double rLastData) {
  double rRatio, timeBusy, timeOther;

  // ASSUME: n copies implies our uow time is reduced by n (less cpu scaling)
  rRatio = (pAppNode->pShared->pNodeInfoTuning->rMips / rMipsDest);

  // linearly scaled compute time, for this copy's portion
  timeBusy = ((rRatio * pAppNode->pShared->rCompTime) /
	      pAppNode->pShared->nCopies);

  // unscaled stream other time, for this copy's portion
#if 1 /* ***MDB*** */
  if (Global.wFeatures & PLANNER_USE_PIPE &&
      Global.wFeatures & PLANNER_COPIES_1) {
    // don't include read stall time from tuning run
    // and since we are creating copies, we can assume
    // that we will lose write stalls too
    timeOther = ((pAppNode->pShared->rProcessTime -
		  pAppNode->pShared->rCompTime -
		  pAppNode->pShared->rReadTime -
		  pAppNode->pShared->rWriteTime) /
		 pAppNode->pShared->nCopies);
  } else {
    // don't include read stall time from tuning run
    timeOther = ((pAppNode->pShared->rProcessTime -
		  pAppNode->pShared->rCompTime /* -
		  pAppNode->pShared->rReadTime*/) /
		 pAppNode->pShared->nCopies);
  }
#else
  timeOther = ((pAppNode->pShared->rProcessTime -
		pAppNode->pShared->rCompTime) /
	       pAppNode->pShared->nCopies);
#endif
  if (timeOther < 0.0) timeOther = 0.0;

  assert(ptimeExtent && ptimeBusy);
  *ptimeExtent = timeBusy + timeOther;
  *ptimeBusy = timeBusy;


  // check to see if we need to increase the time due to read stalls
  //
  // ASSUME: we cannot finish before our predecessors do.  If in
  // fact we should finish earlier, but have to wait for read data
  // to arrive, then we have read stalls.  This will later affect
  // our output, so we avoid creating too many copies later assuming
  // a higher output rate.
  if (Global.wFeatures & PLANNER_USE_PIPE) {
#if defined(DEBUG_MISC)
    cout << endl
	 << "DEBUG: CheckReadStalls(): check *ptimeExtent=" << *ptimeExtent
	 << " < rLastData-rEST=" << rLastData << endl;
#endif /* DEBUG_MISC */
    if (*ptimeExtent < rLastData) {
#if defined(DEBUG_MISC)
      if ((rLastData - *ptimeExtent) > DOUBLE_EPSILON) {
	cout << endl
	     << "DEBUG: CheckReadStalls(): Adding "
	     << (rLastData - *ptimeExtent)
	     << " to compensate for slower pred" << endl;
      }
#endif /* DEBUG_MISC */
      *ptimeExtent = rLastData;
    }
  }
  return;
}


// assuming all pred have valid EWT/EFT fields based on actual cluster
// assignments, we compute the possible EST of the given appnode, to
// the given cluster.
// input: pNode - for node cluster case, NULL for container cluster case
//        pwColocation - degree of colocation (used for breaking ties)
// output: *prLastData = the time for the last piece of data to arrive
//                       (used later in CheckReadStalls())
double DC_Topology_t::ComputeEST(DC_AppNode *pAppNode,
				 DC_TopologyCluster *pCluster,
				 DC_TopologyNode *pNode,
				 int *pwColocation,
				 double *prLastData) {
  // traditional dataflow assumption: all data is read and exists
  // locally before processing starts, then all output data it written
  // at the end.
  // vs.
  // our pipelined assumption: we assume that the delay in starting a
  // filter is just the pipe fill time - (total volume / # writes).

  double rESTPipe, rESTNonPipe;

  // transfer time == EST (earliest start time)
  //
  // ASSUME: copies will balance rates enough to get pipelining, so we
  // only need to worry about the first cycle's data to arrive
  DC_AppNodeInsOuts *pInsOuts;
  unsigned int nins = pAppNode->daIns.NumElem();
  double rTmp, rTmpPipe, rTmpNonPipe;
  int wColocation = 0;

  // how long to wait until we receive enough data -- max of all
  // transfer times from all copies of writers across all ins.
  rESTPipe = rESTNonPipe = 0.0;
  // ASSUME: for a producer with no ins, it starts at time 0.0
  for (unsigned int i=0; i<nins; ++i) {
    pInsOuts = pAppNode->daIns[i];
    // volume: pInsOuts->wVolumeKb_Cycle
    // src:    pInsOuts->pLink->daSrc[*]->pNodeMap->pCluster
    // dst:    pCluster fcn arg
    // => use one with the longest delay
    for (unsigned int j=0; j<pInsOuts->pLink->daSrc.NumElem(); ++j) {
      DC_AppNode *pAppNodeSrc = pInsOuts->pLink->daSrc[j];
      assert(pAppNodeSrc->pNodeMap);
      DC_TopologyCluster *pClusterSrc = pAppNodeSrc->pNodeMap->pCluster;
#if !defined(USE_COMPUTENETWORKXFER)
      DC_TopologyNetworkCost netcost;
#endif

      bool fSameHost = (pNode == pAppNodeSrc->pNodeMap ? true : false);
      if (fSameHost) ++wColocation;

      // BUGBUG: each daSrc[] is a copy, and this loop finds a single
      // copy to exclusively send data to us in the least time, which
      // is not realistically what happens.  Instead, we should assume
      // some (optimal) distribution policy and find the fastest way
      // those buffers can get to us.  The problem is we run the same
      // alg for all copies of the receiveing filter, and there is no
      // easy method to distinguish which buffers are allocated to a
      // *particular* receiving filter.
      //
      // Perhaps we can just tag sender copies to say how many cycles
      // of data are reserved for pipefilling in other receiving
      // filters.  This allocation is similar to what is needed for
      // computing the steady state number of copies that are needed.


      // BUGBUG: we need to instead compute the time to get *enough*
      // data (that required for our local cycle) to us.  As is below,
      // we just wait for the first write from our pred's local cycle,
      // which may or may not be correlated with our cycle.  This
      // needs to be fixed, and is complicated by pred copies as
      // described above.

#if defined(USE_COMPUTENETWORKXFER)
      if (fSameHost) {
	rTmp = DOUBLE_EPSILON;
	//rTmp = 0.0;
	rTmpPipe    = pAppNodeSrc->rEWT + rTmp;
	rTmpNonPipe = pAppNodeSrc->rEFT + rTmp;
      } else {
	long wVolumeKb;
	if (!(Global.wFeatures & PLANNER_USE_PIPE)) {
	  // here we don't model cycles, so to avoid pinned producer copies
	  // from appearing to send too much, we take care of multiple
	  // pinned producer predecessors.
	  wVolumeKb = (pInsOuts->wVolumeKb_Cycle /
		       pInsOuts->pLink->daSrc.NumElem());
	} else {
	  wVolumeKb = pInsOuts->wVolumeKb_Cycle;
	}
      
	// 0.0 for timeExtent means "as fast as possible"
	rTmpPipe    = ComputeNetworkXfer(pClusterSrc, pCluster, wVolumeKb,
					 pAppNodeSrc->rEWT, 0.0, false);
	rTmpNonPipe = ComputeNetworkXfer(pClusterSrc, pCluster, wVolumeKb,
					 pAppNodeSrc->rEFT, 0.0, false);
      }
#if defined(DEBUG_COMPUTE_EST)
      cout << "ComputeEST(): ins[" << i
	   << "].cs" << j << " on cluster " << pClusterSrc->sbName
	   << " to " << pCluster->sbName
	//<< ", cost=" << netcost
	   << ", vol=" << pInsOuts->wVolumeKb_Cycle
	   << ", time_pipe=" << rTmpPipe
	   << ", time_nonpipe=" << rTmpNonPipe << ", (network=" << rTmp << ')'
	   << endl;
#endif
#else /* USE_COMPUTENETWORKXFER */
      if (fSameHost) {
	rTmp = DOUBLE_EPSILON;
	//rTmp = 0.0;
      } else {
	netcost = ComputeNetworkCost(pClusterSrc, pCluster);

	if (!(Global.wFeatures & PLANNER_USE_PIPE)) {
	  // here we don't model cycles, so to avoid pinned producer copies
	  // from appearing to send too much, we take care of multiple
	  // pinned producer predecessors.
	  rTmp = netcost.rLatency + ((double(pInsOuts->wVolumeKb_Cycle/
					     pInsOuts->pLink->daSrc.NumElem())
				      / netcost.rBandwidth)
				     * double(8.0*1024.0/1000000.0));
	} else {
	  rTmp = netcost.rLatency + ((double(pInsOuts->wVolumeKb_Cycle)
				      / netcost.rBandwidth)
				     * double(8.0*1024.0/1000000.0));
	}
      }
      rTmpPipe    = pAppNodeSrc->rEWT + rTmp;
      rTmpNonPipe = pAppNodeSrc->rEFT + rTmp;
#if defined(DEBUG_COMPUTE_EST)
      cout << "ComputeEST(): ins[" << i
	   << "].cs" << j << " on cluster " << pClusterSrc->sbName
	   << " to " << pCluster->sbName
	   << ", cost=" << netcost
	   << ", vol=" << pInsOuts->wVolumeKb_Cycle
	   << ", time_pipe=" << rTmpPipe
	   << ", time_nonpipe=" << rTmpNonPipe << ", (network=" << rTmp << ')'
	   << endl;
#endif
#endif /* USE_COMPUTENETWORKXFER */

      if (rESTPipe < rTmpPipe) rESTPipe = rTmpPipe;  // max
      if (rESTNonPipe < rTmpNonPipe) rESTNonPipe = rTmpNonPipe;  // max
    }
  }

#if defined(DEBUG_COMPUTE_EST)
  cout << "ComputeEST(): min: time_pipe=" << rESTPipe
       << ", time_nonpipe=" << rESTNonPipe << endl;
#endif
  *prLastData = rESTNonPipe;
  if (pwColocation) *pwColocation = wColocation;

  if (Global.wFeatures & PLANNER_USE_PIPE) {
#if defined(DEBUG_COMPUTE_EST)
    cout << "rEST=" << rESTPipe << endl;
#endif
    return rESTPipe;
  }

#if defined(DEBUG_COMPUTE_EST)
  cout << "rEST=" << rESTNonPipe << endl;
#endif
  return rESTNonPipe;
}

// given the computed EST on this node, we compute the resulting EFT
// in: if pNode == NULL -> sched to cluster, else sched to the actual node
double DC_Topology_t::ComputeEFT(DC_AppNode *pAppNode,
				 DC_TopologyCluster *pCluster,
				 DC_TopologyNode *pNode,
				 double rEST,
				 double rLastData) {
  double rEFT;

  // computation time
  //
  // 10/05/2001 : mdb - Fixed bug where smp.Check() output was treated as
  //                    a time extent instead of absolute time as it is.
  //
  // ASSUME: copies will be created so the reader can keep up with us
  if (pCluster->fNodeCluster && pNode) {
    double timeExtent, timeBusy;
    ScaleCpuTimes(pAppNode, pNode->smp.rMips,
		  &timeExtent, &timeBusy, rLastData-rEST);
    rEFT = pNode->smp.Check(rEST, timeExtent, timeBusy, pAppNode);
#if defined(DEBUG_COMPUTE_EFT)
    cout << "ComputeEFT(): nodecluster EST=" << rEST
	 << " -> EFT=" << rEFT << endl;
#endif
  } else {  // for a container cluster, or a node cluster in summary
    // TODO: use smp as a cluster model?  how would this be done?
    rEFT = rEST +
      pAppNode->rWeight * (pAppNode->pShared->pNodeInfoTuning->rMips /
			   (pCluster->ClusterInfoRemain.rMips /
			    pCluster->ClusterInfo.nCpu));
#if defined(DEBUG_COMPUTE_EFT)
    cout << "ComputeEFT(): container cluster EST=" << rEST
	 << " -> EFT=" << rEFT << endl;
#endif
  }
  return rEFT;
}

//
// We want to compute this exactly, but that would require mapping and
// reevaluating if more copies are needed, where existing copies would
// need to be torn down, and re-mapped with correct volume and EST
// information.  It is assumed to be to computationally intensive, and
// a way to approximate the number of copies needed in a virtual sense
// would be better for an online alg.  (A counter argument is that how
// unbalanced an application is will determine how many copies are
// needed, hence how many re-mappings.  It is hard to believe the
// number of copies will be that large.)  Possibly a hybrid approach,
// where we estimate the number of copies first, and then possibly add
// more as needed once we actually instantiate the copies, and see
// what the resulting rates will be.
//
// For now, we do the simple approach, and just estimate the number of
// copies needed from the tuning run stats.
//
int DC_Topology_t::ComputeNeededCopies(DC_App_t &app, DC_AppNode *pAppNode) {
  unsigned int i, j;

  if (!pAppNode || pAppNode->daIns.NumElem() == 0) return 1;  // producer?
  assert(pAppNode->pShared->nCopies == 1);

  // NOTE: for ComputeEST() we cared about the first piece of data
  // arriving, and assumed the others would be overlapped by the
  // pipelining benefit.  Here, we are trying to insure this will
  // actually happen, and instead care about the overall rates and how
  // many copies of the current filter are needed to keep up with the
  // rate of incoming buffers.

  // We statically compute number of copies, virtually based on tuning
  // run.  The sources may have copies as well.  If the sources are
  // producers, we get an "instant" speedup if on different hosts.
  int n, nCopies;

  // TODO: if we don't get the number of copies just right, we introduce
  // stalls, which should be taken into consideration, since that could
  // cause too many unneeded later-stage copies.
  
  //SUM(cycles/sec for each src copy) -> aggr cycles per sec for "src"
  double rSrcRateVolAggr, rCopies;
  for (i=0; i<pAppNode->daIns.NumElem(); ++i) {
    DC_AppNodeInsOuts *pInsOuts = pAppNode->daIns[i];

    rSrcRateVolAggr = 0.0;
    for (j=0; j<pInsOuts->pLink->daSrc.NumElem(); ++j) {
      DC_AppNode *pAppNodeSrc = pInsOuts->pLink->daSrc[j];
      DC_AppNodeInsOuts *pInsOutsPred = app.FindPredInsOuts(pAppNodeSrc,
							    pInsOuts->pLink);
      assert(pInsOutsPred);
      rSrcRateVolAggr += (pAppNodeSrc->rCycleRate *
			  double(pInsOutsPred->wVolumeKb_Cycle));
    }
    // NOTE: at this point, we don't know our actual rCycleRate, since
    // we are not yet mapped to a host.  Instead, we approximate it
    // here using the tuning run.  The drawback is if the tuning run
    // was on hosts that are much different than those actually mapped
    // to, the number of copies can be way off.
    //
    // The alternative is to place k copies, compute the exact rates,
    // and see if we need more.  If more neded, remove k copies, and
    // try k+1.  Continue until we have enough.
#if 1
    // 10/15/2001 : mdb - Less the write time, since should not stall.
    double rCycleRateApprox = (double(pAppNode->pShared->wNumber_Cycle) /
			       (pAppNode->pShared->rProcessTime -
				pAppNode->pShared->rWriteTime));
#else
    double rCycleRateApprox = (double(pAppNode->pShared->wNumber_Cycle) /
			       pAppNode->pShared->rProcessTime);
#endif
    rCopies = (rSrcRateVolAggr /
	       (rCycleRateApprox * double(pInsOuts->wVolumeKb_Cycle)));
    n = int(ceil(rCopies));

#if defined(DEBUG_MISC)
    cout << "ComputeNeededCopies(): " << pAppNode->sbName
	 << " ins[" << i << ']'
	 << " rSrcRateVolAggr=" << rSrcRateVolAggr
	 << " rCycleRateApprox=" << rCycleRateApprox
	 << " wVolumeKb=" << pInsOuts->wVolumeKb_Cycle
	 << ", rCopies=" << rCopies << ", n=" << n
	 << endl;
#endif /* DEBUG_MISC */
    if (n < 1) n = 1;  // need at least 1 copy
    // keep min nCopies
    if (i == 0 || nCopies > n) nCopies = n;
  }

  // obey user-specified layout restriction on copies
  if (pAppNode->pShared->nCopiesMin != -1 &&
      nCopies < pAppNode->pShared->nCopiesMin)
    nCopies = pAppNode->pShared->nCopiesMin;
  if (pAppNode->pShared->nCopiesMax != -1 &&
      nCopies > pAppNode->pShared->nCopiesMax)
    nCopies = pAppNode->pShared->nCopiesMax;

#if defined(DEBUG_MISC)
  cout << "ComputeNeededCopies(): " << pAppNode->sbName 
       << " min nCopies=" << nCopies << endl;
#endif /* DEBUG_MISC */

  return nCopies;
}


//
//  planner code
//

//
// The goal is to partition app nodes to available clusters.  We start
// from all pinned nodes, where a mapping is predetermined because
// that is where the input data lives, or where the client is located,
// or where the output is supposed to reside.  Each pinned node is
// considered RR in any order (or maybe in order of largest output or
// input volume size).  For each pinned node, we consider a neighbor
// node, which either outputs to or inputs from the pinned node.  The
// order neighbors are considered is ??? (maybe arbitrary, maybe
// b_level, etc, maybe depend on if input or output and where in some
// level sorted order they are).
//
// For each candidate node, we need to choose how many copies, and
// where each should run.  We continue finding nodes to run on in a
// greedy fashion.  [Problem: we fill up all nodes early in the
// pipeline, leaving few for later stages.  Maybe this is ok if the
// number of hosts is very large in comparision to the number of app
// nodes.  Alternative is to just do partitioning w/o copies, and make
// a final pass to add copies as needed.]
//
// Refinenment: Since we only consider the source-sink data transfer
// cost and the sink computation cost, we ignore the cost of the sink
// outputting and where the sink's sinks are located.  When the
// initial partitioning finishes, the last link(s) between each
// seperate region may be a very bad choice.  To fix this we move
// sources/sinks around to compensate.
//
// For recursive partitioning, we need to summarize the input and
// output to the partition, store this somewhere in the stream (most
// likely), and then call recursively.  Lower calls see an stream
// endpoint that is not in the subset (by some flag in the stream, or
// by the setid of the other app node).  Maybe this should just entail
// a single pass of the current graph, and decide what the
// inter-partition rates will be and store in the stream.
//


// generic planner framework
int planner(DC_Topology_t &topology, DC_App_t &app, ConfigInfo &cfgTuneRun) {
  unsigned int i;
  DC_AppNode  *pAppNode, *pAppNodeTmp;
  DC_TopologyCluster *pCluster;
  // ASSUME: only one top-level cluster
  assert(topology.daClusterRoot.NumElem() == 1);
  DC_TopologyCluster *pClusterRoot = topology.daClusterRoot[0];

  DArrayPtr<DC_TopologyCluster*> daClusterStack;
  DC_TopologyCluster *pSubCluster;
  DC_TopologyNode    *pNode, *pNodeSave;

  DC_Heap<double,DC_AppNode> heap;
  for (i=0; i<app.daNodes.NumElem(); ++i) {
    // add to heap
    pAppNode = app.daNodes[i];
    if (pAppNode->daIns.NumElem() == 0 && pAppNode->fPinned)
      continue;  // skip already assigned producers
    heap.insert(pAppNode->b_level, pAppNode);
  }

  // for each appnode in heap order
  double rESTSave, rEFTSave, rLastDataSave;
  double rTmpEST, rTmpEFT, rTmpLastData;
  int wColocationSave;
  int wTmpColocation;
  int iCopy, nCopies;
  while ((pAppNode = heap.extract_max()) != NULL) {
    if (pAppNode->fPinned) {
      // map it now to update times
      rTmpEST = topology.ComputeEST(pAppNode, pAppNode->pNodeMap->pCluster,
				    pAppNode->pNodeMap, NULL, &rTmpLastData);
      rTmpEFT = topology.ComputeEFT(pAppNode, pAppNode->pNodeMap->pCluster,
				    pAppNode->pNodeMap, rTmpEST, rTmpLastData);
      topology.AssignMapNode(pAppNode, pAppNode->pNodeMap, 1,
			     rTmpEST, rLastDataSave);
      continue;
    }
    // generate as many copies as needed to satisfy dataflow requirements
    nCopies = 1;
    if (Global.wFeatures & PLANNER_COPIES_1) {
      // NOTE: its seems more natural to instead just keep iteratively
      // adding another copy until the rates are satisfied, but that
      // has a problem.  We base the placement decision on volume and
      // timing information that will be based on less copies, and
      // then later add another copy, which may cause a bad decision.
      // Instead, we are just balancing it virtually, and then
      // scheduling those copies.  The refinement phase can always be
      // modified to add copies, but that may be as complex as the
      // above description of adding them one at a time.
      nCopies = topology.ComputeNeededCopies(app, pAppNode);

      // BUGBUG: we need to create all copies we need, or think we
      // need at once, similar to the pinned code.  Then later when we
      // do the assignment, we get true volume and time information
      // due to nCopies > 1.
    }
    // create all needed copies, so we know the total for mapping
    for (iCopy=1; iCopy<nCopies; ++iCopy) {
      pAppNodeTmp = app.CreateCopy(pAppNode);
    }
    assert(nCopies == pAppNode->pShared->nCopies);

    // now map the copies
    for (iCopy=0; iCopy<nCopies; ++iCopy) {
      pAppNodeTmp = pAppNode->pShared->daCopies[iCopy];
      pNodeSave = NULL;

      // starting at the cluster root, we descend choosing a subcluster for
      // this appnode copy, and eventually an actual host assignment.
      daClusterStack.Append(pClusterRoot);  // push
      while (daClusterStack.NumElem() > 0) {
	pCluster = daClusterStack.GetLast(); daClusterStack.DeAppend();  // pop

#if defined(DEBUG_MISC)
	cout << endl
	     << "--- planning cluster " << pCluster->sbName << endl;
#endif /* DEBUG_MISC */
	if (pCluster->fNodeCluster) {
	  //
	  // node cluster assignment
	  //
#if defined(DEBUG_MISC)
	  pAppNodeTmp->PrintFullName(cout);
	  cout << ": assign to actual hosts ..." << endl;
#endif /* DEBUG_MISC */

	  // consider all actual nodes in this cluster
#if defined(ENABLE_IDLE_NODE_PRUNE)
	  bool fIdleFlag = false;
#endif /* ENABLE_IDLE_NODE_PRUNE */
	  for (i=0; i<pCluster->daNodes.NumElem(); ++i) {
	    pNode = pCluster->daNodes[i];

#if defined(ENABLE_IDLE_NODE_PRUNE)
	    if (!fIdleFlag) {
	      if (pNode->smp.isAllEmpty()) fIdleFlag = true;  // first
	    } else {
	      if (pNode->smp.isAllEmpty()) {  // second+
		cout << "Note: hit idle nodes, prune search space" << endl;
		break;
	      }
	    }
#endif /* ENABLE_IDLE_NODE_PRUNE */

#if defined(DEBUG_MISC)
	    cout << "- considering node " << pNode->sbName << " ..." << endl;
#endif /* DEBUG_MISC */
	    rTmpEST = topology.ComputeEST(pAppNodeTmp, pCluster,
					  pNode, &wTmpColocation,
					  &rTmpLastData);
	    rTmpEFT = topology.ComputeEFT(pAppNodeTmp, pCluster,
					  pNode, rTmpEST, rTmpLastData);
#if defined(DEBUG_MISC)
	    cout << "= node " << pNode->sbName
		 << " EST=" << rTmpEST << ", EFT= " << rTmpEFT
		 << ", colocation=" << wTmpColocation << endl;
#endif /* DEBUG_MISC */
	    // we use a measure of colocation as the tie-breaker
	    if (!pNodeSave ||  // first?
		(rEFTSave > rTmpEFT) ||  // new min
		(rEFTSave == rTmpEFT &&
		 wColocationSave < wTmpColocation)) {  // better colocation?
	      rEFTSave = rTmpEFT;
	      rESTSave = rTmpEST;
	      rLastDataSave = rTmpLastData;
	      pNodeSave = pNode;
	      wColocationSave = wTmpColocation;
	    }
	  }
	} else {
	  //
	  // collection cluster partitioning
	  //
#if defined(DEBUG_MISC)
	  pAppNodeTmp->PrintFullName(cout);
	  cout << ": check subclusters ..." << endl;
#endif /* DEBUG_MISC */
	  assert(pCluster->daSubClusters.NumElem() > 0);

	  if (Global.wFeatures & PLANNER_USE_ALL) {
	    // follow all sub-clusters
	    for (i=0; i<pCluster->daSubClusters.NumElem(); ++i) {
	      pSubCluster = pCluster->daSubClusters[i];
	      daClusterStack.Append(pSubCluster);  // push
	    }
	  } else {
	    // choose a min sub-cluster to follow (prune)
	    double rEFTMinCluster;
	    DC_TopologyCluster *pClusterMin = NULL;
	    for (i=0; i<pCluster->daSubClusters.NumElem(); ++i) {
	      pSubCluster = pCluster->daSubClusters[i];
#if defined(DEBUG_MISC)
	      cout << "- considering sub-cluster "
		   << pSubCluster->sbName << " ...";
#endif /* DEBUG_MISC */
	      rTmpEST = topology.ComputeEST(pAppNodeTmp, pSubCluster,
					    NULL, NULL, &rTmpLastData);
	      rTmpEFT = topology.ComputeEFT(pAppNodeTmp, pSubCluster,
					    NULL, rTmpEST, rTmpLastData);
#if defined(DEBUG_MISC)
	      cout << " EFT= " << rTmpEFT << endl;
#endif /* DEBUG_MISC */
	      if (!pClusterMin || rEFTMinCluster > rTmpEFT) {  // first? || new min?
		rEFTMinCluster = rTmpEFT;
		pClusterMin = pSubCluster;
	      }
	    }
	    assert(pClusterMin);
#if defined(DEBUG_MISC)
	    cout << "-> min EFT= " << rEFTMinCluster << " on cluster "
		 << pClusterMin->sbName << endl;
#endif /* DEBUG_MISC */
	    daClusterStack.Append(pClusterMin);  // push
	  }
	}

      }  // while daClusterStack

#if defined(DEBUG_MISC)
      cout << "-> min EFT= " << rEFTSave << " on node "
	   << pNodeSave->sbName << " rLastDataSave=" << rLastDataSave << endl;
#endif /* DEBUG_MISC */
      // do the assignment
      topology.AssignMapNode(pAppNodeTmp, pNodeSave, 1, rESTSave, rLastDataSave);
#if defined(DEBUG_MISC)
      cout << "= assign " << pAppNodeTmp->sbName
	   << " to node " << pNodeSave->sbName
	   << " EST=" << pAppNodeTmp->rEST << '(' << rESTSave << ')'
	   << ", EFT= " << pAppNodeTmp->rEFT << '(' << rEFTSave << ')'
	   << endl;
#endif /* DEBUG_MISC */
#if defined(DEBUG_CPU_ASSIGNMENT)
      pNode->smp.Dump("after assigning", cout);
#endif

    }  // for iCopy
  }  // while pAppNode

  // now that we have an initial placement, the refinement phase starts
  if (Global.wFeatures & PLANNER_REFINE) {
    // TODO: refine the edges that are marked as "unscheduled"
  }
  return 0;
}


// output by filling cfgPlacement with the resulting placement
int GeneratePlacement(ConfigInfo &cfgPlacement,
		      DC_Topology_t &topology, DC_App_t &app,
		      ConfigInfo &cfgTuneRun) {
  const int maxbuf = 80;
  char buf[maxbuf];
  cfgPlacement.Flush();
  unsigned int iSection = cfgPlacement.AddSection("placement");
  unsigned int i;

  // init
  for (i=0; i<app.daNodes.NumElem(); ++i) {
    DC_AppNode *pAppNode = app.daNodes[i];
    pAppNode->eColor = COLOR_BLACK;
    if (!pAppNode->pNodeMap) {
      cout << "ERROR: no mapping found for " << pAppNode->sbName << endl;
      continue;
    }
    pAppNode->pNodeMap->eColor = COLOR_BLACK;
#if defined(DEBUG_CPU_ASSIGNMENT)
    cout << "placement of filter " << pAppNode->sbName
	 << " on host " << pAppNode->pNodeMap->sbName
	 << endl;
    //pAppNode->pNodeMap->smp.Dump("", cout);
#endif
  }

  int nCopies;
  DC_AppNode *pAppNodePass;
  do {
    pAppNodePass = NULL;
    nCopies = 0;
    for (i=0; i<app.daNodes.NumElem(); ++i) {
      DC_AppNode *pAppNode = app.daNodes[i];
      if (pAppNode->eColor == COLOR_WHITE) continue;  // already done?
      if (!pAppNodePass) pAppNodePass = pAppNode;  // first in new loop?
      if (pAppNode->pShared != pAppNodePass->pShared) continue;  // this pass?
      if (pAppNode->pNodeMap != pAppNodePass->pNodeMap) continue;  // diff map?
      pAppNode->eColor = COLOR_WHITE;  // mark as processed
      //cout << "adding " << pAppNode->nCopiesMap << " copies for " << pAppNode->sbName << ':' << pAppNode->wCopyNum << " -> " << nCopies + pAppNode->nCopiesMap << endl;
      nCopies += pAppNode->nCopiesMap;
    }
    if (pAppNodePass) {
#if 1
      // 10/16/2001 : mdb - Added this.
      //
      // ASSUME: It makes no sense to have more copies of a given filter
      // than the number of cpus.  If we send more work to a filter that
      // has utilization < 1, then its time_busy will increase.  Since
      // we already scheduled the previous copies, we allow "virtual"
      // copies of the same filter to be mapped to the same node, and then
      // strip them away here.  This does mess up the pretty output, but
      // there is not much we can do about it quickly.
      if (nCopies > pAppNodePass->pNodeMap->smp.nCpu &&
	  pAppNodePass->daIns.NumElem() > 0) {
	cout << "*** reducing number of "
	     << pAppNodePass->sbName << " copies on "
	     << pAppNodePass->pNodeMap->sbName << " from "
	     << nCopies << " to " << pAppNodePass->pNodeMap->smp.nCpu << endl;
	nCopies = pAppNodePass->pNodeMap->smp.nCpu;
      }
#endif

      snprintf(buf, maxbuf, "%s %d",
	       pAppNodePass->pNodeMap->sbName, nCopies);
      char bufName[maxbuf];
      // the computation of copysets needs to be changed to really represent
      // all copies on a host, instead of copy number.
      //snprintf(bufName, maxbuf,
      //"%s:%d", pAppNodePass->sbName, pAppNodePass->wCopyNum);
      snprintf(bufName, maxbuf, "%s", pAppNodePass->sbName);
      cfgPlacement.AddEntry(iSection, bufName, buf);

      if (pAppNodePass->pNodeMap->eColor == COLOR_BLACK) {
	pAppNodePass->pNodeMap->eColor = COLOR_WHITE;
	//pAppNodePass->pNodeMap->smp.Dump("", cout);
      }
    }
  } while (pAppNodePass);
  return 0;
}

int OutputPlacement(ConfigInfo &cfgPlacement, char *sbFile,
		    int argc, char *argv[],
		    double tPlanner, double rEFT, double rOptEFT) {
  cout << "OutputPlacement: Creating \"" << sbFile << "\" for placement ..."
       << endl;
  ofstream fout;
  fout.open(sbFile, ios::out);
  if (fout.fail()) return 1;

  fout << "# automatically generated file" << endl
       << "# DataCutter planner - placement spec" << endl;
  for (int i=0; i<argc; ++i) {
    fout << "# argv[" << i << "] = " << argv[i] << endl;
  }
  fout << endl
       << "rOptEFT= " << rOptEFT << endl
       << "tPlanner= " << tPlanner << endl
       << "rEFT= " << rEFT << endl
       << endl
       << cfgPlacement;

  fout.close();
  return fout.fail() ? 1 : 0;
}


//
//  stand alone computation for lower bound (OPT) - should be called before
//  graph is mutated.
//
//  assume: not more than 1 pinned consumer
//  ignore: internal pinned filters
//
double DC_Topology_t::ComputeLowerBound(DC_App_t &app) {
  // time is the max of:
  // - producer computation - since we have to read the data
  // - one cycle of all other filters, since we could run a copy for
  //   each cycle
  // and if a pinned consumer we add:
  // - the network transfer of the minimum cycle volume from the src
  //   pinned consumer
  double rEFT = 0.0, rTmp;
  bool fAny;
  unsigned int i, j, nProducer;
  DC_AppNode *pAppNode, *pAppNodePinnedConsumer;
  DArrayPtr<DC_AppNode*> daTopoList;

  // max producer time - they are pinned and we must wait for the slowest
  fAny = false;
  pAppNodePinnedConsumer = NULL;
  nProducer = 0;
  for (i=0; i<app.daNodes.NumElem(); ++i) {
    pAppNode = app.daNodes[i];
    if (pAppNode->daOuts.NumElem() == 0 && pAppNode->fPinned) {
      assert(pAppNodePinnedConsumer == NULL);  // assume not more than 1
      pAppNodePinnedConsumer = pAppNode;
    }
    if (pAppNode->daIns.NumElem() > 0) continue;  // not a producer
    // producer
    ++nProducer;
    double r = pAppNode->pShared->rCompTime / pAppNode->pShared->nCopies;
    if (!fAny || rTmp < r) {  // max
      fAny = true;
      rTmp = r;
    }
  }
  rEFT = rTmp;
#if defined(DEBUG_COMPUTE_LB)
  cout << "LB: producer time: " << rEFT << endl;
#endif /* DEBUG_COMPUTE_LB */
  // all other filters - each runs 1 cycle if #copies == #cycles (best case)
  for (i=0; i<app.daNodes.NumElem(); ++i) {
    pAppNode = app.daNodes[i];
    if (pAppNode->daIns.NumElem() == 0) continue;  // a producer
    // find the min time to execute 1 cycle of this filter on any host
    rTmp = ComputeLowerBound_Aux(pAppNode);
    if (!fAny || rEFT < rTmp) {  // max
      fAny = true;
      rEFT = rTmp;
    }
  }
#if defined(DEBUG_COMPUTE_LB)
  cout << "LB: factor in 1 cycle of others: " << rEFT << endl;
#endif /* DEBUG_COMPUTE_LB */

  if (pAppNodePinnedConsumer) {
    // here the min cost is some filters at the client with network
    // cost smallest required network transfer from any producer.
    //
    // we approximate this by just adding in the smallest volume
    // network cost from any pinned producer to the pinned consumer.
    double rNet_Cycle, rNet_Total;
    int wMinVol_Cycle, wMinVol_Total;
    fAny = false;
    for (i=0; i<app.daNodes.NumElem(); ++i) {
      pAppNode = app.daNodes[i];
      for (j=0; j<pAppNode->daIns.NumElem(); ++j) {
	DC_AppNodeInsOuts *pInsOuts = pAppNode->daIns[j];
	if (!fAny || wMinVol_Cycle > pInsOuts->wVolumeKb_Cycle) {  // min
	  wMinVol_Cycle = pInsOuts->wVolumeKb_Cycle;
	}
	if (!fAny || wMinVol_Total > pInsOuts->wVolumeKb) {  // min
	  wMinVol_Total = pInsOuts->wVolumeKb;
	}
	fAny = true;
      }
    }
#if defined(DEBUG_COMPUTE_LB)
    cout << "LB: min vol_cycle=" << wMinVol_Cycle
	 << " vol_total=" << wMinVol_Total << endl;
#endif /* DEBUG_COMPUTE_LB */
    // compute the network cost from any producer to the single pinned consumer
#if !defined(USE_COMPUTENETWORKXFER)
    DC_TopologyNetworkCost netcost;
#endif
    fAny = false;
    for (i=0; i<app.daNodes.NumElem(); ++i) {
      pAppNode = app.daNodes[i];
      if (pAppNode->daIns.NumElem() > 0) continue;  // not a producer
      // producer
      if (pAppNodePinnedConsumer->pNodeMap == pAppNode->pNodeMap) continue;
#if defined(USE_COMPUTENETWORKXFER)
      // 0.0 for timeStart (what should this be instead?)
      // 0.0 for timeExtent means "as fast as possible"
      rTmp = ComputeNetworkXfer(pAppNode->pNodeMap->pCluster,
				pAppNodePinnedConsumer->pNodeMap->pCluster,
				wMinVol_Cycle,
				0.0, 0.0, false);
      if (!fAny || rNet_Cycle < rTmp) {  // max
	rNet_Cycle = rTmp;
      }
      // 0.0 for timeStart (what should this be instead?)
      // 0.0 for timeExtent means "as fast as possible"
      rTmp = ComputeNetworkXfer(pAppNode->pNodeMap->pCluster,
				pAppNodePinnedConsumer->pNodeMap->pCluster,
				wMinVol_Total,
				0.0, 0.0, false);
      if (!fAny || rNet_Total < rTmp) {  // max
	rNet_Total = rTmp;
      }
#else /* USE_COMPUTENETWORKXFER */
      netcost =
	ComputeNetworkCost(pAppNode->pNodeMap->pCluster,
			   pAppNodePinnedConsumer->pNodeMap->pCluster);
      rTmp = netcost.rLatency + ((double(wMinVol_Cycle)
				  / netcost.rBandwidth)
				 * double(8.0*1024.0/1000000.0));
      if (!fAny || rNet_Cycle < rTmp) {  // max
	rNet_Cycle = rTmp;
      }
      rTmp = netcost.rLatency + ((double(wMinVol_Total)
				  / netcost.rBandwidth)
				 * double(8.0*1024.0/1000000.0));
      if (!fAny || rNet_Total < rTmp) {  // max
	rNet_Total = rTmp;
      }
#endif /* USE_COMPUTENETWORKXFER */
      fAny = true;
    }
#if defined(DEBUG_COMPUTE_LB)
    cout << "LB: net_cycle=" << rNet_Cycle
	 << " net_total=" << rNet_Total << endl;
#endif /* DEBUG_COMPUTE_LB */
    if (rEFT + rNet_Cycle > rNet_Total) {
      rEFT = rEFT + rNet_Cycle;  // time to send smallest 1 cycle volume
    } else {
      rEFT = rNet_Total;  // xfer time dwarfs all else
    }
  }

#if defined(DEBUG_COMPUTE_LB)
  cout << "LB: return: " << rEFT << endl;
#endif /* DEBUG_COMPUTE_LB */
  return rEFT;
}
double DC_Topology_t::ComputeLowerBound_Aux(DC_AppNode *pAppNode) {
  // go through all nodes, and see how long 1 cycle will take
  DArrayPtr<DC_TopologyCluster*> daClusterStack;
  DC_TopologyCluster *pCluster, *pSubCluster;
  DC_TopologyNode    *pNode;
  unsigned int i;

  double rEFT, rTmp;
  bool fAny = false;

  assert(daClusterRoot.NumElem() == 1);
  daClusterStack.Append(daClusterRoot[0]);  // push
  while (daClusterStack.NumElem() > 0) {
    pCluster = daClusterStack.GetLast(); daClusterStack.DeAppend();  // pop
    if (!pCluster->fNodeCluster) {
      // follow all sub-clusters
      for (i=0; i<pCluster->daSubClusters.NumElem(); ++i) {
	pSubCluster = pCluster->daSubClusters[i];
	daClusterStack.Append(pSubCluster);  // push
      }
      continue;
    }
    // go through the nodes
    for (i=0; i<pCluster->daNodes.NumElem(); ++i) {
      pNode = pCluster->daNodes[i];
      double timeExtent, timeBusy;
      ScaleCpuTimes(pAppNode, pNode->smp.rMips,
		    &timeExtent, &timeBusy, 0.0);
      rTmp = timeExtent / pAppNode->pShared->wNumber_Cycle;
      //cout << "LB_aux: " << pAppNode->sbName << " extent=" << timeExtent
      //<< " 1 cycle=" << rTmp << endl;
      if (!fAny || rEFT > rTmp) {  // min
	fAny = true;
	rEFT = rTmp;
      }
    }
  }
  //cout << "LB_aux: max 1 cycle: " << rEFT << endl;
  return rEFT;
}


//
//  ps output stuff
//


// output the current AppNode graph as dot input
//
// external software: dot (part of graph vis package)
// installed: scavenger cluster piranha.cs.umd.edu:~beynon/util/gv1.4/...
// url: http://www.research.att.com/sw/tools/graphviz/
int DC_App_t::PsOutput(char *sbFile) {
  cout << "PsOutput: Creating \"" << sbFile << "\" for component graph ..."
       << endl;
  const unsigned int wNetMin = 2;
  unsigned int i, j, iLink;
  ofstream fout;
  fout.open(sbFile, ios::out);
  if (fout.fail()) return 1;

  fout << "// automatically generated dot file" << endl
       << "// DataCutter planner - AppNode graph" << endl
       << "digraph G {" << endl
       << "  rankdir=LR;" << endl;
  for (i=0; i<daNodes.NumElem(); ++i) {  // nodes
    DC_AppNode *pAppNode = daNodes[i];
    fout << "  " << pAppNode->sbName << '_' << pAppNode->wCopyNum
	 << " [label=\"" << pAppNode->sbName;
#if 0 /* cannot control the order the copies appear in, so forget number */
    if (pAppNode->pShared->nCopies > 1)
      fout << " " << pAppNode->wCopyNum;
#endif
    fout << "\"]" << ';' << endl;
  }
  for (iLink=0; iLink<daLinks.NumElem(); ++iLink) {  // nets
    DC_AppLink *pAppLink = daLinks[iLink];
    if (pAppLink->daSrc.NumElem() > wNetMin ||
	pAppLink->daDst.NumElem() > wNetMin) {
      fout << "  net_" << (unsigned int)pAppLink
	   << " [label=\"\",shape=circle,style=filled,color=black"
	   << ",height=.05,width=.05]"
	   << ';' << endl;
    }
  }
  for (iLink=0; iLink<daLinks.NumElem(); ++iLink) {  // edges
    DC_AppLink *pAppLink = daLinks[iLink];
    DC_AppNode *pAppNodeSrc, *pAppNodeDst;
    if (pAppLink->daSrc.NumElem() > wNetMin ||
	pAppLink->daDst.NumElem() > wNetMin) {  // use a net?
      for (i=0; i<pAppLink->daSrc.NumElem(); ++i) {
	pAppNodeSrc = pAppLink->daSrc[i];
	fout << "  " << pAppNodeSrc->sbName << '_' << pAppNodeSrc->wCopyNum
	     << "->" << "net_" << (unsigned int)pAppLink
	     << " [dir=none]"
	     << ';';
      }
      for (j=0; j<pAppLink->daDst.NumElem(); ++j) {
	pAppNodeDst = pAppLink->daDst[j];
	fout << "  " << "net_" << (unsigned int)pAppLink
	     << "->" << pAppNodeDst->sbName << '_' << pAppNodeDst->wCopyNum
	     << ';' << endl;
      }
    } else {  // many to many?
      for (i=0; i<pAppLink->daSrc.NumElem(); ++i) {
	pAppNodeSrc = pAppLink->daSrc[i];
	for (j=0; j<pAppLink->daDst.NumElem(); ++j) {
	  pAppNodeDst = pAppLink->daDst[j];
	  fout << "  " << pAppNodeSrc->sbName << '_' << pAppNodeSrc->wCopyNum
	       << "->" << pAppNodeDst->sbName << '_' << pAppNodeDst->wCopyNum
	       << ';' << endl;
	}
      }
    }
  }
  fout << "}" << endl;

  fout.close();
  return fout.fail() ? 1 : 0;
}

// output the current Topology graph as dot input
//
// external software: dot (part of graph vis package)
// installed: scavenger cluster piranha.cs.umd.edu:~beynon/util/gv1.4/...
// url: http://www.research.att.com/sw/tools/graphviz/
int DC_Topology_t::PsOutput_Aux(ofstream &fout, DC_TopologyCluster *pCluster,
				int wLevel) {
  unsigned int i;
  for (i=wLevel*2; i; --i) fout << ' ';  // indent
  fout << "subgraph \"cluster_" << pCluster->sbName << "\" {" << endl;
  for (i=wLevel*2; i; --i) fout << ' ';  // indent
  fout << "  rankdir=" << ((wLevel == wLevel/2*2) ? "TB":"LR") << ';'
       << " margin=\".5,.5\";" << endl;
  if (pCluster->fNodeCluster) {
    for (i=wLevel*2; i; --i) fout << ' ';  // indent
    if (pCluster->daNodes.NumElem() > 1) {  // use node ... node
      fout << "  " << pCluster->daNodes[0]->sbName << "->"
	   << pCluster->daNodes.GetLast()->sbName
	   << " [dir=none,style=dotted];" << endl;
    } else {
      fout << "  " << pCluster->daNodes[0]->sbName << ';' << endl;
    }
  } else {
    for (i=0; i<pCluster->daSubClusters.NumElem(); ++i) {
      PsOutput_Aux(fout, pCluster->daSubClusters[i], wLevel+1);
    }
  }
  for (i=wLevel*2; i; --i) fout << ' ';  // indent
  fout << "  label=\"" << pCluster->sbName << "\";" << endl;
  for (i=wLevel*2; i; --i) fout << ' ';  // indent
  fout << '}' << endl;
  return 0;
}
int DC_Topology_t::PsOutput(char *sbFile) {
  cout << "PsOutput: Creating \"" << sbFile << "\" for topology graph ..."
       << endl;
  ofstream fout;
  fout.open(sbFile, ios::out);
  if (fout.fail()) return 1;

  fout << "// automatically generated dot file" << endl
       << "// DataCutter planner - cluster topology graph" << endl
       << "digraph G {" << endl;
  for (unsigned int i=0; i<daClusterRoot.NumElem(); ++i) {
    PsOutput_Aux(fout, daClusterRoot[i], 1);
  }
  fout << "}" << endl;

  fout.close();
  return fout.fail() ? 1 : 0;
}

// output the current util schedule to the given file as Ploticus input
//
// external software: ploticus 2.00 (with my cluster-segment-bar patches)
// installed: scavenger cluster piranha.cs.umd.edu:~beynon/util/ploticus/...
// url: http://www.sgpr.net/
template <class T>
void DC_CapacityModel<T>::PsOutputUtil(ostream &os, char *sbName) {
  for (int iResource=0; iResource<nResources; ++iResource) {
    for (unsigned int i=0; i<rgTimeline[iResource].da.NumElem(); ++i) {
      CapacityTimelineEntry *pentry = rgTimeline[iResource].da[i];

      if (pentry->rCapacity == rCapacityMax) continue;  // skip empty

      os << "  " << sbName << ' ' << iResource << ' '
	 << pentry->timeStart << ' '
	 << (pentry->timeStart + pentry->timeExtent)
	 << " \"" << (rCapacityMax - pentry->rCapacity) << ' '
	//<< '{'
	;
      
      for (unsigned int j=0; j<pentry->daMembers.NumElem(); ++j) {
	MemberDetail *pMember = pentry->daMembers[j];
	if (j > 0) os << ',';
	if (pMember->pMemberObj) {
	  os << pMember->pMemberObj->sbName;
	  //pMember->pAppNode->PrintFullName(os);
	  //os << ' ' << pMember->rCapacityUse;
	} else {
	  os << "?";
	}
      }
      os
	//<< '}'
	<< '\"' << endl;
    }
  }
};
int PsOutputUtil_Aux(ostream &os, DC_TopologyCluster *pCluster,
		     DArrayPtr<DC_TopologyNode*> &daSchedNodes,
		     int &wMaxCpu, double &rMaxTime) {
  unsigned int i;
  if (pCluster->fNodeCluster) {
    for (i=0; i<pCluster->daNodes.NumElem(); ++i) {
      DC_TopologyNode *pNode = pCluster->daNodes[i];
      if (pNode->smp.isAllEmpty()) continue;

      daSchedNodes.AppendIfNotExist(pNode);
      if (wMaxCpu < pNode->smp.nResources) wMaxCpu = pNode->smp.nResources;

      double rTmp = pNode->smp.getMaxTime();
      if (rMaxTime < rTmp) rMaxTime = rTmp;

      // output this one
      pNode->smp.PsOutputUtil(os, pNode->sbName);
    }
  } else {
    for (i=0; i<pCluster->daSubClusters.NumElem(); ++i) {
      PsOutputUtil_Aux(os, pCluster->daSubClusters[i], daSchedNodes,
		       wMaxCpu, rMaxTime);
    }
  }
  return 0;  
}
int PsOutputUtil(DC_Topology_t &topology, DC_App_t &app, char *sbFile) {
  cout << "PsOutputUtil: Creating \"" << sbFile << "\" for schedule util ..."
       << endl;
  unsigned int i;
  ofstream fout;
  fout.open(sbFile, ios::out);
  if (fout.fail()) return 1;

  fout.setf(ios::fixed, ios::floatfield);

  fout << "// automatically generated ploticus file" << endl
       << "// DataCutter planner - runtime utilization graph" << endl;

  // proc getdata
  fout << "#proc getdata" << endl
       << "data:" << endl
       << setprecision(2);
  // build a set of hosts that have something scheduled
  DArrayPtr<DC_TopologyNode*> daSchedNodes;
  int wMaxCpu = 0;
  double rMaxTime = 0.0;
  for (i=0; i<topology.daClusterRoot.NumElem(); ++i) {
    PsOutputUtil_Aux(fout, topology.daClusterRoot[i], daSchedNodes,
		     wMaxCpu, rMaxTime);
  }
  fout << endl;

  int wTics;
  int wXRangeMax;
  if (rMaxTime <= 10.0) {
    wTics = 1;
    wXRangeMax = 10;
  } else if (rMaxTime <= 25.0) {
    wTics = 5;
    wXRangeMax = 25;
  } else if (rMaxTime <= 50.0) {
    wTics = 10;
    wXRangeMax = 50;
  } else if (rMaxTime <= 100.0) {
    wTics = 10;
    wXRangeMax = 100;
  } else if (rMaxTime <= 500.0) {
    wTics = 100;
    wXRangeMax = 500;
  } else if (rMaxTime <= 250.0) {
    wTics = 100;
    wXRangeMax = 250;
  } else if (rMaxTime <= 1000.0) {
    wTics = 100;
    wXRangeMax = 1000;
  } else {
    wTics = 200;
    wXRangeMax = (int)rMaxTime;    
  }

  // proc areadef and axes
  fout << "#proc areadef" << endl;
  if (Global.fDoScheduling) {
    fout << "  title: Schedule Host Utilization "
	 << " (" << Global.sbFeatureString << ')'
	 << endl;
  } else {
    fout << "  title: Runtime Host Utilization" << endl;
  }
  fout << "  rectangle: 1 1 5 "
       << (0.2 * (double)(daSchedNodes.NumElem() * wMaxCpu) + 1.0) << endl
       << "  //xscaletype: time" << endl
       << "  xrange: 0.0 " << wXRangeMax << endl;
  fout << "  yscaletype: categories" << endl
       << "  ycategories: " << endl;
  for (i=0; i<daSchedNodes.NumElem(); ++i) {
    fout << "    " << daSchedNodes[i]->sbName << endl;
  }
  fout << endl
       << "#proc xaxis" << endl
       << "  stubs: inc " << wTics << endl
       << endl
       << "#proc yaxis" << endl
       << "  stubs: usecategories" << endl
       << endl;

  // proc bars
  for (i=0; i<(unsigned int)wMaxCpu; ++i) {
    fout << "#proc bars" << endl
	 << "  select: @2 = " << i << endl
	 << "  axis: x" << endl
	 << "  locfield: 1" << endl
	 << "  segmentfields: 3 4" << endl
	 << "  color: gray(0.8)" << endl
	 << "  cluster: " << i+1 << " / " << wMaxCpu << endl
	 << "  labelfield: 5" << endl
	 << "  longwayslabel: yes" << endl
	 << "  labeldetails: adjust=0,0.02 size=6" << endl
	 << endl;
  }

  fout.close();
  return fout.fail() ? 1 : 0;
}


// output the current util schedule to the given file as Ploticus input
//
// external software: ploticus 2.00 (with my cluster-segment-bar patches)
// installed: scavenger cluster piranha.cs.umd.edu:~beynon/util/ploticus/...
// url: http://www.sgpr.net/
int PsOutputSched(DC_Topology_t &topology, DC_App_t &app, char *sbFile) {
  cout << "PsOutputSched: Creating \"" << sbFile << "\" for schedule ..."
       << endl;
  unsigned int i;
  ofstream fout;
  fout.open(sbFile, ios::out);
  if (fout.fail()) return 1;

  fout.setf(ios::fixed, ios::floatfield);

  fout << "// automatically generated ploticus file" << endl
       << "// DataCutter planner - runtime schedule graph" << endl;

  // proc getdata
  fout << "#proc getdata" << endl
       << "data:" << endl
       << setprecision(2);
  // build a set of hosts that have something scheduled
  DArrayPtr<DC_TopologyNode*> daSchedNodes;
  int nItems = 0;
  double rMaxTime = 0.0;
  for (i=0; i<app.daNodes.NumElem(); ++i) {
    DC_AppNode *pAppNode = app.daNodes[i];
    fout << "  ";
    pAppNode->PrintFullName(fout);
    fout << ' '
	 << pAppNode->wCopyNum << ' '
	 << pAppNode->rEST << ' ' << pAppNode->rEFT << ' '
	 << '\"' << pAppNode->sbName
	 << " (" << pAppNode->pNodeMap->sbName << ")\""
	 << endl;
    if (rMaxTime < pAppNode->rEFT) rMaxTime = pAppNode->rEFT;
    ++nItems;
  }
  fout << endl;

  int wTics;
  int wXRangeMax;
  if (rMaxTime <= 10.0) {
    wTics = 1;
    wXRangeMax = 10;
  } else if (rMaxTime <= 25.0) {
    wTics = 5;
    wXRangeMax = 25;
  } else if (rMaxTime <= 50.0) {
    wTics = 10;
    wXRangeMax = 50;
  } else if (rMaxTime <= 100.0) {
    wTics = 10;
    wXRangeMax = 100;
  } else if (rMaxTime <= 500.0) {
    wTics = 100;
    wXRangeMax = 500;
  } else if (rMaxTime <= 250.0) {
    wTics = 100;
    wXRangeMax = 250;
  } else if (rMaxTime <= 1000.0) {
    wTics = 100;
    wXRangeMax = 1000;
  } else {
    wTics = 200;
    wXRangeMax = (int)rMaxTime;    
  }

  // proc areadef and axes
  fout << "#proc areadef" << endl;
  if (Global.fDoScheduling) {
    fout << "  title: Schedule Timing "
	 << " (" << Global.sbFeatureString << ") "
	 << "EFT=" << app.rEFT;
    if (Global.wFeatures & PLANNER_USE_PIPE) {
      fout << " LB=" << app.rOptEFT;
    }
    fout << endl;
  } else {
    fout << "  title: Runtime Timing "
	 << "EFT=" << app.rEFT
	 << endl;
  }
  fout << "  rectangle: 1 1 5 " << (0.2 * (double)nItems + 1.0) << endl
       << "  xrange: 0.0 " << wXRangeMax << endl
       << "  yrange: 0 " << nItems+1 << endl
       << endl
       << "#proc xaxis" << endl
       << "  stubs: inc " << wTics << endl
       << endl
       << "#proc yaxis" << endl
       << "  tics: none" << endl
       << "  stubs: none" << endl
       << endl;

  // proc bars
  fout << "#proc bars" << endl
       << "  axis: x" << endl
       << "  segmentfields: 3 4" << endl
       << "  color: gray(0.8)" << endl
       << "  labelfield: 5" << endl
       << "  longwayslabel: yes" << endl
       << "  labeldetails: adjust=0,0.02 size=6" << endl
       << endl;

  fout.close();
  return fout.fail() ? 1 : 0;
}

// output the current placement spec as LaTeX input
//
// external software: LaTeX
int PsOutputPlacement(ConfigInfo &cfgPlacement, char *sbFile) {
  ofstream fout;
  fout.open(sbFile, ios::out);
  if (fout.fail()) return 1;

  fout << "% automatically generated LaTeX file" << endl
       << "% DataCutter planner - placement spec" << endl;

  fout << "\\begin{tabular}{|lcl|}" << endl
       << "\\hline" << endl;

  const char *sbSection, *sbName;

  sbSection = cfgPlacement.FirstSection();
  while (sbSection != NULL) {
    fout << "\\multicolumn{3}{|l|}{\\bf [" << sbSection << "]} \\\\"
	 << endl;
    sbName = cfgPlacement.FirstName();
    while (sbName != NULL) {
      fout << sbName << " & = & " << cfgPlacement.CurrentValue() << " \\\\"
	   << endl;
      sbName = cfgPlacement.NextName();
    }
    sbSection = cfgPlacement.NextSection();
  }

  fout << "\\hline" << endl
       << "\\end{tabular}" << endl;

 fout.close();
  return fout.fail() ? 1 : 0;
}


// read the tuning run, and pre-load the app and topology as needed to
// pretend the scheduler produced output.  This corresponds to
// Global.fDoScheduling == false.  Some of the psoutput charts won't be
// sensical, because for example the smp model is not really scheduled, so
// the host utilization timing graph will be empty.
int ShowRun(DC_Topology_t &topology, DC_App_t &app, ConfigInfo &cfgTuneRun) {
  const int maxBuf = 80;
  char buf[maxBuf];

  cout << cfgTuneRun << endl;

  unsigned int i;
  DC_AppNode *pAppNode, *pAppNodeTmp;

  // init as "to be processed"
  app.reset_color(COLOR_BLACK);

  for (i=0; i<app.daNodes.NumElem(); ++i) {
    pAppNode = app.daNodes[i];
    if (pAppNode->eColor != COLOR_BLACK) continue;
    pAppNode->eColor = COLOR_WHITE;

    cout << "appnode=" << pAppNode->sbName << endl;
    // look through run output, and see how many were run
    snprintf(buf, maxBuf, "filter.%s", pAppNode->sbName);

    int nCopies = 0;
    const char *sbSection, *sbName;
    sbSection = cfgTuneRun.FirstSection();
    while (sbSection != NULL) {
      if (strcmp(sbSection, buf) == 0) {
	if (nCopies++ == 0) {
	  pAppNodeTmp = pAppNode;
	} else {
	  pAppNodeTmp = app.CreateCopy(pAppNode);
	}
	DC_TopologyCluster *pCluster = NULL;
	DC_TopologyNode    *pNode = NULL;
	double rProcess = 0.0, rReadFirst = 0.0, rWriteFirst = 0.0;
	bool fProducer = false;
	// walk through all name value entries and grab values we need
	sbName = cfgTuneRun.FirstName();
	while (sbName != NULL) {
	  cout << "- " << sbName << " = " << cfgTuneRun.CurrentValue() << endl;
	  if (strcmp(sbName, "placement") == 0) {
	    pCluster = 
	      topology.FindHost(cfgTuneRun.Find(buf,
						cfgTuneRun.CurrentValue()));
	    pAppNode->pShared->pNodeInfoTuning = &pCluster->NodeInfo;

	    pCluster = topology.FindHost(cfgTuneRun.CurrentValue(), &pNode);
	    assert(pCluster);
	    assert(pCluster->fNodeCluster);
	  } else if (strcmp(sbName, "process_time") == 0) {
	    rProcess = Atof(cfgTuneRun.CurrentValue());
	  } else if (strcmp(sbName, "process.readfirst_time") == 0) {
	    rReadFirst = Atof(cfgTuneRun.CurrentValue());
	  } else if (strcmp(sbName, "process.writefirst_time") == 0) {
	    rWriteFirst = Atof(cfgTuneRun.CurrentValue());
	  } else if (strcmp(sbName, "ins") == 0) {
	    if (strlen(cfgTuneRun.CurrentValue()) == 0) fProducer = true;
	  }
	  sbName = cfgTuneRun.NextName();
	}

	assert(pNode);
	pAppNodeTmp->pNodeMap = pNode;
	pAppNodeTmp->nCopiesMap = 1;
	pAppNodeTmp->rEST = (fProducer ? 0.0 : rReadFirst);
	pAppNodeTmp->rEFT = rProcess;
	pAppNodeTmp->rEWT = (rWriteFirst > 0.0 ? rWriteFirst : 0.0);
      }
      sbSection = cfgTuneRun.NextSection();
    }
    
  }
  return 0;
}


//
//  main
//

int usage(char *sbProg, char *sbMsg=NULL) {
  if (sbMsg) cerr << "error: \"" << sbMsg << '\"' << endl;
  cerr << "usage: " << sbProg
       << " [-ps] [-showrun]"
       << " <topo_cfg> <layout_cfg> <tunerun_cfgout> <flist>" << endl
       << "where"
       << "\t-ps = turns on psoutput of app graphs and schedule" << endl
       << "\t-showrun = instead of scheduling, use real run info" << endl
       << endl
       << "\t<topo_cfg>    = heirarchical cluster topology" << endl
       << "\t<layout_cfg>  = filter layout of stream connectivity" << endl
       << "\t<tunerun_cfg> = filterexit stats from a tuning run" << endl
       << "\t<flist> = set of features" << endl
       << "\t    +<feature>[+<feature>...] : manual list of features" << endl
       << "\t\t+ALL\t: instead of pruning clusters, try all clusters" << endl
       << "\t\t+PIPE\t: use EWT instead of EFT for pipelined execution" << endl
       << "\t\t+NET\t: model network as a shared capacity resource" << endl
       << "\t\t+COPY\t: create copies based on tuning run assumption" << endl
       << "\t\t+COPY2\t: create copies iteratively with real mappings" << endl
       << "\t\t+REFINE\t: final pass to fix greedy artifacts" << endl
       << endl;
  exit(1);
  return 1;  // not needed
}

int main(int argc, char *argv[], char *envp[]) {
  DC_Topology_t topology;
  DC_App_t app;
  ConfigInfo cfgTuneRun, cfgPlacement;
  Timer t;
  const int maxBuf = 80;
  char bufTmp[maxBuf];

  (void) StandardInit(argc, argv);
  DBG_InstallCoreSigHandlers(argv[0], envp);
  //cout.setf(ios::fixed, ios::floatfield);

  // look for options
  bool fPsOutput = false;
  Global.fDoScheduling = true;
  int iArg;
  for (iArg=1; iArg<argc; ++iArg) {
    if (argv[iArg][0] != '-') break;  // done with options
    if (strcmp(argv[iArg], "-ps") == 0) {
      fPsOutput = true;
    } else if (strcmp(argv[iArg], "-showrun") == 0) {
      Global.fDoScheduling = false;
    } else {
      snprintf(bufTmp, maxBuf, "unknown option '%s'", argv[iArg]);
      usage(argv[0], bufTmp);
    }
  }
  int nArg = argc - iArg;
  if (nArg < 1) usage(argv[0], "missing <topo_cfg> argument");
  if (nArg < 2) usage(argv[0], "missing <layout_cfg> argument");
  if (nArg < 3) usage(argv[0], "missing <tune_cfg> argument");
  if (Global.fDoScheduling) {
    if (nArg < 4) usage(argv[0], "missing <map_sched_alg> argument");
  }

  if (topology.ReadFile(argv[iArg+0]) != 0)
    usage(argv[0], "invalid <topo_cfg> file");
  if (app.ReadFile(argv[iArg+1]) != 0)
    usage(argv[0], "invalid <layout_cfg> file");
  // since filterexit stats can contain duplicated sections, set cfg opts
  cfgTuneRun.ClrOpt(ConfigInfo::ADD_DUPSEC_USEOLD);
  cfgTuneRun.ClrOpt(ConfigInfo::ADD_DUPSEC_REPLACE);
  cfgTuneRun.SetOpt(ConfigInfo::ADD_DUPSEC_APPEND);
  if (cfgTuneRun.ReadFile(argv[iArg+2]) != 0)
    usage(argv[0], "invalid <tune_cfg> file");

  Global.sbFeatureString = argv[iArg+3];
  Global.wFeatures = 0;
  if (Global.fDoScheduling) {
    if (argv[iArg+3][0] == '+') {  // fmt: +<feature>[+<feature>...]
      Str str(argv[iArg+3]+1);
      char *sb;
      ++Global.sbFeatureString;  // skip + in psoutput
      while ((sb = str.TokenNoDup("+"))) {
	if (strcmp(sb, "ALL") == 0)
	  Global.wFeatures |= PLANNER_USE_ALL;
	else if (strcmp(sb, "PIPE") == 0)
	  Global.wFeatures |= PLANNER_USE_PIPE;
	else if (strcmp(sb, "NET") == 0)
	  Global.wFeatures |= PLANNER_NETMODEL;
	else if (strcmp(sb, "COPY") == 0)
	  Global.wFeatures |= PLANNER_COPIES_1;
	else if (strcmp(sb, "COPY2") == 0)
	  Global.wFeatures |= PLANNER_COPIES_2;
	else if (strcmp(sb, "REFINE") == 0)
	  Global.wFeatures |= PLANNER_REFINE;
	else {
	  snprintf(bufTmp, maxBuf, "unknown <map_sched_alg> feature '%s'", sb);
	  usage(argv[0], bufTmp);
	}
      }
    }
  }

  topology.Process();
  //cout << "*** DEBUG: topology..." << endl << topology << endl;
  app.Process(cfgTuneRun);
  topology.ProcessApp(app, cfgTuneRun);  // also creates pinned copies
  app.ComputeLevels();
  //cout << "*** DEBUG: app..." << endl << app << endl;

  if (fPsOutput) {
    topology.PsOutput("psoutput-topology.dot");
    app.PsOutput("psoutput-app.dot");
  }

  app.rOptEFT = topology.ComputeLowerBound(app);
  cout << "OptEFT = " << app.rOptEFT << endl;

  int ret;
  if (Global.fDoScheduling) {
    //cout << "*** DEBUG: main(): calling planner()..." << endl;
    t.Start();
    ret = planner(topology, app, cfgTuneRun);
    t.Stop();
    cout << "tPlanner = " << t << " seconds" << endl;
  } else {
    ret = ShowRun(topology, app, cfgTuneRun);
  }

  app.rEFT = app.GetMakeSpan();
  cout << "EFT = " << app.rEFT << endl;

  GeneratePlacement(cfgPlacement, topology, app, cfgTuneRun);
  if (fPsOutput) {
    app.PsOutput("psoutput-app-final.dot");
    PsOutputUtil(topology, app, "psoutput-schedule-util.plot");
    PsOutputSched(topology, app, "psoutput-schedule.plot");
    PsOutputPlacement(cfgPlacement, "psoutput-placement.tex");
  }
  cout << "*** DEBUG: output placement..." << endl << cfgPlacement << endl;
  OutputPlacement(cfgPlacement, "planner_placement.cfg",
		  argc, argv, t.Elapsed(), app.rEFT, app.rOptEFT);

#if 0
  // testing smp model stuff ...
  DC_SmpModel smp;
  double r;
  smp.init(2, 1000, 2048);
  smp.Dump("start");
  r = smp.Assign(20, 50, 10, NULL); cout << "EFT=" << r << endl;
  smp.Dump("after assign(20, 50, 10)");
  r = smp.Assign(20, 50, 10, NULL); cout << "EFT=" << r << endl;
  smp.Dump("after assign(20, 50, 10)");
  r = smp.Assign(40, 40, 10, NULL); cout << "EFT=" << r << endl;
  smp.Dump("after assign(40, 40, 10)");
  r = smp.Assign(90, 10, 10, NULL); cout << "EFT=" << r << endl;
  smp.Dump("after assign(90, 10, 10)");
  r = smp.Assign(10, 200, 20, NULL); cout << "EFT=" << r << endl;
  smp.Dump("after assign(10, 200, 20)");
  r = smp.Assign(10, 100, 50, NULL); cout << "EFT=" << r << endl;
  smp.Dump("after assign(10, 100, 50)");
  r = smp.Assign(10, 100, 50, NULL); cout << "EFT=" << r << endl;
  smp.Dump("after assign(10, 100, 50)");
  smp.UnAssign(NULL);
  smp.Dump("after unassign(NULL)");
  smp.UnAssign(NULL);
  smp.Dump("after unassign(NULL)");
#endif

  cout << "*** DEBUG: main(): done, ret=" << ret << endl;
  return ret;
}


// ****************************************************************************
