// ****************************************************************************
//  Name: remoteprocessstate.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 08/12/2000 : mdb - Created.
//        11/29/2000 : mdb - Converted to use DC_Mutex/DC_Cond.
//        02/22/2001 : mdb - Started StreamBufferQueue to handle copies.
//        03/07/2001 : mdb - Rewritten many part to start filter copies.
//        03/13/2001 : mdb - Sweeping changes to support multiple copies.
//        04/12/2001 : mdb - Fixed bad bug in WriteBuffer_DATA() where we read
//                           the orig pbuf size after a colocated handoff.
//                         - Changed WriteBuffer_DATA() to allow specifying if
//                           a handoff is needed, and auto-create a copy only
//                           when necessary.
//        07/17/2001 : mdb - Fixed omission of colocated producer DATA_ACKs.
//        07/27/2001 : mdb - Implemented stream write policy subsets.
//        01/10/2002 : mdb - Compute and transfer 3 sets of rank/max values.
//  Desc: State and supporting routines for the remote lib process.
// ****************************************************************************

#include "lib/dc_str.h"
#include "lib/dc_memory.h"
#include "lib/dc_threadclass.h"
#include "lib/dc_lists.h"
#include "lib/dc_random.h"

#include "remoteprocessstate.h"
#include "packet.h"
#include "packet-tags.h"
#include "thread-filter.h"
#include "thread-reader.h"
#include "util.h"
#include "dc_error.h"
#include "consoleprocessstate.h"  // needed for pseudo-filter setup signaling

#include <assert.h>

#ifdef _WIN32
#include "win32.h"
#endif

#if defined(DCDBG)
//#define DCDBG_FILTERSERVICE_REMOTEPROCESS
//#define DCDBG_FILTERSERVICE_HELLOACK
//#define DCDBG_FILTERSERVICE_STARTINSTANCE
//#define DCDBG_FILTERSERVICE_STREAMCONNECT
//#define DCDBG_FILTERSERVICE_APPENDWORK
//#define DCDBG_FILTERSERVICE_WRITEBUFFER
//#define DCDBG_FILTERSERVICE_DATA
//#define DCDBG_FILTERSERVICE_DATA_DUMPBUF
//#define DCDBG_FILTERSERVICE_CLEANUP_INSTANCE
#endif /* DCDBG */


/** 
 * ctor
 */
DC_RemoteProcessState::DC_RemoteProcessState(DC_FilterService_t *pDC_in)
  : pDC(pDC_in), fInit(false), sbHostMe(NULL), lcFullWorkQs(3) {
}

/**
 * dtor
 */
DC_RemoteProcessState::~DC_RemoteProcessState(void) {
  // wait for all running filters to exit?
  // No -- we wait for the console process to dictate exit, and the
  // reader thread exits, which the main thread joins with.

  if (fInit) {
    delete [] sbHostMe;
  }
}

/**
 *  Contact console process, and pass our contact info.  We have no
 *  knowledge of any filter instances to run yet -- that comes later
 *  when the console node creates a filter instance and sends us the
 *  STARTINSTANCE command.
 */
int DC_RemoteProcessState::init(char *sbArg) {
  char *sbHostConsole;
  int   wPortConsole;

  fInit = true;

  // parse command line arg
  //   format: <iplacement>,<consosle_host>,<console_port>,<my_hostname>
  //
  // iplacement - console process index for us, not used in remote process
  // console_host, console_port - contact info (hi) for console process
  // my_hostname - canonical hostname (hi) for our placement
  Str str(sbArg);
  iPlacement = atoi(str.TokenNoDup(","));
  sbHostConsole = str.TokenNoDup(",");
  wPortConsole = atoi(str.TokenNoDup(","));
  sbHostMe = str.Token(",");

#if defined(DCDBG_FILTERSERVICE_REMOTEPROCESS)
  cout << "DC_RemoteProcessState::init(): iPlacement=" << iPlacement
       << ", console=" << sbHostConsole << ':' << wPortConsole
       << ", me=" << sbHostMe
       << endl;
#endif /* DCDBG_FILTERSERVICE_REMOTEPROCESS */

  // build packet with our peer contact info
  MemoryBuf mb(KB_2);
  mb.sprintf("[datacutter.%s.\"%s\"]\n",
	     pDC->sbAppName, pDC->addrLocal.getHostName());
  mb.sprintf("iplacement=%d\n", iPlacement);
  mb.sprintf("pid=%ld\n", static_cast(long int,getpid()));
  mb.sprintf("dns_name=%s\n", pDC->addrLocal.getHostName());
  //mb.sprintf("ip_addr=%s\n", pDC->addrLocal.getHostIpAddress());
  mb.sprintf("port=%d\n", pDC->addrLocal.getPort());
  Packet pkt(TAG_FILTERSERVICE_HELLO, &mb);

  // connect to console (unlocked access ok -- nothing else running yet)
  if (sockConsole.Connect(sbHostConsole, wPortConsole, addrConsole) == 0) {
    cout << "Connected to console " << addrConsole << endl;
  } else {
    perror("DC_RemoteProcessState::init(): failed Connect() to console");
    cerr << addrConsole << endl;
    return -1;
  }

  // send HELLO packet to console
  if (pkt.write(sockConsole) <= 0) {
    cout << "DC_RemoteProcessState::init(): Failed write of HELLO packet"
	 << endl;
    sockConsole.close();
    return -2;
  }
#if defined(DCDBG_FILTERSERVICE_REMOTEPROCESS)
  cout << "Sent HELLO pkt to console" << endl;
#endif /* DCDBG_FILTERSERVICE_REMOTEPROCESS */

  // do not wait for HELLO_ACK here.  Doing this causes a subtle deadlock
  // for the pass-thru case, because StartRemoteProcesses can block on the
  // special <console> case.  We need to make sure this call does not block.
  return 0;
}


/**
 * search for the shared index of a stream
 */
DC_RemoteInStreamShared *
DC_RemoteProcessState::find_instream(char *sbName,
				     DArray<DC_RemoteFilterShared*> &da) {
  DC_RemoteFilterShared   *pFilterShared;
  DC_RemoteInStreamShared *pInStreamShared;
  Types::uint i, j;

  // found colocated output stream, now find corresponding input stream
  for (i=0; i<da.NumElem(); ++i) {
    pFilterShared = da[i];

    for (j=0; j<pFilterShared->daInStreamShared.NumElem(); ++j) {
      pInStreamShared = pFilterShared->daInStreamShared[j];
      if (strcmp(sbName, pInStreamShared->sbName) == 0)
	return pInStreamShared;
    }
  }
  return NULL;  // not found
}

/**
 * search for the copy index of a stream
 */
DC_RemoteInStreamCopy *DC_RemoteProcessState::find_instream(int wInstanceNum,
							    char *sbName) {
  DC_RemoteFilterCopy   *pFilter;
  DC_RemoteInStreamCopy *pInStream;
  Types::uint i, j;

  // found colocated output stream, now find corresponding input stream
  for (i=0; i<daFilterCopy.NumElem(); ++i) {
    pFilter = daFilterCopy[i];
    if (!pFilter || pFilter->pShared->wInstanceNum != wInstanceNum) continue;
    
    for (j=0; j<pFilter->daInStreamCopy.NumElem(); ++j) {
      pInStream = pFilter->daInStreamCopy[j];
      if (strcmp(sbName, pInStream->pShared->sbName) == 0)
	return pInStream;
    }
  }
  return NULL;  // not found
}

/**
 * merge console cfg, and perform initializations
 */
// thread: remote process reader (when recv HELLO_ACK)
//         or console process for pseudo-filter
int DC_RemoteProcessState::ProcessConsoleCfgSettings(void) {
  return 0;
}

/**
 * create new remoteXXXspec entries, spawn filter threads, check pending
 */
// thread: remote process reader
int DC_RemoteProcessState::StartInstance(int wInstanceNum, ConfigInfo &cfg,
					 BaseSocket *psock) {
  char bufSection[10];
  const int maxBufName = 80;
  char bufName[maxBufName];
  char *sb;
  int  ret;
  DC_RemoteFilterShared    *pFilterShared;
  DC_RemoteInStreamShared  *pInStreamShared;
  DC_RemoteOutStreamShared *pOutStreamShared;

  DArray<DC_RemoteFilterShared*> daFilterShared;

#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
  cout << "DC_RemoteProcessState::StartInstance(): entry ..." << endl;
  cout << cfg;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */

  //
  // Parse startintstance spec, create new state for this instance
  //
  Types::uint nFilters, nCopies, nins, nouts;
  Types::uint i, j, k;
  Types::uint wInitialWorkNum;

  nFilters = atoi(FindParm(cfg, "", "num_filters"));
  nCopies = atoi(FindParm(cfg, "", "num_copies"));

// UVC: TODO: solve this problem in Linux
      // (Linux doesn't compile with the following call)
// UVC: in Solaris 8; if you link with standard pthread library you need to
      // call the following function
//  pthread_setconcurrency(nCopies+1);
  
  
#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
  cout << "[reader] StartInstance(): For i" << wInstanceNum
       << ", we will run " << nFilters << " filters" << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */

  // common instance state
  DC_RemoteInstanceShared *pRemoteInstanceShared
    = new DC_RemoteInstanceShared(wInstanceNum, nFilters, nCopies);

  for (i=0; i<nFilters; ++i) {
    snprintf(bufSection, sizeof(bufSection), "%d", i);
    sb = FindParm(cfg, bufSection, "filter");

    // shared filter state
    pFilterShared = new DC_RemoteFilterShared(sb, wInstanceNum);
    pFilterShared->pRemoteInstanceShared = pRemoteInstanceShared;

    pFilterShared->fConsole =
      (FindParm(cfg, bufSection, "console", false) ? true : false);
    if (pFilterShared->fConsole && nFilters != 1) {
      // logic error - we assume a single pseudo-filter at the console
      cerr << "DC_RemoteProcessState::StartInstance(): "
	   << "console set with nFilters=" << nFilters << endl;
      assert(0);
      exit(1);
    }

    sb = FindParm(cfg, bufSection, "copies");
    pFilterShared->fAllowCopies = (strcmp(sb, "yes") == 0 ? true : false);
#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
    cout << "[reader] StartInstance(): Filter \"" << pFilterShared->sbName
	 << "\" " << (pFilterShared->fAllowCopies ? "allows" : "disallows")
	 << " copies" << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */
    pFilterShared->nInitialCopies
      = atoi(FindParm(cfg, bufSection, "numcopies_initial"));
    pFilterShared->rankCopySets
      = atoi(FindParm(cfg, bufSection, "rankcopysets"));
    pFilterShared->maxCopySets
      = atoi(FindParm(cfg, bufSection, "maxcopysets"));
    pFilterShared->rankCopiesStart
      = atoi(FindParm(cfg, bufSection, "rankcopies"));
    pFilterShared->maxCopies
      = atoi(FindParm(cfg, bufSection, "maxcopies"));

    wInitialWorkNum = atoi(FindParm(cfg, bufSection, "worknum"));
    int wQueueSize = atoi(FindParm(cfg, bufSection, "sink_queue_size"));

    snprintf(bufName, maxBufName, "WorkQ:i%d:%s",
	     wInstanceNum, pFilterShared->sbName);
    pFilterShared->WorkQ.init(bufName, atoi(FindParm(cfg, bufSection,
						     "work_queue_size")));
    pFilterShared->WorkQ.setInitialWorkNum(wInitialWorkNum);
    pFilterShared->WorkQ.setNumConsumers(pFilterShared->nInitialCopies);
    pFilterShared->WorkQ.setReaderServerLoop(pDC->pReader);
    pFilterShared->WorkQ.setCountFullWorkQs(&lcFullWorkQs);

    // shared input streams
    nins = atoi(FindParm(cfg, bufSection, "nins"));
    for (j=0; j<nins; ++j) {
      snprintf(bufName, maxBufName, "ins%d_name", j);
      sb = FindParm(cfg, bufSection, bufName);
      pInStreamShared = new DC_RemoteInStreamShared(sb, wInstanceNum);
      pInStreamShared->iStream
	= pFilterShared->daInStreamShared.Append(pInStreamShared);

      snprintf(bufName, maxBufName, "ins%d_filter", j);
      sb = FindParm(cfg, bufSection, bufName);
      pInStreamShared->sbFilterSource = StrDup(sb);

      snprintf(bufName, maxBufName, "SinkQ:i%d:%s",
	       wInstanceNum, pInStreamShared->sbName);
      pInStreamShared->SinkQ.init(bufName, wQueueSize,
				  &(pFilterShared->mutexAllIns),
				  &(pFilterShared->cvAllInsNotEmpty));
      pInStreamShared->SinkQ.setInitialWorkNum(wInitialWorkNum);

#if 0 /* MDB */
      // we need to know the writers policy here to produce ACKs for DD
      snprintf(bufName, maxBufName, "ins%d_write_policy", j);
      sb = FindParm(cfg, bufSection, bufName);
      DC_StreamWritePolicy_t ePolicy = DC_StreamWritePolicy_toenum(sb);
      pInStreamShared->SinkQ.setStreamWritePolicy(ePolicy);
#endif
      // note: for an ins we currently only need to know if the DD
      // policy is used, which will need acks

      // check inputs to setup SinkQ producer counts
      for (k=0; ; ++k) {
	snprintf(bufName, maxBufName, "ins%d_host%d", j, k);
	sb = FindParm(cfg, bufSection, bufName, false);
	if (!sb) break;

	snprintf(bufName, maxBufName, "ins%d_numcopies%d", j, k);
	int n = atoi(FindParm(cfg, bufSection, bufName));
	if (strcmp(sb, "<colocated>") == 0) {
	  //pInStreamShared->SinkQ.localAddProducer(n);
	  // we don't count these here, because we want in the future
	  // to dynamically add new local copies, so this should be
	  // self-contained in the StartInstance_NewCopy() fcn.
	} else {
	  // we count these here, because there is no way to know
	  // they exist w/o pre-allocating them, thus avoiding a race
	  // in SinkQ to decide on merged EOW boundaries
	  pInStreamShared->SinkQ.remoteAddProducer(n);
	}
      }
    }

    // shared output streams
    nouts = atoi(FindParm(cfg, bufSection, "nouts"));
    for (j=0; j<nouts; ++j) {
      snprintf(bufName, maxBufName, "outs%d_name", j);
      sb = FindParm(cfg, bufSection, bufName);
      pOutStreamShared = new DC_RemoteOutStreamShared(sb, wInstanceNum);
      pOutStreamShared->iStream
	= pFilterShared->daOutStreamShared.Append(pOutStreamShared);

      snprintf(bufName, maxBufName, "outs%d_filter", j);
      sb = FindParm(cfg, bufSection, bufName);
      pOutStreamShared->sbFilterSink = StrDup(sb);

      snprintf(bufName, maxBufName, "outs%d_write_policy", j);
      sb = FindParm(cfg, bufSection, bufName);
      pOutStreamShared->eStreamWritePolicy = DC_StreamWritePolicy_toenum(sb);
      if (pOutStreamShared->eStreamWritePolicy == DC_StreamWritePolicy_UD) {
	snprintf(bufName, maxBufName, "outs%d_write_policy_UD", j);
	sb = FindParm(cfg, bufSection, bufName);
	// create policy object (once for all copies)
	if (pDC->fcnPolicyFactory) {
	  pOutStreamShared->pWritePolicyUD
	    = (*(pDC->fcnPolicyFactory))(sb);
	  if (pOutStreamShared->pWritePolicyUD == NULL) {
	    cerr << "*** Error: Specified UD write policy \"" << sb
		 << "\", but policy factory returned NULL to create it"
		 << endl;
	    return DC_ERR_Error;  // TODO: specific error for this case
	  }
	} else {
	  cerr << "*** Error: Specified UD write policy \"" << sb
	       << "\", but no policy factory defined"
	       << endl;
	  return DC_ERR_Error;  // TODO: specific error for this case
	}
      }

      // check for all outputs to copy sets
      for (k=0; ; ++k) {
	snprintf(bufName, maxBufName, "outs%d_host%d", j, k);
	sb = FindParm(cfg, bufSection, bufName, false);
	if (!sb) break;

	snprintf(bufName, maxBufName, "outs%d_numcopies%d", j, k);
	int n = atoi(FindParm(cfg, bufSection, bufName));
	DC_RemoteOutStreamShared::ConnInfo_t *pConnInfo =
	  new DC_RemoteOutStreamShared::ConnInfo_t;
	if (strcmp(sb, "<colocated>") == 0) {
	  pConnInfo->fColocated = true;
	  pOutStreamShared->pInStreamShared
	    = (DC_RemoteInStreamShared*)1; // dummy value to flag below
	} else {
	  pConnInfo->fColocated = false;
	  snprintf(bufName, maxBufName, "outs%d_port%d", j, k);
	  pConnInfo->addr.Set(sb, atoi(FindParm(cfg, bufSection, bufName)));
	}
	pConnInfo->nCopies = n;
	pOutStreamShared->daConnInfo.Append(pConnInfo);
      }
    }

    // save ptr to be able to fillin cross references
    daFilterShared.Append(pFilterShared);
  }

  // fillin colocated cross references
  for (i=0; i<daFilterShared.NumElem(); ++i) {
    pFilterShared = daFilterShared[i];

    for (j=0; j<pFilterShared->daOutStreamShared.NumElem(); ++j) {
      pOutStreamShared = pFilterShared->daOutStreamShared[j];
      if (pOutStreamShared->pInStreamShared == NULL) continue;

      pOutStreamShared->pInStreamShared
	= find_instream(pOutStreamShared->sbName, daFilterShared);
    }
  }


  // At this point, all the shared data is created for the entire
  // instance.  Now generate the actual copies using the shared data.  

  mutexFilterCopy.lock();

  for (i=0; i<daFilterShared.NumElem(); ++i) {
    pFilterShared = daFilterShared[i];

    StartInstance_AddNewCopies(pFilterShared,
			       pFilterShared->nInitialCopies,
			       wInitialWorkNum);
  }

  mutexFilterCopy.unlock();

  // check pending list for this new instance
#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
  cout << "[reader] StartInstance(): checking pending list for i"
       << wInstanceNum << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */
  // no need to lock daPending because it is only accessed through this
  // reader thread.
  DC_RemotePending *pPending;
  for (k=0; k<daPending.NumElem(); ++k) {
    pPending = daPending[k];
    if (! pPending || wInstanceNum != pPending->wInstanceNum) continue;

    // this existing premature connect request is for this new instance
    ret = StreamConnect(wInstanceNum, pPending->sbStreamName,
			pPending->iConnInfo, pPending->eStreamWritePolicy,
			pPending->pMap, pPending->psock,
			false);
    if (ret != 0) { // this should never happen!
      cerr << ">>> FATAL ERROR: DC_RemoteProcessState::StreamConnect(): "
	   << "connect to new instance failed" << endl;
      exit(1);
    }
    // send ack
    ret = StreamConnectAck(wInstanceNum, pPending->iConnInfo, pPending->psock);
    if (ret != 0) {
      // BUGBUG: cleanup?  how do we close the socket from here?
    }

    // remove from pending list
    daPending[k] = NULL;
    delete [] pPending->sbStreamName;
    delete pPending;
  }

  // We want to start source filters and let them work on data while
  // the later filters are starting, without a barrier to impede
  // progress.  Doing this perfectly would require the console process
  // do a topological sort on the filter graph, and send out
  // STARTINSTANCE packets in this order.

  // lazily open stream connections, now that the filter threads have started
  StartInstance_OpenStreamConn(wInstanceNum);

  return DC_ERR_OK;
}

/**
 * given a ptr to shared filter state, create a new copy and start thread
 * storing all state in daFilterCopy()
 */
// locks: assume mutexFilterCopy is held
// thread: remote process reader
int DC_RemoteProcessState::StartInstance_AddNewCopies(DC_RemoteFilterShared
						      *pFilterShared,
						      int wNewCopies,
						      int wInitialWorkNum) {
  DC_RemoteFilterCopy      *pFilter;
  DC_RemoteInStreamShared  *pInStreamShared;
  DC_RemoteInStreamCopy    *pInStream;
  DC_RemoteOutStreamShared *pOutStreamShared;
  DC_RemoteOutStreamCopy   *pOutStream;
  Types::uint j, k;
  Random rand;

#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
  cout << "[reader] StartInstance_AddNewCopies(): creating " << wNewCopies
       << " copies of " << pFilterShared->sbName << " ..." << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */

  for (int iCopy=0; iCopy<wNewCopies; ++iCopy) {
    pFilter = new DC_RemoteFilterCopy(pFilterShared, iCopy, wInitialWorkNum);
#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
    cout << "[reader] StartInstance_AddNewCopies(): creating "
	 << pFilter->pShared->sbName << ".c" << iCopy << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */

    // look for a vacated spot first, if none append
    bool fFound = false;
    for (k=0; k<daFilterCopy.NumElem(); ++k) {
      if (! daFilterCopy[k]) { fFound = true; break; }
    }
    if (fFound) {
      daFilterCopy.Set(k, pFilter);
      pFilter->iFilter = k;
    } else {
      pFilter->iFilter = daFilterCopy.Append(pFilter);
    }

    // input streams
    for (j=0; j<pFilterShared->daInStreamShared.NumElem(); ++j) {
      pInStreamShared = pFilterShared->daInStreamShared[j];

      pInStream = new DC_RemoteInStreamCopy(pInStreamShared, pFilter);
      pFilter->daInStreamCopy.Append(pInStream);

      // we are a consumer
      pInStreamShared->SinkQ.localAddConsumer();
    }

    // output streams
    for (j=0; j<pFilterShared->daOutStreamShared.NumElem(); ++j) {
      pOutStreamShared = pFilterShared->daOutStreamShared[j];

      pOutStream = new DC_RemoteOutStreamCopy(pOutStreamShared, pFilter);
      // randomly assign intial copyset for output to avoid hotspots
      pOutStream->iConnWrite =
	rand.GetRandRange(0, pOutStreamShared->daConnInfo.NumElem()-1);
      pFilter->daOutStreamCopy.Append(pOutStream);

      pOutStream->wWorkNum = wInitialWorkNum;

#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
      cout << "[reader] StartInstance_AddNewCopies(): outs copy for "
	   << pOutStreamShared->sbName << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */

      // all copy outputs
      // 1 colocated/many remote?
      for (k=0; k<pOutStreamShared->daConnInfo.NumElem(); ++k) {
	DC_RemoteOutStreamCopy::ConnInfo_t *pConnInfo =
	  new DC_RemoteOutStreamCopy::ConnInfo_t;
	if (pOutStreamShared->daConnInfo[k]->fColocated) {
	  // we are a producer
	  pOutStreamShared->pInStreamShared->SinkQ.localAddProducer();
	}
	pOutStream->daConnInfo.Append(pConnInfo);
	// conn is opened by caller after they are all created
      }

      // setup for sliding window counters, one per output stream
      // TODO: get threshold value from console config file, and it
      //       should be specified in StartInstance spec.  For now
      //       it is hard-coded to 3.
      pOutStream->lcUnAckedBuffers
	.init(pOutStreamShared->daConnInfo.NumElem(), 0, 3);
    }

    // no cross references to fill in, because we have the shared xref

    if (!pFilterShared->fConsole) {
      //
      // start filter threads
      //
      thread_filter_arg_t *pargFilter;

      // start the filter thread
      pargFilter = new thread_filter_arg_t;  // freed at end of thread fcn
      pargFilter->pDC = pDC;
      pargFilter->pRemote = pDC->pRemote;
      pargFilter->pargc = pDC->pargc;
      pargFilter->pargv = pDC->pargv;
      pFilter->thread_detach();
      int ret = pFilter->thread_create(NULL, static_cast(void*,pargFilter));
      if (ret != 0) {
	cout << "DC_RemoteProcessState::StartInstance(): "
	     << "failed i" << pFilter->pShared->wInstanceNum
	     << ", filter " << pFilter->pShared->sbName
	     << " thread creation"
	     << " (" << ret << " - " << strerror(ret) << ")"
	     << endl;
	return DC_ERR_RemoteFailThreadCreate;
      } else {
#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
	cout << "[reader] Created thread " << (int)pFilter->tid
	     << " for i" << pFilter->pShared->wInstanceNum
	     << ", filter " << pFilter->pShared->sbName << ".c" << iCopy
	     << " ..." << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */
      }

      // We don't need filter barriers any longer ... everything is
      // asynchronous and there are not memory visibility issues, since
      // all filter abstractions are setup before the filter thread is
      // started.  semOpenAck will be the only communication once the
      // the streams are connected.
    } else {
      // console pseudo-filter
      pFilter->_init();
    }
  }

  // special-case signaling of console process pseudo filter
  // (should only be 1 copy)
  if (pFilterShared->fConsole) {
    // here we need to signal the console main thread to allow it to
    // continue
    pDC->pConsole->mutexRemoteFilter.lock();
#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
    cout << "[reader] created <console> filter" << endl;
#endif
    assert(pDC->pConsole->pRemoteFilter == NULL);
    pDC->pConsole->pRemoteFilter = pFilter;
#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
    cout << "[reader] StartInstance(): signaling console that i"
	 << pFilter->pShared->wInstanceNum << " processed ..." << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */
    pDC->pConsole->cvRemoteFilter.signal();
    pDC->pConsole->mutexRemoteFilter.unlock();      
  }

  return DC_ERR_OK;
}

/**
 * Look through daFilterCopy(), and open any needed stream connections for
 * the given instance.
 */
// thread: remote process reader
int DC_RemoteProcessState::StartInstance_OpenStreamConn(int wInstanceNum) {
  DC_RemoteFilterCopy      *pFilter;
  DC_RemoteOutStreamCopy   *pOutStream;
  DC_RemoteOutStreamShared *pOutStreamShared;
  DC_RemoteOutStreamCopy::ConnInfo_t   *pConnInfo;
  DC_RemoteOutStreamShared::ConnInfo_t *pConnInfoShared;
  Types::uint i, j, k;
  
  // BUGBUG: The reader thread will block during the connect (TCP
  // three-way handshake).  If this is a problem, we can use a
  // dedicated static or temp detached thread for opening connections.

  // Filter threads will block on their first write operation to a
  // stream that has not been ACKed, until the reader recvs the ACK
  // and enables the stream.  If the other end is not alive yet
  // (meaning the STARTINSTANCE has not arrived), the filter write
  // operation will block on this side.

  //
  //  connect outgoing streams
  //
  MemoryBuf mb(KB_1);
  Packet packetOut;

  for (i=0; i<daFilterCopy.NumElem(); ++i) {
    pFilter = daFilterCopy[i];
    if (!pFilter || pFilter->pShared->wInstanceNum != wInstanceNum) continue;

    for (j=0; j<pFilter->daOutStreamCopy.NumElem(); ++j) {
      pOutStream = pFilter->daOutStreamCopy[j];
      pOutStreamShared = pFilter->pShared->daOutStreamShared[j];

      // consider array of copy sets
      for (k=0; k<pOutStreamShared->daConnInfo.NumElem(); ++k) {
	pConnInfoShared = pOutStreamShared->daConnInfo[k];
	pConnInfo = pOutStream->daConnInfo[k];
	if (pConnInfoShared->fColocated) continue;

	// open socket
	if (pConnInfo->sock.Connect(pConnInfoShared->addr) != 0) {
	  cout << "StartInstance_OpenStreamConn(): "
	       << "failed Connect() to remote process "
	       << pConnInfoShared->addr << endl
	       << " - " << strerror(errno) << endl;
	  // BUGBUG: cleanup?
	  return DC_ERR_Error;
	}
#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
	cout << "[reader] StartInstance_OpenStreamConn(): "
	     << "Connected to remote process "
	     << pConnInfoShared->addr << endl;
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */

	// add socket info to reader server loop
	// (this is ok, because the reader thread calls us)
	pConnInfo->iClient =
	  pDC->pReader->AddExternalConn(&pConnInfo->sock,
					&pConnInfoShared->addr);
	ReaderConnMap *pMap =
	  (ReaderConnMap*)pDC->pReader->getClientMapValue(pConnInfo->iClient);
	pMap->pOutStream = pOutStream;
	pMap->wInstanceNum = wInstanceNum;
#if defined(DCDBG_READERCONNNAME)
	{
	  char bufTmp[80];
	  snprintf(bufTmp, 80, "<conn %d - source i%d,%s>",
		   pConnInfo->iClient, wInstanceNum, pOutStreamShared->sbName);
	  if (pMap->sbName) delete [] pMap->sbName;
	  pMap->sbName = StrDup(bufTmp);
	}
#endif /* DCDBG_READERCONNNAME */

	// send STREAMCONNECT packets
	mb.setSize(0);
	mb.sprintf("instance=%d\n", pFilter->pShared->wInstanceNum);
	mb.sprintf("stream=%s\n", pOutStreamShared->sbName);
	mb.sprintf("iconninfo=%d\n", k);
	mb.sprintf("streamwritepolicy=%s\n",
		   DC_StreamWritePolicy_tostr(pOutStreamShared->
					      eStreamWritePolicy));
	//mb.sprintf("copysetid=%d\n", ???);
	packetOut.set(TAG_FILTERSERVICE_STREAMCONNECT, &mb);
	packetOut.setExtra(pFilter->pShared->wInstanceNum);
	// send packet to peer remote process
	if (packetOut.write(pConnInfo->sock) <= 0) {
	  cerr << "write_nocopy(): Failed write of STREAM_CONNECT packet."
	       << endl;
	  // BUGBUG: cleanup?
	  return DC_ERR_FailedConnect;
	}
#if defined(DCDBG_FILTERSERVICE_STARTINSTANCE)
	cout << "[reader] StartInstance_OpenStreamConn():"
	     << " Sent STREAM_CONNECT packet ..." << endl
	     << mb.getPtr();
#endif /* DCDBG_FILTERSERVICE_STARTINSTANCE */
      }
    }
  }
  return DC_ERR_OK;
}

/**
 * handle an incoming stream connect request (realtime or pending)
 */
// thread: remote process reader
int DC_RemoteProcessState::
StreamConnect(int wInstanceNum, char *sbStreamName, int iConnInfo,
	      DC_StreamWritePolicy_t eStreamWritePolicy,
	      ReaderConnMap *pMap, BaseSocket *psock, bool fDoPending) {
  // stream names are unique per instance, so we don't need anything
  // else like the filter name.

  // find this filter
  DC_RemoteFilterCopy   *pFilter;
  DC_RemoteInStreamCopy *pInStream = NULL;
  Types::uint i, j;
  bool fFound = false;

  mutexFilterCopy.lock();  // BUGBUG: not sure if we need this

  for (i=0; i<daFilterCopy.NumElem(); ++i) {
    pFilter = daFilterCopy[i];
    if (!pFilter || pFilter->pShared->wInstanceNum != wInstanceNum) continue;

    for (j=0; j<pFilter->daInStreamCopy.NumElem(); ++j) {
      pInStream = pFilter->daInStreamCopy[j];
      if (strcmp(sbStreamName, pInStream->pShared->sbName) == 0) {
	fFound = true;
	break;
      }
    }
    if (fFound) break;  // prune
  }

  mutexFilterCopy.unlock();

  // ASSUME: if we don't send back an ACK immediately, the
  // other side will block, and we will receive nothing else
  // on this socket in the interim.
  if (fFound) {
    pMap->pInStream = pInStream;
    pMap->wInstanceNum = wInstanceNum;
    pMap->eStreamWritePolicy = eStreamWritePolicy;
#if 0  /* we compensated for remote copies when the SinkQ was created */
    // add this connection
    pMap->pInStream->pShared->SinkQ.remoteAddProducer();
#endif
    return 0;
  }

  if (fDoPending) {  // we only do this if called when STREAMCONNECT recv
    // not found - add to set of pending stream connects
    DC_RemotePending *pPending = new DC_RemotePending;
    pPending->wInstanceNum = wInstanceNum;
    pPending->sbStreamName = StrDup(sbStreamName);
    pPending->iConnInfo = iConnInfo;
    pPending->eStreamWritePolicy = eStreamWritePolicy;
    pPending->pMap = pMap;
    pPending->psock = psock;

    // find free spot, or append
    bool fEmptySpot = false;
    Types::uint k;
    for (k=0; k<daPending.NumElem(); ++k) {
      if (! daPending[k]) { fEmptySpot = true; break; }
    }
    if (fEmptySpot) {
      daPending.Set(k, pPending);
    } else {
      k = daPending.Append(pPending);
    }
#if defined(DCDBG_FILTERSERVICE_STREAMCONNECT)
    cout << "StreamConnect(): adding to pending list (k=" << k << ")" << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMCONNECT */
  }

  return -1;
}

/**
 * send a stream connect ack
 */
// thread: remote process reader
int DC_RemoteProcessState::StreamConnectAck(int wInstanceNum, int iConnInfo,
					    BaseSocket *psock) {
  Packet packetOut;
  // send ack
  packetOut.set(TAG_FILTERSERVICE_STREAMCONNECT_ACK, NULL);
  packetOut.setExtra(iConnInfo);
  if (packetOut.write(*psock) <= 0) {
    cerr << "Send of STREAMCONNECT packet failed." << endl;
    return -1;  // some error writing the client socket
  }
#if defined(DCDBG_FILTERSERVICE_STREAMCONNECT)
  cout << "Sent STREAMCONNECT_ACK packet" << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMCONNECT */
  return 0;
}

int DC_RemoteProcessState::CancelWork(int wInstanceNum, int wWorkNum) {
  DC_RemoteFilterCopy *pFilter;
  Types::uint i;

  mutexFilterCopy.lock();
  for (i = 0; i < daFilterCopy.NumElem(); ++i) {
    pFilter = daFilterCopy[i];
    if (!pFilter ||
	pFilter->pShared->wInstanceNum != wInstanceNum ||
	pFilter->wCopyNum != 0) {
      continue;
    }
    pFilter->pShared->cancellations.Append(wWorkNum);
  }
  mutexFilterCopy.unlock();
  return DC_ERR_OK;
}


/**
 * append the buffer into the work queue for the given instance
 */
// thread: remote process reader
int DC_RemoteProcessState::AppendWork(int wInstanceNum, int wWorkNum,
				      DC_Buffer_t *pbuf,
				      int iConnDisableIfFull) {
  // ASSUME: Since the same entity that sends the start instance pkt will
  // also be the only one to send the append work pkt (over the same socket),
  // thus we know the instance will be created already.
  int wRtn = DC_ERR_RemoteInstanceError;

  DC_Work_t *pwork;
  DC_RemoteFilterCopy *pFilter;
  Types::uint i;
  bool fFound = false;

  // BUGBUG: I don't think we need to lock this because the reader is the only
  // thread that can change the filterspec da structs.
  mutexFilterCopy.lock();

#if defined(DCDBG_FILTERSERVICE_APPENDWORK)
  cout << "[reader] RemoteProcessState::AppendWork() i" << wInstanceNum
       << ", work=" << wWorkNum << " ..." << endl;
#endif /* DCDBG_FILTERSERVICE_APPENDWORK */

  for (i=0; i<daFilterCopy.NumElem(); ++i) {
    pFilter = daFilterCopy[i];
    if (!pFilter ||
	pFilter->pShared->wInstanceNum != wInstanceNum ||
	pFilter->wCopyNum != 0)  // append to a copy set once
      continue;
    // BUGBUG: the above hack to check wCopyNum for only appending work to
    // a copy set's shared WorkQ only work is copy 0 is always there.  If not
    // we need to store a wWorkNum count in the pShared object, and check that.
    fFound = true;

#if defined(DCDBG_FILTERSERVICE_APPENDWORK)
    cout << "RemoteProcessState::AppendWork() append work=" << wWorkNum
	 << " to filter " << pFilter->pShared->sbName
	 << ".c" << pFilter->wCopyNum << endl;
#endif /* DCDBG_FILTERSERVICE_APPENDWORK */

    // We make a duplicate of the work buffer for each local filter,
    // because the buffers are not thread safe.
    wRtn = DC_ERR_OK;
    pwork = new DC_Work_t(0);
    pwork->buf.DeepCopy(*pbuf);
    pwork->wWorkNum = wWorkNum;

    // this will modify lcFullWorkQs with the count of queues that are full
    if (pFilter->pShared->WorkQ.remoteEnqueue(pwork, iConnDisableIfFull)) {
      cerr << "DC_RemoteProcessState::AppendWork(): "
	   << "failed append of new work to filter "
	   << pFilter->pShared->sbName << endl;
      // BUGBUG: cleanup?
      wRtn = DC_ERR_RemoteInstanceError;
      break;
    }
  }

  if (!fFound) {
    cout << "DC_RemoteProcessState::AppendWork(): "
	 << "failed append of new work to any filter for i"
	 << wInstanceNum << endl;
  }

  mutexFilterCopy.unlock();

  // check if conn should be disabled
  lcFullWorkQs.lockInternal();
  if (lcFullWorkQs.get_u(0) > 0 && lcFullWorkQs.get_u(2) == 0) {
    lcFullWorkQs.set_u(1, iConnDisableIfFull);
    lcFullWorkQs.set_u(2, 1);
#if defined(DCDBG_FILTERSERVICE_APPENDWORK)
    cout << "RemoteProcessState::AppendWork() disabling conn="
	 << iConnDisableIfFull << endl;
#endif /* DCDBG_FILTERSERVICE_APPENDWORK */
    pDC->pReader->DisableConn(iConnDisableIfFull);
  }
  lcFullWorkQs.unlockInternal();

  return wRtn;
}

/**
 * as directed by console process, we append a special STOP work buffer to
 * signal the filter-thread to stop
 */
// thread: remote process reader
int DC_RemoteProcessState::StopInstance(int wInstanceNum) {
  // ASSUME: Since the same entity that sends the start instance pkt will
  // also be the only one to send the append work pkt (over the same socket),
  // thus we know the instance will be created already.
  int wRtn = DC_ERR_RemoteInstanceError;

  DC_Work_t *pwork;
  DC_RemoteFilterCopy *pFilter;
  Types::uint i;

  // BUGBUG: I don't think we need to lock this because the reader is the only
  // thread that can change the filterspec da structs.
  mutexFilterCopy.lock();

  for (i=0; i<daFilterCopy.NumElem(); ++i) {
    pFilter = daFilterCopy[i];
    if (!pFilter || pFilter->pShared->wInstanceNum != wInstanceNum) continue;

    wRtn = DC_ERR_OK;
    pwork = NULL;  // NULL work means STOP
    if (pFilter->pShared->WorkQ.remoteEnqueue(pwork)) {
      cerr << "DC_RemoteProcessState::StopInstance(): "
	   << "failed append of new work to filter "
	   << pFilter->pShared->sbName << endl;
      // BUGBUG: cleanup?
      wRtn = DC_ERR_RemoteInstanceError;
    }
  }

  mutexFilterCopy.unlock();
  return wRtn;
}

/**
 * remove the state for this instance in the remote process.  The last running
 * filter of an instance will send a ReaderCmd to the reader, which in turn
 * the reader will call this fcn for cleanup.
 */
// thread: remote process reader
int DC_RemoteProcessState::CleanupInstance(int wInstanceNum) {
  DC_RemoteFilterCopy    *pFilter;
  DC_RemoteInStreamCopy  *pInStream;
  DC_RemoteOutStreamCopy *pOutStream;
  Types::uint i, j;

#if defined(DCDBG_FILTERSERVICE_CLEANUP_INSTANCE)
  cout << "CleanupInstance(): last filter for instance=" << wInstanceNum
       << " exited" << endl;
#endif /* DCDBG_FILTERSERVICE_CLEANUP_INSTANCE */

  mutexFilterCopy.lock();

  // ASSUME: the normal exit of the filter should have called
  // CloseOutputWaitInput(), so all sockets should be closed, and the
  // reader should be consistent with this in terms of its rgClient[]
  // array.

  // TODO: make sure this assumption is true when the filter exits on
  // any path.  There is a race betweeen STREAMCONNECT that are sent
  // out by the reader thread, and a filter exiting quickly.  We
  // should block until all streams are created, or fail trying, and
  // then exit gracefully.  Can this deadlock waiting for
  // STREAMCONNECT to finish on a node where the filter instance fails
  // or all quickly exit?  Need to check and make sure it is
  // reasonable.  Maybe record wInstanceNum in the reader, and if it
  // sees a STREAMCONNECT for an instance that is <= the current count,
  // and it sees nothing running for it, return failure
  // STREAMCONNECT_ACK.

  DC_RemoteInstanceShared *pRemoteInstanceShared = NULL;

  // TODO: need to keep track of the number of copies, and cleanup
  // shared state (DC_RemoteFilterShared, DC_Remote{In/Out}StreamShared)
  // similar to how the instance level state is kept.

  for (i=0; i<daFilterCopy.NumElem(); ++i) {
    pFilter = daFilterCopy[i];
    if (!pFilter || pFilter->pShared->wInstanceNum != wInstanceNum) continue;

    for (j=0; j<pFilter->daInStreamCopy.NumElem(); ++j) {
      pInStream = pFilter->daInStreamCopy[j];
      // reader already closed connections
      // TODO: close down queue? -- Queue already closed down when filter exits
      delete pInStream;
    }
    for (j=0; j<pFilter->daOutStreamCopy.NumElem(); ++j) {
      pOutStream = pFilter->daOutStreamCopy[j];
      // reader already closed connections
      delete pOutStream;
    }

    // drain and shutdown work queue
    pFilter->pShared->WorkQ.localShutdown();

    // remove filter struct
    daFilterCopy[i] = NULL;
    pRemoteInstanceShared = pFilter->pShared->pRemoteInstanceShared;  // save
    delete pFilter;
  }

  // cleanup of common instance state
  if (pRemoteInstanceShared) delete pRemoteInstanceShared;

  mutexFilterCopy.unlock();
  return 0;
}

/**
 * Given a buffer, automatically decide which set of copies to send it
 * to.  For colocated copy sets, we deposit a buffer directly into
 * their SinkQ, and for remote copy sets we send over the socket.  The
 * buffer is duplicated and/or consumed as determined by
 * fBufferHandoff, colocated vs. remote destination, and the write
 * success.
 *
 * pShared->eStreamWritePolicy controls writing to multiple copy sets
 *   RR  = round robin per copy set (default)
 *   WRR = weight based on # copies per copy set
 *   DD  = demand driven push flows based on acks
 *   UD  = user defined (callback fcn)
 *
 * fBufferHandoff - f=called by write(), t=called by write_nocopy()
 *
 * rtn: DC_ERR_OK = write ok
 *      DC_ERR_FailedWrite = write failed
 *
 */
// thread: remote filter
int DC_RemoteOutStreamCopy::WriteBuffer_DATA(DC_Buffer_t *pbuf,
					     DC_RemoteFilterCopy *pFilter,
					     bool fBufferHandoff,
					     void *pUserArg) {
  // stats
  if (!pFilter->fInstantWriteFirst) {
    pFilter->fInstantWriteFirst = true;
    pFilter->rInstantWriteFirst = pFilter->tProcess.Snapshot();
  }

  DC_RemoteOutStreamCopy::ConnInfo_t   *pConnInfo;
  DC_RemoteOutStreamShared::ConnInfo_t *pConnInfoShared;

  int  wBufSize = pbuf->getSize();  // save for accumulating after send

  int  iConnSave = iConnWrite;
  bool fDone = false, fDoConsume = false;

  // RR  == round robin per copy set (default)
  // WRR == weight based on # copies per copy set
  // DD  == demand driven based on DATA_ACK receipt
  // UD  == user defined (callback based)

  int minCount;  // used for _DD policy

  assert(iConnWrite >= 0);
  assert(iConnWrite < static_cast(int,pShared->daConnInfo.NumElem()));

  do {  // we are going to write to ConnInfo[iConnWrite] ...

    // demand driven: find the connection that has the least unacked
    // buffers (fastest consumer), and set iConnWrite for this one
    // 07/17/2001 : mdb - Ties are broken by choosing the colocated
    //                    filter, otherwise the first in copyset order
    //                    starting from iConnWrite is used.
    // 07/27/2001 : mdb - Fixed lt/gt bug in DD policy, causing skewing.
    if (pShared->eStreamWritePolicy == DC_StreamWritePolicy_DD) {
      int iConnTmp;
      lcUnAckedBuffers.lockInternal();
#if 0
      cout << "DEBUG:DD on entry iConnWrite=" << iConnWrite << endl;
      for (unsigned int i=0; i<pShared->daConnInfo.NumElem(); ++i) { cout << "DEBUG:DD [" << i << "]=" << lcUnAckedBuffers.get_u(i) << endl; }  // DEBUG
#endif
      iConnTmp = iConnWrite;
      int iConnMin = -1;
      while (1) {
	do {
	  if (! daConnInfo[iConnTmp]->fEOS &&
	      lcUnAckedBuffers.diffHigh_u(iConnTmp) >= 0) {
	    // an open conn, with room for more unacked writes
	    if (iConnMin == -1) {
	      iConnMin = iConnTmp;
	    } else {
	      int wDiff = lcUnAckedBuffers.diffHigh_u(iConnTmp) -
		lcUnAckedBuffers.diffHigh_u(iConnMin);
	      // a better min, or a tie and the new one is colocated 
	      if (wDiff > 0 ||  // new min
		  (wDiff == 0 && pShared->daConnInfo[iConnTmp]->fColocated))
		iConnMin = iConnTmp;
	    }
	  }
	  ++iConnTmp;
	  if ((unsigned int)iConnTmp >= pShared->daConnInfo.NumElem())
	    iConnTmp = 0;
	} while (iConnTmp != iConnSave);  // make one full pass

	if (iConnMin != -1) {
	  iConnWrite = iConnMin;
	  minCount = lcUnAckedBuffers.get_u(iConnMin);
	  break;
	}

#if 0
	cout << "DEBUG:DD sleeping until some ACKs arrive ..." << endl;
#endif
	lcUnAckedBuffers.waitHigh();  // sleep until recv enough acks
      }
      // we assume this one will be written to, so bump up that
      // counter now so we don't need another lock/unlock or to hold
      // the lock during the write (cleanup if write fails?)
      lcUnAckedBuffers.inc_u(iConnWrite);
      lcUnAckedBuffers.unlockInternal();
    } else if (pShared->eStreamWritePolicy == DC_StreamWritePolicy_UD) {
      // user defined: we write to the connection the application
      // chooses.  if it fails, we don't try other connections.
      iConnWrite = pShared->pWritePolicyUD
	->write_copyset(pShared->daConnInfo.NumElem(),
			pShared->sbName, pbuf, pUserArg);
      iConnSave = iConnWrite;  // force exit after this iteration
    }

    // attempt send to iConnWrite
    pConnInfo = daConnInfo[iConnWrite];
    pConnInfoShared = pShared->daConnInfo[iConnWrite];
#if defined(DCDBG_FILTERSERVICE_WRITEBUFFER)
    cout << lock
	 << "WriteBuffer DATA (" << pbuf->getSize() << ",work="
	 << wWorkNum << ") from "
	 << pFilter->pShared->sbName << " to " << pShared->sbFilterSink
	 << " (copyset=" << iConnWrite << ",stream=" << pShared->sbName
	 << ",fd=" << pConnInfo->sock.getFd() << ")";
    if (pShared->eStreamWritePolicy == DC_StreamWritePolicy_DD) {
      cout << " DD:minCount=" << minCount;
    }
    cout << endl << unlock;
#if defined(DCDBG_FILTERSERVICE_DATA_DUMPBUF)
    pbuf->Dump();
#endif /* DCDBG_FILTERSERVICE_DATA_DUMPBUF */
#endif /* DCDBG_FILTERSERVICE_WRITEBUFFER */

    if (! pConnInfo->fEOS) {
      if (pConnInfoShared->fColocated) {  // put directly into sink's queue
	if (fBufferHandoff) {
	  fDone = WriteBuffer_Colocated(pbuf, pFilter, iConnWrite);
	  // NOTE: fDone==t -> cannot use pbuf any longer
	} else {
	  DC_Buffer_t *pbufTmp = new DC_Buffer_t(*pbuf);
	  fDone = WriteBuffer_Colocated(pbufTmp, pFilter, iConnWrite);
	  if (!fDone) pbufTmp->consume();
	}
      } else {  // send over socket
	// dont need a copy here for either case
	fDone = WriteBuffer_Remote(pbuf, pFilter, iConnWrite);
	if (fDone && fBufferHandoff) fDoConsume = true;
      }
    }

    // advance "first considered" connection for next write operation
    if (pShared->eStreamWritePolicy == DC_StreamWritePolicy_WRR) {
      // only advance if we have written one to *each* copy in the set
      if (++nCopiesWritten >= pConnInfoShared->nCopies) {
	nCopiesWritten = 0;
	++iConnWrite;
	if ((unsigned int)iConnWrite >= pShared->daConnInfo.NumElem())
	  iConnWrite = 0;
      }
    } else {  // _RR || _DD
      // always advance
      ++iConnWrite;
      if ((unsigned int)iConnWrite >= pShared->daConnInfo.NumElem())
	iConnWrite = 0;
    }
  } while (!fDone && iConnWrite != iConnSave);  // make one full pass

  if (!fDone) {
    // failed write to all connections, fail
#if 1 || defined(DCDBG_FILTERSERVICE_CLOSE)
    cout << "WriteBuffer_DATA(): discard DATA - input stream must be closed"
	 << endl;
#endif /* DCDBG_FILTERSERVICE_CLOSE */
    //pbuf->consume();
    return DC_ERR_FailedWrite;
  }
  // fDone -> write to somewhere succeeded
  mWriteBytes.NewValue(wBufSize);  // stats
  if (fDoConsume) pbuf->consume();
  return DC_ERR_OK;
}

/**
 * Given a control buffer, send it to all sets of copies, and copy as
 * needed for colocated copies.  Consume the given buffer at the end
 * regardless of success or failure.
 *
 * rtn: DC_ERR_OK = all writes ok
 *      DC_ERR_FailedWrite = at least one failed
 */
int DC_RemoteOutStreamCopy::WriteBuffer_EOWEOS(DC_Buffer_t *pbuf,
					       DC_RemoteFilterCopy *pFilter) {
  DC_RemoteOutStreamCopy::ConnInfo_t *pConnInfo;
  DC_RemoteOutStreamShared::ConnInfo_t *pConnInfoShared;

  bool fSetEOS = (pbuf->eType == DC_Buffer_t::EOS);
  bool fRtn = true;
  bool fOk;

  // replicate EOW/EOS buffers to all the consumer copy sets
  for (Types::uint i=0; i<pShared->daConnInfo.NumElem(); ++i) {
    pConnInfo = daConnInfo[i];
    pConnInfoShared = pShared->daConnInfo[i];
    if (pConnInfo->fEOS) continue;

#if defined(DCDBG_FILTERSERVICE_WRITEBUFFER)
    cout << lock
	 << "WriteBuffer EOWEOS (" << pbuf->getSize() << ",work="
	 << wWorkNum << ") from "
	 << pFilter->pShared->sbName << " to " << pShared->sbFilterSink
	 << " (copyset=" << i << ",stream=" << pShared->sbName
	 << ",fd=" << pConnInfo->sock.getFd() << ")" << endl
	 << unlock;
#endif /* DCDBG_FILTERSERVICE_WRITEBUFFER */

    if (pConnInfoShared->fColocated) {  // put directly into sink's queue
      DC_Buffer_t *pbufTmp = new DC_Buffer_t(*pbuf);
      fOk = WriteBuffer_Colocated(pbufTmp, pFilter, i);
      if (!fOk) pbufTmp->consume();
    } else {  // send over socket
      fOk = WriteBuffer_Remote(pbuf, pFilter, i);
    }
    if (!fOk) fRtn = false;  // this one failed?
    if (fSetEOS) pConnInfo->fEOS = true;
  }

  if (pbuf->eType == DC_Buffer_t::EOW) ++wWorkNum;
  pbuf->consume();

  return (fOk ? DC_ERR_OK : DC_ERR_FailedWrite);
}

/**
 * write the given buffer to a colocated filter, no consume takes place
 * rtn: true=success, pbuf handoff - don't use buffer any longer
 *      false=failure, pbuf not handed off
 */
bool DC_RemoteOutStreamCopy::WriteBuffer_Colocated(DC_Buffer_t *pbuf,
						  DC_RemoteFilterCopy *pFilter,
						   int iConnWrite) {
  DC_RemoteOutStreamCopy::ConnInfo_t *pConnInfo = daConnInfo[iConnWrite];
#if defined(DCDBG_FILTERSERVICE_DATA)
  char *sbType;
  if (pbuf->eType == DC_Buffer_t::EOW) sbType = "EOW";
  else if (pbuf->eType == DC_Buffer_t::EOS) sbType = "EOS";
  else sbType = "DATA";
#if 0
  if (pbuf->eType == DC_Buffer_t::EOW || pbuf->eType == DC_Buffer_t::EOS)
#endif
  cout << lock
       << "CoSent " << sbType << " (" << pbuf->getSize() << ",work="
       << wWorkNum << ")  from "
       << pFilter->pShared->sbName << " to " << pShared->sbFilterSink
       << " over stream " << pShared->sbName << " (queue "
       << hex << (long)(&pShared->pInStreamShared->SinkQ) << dec
       << ")" << endl
       << unlock;
#if defined(DCDBG_FILTERSERVICE_DATA_DUMPBUF)
  pbuf->Dump();
#endif /* DCDBG_FILTERSERVICE_DATA_DUMPBUF */
#endif
  pbuf->resetExtract(); // in case previous filter extracted something
  if (pShared->pInStreamShared->
      SinkQ.localEnqueue(wWorkNum, pbuf, pShared->eStreamWritePolicy,
			 iConnWrite, &lcUnAckedBuffers)) {
    // queue shutdown == forced EOS from colocated sink
    pConnInfo->fEOS = true;
    return false;
  }
  // enqueue succeeded
  return true;
}

/**
 * write the given buffer to a remote process, no consume takes place
 */
bool DC_RemoteOutStreamCopy::WriteBuffer_Remote(DC_Buffer_t *pbuf,
						DC_RemoteFilterCopy *pFilter,
						int iConnWrite) {
  DC_RemoteOutStreamCopy::ConnInfo_t *pConnInfo = daConnInfo[iConnWrite];
  // is the async STREAMOPEN still happening?
  if (! pConnInfo->fOpened) {
#if defined(DCDBG_FILTERSERVICE_OPEN)
    bool fNotOpen = pConnInfo->semOpenAck.wait_testblock();
    if (fNotOpen) {
      cout << "WriteBuffer(): STREAMOPEN not finished from "
	   << pFilter->pShared->sbName << " to " << pShared->sbFilterSink
	   << " over stream " << pShared->sbName << " ..." << endl;
    }
#endif /* DCDBG_FILTERSERVICE_OPEN */

    // wait until reader has received open ACK
    // BUGBUG: how to unblock this for a forced shutdown?
    pConnInfo->semOpenAck.wait();
    pConnInfo->fOpened = true;

#if defined(DCDBG_FILTERSERVICE_OPEN)
    if (fNotOpen) {
      cout << "WriteBuffer(): STREAMOPEN finished from "
	   << pFilter->pShared->sbName << " to " << pShared->sbFilterSink
	   << " over stream " << pShared->sbName << "." << endl;
    }
#endif /* DCDBG_FILTERSERVICE_OPEN */
  }

  // send the buffer to the remote sink
  Packet pkt(TAG_FILTERSERVICE_DATA, pbuf);
  // TODO: create buffer "tags" for sending?  It may be useful for
  // the application to use, perhaps in identifying the fixed size
  // pieces of a variable sized "chunk".
  //pkt.setExtra(XXX);
  if (pbuf->eType == DC_Buffer_t::EOW)
    pkt.setTag(TAG_FILTERSERVICE_EOW);
  else if (pbuf->eType == DC_Buffer_t::EOS)
    pkt.setTag(TAG_FILTERSERVICE_EOS);

  if (pkt.write(pConnInfo->sock) <= 0) {
    cout << "WriteBuffer(): Failed write of data from "
	 << pFilter->pShared->sbName << " to "
	 << pShared->sbFilterSink << " over stream "
	 << pShared->sbName << " (fd="
	 << pConnInfo->sock.getFd() << ")"
	 << " - " << strerror(errno) << endl << pkt << endl;
    // assume socket is dead
    pConnInfo->fEOS = true;
    // TODO: close the socket as before?
    return false;
  }
#if defined(DCDBG_FILTERSERVICE_DATA)
  char *sbType;
  if (pbuf->eType == DC_Buffer_t::EOW) sbType = "EOW";
  else if (pbuf->eType == DC_Buffer_t::EOS) sbType = "EOS";
  else sbType = "DATA";
#if 0
  if (pbuf->eType == DC_Buffer_t::EOW || pbuf->eType == DC_Buffer_t::EOS)
#endif
  cout << lock
       << "Sent " << sbType << " (" << pbuf->getSize() << ",work="
       << wWorkNum << ") from "
       << pFilter->pShared->sbName << " to " << pShared->sbFilterSink
       << " over stream " << pShared->sbName
       << " (fd=" << pConnInfo->sock.getFd() << ")" << endl
       << unlock;
#if defined(DCDBG_FILTERSERVICE_DATA_DUMPBUF)
  pbuf->Dump();
#endif /* DCDBG_FILTERSERVICE_DATA_DUMPBUF */
#endif
  return true;
}


// ****************************************************************************
