// -*- c++ -*-
// ****************************************************************************
//  Name: remoteprocessstate.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 08/12/2000 : mdb - Created.
//        11/29/2000 : mdb - Converted to use DC_Mutex/DC_Cond.
//        02/22/2001 : mdb - Started StreamBufferQueue to handle copies.
//        03/06/2001 : mdb - Restructured to start filter copies.
//  Desc: State and supporting routines for the remote lib process.
// ****************************************************************************

#if !defined(H_DC_REMOTEPROCESSSTATE)
#define H_DC_REMOTEPROCESSSTATE

#ifdef _WIN32
#include <process.h>
#endif

#include <pthread.h>

#include "lib/dc_configinfo.h"
#include "lib/dc_threadclass.h"
#include "lib/dc_guard.h"
#include "lib/dc_semaphore.h"
#include "lib/dc_queue.h"
#include "hostinfo.h"

#include "dc_filterservice.h"
#include "thread-reader.h"
#include "streambufferqueue.h"
#include "workqueue.h"
#include "dc_policy.h"


/// state common to all filters for a given instance
class DC_RemoteInstanceShared {
private:
  DC_Mutex mutex;
  int wInstanceNum;
  int nFilters;
  int nCopies;

public:
  DC_RemoteInstanceShared(int wInstanceNum_in, int nFilters_in, int nCopies_in)
    : wInstanceNum(wInstanceNum_in), nFilters(nFilters_in),
      nCopies(nCopies_in) {
  };

#if 0
  /// called on exit of the last filter for an instance
  int FilterExit(void) {
    DC_Guard guard(mutex);
    return --nFilters;
  };
#endif
  /// called on exit of the last copy of a filter for an instance
  int CopyExit(void) {
    DC_Guard guard(mutex);
    return --nCopies;
  };
};


/// shared stream state among all filter copies
class DC_RemoteInStreamShared {
public:
  char       *sbName;
  int         wInstanceNum;
  Types::uint iStream;  ///< this stream's index in enclosing da

  char       *sbFilterSource; ///< producer filter name

  /// shared stream Q - can have multiple local and remote copies as the source for buffers being read from this stream
  DC_StreamBufferQueue SinkQ;  

public:
  DC_RemoteInStreamShared(char *sbName_in, int wInstanceNum_in)
    : sbName(StrDup(sbName_in)), wInstanceNum(wInstanceNum_in),
      sbFilterSource(NULL) {
  };
  ~DC_RemoteInStreamShared(void) {
    TestDelete(sbName);
    TestDeleteNull(sbFilterSource);
  };
};

/// per-copy stream state
class DC_RemoteInStreamCopy {
public:
  DC_RemoteInStreamShared *pShared;
  DC_RemoteFilterCopy     *pFilter;

  int   wBufSize;  // TODO: negotiated buffer size
  bool  fEOS;  ///< used by reader thread when EOS packet is read/enqueued

  // runtime stats
  Metric mReadBytes;  ///< num bytes read from this stream
  Timer  tRead;

public:
  DC_RemoteInStreamCopy(DC_RemoteInStreamShared *pInStreamShared,
			DC_RemoteFilterCopy     *pFilterCopy)
    : pShared(pInStreamShared), pFilter(pFilterCopy), fEOS(false) {
  };
  ~DC_RemoteInStreamCopy(void) {
  };
};


/// shared stream state among all filter copies
class DC_RemoteOutStreamShared {
public:
  char       *sbName;
  int         wInstanceNum;
  Types::uint iStream;  ///< this stream's index in enclosing da

  char       *sbFilterSink; ///< the filter name at the other end of this stream

  /// output streams can only use a single output policy for all sink copysets
  int         wCopySetId;
  DC_StreamWritePolicy_t eStreamWritePolicy;  ///< copied from pRemote obj
  DC_Policy_Base_t      *pWritePolicyUD;

  // with copies, we can have up to 1 colocated, and any number of remote

  /// colocated
  DC_RemoteInStreamShared *pInStreamShared;
  /// 1 colocated/ many remote
  struct ConnInfo_t {
    bool        fColocated;
    int         nCopies;
    InetAddress addr;
  };
  DArray<ConnInfo_t*> daConnInfo;

public:
  DC_RemoteOutStreamShared(char *sbName_in, int wInstanceNum_in)
    : sbName(StrDup(sbName_in)), wInstanceNum(wInstanceNum_in),
      sbFilterSink(NULL), pWritePolicyUD(NULL), pInStreamShared(NULL) {
  };
  ~DC_RemoteOutStreamShared(void) {
    TestDelete(sbName);
    TestDeleteNull(sbFilterSink);
    STestDelete(pWritePolicyUD);
  };
};

/// per-copy stream state
class DC_RemoteOutStreamCopy {
public:
  DC_RemoteOutStreamShared *pShared;
  DC_RemoteFilterCopy      *pFilter;

  int   wBufSize;  // TODO: negotiated buffer size
  //bool  fEOS;  //TODO: EOS fixes

  int   wWorkNum; ///< current worknum

  /**
   * for copies, we can have up to 1 set colocated, and any number of
   * remote sets colocated
   * 1 colocated/many remote
   */
  class ConnInfo_t {
  public:
    int iClient;  ///< index in Reader ServerLoop (used for closing)
    ClientSocket sock;
    DC_Semaphore semOpenAck;  ///< signal async receive of STREAMCONNECT_ACK
    bool fOpened; ///< state used by WriteBuffer
    bool fEOS;

    ConnInfo_t(void) : iClient(-1), fOpened(false), fEOS(false) {};
  };
  DArray<ConnInfo_t*> daConnInfo;

  /// demand driven sliding windows
  DC_LockedCounterThreshold lcUnAckedBuffers;

  // runtime stats
  Metric mWriteBytes;  ///< num bytes written to this stream
  Timer  tWrite;       ///< cumulative time spent in write operations

  // state used by WriteBuffer for round-robin writes
  int iConnWrite;      ///< next copyset to write to
  int nCopiesWritten;  ///< number of writes to the current copyset

public:
  DC_RemoteOutStreamCopy(DC_RemoteOutStreamShared *pOutStreamShared,
			 DC_RemoteFilterCopy      *pFilterCopy)
    : pShared(pOutStreamShared), pFilter(pFilterCopy),
      iConnWrite(0), nCopiesWritten(0) {};
  ~DC_RemoteOutStreamCopy(void) {
  };

  int WriteBuffer_DATA(DC_Buffer_t *pbuf, DC_RemoteFilterCopy *pFilter,
		       bool fBufferHandoff, void *pUserArg);
  int WriteBuffer_EOWEOS(DC_Buffer_t *pbuf, DC_RemoteFilterCopy *pFilter);
  bool WriteBuffer_Colocated(DC_Buffer_t *pbuf, DC_RemoteFilterCopy *pFilter,
			     int iConnWrite);
  bool WriteBuffer_Remote(DC_Buffer_t *pbuf, DC_RemoteFilterCopy *pFilter,
			  int iConnWrite);
};


/// shared state among all filter copies
class DC_RemoteFilterShared {
public:
  char *sbName;
  int   wInstanceNum;
  DC_RemoteInstanceShared *pRemoteInstanceShared;
  bool  fConsole;  //< T==this is a pseudo-filter in the console process
  bool  fAllowCopies;

  int   nInitialCopies;
  int   nCopies;  // TODO
  int   nCopiesLow, nCopiesHigh;  // TODO
  int   rankCopySets, maxCopySets;
  int   rankCopiesStart, maxCopies;

  DC_WorkQueue WorkQ;
  DArray<int> cancellations; // CLH - added for new filter return types
  
  DC_Mutex mutexAllIns;
  DC_Cond  cvAllInsNotEmpty;

  /// indexes of these da's match DC_RemoteFilterCopy::daXXStreamCopy;
  DArray<DC_RemoteInStreamShared*>  daInStreamShared;
  DArray<DC_RemoteOutStreamShared*> daOutStreamShared;

public:
  DC_RemoteFilterShared(char *sbName_in, int wInstanceNum_in)
    : sbName(StrDup(sbName_in)), wInstanceNum(wInstanceNum_in),
      nInitialCopies(0), nCopies(0) {
  };
  ~DC_RemoteFilterShared(void) {
    TestDeleteNull(sbName);
  };
};


/// per-copy filter state
class DC_RemoteFilterCopy : public ThreadClass {
public:
  DC_RemoteFilterShared *pShared;
  int         wCopyNum;
  Types::uint iFilter;   ///< our index in RemoteProcessState::daFilterCopy
  char       *sbFullName;  ///< [t#,i#,<filter>.c#]
  char       *sbCopyName;  ///< <filter>.c#

  int         wWorkNum;  ///< holds initial and filter-level current worknum

  /// indexes of these da's match DC_RemoteFilterShared::daXXStreamShared;
  DArray<DC_RemoteInStreamCopy*>  daInStreamCopy;
  DArray<DC_RemoteOutStreamCopy*> daOutStreamCopy;

  /// this is for per-ins state used by arg_t::insReadAny();
  struct insReadAnyState_t {
    bool fStreamSet;  ///< stream considered
  };
  insReadAnyState_t *rgReadAnyState;
  int iStreamReadAnyState;

  // runtime stats
  bool   fInProcessCall;  ///< to avoid counting "system" calls to read

      // stats
  Timer  tWallClock;  // lifetime of filter
  Timer  tWorkQ;      // time for blocking for work
  Timer  tInit;       // init() calls
  Timer  tFinalize;   // finalize() calls
  Timer  tProcess;    ///< process() calls
  // stats for i/o time within process() is in the Stream objects    
  bool   fReadFirst;  ///< either blocking read or insReadAny
  Timer  tReadFirst;
  Timer  tReadAny;  ///< time blocked in insReadAny()
  bool   fInstantWriteFirst;
  double rInstantWriteFirst; ///< time elapsed until first write


  double dLastPerformanceSendTime;
  void   initPerformanceUpdate();  
  void   sendPerformanceUpdate(bool forcesend=false);
    
public:
  DC_RemoteFilterCopy(DC_RemoteFilterShared *pFilterShared, int wCopyNum_in,
		      int wInitialWorkNum)
    : pShared(pFilterShared), wCopyNum(wCopyNum_in),
      sbFullName(NULL), sbCopyName(NULL),
      wWorkNum(wInitialWorkNum), rgReadAnyState(NULL), iStreamReadAnyState(0),
      fInProcessCall(false), fReadFirst(true), fInstantWriteFirst(false) {
  };
  ~DC_RemoteFilterCopy(void) {
    TestDelete(sbFullName);
    TestDelete(sbCopyName);
    TestDelete(rgReadAnyState);
  };

  void *thread_run(void *pArg);  ///< where filter threads start execution
  int thread_run_filter(void *pArg, MemoryBuf &mb);
  void _init(void);
};



class ReaderConnMap;  // defined in thread-reader.h

/// contains STREAM_CONNECT info for instances not yet created
class DC_RemotePending {
public:
  int   wInstanceNum;
  char *sbStreamName;
  int   iConnInfo;
  DC_StreamWritePolicy_t eStreamWritePolicy;  // for sink side
  ReaderConnMap *pMap;
  BaseSocket    *psock;
};

/// Represents the runtime state for the Remote Process
class DC_RemoteProcessState {
public:
  DC_FilterService_t *pDC;
  ConfigInfo cfg;
  bool fInit;

  /// all filters for all instances
  DArray<DC_RemoteFilterCopy*> daFilterCopy;
  DC_Mutex mutexFilterCopy;

  /// collection of premature STREAM_CONNECT requests
  DArray<DC_RemotePending*> daPending;
  DC_Mutex mutexPending;

  DC_Mutex mutexFilterLock;
  /// used for DC.FindAppConfigXXX(), and probably ConsolePrint() when written
  DC_Mutex mutexFilterLock_Internal;

  Types::uint iPlacement;  ///< our entry in console process daPlacementSpec[]
  char       *sbHostMe;

  /// sock to console process
  InetAddress  addrConsole;
  ClientSocket sockConsole;
  DC_Mutex     mutexSockConsole;

  // [0] = count of all full workQs
  // [1] = iConn to disable/enable
  // [2] = disabled=1, enabled=0
  DC_LockedCounter lcFullWorkQs;  ///< count of all full workQs

  // misc settings

public:
  DC_RemoteProcessState(DC_FilterService_t *pDC_in);
  ~DC_RemoteProcessState(void);
  int init(char *sbArg);

  void _FilterLock(DC_Mutex *pmutex) { pmutex->lock(); };
  void _FilterUnlock(DC_Mutex *pmutex) { pmutex->unlock(); };

  DC_RemoteInStreamShared *find_instream(char *sbName,
					 DArray<DC_RemoteFilterShared*> &da);
  DC_RemoteInStreamCopy *find_instream(int wInstanceNum, char *sbName);

  int ProcessConsoleCfgSettings(void);

  int StartInstance(int wInstanceNum, ConfigInfo &cfg, BaseSocket *psock);
  int StartInstance_AddNewCopies(DC_RemoteFilterShared *pFilterShared,
				 int nNewCopies, int wInitialWorkNum);
  int StartInstance_OpenStreamConn(int wInstanceNum);
  int StreamConnect(int wInstanceNum, char *sbStreamName, int iConnInfo,
		    DC_StreamWritePolicy_t eStreamWritePolicy,
		    ReaderConnMap *pMap, BaseSocket *psock,
		    bool fDoPending);
  int StreamConnectAck(int wInstanceNum, int iConnInfo, BaseSocket *psock);
  int CancelWork(int wInstanceNum, int wWorkNum); // CLH - added for new return types
  int AppendWork(int wInstanceNum, int wWorkNum, DC_Buffer_t *pbuf,
		 int iConnDisableIfFull=-1);
  int StopInstance(int wInstanceNum);
  int CleanupInstance(int wInstanceNum);
};


#endif /* H_DC_REMOTEPROCESSSTATE */
// ****************************************************************************
