// ****************************************************************************
//  Name: server-loop-base.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 05/16/1997 : mdb - Created.
//        08/20/1999 : mdb - Removed callback struct and use virtual fcns in
//                           a derived class.
//                         - Moved supporting structs and types inside class.
//                         - Changed name from ServerLoop to ServerLoop_Base
//        12/01/1999 : mdb - Fixed bug in OnePass where some signals could be
//                           postponed if received while not in select().
//        01/19/2000 : mdb - Cleaned up all the various close procedures.
//        04/08/2000 : mdb - Fixed race when signal not during select() call.
//                           Redesigned signal interface to be simpler.  Now
//                           signal handlers write an int to the pipe, and
//                           SigProcess() will be passed this int.
//  Desc: 
// ****************************************************************************
#if defined(AIX)
#  include <strings.h>  // for bzero() which is #def from FD_ZERO
#endif /* AIX */
#include <errno.h>
#include <string.h>

#include "lib/dc_machine.h"
#include "server-loop-base.h"

#if defined(DCDBG)
//#define DCDBG_SERVERLOOP
#endif /* DCDBG */


ServerLoop_Base::~ServerLoop_Base(void) {
  Flush();
}

void ServerLoop_Base::Init(ServerSocket &sockLocal_in, int maxClients_in) {
  psockLocal = &sockLocal_in;

  numClients = 0;
  maxClients = maxClients_in;
  rgClients = new ClientConn_t[maxClients];

  iClientStart = iClientMax = 0;

  for (int i=0; i<maxClients; ++i) {
    rgClients[i].psock = NULL;
    rgClients[i].paddr = NULL;
    rgClients[i].fExternal = false;
    rgClients[i].fEnabled = true;
    rgClients[i].map = NULL;
  }

  tvNoDelay.tv_sec = 0;
  tvNoDelay.tv_usec = 0;
  tvDelay.tv_sec = 1;
  tvDelay.tv_usec = 0;
  tv = tvDelay;

  fdMaxStatic = fdMax = psockLocal->getFd();

  FD_ZERO(&fdsetIn);
  FD_SET(psockLocal->getFd(), &fdsetIn);  // for new client connections

  fSig = false;

  nSelectErrors = 0;
}

/**
 * This sets the signal mask for the signals to be blocked while SigProcess()
 * is called.  It should contain all signals that you want to react to in
 * SigProcess().  Other signals can be delivered, such as sig_pipe, and
 * not included in this mask setting.
 * rtn: pipe fd used by signal handler to write an int for SigProcess() use
 */
#ifndef _WIN32
int ServerLoop_Base::setSig(sigset_t maskSigAny_in) {
  fSig = true;
  maskSigAny = maskSigAny_in;
  if (pipe(rgfdSigPipe) == -1) {
    cerr << ErrorHeader(wERROR) << "ServerLoop_Base(): Failed pipe() call - "
	 << strerror(errno) << endl;
    return -1;
  }
  // update for this static fd in addition to the local server socket
  if (rgfdSigPipe[0] > fdMax) {
    fdMaxStatic = fdMax = rgfdSigPipe[0];
  }
  FD_SET(rgfdSigPipe[0], &fdsetIn);  // for signal handler to select IPC
  return rgfdSigPipe[1];
}
#endif

int ServerLoop_Base::Flush(void) {
  CloseAllConn();
  TestDeleteNull(rgClients);
  return 0;
}


void ServerLoop_Base::CloseConn(int i) {
  int fd;
  fd = rgClients[i].psock->getFd();
  if (fd == -1) return;  // already closed?  03/01/2001 : mdb - for bugfix

#if defined(DCDBG_SERVERLOOP)
  cout << "CloseConn(): Closing [" << i << "] " << fd << endl;
#endif /* DCDBG_SERVERLOOP */

  MapProcess(i, CLOSE_CONN);

  FD_CLR(fd, &fdsetIn);
  rgClients[i].psock->close();  // assigns -1 to fd
  if (! rgClients[i].fExternal) {
#if defined(DCDBG_SERVERLOOP)
    cout << "CloseConn(): delete [" << i << "] " << endl;
#endif /* DCDBG_SERVERLOOP */
    delete rgClients[i].psock;
    delete rgClients[i].paddr;
  }
  rgClients[i].psock = NULL;
  rgClients[i].paddr = NULL;
  --numClients;

  if (fd == fdMax)     fdMax = FindMaxFd(fdMaxStatic);
  if (i == iClientMax) iClientMax = FindMaxClientI(iClientMax);

  if (numClients == 0) {
    tv = tvDelay;
  }
}

void ServerLoop_Base::CloseAllConn(void) {
  int i;
  for (i=0; numClients; ++i) {
    if (rgClients[i].psock &&
	rgClients[i].psock->getFd() != -1) {
      CloseConn(i);  // this decrements numClients
    }
  }
  if (fSig) {
    if (rgfdSigPipe[0] != -1) close(rgfdSigPipe[0]);
    if (rgfdSigPipe[1] != -1) close(rgfdSigPipe[1]);
  }
  return;
}

int ServerLoop_Base::FindMaxFd(int fdSock) {
  int i, n, fd, fdMax = fdSock;
  for (i=n=0; n<numClients; ++i) {
    if (rgClients[i].psock &&
	(fd = rgClients[i].psock->getFd()) != -1) {
      ++n;
      if (rgClients[i].fEnabled && fd>fdMax) fdMax = fd;
    }
  }
  return fdMax;
}

int ServerLoop_Base::FindMaxClientI(int iOldMax) {
  int i;
  if (numClients > 0) {
    for (i=iOldMax; i>=0; --i) {
      if (rgClients[i].psock &&
	  rgClients[i].psock->getFd() != -1 &&
	  rgClients[i].fEnabled) {
	return i;
      }
    }
  }
  return 0;
}

/**
 * rtn: <0=error, else index to rgClients[]
 */
int ServerLoop_Base::AddNewConn(BaseSocket *psock, InetAddress *paddr,
				bool fExternal) {
  int i, fd;
  if (!psock || (fd = psock->getFd()) == -1) return -1; // bad sock

  // find an empty client slot
  for (i=0; i<maxClients; ++i) {
    if (rgClients[i].psock != NULL) continue;

    rgClients[i].psock = psock;
    rgClients[i].paddr = paddr;
    rgClients[i].fExternal = fExternal;
    rgClients[i].fEnabled = true;
    rgClients[i].map = NULL;
    ++numClients;

#if defined(DCDBG_SERVERLOOP)
    cout << "AddNewConn(): Adding [" << i << "] "
	 << psock->getFd() << ", " << *paddr << ", "
	 << (fExternal ? "external" : "internal")
	 << " numClients=" << numClients << endl;
#endif /* DCDBG_SERVERLOOP */

    fd = psock->getFd();
    FD_SET(fd, &fdsetIn);

    if (fd > fdMax)     fdMax = fd;
    if (i > iClientMax) iClientMax = i;

    tv = tvNoDelay;
      
    // BUGBUG: don't know if I want to do it this way any longer
    if (!IsBigEndian()) rgClients[i].psock->setEndian(1, 1);

    MapProcess(i, ADD_CONN);

    return i;
  }
  return -2;  // no room
}


// ****************************************************************************
//  Name: DisableConn,EnableConn
//  Date: 02/24/2001 : mdb - Created.
// ****************************************************************************
/**
 *  Temporarily pauses/restores the connection from being checked.
 *  Rtn:  int -
 */
int ServerLoop_Base::DisableConn(int i) {
  if (!rgClients[i].fEnabled) return 0;  // already disabled

  int fd = rgClients[i].psock->getFd();

#if defined(DCDBG_SERVERLOOP)
  cout << "DisableConn(): Disabling [" << i << "] " << fd << endl;
#endif /* DCDBG_SERVERLOOP */
  rgClients[i].fEnabled = false;
  FD_CLR(fd, &fdsetIn);

  if (fd == fdMax)     fdMax = FindMaxFd(fdMaxStatic);
  if (i == iClientMax) iClientMax = FindMaxClientI(iClientMax);
  return 0;
}

int ServerLoop_Base::EnableConn(int i) {
  if (rgClients[i].fEnabled) return 0;  // already enabled

  int fd = rgClients[i].psock->getFd();

#if defined(DCDBG_SERVERLOOP)
  cout << "EnableConn(): Enabling [" << i << "] " << fd << endl;
#endif /* DCDBG_SERVERLOOP */
  rgClients[i].fEnabled = true;
  FD_SET(fd, &fdsetIn);

  if (fd > fdMax)     fdMax = fd;
  if (i > iClientMax) iClientMax = i;

  tv = tvNoDelay;
  return 0;
}


// ****************************************************************************
//  Name: AddExternalConn
//  Date: 08/16/2000 : mdb - Changed to support external setting of user map.
// ****************************************************************************
/**
 *  Adds an externally created socket to the internal server loop for
 *  processing.  The only difference is they are not deleted in the end
 *  as are internally accepted sockets.
 *
 *  Rtn:  int -
 */
int ServerLoop_Base::AddExternalConn(BaseSocket *psock, InetAddress *paddr) {
  return AddNewConn(psock, paddr, true);
}


// ****************************************************************************
//  Name: OnePass
//  Date: 03/06/1998 : mdb - Rewrote to only make one sweep.
//        08/14/1999 : mdb - Added signal handling to the sweep, as in noded.
//        11/02/2000 : mdb - Fixed subtle bug with AddNewConn() changing
//                           numClients, potentially reading the same socket
//                           twice.  (thanks to Chialin Chang for this fix)
// ****************************************************************************
/**
 *  Checks all known connections for one packet on each
 *
 *  Rtn:  int - 1=error, 0=may have processed packets
 */

int ServerLoop_Base::OnePass(bool fBlock) {
  int i, n, wTmp, wRtnSig;
  int fd;
  InetAddress addr;
  fd_set fdsetOut;


  // check descriptors
  do {
    fdsetOut = fdsetIn;
    if (fBlock) {
      wTmp = ::select(fdMax+1, &fdsetOut, NULL, NULL, NULL);
    } else {
      wTmp = ::select(fdMax+1, &fdsetOut, NULL, NULL, &tv);
    }
  } while (wTmp == -1 && errno == EINTR);
  if (wTmp == -1) {
    int err = errno;
    if (++nSelectErrors <= 4) {
      cout << "ServerLoop_Base::OnePass(): select failure (errno=" << err
	   << ") : " << strerror(err) << ' ' << nSelectErrors << endl;
    }
#ifndef _WIN32
	return SLE_ERROR;
#else
	return SLE_ERROR_WIN;
#endif
  } else if (wTmp == 0) {
    // do some computation
    WaitProcess();
    return SLE_OK;  // nothing else to do
  }
#if defined(DCDBG_SERVERLOOP)
  cout << "ServerLoop_Base::OnePass(): select ret=" << wTmp << endl;
#endif /* DCDBG_SERVERLOOP */

  // any delivered signals?
  if (fSig && rgfdSigPipe[0] != -1 && FD_ISSET(rgfdSigPipe[0], &fdsetOut)) {
    int pipeval;
    read(rgfdSigPipe[0], &pipeval, sizeof(pipeval));
    
#ifndef _WIN32
    sigprocmask(SIG_BLOCK, &maskSigAny, NULL);
#endif

#if defined(DCDBG_SERVERLOOP)
    cout << "OnePass(): CloseAllConn() and return SLE_SHUTDOWN\n";
#endif /* DCDBG_SERVERLOOP */
    wRtnSig = SigProcess(pipeval);
    switch (wRtnSig) {
    case SLE_CB_SHUTDOWN:  // shutdown
      // for now, just close all connections, and exit
#if defined(DCDBG_SERVERLOOP)
      cout << "OnePass(): CloseAllConn() and return SLE_SHUTDOWN\n";
#endif /* DCDBG_SERVERLOOP */
      CloseAllConn();
      return SLE_SHUTDOWN;
      break;
    }
#ifndef _WIN32
    sigprocmask(SIG_UNBLOCK, &maskSigAny, NULL);    
#endif
  }

  // anything pending on client connections?
  if (numClients > 0) {
    for (i=iClientStart, n=0; n<numClients; ) {
      if (rgClients[i].psock && (fd = rgClients[i].psock->getFd()) != -1) {
	++n;
	// check if still enabled (could have been disabled in this loop)
	if (rgClients[i].fEnabled && FD_ISSET(fd, &fdsetOut)) {
#if defined(DCDBG_SERVERLOOP)
	  cout << "ServerLoop_Base(): ConnProcess on connection from "
	  	 << *rgClients[i].paddr << " (fd=" << fd << ")\n";
#endif /* DCDBG_SERVERLOOP */

	  // data to read on client connection
	  switch (ClientProcess(i)) {
	  case SLE_CB_CLOSECONN:  // connection closed
#if defined(DCDBG_SERVERLOOP)
	    cout << "ServerLoop_Base::OnePass(): Closed connection from "
		 << *rgClients[i].paddr << endl;
#endif /* DCDBG_SERVERLOOP */
	    CloseConn(i);
	    break;
	  case SLE_CB_SHUTDOWN:  // shutdown
	    // for now, just close all connections, and exit
	    CloseAllConn();
	    return SLE_SHUTDOWN;
	    break;
	  }

	  // FIX: This is to help with problems seen when processing a
	  // given iClient entry can cause others to be added or
	  // removed that may still need to be handled in *this*
	  // fdsetOut check loop.
	  FD_CLR(fd, &fdsetOut);
	}

	if (numClients == 0) break;
      }
      ++i;
      if (i>iClientMax) i=0;

      // FIX: sanity check to stop the sweep of connections is we hit
      // the one we started with.
      if (i == iClientStart) break;
    }
    ++iClientStart; if (iClientStart>iClientMax) iClientStart=0;
  }

  // any new connections?
  if (FD_ISSET(psockLocal->getFd(), &fdsetOut)) {
    if (numClients < maxClients) {
      BaseSocket  *psock = new BaseSocket;
      InetAddress *paddr = new InetAddress;

      if ((fd = psockLocal->accept(*paddr)) >= 0) {
	psock->getFd() = fd;
	i = AddNewConn(psock, paddr, false);
#if defined(DCDBG_SERVERLOOP)
	cout << "ServerLoop_Base::OnePass(): Accepted connection from "
	     << *paddr << " (fd=" << fd << ")\n";
#endif /* DCDBG_SERVERLOOP */
      } else {
	delete [] psock;
	delete [] paddr;
	cerr << "ServerLoop_Base::OnePass(): Failed accept() call" << endl;
      }
    } else {
      // BUGBUG: Perhaps better is to catch this ahead of time.  If there is
      //         no room, don't add the listen socket to the select set, as
      //         described in [Stevens] UNPv2.

      // no room for more clients
      close(psockLocal->accept(addr));
      cerr << "ServerLoop_Base::OnePass(): Refused connection to "
	   << addr << endl;
    }
  }

  return SLE_OK;
}


int ServerLoop_Base::MapProcess(int i, enum eMapReason e) {
#if 0
  // example
  switch (e) {
  case ADD_CONN:
    void *p = new mystruct;
    p->somefield = 123;
    setClientMapValue(i, p);
    break;
  case DISABLE_CONN:
    break;
  case ENABLE_CONN:
    break;
  case CLOSE_CONN:
    p = getClientMapValue(i);
    setClientMapValue(i, NULL);
    delete p;
    break;
  }
#endif
  return SLE_CB_OK;
}

// ****************************************************************************
