// -*- c++ -*-
// ****************************************************************************
//  Name: server-loop-base.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 06/26/1997 : mdb - Created.
//        08/20/1999 : mdb - Removed callback struct and use virtual fcns in
//                           a derived class.
//                         - Moved supporting structs and types inside class.
//                         - Changed name from ServerLoop to ServerLoop_Base
//        08/25/1999 : mdb - Need to add sockets that were created externally
//                           to the loop processing.
//        04/07/2000 : mdb - Added userdef mapping of client connections.
//        04/08/2000 : mdb - Fixed race when signal not during select() call.
//                           Redesigned signal interface to be simpler.  Now
//                           signal handlers write an int to the pipe, and
//                           SigProcess() will be passed this int.
//        02/22/2001 : mdb - Can enable/disable connection from consideration.
//  Desc: 
// ****************************************************************************

#if !defined(H_DC_SERVER_LOOP_BASE)
#define H_DC_SERVER_LOOP_BASE

#ifdef _WIN32
#include <pthread.h>
#endif

#include <sys/types.h>
#include <signal.h>

#include "lib/dc_socket.h"
#include "lib/dc_timer.h"

#define SERVER_LOOP_DEFAULT_MAX_CLIENTS  (64)


class ServerLoop_Base {
public:
  typedef struct {
    BaseSocket  *psock;    ///< fd != -1 --> live connection
    InetAddress *paddr;

    /**
     * f=sock and addr when we accepted it server style (allocated)
     * t=external sock created (presumably) by a connect call (not allocated)
     */
    bool fExternal;

    /**
     * t=normal
     * f=do not check for activity in select()
     */
    bool fEnabled;

    void *map;  ///< user defined map value for the client connection
  } ClientConn_t;

  /// returned by OnePass
  enum onepass_e {
    SLE_SHUTDOWN = 1,
    SLE_OK       = 0,
#ifndef _WIN32
	SLE_ERROR    = -1
#else
    SLE_ERROR_WIN = -1
#endif
  };

  /// for client callbacks to return
  enum callback_e {
    SLE_CB_OK        = 0,
    SLE_CB_CLOSECONN = -1,
    SLE_CB_SHUTDOWN  = -2
  };

protected:
  ServerSocket *psockLocal;  ///< listen socket
  bool     fSig;
#ifndef _WIN32
  sigset_t maskSigAny;
#endif

  int      rgfdSigPipe[2];

  fd_set fdsetIn;
  int iClientStart, iClientMax;  ///< used to make select fair
  struct timeval tv, tvNoDelay, tvDelay;
  int fdMax, fdMaxStatic;

  int nSelectErrors;

  void CloseAllConn(void);
  int  FindMaxFd(int fdSock);
  int  FindMaxClientI(int iOldMax);

  int  AddNewConn(BaseSocket *psock, InetAddress *paddr, bool fExternal);
  void CloseConn(int i);

public:
  int           numClients;
  int           maxClients;
  ClientConn_t *rgClients;  ///< used for storing client connections
  
public:
  ServerLoop_Base(void)
    : fSig(false), numClients(0), maxClients(0), rgClients(NULL) {};
  virtual ~ServerLoop_Base(void);

  /// must be called before any other member fcns
  void Init(ServerSocket &sockLocal_in, 
	    int maxClients_in = SERVER_LOOP_DEFAULT_MAX_CLIENTS);

#ifndef _WIN32
  sigset_t &getMaskSigAny(void) { return maskSigAny; };
  int setSig(sigset_t maskSigAny_in);
#endif

  bool isDisabled(int i) { return !rgClients[i].fEnabled; };
  int  DisableConn(int i);
  bool isEnabled(int i) { return rgClients[i].fEnabled; };
  int  EnableConn(int i);

  int AddExternalConn(BaseSocket *psock, InetAddress *paddr);
  int OnePass(bool fBlock=true);

  int Flush();

  void *getClientMapValue(int i) { return rgClients[i].map; };
  void setClientMapValue(int i, void *map) { rgClients[i].map = map; };
  /// for calling MapProcess
  enum eMapReason {ADD_CONN, DISABLE_CONN, ENABLE_CONN, CLOSE_CONN};

public:  // callback routines to be redefined in derived class
  virtual int ClientProcess(int i) = 0;
  virtual int WaitProcess(void) { return SLE_CB_OK; };
  virtual int SigProcess(int pipeval) { return SLE_CB_OK; };

  virtual int MapProcess(int i, enum eMapReason e);
};


#endif /* H_DC_SERVER_LOOP_BASE */
// ****************************************************************************
