// ****************************************************************************
//  Name: streambufferqueue.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 02/22/2001 : mdb - Created.
//        03/02/2001 : mdb - Fixed hang bug in localDequeue().
//        04/17/2001 : mdb - Rewrote localDequeue() to simplify and hopefully
//                           fix the dequeue zero len bug.  Don't count EOW/EOS
//                           as strictly -- instead base on wWorkNum.
//        04/22/2001 : mdb - Use a container for items in queue to add psock.
//        07/17/2001 : mdb - Fixed bug to DATA_ACK a colocated producer.
//        07/27/2001 : mdb - Implemented write policy subsets.
//  Desc: Part of remote lib process stream abstraction.  Layered above simple
//        queue to support multiple local/remote producers and multiple
//        consumers. (copies)  The invariant of all work k buffers appear
//        before work k+1 delimited by an EOW is preserved.
// ****************************************************************************

#include <assert.h>

#include "lib/dc_memory.h"

#include "dc_filterservice.h"
#include "remoteprocessstate.h"
#include "streambufferqueue.h"
#include "packet.h"
#include "packet-tags.h"

#if defined(DCDBG)
//#define DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE
//#define DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE_DATAACK
#endif /* DCDBG */

#define BUGFIX

/**
 * ctor
 */
DC_StreamBufferQueue::DC_StreamBufferQueue(void)
  : wEnqueueWorkNum(0), nProducerEOW(0), nProducerEOS(0),
    nLocalProducers(0), nRemoteProducers(0), nTotalProducers(0),
    wDequeueWorkNum(0), nConsumerEOW(0), nConsumerEOS(0), fConsumerEOS(false),
    nTotalConsumers(0), fFirstDequeue(false), nEarlyDequeue(0),
    fRemoteDisabledEarly(false), fRemoteDisabledFull(false),
    pSL(NULL) {
}

/**
 * dtor
 */
DC_StreamBufferQueue::~DC_StreamBufferQueue(void) {
}

#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
#include <strstream.h>
void DC_StreamBufferQueue::_DumpState(ostream &os) {
  char bufTmp[2048];
  ostrstream sout(bufTmp, 2048);

  sout << "[t" << (int)pthread_self() << "] StreamBufferQueue()::_DumpState():"
       << endl
       << "  name=" << sbName
       << ", num=" << num << ", max=" << max
       << ", shutdown=" << (buf ? "no" : "yes") << endl
       << "  enqueue: work=" << wEnqueueWorkNum
       << ", EOW=" << nProducerEOW
       << ", EOS=" << nProducerEOS
       << ", local=" << nLocalProducers
       << ", remote=" << nRemoteProducers
       << ", tot=" << nTotalProducers << endl
       << "  dequeue: work=" << wDequeueWorkNum
       << ", EOW=" << nConsumerEOW
       << ", EOS=" << nConsumerEOS
       << ", tot=" << nTotalConsumers << endl;

  sout << "  early: "
       << (fRemoteDisabledEarly ? "T" : "F") << " ";
  for (unsigned int i=0; i<daRemoteDisabledEarly.NumElem(); ++i) {
    RemoteDisabled rd = daRemoteDisabledEarly.Get(i);
    if (rd.wWorkNum == -1) continue;
    sout << "<w" << rd.wWorkNum << ",c" << rd.iConn << "> ";
  }
  sout << endl;

  sout << "  full: "
       << (fRemoteDisabledFull ? "T" : "F") << " ";
  for (int i=listRemoteDisabledFull.StartIter(); i; --i) {
    RemoteDisabled rd = listRemoteDisabledFull.Next();
    if (rd.wWorkNum == -1) continue;
    sout << "<w" << rd.wWorkNum << ",c" << rd.iConn << "> ";
  }
  sout << endl;

  sout << '\0';
  os << bufTmp;
}
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */

// unlocked -- assume caller holds the lock
int DC_StreamBufferQueue::_IncrementEnqueueWorkNum(bool fReaderThread) {
  // ASSUME: the next set of work to expect must be incremented by exactly one
  ++wEnqueueWorkNum;

  // Note: since fRemoteDisabledEarly, we know nRemoteProducers>0
  if (fRemoteDisabledEarly) {
    // look thru list of disabled iConns, and enable them as needed
    fRemoteDisabledEarly = false;  // assume we will enable them all
    for (unsigned int i=0; i<daRemoteDisabledEarly.NumElem(); ++i) {
      RemoteDisabled rd = daRemoteDisabledEarly.Get(i);
      if (rd.wWorkNum == -1) continue;
      if (rd.wWorkNum == wEnqueueWorkNum) {
	// enable this connection
	if (fReaderThread) {
	  // call the Enable fcn directly
	  pSL->EnableConn(rd.iConn);
	} else {
	  // send ReaderCmd so this is done in the context of the reader thread
	  pSL->ReaderCmd_EnableConn(rd.iConn);
	}
	rd.wWorkNum = -1;  // "remove" from list
	daRemoteDisabledEarly.Set(i, rd);
      } else {
	fRemoteDisabledEarly = true;  // some left in list
      }
    }
  }
  if (nLocalProducers) {
    cvLocalProducerReady.broadcast();
  }
  return 0;
}

// unlocked -- assume caller holds the lock
int DC_StreamBufferQueue::_IncrementDequeueWorkNum(void) {
  // ASSUME: the next set of work to expect must be incremented by exactly one
  ++wDequeueWorkNum;

  if (nTotalConsumers) {
    cvLocalConsumerReady.broadcast();	
  }
  return 0;
}

int DC_StreamBufferQueue::setInitialWorkNum(int wWorkNum) {
  lockInternal();
  wEnqueueWorkNum = wWorkNum;
  wDequeueWorkNum = wWorkNum;
  unlockInternal();
  return 0;
}

#if 0  /* MDB */
/**
 * used for later localDequeue() calls to decide if a DATA_ACK is needed
 */
void DC_StreamBufferQueue::setStreamWritePolicy(DC_StreamWritePolicy_t
						eStreamWritePolicy_in) {
  eStreamWritePolicy = eStreamWritePolicy_in;
}
#endif

int DC_StreamBufferQueue::localAddProducer(int n) {
  lockInternal();
  // BUGBUG: What if a collective EOW/EOS is in progress?
  nLocalProducers += n;
  nTotalProducers += n;
  unlockInternal();
  return 0;
}

int DC_StreamBufferQueue::localRemoveProducer(void) {
  lockInternal();
  // BUGBUG: What if a collective EOW/EOS is in progress?
  --nLocalProducers;
  --nTotalProducers;
  unlockInternal();
  return 0;
}

// thread: remote process filter
int DC_StreamBufferQueue::
localEnqueue(int wWorkNum, DC_Buffer_t *pbuf,
	     DC_StreamWritePolicy_t eStreamWritePolicy,
	     int iConn, DC_LockedCounterThreshold *plcUnAckedBuffers) {
  int wRtn;
  DC_StreamBufferQueue_Item item;
  lockInternal();

#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
  char sbFilterName[80];
  snprintf(sbFilterName, 80, "[t%d,filter,w%d]: localEnqueue(): ",
	   (int)pthread_self(), wWorkNum);
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */

#if 0
  assert(wWorkNum >= wEnqueueWorkNum);
#else
  if (wWorkNum < wEnqueueWorkNum) {
    cerr << "localEnqueue(): ERROR: buffer for work " << wWorkNum
	 << " < enqueue work " << wEnqueueWorkNum << endl
	 << "- for buffer " << *pbuf << endl;
  }
#endif

#if !defined(BUGFIX)
  // bump collective counts as needed
  if (pbuf->eType == DC_Buffer_t::EOW) {
    ++nProducerEOW;
  } else if (pbuf->eType == DC_Buffer_t::EOS) {
    ++nProducerEOS;
  }
#endif /* BUGFIX */

  // block source filter if too early
  while (wWorkNum > wEnqueueWorkNum) {
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
    cout << sbFilterName << "block on future work=" << wWorkNum
	 << " (enqueue work=" << wEnqueueWorkNum << ")" << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
    cvLocalProducerReady.wait(getInternal());  // sleep if too early
    if (isShutdown(false)) {
      unlockInternal();
      return QUEUE_SHUTDOWN;
    }
  }

  // add to the queue
  switch (pbuf->eType) {
  case DC_Buffer_t::DATA:
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
    cout << sbFilterName << "enqueue DATA for work=" << wWorkNum
	 << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
    item.pbuf = pbuf;
    item.eStreamWritePolicy = eStreamWritePolicy;
    item.iConn = iConn;
    item.psock = NULL;
    item.plcUnAckedBuffers = plcUnAckedBuffers;
    wRtn = enqueue(&item, false);  //TODO: make sure ok (pass & local storage)
    break;
  case DC_Buffer_t::EOW:
#if !defined(BUGFIX)
    if (nProducerEOW == nTotalProducers) {  // last?
#else /* BUGFIX */
    if (++nProducerEOW == nTotalProducers) {  // last?
#endif /* BUGFIX */
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
      cout << sbFilterName << "enqueue EOW work=" << wWorkNum << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
      item.pbuf = pbuf;
      item.eStreamWritePolicy = eStreamWritePolicy;
      item.iConn = -1;
      item.psock = NULL;
      item.plcUnAckedBuffers = NULL;
      wRtn = enqueue(&item, false);
      nProducerEOW = 0; _IncrementEnqueueWorkNum(false);
    } else {
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
      cout << sbFilterName << "consume EOW work=" << wWorkNum
	   << " (" << nProducerEOW << "/" << nTotalProducers << " copies)"
	   << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
      pbuf->consume();
      wRtn = 0;
    }    
    break;
  case DC_Buffer_t::EOS:
#if !defined(BUGFIX)
    if (nProducerEOS == nTotalProducers) {  // last?
#else /* BUGFIX */
    if (++nProducerEOS == nTotalProducers) {  // last?
#endif /* BUGFIX */
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
      cout << sbFilterName << "enqueue EOS work=" << wWorkNum << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
      item.pbuf = pbuf;
      item.eStreamWritePolicy = eStreamWritePolicy;
      item.iConn = -1;
      item.psock = NULL;
      item.plcUnAckedBuffers = NULL;
      wRtn = enqueue(&item, false);
      nProducerEOS = 0;
    } else {
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
      cout << sbFilterName << "consume EOS work=" << wWorkNum
	   << " (" << nProducerEOS << "/" << nTotalProducers << " copies)"
	   << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
      pbuf->consume();
      wRtn = 0;
    }    
    break;
  }

  unlockInternal();
  return wRtn;
}


int DC_StreamBufferQueue::remoteAddProducer(int n) {
  lockInternal();
  // BUGBUG: What if a collective EOW/EOS is in progress?
  nRemoteProducers += n;
  nTotalProducers += n;
  unlockInternal();
  return 0;
}
int DC_StreamBufferQueue::remoteRemoveProducer(void) {
  lockInternal();
  // BUGBUG: What if a collective EOW/EOS is in progress?
  --nRemoteProducers;
  --nTotalProducers;
  unlockInternal();
  return 0;
}

/**
 * rtn: bool - t=ok to enqueue, f=not ok to enqueue and conn is disabled
 */
// unlocked -- assume caller holds the lock
// thread: reader
bool DC_StreamBufferQueue::remoteCheckWorkNum(int wWorkNum,
					      DC_ReaderServerLoop *pSL_in,
					      int iConn) {
  // make sure that the work to be read from the socket is not too far ahead

#if 0
  assert(wWorkNum >= wEnqueueWorkNum);
#else
  if (wWorkNum < wEnqueueWorkNum) {
    cerr << "remoteCheckWorkNum(): ERROR: buffer for work " << wWorkNum
	 << " < enqueue work " << wEnqueueWorkNum << endl;
  }
#endif
  bool fOk = (wWorkNum == wEnqueueWorkNum);
  if (!fOk) {
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
    cout << "[reader] remoteCheckWorkNum(): disable conn " << iConn
	 << ", ahead work=" << wWorkNum << " enqueue work=" << wEnqueueWorkNum
	 << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
    pSL = pSL_in;
    // While holding the lock, we need to disable the sockets in the
    // server loop.  This avoids the race if we were to do this in the
    // caller: If the caller was preempted right after this call but
    // before the call to DisableConn(), the last call to
    // localProducers() could slip in.
    pSL->DisableConn(iConn);

    // Enabled by either the reader thread or a filter thread,
    // depending on where the source copies live, and the order of UOW
    // completion.

    // append to list
    fRemoteDisabledEarly = true;
    RemoteDisabled rd;
    rd.iConn = iConn;
    rd.wWorkNum = wWorkNum;
    unsigned int i;
    for (i=0; i<daRemoteDisabledEarly.NumElem(); ++i) {
      RemoteDisabled rdTmp = daRemoteDisabledEarly.Get(i);
      if (rdTmp.wWorkNum == -1) break;
    }
    if (i < daRemoteDisabledEarly.NumElem()) {
      daRemoteDisabledEarly.Set(i, rd);
    } else {
      daRemoteDisabledEarly.Append(rd);
    }
  }
  return fOk;
}

/**
 * queue is full, disable the remote connection until later
 */
// unlocked -- assume caller holds the lock
// thread: reader
int DC_StreamBufferQueue::remoteDisableFull(DC_ReaderServerLoop *pSL_in,
					    int iConn) {
  // NOTE: wWorkNum is not really needed or used
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
  cout << "[reader] remoteDisableFull(): disable conn " << iConn << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
  pSL = pSL_in;
  pSL->DisableConn(iConn);

  // append to list
  fRemoteDisabledFull = true;
  RemoteDisabled rd;
  rd.iConn = iConn;
  rd.wWorkNum = 0;
  listRemoteDisabledFull.InsTail(rd);
  return 0;
}
// unlocked -- assume caller holds the lock
// thread: reader
int DC_StreamBufferQueue::
remoteEnqueue(int wWorkNum, DC_Buffer_t *pbuf,
	      DC_StreamWritePolicy_t eStreamWritePolicy,
	      int iConn, BaseSocket *psock) {
#if 0
  // should not be called if remoteCheckWorkNum() failed
  assert(wWorkNum == wEnqueueWorkNum);
#else
  if (wWorkNum < wEnqueueWorkNum) {
    cerr << "remoteEnqueue(): ERROR: buffer for work " << wWorkNum
	 << " < enqueue work " << wEnqueueWorkNum << endl;
  } else if (wWorkNum > wEnqueueWorkNum) {
    cerr << "remoteEnqueue(): ERROR: buffer for work " << wWorkNum
	 << " > enqueue work " << wEnqueueWorkNum << endl;
  }
#endif

  // add to the queue
  int wRtn;
  DC_StreamBufferQueue_Item item;
  switch (pbuf->eType) {
  case DC_Buffer_t::DATA:
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
    cout << "[reader] remoteEnqueue(): enqueue DATA for work=" << wWorkNum
	 << " save iConn=" << iConn << ", psock->fd=" << psock->getFd()
	 << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
    item.pbuf = pbuf;
    item.eStreamWritePolicy = eStreamWritePolicy;
    item.iConn = iConn;
    item.psock = psock;
    item.plcUnAckedBuffers = NULL;
    wRtn = enqueue(&item, false);
    break;
  case DC_Buffer_t::EOW:
    if (++nProducerEOW == nTotalProducers) {  // last?
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
      cout << "[reader] remoteEnqueue(): enqueue EOW for work=" << wWorkNum
	   << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
      item.pbuf = pbuf;
      item.eStreamWritePolicy = eStreamWritePolicy;
      item.iConn = -1;
      item.psock = NULL;
      item.plcUnAckedBuffers = NULL;
      wRtn = enqueue(&item, false);
      nProducerEOW = 0; _IncrementEnqueueWorkNum(true);
    } else {
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
      cout << "[reader] remoteEnqueue(): consume EOW work=" << wWorkNum
	   << " (" << nProducerEOW << "/" << nTotalProducers << " copies)"
	   << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
      pbuf->consume();
      wRtn = 0;
    }
    break;
  case DC_Buffer_t::EOS:
    if (++nProducerEOS == nTotalProducers) {  // last?
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
      cout << "[reader] remoteEnqueue(): enqueue EOS for work=" << wWorkNum
	   << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
      item.pbuf = pbuf;
      item.eStreamWritePolicy = eStreamWritePolicy;
      item.iConn = -1;
      item.psock = NULL;
      item.plcUnAckedBuffers = NULL;
      wRtn = enqueue(&item, false);
      nProducerEOS = 0;
    } else {
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
      cout << "[reader] remoteEnqueue(): consume EOS work=" << wWorkNum
	   << " (" << nProducerEOS << "/" << nTotalProducers << " copies)"
	   << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
      pbuf->consume();
      wRtn = 0;
    }
    break;
  }
  return wRtn;
}


int DC_StreamBufferQueue::localAddConsumer(int n) {
  lockInternal();
  nTotalConsumers += n;
  unlockInternal();
  return 0;
}

/**
 * in: fLock - needed by localShutdown() calls
 */
// thread: remote process filter
int DC_StreamBufferQueue::localDequeue(int wWorkNum, DC_Buffer_t **ppbuf,
				       bool fLock) {
  DC_StreamBufferQueue_Item item;

  if (fLock) lockInternal();

  if (isShutdown(false)) {
    if (fLock) unlockInternal();
    return QUEUE_SHUTDOWN;
  }

#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
  char sbFilterName[80];
  snprintf(sbFilterName, 80, "[t%d,filter,w%d]: localDequeue(): ",
	   (int)pthread_self(), wWorkNum);
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */

  // BUGBUG: (race fairness) if the queue becomes not-empty with this
  // call we enable the remote sockets.  After the reader enables
  // them, but before it can call select() again, read data, and
  // enqueue anything, a local copy gets the mutex and enqueues.  Now
  // the sockets are enabled, but the queue is full again, which means
  // they are disabled again.  The fairness issue is caused by a local
  // filter copy effectively having higher priority than remote copies
  // due to the less overhead for enqueuing.  This is ignored for now.

  // block consumers with early worknum
  while (wWorkNum > wDequeueWorkNum) {
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
    cout << sbFilterName << "block on future work=" << wWorkNum
	 << " (dequeue work=" << wDequeueWorkNum << ")" << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
    if (++nEarlyDequeue == nTotalConsumers) {  // last consumer early?
      nEarlyDequeue = 0;
      // drain queue up to next EOW and wake early consumers
      while (probe(false) > 0) {
	if (dequeue(&item, -1, false) != QUEUE_OK) break;
#if 1 || defined(DCDBG_FILTERSERVICE_QUEUE)
	cout << "localDequeue(): all-consumer-early drain "
	     << *item.pbuf << endl;
#endif /* DCDBG_FILTERSERVICE_QUEUE */
	if (item.pbuf->eType == DC_Buffer_t::EOW) {
	  _IncrementDequeueWorkNum();
	  break;
	}
	item.pbuf->consume();
      }
    } else {  // sleep if too early?
      cvLocalConsumerReady.wait(getInternal());  /*** RELEASE MUTEX ***/
    }
    if (isShutdown(false)) {
      if (fLock) unlockInternal();
      return QUEUE_SHUTDOWN;
    }
  }

  // since a dequeue of a EOW/EOS buffer must be replicated to all
  // copies, we cannot have >1 consumer block in dequeue() on an empty
  // queue in case the next buffer read in EOW/EOS.  Here, we simply
  // block other consumers from the actual dequeue() call while
  // another is blocked in dequeue().  We do this before the collective
  // EOW/EOS checks because the other consumer may be reading EOW/EOS.
  while (fFirstDequeue) {
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
    cout << sbFilterName << "block because another filter is dequeuing"
	 << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
    cvFirstDequeue.wait(getInternal());  /*** RELEASE MUTEX ***/
    if (isShutdown(false)) {
      if (fLock) unlockInternal();
      return QUEUE_SHUTDOWN;
    }
  }

  // collective EOW/EOS handling
  if (wWorkNum < wDequeueWorkNum) {  // need an EOW?
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
    cout << sbFilterName << "generate fake EOW work=" << wWorkNum
	 << " (wDequeueWorkNum=" << wDequeueWorkNum << ')'
	 << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
    if (fLock) unlockInternal();
    *ppbuf = new DC_Buffer_t;
    (*ppbuf)->eType = DC_Buffer_t::EOW;
    return QUEUE_OK;
  } else if (fConsumerEOS) {  // need an EOS?
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
    cout << sbFilterName << "generate fake EOS work=" << wWorkNum
	 << " (wDequeueWorkNum=" << wDequeueWorkNum << ')'
	 << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
    if (fLock) unlockInternal();
    *ppbuf = new DC_Buffer_t;
    (*ppbuf)->eType = DC_Buffer_t::EOS;
    return QUEUE_OK;
  }

  // this thread will do a real dequeue() ...

#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
  cout << sbFilterName << "before Q.dequeue()" << endl;
  //_DumpState(cout);
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
  int wRtn;
  fFirstDequeue = true;  // set here in case we block/lose mutex in dequeue  
  wRtn = dequeue(&item, -1, false);  /*** RELEASE MUTEX ***/
  *ppbuf = item.pbuf;
  fFirstDequeue = false;  // release other consumers
#if defined(ENABLE_STREAMBUFFERQUEUE_STATS)
  mQueueSizeDequeue.NewValue(probe(false));
#endif /* ENABLE_STREAMBUFFERQUEUE_STATS */

  if (nTotalConsumers > 1) cvFirstDequeue.broadcast(); // wake others
  // this must be a broadcast: if signal woke 1 of k cvFirstDequeue sleepers,
  // and that one returns above with a fake EOW/EOS, the other k-1 are left
  // sleeping forever.  The alternative soln is to signal on all exit paths
  // above (collective EOW/EOS).

  if (wRtn == QUEUE_SHUTDOWN) {
    if (fLock) unlockInternal();
    return QUEUE_SHUTDOWN;
  }

  if (fRemoteDisabledFull) {
    int nSpace = probeSpace(false);
    assert(nSpace > 0);  // since we just dequeued
    
    // the queue was full, and now it is not.  The local producers
    // don't need any special handling, because they are blocked on
    // the queue-internal cv not-full.  But the remote sockets need
    // to be re-enabled.  We send a readercmd to enable them all.

    // Note: since fRemoteDisabledFull, we know nRemoteProducers>0
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
    cout << sbFilterName << "Q now not-full, signal producers. ****"
	 << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
    fRemoteDisabledFull = false;  // assume we will enable them all
    // look thru list of disabled iConns, and enable them as needed
    while (listRemoteDisabledFull.getNum() > 0) {
      RemoteDisabled rd = listRemoteDisabledFull.First();
      if (--nSpace >= 0) {  // enable this connection?
	// send ReaderCmd so this is done in the context of the reader thread
	pSL->ReaderCmd_EnableConn(rd.iConn);
	listRemoteDisabledFull.DelHead();
      } else {
	fRemoteDisabledFull = true;  // we don't have space for more
	break;
      }
    }
  }

  // process the dequeued buffer as needed
  if ((*ppbuf)->eType == DC_Buffer_t::EOW) {
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
    cout << sbFilterName << "dequeue real EOW for work=" << wWorkNum
	 << ", from iConn=" << item.iConn << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
    _IncrementDequeueWorkNum();
    _BroadcastNotEmpty();  // base class
  } else if ((*ppbuf)->eType == DC_Buffer_t::EOS) {
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
    cout << sbFilterName << "dequeue real EOS for work=" << wWorkNum
	 << ", from iConn=" << item.iConn << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
    fConsumerEOS = true;
    _BroadcastNotEmpty();  // base class
  } else {
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
    cout << sbFilterName << "dequeue DATA for work=" << wWorkNum
	 << ", from iConn=" << item.iConn
#if 0
	 << ' ' << **ppbuf
#endif
	 << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */

    // send DATA_ACK?
    if (item.eStreamWritePolicy == DC_StreamWritePolicy_DD) {
      if (item.psock && item.psock->getFd() != -1) {
	// no sync needed among local copies (the copy set of consumers
	// on this node), since we already hold a copy set level lock in
	// this SinkQ.
	Packet pkt(TAG_FILTERSERVICE_DATA_ACK, NULL);
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE_DATAACK)
	cout << "SinkQ.localDequeue - send DATA_ACK on iConn="
	     << item.iConn << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE_DATAACK */
	pkt.write(*item.psock);
      } else if (item.plcUnAckedBuffers) {
	// ack local writer
	item.plcUnAckedBuffers->decSignalHigh(item.iConn);
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE_DATAACK)
	cout << "SinkQ.localDequeue - direct DATA_ACK for iConn="
	     << item.iConn << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE_DATAACK */
      }
    }
  }

  if (fLock) unlockInternal();
  return wRtn;
}

/**
 * return: <0=shutdown, 0=data to read, 1=no data to read
 */
int DC_StreamBufferQueue::localProbe(int wWorkNum, bool fLock) {
  int wRtn;
  if (fLock) lockInternal();
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
  _DumpState(cout);
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
  if (isShutdown(false)) {
    wRtn = QUEUE_SHUTDOWN;
  } else if (wWorkNum < wDequeueWorkNum ||  // fake EOW
	     fConsumerEOS ||  // fake EOS
	     (wWorkNum == wDequeueWorkNum && probe(false) > 0)) {
    wRtn = 0;
  } else {
    wRtn = 1;
  }
  if (fLock) unlockInternal();
  return wRtn;
}

int DC_StreamBufferQueue::localShutdown(bool fLock) {
  // drain and shutdown SinkQ
  DC_StreamBufferQueue_Item item;
  if (fLock) lockInternal();
  while (probe(false) > 0) {
    if (dequeue(&item, -1, false) != QUEUE_OK) break;
#if defined(DCDBG_FILTERSERVICE_QUEUE)
    cout << "DC_ShutdownSinkQ(): drain queue " << *item.pbuf << endl;
#endif /* DCDBG_FILTERSERVICE_QUEUE */
    item.pbuf->consume();
  }
  shutdown(false);
  if (fLock) unlockInternal();
  // wake any waiters
  if (nLocalProducers) cvLocalProducerReady.broadcast();
  if (nTotalConsumers) cvLocalConsumerReady.broadcast();
  if (nTotalConsumers > 1) cvFirstDequeue.broadcast();
  return 0;
}


// ****************************************************************************
