// -*- c++ -*-
// ****************************************************************************
//  Name: streambufferqueue.h
//  Athr: Michael D. Beynon (mdb)
//  Date: 02/22/2001 : mdb - Created.
//        03/02/2001 : mdb - Fixed hang bug in localDequeue().
//        04/22/2001 : mdb - Use a container for items in queue to add iConn.
//  Desc: Part of remote lib process stream abstraction.  Layered above simple
//        queue to support multiple local/remote producers and multiple
//        consumers. (copies)  The invariant of all work k buffers appear
//        before work k+1 delimited by an EOW is preserved.
// ****************************************************************************

#if !defined(H_DC_STREAMBUFFERQUEUE)
#define H_DC_STREAMBUFFERQUEUE

#include "lib/dc_mutex.h"
#include "lib/dc_queue.h"
#include "lib/dc_darray.h"
#include "lib/dc_lists.h"
#include "lib/dc_lockedcounter.h"
#include "lib/dc_metric.h"

#include "dc_buffer.h"
#include "thread-reader.h"
#include "dc_policy.h"

#define ENABLE_STREAMBUFFERQUEUE_STATS

/**
 * Shared queue that supports multiple producers and consumers that
 * are required for multiple transparent copies.  Both remote and
 * local writers are synchronized to preserve the following partial
 * order: all buffers for work k must appear in the queue before work
 * k+1.  The main work here is to have all producers (local and
 * remote) all enqueue an EOW/EOS and there will be one actual EOW/EOS
 * enqueued.  Similarly, local dequeues will turn a single EOW/EOS in
 * the queue into one for each consumer.
 *
 * Local producers will block in localEnqueue() and the reader thread
 * will first check if the sockets for remote producers should be
 * read.  Both methods will pause the producers, while not blocking
 * the reader so progress can be made.
 *
 * this is external to provide visibility in subclassing DC_QueueT<>
 */
class DC_StreamBufferQueue_Item {
public:
  DC_Buffer_t *pbuf;
  DC_StreamWritePolicy_t eStreamWritePolicy;
  int          iConn;  ///< -1 for EOW/EOS buffers, !-1 for DATA buffers
  BaseSocket  *psock;  ///< !NULL == remote sender for Ack
  DC_LockedCounterThreshold *plcUnAckedBuffers;  ///< !NULL == local sender
};

class DC_StreamBufferQueue : public DC_QueueT<DC_StreamBufferQueue_Item> {
protected:
  // enqueue producer side of queue
  int  wEnqueueWorkNum;  ///< current value
  int  nProducerEOW, nProducerEOS;
  int  nLocalProducers, nRemoteProducers;
  int  nTotalProducers;
  // dequeue consumer side of queue
  int  wDequeueWorkNum;
  int  nConsumerEOW, nConsumerEOS;
  bool fConsumerEOS;
  int  nTotalConsumers;
  bool fFirstDequeue;  ///< for handling a blocking dequeue and EOW/EOS
  int  nEarlyDequeue;

  DC_Cond cvLocalProducerReady;  ///< local producers that write too early
  DC_Cond cvLocalConsumerReady;  ///< local consumers that read too early
  DC_Cond cvFirstDequeue;  ///< predicate: fFirstDequeue

  struct RemoteDisabled { int iConn, wWorkNum; };
  bool fRemoteDisabledEarly, fRemoteDisabledFull;
  DArray<RemoteDisabled> daRemoteDisabledEarly;
  SLList<RemoteDisabled> listRemoteDisabledFull;
  DC_ReaderServerLoop *pSL;

public:
#if defined(ENABLE_STREAMBUFFERQUEUE_STATS)
  Metric mQueueSizeDequeue;
#endif /* ENABLE_STREAMBUFFERQUEUE_STATS */
 
protected:
  int  _IncrementEnqueueWorkNum(bool fReaderThread);
  int  _IncrementDequeueWorkNum(void);
  void _DumpState(ostream &os);

public:
  DC_StreamBufferQueue();
  ~DC_StreamBufferQueue();

  /// set initial worknum to expect (should be called before anything else)
  int  setInitialWorkNum(int wWorkNum);
#if 0 /* MDB */
  void setStreamWritePolicy(DC_StreamWritePolicy_t eStreamWritePolicy_in);
#endif

  int  localAddProducer(int n=1);  ///< colocated filters
  int  localRemoveProducer(void);
  int  localEnqueue(int wWorkNum, DC_Buffer_t *pbuf,
		    DC_StreamWritePolicy_t eStreamWritePolicy,
		    int iConn, DC_LockedCounterThreshold *plcUnAckedBuffers);

  int  remoteAddProducer(int n=1); ///< remote filters via reader thread
  int  remoteRemoveProducer(void);
  bool remoteCheckWorkNum(int wWorkNum, DC_ReaderServerLoop *pSL, int iConn);
  int  remoteDisableFull(DC_ReaderServerLoop *pSL, int iConn);
  int  remoteEnqueue(int wWorkNum, DC_Buffer_t *pbuf,
		     DC_StreamWritePolicy_t eStreamWritePolicy,
		     int iConn, BaseSocket *psockAck);

  int  localAddConsumer(int n=1);
  int  localDequeue(int wWorkNum, DC_Buffer_t **ppbuf, bool fLock=true);
  int  localProbe(int wWorkNum, bool fLock=true);
  int  localShutdown(bool fLock=true);  ///< drain and shutdown Q
};


#endif /* H_DC_STREAMBUFFERQUEUE */
// ****************************************************************************
