// ****************************************************************************
//  Name: thread-filter.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 04/08/2000 : mdb - Split into a seperate source files.
//        11/29/2000 : mdb - Converted to use DC_Mutex/DC_Cond.
//        03/04/2001 : mdb - Removed insSelect() because it would not work
//                           with a shared SinkQ or UOW boundaries in a nice
//                           way.  Added insReadAny() to replace its use.
//        03/13/2001 : mdb - Sweeping changes to support multiple copies.
//        03/20/2001 : mdb - Fixed unlocked concurrent access to sockConsole.
//        04/12/2001 : mdb - Changed write() and write_nocopy() to avoid an
//                           extra copy for a common case.
//        06/10/2001 : mdb - Fixed small time collection bug where computation
//                           time for a filter was negative.  Added a flag to
//                           filter state to avoid collecting time for system
//                           calls to read any (CloseOutputWaitInput).
//        08/08/2001 : mdb - Changed filterstats output to include filter name
//                           in stream stats sections.
//        01/04/2002 : mdb - Include instance num in all stats output.
//        02/15/2002 : mdb - Added DrainInput() for filters that return with
//                           unread buffers for wWorkNum.
//  Desc: Filter execution thread and supporting functions.
// ****************************************************************************

#include "dc_filterservice.h"
#include "remoteprocessstate.h"
#include "packet.h"
#include "packet-tags.h"
#include "util.h"

#include "thread-filter.h"
#include "thread-reader.h"

#ifdef _WIN32
#include "win32.h"
#endif

#include <stdarg.h>
#include <errno.h>


#if defined(DCDBG)
//#define DCDBG_FILTERSERVICE_DATA
//#define DCDBG_FILTERSERVICE_DATA_DUMPBUF
//#define DCDBG_FILTERSERVICE_FILTER_THREAD
//#define DCDBG_FILTERSERVICE_FILTEREXIT
//#define DCDBG_FILTERSERVICE_OPEN
//#define DCDBG_FILTERSERVICE_CLOSE
//#define DCDBG_FILTERSERVICE_DRAIN
//#define DCDBG_FILTERSERVICE_INSSELECT
//#define DCDBG_FILTERSERVICE_QUEUE
#endif /* DCDBG */

//
//  filter thread(s)
//

//#define DCDBG_FILTERSERVICE_FILTEREXIT

/**
 * this fcn is used to catch all return points for cleanup in
 * thread_filter_run_aux()
 */
void *DC_RemoteFilterCopy::thread_run(void *parg_in) {
  thread_filter_arg_t *parg = static_cast(thread_filter_arg_t*,parg_in);
  MemoryBuf mb(KB_8);

  // execute filter to completion
  int wExtra = thread_run_filter(parg_in, mb);

#if defined(DCDBG_FILTERSERVICE_FILTEREXIT)
  cout << sbFullName << " sending FILTEREXIT to console." << endl;
#endif /* DCDBG_FILTERSERVICE_FILTEREXIT */
  Packet pkt(TAG_FILTERSERVICE_FILTEREXIT, &mb);
  pkt.setExtra(wExtra);
  parg->pRemote->mutexSockConsole.lock();
  if (pkt.write(parg->pRemote->sockConsole) <= 0) {
    cout << "Error sending FILTEREXIT to console"
	 << " fd=" << parg->pRemote->sockConsole.getFd()
	 << " - " << strerror(errno)
	 << endl;
    // BAD -- the console will never think the instance is done.
  }
  parg->pRemote->mutexSockConsole.unlock();

  // cleanup of the remote instance state?
  if (pShared->pRemoteInstanceShared->CopyExit() == 0) {
#if defined(DCDBG_FILTERSERVICE_FILTEREXIT)
    cout << sbFullName << " sending INSTANCEEXIT to console." << endl;
#endif /* DCDBG_FILTERSERVICE_FILTEREXIT */
    pkt.set(TAG_FILTERSERVICE_INSTANCEEXIT, NULL);
    pkt.setExtra(pShared->wInstanceNum);
    parg->pRemote->mutexSockConsole.lock();
    if (pkt.write(parg->pRemote->sockConsole) <= 0) {
      cout << "Error sending INSTANCEEXIT to console"
	   << " fd=" << parg->pRemote->sockConsole.getFd()
	   << " - " << strerror(errno)
	   << endl;
      // BAD -- the console will never think the instance is done.
    }
    parg->pRemote->mutexSockConsole.unlock();

    // FIX: This allows the reader thread to cleanup the instance
    // state, which also means it can do the right thing in the
    // rgClients array of connections to avoid subtle consistency
    // problems.
    parg->pDC->pReader->ReaderCmd_CleanupInstance(pShared->wInstanceNum);
  }
  delete parg;
  return NULL;
}

/**
 * called either by a new filter thread context, or by the console
 * thread for a pseudo <console> filter.  This allows for common
 * inits for either case.
 */
// assume: pShared, wCopyNum are valid
void DC_RemoteFilterCopy::_init(void) {
  int lenTmp;
  lenTmp = (int)(strlen(pShared->sbName)+20);
  sbFullName = new char[lenTmp];

      // UVC : no need to thread id and lets make names globally unique by adding global rank
#if 0
  snprintf(sbFullName, lenTmp, "[t%d,i%d,%s.c%d]", (int)pthread_self(),
	   pShared->wInstanceNum, pShared->sbName, wCopyNum);
  lenTmp = (int)(strlen(pShared->sbName)+7);
  sbCopyName = new char[lenTmp];
  snprintf(sbCopyName, lenTmp, "%s.c%d", pShared->sbName, wCopyNum);
#else
  snprintf(sbFullName, lenTmp, "[i%d,%s.c%d]", pShared->wInstanceNum,
           pShared->sbName, pShared->rankCopiesStart + wCopyNum);
  lenTmp = strlen(pShared->sbName)+7;
  sbCopyName = new char[lenTmp];
  snprintf(sbCopyName, lenTmp, "%s.c%d", pShared->sbName, pShared->rankCopiesStart + wCopyNum);
#endif
  return;
}

/**
 * Return: DC_ERR_OK=success, else=failure
 */
int DC_RemoteFilterCopy::thread_run_filter(void *parg_in,
					   MemoryBuf &mb) {
  thread_filter_arg_t *parg = static_cast(thread_filter_arg_t*,parg_in);
  int k;

  _init();


  tWallClock.Start();

#if defined(DCDBG_FILTERSERVICE_FILTER_THREAD)
  cout << sbFullName << " entry parg=<";
  cout << "pDC=" << hex << (long)(parg->pDC) << dec << ',';
  cout << "pRemote=" << hex << (long)(parg->pRemote) << dec << ',';
  cout << "argc=" << (parg->pargc ? *(parg->pargc) : -1) << ">" << endl;
#endif /* DCDBG_FILTERSERVICE_FILTER_THREAD */

  DC_Filter_Base_t::arg_t arg;
  arg._init(this, parg->pDC);

  mb.sprintf("[filter.%s]\n", pShared->sbName);
  mb.sprintf("instance=%d\n", pShared->wInstanceNum);
  mb.sprintf("filter=%s\n", pShared->sbName);
  mb.sprintf("rank_copyset=%d\n", wCopyNum);
  mb.sprintf("max_copyset=%d\n", pShared->nInitialCopies);
  mb.sprintf("placement=%s\n", parg->pRemote->sbHostMe);
  // we include these because layouts can be dynamically generated
  mb.sprintf("ins=");
  for (k=0; k<arg.nins; ++k) {
    mb.sprintf("%s ", arg.ins[k].pInStream->pShared->sbName);
  }
  mb.sprintf("\n");
  mb.sprintf("outs=");
  for (k=0; k<arg.nouts; ++k) {
    mb.sprintf("%s ", arg.outs[k].pOutStream->pShared->sbName);
  }
  mb.sprintf("\n");

  // create a new filter object using provided filter factory and name
#if defined(DCDBG_FILTERSERVICE_FILTER_THREAD)
  cout << sbFullName << " creating new filter object ..." << endl;
#endif /* DCDBG_FILTERSERVICE_FILTER_THREAD */
  DC_Filter_Base_t *pUserFilter
    = (*(parg->pDC->fcnFilterFactory))(pShared->sbName);
  if (pUserFilter == NULL) {
    cout << sbFullName << " failed new filter object creation" << endl;
    // BUGBUG: other cleanup?
    return DC_ERR_RemoteFailObjCreate;
  }

  // fillin part of base class members we know
  pUserFilter->pDC = parg->pDC;
  pUserFilter->sbFilterName = pShared->sbName;

  DC_Filter_Base_t::initarg_t initarg;
  initarg.sbFilterName = pShared->sbName;
  initarg.sbFullName = sbFullName;
  initarg.sbCopyName = sbCopyName;
  initarg.argc = *(parg->pargc);
  initarg.argv = *(parg->pargv);
  initarg.pFilter = this;

  DC_Work_t *pwork = NULL;
  DC_RTN_t eRtn;
  int nWork = 0;
  int wExitStatus = DC_ERR_OK;

  //
  // Work reception loop
  //
  while (1) {
    wExitStatus = DC_ERR_OK;
    ++nWork;
    if (pwork) {  // not the first iteration
      delete pwork;
      ++wWorkNum;
    }

#if defined(DCDBG_FILTERSERVICE_FILTER_THREAD)
    cout << sbFullName << " checking for work ... "
	 << "(" << pShared->WorkQ.probe() << " entries in work queue)"
	 << endl;
#endif /* DCDBG_FILTERSERVICE_FILTER_THREAD */

    tWorkQ.Start();
    int ret = pShared->WorkQ.localDequeue(wWorkNum, &pwork);
    tWorkQ.Stop();

    if (ret) {
      cerr << sbFullName << " failed dequeue from work queue" << endl;
      delete pUserFilter;
      // BUGBUG: other cleanup?
      return DC_ERR_RemoteFilterFail;
    } else if (! pwork) { // console process forced EndOfFilter
      // DESIGN: the application does not explicitly know that
      // process() will never be called again.  The semantics of
      // returning DC_RTN_EndOfWork are you *may* be called again.
      // Any final cleanup can be done in the C++ destructor.
#if defined(DCDBG_FILTERSERVICE_FILTER_THREAD)
      cout << sbFullName << " found NULL work -> EndOfFilter" << endl;
#endif /* DCDBG_FILTERSERVICE_FILTER_THREAD */
      break;
    } else if (wWorkNum != pwork->wWorkNum) {
      cerr << sbFullName << " expected work " << wWorkNum
	   << " but found work " << pwork->wWorkNum << endl;
      delete pUserFilter;
      // BUGBUG: other cleanup?
      return DC_ERR_RemoteFilterFail;
      // alternatively we could try to gracefully continue.
      // if wWorkNum > pwork->wWorkNum then
      //   failure
      // else if wWorkNum < pwork->wWorkNum then
      //   emit enough EOW packets to make up difference
    }

    for (unsigned int i = 0; i < pShared->cancellations.NumElem(); ++i) {
      if (pShared->cancellations[i] == wWorkNum) {
	cout << "Cancelling workID " << wWorkNum << endl;
	parg->pRemote->mutexFilterCopy.lock();
	pShared->cancellations.Remove(i);
	parg->pRemote->mutexFilterCopy.unlock();
	wExitStatus = DC_ERR_WorkFail;
	break;
      }
    }
    if (wExitStatus == DC_ERR_WorkFail) {
      MemoryBuf mb(sizeof(int));
      mb.Append(wExitStatus);
      Packet pkt(TAG_FILTERSERVICE_APPENDWORK_DONE, &mb);
      pkt.setExtra(DC_EncodeInt(pShared->wInstanceNum, wWorkNum));
      parg->pRemote->mutexSockConsole.lock(); //UVC
      if (pkt.write(parg->pRemote->sockConsole) <= 0) {
	cout << "Error sending APPENDWORK_DONE to console." << endl;
      }
      parg->pRemote->mutexSockConsole.unlock(); //UVC
      arg.pwork = pwork;
      arg.CloseOutputWaitInput(wWorkNum, true);
      arg.DrainInput(wWorkNum);
      arg._resetInsEOW();
      continue;
    }
    
#if defined(DCDBG_FILTERSERVICE_FILTER_THREAD)
    cout << sbFullName << " calling DC_Filter_Base_t::init() for work="
	 << wWorkNum << " ..." << endl;
#endif /* DCDBG_FILTERSERVICE_FILTER_THREAD */
    initarg.pwork = pwork;
    tInit.Start();
    eRtn = pUserFilter->init(initarg);  // BUGBUG: ignore the init() return value?
    tInit.Stop();
#if defined(DCDBG_FILTERSERVICE_FILTER_THREAD)
    cout << sbFullName << " done DC_Filter_Base_t::init() for work="
	 << wWorkNum << '.' << endl;
#endif /* DCDBG_FILTERSERVICE_FILTER_THREAD */
    
    // TODO: negotiate buffer sizes here - problem is this will
    // interfere with the async STREAMCONNECT process, which only
    // happens the first init,process,finalize cycle.  Later
    // iterations just reuse the stream.

    if (eRtn == DC_RTN_OK) { // CLH - Only call process if init succeeded      
#if defined(DCDBG_FILTERSERVICE_FILTER_THREAD)
      cout << sbFullName << " calling DC_Filter_Base_t::process() for work="
	   << wWorkNum << " ..." << endl;
#endif /* DCDBG_FILTERSERVICE_FILTER_THREAD */
      arg.pwork = pwork;
      
      tProcess.Start();
      fInProcessCall = true;
      eRtn = pUserFilter->process(arg);  // return: EndOfWork or EndOfFilter
      fInProcessCall = false;
      tProcess.Stop();

#if defined(DCDBG_FILTERSERVICE_FILTER_THREAD)
      cout << sbFullName << " done DC_Filter_Base_t::process() for work="
	   << wWorkNum << ", ret="
	   << endl;
      //<< (eRtn == DC_RTN_EndOfFilter ? "EndOfFilter" : "EndOfWork")
#endif /* DCDBG_FILTERSERVICE_FILTER_THREAD */
    }

    // CLH - handling for new filter return codes
    switch (eRtn) {
    case DC_RTN_OK:
      wExitStatus = DC_ERR_OK;
      break;
    case DC_RTN_CONT:
      wExitStatus = DC_ERR_WorkFail;
      break;
      //case DC_RTN_TERM:
      //wExitStatus = DC_ERR_RemoteFilterFail;
      //break;
    case DC_RTN_ABRT:
      wExitStatus = DC_ERR_InstanceFail;
      break;
    }
    
    // send EndOfWork packet to console?
    MemoryBuf mb(sizeof(int));
    mb.Append(wExitStatus);
    Packet pkt(TAG_FILTERSERVICE_APPENDWORK_DONE, &mb);
    pkt.setExtra(DC_EncodeInt(pShared->wInstanceNum, wWorkNum));
    // console only cares about the count of filters done
    parg->pRemote->mutexSockConsole.lock(); // UVC
    if (pkt.write(parg->pRemote->sockConsole) <= 0) {
      cout << "Error sending APPENDWORK_DONE to console." << endl;
      // BAD -- the console will never think the work is done.
    }
    parg->pRemote->mutexSockConsole.unlock(); // UVC
    if (eRtn == DC_RTN_ABRT) {
      break;
    }
    
#if defined(DCDBG_FILTERSERVICE_FILTER_THREAD)
    cout << sbFullName << " sent APPENDWORK_DONE" << endl;
#endif /* DCDBG_FILTERSERVICE_FILTER_THREAD */
    arg.CloseOutputWaitInput(wWorkNum, true);
#if defined(DCDBG_FILTERSERVICE_FILTER_THREAD)
    cout << sbFullName << " calling DC_Filter_Base_t::finalize() for work="
	 << wWorkNum << " ..." << endl;
#endif /* DCDBG_FILTERSERVICE_FILTER_THREAD */

    tFinalize.Start();
    // CLH - Right now, the return value of finalize is not considered.
    eRtn = pUserFilter->finalize();
    tFinalize.Stop();

#if defined(DCDBG_FILTERSERVICE_FILTER_THREAD)
    cout << sbFullName << " done DC_Filter_Base_t::finalize() for work="
	 << wWorkNum << '.' << endl;
#endif /* DCDBG_FILTERSERVICE_FILTER_THREAD */

    // insure all input buffers were read for this UOW, and reset the
    // flags to allow application read() calls
    arg.DrainInput(wWorkNum);
    arg._resetInsEOW();
  }
  // BUGBUG: For applications that do not fit nicely into this
  // model, they need to close the streams in whatever order makes
  // sense for them.  If all input streams are closed by the
  // application before reaching this point, the call will not
  // block.  (this may not apply to the EndOfWork case)
  arg.CloseOutputWaitInput(wWorkNum, false);

  tWallClock.Stop();

  double rReadTime = 0.0;
  double rWriteTime = 0.0;
  for (k=0; k<arg.nins; ++k) {
    rReadTime += arg.ins[k].pInStream->tRead.Elapsed();
  }
  for (k=0; k<arg.nouts; ++k) {
    rWriteTime += arg.outs[k].pOutStream->tWrite.Elapsed();
  }

  // add stats about this filter execution
  mb.sprintf("num_work=%d\n", nWork);
  mb.sprintf("wall_clock_time=%f\n", tWallClock.Elapsed());
  mb.sprintf("workq_time=%f\n", tWorkQ.Elapsed());
  mb.sprintf("init_time=%f\n", tInit.Elapsed());
  mb.sprintf("process_time=%f\n", tProcess.Elapsed());
  mb.sprintf("process.read_time=%f\n", rReadTime);
  mb.sprintf("process.write_time=%f\n", rWriteTime);
  mb.sprintf("process.readany_time=%f\n", tReadAny.Elapsed());
  mb.sprintf("process.readfirst_time=%f\n", tReadFirst.Elapsed());
  mb.sprintf("process.computation_time=%f\n", (tProcess.Elapsed()
					       - rReadTime - rWriteTime
					       - tReadAny.Elapsed()
					       - tReadFirst.Elapsed()));
  mb.sprintf("process.instant_writefirst=%f\n", rInstantWriteFirst);
  mb.sprintf("finalize_time=%f\n", tFinalize.Elapsed());

  // wall clock time (lifetime of filter)
  // WorkQ time
  // init() time
  // process() time
  //   first buffer read delay
  //   ins[*] read stream time
  //   outs[*] write stream time
  //   insReadAny blocked time
  //   computation time (process() - first - read - write - insReadAny)
  // finalize() time
  // total computation time (init() + process().computation + finalize())
  // total stream i/o time (process().{read,write})
  // total block time (WorkQ + process().{first,read,write}

  for (k=0; k<arg.nins; ++k) {
    mb.sprintf("[filter.%s.ins.%s]\n",
	       pShared->sbName, arg.ins[k].pInStream->pShared->sbName);
    mb.sprintf("instance=%d\n", pShared->wInstanceNum);
    mb.sprintf("number=%ld\n", arg.ins[k].pInStream->mReadBytes.getNum());
    mb.sprintf("volume=%ld\n",
	       static_cast(long,arg.ins[k].pInStream->mReadBytes.getSum()));
    mb.sprintf("read_time=%f\n", arg.ins[k].pInStream->tRead.Elapsed());
#if defined(ENABLE_STREAMBUFFERQUEUE_STATS)
    mb.sprintf("read_queuesize=%ld %f %f %f\n",
	arg.ins[k].pInStream->pShared->SinkQ.mQueueSizeDequeue.getNum(),
	arg.ins[k].pInStream->pShared->SinkQ.mQueueSizeDequeue.getMean(),
	arg.ins[k].pInStream->pShared->SinkQ.mQueueSizeDequeue.getMin(),
	arg.ins[k].pInStream->pShared->SinkQ.mQueueSizeDequeue.getMax());
#endif /* ENABLE_STREAMBUFFERQUEUE_STATS */
  }
  for (k=0; k<arg.nouts; ++k) {
    mb.sprintf("[filter.%s.outs.%s]\n",
	       pShared->sbName, arg.outs[k].pOutStream->pShared->sbName);
    mb.sprintf("instance=%d\n", pShared->wInstanceNum);
    mb.sprintf("number=%ld\n", arg.outs[k].pOutStream->mWriteBytes.getNum());
    mb.sprintf("volume=%ld\n",
	       static_cast(long,arg.outs[k].pOutStream->mWriteBytes.getSum()));
    mb.sprintf("write_time=%f\n", arg.outs[k].pOutStream->tWrite.Elapsed());
  }

  // cleanup
  delete pUserFilter;

  sendPerformanceUpdate(true);
  
#if defined(DCDBG_FILTERSERVICE_FILTER_THREAD)
  cout << sbFullName << " exit" << endl;
#endif /* DCDBG_FILTERSERVICE_FILTER_THREAD */
  return DC_ERR_OK;
}


void   DC_RemoteFilterCopy::initPerformanceUpdate()
{
dLastPerformanceSendTime = w_seconds();
// UVC: do we need any initialization!? 
}

void   DC_RemoteFilterCopy::sendPerformanceUpdate(bool forcesend)
{
double ct = w_seconds();
bool send=forcesend || (ct-dLastPerformanceSendTime>2.5); // don't update less than 1 sec interval

if (send) 
    {
    int k, inst = pShared->wInstanceNum, rank=pShared->rankCopiesStart + wCopyNum;
    
        // UVC: sbFullname contains inst,filtname.c<rank>

    double sa[9];
    long   sal[4];
        
    sa[0] = tWallClock.Snapshot();
    sa[1] = tWorkQ.Snapshot();
    sa[2] = tInit.Snapshot();

    double tprocess = tProcess.Snapshot();

        //read times
    sa[3] = tReadFirst.Snapshot();
    sa[4] = 0.0;
    for (k=0; k<daInStreamCopy.NumElem(); ++k) 
        sa[4] += daInStreamCopy[k]->tRead.Snapshot();
    sa[5] = tReadAny.Snapshot();

        // write time
    sa[7]= 0.0;
    for (k=0; k<daOutStreamCopy.NumElem(); ++k) 
        sa[7] += daOutStreamCopy[k]->tWrite.Snapshot();

        // compute = tporcess - readfirst - read - readany - write
    sa[6] = tprocess - sa[3] - sa[4] - sa[5] - sa[7];
    sa[8] = tFinalize.Snapshot();


    
    
    for (k=0; k<4; ++k)
        sal[k] = 0;
    
        // UVC actually send the data :)

    for (k=0; k<daInStreamCopy.NumElem(); ++k)
        {
            // char *name=daInStreamCopy[k]->pShared->sbName;
        long number=daInStreamCopy[k]->mReadBytes.getNum();
        long sum=static_cast(long, daInStreamCopy[k]->mReadBytes.getSum());
        sal[0] += number;
        sal[1] += sum;
        }
    for (k=0; k<daOutStreamCopy.NumElem(); ++k)
        {
            // char *name=daOutStreamCopy[k]->pShared->sbName;
        long number=daOutStreamCopy[k]->mWriteBytes.getNum();
        long sum=static_cast(long, daOutStreamCopy[k]->mWriteBytes.getSum());
        sal[2] += number;
        sal[3] += sum;
        }


    if (NULL != g_fcnPerfCallback) {
        (*g_fcnPerfCallback)(inst, pShared->sbName, rank, sa, sal);
    }
    
    // don't forget to set last time sent...
    dLastPerformanceSendTime = ct; 
    }
}

//
//  filter thread(s) - stream i/o
//
char *DC_PipeInStream_t::getName(void) {
  return pInStream->pShared->sbName;
}
DC_Buffer_t *DC_PipeInStream_t::read_u(void) {
  DC_Buffer_t *pbuf;
  // user can call isEndOfWork() or isEndOfStream() to distinguish state here
  if (fEOW) {
    if (pFilter->pShared->fConsole) {
      //cout << "read_u(): reset EndOfWork marker for <console> call." << endl;
      fEOW = false;  // hack for pseudo-filter similar to _resetInsEOW()
      // BUGBUG: since we don't have a controlling thread to advance
      // our worknum, we advance it here.  There is a problem if the <console>
      // caller does not read all of each stream before moving onto the next.
      // Unclear how to best deal with this case.
      ++(pFilter->wWorkNum);
    } else {
      cerr << "read_u(): attempted read after EndOfWork marker." << endl;
      return NULL;
    }
  } else if (fEOS) {
    cerr << "read_u(): attempted read after EndOfStream marker." << endl;
    return NULL;
  }

#if defined(DCDBG_FILTERSERVICE_DATA)
  cout << "read(): about to dequeue ..." << endl;
#endif /* DCDBG_FILTERSERVICE_DATA */
  // here the filter thread will block on an empty queue
  if (pInStream->pShared->SinkQ.localDequeue(pFilter->wWorkNum,
					     &pbuf, false)) {
    // (this case should never happen due to how the Reader uses the queue)
    cout << "read_u(): failed localDequeue(): setting NULL" << endl;
    fEOS = true;
    pbuf = NULL;
  } else {
#if defined(DCDBG_FILTERSERVICE_DATA)
    cout << "Read "
	 << (pbuf->eType == DC_Buffer_t::EOW ? "EOW"
	     : (pbuf->eType == DC_Buffer_t::EOS ? "EOS" : "DATA"))
	 << " (" << pbuf->getSize() << ")"
	 << " from " << pInStream->pShared->sbFilterSource
	 << " to " << pFilter->pShared->sbName << " over stream "
	 << pInStream->pShared->sbName
	 << endl;
#if defined(DCDBG_FILTERSERVICE_DATA_DUMPBUF)
    pbuf->Dump();
#endif /* DCDBG_FILTERSERVICE_DATA_DUMPBUF */
#endif /* DCDBG_FILTERSERVICE_DATA */
    // control buffer?
    if (pbuf->eType == DC_Buffer_t::DATA) {
      pInStream->mReadBytes.NewValue(pbuf->getSize());  // stats
    } else {
      // mark stream EOS/EOW when we actually read an EOS/EOW buf
      if (pbuf->eType == DC_Buffer_t::EOS) {
	pInStream->pShared->SinkQ.localShutdown(false);
	fEOS = true;
      } else if (pbuf->eType == DC_Buffer_t::EOW) {
	fEOW = true;
      }
      pbuf->consume();
      pbuf = NULL;
    }
  }

  return pbuf;
}
DC_Buffer_t *DC_PipeInStream_t::read(void) {
  DC_Buffer_t *pbuf;

  if (pFilter->fInProcessCall) {
    if (pInStream->pFilter->fReadFirst) {  // stats
       pInStream->pFilter->tReadFirst.Start();   // UVC: isn't pInStream->pFilter == pFilter ?
    } else {
      pInStream->tRead.Start();
    }
  }

  pInStream->pShared->SinkQ.lockInternal();
  pbuf = read_u();
  pInStream->pShared->SinkQ.unlockInternal();  

  if (pFilter->fInProcessCall) {
    if (pInStream->pFilter->fReadFirst) {  // stats
      pInStream->pFilter->tReadFirst.Stop();   
      pInStream->pFilter->fReadFirst = false;
    } else {
      pInStream->tRead.Stop();
    }
    pFilter->sendPerformanceUpdate();
  }
  return pbuf;
}
int DC_PipeInStream_t::close(void) {
  // THREADS: we don't need to lock this, because there is only one filter
  // thread that can call this fcn, either in the process callback or after
  // the callback ends, when all out streams are closed.

  if (fEOS) {
#if defined(DCDBG_FILTERSERVICE_CLOSE)
    // stream already closed
    cout << "DC_PipeInStream_t::close(): stream "
	 << pInStream->pShared->sbName << " is already closed" << endl;
#endif /* DCDBG_FILTERSERVICE_CLOSE */
    return DC_ERR_StreamEndOfStream;
  }
#if defined(DCDBG_FILTERSERVICE_CLOSE)
  cout << "DC_PipeInStream_t::close(): closing stream "
       << pInStream->pShared->sbName << endl;
#endif /* DCDBG_FILTERSERVICE_CLOSE */

  // COPIES_TODO: we cannot shutdown the shared queue for all copies.
  // Instead we should just remove ourselves as a consumer.  That
  // routine can drain and shutdown the Q is need be.

  // drain and shutdown queue to close in the backward direction
  pInStream->pShared->SinkQ.localShutdown();

  // set flag and reader will drop buffers received for this stream.
  // do this even though read() for the colocated case should do it, because
  // if the input stream is closed, all packets are dropped, and the colocated
  // filter will never notice the EOS if they use insSelect().
  pInStream->fEOS = true;

  fEOS = true;
  return 0;
}


char *DC_PipeOutStream_t::getName(void) {
  return pOutStream->pShared->sbName;
}

/**
 * Write the buffer, and handoff ownership to the next filter.  The
 * filter calling this fcn should not try and use the pbuf ptr when
 * DC_ERR_OK is returned.  If instead the call fails, the pbuf can be
 * read, and the caller must perform any cleanup/recovery.
 */
int DC_PipeOutStream_t::write_nocopy(DC_Buffer_t *pbuf, void *pUserArg) {
  if (!pbuf) return DC_ERR_InvalidBuffer;
  if (fEOW) {
#if defined(DCDBG_FILTERSERVICE_DATA)
    cerr << "write_nocopy(): stream is EndOfWork when writing"
	 << " from " << pFilter->pShared->sbName
	 << " to " << pOutStream->pShared->sbFilterSink
	 << " over stream " << pOutStream->pShared->sbName
	 << endl;
#endif /* DCDBG_FILTERSERVICE_DATA */
    // do we consume this buffer?  no, to allow app to recover
    return DC_ERR_StreamEndOfWork;
  }
  if (fEOS) {
#if defined(DCDBG_FILTERSERVICE_DATA)
    cerr << "write_nocopy(): stream is EndOfStream when writing"
	 << " from " << pFilter->pShared->sbName
	 << " to " << pOutStream->pShared->sbFilterSink
	 << " over stream " << pOutStream->pShared->sbName
	 << endl;
#endif /* DCDBG_FILTERSERVICE_DATA */
    // do we consume this buffer?  no, to allow app to recover
    return DC_ERR_StreamEndOfStream;
  }

  pOutStream->tWrite.Start();  // stats

  int wRtn = pOutStream->WriteBuffer_DATA(pbuf, pFilter, true, pUserArg);
  if (wRtn) close();  // BUGBUG: this may not make sense with multiple copies

  pOutStream->tWrite.Stop();  // stats
  pFilter->sendPerformanceUpdate();

  return wRtn;
}

/**
 * Write the buffer, making a duplicate as needed to guarantee that
 * when this call finishes, the caller can modify pbuf w/o danger of
 * changing what the next filter will see.
 */
int DC_PipeOutStream_t::write(DC_Buffer_t *pbuf, void *pUserArg) {
  if (!pbuf) return DC_ERR_InvalidBuffer;
  if (fEOW) {
#if defined(DCDBG_FILTERSERVICE_DATA)
    cerr << "write(): stream is EndOfWork when writing"
	 << " from " << pFilter->pShared->sbName
	 << " to " << pOutStream->pShared->sbFilterSink
	 << " over stream " << pOutStream->pShared->sbName << endl;
#endif /* DCDBG_FILTERSERVICE_DATA */
    return DC_ERR_StreamEndOfWork;
  }
  if (fEOS) {
#if defined(DCDBG_FILTERSERVICE_DATA)
    cerr << "write(): stream is EndOfStream when writing"
	 << " from " << pFilter->pShared->sbName
	 << " to " << pOutStream->pShared->sbFilterSink
	 << " over stream " << pOutStream->pShared->sbName << endl;
#endif /* DCDBG_FILTERSERVICE_DATA */
    return DC_ERR_StreamEndOfStream;
  }

  pOutStream->tWrite.Start();  // stats

  int wRtn = pOutStream->WriteBuffer_DATA(pbuf, pFilter, false, pUserArg);
  if (wRtn) close();  // BUGBUG: this may not make sense with multiple copies

  pOutStream->tWrite.Stop();  // stats
  pFilter->sendPerformanceUpdate();
  
  return wRtn;
}
int DC_PipeOutStream_t::close(bool fEndOfWork) {
  // THREADS: we don't need to lock this, because there is only one filter
  // thread that can call this fcn, either in the process callback or after
  // the callback ends, when all out streams are closed.

  if (fEOS) {
#if defined(DCDBG_FILTERSERVICE_CLOSE)
    // stream already closed
    cout << "DC_PipeOutStream_t::close(): "
	 << (fEndOfWork ? "EndOfWork" : "EndOfFilter")
	 << " - stream " << pOutStream->pShared->sbName
	 << " is already closed" << endl;
#endif /* DCDBG_FILTERSERVICE_CLOSE */
    return DC_ERR_StreamEndOfStream;
  }
#if defined(DCDBG_FILTERSERVICE_CLOSE)
  cout << "DC_PipeOutStream_t::close(): "
       << (fEndOfWork ? "EndOfWork - marking" : "EndOfFilter - closing")
       << " stream " << pOutStream->pShared->sbName << endl;
#endif /* DCDBG_FILTERSERVICE_CLOSE */

  DC_Buffer_t *pbuf = new DC_Buffer_t;
  pbuf->eType = (fEndOfWork ? DC_Buffer_t::EOW : DC_Buffer_t::EOS);
  pOutStream->WriteBuffer_EOWEOS(pbuf, pFilter);  // send and consume

  if (!fEndOfWork) fEOS = true;
  return 0;
}

/**
 * convenience wrappers to DC. routines
 * searches console config in the section [AppName.FilterName]
 */
char *DC_Filter_Base_t::FindAppFilterConfig(const char *sbName, int wId) {
  // must be freed by the caller
  return pDC->FindAppConfig(sbName, sbFilterName, wId);
}
long DC_Filter_Base_t::FindAppFilterConfigInt(const char *sbName, int wId,
					      bool fInterpretUnits) {
  return pDC->FindAppConfigInt(sbName, sbFilterName, wId,
			       fInterpretUnits);
}
double DC_Filter_Base_t::FindAppFilterConfigDbl(const char *sbName, int wId) {
  return pDC->FindAppConfigDbl(sbName, sbFilterName, wId);
}

// TODO: We will only provide one type of lock for now - across all
// instances (located in DC_RemoteProcessState).  It is possible to do
// instance level locking, or even filter-copy level locking, and will
// be added if needed.
void DC_Filter_Base_t::FilterLock(void) {
  pDC->FilterLock();
}
void DC_Filter_Base_t::FilterUnlock(void) {
  pDC->FilterUnlock();
}


//
//  filter thread(s) - process() arg interface
//  *and* the pseudo-filter console process thread
//

/**
 * our single copyset on the current local host
 */
int DC_Filter_Base_t::initarg_t::rankWithinCopySet(void) {
  return pFilter->wCopyNum;
}
int DC_Filter_Base_t::initarg_t::maxWithinCopySet(void) {
  return pFilter->pShared->nInitialCopies;
}
/**
 * total copyset info for this filter group instance
 */
int DC_Filter_Base_t::initarg_t::rankCopySets(void) {
  return pFilter->pShared->rankCopySets;
}
int DC_Filter_Base_t::initarg_t::maxCopySets(void) {
  return pFilter->pShared->maxCopySets;
}
/**
 * total filter transparent copy info for this filter group instance
 */
int DC_Filter_Base_t::initarg_t::rankCopies(void) {
  return pFilter->pShared->rankCopiesStart + pFilter->wCopyNum;
}
int DC_Filter_Base_t::initarg_t::maxCopies(void) {
  return pFilter->pShared->maxCopies;
}

/**
 * dtor
 */
DC_Filter_Base_t::initarg_t::~initarg_t(void) {
}

/**
 * ctor
 */
DC_Filter_Base_t::arg_t::arg_t(void) {
  nins = nouts = 0;
  ins = NULL;
  outs = NULL;
}

/**
 * dtor
 */
DC_Filter_Base_t::arg_t::~arg_t(void) {
  if (ins) delete [] ins;
  if (outs) delete [] outs;
}

/**
 * called either by a new filter thread context, or by the console
 * thread for a pseudo <console> filter.
 */
int DC_Filter_Base_t::arg_t::_init(DC_RemoteFilterCopy *pFilter_in,
				   DC_FilterService_t *pDC) {
  int i;

  // TODO: do we need to lock *pFilter ?

  // create stream interface objects as needed
  pFilter = pFilter_in;
  sbFilterName = pFilter->pShared->sbName;
  sbFullName = pFilter->sbFullName;
  sbCopyName = pFilter->sbCopyName;

  nins = pFilter->daInStreamCopy.NumElem();
  ins = new DC_PipeInStream_t[nins];
  for (i=0; i<nins; ++i) {
    ins[i].pFilter = pFilter;
    ins[i].pInStream = pFilter->daInStreamCopy[i];
    ins[i].fEOW = false;
    ins[i].fEOS = false;
    ins[i].pDC = pDC;  // still need this?
  }

  nouts = pFilter->daOutStreamCopy.NumElem();
  outs = new DC_PipeOutStream_t[nouts];
  for (i=0; i<nouts; ++i) {
    outs[i].pFilter = pFilter;
    outs[i].pOutStream = pFilter->daOutStreamCopy[i];
    outs[i].fEOW = false;
    outs[i].fEOS = false;
    outs[i].pDC = pDC;  // still need this?
  }
  pwork = NULL;
  return 0;
}

int DC_Filter_Base_t::arg_t::_resetInsEOW(void) {
  for (int i=0; i<nins; ++i) {
    ins[i].fEOW = false;
  }
  return 0;
}

int DC_Filter_Base_t::arg_t::insIndex(char *sbName) {
  // BUGBUG: we could always implement a hash lookup for speed
  for (int i=0; i<nins; ++i) {
    if (strcmp(ins[i].getName(), sbName) == 0) return i;
  }
  return DC_ERR_InvalidStreamName;
}
int DC_Filter_Base_t::arg_t::outsIndex(char *sbName) {
  // BUGBUG: we could always implement a hash lookup for speed
  for (int i=0; i<nouts; ++i) {
    if (strcmp(outs[i].getName(), sbName) == 0) return i;
  }
  return DC_ERR_InvalidStreamName;
}
DC_PipeInStream_t *DC_Filter_Base_t::arg_t::insLookup(char *sbName) {
  int i = insIndex(sbName);
  return (i != -1 ? ins+i : (DC_PipeInStream_t*)NULL);
}
DC_PipeOutStream_t *DC_Filter_Base_t::arg_t::outsLookup(char *sbName) {
  int i = outsIndex(sbName);
  return (i != -1 ? outs+i : (DC_PipeOutStream_t*)NULL);
}

/**
 * block until a buffer available from any input stream
 *
 * BUGBUG: if the caller specifies a rgStreamSet subset of streams,
 * then there is no easy way for them to distinguish allEOW and allEOS
 * within the subset w/o manually testing them all.
 *
 * return: either a buffer is found to read on any of the input
 * streams and is returned, or all input streams are EOW/EOS and NULL
 * is returned to indicate this.
 *
 * block until a buffer available from any input stream.  If all are
 * EOW or EOS, return NULL.
 *
 * Optional: rgStreamSet is an array of bool (one per ins).  If
 * set false, that ins will be ignored when looking for buffers.
 */
DC_Buffer_t *DC_Filter_Base_t::arg_t::insReadAny(int *piFound,
						 bool *rgStreamSet) {
  DC_Buffer_t *pbuf = NULL;
  int i, n;
  int w, nShutdown, nNotEOW, nConsider;

  if (pFilter->fInProcessCall) {
    if (pFilter->fReadFirst) {  // stats
      pFilter->tReadFirst.Start();
    } else {
      pFilter->tReadAny.Start();
    }
  }

  if (! pFilter->rgReadAnyState) {  // allocate only if needed
    pFilter->rgReadAnyState = new DC_RemoteFilterCopy::insReadAnyState_t[nins];
  }
  // mark which to consider
  for (i=0; i<nins; ++i) {
    pFilter->rgReadAnyState[i].fStreamSet
      = (rgStreamSet ? rgStreamSet[i] : true);
  }

  pFilter->pShared->mutexAllIns.lock();

  while (1) {
    // quick probe of state
    nShutdown = nNotEOW = nConsider = 0;
    for (i=pFilter->iStreamReadAnyState, n=0; n<nins; ++n, ++i) {
      if (i==nins) i=0;

      if (!pFilter->rgReadAnyState[i].fStreamSet) continue;  // not considered
      if (ins[i].fEOW || ins[i].fEOS) continue;
      ++nConsider;
      
      // we ignore the return == 2 case
      w = ins[i].pInStream->pShared->SinkQ.localProbe(pFilter->wWorkNum,
						      false);
#if defined(DCDBG_FILTERSERVICE_INSSELECT)
      cout << "insReadAny(): ins[" << i << "] localProbe rtn=" << w << endl;
#endif /* DCDBG_FILTERSERVICE_INSSELECT */
      if (w == 0) {  // should not block?
	pbuf = ins[i].read_u();
	if (pbuf) {
	  if (piFound) *piFound = i;
	  break;
	} else {
	  continue;  // newly fEOW or fEOS
	}
      } else if (w < 0) {
	++nShutdown;
      } else {
	++nNotEOW;
      }
    }
#if defined(DCDBG_FILTERSERVICE_INSSELECT)
    if (pbuf) {
      cout << "insReadAny(): break (pbuf) - read a buffer" << endl;
    } else if (nNotEOW == 0) {
      cout << "insReadAny(): break (nNotEOW == 0) - all are EOW" << endl;
    } else if (nShutdown == nConsider) {
      cout << "insReadAny(): break (nShutdown == nConsider) - all shutdown" << endl;
    }
#endif /* DCDBG_FILTERSERVICE_INSSELECT */
    // done if we read a buffer, all are EOW, or all are shutdown
    if (pbuf || nNotEOW == 0 || nShutdown == nConsider) break;

#if defined(DCDBG_FILTERSERVICE_INSSELECT)
    cout << "insReadAny(): nNotEOW=" << nNotEOW << " nShutdown=" << nShutdown
	 << " nConsider=" << nConsider << " nins=" << nins
	 << " SLEEP" << endl;
#endif /* DCDBG_FILTERSERVICE_INSSELECT */

    //
    // TODO: maybe the best way to fix all this is to special case the
    // reasons for sleeping...
    // (1) worknum too early: sleep until next EOW
    // (2) other filter blocked in dequeue == empty queue: cvAllInsNotEmpty
    //
    // TODO: Track down why DEBUG: 1 wWorkNum=1 > wDequeueWorkNum=0
    //

    // sleep
    pFilter->pShared->cvAllInsNotEmpty.wait(pFilter->pShared->mutexAllIns);
  }
  pFilter->pShared->mutexAllIns.unlock();

  // advance starting point for next call
  ++pFilter->iStreamReadAnyState;
  if (pFilter->iStreamReadAnyState > nins) pFilter->iStreamReadAnyState = 0;

  if (pFilter->fInProcessCall) {
    if (pFilter->fReadFirst) {  // stats
      pFilter->tReadFirst.Stop();
      pFilter->fReadFirst = false;
    } else {
      pFilter->tReadAny.Stop();
    }
    pFilter->sendPerformanceUpdate();    
  }
  return pbuf;
}

/**
 * return: number of ins[] that already hit EOW
 */
int DC_Filter_Base_t::arg_t::insNumEOW(void) {
  int wRtn = 0;
  for (int i=0; i<nins; ++i) {
    if (ins[i].isEndOfWork()) ++wRtn;
  }
  return wRtn;
}
// return: true=if all ins[] are at EOW
bool DC_Filter_Base_t::arg_t::insAllEOW(void) {
  return (nins == insNumEOW());
}


/**
 * insure proper UOW state by draining all input streams
 */
// BUGBUG: we should be smarter and insure that if there are
// multiple copies running, we don't drain buffers away that
// another copy could read.  For now we ignore this.
int DC_Filter_Base_t::arg_t::DrainInput(int wWorkNum) {
  DC_Buffer_t *pbuf;
  int i;

  if (insAllEOW()) return 0;

  while (1) {
    //pbuf = insReadAny(&i);
#if defined(DCDBG_FILTERSERVICE_DRAIN)
    cout << "DrainInput(): " << sbFilterName
	 << " Calling insReadAny() ..." << endl;
#endif /* DCDBG_FILTERSERVICE_DRAIN */
    pbuf = insReadAny(&i);
    if (!pbuf) return 0;
#if defined(DCDBG_FILTERSERVICE_DRAIN)
    cout << "DrainInput(): " << sbFilterName
	 << " discarding unread buf " << *pbuf << endl;
#endif /* DCDBG_FILTERSERVICE_CLOSE */
    pbuf->consume();
  }
}

/**
 * close all outgoing streams, wait for all input streams to see EOS
 * (this is called automatically when process() exits)
 */
int DC_Filter_Base_t::arg_t::CloseOutputWaitInput(int wWorkNum,
						  bool fEndOfWork) {
  int i;
  DC_Buffer_t *pbuf;

  int nOutput=0;
  for (i=0; i<nouts; ++i) {
    if (!outs[i].isEndOfStream()) ++nOutput;
  }
#if defined(DCDBG_FILTERSERVICE_CLOSE)
  cout << "CloseOutputWaitInput(): " << sbFilterName
       << (fEndOfWork ? " EndOfWork to " : " Closing ")
       << nOutput << '/' << nouts << " output streams ..."
       << endl;
#endif /* DCDBG_FILTERSERVICE_CLOSE */
  // EOW/EOS all output streams
  for (i=0; i<nouts; ++i) {
    outs[i].close(fEndOfWork);
  }

  int nInput=0;
  for (i=0; i<nins; ++i) {
    if (!ins[i].isEndOfStream()) ++nInput;
  }
  
  if (fEndOfWork || nInput == 0) {
#if defined(DCDBG_FILTERSERVICE_CLOSE)
    cout << "CloseOutputWaitInput(): " << sbFilterName << " exit" << endl;
#endif /* DCDBG_FILTERSERVICE_CLOSE */
    return DC_ERR_OK;
  }

  // blocking wait for EOW/EOS and discard any incoming data for input
  // streams we are using the insReadAny() mechanism to avoid full
  // buffer deadlock
#if defined(DCDBG_FILTERSERVICE_CLOSE)
  cout << "CloseOutputWaitInput(): " << sbFilterName
       << " Waiting for " << nInput << '/' << nins << " input streams ..."
       << endl;
#endif /* DCDBG_FILTERSERVICE_CLOSE */

  // TODO: we want to have a consumer be able to fEndOfWork, but we
  // cannot just consume and discard all the buffers up to the next
  // EOW, because there could be other copies that can consume and
  // process this data that have not yet fEndOfWork.  An early exiter
  // needs to pause until the last exiter discards all buffers up to
  // the EOW.  The hard part is this needs to coexist with a copy that
  // *does* read all its data up to the EOW marker using read() or
  // insReadAny(), and also coexist with everything else localDequeue
  // does.

  // For EOW: all read until EOW.
  //
  // For exit: If we are a non-last copy that is exiting, do nothing.
  // If we are the last, shutdown Q.  Also, we need to remove ourselves
  // as consumer of ins, and as producer of colocated outs.

  while (1) {
#if defined(DCDBG_FILTERSERVICE_CLOSE)
    cout << "CloseOutputWaitInput(): " << sbFilterName
	 << " Calling insReadAny() ..." << endl;
#endif /* DCDBG_FILTERSERVICE_CLOSE */
    pbuf = insReadAny(&i);
    if (pbuf) {
#if defined(DCDBG_FILTERSERVICE_CLOSE)
      cout << "CloseOutputWaitInput(): " << sbFilterName
	   << " insReadAny() read " << *pbuf << endl;
#endif /* DCDBG_FILTERSERVICE_CLOSE */
      pbuf->consume();
    } else {
#if defined(DCDBG_FILTERSERVICE_CLOSE)
      cout << "CloseOutputWaitInput(): " << i << " " << sbFilterName
	   << " insReadAny() read NULL" << endl;
#endif /* DCDBG_FILTERSERVICE_CLOSE */
      // 02/18/2001 : mdb - Fixed bug where this case was missed, causing
      //                    errors if the EOW was not read by the filter
      //                    before exiting.  The reads() would fail here
      //                    because fEOW was true.
      // reset the EOW marker, so we can consume anything up to EOS
      if (insAllEOW()) {
	_resetInsEOW();
	++(pFilter->wWorkNum);
      } else {
	break;
      }
    }
  }

#if defined(DCDBG_FILTERSERVICE_CLOSE)
  cout << "CloseOutputWaitInput(): " << sbFilterName << " exit" << endl;
#endif /* DCDBG_FILTERSERVICE_CLOSE */
  return DC_ERR_OK;
}


// ****************************************************************************
