// ****************************************************************************
//  Name: thread-reader.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 04/08/2000 : mdb - Split into a seperate source files.
//        02/25/2001 : mdb - Large structural changes to support copies, and
//                           all the attendant problems.
//        07/27/2001 : mdb - Implemented stream write policy subsets.
//  Desc: Reader thread in each remote lib handles all incoming packets.
// ****************************************************************************

#include "lib/dc_barrier.h"

#include "dc_filterservice.h"
#include "packet.h"
#include "packet-tags.h"
#include "consoleprocessstate.h"
#include "remoteprocessstate.h"
#include "util.h"

#include "thread-reader.h"

#ifdef _WIN32
#include "win32.h"
#endif

#if defined(DCDBG)
//#define DCDBG_FILTERSERVICE_READER_THREAD
//#define DCDBG_FILTERSERVICE_CLOSE
//#define DCDBG_FILTERSERVICE_CONN
//#define DCDBG_FILTERSERVICE_HELLOACK
#endif /* DCDBG */

// We stopped using a pthread_cancel mechanism for shutting down the
// reader thread, because async-safeness of syscalls is not portable.
// Instead, we open a connection to the server loop and send a
// shutdown packet.


// this is used internally by the reader to transfer calls made by
// other threads to Reader::ReaderCmd_XXX() fcns to the reader thread
// by using READERCMD packets
class DC_ReaderCmd {
public:
  enum { ADDCONN, ENABLECONN, DISABLECONN, CLOSECONN, CLEANUPINSTANCE } eTag;
  void *p1, *p2;
  int   w;
};


//
//  reader thread server-loop
//


/**
 * ctor
 */
DC_ReaderServerLoop::DC_ReaderServerLoop(DC_FilterService_t *pDC_in)
  : pDC(pDC_in) {
}

void *DC_ReaderServerLoop::thread_run(void *) {
  // job: manage the set of incoming connections, receive data
  // from the network, and place in the appropriate queue.
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
  cout << "reader thread: entry ..." << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
  
  bool fContinue = true;
  bool fLastFail = false;
  while (fContinue) {
    switch (OnePass()) {
#ifdef _WIN32
	case SLE_ERROR_WIN:
#else
    case SLE_ERROR:
#endif

	if (!fLastFail) {
	cout << ErrorHeader(wERROR) << "failed OnePass() call." << endl;
      }
      fLastFail = true;
      // what else to do?
      break;
    case SLE_SHUTDOWN:
      // what else to do?
      fContinue = false;
      break;
    default:
      fLastFail = false;
      break;
    }
  }

#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
  cout << "reader thread: exit" << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
  return NULL;
}


// 09/15/1999 : mdb - Changed reader to have the read fcn allocate a new
//                    buffer.  Since we were doing a deep copy anyway, this
//                    doesn't really hurt performance.
int DC_ReaderServerLoop::ClientProcess(int i) {
  int wTag;
  Packet packetIn;
  Packet packetOut;
  ConfigInfo cfgTmp;
  int wExtra;
  DC_Buffer_t *pbuf;
  ReaderConnMap *pMap;
  int ret;
  int wRtn = SLE_CB_OK;

  // Now we need to read the data from the socket connection, but we
  // don't want the reader to block if the data is TAG_DATA,EOW,EOS
  // and the stream queue is full or if the UOW is not in sync with
  // other copies.  We special case the potentially blocking packets
  // from other control packets based on the pMap settings of the
  // connection.

  // BUGBUG: Just as for the "sleeping" connection description in
  // remoteprocessstate.cpp, we should deal better with the case where
  // we don't want to read from a socket, so we don't eat CPU cycles
  // because every select() call will succeed.  We should mark the
  // connection as "sleeping" if the queue is full, and don't check it
  // in the select() call after that.  When the queue becomes
  // non-full, we should be signalled somehow by a ReaderCmd pkt to
  // unmark the "sleeping" tag so it's checked in future select()
  // calls.  Maybe adding a hook to sleep on the cvNotFull inside the
  // queue will help somehow to do the second part.

  pMap = (ReaderConnMap*)getClientMapValue(i);

  // -------------------------------------
  //  packets on input stream connections
  // -------------------------------------
  if (pMap->pInStream) {
    bool fFull;

    // 04/19/2001 : mdb - Fixed stupid bug not having var name
    DC_Guard guard(pMap->pInStream->pShared->SinkQ.getInternal());

    // check if queue is full
    if (pMap->pInStream->pShared->SinkQ.probeIsFull(fFull, false)) {
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
      cout << "[reader] (skip read, close conn) -- shutdown stream "
	   << pMap->pInStream->pShared->sbName
#if defined(DCDBG_READERCONNNAME)
	    << ' ' << pMap->sbName
#endif /* DCDBG_READERCONNNAME */
	   << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
      // BUGBUG: This is not protected by a mutex
      pMap->pInStream->fEOS = true;  // set for next packet to drop
      return SLE_CB_CLOSECONN;

    } else if (fFull) { // queue is full
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
      cout << "[reader] (skip read, disable conn) -- full stream "
	   << pMap->pInStream->pShared->sbName
#if defined(DCDBG_READERCONNNAME)
	    << ' ' << pMap->sbName
#endif /* DCDBG_READERCONNNAME */
	   << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
      // we disable the connection while holding the lock, and we are
      // re-enabled by SinkQ.localDequeue() causing the queue to
      // become not-full.
      pMap->pInStream->pShared->SinkQ.remoteDisableFull(this, i);
      return SLE_CB_OK;

    } else if (pMap->fCheckWorkNum) {
      pMap->fCheckWorkNum = false;
      if (!pMap->pInStream->pShared->SinkQ.remoteCheckWorkNum(pMap->wWorkNum,
							      this, i)) {
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
	cout << "[reader] (skip read, disable conn) -- work="
	     << pMap->wWorkNum << " sync to stream "
	     << pMap->pInStream->pShared->sbName
#if defined(DCDBG_READERCONNNAME)
	     << ' ' << pMap->sbName
#endif /* DCDBG_READERCONNNAME */
	     << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
	// conn has been disabled and stored by the queue for future
	// enabling when other copies catch up.
	return SLE_CB_OK;
      }
    }
    
    // BUGBUG: We hold the queue mutex for a long time here with the
    // read() and could block the consumer from reading data that is
    // sitting in the queue.  A possible solution is to modify the
    // queue as follows.  The producer first called enqueueReserve()
    // to get a guaranteed reservation, and is given a handle.  Later
    // the consumer calls enqueueFinalize() with the handle and the
    // new data to store in the queue.  The consumer blocks if it
    // tries to dequeue a reserved entry.  All we should need is a
    // single cvReserved that blocks the dequeue until later.  This
    // allows for improved sharing of the queue w/o potentially long
    // locked periods.  We could even allow out-of-order things like
    // dequeue the first available non-reserved entry for the current
    // unit-or-work, etc.

    // queue is not full, read the packet.  packet buffer is allocated
    if ((ret = packetIn.read(*(rgClients[i].psock))) <= 0) {
      // TODO: propagate this as EOS for all input streams that use
      // this connection.  Also, need to call remoteRemoveProducer()
      // to adjust copies counts.  Recovery may not be easily
      // possible.
      if (!pMap->pInStream->fEOS) {  // unexpected socket close?
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
	if (ret == -1) {
	  cerr << "[reader] failed read, close conn " << i
#if defined(DCDBG_READERCONNNAME)
	       << ' ' << (pMap->sbName ? pMap->sbName : "")
#endif /* DCDBG_READERCONNNAME */
	       << " - errno=" << strerror(errno) << endl;
	} else {
	  cerr << "[reader] failed read, close conn " << i
#if defined(DCDBG_READERCONNNAME)
	       << ' ' << (pMap->sbName ? pMap->sbName : "")
#endif /* DCDBG_READERCONNNAME */
	       << endl;
	}
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
      }
      return SLE_CB_CLOSECONN;  // error reading client socket
    }
    wTag = packetIn.getTag();
    wExtra = packetIn.getExtra();

    char *sbTag;
    switch (wTag) {
      //
      //  Remote Process packets
      //
    case TAG_FILTERSERVICE_DATA:  // remote process socket
    case TAG_FILTERSERVICE_EOW:  // remote process socket
    case TAG_FILTERSERVICE_EOS:  // remote process socket

      pbuf = new DC_Buffer_t;  // the object to be enqueued
      switch(wTag) {
      case TAG_FILTERSERVICE_DATA:  // remote process socket
	sbTag = "DATA";
	break;
      case TAG_FILTERSERVICE_EOW:  // remote process socket
	sbTag = "EOW";
	pbuf->eType = DC_Buffer_t::EOW;
	break;
      case TAG_FILTERSERVICE_EOS:  // remote process socket
	sbTag = "EOS";
	pbuf->eType = DC_Buffer_t::EOS;
      }
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
      cout << "[reader] Recv " << sbTag << " pkt for stream "
	   << pMap->pInStream->pShared->sbName
#if defined(DCDBG_READERCONNNAME)
	   << ' ' << pMap->sbName
#endif /* DCDBG_READERCONNNAME */
	   << " (" << packetIn.getSizePayload() << " bytes)" << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */

      // check if stream closed
      if (pMap->pInStream->fEOS) {
	cout << "[reader] drop " << sbTag << " pkt for closed stream" << endl;
	pbuf->consume();
      } else {
	if (packetIn.getPayload()) {  // allow for zero length buffers
	  pbuf->StealCopy(*packetIn.getPayload());
	}
	if (pMap->pInStream->pShared->SinkQ.remoteEnqueue(pMap->wWorkNum, pbuf, 
							  pMap->eStreamWritePolicy,
							  i, rgClients[i].psock)) {
	  // shutdown -- forced EOS from this sink
	  pMap->pInStream->fEOS = true;  // set for next packet to drop
#if defined(DCDBG_FILTERSERVICE_CLOSE)
	  cout << "[reader] discard " << sbTag << " pkt - instream closed"
	       << endl;
#endif /* DCDBG_FILTERSERVICE_CLOSE */
	  pbuf->consume();
	}
	// the following is done regardless of enqueue success
	if (wTag == TAG_FILTERSERVICE_EOW) {
	  // Increment, and set check flag so the new worknum is
	  // checked when the next packet arrives.
	  ++pMap->wWorkNum;
	  pMap->fCheckWorkNum = true;
	} else if (wTag == TAG_FILTERSERVICE_EOS) {
#if defined(DCDBG_FILTERSERVICE_CLOSE)
	  cout << "[reader] setting pInStream->pShared->fEOS for stream "
	       << pMap->pInStream->pShared->sbName << endl;
#endif /* DCDBG_FILTERSERVICE_CLOSE */
	  // BUGBUG: for a collective Q, a single EOS should not stop
	  // the other producer EOS packets.  We need a different way
	  // to drop packets after *all* producer EOS pkts are received.
	  //pMap->pInStream->fEOS = true;  // mark to drop future packets
	}
      }
      break;
    default:
      cerr << ErrorHeader(wERROR)
	   << "ReaderServerLoop(): Unexpected STREAM packet with unknown tag #"
	   << wTag << "." << endl << packetIn << endl;
      sbTag = "UNKNOWN";
      // let it pass thru, and enqueue
    }

    // NOTE: guard unlock when exit this scope

  // ------------------------------------------------------------
  //  packets on other connections (output stream, console, etc)
  // ------------------------------------------------------------
  } else {
    // the buffer in the packet is allocated, since we never set it
    if (packetIn.read(*(rgClients[i].psock)) <= 0) {
      // TODO: go through all input streams for local filters and send an
      // EOS packet for each
#if defined(DCDBG_FILTERSERVICE_CLOSE)
      cerr << "[reader] closing connection " << i << endl;
#endif /* DCDBG_FILTERSERVICE_CLOSE */
      return SLE_CB_CLOSECONN;  // some error reading the client socket
    }
    wTag = packetIn.getTag();
    wExtra = packetIn.getExtra();
    switch (wTag) {
      //
      //  Remote Process packets
      //
    case TAG_FILTERSERVICE_DATA_ACK:  // remote process socket
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
      cout << "[reader] Recv DATA_ACK pkt" << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */

      // update ACK count for this connection.
      if (pMap->pOutStream) {
	pMap->pOutStream->lcUnAckedBuffers.decSignalHigh(pMap->iConnInfo);
      } else {
	// error - maybe filter exited, and the conn was reaped?
      }
      break;
    case TAG_FILTERSERVICE_STREAMCONNECT:  // remote process socket
      // this should be the first packet we receive on remote process socket
      cfgTmp.ReadBuf(packetIn.getPayloadPtr(),
		     packetIn.getSizePayload());
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
      cout << "[reader] Recv STREAMCONNECT pkt for instance " << wExtra << endl;
      cout << cfgTmp;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
#if 0
      if (pMap->pInStream) {
	pMap->pInStream->iClient = i;
      } else {
	// we received a STREAMCONNECT after the instance has exited?
      }
#endif

      {
	// parse cfg
	int wInstanceNum = atoi(FindParm(cfgTmp, "", "instance"));
	char *sbStreamName = FindParm(cfgTmp, "", "stream");
	int iConnInfo = atoi(FindParm(cfgTmp, "", "iconninfo"));
	DC_StreamWritePolicy_t ePolicy
	  = DC_StreamWritePolicy_toenum(FindParm(cfgTmp, "",
						 "streamwritepolicy"));

	// 03/31/2001 : mdb - Set here for incoming conn so it is
	// reaped when the instance is finished.
	pMap->wInstanceNum = wInstanceNum;

	// the following call fills in pMap->{pInStream,wInstanceNum}
	if (pDC->pRemote->StreamConnect(wInstanceNum, sbStreamName, iConnInfo,
					ePolicy,
					pMap, rgClients[i].psock, true) == 0) {
	  // send ack
	  ret = pDC->pRemote->StreamConnectAck(wInstanceNum, iConnInfo,
					       rgClients[i].psock);
	  if (ret != 0) wRtn = SLE_CB_CLOSECONN;  // error writing socket
	}

#if defined(DCDBG_READERCONNNAME)
	{
	  char bufTmp[80];
	  snprintf(bufTmp, 80, "<conn %d - sink i%d,%s>",
		   i, wInstanceNum, sbStreamName);
	  if (pMap->sbName) delete [] pMap->sbName;
	  pMap->sbName = StrDup(bufTmp);
	}
#endif /* DCDBG_READERCONNNAME */
      }
      break;
    case TAG_FILTERSERVICE_STREAMCONNECT_ACK:  // remote process socket
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
      cout << "[reader] Recv STREAMCONNECT_ACK pkt for instance "
	   << pMap->pOutStream->pShared->wInstanceNum
	   << ", iconninfo=" << wExtra << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
      // signal the waiting thread who originally sent the
      // STREAMCONNECT.

      // signal the filter write() fcn the stream is connected
      pMap->iConnInfo = wExtra;  // save for recv DATA_ACKs
      pMap->pOutStream->daConnInfo[wExtra]->semOpenAck.post();
      break;
    case TAG_FILTERSERVICE_STARTINSTANCE:  // console process socket
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
      cout << "[reader] Recv STARTINSTANCE pkt for instance " << wExtra << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
      // turn into a cfg spec
      cfgTmp.ReadBuf(packetIn.getPayloadPtr(),
		     packetIn.getSizePayload());
      // the following call fills in pMap->{pOutStream,wInstanceNum}
      ret = pDC->pRemote->StartInstance(wExtra, cfgTmp, rgClients[i].psock);
      packetOut.set(TAG_FILTERSERVICE_STARTINSTANCE_ACK, NULL);
      packetOut.setExtra(ret);
      if (packetOut.write(*(rgClients[i].psock)) <= 0) {
	wRtn = SLE_CB_CLOSECONN;  // some error writing the client socket
      }
      break;
    case TAG_FILTERSERVICE_STOPINSTANCE:  // console process socket
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
      cout << "[reader] Recv STOPINSTANCE pkt for instance " << wExtra << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
      ret = pDC->pRemote->StopInstance(wExtra);
      break;
    case TAG_FILTERSERVICE_APPENDWORK:  // console process socket
      // Always send an ACK to indicate success or failure, because
      // there is a race between a remote instance deciding to go away
      // on its own, and the arrival of new work.

      // turn MemoryBuf in packet into a DC_Buffer_t
      pbuf = new DC_Buffer_t;
      if (packetIn.getPayload()) {  // allow for zero length buffers
	pbuf->StealCopy(*packetIn.getPayload());
      }
      {
	int wInstanceNum, wWorkNum;
	DC_DecodeInt(wExtra, &wInstanceNum, &wWorkNum);
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
	cout << "[reader] Recv APPENDWORK pkt for instance " << wInstanceNum
	     << " work " << wWorkNum << " size=" << pbuf->getSize() << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
	// this following may disable a client connection
	ret = pDC->pRemote->AppendWork(wInstanceNum, wWorkNum, pbuf, i);
	if (ret != DC_ERR_OK) {
	  cout << "	" << DC_strerror(ret) << endl;
	}
      }
      pbuf->consume();
      packetOut.set(TAG_FILTERSERVICE_APPENDWORK_ACK, NULL);
      packetOut.setExtra(ret);
      if (packetOut.write(*(rgClients[i].psock)) <= 0) {
	wRtn = SLE_CB_CLOSECONN;  // some error writing the client socket
      }
      break;
    case TAG_FILTERSERVICE_APPENDWORK_CANCEL: // CLH - added for new filter return values
      {
	int wInstanceNum, wWorkNum;
	DC_DecodeInt(wExtra, &wInstanceNum, &wWorkNum);
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
	cout << "[reader] Recv APPENDWORK_CANCEL pkt for instance "
	     << wInstanceNum << " work " << wWorkNum << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
	pDC->pRemote->CancelWork(wInstanceNum, wWorkNum);
      }
      break;
    case TAG_FILTERSERVICE_HELLO_ACK:
      cfgTmp.ReadBuf(packetIn.getPayloadPtr(),
		     packetIn.getSizePayload());
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
      cout << "[reader] Recv HELLO_ACK pkt ("
	   << packetIn.getSizePayload() << " bytes)" << endl;
      cout << cfgTmp;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
#if defined(DCDBG_FILTERSERVICE_HELLOACK)
      cout << "[t" << (int)pthread_self() << "]: ";
      cout << "*** BEFORE MERGE" << endl << pDC->cfgConsole << "***" << endl;
#endif /* DCDBG_FILTERSERVICE_HELLOACK */
      pDC->cfgConsole.Merge(cfgTmp);
#if defined(DCDBG_FILTERSERVICE_HELLOACK)
      cout << "[t" << (int)pthread_self() << "]: ";
      cout << "*** AFTER MERGE" << endl << pDC->cfgConsole << "***" << endl;
#endif /* DCDBG_FILTERSERVICE_HELLOACK */
      pDC->pRemote->ProcessConsoleCfgSettings();
      break;
    case TAG_FILTERSERVICE_READERCMDQ: { // localhost socket
      // these are fake local packets sent by other local threads to send
      // commands to the local reader, so that the work is done in the reader's
      // context to avoid thread safety issues and extra synchronization.
      DC_ReaderCmd *pcmd;
      pDC->pReader->ReaderCmdQ.dequeue((void**)&pcmd);
      switch (pcmd->eTag) {
      case DC_ReaderCmd::ADDCONN:
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
	cout << "[reader] Recv READERCMD pkt ADDCONN cmd" << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
	AddExternalConn((BaseSocket *)pcmd->p1, (InetAddress *)pcmd->p2);
	break;
      case DC_ReaderCmd::ENABLECONN:
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
	cout << "[reader] Recv READERCMD pkt ENABLECONN cmd for iConn="
	     << pcmd->w << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
	EnableConn(pcmd->w);
	break;
      case DC_ReaderCmd::DISABLECONN:
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
	cout << "[reader] Recv READERCMD pkt DISABLECONN cmd for iConn="
	     << pcmd->w << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
	DisableConn(pcmd->w);
	break;
      case DC_ReaderCmd::CLOSECONN:
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
	cout << "[reader] Recv READERCMD pkt CLOSECONN cmd for iConn="
	     << pcmd->w << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
	CloseConn(pcmd->w);
	break;
      case DC_ReaderCmd::CLEANUPINSTANCE:
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
	cout << "[reader] Recv READERCMD pkt CLEANUPINSTANCE cmd for i"
	     << pcmd->w << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
	_CloseInstanceConn(pcmd->w);
	pDC->pRemote->CleanupInstance(pcmd->w);
	break;
      default:
	cerr << "DC_ReaderServerLoop(): unknown READERCMD tag="
	     << (int)pcmd->eTag << endl;
	break;
      }
      delete pcmd;
      break;
    }
    case TAG_FILTERSERVICE_READEREXIT:
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
      cout << "[reader] Recv READEREXIT pkt" << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
      wRtn = SLE_CB_SHUTDOWN;
      break;

      //
      //  Console Process packets
      //
    case TAG_FILTERSERVICE_HELLO:
      cfgTmp.ReadBuf(packetIn.getPayloadPtr(),
		     packetIn.getSizePayload());
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
      cout << "[reader] Recv HELLO pkt" << endl;
      //cout << cfgTmp;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */

      // process HELLO and send HELLO_ACK
      if (pDC->pConsole->StartRemoteProcesses_Reader(cfgTmp,
						     rgClients[i].psock)) {
	wRtn = SLE_CB_CLOSECONN;  // some error writing the client socket
      }
      break;
    case TAG_FILTERSERVICE_STARTINSTANCE_ACK:
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
      cout << "[reader] Recv STARTINSTANCE_ACK pkt" << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
      // TODO:
      break;
    case TAG_FILTERSERVICE_APPENDWORK_ACK:  // recv 1 per instance host
      if (wExtra == DC_ERR_OK) {
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
	cout << "[reader] Recv APPENDWORK_ACK pkt wExtra=" << wExtra << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
      } else {
	cout << "[reader] Recv APPENDWORK_ACK pkt wExtra=" << wExtra
	     << " ***ERROR***" << endl;
      }
      // TODO: collect these to make sure all are appended correctly.
      // (wExtra==0).  If not, maybe try and stop it.
      break;
    case TAG_FILTERSERVICE_APPENDWORK_DONE:  // recv 1 per instance filter
      {
	int wInstanceNum, wWorkNum, *pExitStatus;
	DC_DecodeInt(wExtra, &wInstanceNum, &wWorkNum);
	pExitStatus = (int*)packetIn.getPayloadPtr();
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
	cout << "[reader] Recv APPENDWORK_DONE pkt for instance "
	     << wInstanceNum << " work " << wWorkNum 
	     << " return " << *pExitStatus << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
	pDC->pConsole->AppendWorkDone_Reader(wInstanceNum, wWorkNum, *pExitStatus);
      }
      break;
    case TAG_FILTERSERVICE_FILTEREXIT:
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
      cout << "[reader] Recv FILTEREXIT pkt" << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
      cfgTmp.ReadBuf(packetIn.getPayloadPtr(),
		     packetIn.getSizePayload());
      //cout << cfgTmp;
      pDC->pConsole->FilterExit_Reader(wExtra, cfgTmp);
      break;
    case TAG_FILTERSERVICE_INSTANCEEXIT:
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
      cout << "[reader] Recv INSTANCEEXIT pkt" << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
      // this packet is counted as the last copy for a filter
      pDC->pConsole->InstanceExit_Reader(wExtra);
      break;
    case TAG_APPD_DONE:
      cfgTmp.ReadBuf(packetIn.getPayloadPtr(),
		     packetIn.getSizePayload());
#if defined(DCDBG_FILTERSERVICE_READER_THREAD)
      cout << "Recv DONE for "
	   << "pid=" << atoi(cfgTmp.Find(NULL, "pid"))
	   << ",done=" << cfgTmp.Find(NULL, "done")
	   << ",status=" << atoi(cfgTmp.Find(NULL, "status"))
	   << " from appd on " << *rgClients[i].paddr << endl;
#endif /* DCDBG_FILTERSERVICE_READER_THREAD */
      pDC->pConsole->StopRemoteProcesses_Reader();
      break;
    default:
      cerr << ErrorHeader(wERROR)
	   << "ReaderServerLoop(): Unexpected CONTROL packet with unknown tag #"
	   << wTag << "." << endl
	   << packetIn << endl;
    }
  }


  // free allocated packet buffer, if not stolen into a DC_Buffer and enqueued
  if (packetIn.getPayload()) {
    delete packetIn.getPayload();  // the MemoryBuf obj allocated by read()
    packetIn.setBuf(NULL);
  }
  return wRtn;
}

int DC_ReaderServerLoop::MapProcess(int i, enum eMapReason e) {
  ReaderConnMap *pMap;
  switch (e) {
  case ADD_CONN:
#if defined(DCDBG_FILTERSERVICE_CONN)
    cout << "ReaderServerLoop(): MapProcess(" << i << ",ADD_CONN)" << endl;
#endif /* DCDBG_FILTERSERVICE_CONN */
    pMap = new ReaderConnMap;
    pMap->wInstanceNum = -1;
    pMap->wWorkNum = 0; // TODO: this needs to be set to correspond to
                        // when the copy was created
    pMap->fCheckWorkNum = true;
    pMap->pInStream = NULL;
    pMap->pOutStream = NULL;
    pMap->iConnInfo = -1;
#if defined(DCDBG_READERCONNNAME)
    {
      char bufTmp[80];
      snprintf(bufTmp, 80, "<conn %d new>", i);
      pMap->sbName = StrDup(bufTmp);
    }
#endif /* DCDBG_READERCONNNAME */
    setClientMapValue(i, pMap);
    break;
  case DISABLE_CONN:
    break;
  case ENABLE_CONN:
    break;
  case CLOSE_CONN:
    pMap = (ReaderConnMap*)getClientMapValue(i);
#if defined(DCDBG_FILTERSERVICE_CONN)
#if defined(DCDBG_READERCONNNAME)
    cout << "ReaderServerLoop(): MapProcess(" << i << ",CLOSE_CONN) "
	 << pMap->sbName << endl;
#else /* DCDBG_READERCONNNAME */
    cout << "ReaderServerLoop(): MapProcess(" << i << ",CLOSE_CONN)" << endl;
#endif /* DCDBG_READERCONNNAME */
#endif /* DCDBG_FILTERSERVICE_CONN */
    setClientMapValue(i, NULL);
#if defined(DCDBG_READERCONNNAME)
    delete [] pMap->sbName;
#endif /* DCDBG_READERCONNNAME */
    pMap->pInStream = NULL;
    delete pMap;
    break;
  }
  return SLE_CB_OK;
}

// thread: remote process reader
int DC_ReaderServerLoop::_CloseInstanceConn(int wInstanceNum) {
  // go through all connections and close those for the given instance.
  int wRtn = 0;
  for (int i=0; i<maxClients && numClients; ++i) {
    if ((rgClients[i].psock &&
	 rgClients[i].psock->getFd() != -1)) {
      ReaderConnMap *pMap = (ReaderConnMap*)getClientMapValue(i);
      if (pMap && pMap->wInstanceNum == wInstanceNum) {
	++wRtn;
#if 0
	cout << "_CloseInstanceConn(): closing conn " << i << endl;
#endif
	CloseConn(i);
      }
    }
  }
  return wRtn;
}


// thread: console process main
int DC_ReaderServerLoop::ReaderCmd_AddConn(BaseSocket *psock,
					   InetAddress *paddr) {
  DC_ReaderCmd *pcmd = new DC_ReaderCmd;
  pcmd->eTag = DC_ReaderCmd::ADDCONN;
  pcmd->p1 = (void*)psock;
  pcmd->p2 = (void*)paddr;
  ReaderCmdQ.enqueue(pcmd);
  Packet pkt(TAG_FILTERSERVICE_READERCMDQ, NULL);
  pkt.write(pDC->sockReader);
  return 0;
}
// thread: remote process filter
int DC_ReaderServerLoop::ReaderCmd_EnableConn(int iClient) {
  DC_ReaderCmd *pcmd = new DC_ReaderCmd;
  pcmd->eTag = DC_ReaderCmd::ENABLECONN;
  pcmd->w = iClient;
  ReaderCmdQ.enqueue(pcmd);
  Packet pkt(TAG_FILTERSERVICE_READERCMDQ, NULL);
  pkt.write(pDC->sockReader);
  return 0;
}
// thread: ?
int DC_ReaderServerLoop::ReaderCmd_DisableConn(int iClient) {
  DC_ReaderCmd *pcmd = new DC_ReaderCmd;
  pcmd->eTag = DC_ReaderCmd::DISABLECONN;
  pcmd->w = iClient;
  ReaderCmdQ.enqueue(pcmd);
  Packet pkt(TAG_FILTERSERVICE_READERCMDQ, NULL);
  pkt.write(pDC->sockReader);
  return 0;
}
// thread: remote process filter
int DC_ReaderServerLoop::ReaderCmd_CloseConn(int iClient) {
  DC_ReaderCmd *pcmd = new DC_ReaderCmd;
  pcmd->eTag = DC_ReaderCmd::CLOSECONN;
  pcmd->w = iClient;
  ReaderCmdQ.enqueue(pcmd);
  Packet pkt(TAG_FILTERSERVICE_READERCMDQ, NULL);
  pkt.write(pDC->sockReader);
  return 0;
}
// thread: remote process filter
int DC_ReaderServerLoop::ReaderCmd_CleanupInstance(int wInstanceNum) {
  DC_ReaderCmd *pcmd = new DC_ReaderCmd;
  pcmd->eTag = DC_ReaderCmd::CLEANUPINSTANCE;
  pcmd->w = wInstanceNum;
  ReaderCmdQ.enqueue(pcmd);
  Packet pkt(TAG_FILTERSERVICE_READERCMDQ, NULL);
  pkt.write(pDC->sockReader);
  return 0;
}


// ****************************************************************************
