// ****************************************************************************
//  Name: workqueue.cpp
//  Athr: Michael D. Beynon (mdb)
//  Date: 03/03/2001 : mdb - Created.
//        04/17/2001 : mdb - Rewrote localDequeue() according to bugfixes in
//                           streambufferqueue.
//  Desc: Part of remote lib process stream abstraction.  Layered above simple
//        queue to support multiple copies of consumers to collectively read
//        work requests from a single queue.
// ****************************************************************************

#include <assert.h>

#include "lib/dc_memory.h"

#include "dc_filterservice.h"
#include "remoteprocessstate.h"
#include "workqueue.h"

#if defined(DCDBG)
//#define DCDBG_FILTERSERVICE_WORKQUEUE
#endif /* DCDBG */


/**
 * ctor
 */
DC_WorkQueue::DC_WorkQueue(void)
  : wEnqueueWorkNum(0),
    wDequeueWorkNum(0), nConsumerDequeue(0), nTotalConsumers(0),
    fFirstDequeue(false), pworkDequeue(NULL),
    pSL(NULL), plcFullWorkQs(NULL) {
}

/**
 * dtor
 */
DC_WorkQueue::~DC_WorkQueue(void) {
  STestDelete(pworkDequeue);
}

/**
 * >0 if the queue contains at least one item
 */
int DC_WorkQueue::probe(bool fLock) {
  int wRtn;
  if (fLock) pmutex->lock();
  if (isShutdown(false)) {
    wRtn = QUEUE_SHUTDOWN;
  } else {
    wRtn = DC_Queue::probe(false);
    if (nConsumerDequeue > 0) ++wRtn;
  }
  if (fLock) pmutex->unlock();
  return wRtn;
}


// not locked
int DC_WorkQueue::setInitialWorkNum(int wWorkNum) {
  wEnqueueWorkNum = wWorkNum;
  wDequeueWorkNum = wWorkNum;
  return 0;
}
// not locked
int DC_WorkQueue::setNumConsumers(int nConsumers) {
  nTotalConsumers = nConsumers;
#if defined(DCDBG_FILTERSERVICE_WORKQUEUE)
  cout << "WorkQueue::setNumConsumers(): setting nTotalConsumers="
       << nTotalConsumers << endl;
#endif /* DCDBG_FILTERSERVICE_WORKQUEUE */
  return 0;
}
// not locked
int DC_WorkQueue::setReaderServerLoop(DC_ReaderServerLoop *pSL_in) {
  pSL = pSL_in;
  return 0;
}
// not locked
int DC_WorkQueue::setCountFullWorkQs(DC_LockedCounter *plcFullWorkQs_in) {
  plcFullWorkQs = plcFullWorkQs_in;
  return 0;
}


// locked -- unlike StreamBufferQueue
// iConn = connection to disable if the queue is full
// thread: reader
int DC_WorkQueue::remoteEnqueue(DC_Work_t *pwork, int iConn) {
  lockInternal();
  if (pwork) {
#if 0
    assert(pwork->wWorkNum == wEnqueueWorkNum);
#else
    if (pwork->wWorkNum != wEnqueueWorkNum) {
      cerr << "WorkQ.remoteEnqueue(): ERROR: new work " << pwork->wWorkNum
	   << " != enqueue work " << wEnqueueWorkNum << endl;
    }
#endif
  }

  // add to the queue
  int wRtn;
#if defined(DCDBG_FILTERSERVICE_WORKQUEUE)
  if (pwork) {
    cout << "[reader] WorkQ.remoteEnqueue(): enqueue work="
	 << pwork->wWorkNum << endl;
  } else {
    cout << "[reader] WorkQ.remoteEnqueue(): enqueue work=NULL" << endl;
  }
#endif /* DCDBG_FILTERSERVICE_WORKQUEUE */
  // ASSUME: next set of work to expect must be incremented by exactly one
  ++wEnqueueWorkNum;
  wRtn = enqueue(pwork, false);

  // check if now full for disabling connection
  if (pSL && iConn != -1) {
    bool fFull;
    if (probeIsFull(fFull, false) == QUEUE_OK && fFull) {
      // bump up full count
#if defined(DCDBG_FILTERSERVICE_WORKQUEUE)
      int w = plcFullWorkQs->inc(0);
      cout << "[reader] WorkQ.remoteEnqueue(): FULL, add to count"
	   << w << "$$$$$$$$$$" << endl;
#else /* DCDBG_FILTERSERVICE_WORKQUEUE */
      plcFullWorkQs->inc(0);      
#endif /* DCDBG_FILTERSERVICE_WORKQUEUE */
    }
  }
  unlockInternal();
  return wRtn;
}


int DC_WorkQueue::localAddConsumer(void) {
  lockInternal();
  ++nTotalConsumers;
#if defined(DCDBG_FILTERSERVICE_WORKQUEUE)
  cout << "WorkQueue::localAddConsumer(): nTotalConsumers="
       << nTotalConsumers << endl;
#endif /* DCDBG_FILTERSERVICE_WORKQUEUE */
  unlockInternal();
  return 0;
}

// in: fLock - needed by localShutdown() calls
// thread: remote process filter
int DC_WorkQueue::localDequeue(int wWorkNum, DC_Work_t **ppwork,
			       bool fLock) {
  if (fLock) lockInternal();

  if (isShutdown(false)) {
    if (fLock) unlockInternal();
    return QUEUE_SHUTDOWN;
  }

#if defined(DCDBG_FILTERSERVICE_WORKQUEUE)
  char sbFilterName[80];
  snprintf(sbFilterName, 80, "[t%d,filter,w%d]: WorkQ.localDequeue(): ",
	   (int)pthread_self(), wWorkNum);
#endif /* DCDBG_FILTERSERVICE_WORKQUEUE */

#if 0
  assert(wWorkNum >= wDequeueWorkNum);
#else
  if (wWorkNum < wDequeueWorkNum) {
    cerr << "WorkQ.remoteDequeue(): ERROR: looking for work " << wWorkNum
	 << " < dequeue work " << wDequeueWorkNum << endl;
  }
#endif

  // possibly block consumer: early worknum || non-first consumer
  while (wWorkNum > wDequeueWorkNum) {
#if defined(DCDBG_FILTERSERVICE_WORKQUEUE)
    cout << sbFilterName << "block on future work=" << wWorkNum
	 << " (dequeue work=" << wDequeueWorkNum << ")" << endl;
#endif /* DCDBG_FILTERSERVICE_WORKQUEUE */
    cvLocalConsumerReady.wait(getInternal());  // sleep if too early
    if (isShutdown(false)) {
      if (fLock) unlockInternal();
      return QUEUE_SHUTDOWN;
    }
  }

  // since a dequeue of a EOW/EOS buffer must be replicated to all
  // copies, we cannot have >1 consumer block in dequeue() on an empty
  // queue in case the next buffer read in EOW/EOS.  Here, we simply
  // block other consumers from the actual dequeue() call while
  // another is blocked in dequeue().  We do this before the collective
  // EOW/EOS checks because the other consumer may be reading EOW/EOS.
  while (fFirstDequeue) {
#if defined(DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE)
    cout << sbFilterName << "block because another filter is dequeuing"
	 << endl;
#endif /* DCDBG_FILTERSERVICE_STREAMBUFFERQUEUE */
    cvFirstDequeue.wait(getInternal());  /*** RELEASE MUTEX ***/
    if (isShutdown(false)) {
      if (fLock) unlockInternal();
      return QUEUE_SHUTDOWN;
    }
  }

  int wRtn = QUEUE_OK;
  if (nConsumerDequeue == 0) { // we are the first filter copy
#if defined(DCDBG_FILTERSERVICE_WORKQUEUE)
    cout << sbFilterName << "before Q.dequeue()" << endl;
    //_DumpState(cout);
#endif /* DCDBG_FILTERSERVICE_WORKQUEUE */
    fFirstDequeue = true;  // set here in case we block/lose mutex in dequeue
    wRtn = dequeue((void**)&pworkDequeue, -1, false);
    fFirstDequeue = false;  // release other consumers

    if (nTotalConsumers > 1) cvFirstDequeue.broadcast(); // wake others
    // this must be a broadcast: if signal woke 1 of k cvFirstDequeue
    // sleepers, and that one returns above with a fake EOW/EOS, the
    // other k-1 are left sleeping forever.  The alternative soln is
    // to signal on all exit paths above (collective EOW/EOS).

    if (wRtn == QUEUE_SHUTDOWN) {
      if (fLock) unlockInternal();
      return QUEUE_SHUTDOWN;
    }
    if (pworkDequeue) assert(wDequeueWorkNum == pworkDequeue->wWorkNum);

    // unblock console sending work?
    if (probeSpace(false) == 1) {  // this WorkQ just became non-full?
      plcFullWorkQs->lockInternal();
      // last full WorkQ just became non-full, and conn was prev disabled
      if (plcFullWorkQs->dec_u(0) == 0 && plcFullWorkQs->get_u(2) == 1) {
	int iConn = plcFullWorkQs->get_u(1);
	// send ReaderCmd so this is done in the context of the reader thread
#if defined(DCDBG_FILTERSERVICE_WORKQUEUE)
	cout << "[reader] WorkQ.localDequeue(): NOTFULL, re-enable conn="
	     << iConn << "$$$$$$$$$$$" << endl;
#endif /* DCDBG_FILTERSERVICE_WORKQUEUE */
	pSL->ReaderCmd_EnableConn(iConn);	
	plcFullWorkQs->set_u(2, 0);  // enabled now
      }
      plcFullWorkQs->unlockInternal();
    }
  }

  if (++nConsumerDequeue == nTotalConsumers) {  // we are the last copy
    nConsumerDequeue = 0;
    *ppwork = pworkDequeue;
    pworkDequeue = NULL;
    // ASSUME: next set of work to expect must be incremented by exactly one
    ++wDequeueWorkNum;
    if (nTotalConsumers > 0)
      cvLocalConsumerReady.broadcast();  // wake up any sleepers
  } else {
    // make a copy of the work to give to the filter
    if (pworkDequeue) {
      *ppwork = new DC_Work_t(0);
      (*ppwork)->wWorkNum = pworkDequeue->wWorkNum;
      (*ppwork)->buf.DeepCopy(pworkDequeue->buf);
    } else {
      *ppwork = NULL;
    }
  }

  if (fLock) unlockInternal();
  return wRtn;
}

int DC_WorkQueue::localShutdown(void) {
  // drain and shutdown SinkQ
  DC_Work_t *pwork;
  lockInternal();
  while (DC_Queue::probe(false) > 0) {
    if (DC_Queue::dequeue((void**)&pwork, -1, false) == QUEUE_OK && pwork) {
#if defined(DCDBG_FILTERSERVICE_QUEUE)
      cout << "DC_WorkQueue::localShutdown(): drain queue work="
	   << pwork->wWorkNum << ", " << pwork->buf << endl;
#endif /* DCDBG_FILTERSERVICE_QUEUE */
      delete pwork;
    }
  }
  if (nConsumerDequeue > 0 && pworkDequeue) delete pworkDequeue;
  shutdown(false);
  // wake any waiters
  if (nTotalConsumers) cvLocalConsumerReady.broadcast();
  unlockInternal();
  return 0;
}


// ****************************************************************************
